from .codec import ABIEncoder as ABIEncoder
from .registry import registry_packed as registry_packed
from _typeshed import Incomplete
from typing import Final

default_encoder_packed: Final[Incomplete]
encode_packed: Final[Incomplete]
is_encodable_packed: Final[Incomplete]
