from _typeshed import Incomplete
from faster_eth_abi.codec import ABICodec as ABICodec
from faster_eth_abi.registry import registry as registry
from typing import Final

default_codec: Final[Incomplete]
encode: Final[Incomplete]
decode: Final[Incomplete]
is_encodable: Final[Incomplete]
is_encodable_type: Final[Incomplete]
