def zpad(value: bytes, length: int) -> bytes: ...
def zpad32(value: bytes) -> bytes: ...
def zpad_right(value: bytes, length: int) -> bytes: ...
def zpad32_right(value: bytes) -> bytes: ...
def fpad(value: bytes, length: int) -> bytes: ...
def fpad32(value: bytes) -> bytes: ...
