from _typeshed import Incomplete
from faster_eth_abi.abi import (
    decode as decode,
    encode as encode,
    is_encodable as is_encodable,
    is_encodable_type as is_encodable_type,
)

__version__: Incomplete
