# AUTOGENERATED: Run scripts/generate_overload_tests.py to update.
# Requires mypy extras
# To run: pip install mypy && mypy tests/typecheck

# alias some things so we can minimize test data size in source control
from typing import Any as A
from typing import Tuple as T
from typing import Union as U
from typing_extensions import assert_type as ck
from eth_typing import HexAddress as HA
from faster_eth_abi import abi

# more aliases to save a bunch more chars
bo = bool
by = bytes
i = int
s = str

TA = T[A, ...]

THA = T[HA, ...]
Tbo = T[bo, ...]
Tby = T[by, ...]
Ti = T[i, ...]
Ts = T[s, ...]
TUa_bo = T[U[HA, bo], ...]
TUa_by = T[U[HA, by], ...]
TUa_i = T[U[HA, i], ...]
TUa_s = T[U[HA, s], ...]
TUbo_by = T[U[bo, by], ...]
TUbo_i = T[U[bo, i], ...]
TUbo_s = T[U[bo, s], ...]
TUby_i = T[U[by, i], ...]
TUby_s = T[U[by, s], ...]
TUi_s = T[U[i, s], ...]
TUa_bo_by = T[U[bo, by, HA], ...]
TUbo_by_i = T[U[bo, by, i], ...]
TUbo_by_s = T[U[bo, by, s], ...]
TUa_bo_i = T[U[bo, i, HA], ...]
TUa_by_i = T[U[by, i, HA], ...]
TUby_i_s = T[U[by, i, s], ...]

# the callable we are checking
f = abi.decode

# input data
D = b"\x00" * 32
ck(f(('address[]', 'uint232', 'bytes5'), D), T[A, i, by])  # fixed case 30920
ck(f(('address[]', 'uint232', 'bytes6'), D), T[A, i, by])  # fixed case 30921
ck(f(('address[]', 'uint232', 'bytes7'), D), T[A, i, by])  # fixed case 30922
ck(f(('address[]', 'uint232', 'bytes8'), D), T[A, i, by])  # fixed case 30923
ck(f(('address[]', 'uint232', 'bytes9'), D), T[A, i, by])  # fixed case 30924
ck(f(('address[]', 'uint232', 'int104'), D), T[A, i, i])  # fixed case 30925
ck(f(('address[]', 'uint232', 'int112'), D), T[A, i, i])  # fixed case 30926
ck(f(('address[]', 'uint232', 'int120'), D), T[A, i, i])  # fixed case 30927
ck(f(('address[]', 'uint232', 'int128'), D), T[A, i, i])  # fixed case 30928
ck(f(('address[]', 'uint232', 'int136'), D), T[A, i, i])  # fixed case 30929
ck(f(('address[]', 'uint232', 'int144'), D), T[A, i, i])  # fixed case 30930
ck(f(('address[]', 'uint232', 'int152'), D), T[A, i, i])  # fixed case 30931
ck(f(('address[]', 'uint232', 'int16'), D), T[A, i, i])  # fixed case 30932
ck(f(('address[]', 'uint232', 'int160'), D), T[A, i, i])  # fixed case 30933
ck(f(('address[]', 'uint232', 'int168'), D), T[A, i, i])  # fixed case 30934
ck(f(('address[]', 'uint232', 'int176'), D), T[A, i, i])  # fixed case 30935
ck(f(('address[]', 'uint232', 'int184'), D), T[A, i, i])  # fixed case 30936
ck(f(('address[]', 'uint232', 'int192'), D), T[A, i, i])  # fixed case 30937
ck(f(('address[]', 'uint232', 'int200'), D), T[A, i, i])  # fixed case 30938
ck(f(('address[]', 'uint232', 'int208'), D), T[A, i, i])  # fixed case 30939
ck(f(('address[]', 'uint232', 'int216'), D), T[A, i, i])  # fixed case 30940
ck(f(('address[]', 'uint232', 'int224'), D), T[A, i, i])  # fixed case 30941
ck(f(('address[]', 'uint232', 'int232'), D), T[A, i, i])  # fixed case 30942
ck(f(('address[]', 'uint232', 'int24'), D), T[A, i, i])  # fixed case 30943
ck(f(('address[]', 'uint232', 'int240'), D), T[A, i, i])  # fixed case 30944
ck(f(('address[]', 'uint232', 'int248'), D), T[A, i, i])  # fixed case 30945
ck(f(('address[]', 'uint232', 'int256'), D), T[A, i, i])  # fixed case 30946
ck(f(('address[]', 'uint232', 'int32'), D), T[A, i, i])  # fixed case 30947
ck(f(('address[]', 'uint232', 'int40'), D), T[A, i, i])  # fixed case 30948
ck(f(('address[]', 'uint232', 'int48'), D), T[A, i, i])  # fixed case 30949
ck(f(('address[]', 'uint232', 'int56'), D), T[A, i, i])  # fixed case 30950
ck(f(('address[]', 'uint232', 'int64'), D), T[A, i, i])  # fixed case 30951
ck(f(('address[]', 'uint232', 'int72'), D), T[A, i, i])  # fixed case 30952
ck(f(('address[]', 'uint232', 'int8'), D), T[A, i, i])  # fixed case 30953
ck(f(('address[]', 'uint232', 'int80'), D), T[A, i, i])  # fixed case 30954
ck(f(('address[]', 'uint232', 'int88'), D), T[A, i, i])  # fixed case 30955
ck(f(('address[]', 'uint232', 'int96'), D), T[A, i, i])  # fixed case 30956
ck(f(('address[]', 'uint232', 'string'), D), T[A, i, s])  # fixed case 30957
ck(f(('address[]', 'uint232', 'string[]'), D), T[A, i, A])  # fixed case 30958
ck(f(('address[]', 'uint232', 'uint104'), D), T[A, i, i])  # fixed case 30959
ck(f(('address[]', 'uint232', 'uint112'), D), T[A, i, i])  # fixed case 30960
ck(f(('address[]', 'uint232', 'uint120'), D), T[A, i, i])  # fixed case 30961
ck(f(('address[]', 'uint232', 'uint128'), D), T[A, i, i])  # fixed case 30962
ck(f(('address[]', 'uint232', 'uint136'), D), T[A, i, i])  # fixed case 30963
ck(f(('address[]', 'uint232', 'uint144'), D), T[A, i, i])  # fixed case 30964
ck(f(('address[]', 'uint232', 'uint152'), D), T[A, i, i])  # fixed case 30965
ck(f(('address[]', 'uint232', 'uint16'), D), T[A, i, i])  # fixed case 30966
ck(f(('address[]', 'uint232', 'uint160'), D), T[A, i, i])  # fixed case 30967
ck(f(('address[]', 'uint232', 'uint168'), D), T[A, i, i])  # fixed case 30968
ck(f(('address[]', 'uint232', 'uint176'), D), T[A, i, i])  # fixed case 30969
ck(f(('address[]', 'uint232', 'uint184'), D), T[A, i, i])  # fixed case 30970
ck(f(('address[]', 'uint232', 'uint192'), D), T[A, i, i])  # fixed case 30971
ck(f(('address[]', 'uint232', 'uint200'), D), T[A, i, i])  # fixed case 30972
ck(f(('address[]', 'uint232', 'uint208'), D), T[A, i, i])  # fixed case 30973
ck(f(('address[]', 'uint232', 'uint216'), D), T[A, i, i])  # fixed case 30974
ck(f(('address[]', 'uint232', 'uint224'), D), T[A, i, i])  # fixed case 30975
ck(f(('address[]', 'uint232', 'uint232'), D), T[A, i, i])  # fixed case 30976
ck(f(('address[]', 'uint232', 'uint24'), D), T[A, i, i])  # fixed case 30977
ck(f(('address[]', 'uint232', 'uint240'), D), T[A, i, i])  # fixed case 30978
ck(f(('address[]', 'uint232', 'uint248'), D), T[A, i, i])  # fixed case 30979
ck(f(('address[]', 'uint232', 'uint256'), D), T[A, i, i])  # fixed case 30980
ck(f(('address[]', 'uint232', 'uint32'), D), T[A, i, i])  # fixed case 30981
ck(f(('address[]', 'uint232', 'uint40'), D), T[A, i, i])  # fixed case 30982
ck(f(('address[]', 'uint232', 'uint48'), D), T[A, i, i])  # fixed case 30983
ck(f(('address[]', 'uint232', 'uint56'), D), T[A, i, i])  # fixed case 30984
ck(f(('address[]', 'uint232', 'uint64'), D), T[A, i, i])  # fixed case 30985
ck(f(('address[]', 'uint232', 'uint72'), D), T[A, i, i])  # fixed case 30986
ck(f(('address[]', 'uint232', 'uint8'), D), T[A, i, i])  # fixed case 30987
ck(f(('address[]', 'uint232', 'uint80'), D), T[A, i, i])  # fixed case 30988
ck(f(('address[]', 'uint232', 'uint88'), D), T[A, i, i])  # fixed case 30989
ck(f(('address[]', 'uint232', 'uint96'), D), T[A, i, i])  # fixed case 30990
ck(f(('address[]', 'uint232', '?'), D), T[A, i, A])  # fixed case 30991
ck(f(('address[]', 'uint24', 'address'), D), T[A, i, HA])  # fixed case 30992
ck(f(('address[]', 'uint24', 'address[]'), D), T[A, i, A])  # fixed case 30993
ck(f(('address[]', 'uint24', 'bool'), D), T[A, i, bo])  # fixed case 30994
ck(f(('address[]', 'uint24', 'bool[]'), D), T[A, i, A])  # fixed case 30995
ck(f(('address[]', 'uint24', 'bytes'), D), T[A, i, by])  # fixed case 30996
ck(f(('address[]', 'uint24', 'bytes1'), D), T[A, i, by])  # fixed case 30997
ck(f(('address[]', 'uint24', 'bytes10'), D), T[A, i, by])  # fixed case 30998
ck(f(('address[]', 'uint24', 'bytes11'), D), T[A, i, by])  # fixed case 30999
ck(f(('address[]', 'uint24', 'bytes12'), D), T[A, i, by])  # fixed case 31000
ck(f(('address[]', 'uint24', 'bytes13'), D), T[A, i, by])  # fixed case 31001
ck(f(('address[]', 'uint24', 'bytes14'), D), T[A, i, by])  # fixed case 31002
ck(f(('address[]', 'uint24', 'bytes15'), D), T[A, i, by])  # fixed case 31003
ck(f(('address[]', 'uint24', 'bytes16'), D), T[A, i, by])  # fixed case 31004
ck(f(('address[]', 'uint24', 'bytes17'), D), T[A, i, by])  # fixed case 31005
ck(f(('address[]', 'uint24', 'bytes18'), D), T[A, i, by])  # fixed case 31006
ck(f(('address[]', 'uint24', 'bytes19'), D), T[A, i, by])  # fixed case 31007
ck(f(('address[]', 'uint24', 'bytes2'), D), T[A, i, by])  # fixed case 31008
ck(f(('address[]', 'uint24', 'bytes20'), D), T[A, i, by])  # fixed case 31009
ck(f(('address[]', 'uint24', 'bytes21'), D), T[A, i, by])  # fixed case 31010
ck(f(('address[]', 'uint24', 'bytes22'), D), T[A, i, by])  # fixed case 31011
ck(f(('address[]', 'uint24', 'bytes23'), D), T[A, i, by])  # fixed case 31012
ck(f(('address[]', 'uint24', 'bytes24'), D), T[A, i, by])  # fixed case 31013
ck(f(('address[]', 'uint24', 'bytes25'), D), T[A, i, by])  # fixed case 31014
ck(f(('address[]', 'uint24', 'bytes26'), D), T[A, i, by])  # fixed case 31015
ck(f(('address[]', 'uint24', 'bytes27'), D), T[A, i, by])  # fixed case 31016
ck(f(('address[]', 'uint24', 'bytes28'), D), T[A, i, by])  # fixed case 31017
ck(f(('address[]', 'uint24', 'bytes29'), D), T[A, i, by])  # fixed case 31018
ck(f(('address[]', 'uint24', 'bytes3'), D), T[A, i, by])  # fixed case 31019
ck(f(('address[]', 'uint24', 'bytes30'), D), T[A, i, by])  # fixed case 31020
ck(f(('address[]', 'uint24', 'bytes31'), D), T[A, i, by])  # fixed case 31021
ck(f(('address[]', 'uint24', 'bytes32'), D), T[A, i, by])  # fixed case 31022
ck(f(('address[]', 'uint24', 'bytes4'), D), T[A, i, by])  # fixed case 31023
ck(f(('address[]', 'uint24', 'bytes5'), D), T[A, i, by])  # fixed case 31024
ck(f(('address[]', 'uint24', 'bytes6'), D), T[A, i, by])  # fixed case 31025
ck(f(('address[]', 'uint24', 'bytes7'), D), T[A, i, by])  # fixed case 31026
ck(f(('address[]', 'uint24', 'bytes8'), D), T[A, i, by])  # fixed case 31027
ck(f(('address[]', 'uint24', 'bytes9'), D), T[A, i, by])  # fixed case 31028
ck(f(('address[]', 'uint24', 'int104'), D), T[A, i, i])  # fixed case 31029
ck(f(('address[]', 'uint24', 'int112'), D), T[A, i, i])  # fixed case 31030
ck(f(('address[]', 'uint24', 'int120'), D), T[A, i, i])  # fixed case 31031
ck(f(('address[]', 'uint24', 'int128'), D), T[A, i, i])  # fixed case 31032
ck(f(('address[]', 'uint24', 'int136'), D), T[A, i, i])  # fixed case 31033
ck(f(('address[]', 'uint24', 'int144'), D), T[A, i, i])  # fixed case 31034
ck(f(('address[]', 'uint24', 'int152'), D), T[A, i, i])  # fixed case 31035
ck(f(('address[]', 'uint24', 'int16'), D), T[A, i, i])  # fixed case 31036
ck(f(('address[]', 'uint24', 'int160'), D), T[A, i, i])  # fixed case 31037
ck(f(('address[]', 'uint24', 'int168'), D), T[A, i, i])  # fixed case 31038
ck(f(('address[]', 'uint24', 'int176'), D), T[A, i, i])  # fixed case 31039
ck(f(('address[]', 'uint24', 'int184'), D), T[A, i, i])  # fixed case 31040
ck(f(('address[]', 'uint24', 'int192'), D), T[A, i, i])  # fixed case 31041
ck(f(('address[]', 'uint24', 'int200'), D), T[A, i, i])  # fixed case 31042
ck(f(('address[]', 'uint24', 'int208'), D), T[A, i, i])  # fixed case 31043
ck(f(('address[]', 'uint24', 'int216'), D), T[A, i, i])  # fixed case 31044
ck(f(('address[]', 'uint24', 'int224'), D), T[A, i, i])  # fixed case 31045
ck(f(('address[]', 'uint24', 'int232'), D), T[A, i, i])  # fixed case 31046
ck(f(('address[]', 'uint24', 'int24'), D), T[A, i, i])  # fixed case 31047
ck(f(('address[]', 'uint24', 'int240'), D), T[A, i, i])  # fixed case 31048
ck(f(('address[]', 'uint24', 'int248'), D), T[A, i, i])  # fixed case 31049
ck(f(('address[]', 'uint24', 'int256'), D), T[A, i, i])  # fixed case 31050
ck(f(('address[]', 'uint24', 'int32'), D), T[A, i, i])  # fixed case 31051
ck(f(('address[]', 'uint24', 'int40'), D), T[A, i, i])  # fixed case 31052
ck(f(('address[]', 'uint24', 'int48'), D), T[A, i, i])  # fixed case 31053
ck(f(('address[]', 'uint24', 'int56'), D), T[A, i, i])  # fixed case 31054
ck(f(('address[]', 'uint24', 'int64'), D), T[A, i, i])  # fixed case 31055
ck(f(('address[]', 'uint24', 'int72'), D), T[A, i, i])  # fixed case 31056
ck(f(('address[]', 'uint24', 'int8'), D), T[A, i, i])  # fixed case 31057
ck(f(('address[]', 'uint24', 'int80'), D), T[A, i, i])  # fixed case 31058
ck(f(('address[]', 'uint24', 'int88'), D), T[A, i, i])  # fixed case 31059
ck(f(('address[]', 'uint24', 'int96'), D), T[A, i, i])  # fixed case 31060
ck(f(('address[]', 'uint24', 'string'), D), T[A, i, s])  # fixed case 31061
ck(f(('address[]', 'uint24', 'string[]'), D), T[A, i, A])  # fixed case 31062
ck(f(('address[]', 'uint24', 'uint104'), D), T[A, i, i])  # fixed case 31063
ck(f(('address[]', 'uint24', 'uint112'), D), T[A, i, i])  # fixed case 31064
ck(f(('address[]', 'uint24', 'uint120'), D), T[A, i, i])  # fixed case 31065
ck(f(('address[]', 'uint24', 'uint128'), D), T[A, i, i])  # fixed case 31066
ck(f(('address[]', 'uint24', 'uint136'), D), T[A, i, i])  # fixed case 31067
ck(f(('address[]', 'uint24', 'uint144'), D), T[A, i, i])  # fixed case 31068
ck(f(('address[]', 'uint24', 'uint152'), D), T[A, i, i])  # fixed case 31069
ck(f(('address[]', 'uint24', 'uint16'), D), T[A, i, i])  # fixed case 31070
ck(f(('address[]', 'uint24', 'uint160'), D), T[A, i, i])  # fixed case 31071
ck(f(('address[]', 'uint24', 'uint168'), D), T[A, i, i])  # fixed case 31072
ck(f(('address[]', 'uint24', 'uint176'), D), T[A, i, i])  # fixed case 31073
ck(f(('address[]', 'uint24', 'uint184'), D), T[A, i, i])  # fixed case 31074
ck(f(('address[]', 'uint24', 'uint192'), D), T[A, i, i])  # fixed case 31075
ck(f(('address[]', 'uint24', 'uint200'), D), T[A, i, i])  # fixed case 31076
ck(f(('address[]', 'uint24', 'uint208'), D), T[A, i, i])  # fixed case 31077
ck(f(('address[]', 'uint24', 'uint216'), D), T[A, i, i])  # fixed case 31078
ck(f(('address[]', 'uint24', 'uint224'), D), T[A, i, i])  # fixed case 31079
ck(f(('address[]', 'uint24', 'uint232'), D), T[A, i, i])  # fixed case 31080
ck(f(('address[]', 'uint24', 'uint24'), D), T[A, i, i])  # fixed case 31081
ck(f(('address[]', 'uint24', 'uint240'), D), T[A, i, i])  # fixed case 31082
ck(f(('address[]', 'uint24', 'uint248'), D), T[A, i, i])  # fixed case 31083
ck(f(('address[]', 'uint24', 'uint256'), D), T[A, i, i])  # fixed case 31084
ck(f(('address[]', 'uint24', 'uint32'), D), T[A, i, i])  # fixed case 31085
ck(f(('address[]', 'uint24', 'uint40'), D), T[A, i, i])  # fixed case 31086
ck(f(('address[]', 'uint24', 'uint48'), D), T[A, i, i])  # fixed case 31087
ck(f(('address[]', 'uint24', 'uint56'), D), T[A, i, i])  # fixed case 31088
ck(f(('address[]', 'uint24', 'uint64'), D), T[A, i, i])  # fixed case 31089
ck(f(('address[]', 'uint24', 'uint72'), D), T[A, i, i])  # fixed case 31090
ck(f(('address[]', 'uint24', 'uint8'), D), T[A, i, i])  # fixed case 31091
ck(f(('address[]', 'uint24', 'uint80'), D), T[A, i, i])  # fixed case 31092
ck(f(('address[]', 'uint24', 'uint88'), D), T[A, i, i])  # fixed case 31093
ck(f(('address[]', 'uint24', 'uint96'), D), T[A, i, i])  # fixed case 31094
ck(f(('address[]', 'uint24', '?'), D), T[A, i, A])  # fixed case 31095
ck(f(('address[]', 'uint240', 'address'), D), T[A, i, HA])  # fixed case 31096
ck(f(('address[]', 'uint240', 'address[]'), D), T[A, i, A])  # fixed case 31097
ck(f(('address[]', 'uint240', 'bool'), D), T[A, i, bo])  # fixed case 31098
ck(f(('address[]', 'uint240', 'bool[]'), D), T[A, i, A])  # fixed case 31099
ck(f(('address[]', 'uint240', 'bytes'), D), T[A, i, by])  # fixed case 31100
ck(f(('address[]', 'uint240', 'bytes1'), D), T[A, i, by])  # fixed case 31101
ck(f(('address[]', 'uint240', 'bytes10'), D), T[A, i, by])  # fixed case 31102
ck(f(('address[]', 'uint240', 'bytes11'), D), T[A, i, by])  # fixed case 31103
ck(f(('address[]', 'uint240', 'bytes12'), D), T[A, i, by])  # fixed case 31104
ck(f(('address[]', 'uint240', 'bytes13'), D), T[A, i, by])  # fixed case 31105
ck(f(('address[]', 'uint240', 'bytes14'), D), T[A, i, by])  # fixed case 31106
ck(f(('address[]', 'uint240', 'bytes15'), D), T[A, i, by])  # fixed case 31107
ck(f(('address[]', 'uint240', 'bytes16'), D), T[A, i, by])  # fixed case 31108
ck(f(('address[]', 'uint240', 'bytes17'), D), T[A, i, by])  # fixed case 31109
ck(f(('address[]', 'uint240', 'bytes18'), D), T[A, i, by])  # fixed case 31110
ck(f(('address[]', 'uint240', 'bytes19'), D), T[A, i, by])  # fixed case 31111
ck(f(('address[]', 'uint240', 'bytes2'), D), T[A, i, by])  # fixed case 31112
ck(f(('address[]', 'uint240', 'bytes20'), D), T[A, i, by])  # fixed case 31113
ck(f(('address[]', 'uint240', 'bytes21'), D), T[A, i, by])  # fixed case 31114
ck(f(('address[]', 'uint240', 'bytes22'), D), T[A, i, by])  # fixed case 31115
ck(f(('address[]', 'uint240', 'bytes23'), D), T[A, i, by])  # fixed case 31116
ck(f(('address[]', 'uint240', 'bytes24'), D), T[A, i, by])  # fixed case 31117
ck(f(('address[]', 'uint240', 'bytes25'), D), T[A, i, by])  # fixed case 31118
ck(f(('address[]', 'uint240', 'bytes26'), D), T[A, i, by])  # fixed case 31119
ck(f(('address[]', 'uint240', 'bytes27'), D), T[A, i, by])  # fixed case 31120
ck(f(('address[]', 'uint240', 'bytes28'), D), T[A, i, by])  # fixed case 31121
ck(f(('address[]', 'uint240', 'bytes29'), D), T[A, i, by])  # fixed case 31122
ck(f(('address[]', 'uint240', 'bytes3'), D), T[A, i, by])  # fixed case 31123
ck(f(('address[]', 'uint240', 'bytes30'), D), T[A, i, by])  # fixed case 31124
ck(f(('address[]', 'uint240', 'bytes31'), D), T[A, i, by])  # fixed case 31125
ck(f(('address[]', 'uint240', 'bytes32'), D), T[A, i, by])  # fixed case 31126
ck(f(('address[]', 'uint240', 'bytes4'), D), T[A, i, by])  # fixed case 31127
ck(f(('address[]', 'uint240', 'bytes5'), D), T[A, i, by])  # fixed case 31128
ck(f(('address[]', 'uint240', 'bytes6'), D), T[A, i, by])  # fixed case 31129
ck(f(('address[]', 'uint240', 'bytes7'), D), T[A, i, by])  # fixed case 31130
ck(f(('address[]', 'uint240', 'bytes8'), D), T[A, i, by])  # fixed case 31131
ck(f(('address[]', 'uint240', 'bytes9'), D), T[A, i, by])  # fixed case 31132
ck(f(('address[]', 'uint240', 'int104'), D), T[A, i, i])  # fixed case 31133
ck(f(('address[]', 'uint240', 'int112'), D), T[A, i, i])  # fixed case 31134
ck(f(('address[]', 'uint240', 'int120'), D), T[A, i, i])  # fixed case 31135
ck(f(('address[]', 'uint240', 'int128'), D), T[A, i, i])  # fixed case 31136
ck(f(('address[]', 'uint240', 'int136'), D), T[A, i, i])  # fixed case 31137
ck(f(('address[]', 'uint240', 'int144'), D), T[A, i, i])  # fixed case 31138
ck(f(('address[]', 'uint240', 'int152'), D), T[A, i, i])  # fixed case 31139
ck(f(('address[]', 'uint240', 'int16'), D), T[A, i, i])  # fixed case 31140
ck(f(('address[]', 'uint240', 'int160'), D), T[A, i, i])  # fixed case 31141
ck(f(('address[]', 'uint240', 'int168'), D), T[A, i, i])  # fixed case 31142
ck(f(('address[]', 'uint240', 'int176'), D), T[A, i, i])  # fixed case 31143
ck(f(('address[]', 'uint240', 'int184'), D), T[A, i, i])  # fixed case 31144
ck(f(('address[]', 'uint240', 'int192'), D), T[A, i, i])  # fixed case 31145
ck(f(('address[]', 'uint240', 'int200'), D), T[A, i, i])  # fixed case 31146
ck(f(('address[]', 'uint240', 'int208'), D), T[A, i, i])  # fixed case 31147
ck(f(('address[]', 'uint240', 'int216'), D), T[A, i, i])  # fixed case 31148
ck(f(('address[]', 'uint240', 'int224'), D), T[A, i, i])  # fixed case 31149
ck(f(('address[]', 'uint240', 'int232'), D), T[A, i, i])  # fixed case 31150
ck(f(('address[]', 'uint240', 'int24'), D), T[A, i, i])  # fixed case 31151
ck(f(('address[]', 'uint240', 'int240'), D), T[A, i, i])  # fixed case 31152
ck(f(('address[]', 'uint240', 'int248'), D), T[A, i, i])  # fixed case 31153
ck(f(('address[]', 'uint240', 'int256'), D), T[A, i, i])  # fixed case 31154
ck(f(('address[]', 'uint240', 'int32'), D), T[A, i, i])  # fixed case 31155
ck(f(('address[]', 'uint240', 'int40'), D), T[A, i, i])  # fixed case 31156
ck(f(('address[]', 'uint240', 'int48'), D), T[A, i, i])  # fixed case 31157
ck(f(('address[]', 'uint240', 'int56'), D), T[A, i, i])  # fixed case 31158
ck(f(('address[]', 'uint240', 'int64'), D), T[A, i, i])  # fixed case 31159
ck(f(('address[]', 'uint240', 'int72'), D), T[A, i, i])  # fixed case 31160
ck(f(('address[]', 'uint240', 'int8'), D), T[A, i, i])  # fixed case 31161
ck(f(('address[]', 'uint240', 'int80'), D), T[A, i, i])  # fixed case 31162
ck(f(('address[]', 'uint240', 'int88'), D), T[A, i, i])  # fixed case 31163
ck(f(('address[]', 'uint240', 'int96'), D), T[A, i, i])  # fixed case 31164
ck(f(('address[]', 'uint240', 'string'), D), T[A, i, s])  # fixed case 31165
ck(f(('address[]', 'uint240', 'string[]'), D), T[A, i, A])  # fixed case 31166
ck(f(('address[]', 'uint240', 'uint104'), D), T[A, i, i])  # fixed case 31167
ck(f(('address[]', 'uint240', 'uint112'), D), T[A, i, i])  # fixed case 31168
ck(f(('address[]', 'uint240', 'uint120'), D), T[A, i, i])  # fixed case 31169
ck(f(('address[]', 'uint240', 'uint128'), D), T[A, i, i])  # fixed case 31170
ck(f(('address[]', 'uint240', 'uint136'), D), T[A, i, i])  # fixed case 31171
ck(f(('address[]', 'uint240', 'uint144'), D), T[A, i, i])  # fixed case 31172
ck(f(('address[]', 'uint240', 'uint152'), D), T[A, i, i])  # fixed case 31173
ck(f(('address[]', 'uint240', 'uint16'), D), T[A, i, i])  # fixed case 31174
ck(f(('address[]', 'uint240', 'uint160'), D), T[A, i, i])  # fixed case 31175
ck(f(('address[]', 'uint240', 'uint168'), D), T[A, i, i])  # fixed case 31176
ck(f(('address[]', 'uint240', 'uint176'), D), T[A, i, i])  # fixed case 31177
ck(f(('address[]', 'uint240', 'uint184'), D), T[A, i, i])  # fixed case 31178
ck(f(('address[]', 'uint240', 'uint192'), D), T[A, i, i])  # fixed case 31179
ck(f(('address[]', 'uint240', 'uint200'), D), T[A, i, i])  # fixed case 31180
ck(f(('address[]', 'uint240', 'uint208'), D), T[A, i, i])  # fixed case 31181
ck(f(('address[]', 'uint240', 'uint216'), D), T[A, i, i])  # fixed case 31182
ck(f(('address[]', 'uint240', 'uint224'), D), T[A, i, i])  # fixed case 31183
ck(f(('address[]', 'uint240', 'uint232'), D), T[A, i, i])  # fixed case 31184
ck(f(('address[]', 'uint240', 'uint24'), D), T[A, i, i])  # fixed case 31185
ck(f(('address[]', 'uint240', 'uint240'), D), T[A, i, i])  # fixed case 31186
ck(f(('address[]', 'uint240', 'uint248'), D), T[A, i, i])  # fixed case 31187
ck(f(('address[]', 'uint240', 'uint256'), D), T[A, i, i])  # fixed case 31188
ck(f(('address[]', 'uint240', 'uint32'), D), T[A, i, i])  # fixed case 31189
ck(f(('address[]', 'uint240', 'uint40'), D), T[A, i, i])  # fixed case 31190
ck(f(('address[]', 'uint240', 'uint48'), D), T[A, i, i])  # fixed case 31191
ck(f(('address[]', 'uint240', 'uint56'), D), T[A, i, i])  # fixed case 31192
ck(f(('address[]', 'uint240', 'uint64'), D), T[A, i, i])  # fixed case 31193
ck(f(('address[]', 'uint240', 'uint72'), D), T[A, i, i])  # fixed case 31194
ck(f(('address[]', 'uint240', 'uint8'), D), T[A, i, i])  # fixed case 31195
ck(f(('address[]', 'uint240', 'uint80'), D), T[A, i, i])  # fixed case 31196
ck(f(('address[]', 'uint240', 'uint88'), D), T[A, i, i])  # fixed case 31197
ck(f(('address[]', 'uint240', 'uint96'), D), T[A, i, i])  # fixed case 31198
ck(f(('address[]', 'uint240', '?'), D), T[A, i, A])  # fixed case 31199
ck(f(('address[]', 'uint248', 'address'), D), T[A, i, HA])  # fixed case 31200
ck(f(('address[]', 'uint248', 'address[]'), D), T[A, i, A])  # fixed case 31201
ck(f(('address[]', 'uint248', 'bool'), D), T[A, i, bo])  # fixed case 31202
ck(f(('address[]', 'uint248', 'bool[]'), D), T[A, i, A])  # fixed case 31203
ck(f(('address[]', 'uint248', 'bytes'), D), T[A, i, by])  # fixed case 31204
ck(f(('address[]', 'uint248', 'bytes1'), D), T[A, i, by])  # fixed case 31205
ck(f(('address[]', 'uint248', 'bytes10'), D), T[A, i, by])  # fixed case 31206
ck(f(('address[]', 'uint248', 'bytes11'), D), T[A, i, by])  # fixed case 31207
ck(f(('address[]', 'uint248', 'bytes12'), D), T[A, i, by])  # fixed case 31208
ck(f(('address[]', 'uint248', 'bytes13'), D), T[A, i, by])  # fixed case 31209
ck(f(('address[]', 'uint248', 'bytes14'), D), T[A, i, by])  # fixed case 31210
ck(f(('address[]', 'uint248', 'bytes15'), D), T[A, i, by])  # fixed case 31211
ck(f(('address[]', 'uint248', 'bytes16'), D), T[A, i, by])  # fixed case 31212
ck(f(('address[]', 'uint248', 'bytes17'), D), T[A, i, by])  # fixed case 31213
ck(f(('address[]', 'uint248', 'bytes18'), D), T[A, i, by])  # fixed case 31214
ck(f(('address[]', 'uint248', 'bytes19'), D), T[A, i, by])  # fixed case 31215
ck(f(('address[]', 'uint248', 'bytes2'), D), T[A, i, by])  # fixed case 31216
ck(f(('address[]', 'uint248', 'bytes20'), D), T[A, i, by])  # fixed case 31217
ck(f(('address[]', 'uint248', 'bytes21'), D), T[A, i, by])  # fixed case 31218
ck(f(('address[]', 'uint248', 'bytes22'), D), T[A, i, by])  # fixed case 31219
ck(f(('address[]', 'uint248', 'bytes23'), D), T[A, i, by])  # fixed case 31220
ck(f(('address[]', 'uint248', 'bytes24'), D), T[A, i, by])  # fixed case 31221
ck(f(('address[]', 'uint248', 'bytes25'), D), T[A, i, by])  # fixed case 31222
ck(f(('address[]', 'uint248', 'bytes26'), D), T[A, i, by])  # fixed case 31223
ck(f(('address[]', 'uint248', 'bytes27'), D), T[A, i, by])  # fixed case 31224
ck(f(('address[]', 'uint248', 'bytes28'), D), T[A, i, by])  # fixed case 31225
ck(f(('address[]', 'uint248', 'bytes29'), D), T[A, i, by])  # fixed case 31226
ck(f(('address[]', 'uint248', 'bytes3'), D), T[A, i, by])  # fixed case 31227
ck(f(('address[]', 'uint248', 'bytes30'), D), T[A, i, by])  # fixed case 31228
ck(f(('address[]', 'uint248', 'bytes31'), D), T[A, i, by])  # fixed case 31229
ck(f(('address[]', 'uint248', 'bytes32'), D), T[A, i, by])  # fixed case 31230
ck(f(('address[]', 'uint248', 'bytes4'), D), T[A, i, by])  # fixed case 31231
ck(f(('address[]', 'uint248', 'bytes5'), D), T[A, i, by])  # fixed case 31232
ck(f(('address[]', 'uint248', 'bytes6'), D), T[A, i, by])  # fixed case 31233
ck(f(('address[]', 'uint248', 'bytes7'), D), T[A, i, by])  # fixed case 31234
ck(f(('address[]', 'uint248', 'bytes8'), D), T[A, i, by])  # fixed case 31235
ck(f(('address[]', 'uint248', 'bytes9'), D), T[A, i, by])  # fixed case 31236
ck(f(('address[]', 'uint248', 'int104'), D), T[A, i, i])  # fixed case 31237
ck(f(('address[]', 'uint248', 'int112'), D), T[A, i, i])  # fixed case 31238
ck(f(('address[]', 'uint248', 'int120'), D), T[A, i, i])  # fixed case 31239
ck(f(('address[]', 'uint248', 'int128'), D), T[A, i, i])  # fixed case 31240
ck(f(('address[]', 'uint248', 'int136'), D), T[A, i, i])  # fixed case 31241
ck(f(('address[]', 'uint248', 'int144'), D), T[A, i, i])  # fixed case 31242
ck(f(('address[]', 'uint248', 'int152'), D), T[A, i, i])  # fixed case 31243
ck(f(('address[]', 'uint248', 'int16'), D), T[A, i, i])  # fixed case 31244
ck(f(('address[]', 'uint248', 'int160'), D), T[A, i, i])  # fixed case 31245
ck(f(('address[]', 'uint248', 'int168'), D), T[A, i, i])  # fixed case 31246
ck(f(('address[]', 'uint248', 'int176'), D), T[A, i, i])  # fixed case 31247
ck(f(('address[]', 'uint248', 'int184'), D), T[A, i, i])  # fixed case 31248
ck(f(('address[]', 'uint248', 'int192'), D), T[A, i, i])  # fixed case 31249
ck(f(('address[]', 'uint248', 'int200'), D), T[A, i, i])  # fixed case 31250
ck(f(('address[]', 'uint248', 'int208'), D), T[A, i, i])  # fixed case 31251
ck(f(('address[]', 'uint248', 'int216'), D), T[A, i, i])  # fixed case 31252
ck(f(('address[]', 'uint248', 'int224'), D), T[A, i, i])  # fixed case 31253
ck(f(('address[]', 'uint248', 'int232'), D), T[A, i, i])  # fixed case 31254
ck(f(('address[]', 'uint248', 'int24'), D), T[A, i, i])  # fixed case 31255
ck(f(('address[]', 'uint248', 'int240'), D), T[A, i, i])  # fixed case 31256
ck(f(('address[]', 'uint248', 'int248'), D), T[A, i, i])  # fixed case 31257
ck(f(('address[]', 'uint248', 'int256'), D), T[A, i, i])  # fixed case 31258
ck(f(('address[]', 'uint248', 'int32'), D), T[A, i, i])  # fixed case 31259
ck(f(('address[]', 'uint248', 'int40'), D), T[A, i, i])  # fixed case 31260
ck(f(('address[]', 'uint248', 'int48'), D), T[A, i, i])  # fixed case 31261
ck(f(('address[]', 'uint248', 'int56'), D), T[A, i, i])  # fixed case 31262
ck(f(('address[]', 'uint248', 'int64'), D), T[A, i, i])  # fixed case 31263
ck(f(('address[]', 'uint248', 'int72'), D), T[A, i, i])  # fixed case 31264
ck(f(('address[]', 'uint248', 'int8'), D), T[A, i, i])  # fixed case 31265
ck(f(('address[]', 'uint248', 'int80'), D), T[A, i, i])  # fixed case 31266
ck(f(('address[]', 'uint248', 'int88'), D), T[A, i, i])  # fixed case 31267
ck(f(('address[]', 'uint248', 'int96'), D), T[A, i, i])  # fixed case 31268
ck(f(('address[]', 'uint248', 'string'), D), T[A, i, s])  # fixed case 31269
ck(f(('address[]', 'uint248', 'string[]'), D), T[A, i, A])  # fixed case 31270
ck(f(('address[]', 'uint248', 'uint104'), D), T[A, i, i])  # fixed case 31271
ck(f(('address[]', 'uint248', 'uint112'), D), T[A, i, i])  # fixed case 31272
ck(f(('address[]', 'uint248', 'uint120'), D), T[A, i, i])  # fixed case 31273
ck(f(('address[]', 'uint248', 'uint128'), D), T[A, i, i])  # fixed case 31274
ck(f(('address[]', 'uint248', 'uint136'), D), T[A, i, i])  # fixed case 31275
ck(f(('address[]', 'uint248', 'uint144'), D), T[A, i, i])  # fixed case 31276
ck(f(('address[]', 'uint248', 'uint152'), D), T[A, i, i])  # fixed case 31277
ck(f(('address[]', 'uint248', 'uint16'), D), T[A, i, i])  # fixed case 31278
ck(f(('address[]', 'uint248', 'uint160'), D), T[A, i, i])  # fixed case 31279
ck(f(('address[]', 'uint248', 'uint168'), D), T[A, i, i])  # fixed case 31280
ck(f(('address[]', 'uint248', 'uint176'), D), T[A, i, i])  # fixed case 31281
ck(f(('address[]', 'uint248', 'uint184'), D), T[A, i, i])  # fixed case 31282
ck(f(('address[]', 'uint248', 'uint192'), D), T[A, i, i])  # fixed case 31283
ck(f(('address[]', 'uint248', 'uint200'), D), T[A, i, i])  # fixed case 31284
ck(f(('address[]', 'uint248', 'uint208'), D), T[A, i, i])  # fixed case 31285
ck(f(('address[]', 'uint248', 'uint216'), D), T[A, i, i])  # fixed case 31286
ck(f(('address[]', 'uint248', 'uint224'), D), T[A, i, i])  # fixed case 31287
ck(f(('address[]', 'uint248', 'uint232'), D), T[A, i, i])  # fixed case 31288
ck(f(('address[]', 'uint248', 'uint24'), D), T[A, i, i])  # fixed case 31289
ck(f(('address[]', 'uint248', 'uint240'), D), T[A, i, i])  # fixed case 31290
ck(f(('address[]', 'uint248', 'uint248'), D), T[A, i, i])  # fixed case 31291
ck(f(('address[]', 'uint248', 'uint256'), D), T[A, i, i])  # fixed case 31292
ck(f(('address[]', 'uint248', 'uint32'), D), T[A, i, i])  # fixed case 31293
ck(f(('address[]', 'uint248', 'uint40'), D), T[A, i, i])  # fixed case 31294
ck(f(('address[]', 'uint248', 'uint48'), D), T[A, i, i])  # fixed case 31295
ck(f(('address[]', 'uint248', 'uint56'), D), T[A, i, i])  # fixed case 31296
ck(f(('address[]', 'uint248', 'uint64'), D), T[A, i, i])  # fixed case 31297
ck(f(('address[]', 'uint248', 'uint72'), D), T[A, i, i])  # fixed case 31298
ck(f(('address[]', 'uint248', 'uint8'), D), T[A, i, i])  # fixed case 31299
ck(f(('address[]', 'uint248', 'uint80'), D), T[A, i, i])  # fixed case 31300
ck(f(('address[]', 'uint248', 'uint88'), D), T[A, i, i])  # fixed case 31301
ck(f(('address[]', 'uint248', 'uint96'), D), T[A, i, i])  # fixed case 31302
ck(f(('address[]', 'uint248', '?'), D), T[A, i, A])  # fixed case 31303
ck(f(('address[]', 'uint256', 'address'), D), T[A, i, HA])  # fixed case 31304
ck(f(('address[]', 'uint256', 'address[]'), D), T[A, i, A])  # fixed case 31305
ck(f(('address[]', 'uint256', 'bool'), D), T[A, i, bo])  # fixed case 31306
ck(f(('address[]', 'uint256', 'bool[]'), D), T[A, i, A])  # fixed case 31307
ck(f(('address[]', 'uint256', 'bytes'), D), T[A, i, by])  # fixed case 31308
ck(f(('address[]', 'uint256', 'bytes1'), D), T[A, i, by])  # fixed case 31309
ck(f(('address[]', 'uint256', 'bytes10'), D), T[A, i, by])  # fixed case 31310
ck(f(('address[]', 'uint256', 'bytes11'), D), T[A, i, by])  # fixed case 31311
ck(f(('address[]', 'uint256', 'bytes12'), D), T[A, i, by])  # fixed case 31312
ck(f(('address[]', 'uint256', 'bytes13'), D), T[A, i, by])  # fixed case 31313
ck(f(('address[]', 'uint256', 'bytes14'), D), T[A, i, by])  # fixed case 31314
ck(f(('address[]', 'uint256', 'bytes15'), D), T[A, i, by])  # fixed case 31315
ck(f(('address[]', 'uint256', 'bytes16'), D), T[A, i, by])  # fixed case 31316
ck(f(('address[]', 'uint256', 'bytes17'), D), T[A, i, by])  # fixed case 31317
ck(f(('address[]', 'uint256', 'bytes18'), D), T[A, i, by])  # fixed case 31318
ck(f(('address[]', 'uint256', 'bytes19'), D), T[A, i, by])  # fixed case 31319
ck(f(('address[]', 'uint256', 'bytes2'), D), T[A, i, by])  # fixed case 31320
ck(f(('address[]', 'uint256', 'bytes20'), D), T[A, i, by])  # fixed case 31321
ck(f(('address[]', 'uint256', 'bytes21'), D), T[A, i, by])  # fixed case 31322
ck(f(('address[]', 'uint256', 'bytes22'), D), T[A, i, by])  # fixed case 31323
ck(f(('address[]', 'uint256', 'bytes23'), D), T[A, i, by])  # fixed case 31324
ck(f(('address[]', 'uint256', 'bytes24'), D), T[A, i, by])  # fixed case 31325
ck(f(('address[]', 'uint256', 'bytes25'), D), T[A, i, by])  # fixed case 31326
ck(f(('address[]', 'uint256', 'bytes26'), D), T[A, i, by])  # fixed case 31327
ck(f(('address[]', 'uint256', 'bytes27'), D), T[A, i, by])  # fixed case 31328
ck(f(('address[]', 'uint256', 'bytes28'), D), T[A, i, by])  # fixed case 31329
ck(f(('address[]', 'uint256', 'bytes29'), D), T[A, i, by])  # fixed case 31330
ck(f(('address[]', 'uint256', 'bytes3'), D), T[A, i, by])  # fixed case 31331
ck(f(('address[]', 'uint256', 'bytes30'), D), T[A, i, by])  # fixed case 31332
ck(f(('address[]', 'uint256', 'bytes31'), D), T[A, i, by])  # fixed case 31333
ck(f(('address[]', 'uint256', 'bytes32'), D), T[A, i, by])  # fixed case 31334
ck(f(('address[]', 'uint256', 'bytes4'), D), T[A, i, by])  # fixed case 31335
ck(f(('address[]', 'uint256', 'bytes5'), D), T[A, i, by])  # fixed case 31336
ck(f(('address[]', 'uint256', 'bytes6'), D), T[A, i, by])  # fixed case 31337
ck(f(('address[]', 'uint256', 'bytes7'), D), T[A, i, by])  # fixed case 31338
ck(f(('address[]', 'uint256', 'bytes8'), D), T[A, i, by])  # fixed case 31339
ck(f(('address[]', 'uint256', 'bytes9'), D), T[A, i, by])  # fixed case 31340
ck(f(('address[]', 'uint256', 'int104'), D), T[A, i, i])  # fixed case 31341
ck(f(('address[]', 'uint256', 'int112'), D), T[A, i, i])  # fixed case 31342
ck(f(('address[]', 'uint256', 'int120'), D), T[A, i, i])  # fixed case 31343
ck(f(('address[]', 'uint256', 'int128'), D), T[A, i, i])  # fixed case 31344
ck(f(('address[]', 'uint256', 'int136'), D), T[A, i, i])  # fixed case 31345
ck(f(('address[]', 'uint256', 'int144'), D), T[A, i, i])  # fixed case 31346
ck(f(('address[]', 'uint256', 'int152'), D), T[A, i, i])  # fixed case 31347
ck(f(('address[]', 'uint256', 'int16'), D), T[A, i, i])  # fixed case 31348
ck(f(('address[]', 'uint256', 'int160'), D), T[A, i, i])  # fixed case 31349
ck(f(('address[]', 'uint256', 'int168'), D), T[A, i, i])  # fixed case 31350
ck(f(('address[]', 'uint256', 'int176'), D), T[A, i, i])  # fixed case 31351
ck(f(('address[]', 'uint256', 'int184'), D), T[A, i, i])  # fixed case 31352
ck(f(('address[]', 'uint256', 'int192'), D), T[A, i, i])  # fixed case 31353
ck(f(('address[]', 'uint256', 'int200'), D), T[A, i, i])  # fixed case 31354
ck(f(('address[]', 'uint256', 'int208'), D), T[A, i, i])  # fixed case 31355
ck(f(('address[]', 'uint256', 'int216'), D), T[A, i, i])  # fixed case 31356
ck(f(('address[]', 'uint256', 'int224'), D), T[A, i, i])  # fixed case 31357
ck(f(('address[]', 'uint256', 'int232'), D), T[A, i, i])  # fixed case 31358
ck(f(('address[]', 'uint256', 'int24'), D), T[A, i, i])  # fixed case 31359
ck(f(('address[]', 'uint256', 'int240'), D), T[A, i, i])  # fixed case 31360
ck(f(('address[]', 'uint256', 'int248'), D), T[A, i, i])  # fixed case 31361
ck(f(('address[]', 'uint256', 'int256'), D), T[A, i, i])  # fixed case 31362
ck(f(('address[]', 'uint256', 'int32'), D), T[A, i, i])  # fixed case 31363
ck(f(('address[]', 'uint256', 'int40'), D), T[A, i, i])  # fixed case 31364
ck(f(('address[]', 'uint256', 'int48'), D), T[A, i, i])  # fixed case 31365
ck(f(('address[]', 'uint256', 'int56'), D), T[A, i, i])  # fixed case 31366
ck(f(('address[]', 'uint256', 'int64'), D), T[A, i, i])  # fixed case 31367
ck(f(('address[]', 'uint256', 'int72'), D), T[A, i, i])  # fixed case 31368
ck(f(('address[]', 'uint256', 'int8'), D), T[A, i, i])  # fixed case 31369
ck(f(('address[]', 'uint256', 'int80'), D), T[A, i, i])  # fixed case 31370
ck(f(('address[]', 'uint256', 'int88'), D), T[A, i, i])  # fixed case 31371
ck(f(('address[]', 'uint256', 'int96'), D), T[A, i, i])  # fixed case 31372
ck(f(('address[]', 'uint256', 'string'), D), T[A, i, s])  # fixed case 31373
ck(f(('address[]', 'uint256', 'string[]'), D), T[A, i, A])  # fixed case 31374
ck(f(('address[]', 'uint256', 'uint104'), D), T[A, i, i])  # fixed case 31375
ck(f(('address[]', 'uint256', 'uint112'), D), T[A, i, i])  # fixed case 31376
ck(f(('address[]', 'uint256', 'uint120'), D), T[A, i, i])  # fixed case 31377
ck(f(('address[]', 'uint256', 'uint128'), D), T[A, i, i])  # fixed case 31378
ck(f(('address[]', 'uint256', 'uint136'), D), T[A, i, i])  # fixed case 31379
ck(f(('address[]', 'uint256', 'uint144'), D), T[A, i, i])  # fixed case 31380
ck(f(('address[]', 'uint256', 'uint152'), D), T[A, i, i])  # fixed case 31381
ck(f(('address[]', 'uint256', 'uint16'), D), T[A, i, i])  # fixed case 31382
ck(f(('address[]', 'uint256', 'uint160'), D), T[A, i, i])  # fixed case 31383
ck(f(('address[]', 'uint256', 'uint168'), D), T[A, i, i])  # fixed case 31384
ck(f(('address[]', 'uint256', 'uint176'), D), T[A, i, i])  # fixed case 31385
ck(f(('address[]', 'uint256', 'uint184'), D), T[A, i, i])  # fixed case 31386
ck(f(('address[]', 'uint256', 'uint192'), D), T[A, i, i])  # fixed case 31387
ck(f(('address[]', 'uint256', 'uint200'), D), T[A, i, i])  # fixed case 31388
ck(f(('address[]', 'uint256', 'uint208'), D), T[A, i, i])  # fixed case 31389
ck(f(('address[]', 'uint256', 'uint216'), D), T[A, i, i])  # fixed case 31390
ck(f(('address[]', 'uint256', 'uint224'), D), T[A, i, i])  # fixed case 31391
ck(f(('address[]', 'uint256', 'uint232'), D), T[A, i, i])  # fixed case 31392
ck(f(('address[]', 'uint256', 'uint24'), D), T[A, i, i])  # fixed case 31393
ck(f(('address[]', 'uint256', 'uint240'), D), T[A, i, i])  # fixed case 31394
ck(f(('address[]', 'uint256', 'uint248'), D), T[A, i, i])  # fixed case 31395
ck(f(('address[]', 'uint256', 'uint256'), D), T[A, i, i])  # fixed case 31396
ck(f(('address[]', 'uint256', 'uint32'), D), T[A, i, i])  # fixed case 31397
ck(f(('address[]', 'uint256', 'uint40'), D), T[A, i, i])  # fixed case 31398
ck(f(('address[]', 'uint256', 'uint48'), D), T[A, i, i])  # fixed case 31399
ck(f(('address[]', 'uint256', 'uint56'), D), T[A, i, i])  # fixed case 31400
ck(f(('address[]', 'uint256', 'uint64'), D), T[A, i, i])  # fixed case 31401
ck(f(('address[]', 'uint256', 'uint72'), D), T[A, i, i])  # fixed case 31402
ck(f(('address[]', 'uint256', 'uint8'), D), T[A, i, i])  # fixed case 31403
ck(f(('address[]', 'uint256', 'uint80'), D), T[A, i, i])  # fixed case 31404
ck(f(('address[]', 'uint256', 'uint88'), D), T[A, i, i])  # fixed case 31405
ck(f(('address[]', 'uint256', 'uint96'), D), T[A, i, i])  # fixed case 31406
ck(f(('address[]', 'uint256', '?'), D), T[A, i, A])  # fixed case 31407
ck(f(('address[]', 'uint32', 'address'), D), T[A, i, HA])  # fixed case 31408
ck(f(('address[]', 'uint32', 'address[]'), D), T[A, i, A])  # fixed case 31409
ck(f(('address[]', 'uint32', 'bool'), D), T[A, i, bo])  # fixed case 31410
ck(f(('address[]', 'uint32', 'bool[]'), D), T[A, i, A])  # fixed case 31411
ck(f(('address[]', 'uint32', 'bytes'), D), T[A, i, by])  # fixed case 31412
ck(f(('address[]', 'uint32', 'bytes1'), D), T[A, i, by])  # fixed case 31413
ck(f(('address[]', 'uint32', 'bytes10'), D), T[A, i, by])  # fixed case 31414
ck(f(('address[]', 'uint32', 'bytes11'), D), T[A, i, by])  # fixed case 31415
ck(f(('address[]', 'uint32', 'bytes12'), D), T[A, i, by])  # fixed case 31416
ck(f(('address[]', 'uint32', 'bytes13'), D), T[A, i, by])  # fixed case 31417
ck(f(('address[]', 'uint32', 'bytes14'), D), T[A, i, by])  # fixed case 31418
ck(f(('address[]', 'uint32', 'bytes15'), D), T[A, i, by])  # fixed case 31419
ck(f(('address[]', 'uint32', 'bytes16'), D), T[A, i, by])  # fixed case 31420
ck(f(('address[]', 'uint32', 'bytes17'), D), T[A, i, by])  # fixed case 31421
ck(f(('address[]', 'uint32', 'bytes18'), D), T[A, i, by])  # fixed case 31422
ck(f(('address[]', 'uint32', 'bytes19'), D), T[A, i, by])  # fixed case 31423
ck(f(('address[]', 'uint32', 'bytes2'), D), T[A, i, by])  # fixed case 31424
ck(f(('address[]', 'uint32', 'bytes20'), D), T[A, i, by])  # fixed case 31425
ck(f(('address[]', 'uint32', 'bytes21'), D), T[A, i, by])  # fixed case 31426
ck(f(('address[]', 'uint32', 'bytes22'), D), T[A, i, by])  # fixed case 31427
ck(f(('address[]', 'uint32', 'bytes23'), D), T[A, i, by])  # fixed case 31428
ck(f(('address[]', 'uint32', 'bytes24'), D), T[A, i, by])  # fixed case 31429
ck(f(('address[]', 'uint32', 'bytes25'), D), T[A, i, by])  # fixed case 31430
ck(f(('address[]', 'uint32', 'bytes26'), D), T[A, i, by])  # fixed case 31431
ck(f(('address[]', 'uint32', 'bytes27'), D), T[A, i, by])  # fixed case 31432
ck(f(('address[]', 'uint32', 'bytes28'), D), T[A, i, by])  # fixed case 31433
ck(f(('address[]', 'uint32', 'bytes29'), D), T[A, i, by])  # fixed case 31434
ck(f(('address[]', 'uint32', 'bytes3'), D), T[A, i, by])  # fixed case 31435
ck(f(('address[]', 'uint32', 'bytes30'), D), T[A, i, by])  # fixed case 31436
ck(f(('address[]', 'uint32', 'bytes31'), D), T[A, i, by])  # fixed case 31437
ck(f(('address[]', 'uint32', 'bytes32'), D), T[A, i, by])  # fixed case 31438
ck(f(('address[]', 'uint32', 'bytes4'), D), T[A, i, by])  # fixed case 31439
ck(f(('address[]', 'uint32', 'bytes5'), D), T[A, i, by])  # fixed case 31440
ck(f(('address[]', 'uint32', 'bytes6'), D), T[A, i, by])  # fixed case 31441
ck(f(('address[]', 'uint32', 'bytes7'), D), T[A, i, by])  # fixed case 31442
ck(f(('address[]', 'uint32', 'bytes8'), D), T[A, i, by])  # fixed case 31443
ck(f(('address[]', 'uint32', 'bytes9'), D), T[A, i, by])  # fixed case 31444
ck(f(('address[]', 'uint32', 'int104'), D), T[A, i, i])  # fixed case 31445
ck(f(('address[]', 'uint32', 'int112'), D), T[A, i, i])  # fixed case 31446
ck(f(('address[]', 'uint32', 'int120'), D), T[A, i, i])  # fixed case 31447
ck(f(('address[]', 'uint32', 'int128'), D), T[A, i, i])  # fixed case 31448
ck(f(('address[]', 'uint32', 'int136'), D), T[A, i, i])  # fixed case 31449
ck(f(('address[]', 'uint32', 'int144'), D), T[A, i, i])  # fixed case 31450
ck(f(('address[]', 'uint32', 'int152'), D), T[A, i, i])  # fixed case 31451
ck(f(('address[]', 'uint32', 'int16'), D), T[A, i, i])  # fixed case 31452
ck(f(('address[]', 'uint32', 'int160'), D), T[A, i, i])  # fixed case 31453
ck(f(('address[]', 'uint32', 'int168'), D), T[A, i, i])  # fixed case 31454
ck(f(('address[]', 'uint32', 'int176'), D), T[A, i, i])  # fixed case 31455
ck(f(('address[]', 'uint32', 'int184'), D), T[A, i, i])  # fixed case 31456
ck(f(('address[]', 'uint32', 'int192'), D), T[A, i, i])  # fixed case 31457
ck(f(('address[]', 'uint32', 'int200'), D), T[A, i, i])  # fixed case 31458
ck(f(('address[]', 'uint32', 'int208'), D), T[A, i, i])  # fixed case 31459
ck(f(('address[]', 'uint32', 'int216'), D), T[A, i, i])  # fixed case 31460
ck(f(('address[]', 'uint32', 'int224'), D), T[A, i, i])  # fixed case 31461
ck(f(('address[]', 'uint32', 'int232'), D), T[A, i, i])  # fixed case 31462
ck(f(('address[]', 'uint32', 'int24'), D), T[A, i, i])  # fixed case 31463
ck(f(('address[]', 'uint32', 'int240'), D), T[A, i, i])  # fixed case 31464
ck(f(('address[]', 'uint32', 'int248'), D), T[A, i, i])  # fixed case 31465
ck(f(('address[]', 'uint32', 'int256'), D), T[A, i, i])  # fixed case 31466
ck(f(('address[]', 'uint32', 'int32'), D), T[A, i, i])  # fixed case 31467
ck(f(('address[]', 'uint32', 'int40'), D), T[A, i, i])  # fixed case 31468
ck(f(('address[]', 'uint32', 'int48'), D), T[A, i, i])  # fixed case 31469
ck(f(('address[]', 'uint32', 'int56'), D), T[A, i, i])  # fixed case 31470
ck(f(('address[]', 'uint32', 'int64'), D), T[A, i, i])  # fixed case 31471
ck(f(('address[]', 'uint32', 'int72'), D), T[A, i, i])  # fixed case 31472
ck(f(('address[]', 'uint32', 'int8'), D), T[A, i, i])  # fixed case 31473
ck(f(('address[]', 'uint32', 'int80'), D), T[A, i, i])  # fixed case 31474
ck(f(('address[]', 'uint32', 'int88'), D), T[A, i, i])  # fixed case 31475
ck(f(('address[]', 'uint32', 'int96'), D), T[A, i, i])  # fixed case 31476
ck(f(('address[]', 'uint32', 'string'), D), T[A, i, s])  # fixed case 31477
ck(f(('address[]', 'uint32', 'string[]'), D), T[A, i, A])  # fixed case 31478
ck(f(('address[]', 'uint32', 'uint104'), D), T[A, i, i])  # fixed case 31479
ck(f(('address[]', 'uint32', 'uint112'), D), T[A, i, i])  # fixed case 31480
ck(f(('address[]', 'uint32', 'uint120'), D), T[A, i, i])  # fixed case 31481
ck(f(('address[]', 'uint32', 'uint128'), D), T[A, i, i])  # fixed case 31482
ck(f(('address[]', 'uint32', 'uint136'), D), T[A, i, i])  # fixed case 31483
ck(f(('address[]', 'uint32', 'uint144'), D), T[A, i, i])  # fixed case 31484
ck(f(('address[]', 'uint32', 'uint152'), D), T[A, i, i])  # fixed case 31485
ck(f(('address[]', 'uint32', 'uint16'), D), T[A, i, i])  # fixed case 31486
ck(f(('address[]', 'uint32', 'uint160'), D), T[A, i, i])  # fixed case 31487
ck(f(('address[]', 'uint32', 'uint168'), D), T[A, i, i])  # fixed case 31488
ck(f(('address[]', 'uint32', 'uint176'), D), T[A, i, i])  # fixed case 31489
ck(f(('address[]', 'uint32', 'uint184'), D), T[A, i, i])  # fixed case 31490
ck(f(('address[]', 'uint32', 'uint192'), D), T[A, i, i])  # fixed case 31491
ck(f(('address[]', 'uint32', 'uint200'), D), T[A, i, i])  # fixed case 31492
ck(f(('address[]', 'uint32', 'uint208'), D), T[A, i, i])  # fixed case 31493
ck(f(('address[]', 'uint32', 'uint216'), D), T[A, i, i])  # fixed case 31494
ck(f(('address[]', 'uint32', 'uint224'), D), T[A, i, i])  # fixed case 31495
ck(f(('address[]', 'uint32', 'uint232'), D), T[A, i, i])  # fixed case 31496
ck(f(('address[]', 'uint32', 'uint24'), D), T[A, i, i])  # fixed case 31497
ck(f(('address[]', 'uint32', 'uint240'), D), T[A, i, i])  # fixed case 31498
ck(f(('address[]', 'uint32', 'uint248'), D), T[A, i, i])  # fixed case 31499
ck(f(('address[]', 'uint32', 'uint256'), D), T[A, i, i])  # fixed case 31500
ck(f(('address[]', 'uint32', 'uint32'), D), T[A, i, i])  # fixed case 31501
ck(f(('address[]', 'uint32', 'uint40'), D), T[A, i, i])  # fixed case 31502
ck(f(('address[]', 'uint32', 'uint48'), D), T[A, i, i])  # fixed case 31503
ck(f(('address[]', 'uint32', 'uint56'), D), T[A, i, i])  # fixed case 31504
ck(f(('address[]', 'uint32', 'uint64'), D), T[A, i, i])  # fixed case 31505
ck(f(('address[]', 'uint32', 'uint72'), D), T[A, i, i])  # fixed case 31506
ck(f(('address[]', 'uint32', 'uint8'), D), T[A, i, i])  # fixed case 31507
ck(f(('address[]', 'uint32', 'uint80'), D), T[A, i, i])  # fixed case 31508
ck(f(('address[]', 'uint32', 'uint88'), D), T[A, i, i])  # fixed case 31509
ck(f(('address[]', 'uint32', 'uint96'), D), T[A, i, i])  # fixed case 31510
ck(f(('address[]', 'uint32', '?'), D), T[A, i, A])  # fixed case 31511
ck(f(('address[]', 'uint40', 'address'), D), T[A, i, HA])  # fixed case 31512
ck(f(('address[]', 'uint40', 'address[]'), D), T[A, i, A])  # fixed case 31513
ck(f(('address[]', 'uint40', 'bool'), D), T[A, i, bo])  # fixed case 31514
ck(f(('address[]', 'uint40', 'bool[]'), D), T[A, i, A])  # fixed case 31515
ck(f(('address[]', 'uint40', 'bytes'), D), T[A, i, by])  # fixed case 31516
ck(f(('address[]', 'uint40', 'bytes1'), D), T[A, i, by])  # fixed case 31517
ck(f(('address[]', 'uint40', 'bytes10'), D), T[A, i, by])  # fixed case 31518
ck(f(('address[]', 'uint40', 'bytes11'), D), T[A, i, by])  # fixed case 31519
ck(f(('address[]', 'uint40', 'bytes12'), D), T[A, i, by])  # fixed case 31520
ck(f(('address[]', 'uint40', 'bytes13'), D), T[A, i, by])  # fixed case 31521
ck(f(('address[]', 'uint40', 'bytes14'), D), T[A, i, by])  # fixed case 31522
ck(f(('address[]', 'uint40', 'bytes15'), D), T[A, i, by])  # fixed case 31523
ck(f(('address[]', 'uint40', 'bytes16'), D), T[A, i, by])  # fixed case 31524
ck(f(('address[]', 'uint40', 'bytes17'), D), T[A, i, by])  # fixed case 31525
ck(f(('address[]', 'uint40', 'bytes18'), D), T[A, i, by])  # fixed case 31526
ck(f(('address[]', 'uint40', 'bytes19'), D), T[A, i, by])  # fixed case 31527
ck(f(('address[]', 'uint40', 'bytes2'), D), T[A, i, by])  # fixed case 31528
ck(f(('address[]', 'uint40', 'bytes20'), D), T[A, i, by])  # fixed case 31529
ck(f(('address[]', 'uint40', 'bytes21'), D), T[A, i, by])  # fixed case 31530
ck(f(('address[]', 'uint40', 'bytes22'), D), T[A, i, by])  # fixed case 31531
ck(f(('address[]', 'uint40', 'bytes23'), D), T[A, i, by])  # fixed case 31532
ck(f(('address[]', 'uint40', 'bytes24'), D), T[A, i, by])  # fixed case 31533
ck(f(('address[]', 'uint40', 'bytes25'), D), T[A, i, by])  # fixed case 31534
ck(f(('address[]', 'uint40', 'bytes26'), D), T[A, i, by])  # fixed case 31535
ck(f(('address[]', 'uint40', 'bytes27'), D), T[A, i, by])  # fixed case 31536
ck(f(('address[]', 'uint40', 'bytes28'), D), T[A, i, by])  # fixed case 31537
ck(f(('address[]', 'uint40', 'bytes29'), D), T[A, i, by])  # fixed case 31538
ck(f(('address[]', 'uint40', 'bytes3'), D), T[A, i, by])  # fixed case 31539
ck(f(('address[]', 'uint40', 'bytes30'), D), T[A, i, by])  # fixed case 31540
ck(f(('address[]', 'uint40', 'bytes31'), D), T[A, i, by])  # fixed case 31541
ck(f(('address[]', 'uint40', 'bytes32'), D), T[A, i, by])  # fixed case 31542
ck(f(('address[]', 'uint40', 'bytes4'), D), T[A, i, by])  # fixed case 31543
ck(f(('address[]', 'uint40', 'bytes5'), D), T[A, i, by])  # fixed case 31544
ck(f(('address[]', 'uint40', 'bytes6'), D), T[A, i, by])  # fixed case 31545
ck(f(('address[]', 'uint40', 'bytes7'), D), T[A, i, by])  # fixed case 31546
ck(f(('address[]', 'uint40', 'bytes8'), D), T[A, i, by])  # fixed case 31547
ck(f(('address[]', 'uint40', 'bytes9'), D), T[A, i, by])  # fixed case 31548
ck(f(('address[]', 'uint40', 'int104'), D), T[A, i, i])  # fixed case 31549
ck(f(('address[]', 'uint40', 'int112'), D), T[A, i, i])  # fixed case 31550
ck(f(('address[]', 'uint40', 'int120'), D), T[A, i, i])  # fixed case 31551
ck(f(('address[]', 'uint40', 'int128'), D), T[A, i, i])  # fixed case 31552
ck(f(('address[]', 'uint40', 'int136'), D), T[A, i, i])  # fixed case 31553
ck(f(('address[]', 'uint40', 'int144'), D), T[A, i, i])  # fixed case 31554
ck(f(('address[]', 'uint40', 'int152'), D), T[A, i, i])  # fixed case 31555
ck(f(('address[]', 'uint40', 'int16'), D), T[A, i, i])  # fixed case 31556
ck(f(('address[]', 'uint40', 'int160'), D), T[A, i, i])  # fixed case 31557
ck(f(('address[]', 'uint40', 'int168'), D), T[A, i, i])  # fixed case 31558
ck(f(('address[]', 'uint40', 'int176'), D), T[A, i, i])  # fixed case 31559
ck(f(('address[]', 'uint40', 'int184'), D), T[A, i, i])  # fixed case 31560
ck(f(('address[]', 'uint40', 'int192'), D), T[A, i, i])  # fixed case 31561
ck(f(('address[]', 'uint40', 'int200'), D), T[A, i, i])  # fixed case 31562
ck(f(('address[]', 'uint40', 'int208'), D), T[A, i, i])  # fixed case 31563
ck(f(('address[]', 'uint40', 'int216'), D), T[A, i, i])  # fixed case 31564
ck(f(('address[]', 'uint40', 'int224'), D), T[A, i, i])  # fixed case 31565
ck(f(('address[]', 'uint40', 'int232'), D), T[A, i, i])  # fixed case 31566
ck(f(('address[]', 'uint40', 'int24'), D), T[A, i, i])  # fixed case 31567
ck(f(('address[]', 'uint40', 'int240'), D), T[A, i, i])  # fixed case 31568
ck(f(('address[]', 'uint40', 'int248'), D), T[A, i, i])  # fixed case 31569
ck(f(('address[]', 'uint40', 'int256'), D), T[A, i, i])  # fixed case 31570
ck(f(('address[]', 'uint40', 'int32'), D), T[A, i, i])  # fixed case 31571
ck(f(('address[]', 'uint40', 'int40'), D), T[A, i, i])  # fixed case 31572
ck(f(('address[]', 'uint40', 'int48'), D), T[A, i, i])  # fixed case 31573
ck(f(('address[]', 'uint40', 'int56'), D), T[A, i, i])  # fixed case 31574
ck(f(('address[]', 'uint40', 'int64'), D), T[A, i, i])  # fixed case 31575
ck(f(('address[]', 'uint40', 'int72'), D), T[A, i, i])  # fixed case 31576
ck(f(('address[]', 'uint40', 'int8'), D), T[A, i, i])  # fixed case 31577
ck(f(('address[]', 'uint40', 'int80'), D), T[A, i, i])  # fixed case 31578
ck(f(('address[]', 'uint40', 'int88'), D), T[A, i, i])  # fixed case 31579
ck(f(('address[]', 'uint40', 'int96'), D), T[A, i, i])  # fixed case 31580
ck(f(('address[]', 'uint40', 'string'), D), T[A, i, s])  # fixed case 31581
ck(f(('address[]', 'uint40', 'string[]'), D), T[A, i, A])  # fixed case 31582
ck(f(('address[]', 'uint40', 'uint104'), D), T[A, i, i])  # fixed case 31583
ck(f(('address[]', 'uint40', 'uint112'), D), T[A, i, i])  # fixed case 31584
ck(f(('address[]', 'uint40', 'uint120'), D), T[A, i, i])  # fixed case 31585
ck(f(('address[]', 'uint40', 'uint128'), D), T[A, i, i])  # fixed case 31586
ck(f(('address[]', 'uint40', 'uint136'), D), T[A, i, i])  # fixed case 31587
ck(f(('address[]', 'uint40', 'uint144'), D), T[A, i, i])  # fixed case 31588
ck(f(('address[]', 'uint40', 'uint152'), D), T[A, i, i])  # fixed case 31589
ck(f(('address[]', 'uint40', 'uint16'), D), T[A, i, i])  # fixed case 31590
ck(f(('address[]', 'uint40', 'uint160'), D), T[A, i, i])  # fixed case 31591
ck(f(('address[]', 'uint40', 'uint168'), D), T[A, i, i])  # fixed case 31592
ck(f(('address[]', 'uint40', 'uint176'), D), T[A, i, i])  # fixed case 31593
ck(f(('address[]', 'uint40', 'uint184'), D), T[A, i, i])  # fixed case 31594
ck(f(('address[]', 'uint40', 'uint192'), D), T[A, i, i])  # fixed case 31595
ck(f(('address[]', 'uint40', 'uint200'), D), T[A, i, i])  # fixed case 31596
ck(f(('address[]', 'uint40', 'uint208'), D), T[A, i, i])  # fixed case 31597
ck(f(('address[]', 'uint40', 'uint216'), D), T[A, i, i])  # fixed case 31598
ck(f(('address[]', 'uint40', 'uint224'), D), T[A, i, i])  # fixed case 31599
ck(f(('address[]', 'uint40', 'uint232'), D), T[A, i, i])  # fixed case 31600
ck(f(('address[]', 'uint40', 'uint24'), D), T[A, i, i])  # fixed case 31601
ck(f(('address[]', 'uint40', 'uint240'), D), T[A, i, i])  # fixed case 31602
ck(f(('address[]', 'uint40', 'uint248'), D), T[A, i, i])  # fixed case 31603
ck(f(('address[]', 'uint40', 'uint256'), D), T[A, i, i])  # fixed case 31604
ck(f(('address[]', 'uint40', 'uint32'), D), T[A, i, i])  # fixed case 31605
ck(f(('address[]', 'uint40', 'uint40'), D), T[A, i, i])  # fixed case 31606
ck(f(('address[]', 'uint40', 'uint48'), D), T[A, i, i])  # fixed case 31607
ck(f(('address[]', 'uint40', 'uint56'), D), T[A, i, i])  # fixed case 31608
ck(f(('address[]', 'uint40', 'uint64'), D), T[A, i, i])  # fixed case 31609
ck(f(('address[]', 'uint40', 'uint72'), D), T[A, i, i])  # fixed case 31610
ck(f(('address[]', 'uint40', 'uint8'), D), T[A, i, i])  # fixed case 31611
ck(f(('address[]', 'uint40', 'uint80'), D), T[A, i, i])  # fixed case 31612
ck(f(('address[]', 'uint40', 'uint88'), D), T[A, i, i])  # fixed case 31613
ck(f(('address[]', 'uint40', 'uint96'), D), T[A, i, i])  # fixed case 31614
ck(f(('address[]', 'uint40', '?'), D), T[A, i, A])  # fixed case 31615
ck(f(('address[]', 'uint48', 'address'), D), T[A, i, HA])  # fixed case 31616
ck(f(('address[]', 'uint48', 'address[]'), D), T[A, i, A])  # fixed case 31617
ck(f(('address[]', 'uint48', 'bool'), D), T[A, i, bo])  # fixed case 31618
ck(f(('address[]', 'uint48', 'bool[]'), D), T[A, i, A])  # fixed case 31619
ck(f(('address[]', 'uint48', 'bytes'), D), T[A, i, by])  # fixed case 31620
ck(f(('address[]', 'uint48', 'bytes1'), D), T[A, i, by])  # fixed case 31621
ck(f(('address[]', 'uint48', 'bytes10'), D), T[A, i, by])  # fixed case 31622
ck(f(('address[]', 'uint48', 'bytes11'), D), T[A, i, by])  # fixed case 31623
ck(f(('address[]', 'uint48', 'bytes12'), D), T[A, i, by])  # fixed case 31624
ck(f(('address[]', 'uint48', 'bytes13'), D), T[A, i, by])  # fixed case 31625
ck(f(('address[]', 'uint48', 'bytes14'), D), T[A, i, by])  # fixed case 31626
ck(f(('address[]', 'uint48', 'bytes15'), D), T[A, i, by])  # fixed case 31627
ck(f(('address[]', 'uint48', 'bytes16'), D), T[A, i, by])  # fixed case 31628
ck(f(('address[]', 'uint48', 'bytes17'), D), T[A, i, by])  # fixed case 31629
ck(f(('address[]', 'uint48', 'bytes18'), D), T[A, i, by])  # fixed case 31630
ck(f(('address[]', 'uint48', 'bytes19'), D), T[A, i, by])  # fixed case 31631
ck(f(('address[]', 'uint48', 'bytes2'), D), T[A, i, by])  # fixed case 31632
ck(f(('address[]', 'uint48', 'bytes20'), D), T[A, i, by])  # fixed case 31633
ck(f(('address[]', 'uint48', 'bytes21'), D), T[A, i, by])  # fixed case 31634
ck(f(('address[]', 'uint48', 'bytes22'), D), T[A, i, by])  # fixed case 31635
ck(f(('address[]', 'uint48', 'bytes23'), D), T[A, i, by])  # fixed case 31636
ck(f(('address[]', 'uint48', 'bytes24'), D), T[A, i, by])  # fixed case 31637
ck(f(('address[]', 'uint48', 'bytes25'), D), T[A, i, by])  # fixed case 31638
ck(f(('address[]', 'uint48', 'bytes26'), D), T[A, i, by])  # fixed case 31639
ck(f(('address[]', 'uint48', 'bytes27'), D), T[A, i, by])  # fixed case 31640
ck(f(('address[]', 'uint48', 'bytes28'), D), T[A, i, by])  # fixed case 31641
ck(f(('address[]', 'uint48', 'bytes29'), D), T[A, i, by])  # fixed case 31642
ck(f(('address[]', 'uint48', 'bytes3'), D), T[A, i, by])  # fixed case 31643
ck(f(('address[]', 'uint48', 'bytes30'), D), T[A, i, by])  # fixed case 31644
ck(f(('address[]', 'uint48', 'bytes31'), D), T[A, i, by])  # fixed case 31645
ck(f(('address[]', 'uint48', 'bytes32'), D), T[A, i, by])  # fixed case 31646
ck(f(('address[]', 'uint48', 'bytes4'), D), T[A, i, by])  # fixed case 31647
ck(f(('address[]', 'uint48', 'bytes5'), D), T[A, i, by])  # fixed case 31648
ck(f(('address[]', 'uint48', 'bytes6'), D), T[A, i, by])  # fixed case 31649
ck(f(('address[]', 'uint48', 'bytes7'), D), T[A, i, by])  # fixed case 31650
ck(f(('address[]', 'uint48', 'bytes8'), D), T[A, i, by])  # fixed case 31651
ck(f(('address[]', 'uint48', 'bytes9'), D), T[A, i, by])  # fixed case 31652
ck(f(('address[]', 'uint48', 'int104'), D), T[A, i, i])  # fixed case 31653
ck(f(('address[]', 'uint48', 'int112'), D), T[A, i, i])  # fixed case 31654
ck(f(('address[]', 'uint48', 'int120'), D), T[A, i, i])  # fixed case 31655
ck(f(('address[]', 'uint48', 'int128'), D), T[A, i, i])  # fixed case 31656
ck(f(('address[]', 'uint48', 'int136'), D), T[A, i, i])  # fixed case 31657
ck(f(('address[]', 'uint48', 'int144'), D), T[A, i, i])  # fixed case 31658
ck(f(('address[]', 'uint48', 'int152'), D), T[A, i, i])  # fixed case 31659
ck(f(('address[]', 'uint48', 'int16'), D), T[A, i, i])  # fixed case 31660
ck(f(('address[]', 'uint48', 'int160'), D), T[A, i, i])  # fixed case 31661
ck(f(('address[]', 'uint48', 'int168'), D), T[A, i, i])  # fixed case 31662
ck(f(('address[]', 'uint48', 'int176'), D), T[A, i, i])  # fixed case 31663
ck(f(('address[]', 'uint48', 'int184'), D), T[A, i, i])  # fixed case 31664
ck(f(('address[]', 'uint48', 'int192'), D), T[A, i, i])  # fixed case 31665
ck(f(('address[]', 'uint48', 'int200'), D), T[A, i, i])  # fixed case 31666
ck(f(('address[]', 'uint48', 'int208'), D), T[A, i, i])  # fixed case 31667
ck(f(('address[]', 'uint48', 'int216'), D), T[A, i, i])  # fixed case 31668
ck(f(('address[]', 'uint48', 'int224'), D), T[A, i, i])  # fixed case 31669
ck(f(('address[]', 'uint48', 'int232'), D), T[A, i, i])  # fixed case 31670
ck(f(('address[]', 'uint48', 'int24'), D), T[A, i, i])  # fixed case 31671
ck(f(('address[]', 'uint48', 'int240'), D), T[A, i, i])  # fixed case 31672
ck(f(('address[]', 'uint48', 'int248'), D), T[A, i, i])  # fixed case 31673
ck(f(('address[]', 'uint48', 'int256'), D), T[A, i, i])  # fixed case 31674
ck(f(('address[]', 'uint48', 'int32'), D), T[A, i, i])  # fixed case 31675
ck(f(('address[]', 'uint48', 'int40'), D), T[A, i, i])  # fixed case 31676
ck(f(('address[]', 'uint48', 'int48'), D), T[A, i, i])  # fixed case 31677
ck(f(('address[]', 'uint48', 'int56'), D), T[A, i, i])  # fixed case 31678
ck(f(('address[]', 'uint48', 'int64'), D), T[A, i, i])  # fixed case 31679
ck(f(('address[]', 'uint48', 'int72'), D), T[A, i, i])  # fixed case 31680
ck(f(('address[]', 'uint48', 'int8'), D), T[A, i, i])  # fixed case 31681
ck(f(('address[]', 'uint48', 'int80'), D), T[A, i, i])  # fixed case 31682
ck(f(('address[]', 'uint48', 'int88'), D), T[A, i, i])  # fixed case 31683
ck(f(('address[]', 'uint48', 'int96'), D), T[A, i, i])  # fixed case 31684
ck(f(('address[]', 'uint48', 'string'), D), T[A, i, s])  # fixed case 31685
ck(f(('address[]', 'uint48', 'string[]'), D), T[A, i, A])  # fixed case 31686
ck(f(('address[]', 'uint48', 'uint104'), D), T[A, i, i])  # fixed case 31687
ck(f(('address[]', 'uint48', 'uint112'), D), T[A, i, i])  # fixed case 31688
ck(f(('address[]', 'uint48', 'uint120'), D), T[A, i, i])  # fixed case 31689
ck(f(('address[]', 'uint48', 'uint128'), D), T[A, i, i])  # fixed case 31690
ck(f(('address[]', 'uint48', 'uint136'), D), T[A, i, i])  # fixed case 31691
ck(f(('address[]', 'uint48', 'uint144'), D), T[A, i, i])  # fixed case 31692
ck(f(('address[]', 'uint48', 'uint152'), D), T[A, i, i])  # fixed case 31693
ck(f(('address[]', 'uint48', 'uint16'), D), T[A, i, i])  # fixed case 31694
ck(f(('address[]', 'uint48', 'uint160'), D), T[A, i, i])  # fixed case 31695
ck(f(('address[]', 'uint48', 'uint168'), D), T[A, i, i])  # fixed case 31696
ck(f(('address[]', 'uint48', 'uint176'), D), T[A, i, i])  # fixed case 31697
ck(f(('address[]', 'uint48', 'uint184'), D), T[A, i, i])  # fixed case 31698
ck(f(('address[]', 'uint48', 'uint192'), D), T[A, i, i])  # fixed case 31699
ck(f(('address[]', 'uint48', 'uint200'), D), T[A, i, i])  # fixed case 31700
ck(f(('address[]', 'uint48', 'uint208'), D), T[A, i, i])  # fixed case 31701
ck(f(('address[]', 'uint48', 'uint216'), D), T[A, i, i])  # fixed case 31702
ck(f(('address[]', 'uint48', 'uint224'), D), T[A, i, i])  # fixed case 31703
ck(f(('address[]', 'uint48', 'uint232'), D), T[A, i, i])  # fixed case 31704
ck(f(('address[]', 'uint48', 'uint24'), D), T[A, i, i])  # fixed case 31705
ck(f(('address[]', 'uint48', 'uint240'), D), T[A, i, i])  # fixed case 31706
ck(f(('address[]', 'uint48', 'uint248'), D), T[A, i, i])  # fixed case 31707
ck(f(('address[]', 'uint48', 'uint256'), D), T[A, i, i])  # fixed case 31708
ck(f(('address[]', 'uint48', 'uint32'), D), T[A, i, i])  # fixed case 31709
ck(f(('address[]', 'uint48', 'uint40'), D), T[A, i, i])  # fixed case 31710
ck(f(('address[]', 'uint48', 'uint48'), D), T[A, i, i])  # fixed case 31711
ck(f(('address[]', 'uint48', 'uint56'), D), T[A, i, i])  # fixed case 31712
ck(f(('address[]', 'uint48', 'uint64'), D), T[A, i, i])  # fixed case 31713
ck(f(('address[]', 'uint48', 'uint72'), D), T[A, i, i])  # fixed case 31714
ck(f(('address[]', 'uint48', 'uint8'), D), T[A, i, i])  # fixed case 31715
ck(f(('address[]', 'uint48', 'uint80'), D), T[A, i, i])  # fixed case 31716
ck(f(('address[]', 'uint48', 'uint88'), D), T[A, i, i])  # fixed case 31717
ck(f(('address[]', 'uint48', 'uint96'), D), T[A, i, i])  # fixed case 31718
ck(f(('address[]', 'uint48', '?'), D), T[A, i, A])  # fixed case 31719
ck(f(('address[]', 'uint56', 'address'), D), T[A, i, HA])  # fixed case 31720
ck(f(('address[]', 'uint56', 'address[]'), D), T[A, i, A])  # fixed case 31721
ck(f(('address[]', 'uint56', 'bool'), D), T[A, i, bo])  # fixed case 31722
ck(f(('address[]', 'uint56', 'bool[]'), D), T[A, i, A])  # fixed case 31723
ck(f(('address[]', 'uint56', 'bytes'), D), T[A, i, by])  # fixed case 31724
ck(f(('address[]', 'uint56', 'bytes1'), D), T[A, i, by])  # fixed case 31725
ck(f(('address[]', 'uint56', 'bytes10'), D), T[A, i, by])  # fixed case 31726
ck(f(('address[]', 'uint56', 'bytes11'), D), T[A, i, by])  # fixed case 31727
ck(f(('address[]', 'uint56', 'bytes12'), D), T[A, i, by])  # fixed case 31728
ck(f(('address[]', 'uint56', 'bytes13'), D), T[A, i, by])  # fixed case 31729
ck(f(('address[]', 'uint56', 'bytes14'), D), T[A, i, by])  # fixed case 31730
ck(f(('address[]', 'uint56', 'bytes15'), D), T[A, i, by])  # fixed case 31731
ck(f(('address[]', 'uint56', 'bytes16'), D), T[A, i, by])  # fixed case 31732
ck(f(('address[]', 'uint56', 'bytes17'), D), T[A, i, by])  # fixed case 31733
ck(f(('address[]', 'uint56', 'bytes18'), D), T[A, i, by])  # fixed case 31734
ck(f(('address[]', 'uint56', 'bytes19'), D), T[A, i, by])  # fixed case 31735
ck(f(('address[]', 'uint56', 'bytes2'), D), T[A, i, by])  # fixed case 31736
ck(f(('address[]', 'uint56', 'bytes20'), D), T[A, i, by])  # fixed case 31737
ck(f(('address[]', 'uint56', 'bytes21'), D), T[A, i, by])  # fixed case 31738
ck(f(('address[]', 'uint56', 'bytes22'), D), T[A, i, by])  # fixed case 31739
ck(f(('address[]', 'uint56', 'bytes23'), D), T[A, i, by])  # fixed case 31740
ck(f(('address[]', 'uint56', 'bytes24'), D), T[A, i, by])  # fixed case 31741
ck(f(('address[]', 'uint56', 'bytes25'), D), T[A, i, by])  # fixed case 31742
ck(f(('address[]', 'uint56', 'bytes26'), D), T[A, i, by])  # fixed case 31743
ck(f(('address[]', 'uint56', 'bytes27'), D), T[A, i, by])  # fixed case 31744
ck(f(('address[]', 'uint56', 'bytes28'), D), T[A, i, by])  # fixed case 31745
ck(f(('address[]', 'uint56', 'bytes29'), D), T[A, i, by])  # fixed case 31746
ck(f(('address[]', 'uint56', 'bytes3'), D), T[A, i, by])  # fixed case 31747
ck(f(('address[]', 'uint56', 'bytes30'), D), T[A, i, by])  # fixed case 31748
ck(f(('address[]', 'uint56', 'bytes31'), D), T[A, i, by])  # fixed case 31749
ck(f(('address[]', 'uint56', 'bytes32'), D), T[A, i, by])  # fixed case 31750
ck(f(('address[]', 'uint56', 'bytes4'), D), T[A, i, by])  # fixed case 31751
ck(f(('address[]', 'uint56', 'bytes5'), D), T[A, i, by])  # fixed case 31752
ck(f(('address[]', 'uint56', 'bytes6'), D), T[A, i, by])  # fixed case 31753
ck(f(('address[]', 'uint56', 'bytes7'), D), T[A, i, by])  # fixed case 31754
ck(f(('address[]', 'uint56', 'bytes8'), D), T[A, i, by])  # fixed case 31755
ck(f(('address[]', 'uint56', 'bytes9'), D), T[A, i, by])  # fixed case 31756
ck(f(('address[]', 'uint56', 'int104'), D), T[A, i, i])  # fixed case 31757
ck(f(('address[]', 'uint56', 'int112'), D), T[A, i, i])  # fixed case 31758
ck(f(('address[]', 'uint56', 'int120'), D), T[A, i, i])  # fixed case 31759
ck(f(('address[]', 'uint56', 'int128'), D), T[A, i, i])  # fixed case 31760
ck(f(('address[]', 'uint56', 'int136'), D), T[A, i, i])  # fixed case 31761
ck(f(('address[]', 'uint56', 'int144'), D), T[A, i, i])  # fixed case 31762
ck(f(('address[]', 'uint56', 'int152'), D), T[A, i, i])  # fixed case 31763
ck(f(('address[]', 'uint56', 'int16'), D), T[A, i, i])  # fixed case 31764
ck(f(('address[]', 'uint56', 'int160'), D), T[A, i, i])  # fixed case 31765
ck(f(('address[]', 'uint56', 'int168'), D), T[A, i, i])  # fixed case 31766
ck(f(('address[]', 'uint56', 'int176'), D), T[A, i, i])  # fixed case 31767
ck(f(('address[]', 'uint56', 'int184'), D), T[A, i, i])  # fixed case 31768
ck(f(('address[]', 'uint56', 'int192'), D), T[A, i, i])  # fixed case 31769
ck(f(('address[]', 'uint56', 'int200'), D), T[A, i, i])  # fixed case 31770
ck(f(('address[]', 'uint56', 'int208'), D), T[A, i, i])  # fixed case 31771
ck(f(('address[]', 'uint56', 'int216'), D), T[A, i, i])  # fixed case 31772
ck(f(('address[]', 'uint56', 'int224'), D), T[A, i, i])  # fixed case 31773
ck(f(('address[]', 'uint56', 'int232'), D), T[A, i, i])  # fixed case 31774
ck(f(('address[]', 'uint56', 'int24'), D), T[A, i, i])  # fixed case 31775
ck(f(('address[]', 'uint56', 'int240'), D), T[A, i, i])  # fixed case 31776
ck(f(('address[]', 'uint56', 'int248'), D), T[A, i, i])  # fixed case 31777
ck(f(('address[]', 'uint56', 'int256'), D), T[A, i, i])  # fixed case 31778
ck(f(('address[]', 'uint56', 'int32'), D), T[A, i, i])  # fixed case 31779
ck(f(('address[]', 'uint56', 'int40'), D), T[A, i, i])  # fixed case 31780
ck(f(('address[]', 'uint56', 'int48'), D), T[A, i, i])  # fixed case 31781
ck(f(('address[]', 'uint56', 'int56'), D), T[A, i, i])  # fixed case 31782
ck(f(('address[]', 'uint56', 'int64'), D), T[A, i, i])  # fixed case 31783
ck(f(('address[]', 'uint56', 'int72'), D), T[A, i, i])  # fixed case 31784
ck(f(('address[]', 'uint56', 'int8'), D), T[A, i, i])  # fixed case 31785
ck(f(('address[]', 'uint56', 'int80'), D), T[A, i, i])  # fixed case 31786
ck(f(('address[]', 'uint56', 'int88'), D), T[A, i, i])  # fixed case 31787
ck(f(('address[]', 'uint56', 'int96'), D), T[A, i, i])  # fixed case 31788
ck(f(('address[]', 'uint56', 'string'), D), T[A, i, s])  # fixed case 31789
ck(f(('address[]', 'uint56', 'string[]'), D), T[A, i, A])  # fixed case 31790
ck(f(('address[]', 'uint56', 'uint104'), D), T[A, i, i])  # fixed case 31791
ck(f(('address[]', 'uint56', 'uint112'), D), T[A, i, i])  # fixed case 31792
ck(f(('address[]', 'uint56', 'uint120'), D), T[A, i, i])  # fixed case 31793
ck(f(('address[]', 'uint56', 'uint128'), D), T[A, i, i])  # fixed case 31794
ck(f(('address[]', 'uint56', 'uint136'), D), T[A, i, i])  # fixed case 31795
ck(f(('address[]', 'uint56', 'uint144'), D), T[A, i, i])  # fixed case 31796
ck(f(('address[]', 'uint56', 'uint152'), D), T[A, i, i])  # fixed case 31797
ck(f(('address[]', 'uint56', 'uint16'), D), T[A, i, i])  # fixed case 31798
ck(f(('address[]', 'uint56', 'uint160'), D), T[A, i, i])  # fixed case 31799
ck(f(('address[]', 'uint56', 'uint168'), D), T[A, i, i])  # fixed case 31800
ck(f(('address[]', 'uint56', 'uint176'), D), T[A, i, i])  # fixed case 31801
ck(f(('address[]', 'uint56', 'uint184'), D), T[A, i, i])  # fixed case 31802
ck(f(('address[]', 'uint56', 'uint192'), D), T[A, i, i])  # fixed case 31803
ck(f(('address[]', 'uint56', 'uint200'), D), T[A, i, i])  # fixed case 31804
ck(f(('address[]', 'uint56', 'uint208'), D), T[A, i, i])  # fixed case 31805
ck(f(('address[]', 'uint56', 'uint216'), D), T[A, i, i])  # fixed case 31806
ck(f(('address[]', 'uint56', 'uint224'), D), T[A, i, i])  # fixed case 31807
ck(f(('address[]', 'uint56', 'uint232'), D), T[A, i, i])  # fixed case 31808
ck(f(('address[]', 'uint56', 'uint24'), D), T[A, i, i])  # fixed case 31809
ck(f(('address[]', 'uint56', 'uint240'), D), T[A, i, i])  # fixed case 31810
ck(f(('address[]', 'uint56', 'uint248'), D), T[A, i, i])  # fixed case 31811
ck(f(('address[]', 'uint56', 'uint256'), D), T[A, i, i])  # fixed case 31812
ck(f(('address[]', 'uint56', 'uint32'), D), T[A, i, i])  # fixed case 31813
ck(f(('address[]', 'uint56', 'uint40'), D), T[A, i, i])  # fixed case 31814
ck(f(('address[]', 'uint56', 'uint48'), D), T[A, i, i])  # fixed case 31815
ck(f(('address[]', 'uint56', 'uint56'), D), T[A, i, i])  # fixed case 31816
ck(f(('address[]', 'uint56', 'uint64'), D), T[A, i, i])  # fixed case 31817
ck(f(('address[]', 'uint56', 'uint72'), D), T[A, i, i])  # fixed case 31818
ck(f(('address[]', 'uint56', 'uint8'), D), T[A, i, i])  # fixed case 31819
ck(f(('address[]', 'uint56', 'uint80'), D), T[A, i, i])  # fixed case 31820
ck(f(('address[]', 'uint56', 'uint88'), D), T[A, i, i])  # fixed case 31821
ck(f(('address[]', 'uint56', 'uint96'), D), T[A, i, i])  # fixed case 31822
ck(f(('address[]', 'uint56', '?'), D), T[A, i, A])  # fixed case 31823
ck(f(('address[]', 'uint64', 'address'), D), T[A, i, HA])  # fixed case 31824
ck(f(('address[]', 'uint64', 'address[]'), D), T[A, i, A])  # fixed case 31825
ck(f(('address[]', 'uint64', 'bool'), D), T[A, i, bo])  # fixed case 31826
ck(f(('address[]', 'uint64', 'bool[]'), D), T[A, i, A])  # fixed case 31827
ck(f(('address[]', 'uint64', 'bytes'), D), T[A, i, by])  # fixed case 31828
ck(f(('address[]', 'uint64', 'bytes1'), D), T[A, i, by])  # fixed case 31829
ck(f(('address[]', 'uint64', 'bytes10'), D), T[A, i, by])  # fixed case 31830
ck(f(('address[]', 'uint64', 'bytes11'), D), T[A, i, by])  # fixed case 31831
ck(f(('address[]', 'uint64', 'bytes12'), D), T[A, i, by])  # fixed case 31832
ck(f(('address[]', 'uint64', 'bytes13'), D), T[A, i, by])  # fixed case 31833
ck(f(('address[]', 'uint64', 'bytes14'), D), T[A, i, by])  # fixed case 31834
ck(f(('address[]', 'uint64', 'bytes15'), D), T[A, i, by])  # fixed case 31835
ck(f(('address[]', 'uint64', 'bytes16'), D), T[A, i, by])  # fixed case 31836
ck(f(('address[]', 'uint64', 'bytes17'), D), T[A, i, by])  # fixed case 31837
ck(f(('address[]', 'uint64', 'bytes18'), D), T[A, i, by])  # fixed case 31838
ck(f(('address[]', 'uint64', 'bytes19'), D), T[A, i, by])  # fixed case 31839
ck(f(('address[]', 'uint64', 'bytes2'), D), T[A, i, by])  # fixed case 31840
ck(f(('address[]', 'uint64', 'bytes20'), D), T[A, i, by])  # fixed case 31841
ck(f(('address[]', 'uint64', 'bytes21'), D), T[A, i, by])  # fixed case 31842
ck(f(('address[]', 'uint64', 'bytes22'), D), T[A, i, by])  # fixed case 31843
ck(f(('address[]', 'uint64', 'bytes23'), D), T[A, i, by])  # fixed case 31844
ck(f(('address[]', 'uint64', 'bytes24'), D), T[A, i, by])  # fixed case 31845
ck(f(('address[]', 'uint64', 'bytes25'), D), T[A, i, by])  # fixed case 31846
ck(f(('address[]', 'uint64', 'bytes26'), D), T[A, i, by])  # fixed case 31847
ck(f(('address[]', 'uint64', 'bytes27'), D), T[A, i, by])  # fixed case 31848
ck(f(('address[]', 'uint64', 'bytes28'), D), T[A, i, by])  # fixed case 31849
ck(f(('address[]', 'uint64', 'bytes29'), D), T[A, i, by])  # fixed case 31850
ck(f(('address[]', 'uint64', 'bytes3'), D), T[A, i, by])  # fixed case 31851
ck(f(('address[]', 'uint64', 'bytes30'), D), T[A, i, by])  # fixed case 31852
ck(f(('address[]', 'uint64', 'bytes31'), D), T[A, i, by])  # fixed case 31853
ck(f(('address[]', 'uint64', 'bytes32'), D), T[A, i, by])  # fixed case 31854
ck(f(('address[]', 'uint64', 'bytes4'), D), T[A, i, by])  # fixed case 31855
ck(f(('address[]', 'uint64', 'bytes5'), D), T[A, i, by])  # fixed case 31856
ck(f(('address[]', 'uint64', 'bytes6'), D), T[A, i, by])  # fixed case 31857
ck(f(('address[]', 'uint64', 'bytes7'), D), T[A, i, by])  # fixed case 31858
ck(f(('address[]', 'uint64', 'bytes8'), D), T[A, i, by])  # fixed case 31859
ck(f(('address[]', 'uint64', 'bytes9'), D), T[A, i, by])  # fixed case 31860
ck(f(('address[]', 'uint64', 'int104'), D), T[A, i, i])  # fixed case 31861
ck(f(('address[]', 'uint64', 'int112'), D), T[A, i, i])  # fixed case 31862
ck(f(('address[]', 'uint64', 'int120'), D), T[A, i, i])  # fixed case 31863
ck(f(('address[]', 'uint64', 'int128'), D), T[A, i, i])  # fixed case 31864
ck(f(('address[]', 'uint64', 'int136'), D), T[A, i, i])  # fixed case 31865
ck(f(('address[]', 'uint64', 'int144'), D), T[A, i, i])  # fixed case 31866
ck(f(('address[]', 'uint64', 'int152'), D), T[A, i, i])  # fixed case 31867
ck(f(('address[]', 'uint64', 'int16'), D), T[A, i, i])  # fixed case 31868
ck(f(('address[]', 'uint64', 'int160'), D), T[A, i, i])  # fixed case 31869
ck(f(('address[]', 'uint64', 'int168'), D), T[A, i, i])  # fixed case 31870
ck(f(('address[]', 'uint64', 'int176'), D), T[A, i, i])  # fixed case 31871
ck(f(('address[]', 'uint64', 'int184'), D), T[A, i, i])  # fixed case 31872
ck(f(('address[]', 'uint64', 'int192'), D), T[A, i, i])  # fixed case 31873
ck(f(('address[]', 'uint64', 'int200'), D), T[A, i, i])  # fixed case 31874
ck(f(('address[]', 'uint64', 'int208'), D), T[A, i, i])  # fixed case 31875
ck(f(('address[]', 'uint64', 'int216'), D), T[A, i, i])  # fixed case 31876
ck(f(('address[]', 'uint64', 'int224'), D), T[A, i, i])  # fixed case 31877
ck(f(('address[]', 'uint64', 'int232'), D), T[A, i, i])  # fixed case 31878
ck(f(('address[]', 'uint64', 'int24'), D), T[A, i, i])  # fixed case 31879
ck(f(('address[]', 'uint64', 'int240'), D), T[A, i, i])  # fixed case 31880
ck(f(('address[]', 'uint64', 'int248'), D), T[A, i, i])  # fixed case 31881
ck(f(('address[]', 'uint64', 'int256'), D), T[A, i, i])  # fixed case 31882
ck(f(('address[]', 'uint64', 'int32'), D), T[A, i, i])  # fixed case 31883
ck(f(('address[]', 'uint64', 'int40'), D), T[A, i, i])  # fixed case 31884
ck(f(('address[]', 'uint64', 'int48'), D), T[A, i, i])  # fixed case 31885
ck(f(('address[]', 'uint64', 'int56'), D), T[A, i, i])  # fixed case 31886
ck(f(('address[]', 'uint64', 'int64'), D), T[A, i, i])  # fixed case 31887
ck(f(('address[]', 'uint64', 'int72'), D), T[A, i, i])  # fixed case 31888
ck(f(('address[]', 'uint64', 'int8'), D), T[A, i, i])  # fixed case 31889
ck(f(('address[]', 'uint64', 'int80'), D), T[A, i, i])  # fixed case 31890
ck(f(('address[]', 'uint64', 'int88'), D), T[A, i, i])  # fixed case 31891
ck(f(('address[]', 'uint64', 'int96'), D), T[A, i, i])  # fixed case 31892
ck(f(('address[]', 'uint64', 'string'), D), T[A, i, s])  # fixed case 31893
ck(f(('address[]', 'uint64', 'string[]'), D), T[A, i, A])  # fixed case 31894
ck(f(('address[]', 'uint64', 'uint104'), D), T[A, i, i])  # fixed case 31895
ck(f(('address[]', 'uint64', 'uint112'), D), T[A, i, i])  # fixed case 31896
ck(f(('address[]', 'uint64', 'uint120'), D), T[A, i, i])  # fixed case 31897
ck(f(('address[]', 'uint64', 'uint128'), D), T[A, i, i])  # fixed case 31898
ck(f(('address[]', 'uint64', 'uint136'), D), T[A, i, i])  # fixed case 31899
ck(f(('address[]', 'uint64', 'uint144'), D), T[A, i, i])  # fixed case 31900
ck(f(('address[]', 'uint64', 'uint152'), D), T[A, i, i])  # fixed case 31901
ck(f(('address[]', 'uint64', 'uint16'), D), T[A, i, i])  # fixed case 31902
ck(f(('address[]', 'uint64', 'uint160'), D), T[A, i, i])  # fixed case 31903
ck(f(('address[]', 'uint64', 'uint168'), D), T[A, i, i])  # fixed case 31904
ck(f(('address[]', 'uint64', 'uint176'), D), T[A, i, i])  # fixed case 31905
ck(f(('address[]', 'uint64', 'uint184'), D), T[A, i, i])  # fixed case 31906
ck(f(('address[]', 'uint64', 'uint192'), D), T[A, i, i])  # fixed case 31907
ck(f(('address[]', 'uint64', 'uint200'), D), T[A, i, i])  # fixed case 31908
ck(f(('address[]', 'uint64', 'uint208'), D), T[A, i, i])  # fixed case 31909
ck(f(('address[]', 'uint64', 'uint216'), D), T[A, i, i])  # fixed case 31910
ck(f(('address[]', 'uint64', 'uint224'), D), T[A, i, i])  # fixed case 31911
ck(f(('address[]', 'uint64', 'uint232'), D), T[A, i, i])  # fixed case 31912
ck(f(('address[]', 'uint64', 'uint24'), D), T[A, i, i])  # fixed case 31913
ck(f(('address[]', 'uint64', 'uint240'), D), T[A, i, i])  # fixed case 31914
ck(f(('address[]', 'uint64', 'uint248'), D), T[A, i, i])  # fixed case 31915
ck(f(('address[]', 'uint64', 'uint256'), D), T[A, i, i])  # fixed case 31916
ck(f(('address[]', 'uint64', 'uint32'), D), T[A, i, i])  # fixed case 31917
ck(f(('address[]', 'uint64', 'uint40'), D), T[A, i, i])  # fixed case 31918
ck(f(('address[]', 'uint64', 'uint48'), D), T[A, i, i])  # fixed case 31919
ck(f(('address[]', 'uint64', 'uint56'), D), T[A, i, i])  # fixed case 31920
ck(f(('address[]', 'uint64', 'uint64'), D), T[A, i, i])  # fixed case 31921
ck(f(('address[]', 'uint64', 'uint72'), D), T[A, i, i])  # fixed case 31922
ck(f(('address[]', 'uint64', 'uint8'), D), T[A, i, i])  # fixed case 31923
ck(f(('address[]', 'uint64', 'uint80'), D), T[A, i, i])  # fixed case 31924
ck(f(('address[]', 'uint64', 'uint88'), D), T[A, i, i])  # fixed case 31925
ck(f(('address[]', 'uint64', 'uint96'), D), T[A, i, i])  # fixed case 31926
ck(f(('address[]', 'uint64', '?'), D), T[A, i, A])  # fixed case 31927
ck(f(('address[]', 'uint72', 'address'), D), T[A, i, HA])  # fixed case 31928
ck(f(('address[]', 'uint72', 'address[]'), D), T[A, i, A])  # fixed case 31929
ck(f(('address[]', 'uint72', 'bool'), D), T[A, i, bo])  # fixed case 31930
ck(f(('address[]', 'uint72', 'bool[]'), D), T[A, i, A])  # fixed case 31931
ck(f(('address[]', 'uint72', 'bytes'), D), T[A, i, by])  # fixed case 31932
ck(f(('address[]', 'uint72', 'bytes1'), D), T[A, i, by])  # fixed case 31933
ck(f(('address[]', 'uint72', 'bytes10'), D), T[A, i, by])  # fixed case 31934
ck(f(('address[]', 'uint72', 'bytes11'), D), T[A, i, by])  # fixed case 31935
ck(f(('address[]', 'uint72', 'bytes12'), D), T[A, i, by])  # fixed case 31936
ck(f(('address[]', 'uint72', 'bytes13'), D), T[A, i, by])  # fixed case 31937
ck(f(('address[]', 'uint72', 'bytes14'), D), T[A, i, by])  # fixed case 31938
ck(f(('address[]', 'uint72', 'bytes15'), D), T[A, i, by])  # fixed case 31939
ck(f(('address[]', 'uint72', 'bytes16'), D), T[A, i, by])  # fixed case 31940
ck(f(('address[]', 'uint72', 'bytes17'), D), T[A, i, by])  # fixed case 31941
ck(f(('address[]', 'uint72', 'bytes18'), D), T[A, i, by])  # fixed case 31942
ck(f(('address[]', 'uint72', 'bytes19'), D), T[A, i, by])  # fixed case 31943
ck(f(('address[]', 'uint72', 'bytes2'), D), T[A, i, by])  # fixed case 31944
ck(f(('address[]', 'uint72', 'bytes20'), D), T[A, i, by])  # fixed case 31945
ck(f(('address[]', 'uint72', 'bytes21'), D), T[A, i, by])  # fixed case 31946
ck(f(('address[]', 'uint72', 'bytes22'), D), T[A, i, by])  # fixed case 31947
ck(f(('address[]', 'uint72', 'bytes23'), D), T[A, i, by])  # fixed case 31948
ck(f(('address[]', 'uint72', 'bytes24'), D), T[A, i, by])  # fixed case 31949
ck(f(('address[]', 'uint72', 'bytes25'), D), T[A, i, by])  # fixed case 31950
ck(f(('address[]', 'uint72', 'bytes26'), D), T[A, i, by])  # fixed case 31951
ck(f(('address[]', 'uint72', 'bytes27'), D), T[A, i, by])  # fixed case 31952
ck(f(('address[]', 'uint72', 'bytes28'), D), T[A, i, by])  # fixed case 31953
ck(f(('address[]', 'uint72', 'bytes29'), D), T[A, i, by])  # fixed case 31954
ck(f(('address[]', 'uint72', 'bytes3'), D), T[A, i, by])  # fixed case 31955
ck(f(('address[]', 'uint72', 'bytes30'), D), T[A, i, by])  # fixed case 31956
ck(f(('address[]', 'uint72', 'bytes31'), D), T[A, i, by])  # fixed case 31957
ck(f(('address[]', 'uint72', 'bytes32'), D), T[A, i, by])  # fixed case 31958
ck(f(('address[]', 'uint72', 'bytes4'), D), T[A, i, by])  # fixed case 31959
ck(f(('address[]', 'uint72', 'bytes5'), D), T[A, i, by])  # fixed case 31960
ck(f(('address[]', 'uint72', 'bytes6'), D), T[A, i, by])  # fixed case 31961
ck(f(('address[]', 'uint72', 'bytes7'), D), T[A, i, by])  # fixed case 31962
ck(f(('address[]', 'uint72', 'bytes8'), D), T[A, i, by])  # fixed case 31963
ck(f(('address[]', 'uint72', 'bytes9'), D), T[A, i, by])  # fixed case 31964
ck(f(('address[]', 'uint72', 'int104'), D), T[A, i, i])  # fixed case 31965
ck(f(('address[]', 'uint72', 'int112'), D), T[A, i, i])  # fixed case 31966
ck(f(('address[]', 'uint72', 'int120'), D), T[A, i, i])  # fixed case 31967
ck(f(('address[]', 'uint72', 'int128'), D), T[A, i, i])  # fixed case 31968
ck(f(('address[]', 'uint72', 'int136'), D), T[A, i, i])  # fixed case 31969
ck(f(('address[]', 'uint72', 'int144'), D), T[A, i, i])  # fixed case 31970
ck(f(('address[]', 'uint72', 'int152'), D), T[A, i, i])  # fixed case 31971
ck(f(('address[]', 'uint72', 'int16'), D), T[A, i, i])  # fixed case 31972
ck(f(('address[]', 'uint72', 'int160'), D), T[A, i, i])  # fixed case 31973
ck(f(('address[]', 'uint72', 'int168'), D), T[A, i, i])  # fixed case 31974
ck(f(('address[]', 'uint72', 'int176'), D), T[A, i, i])  # fixed case 31975
ck(f(('address[]', 'uint72', 'int184'), D), T[A, i, i])  # fixed case 31976
ck(f(('address[]', 'uint72', 'int192'), D), T[A, i, i])  # fixed case 31977
ck(f(('address[]', 'uint72', 'int200'), D), T[A, i, i])  # fixed case 31978
ck(f(('address[]', 'uint72', 'int208'), D), T[A, i, i])  # fixed case 31979
ck(f(('address[]', 'uint72', 'int216'), D), T[A, i, i])  # fixed case 31980
ck(f(('address[]', 'uint72', 'int224'), D), T[A, i, i])  # fixed case 31981
ck(f(('address[]', 'uint72', 'int232'), D), T[A, i, i])  # fixed case 31982
ck(f(('address[]', 'uint72', 'int24'), D), T[A, i, i])  # fixed case 31983
ck(f(('address[]', 'uint72', 'int240'), D), T[A, i, i])  # fixed case 31984
ck(f(('address[]', 'uint72', 'int248'), D), T[A, i, i])  # fixed case 31985
ck(f(('address[]', 'uint72', 'int256'), D), T[A, i, i])  # fixed case 31986
ck(f(('address[]', 'uint72', 'int32'), D), T[A, i, i])  # fixed case 31987
ck(f(('address[]', 'uint72', 'int40'), D), T[A, i, i])  # fixed case 31988
ck(f(('address[]', 'uint72', 'int48'), D), T[A, i, i])  # fixed case 31989
ck(f(('address[]', 'uint72', 'int56'), D), T[A, i, i])  # fixed case 31990
ck(f(('address[]', 'uint72', 'int64'), D), T[A, i, i])  # fixed case 31991
ck(f(('address[]', 'uint72', 'int72'), D), T[A, i, i])  # fixed case 31992
ck(f(('address[]', 'uint72', 'int8'), D), T[A, i, i])  # fixed case 31993
ck(f(('address[]', 'uint72', 'int80'), D), T[A, i, i])  # fixed case 31994
ck(f(('address[]', 'uint72', 'int88'), D), T[A, i, i])  # fixed case 31995
ck(f(('address[]', 'uint72', 'int96'), D), T[A, i, i])  # fixed case 31996
ck(f(('address[]', 'uint72', 'string'), D), T[A, i, s])  # fixed case 31997
ck(f(('address[]', 'uint72', 'string[]'), D), T[A, i, A])  # fixed case 31998
ck(f(('address[]', 'uint72', 'uint104'), D), T[A, i, i])  # fixed case 31999
ck(f(('address[]', 'uint72', 'uint112'), D), T[A, i, i])  # fixed case 32000
ck(f(('address[]', 'uint72', 'uint120'), D), T[A, i, i])  # fixed case 32001
ck(f(('address[]', 'uint72', 'uint128'), D), T[A, i, i])  # fixed case 32002
ck(f(('address[]', 'uint72', 'uint136'), D), T[A, i, i])  # fixed case 32003
ck(f(('address[]', 'uint72', 'uint144'), D), T[A, i, i])  # fixed case 32004
ck(f(('address[]', 'uint72', 'uint152'), D), T[A, i, i])  # fixed case 32005
ck(f(('address[]', 'uint72', 'uint16'), D), T[A, i, i])  # fixed case 32006
ck(f(('address[]', 'uint72', 'uint160'), D), T[A, i, i])  # fixed case 32007
ck(f(('address[]', 'uint72', 'uint168'), D), T[A, i, i])  # fixed case 32008
ck(f(('address[]', 'uint72', 'uint176'), D), T[A, i, i])  # fixed case 32009
ck(f(('address[]', 'uint72', 'uint184'), D), T[A, i, i])  # fixed case 32010
ck(f(('address[]', 'uint72', 'uint192'), D), T[A, i, i])  # fixed case 32011
ck(f(('address[]', 'uint72', 'uint200'), D), T[A, i, i])  # fixed case 32012
ck(f(('address[]', 'uint72', 'uint208'), D), T[A, i, i])  # fixed case 32013
ck(f(('address[]', 'uint72', 'uint216'), D), T[A, i, i])  # fixed case 32014
ck(f(('address[]', 'uint72', 'uint224'), D), T[A, i, i])  # fixed case 32015
ck(f(('address[]', 'uint72', 'uint232'), D), T[A, i, i])  # fixed case 32016
ck(f(('address[]', 'uint72', 'uint24'), D), T[A, i, i])  # fixed case 32017
ck(f(('address[]', 'uint72', 'uint240'), D), T[A, i, i])  # fixed case 32018
ck(f(('address[]', 'uint72', 'uint248'), D), T[A, i, i])  # fixed case 32019
ck(f(('address[]', 'uint72', 'uint256'), D), T[A, i, i])  # fixed case 32020
ck(f(('address[]', 'uint72', 'uint32'), D), T[A, i, i])  # fixed case 32021
ck(f(('address[]', 'uint72', 'uint40'), D), T[A, i, i])  # fixed case 32022
ck(f(('address[]', 'uint72', 'uint48'), D), T[A, i, i])  # fixed case 32023
ck(f(('address[]', 'uint72', 'uint56'), D), T[A, i, i])  # fixed case 32024
ck(f(('address[]', 'uint72', 'uint64'), D), T[A, i, i])  # fixed case 32025
ck(f(('address[]', 'uint72', 'uint72'), D), T[A, i, i])  # fixed case 32026
ck(f(('address[]', 'uint72', 'uint8'), D), T[A, i, i])  # fixed case 32027
ck(f(('address[]', 'uint72', 'uint80'), D), T[A, i, i])  # fixed case 32028
ck(f(('address[]', 'uint72', 'uint88'), D), T[A, i, i])  # fixed case 32029
ck(f(('address[]', 'uint72', 'uint96'), D), T[A, i, i])  # fixed case 32030
ck(f(('address[]', 'uint72', '?'), D), T[A, i, A])  # fixed case 32031
ck(f(('address[]', 'uint8', 'address'), D), T[A, i, HA])  # fixed case 32032
ck(f(('address[]', 'uint8', 'address[]'), D), T[A, i, A])  # fixed case 32033
ck(f(('address[]', 'uint8', 'bool'), D), T[A, i, bo])  # fixed case 32034
ck(f(('address[]', 'uint8', 'bool[]'), D), T[A, i, A])  # fixed case 32035
ck(f(('address[]', 'uint8', 'bytes'), D), T[A, i, by])  # fixed case 32036
ck(f(('address[]', 'uint8', 'bytes1'), D), T[A, i, by])  # fixed case 32037
ck(f(('address[]', 'uint8', 'bytes10'), D), T[A, i, by])  # fixed case 32038
ck(f(('address[]', 'uint8', 'bytes11'), D), T[A, i, by])  # fixed case 32039
ck(f(('address[]', 'uint8', 'bytes12'), D), T[A, i, by])  # fixed case 32040
ck(f(('address[]', 'uint8', 'bytes13'), D), T[A, i, by])  # fixed case 32041
ck(f(('address[]', 'uint8', 'bytes14'), D), T[A, i, by])  # fixed case 32042
ck(f(('address[]', 'uint8', 'bytes15'), D), T[A, i, by])  # fixed case 32043
ck(f(('address[]', 'uint8', 'bytes16'), D), T[A, i, by])  # fixed case 32044
ck(f(('address[]', 'uint8', 'bytes17'), D), T[A, i, by])  # fixed case 32045
ck(f(('address[]', 'uint8', 'bytes18'), D), T[A, i, by])  # fixed case 32046
ck(f(('address[]', 'uint8', 'bytes19'), D), T[A, i, by])  # fixed case 32047
ck(f(('address[]', 'uint8', 'bytes2'), D), T[A, i, by])  # fixed case 32048
ck(f(('address[]', 'uint8', 'bytes20'), D), T[A, i, by])  # fixed case 32049
ck(f(('address[]', 'uint8', 'bytes21'), D), T[A, i, by])  # fixed case 32050
ck(f(('address[]', 'uint8', 'bytes22'), D), T[A, i, by])  # fixed case 32051
ck(f(('address[]', 'uint8', 'bytes23'), D), T[A, i, by])  # fixed case 32052
ck(f(('address[]', 'uint8', 'bytes24'), D), T[A, i, by])  # fixed case 32053
ck(f(('address[]', 'uint8', 'bytes25'), D), T[A, i, by])  # fixed case 32054
ck(f(('address[]', 'uint8', 'bytes26'), D), T[A, i, by])  # fixed case 32055
ck(f(('address[]', 'uint8', 'bytes27'), D), T[A, i, by])  # fixed case 32056
ck(f(('address[]', 'uint8', 'bytes28'), D), T[A, i, by])  # fixed case 32057
ck(f(('address[]', 'uint8', 'bytes29'), D), T[A, i, by])  # fixed case 32058
ck(f(('address[]', 'uint8', 'bytes3'), D), T[A, i, by])  # fixed case 32059
ck(f(('address[]', 'uint8', 'bytes30'), D), T[A, i, by])  # fixed case 32060
ck(f(('address[]', 'uint8', 'bytes31'), D), T[A, i, by])  # fixed case 32061
ck(f(('address[]', 'uint8', 'bytes32'), D), T[A, i, by])  # fixed case 32062
ck(f(('address[]', 'uint8', 'bytes4'), D), T[A, i, by])  # fixed case 32063
ck(f(('address[]', 'uint8', 'bytes5'), D), T[A, i, by])  # fixed case 32064
ck(f(('address[]', 'uint8', 'bytes6'), D), T[A, i, by])  # fixed case 32065
ck(f(('address[]', 'uint8', 'bytes7'), D), T[A, i, by])  # fixed case 32066
ck(f(('address[]', 'uint8', 'bytes8'), D), T[A, i, by])  # fixed case 32067
ck(f(('address[]', 'uint8', 'bytes9'), D), T[A, i, by])  # fixed case 32068
ck(f(('address[]', 'uint8', 'int104'), D), T[A, i, i])  # fixed case 32069
ck(f(('address[]', 'uint8', 'int112'), D), T[A, i, i])  # fixed case 32070
ck(f(('address[]', 'uint8', 'int120'), D), T[A, i, i])  # fixed case 32071
ck(f(('address[]', 'uint8', 'int128'), D), T[A, i, i])  # fixed case 32072
ck(f(('address[]', 'uint8', 'int136'), D), T[A, i, i])  # fixed case 32073
ck(f(('address[]', 'uint8', 'int144'), D), T[A, i, i])  # fixed case 32074
ck(f(('address[]', 'uint8', 'int152'), D), T[A, i, i])  # fixed case 32075
ck(f(('address[]', 'uint8', 'int16'), D), T[A, i, i])  # fixed case 32076
ck(f(('address[]', 'uint8', 'int160'), D), T[A, i, i])  # fixed case 32077
ck(f(('address[]', 'uint8', 'int168'), D), T[A, i, i])  # fixed case 32078
ck(f(('address[]', 'uint8', 'int176'), D), T[A, i, i])  # fixed case 32079
ck(f(('address[]', 'uint8', 'int184'), D), T[A, i, i])  # fixed case 32080
ck(f(('address[]', 'uint8', 'int192'), D), T[A, i, i])  # fixed case 32081
ck(f(('address[]', 'uint8', 'int200'), D), T[A, i, i])  # fixed case 32082
ck(f(('address[]', 'uint8', 'int208'), D), T[A, i, i])  # fixed case 32083
ck(f(('address[]', 'uint8', 'int216'), D), T[A, i, i])  # fixed case 32084
ck(f(('address[]', 'uint8', 'int224'), D), T[A, i, i])  # fixed case 32085
ck(f(('address[]', 'uint8', 'int232'), D), T[A, i, i])  # fixed case 32086
ck(f(('address[]', 'uint8', 'int24'), D), T[A, i, i])  # fixed case 32087
ck(f(('address[]', 'uint8', 'int240'), D), T[A, i, i])  # fixed case 32088
ck(f(('address[]', 'uint8', 'int248'), D), T[A, i, i])  # fixed case 32089
ck(f(('address[]', 'uint8', 'int256'), D), T[A, i, i])  # fixed case 32090
ck(f(('address[]', 'uint8', 'int32'), D), T[A, i, i])  # fixed case 32091
ck(f(('address[]', 'uint8', 'int40'), D), T[A, i, i])  # fixed case 32092
ck(f(('address[]', 'uint8', 'int48'), D), T[A, i, i])  # fixed case 32093
ck(f(('address[]', 'uint8', 'int56'), D), T[A, i, i])  # fixed case 32094
ck(f(('address[]', 'uint8', 'int64'), D), T[A, i, i])  # fixed case 32095
ck(f(('address[]', 'uint8', 'int72'), D), T[A, i, i])  # fixed case 32096
ck(f(('address[]', 'uint8', 'int8'), D), T[A, i, i])  # fixed case 32097
ck(f(('address[]', 'uint8', 'int80'), D), T[A, i, i])  # fixed case 32098
ck(f(('address[]', 'uint8', 'int88'), D), T[A, i, i])  # fixed case 32099
ck(f(('address[]', 'uint8', 'int96'), D), T[A, i, i])  # fixed case 32100
ck(f(('address[]', 'uint8', 'string'), D), T[A, i, s])  # fixed case 32101
ck(f(('address[]', 'uint8', 'string[]'), D), T[A, i, A])  # fixed case 32102
ck(f(('address[]', 'uint8', 'uint104'), D), T[A, i, i])  # fixed case 32103
ck(f(('address[]', 'uint8', 'uint112'), D), T[A, i, i])  # fixed case 32104
ck(f(('address[]', 'uint8', 'uint120'), D), T[A, i, i])  # fixed case 32105
ck(f(('address[]', 'uint8', 'uint128'), D), T[A, i, i])  # fixed case 32106
ck(f(('address[]', 'uint8', 'uint136'), D), T[A, i, i])  # fixed case 32107
ck(f(('address[]', 'uint8', 'uint144'), D), T[A, i, i])  # fixed case 32108
ck(f(('address[]', 'uint8', 'uint152'), D), T[A, i, i])  # fixed case 32109
ck(f(('address[]', 'uint8', 'uint16'), D), T[A, i, i])  # fixed case 32110
ck(f(('address[]', 'uint8', 'uint160'), D), T[A, i, i])  # fixed case 32111
ck(f(('address[]', 'uint8', 'uint168'), D), T[A, i, i])  # fixed case 32112
ck(f(('address[]', 'uint8', 'uint176'), D), T[A, i, i])  # fixed case 32113
ck(f(('address[]', 'uint8', 'uint184'), D), T[A, i, i])  # fixed case 32114
ck(f(('address[]', 'uint8', 'uint192'), D), T[A, i, i])  # fixed case 32115
ck(f(('address[]', 'uint8', 'uint200'), D), T[A, i, i])  # fixed case 32116
ck(f(('address[]', 'uint8', 'uint208'), D), T[A, i, i])  # fixed case 32117
ck(f(('address[]', 'uint8', 'uint216'), D), T[A, i, i])  # fixed case 32118
ck(f(('address[]', 'uint8', 'uint224'), D), T[A, i, i])  # fixed case 32119
ck(f(('address[]', 'uint8', 'uint232'), D), T[A, i, i])  # fixed case 32120
ck(f(('address[]', 'uint8', 'uint24'), D), T[A, i, i])  # fixed case 32121
ck(f(('address[]', 'uint8', 'uint240'), D), T[A, i, i])  # fixed case 32122
ck(f(('address[]', 'uint8', 'uint248'), D), T[A, i, i])  # fixed case 32123
ck(f(('address[]', 'uint8', 'uint256'), D), T[A, i, i])  # fixed case 32124
ck(f(('address[]', 'uint8', 'uint32'), D), T[A, i, i])  # fixed case 32125
ck(f(('address[]', 'uint8', 'uint40'), D), T[A, i, i])  # fixed case 32126
ck(f(('address[]', 'uint8', 'uint48'), D), T[A, i, i])  # fixed case 32127
ck(f(('address[]', 'uint8', 'uint56'), D), T[A, i, i])  # fixed case 32128
ck(f(('address[]', 'uint8', 'uint64'), D), T[A, i, i])  # fixed case 32129
ck(f(('address[]', 'uint8', 'uint72'), D), T[A, i, i])  # fixed case 32130
ck(f(('address[]', 'uint8', 'uint8'), D), T[A, i, i])  # fixed case 32131
ck(f(('address[]', 'uint8', 'uint80'), D), T[A, i, i])  # fixed case 32132
ck(f(('address[]', 'uint8', 'uint88'), D), T[A, i, i])  # fixed case 32133
ck(f(('address[]', 'uint8', 'uint96'), D), T[A, i, i])  # fixed case 32134
ck(f(('address[]', 'uint8', '?'), D), T[A, i, A])  # fixed case 32135
ck(f(('address[]', 'uint80', 'address'), D), T[A, i, HA])  # fixed case 32136
ck(f(('address[]', 'uint80', 'address[]'), D), T[A, i, A])  # fixed case 32137
ck(f(('address[]', 'uint80', 'bool'), D), T[A, i, bo])  # fixed case 32138
ck(f(('address[]', 'uint80', 'bool[]'), D), T[A, i, A])  # fixed case 32139
ck(f(('address[]', 'uint80', 'bytes'), D), T[A, i, by])  # fixed case 32140
ck(f(('address[]', 'uint80', 'bytes1'), D), T[A, i, by])  # fixed case 32141
ck(f(('address[]', 'uint80', 'bytes10'), D), T[A, i, by])  # fixed case 32142
ck(f(('address[]', 'uint80', 'bytes11'), D), T[A, i, by])  # fixed case 32143
ck(f(('address[]', 'uint80', 'bytes12'), D), T[A, i, by])  # fixed case 32144
ck(f(('address[]', 'uint80', 'bytes13'), D), T[A, i, by])  # fixed case 32145
ck(f(('address[]', 'uint80', 'bytes14'), D), T[A, i, by])  # fixed case 32146
ck(f(('address[]', 'uint80', 'bytes15'), D), T[A, i, by])  # fixed case 32147
ck(f(('address[]', 'uint80', 'bytes16'), D), T[A, i, by])  # fixed case 32148
ck(f(('address[]', 'uint80', 'bytes17'), D), T[A, i, by])  # fixed case 32149
ck(f(('address[]', 'uint80', 'bytes18'), D), T[A, i, by])  # fixed case 32150
ck(f(('address[]', 'uint80', 'bytes19'), D), T[A, i, by])  # fixed case 32151
ck(f(('address[]', 'uint80', 'bytes2'), D), T[A, i, by])  # fixed case 32152
ck(f(('address[]', 'uint80', 'bytes20'), D), T[A, i, by])  # fixed case 32153
ck(f(('address[]', 'uint80', 'bytes21'), D), T[A, i, by])  # fixed case 32154
ck(f(('address[]', 'uint80', 'bytes22'), D), T[A, i, by])  # fixed case 32155
ck(f(('address[]', 'uint80', 'bytes23'), D), T[A, i, by])  # fixed case 32156
ck(f(('address[]', 'uint80', 'bytes24'), D), T[A, i, by])  # fixed case 32157
ck(f(('address[]', 'uint80', 'bytes25'), D), T[A, i, by])  # fixed case 32158
ck(f(('address[]', 'uint80', 'bytes26'), D), T[A, i, by])  # fixed case 32159
ck(f(('address[]', 'uint80', 'bytes27'), D), T[A, i, by])  # fixed case 32160
ck(f(('address[]', 'uint80', 'bytes28'), D), T[A, i, by])  # fixed case 32161
ck(f(('address[]', 'uint80', 'bytes29'), D), T[A, i, by])  # fixed case 32162
ck(f(('address[]', 'uint80', 'bytes3'), D), T[A, i, by])  # fixed case 32163
ck(f(('address[]', 'uint80', 'bytes30'), D), T[A, i, by])  # fixed case 32164
ck(f(('address[]', 'uint80', 'bytes31'), D), T[A, i, by])  # fixed case 32165
ck(f(('address[]', 'uint80', 'bytes32'), D), T[A, i, by])  # fixed case 32166
ck(f(('address[]', 'uint80', 'bytes4'), D), T[A, i, by])  # fixed case 32167
ck(f(('address[]', 'uint80', 'bytes5'), D), T[A, i, by])  # fixed case 32168
ck(f(('address[]', 'uint80', 'bytes6'), D), T[A, i, by])  # fixed case 32169
ck(f(('address[]', 'uint80', 'bytes7'), D), T[A, i, by])  # fixed case 32170
ck(f(('address[]', 'uint80', 'bytes8'), D), T[A, i, by])  # fixed case 32171
ck(f(('address[]', 'uint80', 'bytes9'), D), T[A, i, by])  # fixed case 32172
ck(f(('address[]', 'uint80', 'int104'), D), T[A, i, i])  # fixed case 32173
ck(f(('address[]', 'uint80', 'int112'), D), T[A, i, i])  # fixed case 32174
ck(f(('address[]', 'uint80', 'int120'), D), T[A, i, i])  # fixed case 32175
ck(f(('address[]', 'uint80', 'int128'), D), T[A, i, i])  # fixed case 32176
ck(f(('address[]', 'uint80', 'int136'), D), T[A, i, i])  # fixed case 32177
ck(f(('address[]', 'uint80', 'int144'), D), T[A, i, i])  # fixed case 32178
ck(f(('address[]', 'uint80', 'int152'), D), T[A, i, i])  # fixed case 32179
ck(f(('address[]', 'uint80', 'int16'), D), T[A, i, i])  # fixed case 32180
ck(f(('address[]', 'uint80', 'int160'), D), T[A, i, i])  # fixed case 32181
ck(f(('address[]', 'uint80', 'int168'), D), T[A, i, i])  # fixed case 32182
ck(f(('address[]', 'uint80', 'int176'), D), T[A, i, i])  # fixed case 32183
ck(f(('address[]', 'uint80', 'int184'), D), T[A, i, i])  # fixed case 32184
ck(f(('address[]', 'uint80', 'int192'), D), T[A, i, i])  # fixed case 32185
ck(f(('address[]', 'uint80', 'int200'), D), T[A, i, i])  # fixed case 32186
ck(f(('address[]', 'uint80', 'int208'), D), T[A, i, i])  # fixed case 32187
ck(f(('address[]', 'uint80', 'int216'), D), T[A, i, i])  # fixed case 32188
ck(f(('address[]', 'uint80', 'int224'), D), T[A, i, i])  # fixed case 32189
ck(f(('address[]', 'uint80', 'int232'), D), T[A, i, i])  # fixed case 32190
ck(f(('address[]', 'uint80', 'int24'), D), T[A, i, i])  # fixed case 32191
ck(f(('address[]', 'uint80', 'int240'), D), T[A, i, i])  # fixed case 32192
ck(f(('address[]', 'uint80', 'int248'), D), T[A, i, i])  # fixed case 32193
ck(f(('address[]', 'uint80', 'int256'), D), T[A, i, i])  # fixed case 32194
ck(f(('address[]', 'uint80', 'int32'), D), T[A, i, i])  # fixed case 32195
ck(f(('address[]', 'uint80', 'int40'), D), T[A, i, i])  # fixed case 32196
ck(f(('address[]', 'uint80', 'int48'), D), T[A, i, i])  # fixed case 32197
ck(f(('address[]', 'uint80', 'int56'), D), T[A, i, i])  # fixed case 32198
ck(f(('address[]', 'uint80', 'int64'), D), T[A, i, i])  # fixed case 32199
ck(f(('address[]', 'uint80', 'int72'), D), T[A, i, i])  # fixed case 32200
ck(f(('address[]', 'uint80', 'int8'), D), T[A, i, i])  # fixed case 32201
ck(f(('address[]', 'uint80', 'int80'), D), T[A, i, i])  # fixed case 32202
ck(f(('address[]', 'uint80', 'int88'), D), T[A, i, i])  # fixed case 32203
ck(f(('address[]', 'uint80', 'int96'), D), T[A, i, i])  # fixed case 32204
ck(f(('address[]', 'uint80', 'string'), D), T[A, i, s])  # fixed case 32205
ck(f(('address[]', 'uint80', 'string[]'), D), T[A, i, A])  # fixed case 32206
ck(f(('address[]', 'uint80', 'uint104'), D), T[A, i, i])  # fixed case 32207
ck(f(('address[]', 'uint80', 'uint112'), D), T[A, i, i])  # fixed case 32208
ck(f(('address[]', 'uint80', 'uint120'), D), T[A, i, i])  # fixed case 32209
ck(f(('address[]', 'uint80', 'uint128'), D), T[A, i, i])  # fixed case 32210
ck(f(('address[]', 'uint80', 'uint136'), D), T[A, i, i])  # fixed case 32211
ck(f(('address[]', 'uint80', 'uint144'), D), T[A, i, i])  # fixed case 32212
ck(f(('address[]', 'uint80', 'uint152'), D), T[A, i, i])  # fixed case 32213
ck(f(('address[]', 'uint80', 'uint16'), D), T[A, i, i])  # fixed case 32214
ck(f(('address[]', 'uint80', 'uint160'), D), T[A, i, i])  # fixed case 32215
ck(f(('address[]', 'uint80', 'uint168'), D), T[A, i, i])  # fixed case 32216
ck(f(('address[]', 'uint80', 'uint176'), D), T[A, i, i])  # fixed case 32217
ck(f(('address[]', 'uint80', 'uint184'), D), T[A, i, i])  # fixed case 32218
ck(f(('address[]', 'uint80', 'uint192'), D), T[A, i, i])  # fixed case 32219
ck(f(('address[]', 'uint80', 'uint200'), D), T[A, i, i])  # fixed case 32220
ck(f(('address[]', 'uint80', 'uint208'), D), T[A, i, i])  # fixed case 32221
ck(f(('address[]', 'uint80', 'uint216'), D), T[A, i, i])  # fixed case 32222
ck(f(('address[]', 'uint80', 'uint224'), D), T[A, i, i])  # fixed case 32223
ck(f(('address[]', 'uint80', 'uint232'), D), T[A, i, i])  # fixed case 32224
ck(f(('address[]', 'uint80', 'uint24'), D), T[A, i, i])  # fixed case 32225
ck(f(('address[]', 'uint80', 'uint240'), D), T[A, i, i])  # fixed case 32226
ck(f(('address[]', 'uint80', 'uint248'), D), T[A, i, i])  # fixed case 32227
ck(f(('address[]', 'uint80', 'uint256'), D), T[A, i, i])  # fixed case 32228
ck(f(('address[]', 'uint80', 'uint32'), D), T[A, i, i])  # fixed case 32229
ck(f(('address[]', 'uint80', 'uint40'), D), T[A, i, i])  # fixed case 32230
ck(f(('address[]', 'uint80', 'uint48'), D), T[A, i, i])  # fixed case 32231
ck(f(('address[]', 'uint80', 'uint56'), D), T[A, i, i])  # fixed case 32232
ck(f(('address[]', 'uint80', 'uint64'), D), T[A, i, i])  # fixed case 32233
ck(f(('address[]', 'uint80', 'uint72'), D), T[A, i, i])  # fixed case 32234
ck(f(('address[]', 'uint80', 'uint8'), D), T[A, i, i])  # fixed case 32235
ck(f(('address[]', 'uint80', 'uint80'), D), T[A, i, i])  # fixed case 32236
ck(f(('address[]', 'uint80', 'uint88'), D), T[A, i, i])  # fixed case 32237
ck(f(('address[]', 'uint80', 'uint96'), D), T[A, i, i])  # fixed case 32238
ck(f(('address[]', 'uint80', '?'), D), T[A, i, A])  # fixed case 32239
ck(f(('address[]', 'uint88', 'address'), D), T[A, i, HA])  # fixed case 32240
ck(f(('address[]', 'uint88', 'address[]'), D), T[A, i, A])  # fixed case 32241
ck(f(('address[]', 'uint88', 'bool'), D), T[A, i, bo])  # fixed case 32242
ck(f(('address[]', 'uint88', 'bool[]'), D), T[A, i, A])  # fixed case 32243
ck(f(('address[]', 'uint88', 'bytes'), D), T[A, i, by])  # fixed case 32244
ck(f(('address[]', 'uint88', 'bytes1'), D), T[A, i, by])  # fixed case 32245
ck(f(('address[]', 'uint88', 'bytes10'), D), T[A, i, by])  # fixed case 32246
ck(f(('address[]', 'uint88', 'bytes11'), D), T[A, i, by])  # fixed case 32247
ck(f(('address[]', 'uint88', 'bytes12'), D), T[A, i, by])  # fixed case 32248
ck(f(('address[]', 'uint88', 'bytes13'), D), T[A, i, by])  # fixed case 32249
ck(f(('address[]', 'uint88', 'bytes14'), D), T[A, i, by])  # fixed case 32250
ck(f(('address[]', 'uint88', 'bytes15'), D), T[A, i, by])  # fixed case 32251
ck(f(('address[]', 'uint88', 'bytes16'), D), T[A, i, by])  # fixed case 32252
ck(f(('address[]', 'uint88', 'bytes17'), D), T[A, i, by])  # fixed case 32253
ck(f(('address[]', 'uint88', 'bytes18'), D), T[A, i, by])  # fixed case 32254
ck(f(('address[]', 'uint88', 'bytes19'), D), T[A, i, by])  # fixed case 32255
ck(f(('address[]', 'uint88', 'bytes2'), D), T[A, i, by])  # fixed case 32256
ck(f(('address[]', 'uint88', 'bytes20'), D), T[A, i, by])  # fixed case 32257
ck(f(('address[]', 'uint88', 'bytes21'), D), T[A, i, by])  # fixed case 32258
ck(f(('address[]', 'uint88', 'bytes22'), D), T[A, i, by])  # fixed case 32259
ck(f(('address[]', 'uint88', 'bytes23'), D), T[A, i, by])  # fixed case 32260
ck(f(('address[]', 'uint88', 'bytes24'), D), T[A, i, by])  # fixed case 32261
ck(f(('address[]', 'uint88', 'bytes25'), D), T[A, i, by])  # fixed case 32262
ck(f(('address[]', 'uint88', 'bytes26'), D), T[A, i, by])  # fixed case 32263
ck(f(('address[]', 'uint88', 'bytes27'), D), T[A, i, by])  # fixed case 32264
ck(f(('address[]', 'uint88', 'bytes28'), D), T[A, i, by])  # fixed case 32265
ck(f(('address[]', 'uint88', 'bytes29'), D), T[A, i, by])  # fixed case 32266
ck(f(('address[]', 'uint88', 'bytes3'), D), T[A, i, by])  # fixed case 32267
ck(f(('address[]', 'uint88', 'bytes30'), D), T[A, i, by])  # fixed case 32268
ck(f(('address[]', 'uint88', 'bytes31'), D), T[A, i, by])  # fixed case 32269
ck(f(('address[]', 'uint88', 'bytes32'), D), T[A, i, by])  # fixed case 32270
ck(f(('address[]', 'uint88', 'bytes4'), D), T[A, i, by])  # fixed case 32271
ck(f(('address[]', 'uint88', 'bytes5'), D), T[A, i, by])  # fixed case 32272
ck(f(('address[]', 'uint88', 'bytes6'), D), T[A, i, by])  # fixed case 32273
ck(f(('address[]', 'uint88', 'bytes7'), D), T[A, i, by])  # fixed case 32274
ck(f(('address[]', 'uint88', 'bytes8'), D), T[A, i, by])  # fixed case 32275
ck(f(('address[]', 'uint88', 'bytes9'), D), T[A, i, by])  # fixed case 32276
ck(f(('address[]', 'uint88', 'int104'), D), T[A, i, i])  # fixed case 32277
ck(f(('address[]', 'uint88', 'int112'), D), T[A, i, i])  # fixed case 32278
ck(f(('address[]', 'uint88', 'int120'), D), T[A, i, i])  # fixed case 32279
ck(f(('address[]', 'uint88', 'int128'), D), T[A, i, i])  # fixed case 32280
ck(f(('address[]', 'uint88', 'int136'), D), T[A, i, i])  # fixed case 32281
ck(f(('address[]', 'uint88', 'int144'), D), T[A, i, i])  # fixed case 32282
ck(f(('address[]', 'uint88', 'int152'), D), T[A, i, i])  # fixed case 32283
ck(f(('address[]', 'uint88', 'int16'), D), T[A, i, i])  # fixed case 32284
ck(f(('address[]', 'uint88', 'int160'), D), T[A, i, i])  # fixed case 32285
ck(f(('address[]', 'uint88', 'int168'), D), T[A, i, i])  # fixed case 32286
ck(f(('address[]', 'uint88', 'int176'), D), T[A, i, i])  # fixed case 32287
ck(f(('address[]', 'uint88', 'int184'), D), T[A, i, i])  # fixed case 32288
ck(f(('address[]', 'uint88', 'int192'), D), T[A, i, i])  # fixed case 32289
ck(f(('address[]', 'uint88', 'int200'), D), T[A, i, i])  # fixed case 32290
ck(f(('address[]', 'uint88', 'int208'), D), T[A, i, i])  # fixed case 32291
ck(f(('address[]', 'uint88', 'int216'), D), T[A, i, i])  # fixed case 32292
ck(f(('address[]', 'uint88', 'int224'), D), T[A, i, i])  # fixed case 32293
ck(f(('address[]', 'uint88', 'int232'), D), T[A, i, i])  # fixed case 32294
ck(f(('address[]', 'uint88', 'int24'), D), T[A, i, i])  # fixed case 32295
ck(f(('address[]', 'uint88', 'int240'), D), T[A, i, i])  # fixed case 32296
ck(f(('address[]', 'uint88', 'int248'), D), T[A, i, i])  # fixed case 32297
ck(f(('address[]', 'uint88', 'int256'), D), T[A, i, i])  # fixed case 32298
ck(f(('address[]', 'uint88', 'int32'), D), T[A, i, i])  # fixed case 32299
ck(f(('address[]', 'uint88', 'int40'), D), T[A, i, i])  # fixed case 32300
ck(f(('address[]', 'uint88', 'int48'), D), T[A, i, i])  # fixed case 32301
ck(f(('address[]', 'uint88', 'int56'), D), T[A, i, i])  # fixed case 32302
ck(f(('address[]', 'uint88', 'int64'), D), T[A, i, i])  # fixed case 32303
ck(f(('address[]', 'uint88', 'int72'), D), T[A, i, i])  # fixed case 32304
ck(f(('address[]', 'uint88', 'int8'), D), T[A, i, i])  # fixed case 32305
ck(f(('address[]', 'uint88', 'int80'), D), T[A, i, i])  # fixed case 32306
ck(f(('address[]', 'uint88', 'int88'), D), T[A, i, i])  # fixed case 32307
ck(f(('address[]', 'uint88', 'int96'), D), T[A, i, i])  # fixed case 32308
ck(f(('address[]', 'uint88', 'string'), D), T[A, i, s])  # fixed case 32309
ck(f(('address[]', 'uint88', 'string[]'), D), T[A, i, A])  # fixed case 32310
ck(f(('address[]', 'uint88', 'uint104'), D), T[A, i, i])  # fixed case 32311
ck(f(('address[]', 'uint88', 'uint112'), D), T[A, i, i])  # fixed case 32312
ck(f(('address[]', 'uint88', 'uint120'), D), T[A, i, i])  # fixed case 32313
ck(f(('address[]', 'uint88', 'uint128'), D), T[A, i, i])  # fixed case 32314
ck(f(('address[]', 'uint88', 'uint136'), D), T[A, i, i])  # fixed case 32315
ck(f(('address[]', 'uint88', 'uint144'), D), T[A, i, i])  # fixed case 32316
ck(f(('address[]', 'uint88', 'uint152'), D), T[A, i, i])  # fixed case 32317
ck(f(('address[]', 'uint88', 'uint16'), D), T[A, i, i])  # fixed case 32318
ck(f(('address[]', 'uint88', 'uint160'), D), T[A, i, i])  # fixed case 32319
ck(f(('address[]', 'uint88', 'uint168'), D), T[A, i, i])  # fixed case 32320
ck(f(('address[]', 'uint88', 'uint176'), D), T[A, i, i])  # fixed case 32321
ck(f(('address[]', 'uint88', 'uint184'), D), T[A, i, i])  # fixed case 32322
ck(f(('address[]', 'uint88', 'uint192'), D), T[A, i, i])  # fixed case 32323
ck(f(('address[]', 'uint88', 'uint200'), D), T[A, i, i])  # fixed case 32324
ck(f(('address[]', 'uint88', 'uint208'), D), T[A, i, i])  # fixed case 32325
ck(f(('address[]', 'uint88', 'uint216'), D), T[A, i, i])  # fixed case 32326
ck(f(('address[]', 'uint88', 'uint224'), D), T[A, i, i])  # fixed case 32327
ck(f(('address[]', 'uint88', 'uint232'), D), T[A, i, i])  # fixed case 32328
ck(f(('address[]', 'uint88', 'uint24'), D), T[A, i, i])  # fixed case 32329
ck(f(('address[]', 'uint88', 'uint240'), D), T[A, i, i])  # fixed case 32330
ck(f(('address[]', 'uint88', 'uint248'), D), T[A, i, i])  # fixed case 32331
ck(f(('address[]', 'uint88', 'uint256'), D), T[A, i, i])  # fixed case 32332
ck(f(('address[]', 'uint88', 'uint32'), D), T[A, i, i])  # fixed case 32333
ck(f(('address[]', 'uint88', 'uint40'), D), T[A, i, i])  # fixed case 32334
ck(f(('address[]', 'uint88', 'uint48'), D), T[A, i, i])  # fixed case 32335
ck(f(('address[]', 'uint88', 'uint56'), D), T[A, i, i])  # fixed case 32336
ck(f(('address[]', 'uint88', 'uint64'), D), T[A, i, i])  # fixed case 32337
ck(f(('address[]', 'uint88', 'uint72'), D), T[A, i, i])  # fixed case 32338
ck(f(('address[]', 'uint88', 'uint8'), D), T[A, i, i])  # fixed case 32339
ck(f(('address[]', 'uint88', 'uint80'), D), T[A, i, i])  # fixed case 32340
ck(f(('address[]', 'uint88', 'uint88'), D), T[A, i, i])  # fixed case 32341
ck(f(('address[]', 'uint88', 'uint96'), D), T[A, i, i])  # fixed case 32342
ck(f(('address[]', 'uint88', '?'), D), T[A, i, A])  # fixed case 32343
ck(f(('address[]', 'uint96', 'address'), D), T[A, i, HA])  # fixed case 32344
ck(f(('address[]', 'uint96', 'address[]'), D), T[A, i, A])  # fixed case 32345
ck(f(('address[]', 'uint96', 'bool'), D), T[A, i, bo])  # fixed case 32346
ck(f(('address[]', 'uint96', 'bool[]'), D), T[A, i, A])  # fixed case 32347
ck(f(('address[]', 'uint96', 'bytes'), D), T[A, i, by])  # fixed case 32348
ck(f(('address[]', 'uint96', 'bytes1'), D), T[A, i, by])  # fixed case 32349
ck(f(('address[]', 'uint96', 'bytes10'), D), T[A, i, by])  # fixed case 32350
ck(f(('address[]', 'uint96', 'bytes11'), D), T[A, i, by])  # fixed case 32351
ck(f(('address[]', 'uint96', 'bytes12'), D), T[A, i, by])  # fixed case 32352
ck(f(('address[]', 'uint96', 'bytes13'), D), T[A, i, by])  # fixed case 32353
ck(f(('address[]', 'uint96', 'bytes14'), D), T[A, i, by])  # fixed case 32354
ck(f(('address[]', 'uint96', 'bytes15'), D), T[A, i, by])  # fixed case 32355
ck(f(('address[]', 'uint96', 'bytes16'), D), T[A, i, by])  # fixed case 32356
ck(f(('address[]', 'uint96', 'bytes17'), D), T[A, i, by])  # fixed case 32357
ck(f(('address[]', 'uint96', 'bytes18'), D), T[A, i, by])  # fixed case 32358
ck(f(('address[]', 'uint96', 'bytes19'), D), T[A, i, by])  # fixed case 32359
ck(f(('address[]', 'uint96', 'bytes2'), D), T[A, i, by])  # fixed case 32360
ck(f(('address[]', 'uint96', 'bytes20'), D), T[A, i, by])  # fixed case 32361
ck(f(('address[]', 'uint96', 'bytes21'), D), T[A, i, by])  # fixed case 32362
ck(f(('address[]', 'uint96', 'bytes22'), D), T[A, i, by])  # fixed case 32363
ck(f(('address[]', 'uint96', 'bytes23'), D), T[A, i, by])  # fixed case 32364
ck(f(('address[]', 'uint96', 'bytes24'), D), T[A, i, by])  # fixed case 32365
ck(f(('address[]', 'uint96', 'bytes25'), D), T[A, i, by])  # fixed case 32366
ck(f(('address[]', 'uint96', 'bytes26'), D), T[A, i, by])  # fixed case 32367
ck(f(('address[]', 'uint96', 'bytes27'), D), T[A, i, by])  # fixed case 32368
ck(f(('address[]', 'uint96', 'bytes28'), D), T[A, i, by])  # fixed case 32369
ck(f(('address[]', 'uint96', 'bytes29'), D), T[A, i, by])  # fixed case 32370
ck(f(('address[]', 'uint96', 'bytes3'), D), T[A, i, by])  # fixed case 32371
ck(f(('address[]', 'uint96', 'bytes30'), D), T[A, i, by])  # fixed case 32372
ck(f(('address[]', 'uint96', 'bytes31'), D), T[A, i, by])  # fixed case 32373
ck(f(('address[]', 'uint96', 'bytes32'), D), T[A, i, by])  # fixed case 32374
ck(f(('address[]', 'uint96', 'bytes4'), D), T[A, i, by])  # fixed case 32375
ck(f(('address[]', 'uint96', 'bytes5'), D), T[A, i, by])  # fixed case 32376
ck(f(('address[]', 'uint96', 'bytes6'), D), T[A, i, by])  # fixed case 32377
ck(f(('address[]', 'uint96', 'bytes7'), D), T[A, i, by])  # fixed case 32378
ck(f(('address[]', 'uint96', 'bytes8'), D), T[A, i, by])  # fixed case 32379
ck(f(('address[]', 'uint96', 'bytes9'), D), T[A, i, by])  # fixed case 32380
ck(f(('address[]', 'uint96', 'int104'), D), T[A, i, i])  # fixed case 32381
ck(f(('address[]', 'uint96', 'int112'), D), T[A, i, i])  # fixed case 32382
ck(f(('address[]', 'uint96', 'int120'), D), T[A, i, i])  # fixed case 32383
ck(f(('address[]', 'uint96', 'int128'), D), T[A, i, i])  # fixed case 32384
ck(f(('address[]', 'uint96', 'int136'), D), T[A, i, i])  # fixed case 32385
ck(f(('address[]', 'uint96', 'int144'), D), T[A, i, i])  # fixed case 32386
ck(f(('address[]', 'uint96', 'int152'), D), T[A, i, i])  # fixed case 32387
ck(f(('address[]', 'uint96', 'int16'), D), T[A, i, i])  # fixed case 32388
ck(f(('address[]', 'uint96', 'int160'), D), T[A, i, i])  # fixed case 32389
ck(f(('address[]', 'uint96', 'int168'), D), T[A, i, i])  # fixed case 32390
ck(f(('address[]', 'uint96', 'int176'), D), T[A, i, i])  # fixed case 32391
ck(f(('address[]', 'uint96', 'int184'), D), T[A, i, i])  # fixed case 32392
ck(f(('address[]', 'uint96', 'int192'), D), T[A, i, i])  # fixed case 32393
ck(f(('address[]', 'uint96', 'int200'), D), T[A, i, i])  # fixed case 32394
ck(f(('address[]', 'uint96', 'int208'), D), T[A, i, i])  # fixed case 32395
ck(f(('address[]', 'uint96', 'int216'), D), T[A, i, i])  # fixed case 32396
ck(f(('address[]', 'uint96', 'int224'), D), T[A, i, i])  # fixed case 32397
ck(f(('address[]', 'uint96', 'int232'), D), T[A, i, i])  # fixed case 32398
ck(f(('address[]', 'uint96', 'int24'), D), T[A, i, i])  # fixed case 32399
ck(f(('address[]', 'uint96', 'int240'), D), T[A, i, i])  # fixed case 32400
ck(f(('address[]', 'uint96', 'int248'), D), T[A, i, i])  # fixed case 32401
ck(f(('address[]', 'uint96', 'int256'), D), T[A, i, i])  # fixed case 32402
ck(f(('address[]', 'uint96', 'int32'), D), T[A, i, i])  # fixed case 32403
ck(f(('address[]', 'uint96', 'int40'), D), T[A, i, i])  # fixed case 32404
ck(f(('address[]', 'uint96', 'int48'), D), T[A, i, i])  # fixed case 32405
ck(f(('address[]', 'uint96', 'int56'), D), T[A, i, i])  # fixed case 32406
ck(f(('address[]', 'uint96', 'int64'), D), T[A, i, i])  # fixed case 32407
ck(f(('address[]', 'uint96', 'int72'), D), T[A, i, i])  # fixed case 32408
ck(f(('address[]', 'uint96', 'int8'), D), T[A, i, i])  # fixed case 32409
ck(f(('address[]', 'uint96', 'int80'), D), T[A, i, i])  # fixed case 32410
ck(f(('address[]', 'uint96', 'int88'), D), T[A, i, i])  # fixed case 32411
ck(f(('address[]', 'uint96', 'int96'), D), T[A, i, i])  # fixed case 32412
ck(f(('address[]', 'uint96', 'string'), D), T[A, i, s])  # fixed case 32413
ck(f(('address[]', 'uint96', 'string[]'), D), T[A, i, A])  # fixed case 32414
ck(f(('address[]', 'uint96', 'uint104'), D), T[A, i, i])  # fixed case 32415
ck(f(('address[]', 'uint96', 'uint112'), D), T[A, i, i])  # fixed case 32416
ck(f(('address[]', 'uint96', 'uint120'), D), T[A, i, i])  # fixed case 32417
ck(f(('address[]', 'uint96', 'uint128'), D), T[A, i, i])  # fixed case 32418
ck(f(('address[]', 'uint96', 'uint136'), D), T[A, i, i])  # fixed case 32419
ck(f(('address[]', 'uint96', 'uint144'), D), T[A, i, i])  # fixed case 32420
ck(f(('address[]', 'uint96', 'uint152'), D), T[A, i, i])  # fixed case 32421
ck(f(('address[]', 'uint96', 'uint16'), D), T[A, i, i])  # fixed case 32422
ck(f(('address[]', 'uint96', 'uint160'), D), T[A, i, i])  # fixed case 32423
ck(f(('address[]', 'uint96', 'uint168'), D), T[A, i, i])  # fixed case 32424
ck(f(('address[]', 'uint96', 'uint176'), D), T[A, i, i])  # fixed case 32425
ck(f(('address[]', 'uint96', 'uint184'), D), T[A, i, i])  # fixed case 32426
ck(f(('address[]', 'uint96', 'uint192'), D), T[A, i, i])  # fixed case 32427
ck(f(('address[]', 'uint96', 'uint200'), D), T[A, i, i])  # fixed case 32428
ck(f(('address[]', 'uint96', 'uint208'), D), T[A, i, i])  # fixed case 32429
ck(f(('address[]', 'uint96', 'uint216'), D), T[A, i, i])  # fixed case 32430
ck(f(('address[]', 'uint96', 'uint224'), D), T[A, i, i])  # fixed case 32431
ck(f(('address[]', 'uint96', 'uint232'), D), T[A, i, i])  # fixed case 32432
ck(f(('address[]', 'uint96', 'uint24'), D), T[A, i, i])  # fixed case 32433
ck(f(('address[]', 'uint96', 'uint240'), D), T[A, i, i])  # fixed case 32434
ck(f(('address[]', 'uint96', 'uint248'), D), T[A, i, i])  # fixed case 32435
ck(f(('address[]', 'uint96', 'uint256'), D), T[A, i, i])  # fixed case 32436
ck(f(('address[]', 'uint96', 'uint32'), D), T[A, i, i])  # fixed case 32437
ck(f(('address[]', 'uint96', 'uint40'), D), T[A, i, i])  # fixed case 32438
ck(f(('address[]', 'uint96', 'uint48'), D), T[A, i, i])  # fixed case 32439
ck(f(('address[]', 'uint96', 'uint56'), D), T[A, i, i])  # fixed case 32440
ck(f(('address[]', 'uint96', 'uint64'), D), T[A, i, i])  # fixed case 32441
ck(f(('address[]', 'uint96', 'uint72'), D), T[A, i, i])  # fixed case 32442
ck(f(('address[]', 'uint96', 'uint8'), D), T[A, i, i])  # fixed case 32443
ck(f(('address[]', 'uint96', 'uint80'), D), T[A, i, i])  # fixed case 32444
ck(f(('address[]', 'uint96', 'uint88'), D), T[A, i, i])  # fixed case 32445
ck(f(('address[]', 'uint96', 'uint96'), D), T[A, i, i])  # fixed case 32446
ck(f(('address[]', 'uint96', '?'), D), T[A, i, A])  # fixed case 32447
ck(f(('address[]', '?', 'address'), D), T[A, A, HA])  # fixed case 32448
ck(f(('address[]', '?', 'address[]'), D), T[A, A, A])  # fixed case 32449
ck(f(('address[]', '?', 'bool'), D), T[A, A, bo])  # fixed case 32450
ck(f(('address[]', '?', 'bool[]'), D), T[A, A, A])  # fixed case 32451
ck(f(('address[]', '?', 'bytes'), D), T[A, A, by])  # fixed case 32452
ck(f(('address[]', '?', 'bytes1'), D), T[A, A, by])  # fixed case 32453
ck(f(('address[]', '?', 'bytes10'), D), T[A, A, by])  # fixed case 32454
ck(f(('address[]', '?', 'bytes11'), D), T[A, A, by])  # fixed case 32455
ck(f(('address[]', '?', 'bytes12'), D), T[A, A, by])  # fixed case 32456
ck(f(('address[]', '?', 'bytes13'), D), T[A, A, by])  # fixed case 32457
ck(f(('address[]', '?', 'bytes14'), D), T[A, A, by])  # fixed case 32458
ck(f(('address[]', '?', 'bytes15'), D), T[A, A, by])  # fixed case 32459
ck(f(('address[]', '?', 'bytes16'), D), T[A, A, by])  # fixed case 32460
ck(f(('address[]', '?', 'bytes17'), D), T[A, A, by])  # fixed case 32461
ck(f(('address[]', '?', 'bytes18'), D), T[A, A, by])  # fixed case 32462
ck(f(('address[]', '?', 'bytes19'), D), T[A, A, by])  # fixed case 32463
ck(f(('address[]', '?', 'bytes2'), D), T[A, A, by])  # fixed case 32464
ck(f(('address[]', '?', 'bytes20'), D), T[A, A, by])  # fixed case 32465
ck(f(('address[]', '?', 'bytes21'), D), T[A, A, by])  # fixed case 32466
ck(f(('address[]', '?', 'bytes22'), D), T[A, A, by])  # fixed case 32467
ck(f(('address[]', '?', 'bytes23'), D), T[A, A, by])  # fixed case 32468
ck(f(('address[]', '?', 'bytes24'), D), T[A, A, by])  # fixed case 32469
ck(f(('address[]', '?', 'bytes25'), D), T[A, A, by])  # fixed case 32470
ck(f(('address[]', '?', 'bytes26'), D), T[A, A, by])  # fixed case 32471
ck(f(('address[]', '?', 'bytes27'), D), T[A, A, by])  # fixed case 32472
ck(f(('address[]', '?', 'bytes28'), D), T[A, A, by])  # fixed case 32473
ck(f(('address[]', '?', 'bytes29'), D), T[A, A, by])  # fixed case 32474
ck(f(('address[]', '?', 'bytes3'), D), T[A, A, by])  # fixed case 32475
ck(f(('address[]', '?', 'bytes30'), D), T[A, A, by])  # fixed case 32476
ck(f(('address[]', '?', 'bytes31'), D), T[A, A, by])  # fixed case 32477
ck(f(('address[]', '?', 'bytes32'), D), T[A, A, by])  # fixed case 32478
ck(f(('address[]', '?', 'bytes4'), D), T[A, A, by])  # fixed case 32479
ck(f(('address[]', '?', 'bytes5'), D), T[A, A, by])  # fixed case 32480
ck(f(('address[]', '?', 'bytes6'), D), T[A, A, by])  # fixed case 32481
ck(f(('address[]', '?', 'bytes7'), D), T[A, A, by])  # fixed case 32482
ck(f(('address[]', '?', 'bytes8'), D), T[A, A, by])  # fixed case 32483
ck(f(('address[]', '?', 'bytes9'), D), T[A, A, by])  # fixed case 32484
ck(f(('address[]', '?', 'int104'), D), T[A, A, i])  # fixed case 32485
ck(f(('address[]', '?', 'int112'), D), T[A, A, i])  # fixed case 32486
ck(f(('address[]', '?', 'int120'), D), T[A, A, i])  # fixed case 32487
ck(f(('address[]', '?', 'int128'), D), T[A, A, i])  # fixed case 32488
ck(f(('address[]', '?', 'int136'), D), T[A, A, i])  # fixed case 32489
ck(f(('address[]', '?', 'int144'), D), T[A, A, i])  # fixed case 32490
ck(f(('address[]', '?', 'int152'), D), T[A, A, i])  # fixed case 32491
ck(f(('address[]', '?', 'int16'), D), T[A, A, i])  # fixed case 32492
ck(f(('address[]', '?', 'int160'), D), T[A, A, i])  # fixed case 32493
ck(f(('address[]', '?', 'int168'), D), T[A, A, i])  # fixed case 32494
ck(f(('address[]', '?', 'int176'), D), T[A, A, i])  # fixed case 32495
ck(f(('address[]', '?', 'int184'), D), T[A, A, i])  # fixed case 32496
ck(f(('address[]', '?', 'int192'), D), T[A, A, i])  # fixed case 32497
ck(f(('address[]', '?', 'int200'), D), T[A, A, i])  # fixed case 32498
ck(f(('address[]', '?', 'int208'), D), T[A, A, i])  # fixed case 32499
ck(f(('address[]', '?', 'int216'), D), T[A, A, i])  # fixed case 32500
ck(f(('address[]', '?', 'int224'), D), T[A, A, i])  # fixed case 32501
ck(f(('address[]', '?', 'int232'), D), T[A, A, i])  # fixed case 32502
ck(f(('address[]', '?', 'int24'), D), T[A, A, i])  # fixed case 32503
ck(f(('address[]', '?', 'int240'), D), T[A, A, i])  # fixed case 32504
ck(f(('address[]', '?', 'int248'), D), T[A, A, i])  # fixed case 32505
ck(f(('address[]', '?', 'int256'), D), T[A, A, i])  # fixed case 32506
ck(f(('address[]', '?', 'int32'), D), T[A, A, i])  # fixed case 32507
ck(f(('address[]', '?', 'int40'), D), T[A, A, i])  # fixed case 32508
ck(f(('address[]', '?', 'int48'), D), T[A, A, i])  # fixed case 32509
ck(f(('address[]', '?', 'int56'), D), T[A, A, i])  # fixed case 32510
ck(f(('address[]', '?', 'int64'), D), T[A, A, i])  # fixed case 32511
ck(f(('address[]', '?', 'int72'), D), T[A, A, i])  # fixed case 32512
ck(f(('address[]', '?', 'int8'), D), T[A, A, i])  # fixed case 32513
ck(f(('address[]', '?', 'int80'), D), T[A, A, i])  # fixed case 32514
ck(f(('address[]', '?', 'int88'), D), T[A, A, i])  # fixed case 32515
ck(f(('address[]', '?', 'int96'), D), T[A, A, i])  # fixed case 32516
ck(f(('address[]', '?', 'string'), D), T[A, A, s])  # fixed case 32517
ck(f(('address[]', '?', 'string[]'), D), T[A, A, A])  # fixed case 32518
ck(f(('address[]', '?', 'uint104'), D), T[A, A, i])  # fixed case 32519
ck(f(('address[]', '?', 'uint112'), D), T[A, A, i])  # fixed case 32520
ck(f(('address[]', '?', 'uint120'), D), T[A, A, i])  # fixed case 32521
ck(f(('address[]', '?', 'uint128'), D), T[A, A, i])  # fixed case 32522
ck(f(('address[]', '?', 'uint136'), D), T[A, A, i])  # fixed case 32523
ck(f(('address[]', '?', 'uint144'), D), T[A, A, i])  # fixed case 32524
ck(f(('address[]', '?', 'uint152'), D), T[A, A, i])  # fixed case 32525
ck(f(('address[]', '?', 'uint16'), D), T[A, A, i])  # fixed case 32526
ck(f(('address[]', '?', 'uint160'), D), T[A, A, i])  # fixed case 32527
ck(f(('address[]', '?', 'uint168'), D), T[A, A, i])  # fixed case 32528
ck(f(('address[]', '?', 'uint176'), D), T[A, A, i])  # fixed case 32529
ck(f(('address[]', '?', 'uint184'), D), T[A, A, i])  # fixed case 32530
ck(f(('address[]', '?', 'uint192'), D), T[A, A, i])  # fixed case 32531
ck(f(('address[]', '?', 'uint200'), D), T[A, A, i])  # fixed case 32532
ck(f(('address[]', '?', 'uint208'), D), T[A, A, i])  # fixed case 32533
ck(f(('address[]', '?', 'uint216'), D), T[A, A, i])  # fixed case 32534
ck(f(('address[]', '?', 'uint224'), D), T[A, A, i])  # fixed case 32535
ck(f(('address[]', '?', 'uint232'), D), T[A, A, i])  # fixed case 32536
ck(f(('address[]', '?', 'uint24'), D), T[A, A, i])  # fixed case 32537
ck(f(('address[]', '?', 'uint240'), D), T[A, A, i])  # fixed case 32538
ck(f(('address[]', '?', 'uint248'), D), T[A, A, i])  # fixed case 32539
ck(f(('address[]', '?', 'uint256'), D), T[A, A, i])  # fixed case 32540
ck(f(('address[]', '?', 'uint32'), D), T[A, A, i])  # fixed case 32541
ck(f(('address[]', '?', 'uint40'), D), T[A, A, i])  # fixed case 32542
ck(f(('address[]', '?', 'uint48'), D), T[A, A, i])  # fixed case 32543
ck(f(('address[]', '?', 'uint56'), D), T[A, A, i])  # fixed case 32544
ck(f(('address[]', '?', 'uint64'), D), T[A, A, i])  # fixed case 32545
ck(f(('address[]', '?', 'uint72'), D), T[A, A, i])  # fixed case 32546
ck(f(('address[]', '?', 'uint8'), D), T[A, A, i])  # fixed case 32547
ck(f(('address[]', '?', 'uint80'), D), T[A, A, i])  # fixed case 32548
ck(f(('address[]', '?', 'uint88'), D), T[A, A, i])  # fixed case 32549
ck(f(('address[]', '?', 'uint96'), D), T[A, A, i])  # fixed case 32550
ck(f(('address[]', '?', '?'), D), T[A, A, A])  # fixed case 32551
ck(f(('bool', 'address', 'address'), D), T[bo, HA, HA])  # fixed case 32552
ck(f(('bool', 'address', 'address[]'), D), T[bo, HA, A])  # fixed case 32553
ck(f(('bool', 'address', 'bool'), D), T[bo, HA, bo])  # fixed case 32554
ck(f(('bool', 'address', 'bool[]'), D), T[bo, HA, A])  # fixed case 32555
ck(f(('bool', 'address', 'bytes'), D), T[bo, HA, by])  # fixed case 32556
ck(f(('bool', 'address', 'bytes1'), D), T[bo, HA, by])  # fixed case 32557
ck(f(('bool', 'address', 'bytes10'), D), T[bo, HA, by])  # fixed case 32558
ck(f(('bool', 'address', 'bytes11'), D), T[bo, HA, by])  # fixed case 32559
ck(f(('bool', 'address', 'bytes12'), D), T[bo, HA, by])  # fixed case 32560
ck(f(('bool', 'address', 'bytes13'), D), T[bo, HA, by])  # fixed case 32561
ck(f(('bool', 'address', 'bytes14'), D), T[bo, HA, by])  # fixed case 32562
ck(f(('bool', 'address', 'bytes15'), D), T[bo, HA, by])  # fixed case 32563
ck(f(('bool', 'address', 'bytes16'), D), T[bo, HA, by])  # fixed case 32564
ck(f(('bool', 'address', 'bytes17'), D), T[bo, HA, by])  # fixed case 32565
ck(f(('bool', 'address', 'bytes18'), D), T[bo, HA, by])  # fixed case 32566
ck(f(('bool', 'address', 'bytes19'), D), T[bo, HA, by])  # fixed case 32567
ck(f(('bool', 'address', 'bytes2'), D), T[bo, HA, by])  # fixed case 32568
ck(f(('bool', 'address', 'bytes20'), D), T[bo, HA, by])  # fixed case 32569
ck(f(('bool', 'address', 'bytes21'), D), T[bo, HA, by])  # fixed case 32570
ck(f(('bool', 'address', 'bytes22'), D), T[bo, HA, by])  # fixed case 32571
ck(f(('bool', 'address', 'bytes23'), D), T[bo, HA, by])  # fixed case 32572
ck(f(('bool', 'address', 'bytes24'), D), T[bo, HA, by])  # fixed case 32573
ck(f(('bool', 'address', 'bytes25'), D), T[bo, HA, by])  # fixed case 32574
ck(f(('bool', 'address', 'bytes26'), D), T[bo, HA, by])  # fixed case 32575
ck(f(('bool', 'address', 'bytes27'), D), T[bo, HA, by])  # fixed case 32576
ck(f(('bool', 'address', 'bytes28'), D), T[bo, HA, by])  # fixed case 32577
ck(f(('bool', 'address', 'bytes29'), D), T[bo, HA, by])  # fixed case 32578
ck(f(('bool', 'address', 'bytes3'), D), T[bo, HA, by])  # fixed case 32579
ck(f(('bool', 'address', 'bytes30'), D), T[bo, HA, by])  # fixed case 32580
ck(f(('bool', 'address', 'bytes31'), D), T[bo, HA, by])  # fixed case 32581
ck(f(('bool', 'address', 'bytes32'), D), T[bo, HA, by])  # fixed case 32582
ck(f(('bool', 'address', 'bytes4'), D), T[bo, HA, by])  # fixed case 32583
ck(f(('bool', 'address', 'bytes5'), D), T[bo, HA, by])  # fixed case 32584
ck(f(('bool', 'address', 'bytes6'), D), T[bo, HA, by])  # fixed case 32585
ck(f(('bool', 'address', 'bytes7'), D), T[bo, HA, by])  # fixed case 32586
ck(f(('bool', 'address', 'bytes8'), D), T[bo, HA, by])  # fixed case 32587
ck(f(('bool', 'address', 'bytes9'), D), T[bo, HA, by])  # fixed case 32588
ck(f(('bool', 'address', 'int104'), D), T[bo, HA, i])  # fixed case 32589
ck(f(('bool', 'address', 'int112'), D), T[bo, HA, i])  # fixed case 32590
ck(f(('bool', 'address', 'int120'), D), T[bo, HA, i])  # fixed case 32591
ck(f(('bool', 'address', 'int128'), D), T[bo, HA, i])  # fixed case 32592
ck(f(('bool', 'address', 'int136'), D), T[bo, HA, i])  # fixed case 32593
ck(f(('bool', 'address', 'int144'), D), T[bo, HA, i])  # fixed case 32594
ck(f(('bool', 'address', 'int152'), D), T[bo, HA, i])  # fixed case 32595
ck(f(('bool', 'address', 'int16'), D), T[bo, HA, i])  # fixed case 32596
ck(f(('bool', 'address', 'int160'), D), T[bo, HA, i])  # fixed case 32597
ck(f(('bool', 'address', 'int168'), D), T[bo, HA, i])  # fixed case 32598
ck(f(('bool', 'address', 'int176'), D), T[bo, HA, i])  # fixed case 32599
ck(f(('bool', 'address', 'int184'), D), T[bo, HA, i])  # fixed case 32600
ck(f(('bool', 'address', 'int192'), D), T[bo, HA, i])  # fixed case 32601
ck(f(('bool', 'address', 'int200'), D), T[bo, HA, i])  # fixed case 32602
ck(f(('bool', 'address', 'int208'), D), T[bo, HA, i])  # fixed case 32603
ck(f(('bool', 'address', 'int216'), D), T[bo, HA, i])  # fixed case 32604
ck(f(('bool', 'address', 'int224'), D), T[bo, HA, i])  # fixed case 32605
ck(f(('bool', 'address', 'int232'), D), T[bo, HA, i])  # fixed case 32606
ck(f(('bool', 'address', 'int24'), D), T[bo, HA, i])  # fixed case 32607
ck(f(('bool', 'address', 'int240'), D), T[bo, HA, i])  # fixed case 32608
ck(f(('bool', 'address', 'int248'), D), T[bo, HA, i])  # fixed case 32609
ck(f(('bool', 'address', 'int256'), D), T[bo, HA, i])  # fixed case 32610
ck(f(('bool', 'address', 'int32'), D), T[bo, HA, i])  # fixed case 32611
ck(f(('bool', 'address', 'int40'), D), T[bo, HA, i])  # fixed case 32612
ck(f(('bool', 'address', 'int48'), D), T[bo, HA, i])  # fixed case 32613
ck(f(('bool', 'address', 'int56'), D), T[bo, HA, i])  # fixed case 32614
ck(f(('bool', 'address', 'int64'), D), T[bo, HA, i])  # fixed case 32615
ck(f(('bool', 'address', 'int72'), D), T[bo, HA, i])  # fixed case 32616
ck(f(('bool', 'address', 'int8'), D), T[bo, HA, i])  # fixed case 32617
ck(f(('bool', 'address', 'int80'), D), T[bo, HA, i])  # fixed case 32618
ck(f(('bool', 'address', 'int88'), D), T[bo, HA, i])  # fixed case 32619
ck(f(('bool', 'address', 'int96'), D), T[bo, HA, i])  # fixed case 32620
ck(f(('bool', 'address', 'string'), D), T[bo, HA, s])  # fixed case 32621
ck(f(('bool', 'address', 'string[]'), D), T[bo, HA, A])  # fixed case 32622
ck(f(('bool', 'address', 'uint104'), D), T[bo, HA, i])  # fixed case 32623
ck(f(('bool', 'address', 'uint112'), D), T[bo, HA, i])  # fixed case 32624
ck(f(('bool', 'address', 'uint120'), D), T[bo, HA, i])  # fixed case 32625
ck(f(('bool', 'address', 'uint128'), D), T[bo, HA, i])  # fixed case 32626
ck(f(('bool', 'address', 'uint136'), D), T[bo, HA, i])  # fixed case 32627
ck(f(('bool', 'address', 'uint144'), D), T[bo, HA, i])  # fixed case 32628
ck(f(('bool', 'address', 'uint152'), D), T[bo, HA, i])  # fixed case 32629
ck(f(('bool', 'address', 'uint16'), D), T[bo, HA, i])  # fixed case 32630
ck(f(('bool', 'address', 'uint160'), D), T[bo, HA, i])  # fixed case 32631
ck(f(('bool', 'address', 'uint168'), D), T[bo, HA, i])  # fixed case 32632
ck(f(('bool', 'address', 'uint176'), D), T[bo, HA, i])  # fixed case 32633
ck(f(('bool', 'address', 'uint184'), D), T[bo, HA, i])  # fixed case 32634
ck(f(('bool', 'address', 'uint192'), D), T[bo, HA, i])  # fixed case 32635
ck(f(('bool', 'address', 'uint200'), D), T[bo, HA, i])  # fixed case 32636
ck(f(('bool', 'address', 'uint208'), D), T[bo, HA, i])  # fixed case 32637
ck(f(('bool', 'address', 'uint216'), D), T[bo, HA, i])  # fixed case 32638
ck(f(('bool', 'address', 'uint224'), D), T[bo, HA, i])  # fixed case 32639
ck(f(('bool', 'address', 'uint232'), D), T[bo, HA, i])  # fixed case 32640
ck(f(('bool', 'address', 'uint24'), D), T[bo, HA, i])  # fixed case 32641
ck(f(('bool', 'address', 'uint240'), D), T[bo, HA, i])  # fixed case 32642
ck(f(('bool', 'address', 'uint248'), D), T[bo, HA, i])  # fixed case 32643
ck(f(('bool', 'address', 'uint256'), D), T[bo, HA, i])  # fixed case 32644
ck(f(('bool', 'address', 'uint32'), D), T[bo, HA, i])  # fixed case 32645
ck(f(('bool', 'address', 'uint40'), D), T[bo, HA, i])  # fixed case 32646
ck(f(('bool', 'address', 'uint48'), D), T[bo, HA, i])  # fixed case 32647
ck(f(('bool', 'address', 'uint56'), D), T[bo, HA, i])  # fixed case 32648
ck(f(('bool', 'address', 'uint64'), D), T[bo, HA, i])  # fixed case 32649
ck(f(('bool', 'address', 'uint72'), D), T[bo, HA, i])  # fixed case 32650
ck(f(('bool', 'address', 'uint8'), D), T[bo, HA, i])  # fixed case 32651
ck(f(('bool', 'address', 'uint80'), D), T[bo, HA, i])  # fixed case 32652
ck(f(('bool', 'address', 'uint88'), D), T[bo, HA, i])  # fixed case 32653
ck(f(('bool', 'address', 'uint96'), D), T[bo, HA, i])  # fixed case 32654
ck(f(('bool', 'address', '?'), D), T[bo, HA, A])  # fixed case 32655
ck(f(('bool', 'address[]', 'address'), D), T[bo, A, HA])  # fixed case 32656
ck(f(('bool', 'address[]', 'address[]'), D), T[bo, A, A])  # fixed case 32657
ck(f(('bool', 'address[]', 'bool'), D), T[bo, A, bo])  # fixed case 32658
ck(f(('bool', 'address[]', 'bool[]'), D), T[bo, A, A])  # fixed case 32659
ck(f(('bool', 'address[]', 'bytes'), D), T[bo, A, by])  # fixed case 32660
ck(f(('bool', 'address[]', 'bytes1'), D), T[bo, A, by])  # fixed case 32661
ck(f(('bool', 'address[]', 'bytes10'), D), T[bo, A, by])  # fixed case 32662
ck(f(('bool', 'address[]', 'bytes11'), D), T[bo, A, by])  # fixed case 32663
ck(f(('bool', 'address[]', 'bytes12'), D), T[bo, A, by])  # fixed case 32664
ck(f(('bool', 'address[]', 'bytes13'), D), T[bo, A, by])  # fixed case 32665
ck(f(('bool', 'address[]', 'bytes14'), D), T[bo, A, by])  # fixed case 32666
ck(f(('bool', 'address[]', 'bytes15'), D), T[bo, A, by])  # fixed case 32667
ck(f(('bool', 'address[]', 'bytes16'), D), T[bo, A, by])  # fixed case 32668
ck(f(('bool', 'address[]', 'bytes17'), D), T[bo, A, by])  # fixed case 32669
ck(f(('bool', 'address[]', 'bytes18'), D), T[bo, A, by])  # fixed case 32670
ck(f(('bool', 'address[]', 'bytes19'), D), T[bo, A, by])  # fixed case 32671
ck(f(('bool', 'address[]', 'bytes2'), D), T[bo, A, by])  # fixed case 32672
ck(f(('bool', 'address[]', 'bytes20'), D), T[bo, A, by])  # fixed case 32673
ck(f(('bool', 'address[]', 'bytes21'), D), T[bo, A, by])  # fixed case 32674
ck(f(('bool', 'address[]', 'bytes22'), D), T[bo, A, by])  # fixed case 32675
ck(f(('bool', 'address[]', 'bytes23'), D), T[bo, A, by])  # fixed case 32676
ck(f(('bool', 'address[]', 'bytes24'), D), T[bo, A, by])  # fixed case 32677
ck(f(('bool', 'address[]', 'bytes25'), D), T[bo, A, by])  # fixed case 32678
ck(f(('bool', 'address[]', 'bytes26'), D), T[bo, A, by])  # fixed case 32679
ck(f(('bool', 'address[]', 'bytes27'), D), T[bo, A, by])  # fixed case 32680
ck(f(('bool', 'address[]', 'bytes28'), D), T[bo, A, by])  # fixed case 32681
ck(f(('bool', 'address[]', 'bytes29'), D), T[bo, A, by])  # fixed case 32682
ck(f(('bool', 'address[]', 'bytes3'), D), T[bo, A, by])  # fixed case 32683
ck(f(('bool', 'address[]', 'bytes30'), D), T[bo, A, by])  # fixed case 32684
ck(f(('bool', 'address[]', 'bytes31'), D), T[bo, A, by])  # fixed case 32685
ck(f(('bool', 'address[]', 'bytes32'), D), T[bo, A, by])  # fixed case 32686
ck(f(('bool', 'address[]', 'bytes4'), D), T[bo, A, by])  # fixed case 32687
ck(f(('bool', 'address[]', 'bytes5'), D), T[bo, A, by])  # fixed case 32688
ck(f(('bool', 'address[]', 'bytes6'), D), T[bo, A, by])  # fixed case 32689
ck(f(('bool', 'address[]', 'bytes7'), D), T[bo, A, by])  # fixed case 32690
ck(f(('bool', 'address[]', 'bytes8'), D), T[bo, A, by])  # fixed case 32691
ck(f(('bool', 'address[]', 'bytes9'), D), T[bo, A, by])  # fixed case 32692
ck(f(('bool', 'address[]', 'int104'), D), T[bo, A, i])  # fixed case 32693
ck(f(('bool', 'address[]', 'int112'), D), T[bo, A, i])  # fixed case 32694
ck(f(('bool', 'address[]', 'int120'), D), T[bo, A, i])  # fixed case 32695
ck(f(('bool', 'address[]', 'int128'), D), T[bo, A, i])  # fixed case 32696
ck(f(('bool', 'address[]', 'int136'), D), T[bo, A, i])  # fixed case 32697
ck(f(('bool', 'address[]', 'int144'), D), T[bo, A, i])  # fixed case 32698
ck(f(('bool', 'address[]', 'int152'), D), T[bo, A, i])  # fixed case 32699
ck(f(('bool', 'address[]', 'int16'), D), T[bo, A, i])  # fixed case 32700
ck(f(('bool', 'address[]', 'int160'), D), T[bo, A, i])  # fixed case 32701
ck(f(('bool', 'address[]', 'int168'), D), T[bo, A, i])  # fixed case 32702
ck(f(('bool', 'address[]', 'int176'), D), T[bo, A, i])  # fixed case 32703
ck(f(('bool', 'address[]', 'int184'), D), T[bo, A, i])  # fixed case 32704
ck(f(('bool', 'address[]', 'int192'), D), T[bo, A, i])  # fixed case 32705
ck(f(('bool', 'address[]', 'int200'), D), T[bo, A, i])  # fixed case 32706
ck(f(('bool', 'address[]', 'int208'), D), T[bo, A, i])  # fixed case 32707
ck(f(('bool', 'address[]', 'int216'), D), T[bo, A, i])  # fixed case 32708
ck(f(('bool', 'address[]', 'int224'), D), T[bo, A, i])  # fixed case 32709
ck(f(('bool', 'address[]', 'int232'), D), T[bo, A, i])  # fixed case 32710
ck(f(('bool', 'address[]', 'int24'), D), T[bo, A, i])  # fixed case 32711
ck(f(('bool', 'address[]', 'int240'), D), T[bo, A, i])  # fixed case 32712
ck(f(('bool', 'address[]', 'int248'), D), T[bo, A, i])  # fixed case 32713
ck(f(('bool', 'address[]', 'int256'), D), T[bo, A, i])  # fixed case 32714
ck(f(('bool', 'address[]', 'int32'), D), T[bo, A, i])  # fixed case 32715
ck(f(('bool', 'address[]', 'int40'), D), T[bo, A, i])  # fixed case 32716
ck(f(('bool', 'address[]', 'int48'), D), T[bo, A, i])  # fixed case 32717
ck(f(('bool', 'address[]', 'int56'), D), T[bo, A, i])  # fixed case 32718
ck(f(('bool', 'address[]', 'int64'), D), T[bo, A, i])  # fixed case 32719
ck(f(('bool', 'address[]', 'int72'), D), T[bo, A, i])  # fixed case 32720
ck(f(('bool', 'address[]', 'int8'), D), T[bo, A, i])  # fixed case 32721
ck(f(('bool', 'address[]', 'int80'), D), T[bo, A, i])  # fixed case 32722
ck(f(('bool', 'address[]', 'int88'), D), T[bo, A, i])  # fixed case 32723
ck(f(('bool', 'address[]', 'int96'), D), T[bo, A, i])  # fixed case 32724
ck(f(('bool', 'address[]', 'string'), D), T[bo, A, s])  # fixed case 32725
ck(f(('bool', 'address[]', 'string[]'), D), T[bo, A, A])  # fixed case 32726
ck(f(('bool', 'address[]', 'uint104'), D), T[bo, A, i])  # fixed case 32727
ck(f(('bool', 'address[]', 'uint112'), D), T[bo, A, i])  # fixed case 32728
ck(f(('bool', 'address[]', 'uint120'), D), T[bo, A, i])  # fixed case 32729
ck(f(('bool', 'address[]', 'uint128'), D), T[bo, A, i])  # fixed case 32730
ck(f(('bool', 'address[]', 'uint136'), D), T[bo, A, i])  # fixed case 32731
ck(f(('bool', 'address[]', 'uint144'), D), T[bo, A, i])  # fixed case 32732
ck(f(('bool', 'address[]', 'uint152'), D), T[bo, A, i])  # fixed case 32733
ck(f(('bool', 'address[]', 'uint16'), D), T[bo, A, i])  # fixed case 32734
ck(f(('bool', 'address[]', 'uint160'), D), T[bo, A, i])  # fixed case 32735
ck(f(('bool', 'address[]', 'uint168'), D), T[bo, A, i])  # fixed case 32736
ck(f(('bool', 'address[]', 'uint176'), D), T[bo, A, i])  # fixed case 32737
ck(f(('bool', 'address[]', 'uint184'), D), T[bo, A, i])  # fixed case 32738
ck(f(('bool', 'address[]', 'uint192'), D), T[bo, A, i])  # fixed case 32739
ck(f(('bool', 'address[]', 'uint200'), D), T[bo, A, i])  # fixed case 32740
ck(f(('bool', 'address[]', 'uint208'), D), T[bo, A, i])  # fixed case 32741
ck(f(('bool', 'address[]', 'uint216'), D), T[bo, A, i])  # fixed case 32742
ck(f(('bool', 'address[]', 'uint224'), D), T[bo, A, i])  # fixed case 32743
ck(f(('bool', 'address[]', 'uint232'), D), T[bo, A, i])  # fixed case 32744
ck(f(('bool', 'address[]', 'uint24'), D), T[bo, A, i])  # fixed case 32745
ck(f(('bool', 'address[]', 'uint240'), D), T[bo, A, i])  # fixed case 32746
ck(f(('bool', 'address[]', 'uint248'), D), T[bo, A, i])  # fixed case 32747
ck(f(('bool', 'address[]', 'uint256'), D), T[bo, A, i])  # fixed case 32748
ck(f(('bool', 'address[]', 'uint32'), D), T[bo, A, i])  # fixed case 32749
ck(f(('bool', 'address[]', 'uint40'), D), T[bo, A, i])  # fixed case 32750
ck(f(('bool', 'address[]', 'uint48'), D), T[bo, A, i])  # fixed case 32751
ck(f(('bool', 'address[]', 'uint56'), D), T[bo, A, i])  # fixed case 32752
ck(f(('bool', 'address[]', 'uint64'), D), T[bo, A, i])  # fixed case 32753
ck(f(('bool', 'address[]', 'uint72'), D), T[bo, A, i])  # fixed case 32754
ck(f(('bool', 'address[]', 'uint8'), D), T[bo, A, i])  # fixed case 32755
ck(f(('bool', 'address[]', 'uint80'), D), T[bo, A, i])  # fixed case 32756
ck(f(('bool', 'address[]', 'uint88'), D), T[bo, A, i])  # fixed case 32757
ck(f(('bool', 'address[]', 'uint96'), D), T[bo, A, i])  # fixed case 32758
ck(f(('bool', 'address[]', '?'), D), T[bo, A, A])  # fixed case 32759
ck(f(('bool', 'bool', 'address'), D), T[bo, bo, HA])  # fixed case 32760
ck(f(('bool', 'bool', 'address[]'), D), T[bo, bo, A])  # fixed case 32761
ck(f(('bool', 'bool', 'bool'), D), T[bo, bo, bo])  # fixed case 32762
ck(f(('bool', 'bool', 'bool[]'), D), T[bo, bo, A])  # fixed case 32763
ck(f(('bool', 'bool', 'bytes'), D), T[bo, bo, by])  # fixed case 32764
ck(f(('bool', 'bool', 'bytes1'), D), T[bo, bo, by])  # fixed case 32765
ck(f(('bool', 'bool', 'bytes10'), D), T[bo, bo, by])  # fixed case 32766
ck(f(('bool', 'bool', 'bytes11'), D), T[bo, bo, by])  # fixed case 32767
ck(f(('bool', 'bool', 'bytes12'), D), T[bo, bo, by])  # fixed case 32768
ck(f(('bool', 'bool', 'bytes13'), D), T[bo, bo, by])  # fixed case 32769
ck(f(('bool', 'bool', 'bytes14'), D), T[bo, bo, by])  # fixed case 32770
ck(f(('bool', 'bool', 'bytes15'), D), T[bo, bo, by])  # fixed case 32771
ck(f(('bool', 'bool', 'bytes16'), D), T[bo, bo, by])  # fixed case 32772
ck(f(('bool', 'bool', 'bytes17'), D), T[bo, bo, by])  # fixed case 32773
ck(f(('bool', 'bool', 'bytes18'), D), T[bo, bo, by])  # fixed case 32774
ck(f(('bool', 'bool', 'bytes19'), D), T[bo, bo, by])  # fixed case 32775
ck(f(('bool', 'bool', 'bytes2'), D), T[bo, bo, by])  # fixed case 32776
ck(f(('bool', 'bool', 'bytes20'), D), T[bo, bo, by])  # fixed case 32777
ck(f(('bool', 'bool', 'bytes21'), D), T[bo, bo, by])  # fixed case 32778
ck(f(('bool', 'bool', 'bytes22'), D), T[bo, bo, by])  # fixed case 32779
ck(f(('bool', 'bool', 'bytes23'), D), T[bo, bo, by])  # fixed case 32780
ck(f(('bool', 'bool', 'bytes24'), D), T[bo, bo, by])  # fixed case 32781
ck(f(('bool', 'bool', 'bytes25'), D), T[bo, bo, by])  # fixed case 32782
ck(f(('bool', 'bool', 'bytes26'), D), T[bo, bo, by])  # fixed case 32783
ck(f(('bool', 'bool', 'bytes27'), D), T[bo, bo, by])  # fixed case 32784
ck(f(('bool', 'bool', 'bytes28'), D), T[bo, bo, by])  # fixed case 32785
ck(f(('bool', 'bool', 'bytes29'), D), T[bo, bo, by])  # fixed case 32786
ck(f(('bool', 'bool', 'bytes3'), D), T[bo, bo, by])  # fixed case 32787
ck(f(('bool', 'bool', 'bytes30'), D), T[bo, bo, by])  # fixed case 32788
ck(f(('bool', 'bool', 'bytes31'), D), T[bo, bo, by])  # fixed case 32789
ck(f(('bool', 'bool', 'bytes32'), D), T[bo, bo, by])  # fixed case 32790
ck(f(('bool', 'bool', 'bytes4'), D), T[bo, bo, by])  # fixed case 32791
ck(f(('bool', 'bool', 'bytes5'), D), T[bo, bo, by])  # fixed case 32792
ck(f(('bool', 'bool', 'bytes6'), D), T[bo, bo, by])  # fixed case 32793
ck(f(('bool', 'bool', 'bytes7'), D), T[bo, bo, by])  # fixed case 32794
ck(f(('bool', 'bool', 'bytes8'), D), T[bo, bo, by])  # fixed case 32795
ck(f(('bool', 'bool', 'bytes9'), D), T[bo, bo, by])  # fixed case 32796
ck(f(('bool', 'bool', 'int104'), D), T[bo, bo, i])  # fixed case 32797
ck(f(('bool', 'bool', 'int112'), D), T[bo, bo, i])  # fixed case 32798
ck(f(('bool', 'bool', 'int120'), D), T[bo, bo, i])  # fixed case 32799
ck(f(('bool', 'bool', 'int128'), D), T[bo, bo, i])  # fixed case 32800
ck(f(('bool', 'bool', 'int136'), D), T[bo, bo, i])  # fixed case 32801
ck(f(('bool', 'bool', 'int144'), D), T[bo, bo, i])  # fixed case 32802
ck(f(('bool', 'bool', 'int152'), D), T[bo, bo, i])  # fixed case 32803
ck(f(('bool', 'bool', 'int16'), D), T[bo, bo, i])  # fixed case 32804
ck(f(('bool', 'bool', 'int160'), D), T[bo, bo, i])  # fixed case 32805
ck(f(('bool', 'bool', 'int168'), D), T[bo, bo, i])  # fixed case 32806
ck(f(('bool', 'bool', 'int176'), D), T[bo, bo, i])  # fixed case 32807
ck(f(('bool', 'bool', 'int184'), D), T[bo, bo, i])  # fixed case 32808
ck(f(('bool', 'bool', 'int192'), D), T[bo, bo, i])  # fixed case 32809
ck(f(('bool', 'bool', 'int200'), D), T[bo, bo, i])  # fixed case 32810
ck(f(('bool', 'bool', 'int208'), D), T[bo, bo, i])  # fixed case 32811
ck(f(('bool', 'bool', 'int216'), D), T[bo, bo, i])  # fixed case 32812
ck(f(('bool', 'bool', 'int224'), D), T[bo, bo, i])  # fixed case 32813
ck(f(('bool', 'bool', 'int232'), D), T[bo, bo, i])  # fixed case 32814
ck(f(('bool', 'bool', 'int24'), D), T[bo, bo, i])  # fixed case 32815
ck(f(('bool', 'bool', 'int240'), D), T[bo, bo, i])  # fixed case 32816
ck(f(('bool', 'bool', 'int248'), D), T[bo, bo, i])  # fixed case 32817
ck(f(('bool', 'bool', 'int256'), D), T[bo, bo, i])  # fixed case 32818
ck(f(('bool', 'bool', 'int32'), D), T[bo, bo, i])  # fixed case 32819
ck(f(('bool', 'bool', 'int40'), D), T[bo, bo, i])  # fixed case 32820
ck(f(('bool', 'bool', 'int48'), D), T[bo, bo, i])  # fixed case 32821
ck(f(('bool', 'bool', 'int56'), D), T[bo, bo, i])  # fixed case 32822
ck(f(('bool', 'bool', 'int64'), D), T[bo, bo, i])  # fixed case 32823
ck(f(('bool', 'bool', 'int72'), D), T[bo, bo, i])  # fixed case 32824
ck(f(('bool', 'bool', 'int8'), D), T[bo, bo, i])  # fixed case 32825
ck(f(('bool', 'bool', 'int80'), D), T[bo, bo, i])  # fixed case 32826
ck(f(('bool', 'bool', 'int88'), D), T[bo, bo, i])  # fixed case 32827
ck(f(('bool', 'bool', 'int96'), D), T[bo, bo, i])  # fixed case 32828
ck(f(('bool', 'bool', 'string'), D), T[bo, bo, s])  # fixed case 32829
ck(f(('bool', 'bool', 'string[]'), D), T[bo, bo, A])  # fixed case 32830
ck(f(('bool', 'bool', 'uint104'), D), T[bo, bo, i])  # fixed case 32831
ck(f(('bool', 'bool', 'uint112'), D), T[bo, bo, i])  # fixed case 32832
ck(f(('bool', 'bool', 'uint120'), D), T[bo, bo, i])  # fixed case 32833
ck(f(('bool', 'bool', 'uint128'), D), T[bo, bo, i])  # fixed case 32834
ck(f(('bool', 'bool', 'uint136'), D), T[bo, bo, i])  # fixed case 32835
ck(f(('bool', 'bool', 'uint144'), D), T[bo, bo, i])  # fixed case 32836
ck(f(('bool', 'bool', 'uint152'), D), T[bo, bo, i])  # fixed case 32837
ck(f(('bool', 'bool', 'uint16'), D), T[bo, bo, i])  # fixed case 32838
ck(f(('bool', 'bool', 'uint160'), D), T[bo, bo, i])  # fixed case 32839
ck(f(('bool', 'bool', 'uint168'), D), T[bo, bo, i])  # fixed case 32840
ck(f(('bool', 'bool', 'uint176'), D), T[bo, bo, i])  # fixed case 32841
ck(f(('bool', 'bool', 'uint184'), D), T[bo, bo, i])  # fixed case 32842
ck(f(('bool', 'bool', 'uint192'), D), T[bo, bo, i])  # fixed case 32843
ck(f(('bool', 'bool', 'uint200'), D), T[bo, bo, i])  # fixed case 32844
ck(f(('bool', 'bool', 'uint208'), D), T[bo, bo, i])  # fixed case 32845
ck(f(('bool', 'bool', 'uint216'), D), T[bo, bo, i])  # fixed case 32846
ck(f(('bool', 'bool', 'uint224'), D), T[bo, bo, i])  # fixed case 32847
ck(f(('bool', 'bool', 'uint232'), D), T[bo, bo, i])  # fixed case 32848
ck(f(('bool', 'bool', 'uint24'), D), T[bo, bo, i])  # fixed case 32849
ck(f(('bool', 'bool', 'uint240'), D), T[bo, bo, i])  # fixed case 32850
ck(f(('bool', 'bool', 'uint248'), D), T[bo, bo, i])  # fixed case 32851
ck(f(('bool', 'bool', 'uint256'), D), T[bo, bo, i])  # fixed case 32852
ck(f(('bool', 'bool', 'uint32'), D), T[bo, bo, i])  # fixed case 32853
ck(f(('bool', 'bool', 'uint40'), D), T[bo, bo, i])  # fixed case 32854
ck(f(('bool', 'bool', 'uint48'), D), T[bo, bo, i])  # fixed case 32855
ck(f(('bool', 'bool', 'uint56'), D), T[bo, bo, i])  # fixed case 32856
ck(f(('bool', 'bool', 'uint64'), D), T[bo, bo, i])  # fixed case 32857
ck(f(('bool', 'bool', 'uint72'), D), T[bo, bo, i])  # fixed case 32858
ck(f(('bool', 'bool', 'uint8'), D), T[bo, bo, i])  # fixed case 32859
ck(f(('bool', 'bool', 'uint80'), D), T[bo, bo, i])  # fixed case 32860
ck(f(('bool', 'bool', 'uint88'), D), T[bo, bo, i])  # fixed case 32861
ck(f(('bool', 'bool', 'uint96'), D), T[bo, bo, i])  # fixed case 32862
ck(f(('bool', 'bool', '?'), D), T[bo, bo, A])  # fixed case 32863
ck(f(('bool', 'bool[]', 'address'), D), T[bo, A, HA])  # fixed case 32864
ck(f(('bool', 'bool[]', 'address[]'), D), T[bo, A, A])  # fixed case 32865
ck(f(('bool', 'bool[]', 'bool'), D), T[bo, A, bo])  # fixed case 32866
ck(f(('bool', 'bool[]', 'bool[]'), D), T[bo, A, A])  # fixed case 32867
ck(f(('bool', 'bool[]', 'bytes'), D), T[bo, A, by])  # fixed case 32868
ck(f(('bool', 'bool[]', 'bytes1'), D), T[bo, A, by])  # fixed case 32869
ck(f(('bool', 'bool[]', 'bytes10'), D), T[bo, A, by])  # fixed case 32870
ck(f(('bool', 'bool[]', 'bytes11'), D), T[bo, A, by])  # fixed case 32871
ck(f(('bool', 'bool[]', 'bytes12'), D), T[bo, A, by])  # fixed case 32872
ck(f(('bool', 'bool[]', 'bytes13'), D), T[bo, A, by])  # fixed case 32873
ck(f(('bool', 'bool[]', 'bytes14'), D), T[bo, A, by])  # fixed case 32874
ck(f(('bool', 'bool[]', 'bytes15'), D), T[bo, A, by])  # fixed case 32875
ck(f(('bool', 'bool[]', 'bytes16'), D), T[bo, A, by])  # fixed case 32876
ck(f(('bool', 'bool[]', 'bytes17'), D), T[bo, A, by])  # fixed case 32877
ck(f(('bool', 'bool[]', 'bytes18'), D), T[bo, A, by])  # fixed case 32878
ck(f(('bool', 'bool[]', 'bytes19'), D), T[bo, A, by])  # fixed case 32879
ck(f(('bool', 'bool[]', 'bytes2'), D), T[bo, A, by])  # fixed case 32880
ck(f(('bool', 'bool[]', 'bytes20'), D), T[bo, A, by])  # fixed case 32881
ck(f(('bool', 'bool[]', 'bytes21'), D), T[bo, A, by])  # fixed case 32882
ck(f(('bool', 'bool[]', 'bytes22'), D), T[bo, A, by])  # fixed case 32883
ck(f(('bool', 'bool[]', 'bytes23'), D), T[bo, A, by])  # fixed case 32884
ck(f(('bool', 'bool[]', 'bytes24'), D), T[bo, A, by])  # fixed case 32885
ck(f(('bool', 'bool[]', 'bytes25'), D), T[bo, A, by])  # fixed case 32886
ck(f(('bool', 'bool[]', 'bytes26'), D), T[bo, A, by])  # fixed case 32887
ck(f(('bool', 'bool[]', 'bytes27'), D), T[bo, A, by])  # fixed case 32888
ck(f(('bool', 'bool[]', 'bytes28'), D), T[bo, A, by])  # fixed case 32889
ck(f(('bool', 'bool[]', 'bytes29'), D), T[bo, A, by])  # fixed case 32890
ck(f(('bool', 'bool[]', 'bytes3'), D), T[bo, A, by])  # fixed case 32891
ck(f(('bool', 'bool[]', 'bytes30'), D), T[bo, A, by])  # fixed case 32892
ck(f(('bool', 'bool[]', 'bytes31'), D), T[bo, A, by])  # fixed case 32893
ck(f(('bool', 'bool[]', 'bytes32'), D), T[bo, A, by])  # fixed case 32894
ck(f(('bool', 'bool[]', 'bytes4'), D), T[bo, A, by])  # fixed case 32895
ck(f(('bool', 'bool[]', 'bytes5'), D), T[bo, A, by])  # fixed case 32896
ck(f(('bool', 'bool[]', 'bytes6'), D), T[bo, A, by])  # fixed case 32897
ck(f(('bool', 'bool[]', 'bytes7'), D), T[bo, A, by])  # fixed case 32898
ck(f(('bool', 'bool[]', 'bytes8'), D), T[bo, A, by])  # fixed case 32899
ck(f(('bool', 'bool[]', 'bytes9'), D), T[bo, A, by])  # fixed case 32900
ck(f(('bool', 'bool[]', 'int104'), D), T[bo, A, i])  # fixed case 32901
ck(f(('bool', 'bool[]', 'int112'), D), T[bo, A, i])  # fixed case 32902
ck(f(('bool', 'bool[]', 'int120'), D), T[bo, A, i])  # fixed case 32903
ck(f(('bool', 'bool[]', 'int128'), D), T[bo, A, i])  # fixed case 32904
ck(f(('bool', 'bool[]', 'int136'), D), T[bo, A, i])  # fixed case 32905
ck(f(('bool', 'bool[]', 'int144'), D), T[bo, A, i])  # fixed case 32906
ck(f(('bool', 'bool[]', 'int152'), D), T[bo, A, i])  # fixed case 32907
ck(f(('bool', 'bool[]', 'int16'), D), T[bo, A, i])  # fixed case 32908
ck(f(('bool', 'bool[]', 'int160'), D), T[bo, A, i])  # fixed case 32909
ck(f(('bool', 'bool[]', 'int168'), D), T[bo, A, i])  # fixed case 32910
ck(f(('bool', 'bool[]', 'int176'), D), T[bo, A, i])  # fixed case 32911
ck(f(('bool', 'bool[]', 'int184'), D), T[bo, A, i])  # fixed case 32912
ck(f(('bool', 'bool[]', 'int192'), D), T[bo, A, i])  # fixed case 32913
ck(f(('bool', 'bool[]', 'int200'), D), T[bo, A, i])  # fixed case 32914
ck(f(('bool', 'bool[]', 'int208'), D), T[bo, A, i])  # fixed case 32915
ck(f(('bool', 'bool[]', 'int216'), D), T[bo, A, i])  # fixed case 32916
ck(f(('bool', 'bool[]', 'int224'), D), T[bo, A, i])  # fixed case 32917
ck(f(('bool', 'bool[]', 'int232'), D), T[bo, A, i])  # fixed case 32918
ck(f(('bool', 'bool[]', 'int24'), D), T[bo, A, i])  # fixed case 32919
ck(f(('bool', 'bool[]', 'int240'), D), T[bo, A, i])  # fixed case 32920
ck(f(('bool', 'bool[]', 'int248'), D), T[bo, A, i])  # fixed case 32921
ck(f(('bool', 'bool[]', 'int256'), D), T[bo, A, i])  # fixed case 32922
ck(f(('bool', 'bool[]', 'int32'), D), T[bo, A, i])  # fixed case 32923
ck(f(('bool', 'bool[]', 'int40'), D), T[bo, A, i])  # fixed case 32924
ck(f(('bool', 'bool[]', 'int48'), D), T[bo, A, i])  # fixed case 32925
ck(f(('bool', 'bool[]', 'int56'), D), T[bo, A, i])  # fixed case 32926
ck(f(('bool', 'bool[]', 'int64'), D), T[bo, A, i])  # fixed case 32927
ck(f(('bool', 'bool[]', 'int72'), D), T[bo, A, i])  # fixed case 32928
ck(f(('bool', 'bool[]', 'int8'), D), T[bo, A, i])  # fixed case 32929
ck(f(('bool', 'bool[]', 'int80'), D), T[bo, A, i])  # fixed case 32930
ck(f(('bool', 'bool[]', 'int88'), D), T[bo, A, i])  # fixed case 32931
ck(f(('bool', 'bool[]', 'int96'), D), T[bo, A, i])  # fixed case 32932
ck(f(('bool', 'bool[]', 'string'), D), T[bo, A, s])  # fixed case 32933
ck(f(('bool', 'bool[]', 'string[]'), D), T[bo, A, A])  # fixed case 32934
ck(f(('bool', 'bool[]', 'uint104'), D), T[bo, A, i])  # fixed case 32935
ck(f(('bool', 'bool[]', 'uint112'), D), T[bo, A, i])  # fixed case 32936
ck(f(('bool', 'bool[]', 'uint120'), D), T[bo, A, i])  # fixed case 32937
ck(f(('bool', 'bool[]', 'uint128'), D), T[bo, A, i])  # fixed case 32938
ck(f(('bool', 'bool[]', 'uint136'), D), T[bo, A, i])  # fixed case 32939
ck(f(('bool', 'bool[]', 'uint144'), D), T[bo, A, i])  # fixed case 32940
ck(f(('bool', 'bool[]', 'uint152'), D), T[bo, A, i])  # fixed case 32941
ck(f(('bool', 'bool[]', 'uint16'), D), T[bo, A, i])  # fixed case 32942
ck(f(('bool', 'bool[]', 'uint160'), D), T[bo, A, i])  # fixed case 32943
ck(f(('bool', 'bool[]', 'uint168'), D), T[bo, A, i])  # fixed case 32944
ck(f(('bool', 'bool[]', 'uint176'), D), T[bo, A, i])  # fixed case 32945
ck(f(('bool', 'bool[]', 'uint184'), D), T[bo, A, i])  # fixed case 32946
ck(f(('bool', 'bool[]', 'uint192'), D), T[bo, A, i])  # fixed case 32947
ck(f(('bool', 'bool[]', 'uint200'), D), T[bo, A, i])  # fixed case 32948
ck(f(('bool', 'bool[]', 'uint208'), D), T[bo, A, i])  # fixed case 32949
ck(f(('bool', 'bool[]', 'uint216'), D), T[bo, A, i])  # fixed case 32950
ck(f(('bool', 'bool[]', 'uint224'), D), T[bo, A, i])  # fixed case 32951
ck(f(('bool', 'bool[]', 'uint232'), D), T[bo, A, i])  # fixed case 32952
ck(f(('bool', 'bool[]', 'uint24'), D), T[bo, A, i])  # fixed case 32953
ck(f(('bool', 'bool[]', 'uint240'), D), T[bo, A, i])  # fixed case 32954
ck(f(('bool', 'bool[]', 'uint248'), D), T[bo, A, i])  # fixed case 32955
ck(f(('bool', 'bool[]', 'uint256'), D), T[bo, A, i])  # fixed case 32956
ck(f(('bool', 'bool[]', 'uint32'), D), T[bo, A, i])  # fixed case 32957
ck(f(('bool', 'bool[]', 'uint40'), D), T[bo, A, i])  # fixed case 32958
ck(f(('bool', 'bool[]', 'uint48'), D), T[bo, A, i])  # fixed case 32959
ck(f(('bool', 'bool[]', 'uint56'), D), T[bo, A, i])  # fixed case 32960
ck(f(('bool', 'bool[]', 'uint64'), D), T[bo, A, i])  # fixed case 32961
ck(f(('bool', 'bool[]', 'uint72'), D), T[bo, A, i])  # fixed case 32962
ck(f(('bool', 'bool[]', 'uint8'), D), T[bo, A, i])  # fixed case 32963
ck(f(('bool', 'bool[]', 'uint80'), D), T[bo, A, i])  # fixed case 32964
ck(f(('bool', 'bool[]', 'uint88'), D), T[bo, A, i])  # fixed case 32965
ck(f(('bool', 'bool[]', 'uint96'), D), T[bo, A, i])  # fixed case 32966
ck(f(('bool', 'bool[]', '?'), D), T[bo, A, A])  # fixed case 32967
ck(f(('bool', 'bytes', 'address'), D), T[bo, by, HA])  # fixed case 32968
ck(f(('bool', 'bytes', 'address[]'), D), T[bo, by, A])  # fixed case 32969
ck(f(('bool', 'bytes', 'bool'), D), T[bo, by, bo])  # fixed case 32970
ck(f(('bool', 'bytes', 'bool[]'), D), T[bo, by, A])  # fixed case 32971
ck(f(('bool', 'bytes', 'bytes'), D), T[bo, by, by])  # fixed case 32972
ck(f(('bool', 'bytes', 'bytes1'), D), T[bo, by, by])  # fixed case 32973
ck(f(('bool', 'bytes', 'bytes10'), D), T[bo, by, by])  # fixed case 32974
ck(f(('bool', 'bytes', 'bytes11'), D), T[bo, by, by])  # fixed case 32975
ck(f(('bool', 'bytes', 'bytes12'), D), T[bo, by, by])  # fixed case 32976
ck(f(('bool', 'bytes', 'bytes13'), D), T[bo, by, by])  # fixed case 32977
ck(f(('bool', 'bytes', 'bytes14'), D), T[bo, by, by])  # fixed case 32978
ck(f(('bool', 'bytes', 'bytes15'), D), T[bo, by, by])  # fixed case 32979
ck(f(('bool', 'bytes', 'bytes16'), D), T[bo, by, by])  # fixed case 32980
ck(f(('bool', 'bytes', 'bytes17'), D), T[bo, by, by])  # fixed case 32981
ck(f(('bool', 'bytes', 'bytes18'), D), T[bo, by, by])  # fixed case 32982
ck(f(('bool', 'bytes', 'bytes19'), D), T[bo, by, by])  # fixed case 32983
ck(f(('bool', 'bytes', 'bytes2'), D), T[bo, by, by])  # fixed case 32984
ck(f(('bool', 'bytes', 'bytes20'), D), T[bo, by, by])  # fixed case 32985
ck(f(('bool', 'bytes', 'bytes21'), D), T[bo, by, by])  # fixed case 32986
ck(f(('bool', 'bytes', 'bytes22'), D), T[bo, by, by])  # fixed case 32987
ck(f(('bool', 'bytes', 'bytes23'), D), T[bo, by, by])  # fixed case 32988
ck(f(('bool', 'bytes', 'bytes24'), D), T[bo, by, by])  # fixed case 32989
ck(f(('bool', 'bytes', 'bytes25'), D), T[bo, by, by])  # fixed case 32990
ck(f(('bool', 'bytes', 'bytes26'), D), T[bo, by, by])  # fixed case 32991
ck(f(('bool', 'bytes', 'bytes27'), D), T[bo, by, by])  # fixed case 32992
ck(f(('bool', 'bytes', 'bytes28'), D), T[bo, by, by])  # fixed case 32993
ck(f(('bool', 'bytes', 'bytes29'), D), T[bo, by, by])  # fixed case 32994
ck(f(('bool', 'bytes', 'bytes3'), D), T[bo, by, by])  # fixed case 32995
ck(f(('bool', 'bytes', 'bytes30'), D), T[bo, by, by])  # fixed case 32996
ck(f(('bool', 'bytes', 'bytes31'), D), T[bo, by, by])  # fixed case 32997
ck(f(('bool', 'bytes', 'bytes32'), D), T[bo, by, by])  # fixed case 32998
ck(f(('bool', 'bytes', 'bytes4'), D), T[bo, by, by])  # fixed case 32999
ck(f(('bool', 'bytes', 'bytes5'), D), T[bo, by, by])  # fixed case 33000
ck(f(('bool', 'bytes', 'bytes6'), D), T[bo, by, by])  # fixed case 33001
ck(f(('bool', 'bytes', 'bytes7'), D), T[bo, by, by])  # fixed case 33002
ck(f(('bool', 'bytes', 'bytes8'), D), T[bo, by, by])  # fixed case 33003
ck(f(('bool', 'bytes', 'bytes9'), D), T[bo, by, by])  # fixed case 33004
ck(f(('bool', 'bytes', 'int104'), D), T[bo, by, i])  # fixed case 33005
ck(f(('bool', 'bytes', 'int112'), D), T[bo, by, i])  # fixed case 33006
ck(f(('bool', 'bytes', 'int120'), D), T[bo, by, i])  # fixed case 33007
ck(f(('bool', 'bytes', 'int128'), D), T[bo, by, i])  # fixed case 33008
ck(f(('bool', 'bytes', 'int136'), D), T[bo, by, i])  # fixed case 33009
ck(f(('bool', 'bytes', 'int144'), D), T[bo, by, i])  # fixed case 33010
ck(f(('bool', 'bytes', 'int152'), D), T[bo, by, i])  # fixed case 33011
ck(f(('bool', 'bytes', 'int16'), D), T[bo, by, i])  # fixed case 33012
ck(f(('bool', 'bytes', 'int160'), D), T[bo, by, i])  # fixed case 33013
ck(f(('bool', 'bytes', 'int168'), D), T[bo, by, i])  # fixed case 33014
ck(f(('bool', 'bytes', 'int176'), D), T[bo, by, i])  # fixed case 33015
ck(f(('bool', 'bytes', 'int184'), D), T[bo, by, i])  # fixed case 33016
ck(f(('bool', 'bytes', 'int192'), D), T[bo, by, i])  # fixed case 33017
ck(f(('bool', 'bytes', 'int200'), D), T[bo, by, i])  # fixed case 33018
ck(f(('bool', 'bytes', 'int208'), D), T[bo, by, i])  # fixed case 33019
ck(f(('bool', 'bytes', 'int216'), D), T[bo, by, i])  # fixed case 33020
ck(f(('bool', 'bytes', 'int224'), D), T[bo, by, i])  # fixed case 33021
ck(f(('bool', 'bytes', 'int232'), D), T[bo, by, i])  # fixed case 33022
ck(f(('bool', 'bytes', 'int24'), D), T[bo, by, i])  # fixed case 33023
ck(f(('bool', 'bytes', 'int240'), D), T[bo, by, i])  # fixed case 33024
ck(f(('bool', 'bytes', 'int248'), D), T[bo, by, i])  # fixed case 33025
ck(f(('bool', 'bytes', 'int256'), D), T[bo, by, i])  # fixed case 33026
ck(f(('bool', 'bytes', 'int32'), D), T[bo, by, i])  # fixed case 33027
ck(f(('bool', 'bytes', 'int40'), D), T[bo, by, i])  # fixed case 33028
ck(f(('bool', 'bytes', 'int48'), D), T[bo, by, i])  # fixed case 33029
ck(f(('bool', 'bytes', 'int56'), D), T[bo, by, i])  # fixed case 33030
ck(f(('bool', 'bytes', 'int64'), D), T[bo, by, i])  # fixed case 33031
ck(f(('bool', 'bytes', 'int72'), D), T[bo, by, i])  # fixed case 33032
ck(f(('bool', 'bytes', 'int8'), D), T[bo, by, i])  # fixed case 33033
ck(f(('bool', 'bytes', 'int80'), D), T[bo, by, i])  # fixed case 33034
ck(f(('bool', 'bytes', 'int88'), D), T[bo, by, i])  # fixed case 33035
ck(f(('bool', 'bytes', 'int96'), D), T[bo, by, i])  # fixed case 33036
ck(f(('bool', 'bytes', 'string'), D), T[bo, by, s])  # fixed case 33037
ck(f(('bool', 'bytes', 'string[]'), D), T[bo, by, A])  # fixed case 33038
ck(f(('bool', 'bytes', 'uint104'), D), T[bo, by, i])  # fixed case 33039
ck(f(('bool', 'bytes', 'uint112'), D), T[bo, by, i])  # fixed case 33040
ck(f(('bool', 'bytes', 'uint120'), D), T[bo, by, i])  # fixed case 33041
ck(f(('bool', 'bytes', 'uint128'), D), T[bo, by, i])  # fixed case 33042
ck(f(('bool', 'bytes', 'uint136'), D), T[bo, by, i])  # fixed case 33043
ck(f(('bool', 'bytes', 'uint144'), D), T[bo, by, i])  # fixed case 33044
ck(f(('bool', 'bytes', 'uint152'), D), T[bo, by, i])  # fixed case 33045
ck(f(('bool', 'bytes', 'uint16'), D), T[bo, by, i])  # fixed case 33046
ck(f(('bool', 'bytes', 'uint160'), D), T[bo, by, i])  # fixed case 33047
ck(f(('bool', 'bytes', 'uint168'), D), T[bo, by, i])  # fixed case 33048
ck(f(('bool', 'bytes', 'uint176'), D), T[bo, by, i])  # fixed case 33049
ck(f(('bool', 'bytes', 'uint184'), D), T[bo, by, i])  # fixed case 33050
ck(f(('bool', 'bytes', 'uint192'), D), T[bo, by, i])  # fixed case 33051
ck(f(('bool', 'bytes', 'uint200'), D), T[bo, by, i])  # fixed case 33052
ck(f(('bool', 'bytes', 'uint208'), D), T[bo, by, i])  # fixed case 33053
ck(f(('bool', 'bytes', 'uint216'), D), T[bo, by, i])  # fixed case 33054
ck(f(('bool', 'bytes', 'uint224'), D), T[bo, by, i])  # fixed case 33055
ck(f(('bool', 'bytes', 'uint232'), D), T[bo, by, i])  # fixed case 33056
ck(f(('bool', 'bytes', 'uint24'), D), T[bo, by, i])  # fixed case 33057
ck(f(('bool', 'bytes', 'uint240'), D), T[bo, by, i])  # fixed case 33058
ck(f(('bool', 'bytes', 'uint248'), D), T[bo, by, i])  # fixed case 33059
ck(f(('bool', 'bytes', 'uint256'), D), T[bo, by, i])  # fixed case 33060
ck(f(('bool', 'bytes', 'uint32'), D), T[bo, by, i])  # fixed case 33061
ck(f(('bool', 'bytes', 'uint40'), D), T[bo, by, i])  # fixed case 33062
ck(f(('bool', 'bytes', 'uint48'), D), T[bo, by, i])  # fixed case 33063
ck(f(('bool', 'bytes', 'uint56'), D), T[bo, by, i])  # fixed case 33064
ck(f(('bool', 'bytes', 'uint64'), D), T[bo, by, i])  # fixed case 33065
ck(f(('bool', 'bytes', 'uint72'), D), T[bo, by, i])  # fixed case 33066
ck(f(('bool', 'bytes', 'uint8'), D), T[bo, by, i])  # fixed case 33067
ck(f(('bool', 'bytes', 'uint80'), D), T[bo, by, i])  # fixed case 33068
ck(f(('bool', 'bytes', 'uint88'), D), T[bo, by, i])  # fixed case 33069
ck(f(('bool', 'bytes', 'uint96'), D), T[bo, by, i])  # fixed case 33070
ck(f(('bool', 'bytes', '?'), D), T[bo, by, A])  # fixed case 33071
ck(f(('bool', 'bytes1', 'address'), D), T[bo, by, HA])  # fixed case 33072
ck(f(('bool', 'bytes1', 'address[]'), D), T[bo, by, A])  # fixed case 33073
ck(f(('bool', 'bytes1', 'bool'), D), T[bo, by, bo])  # fixed case 33074
ck(f(('bool', 'bytes1', 'bool[]'), D), T[bo, by, A])  # fixed case 33075
ck(f(('bool', 'bytes1', 'bytes'), D), T[bo, by, by])  # fixed case 33076
ck(f(('bool', 'bytes1', 'bytes1'), D), T[bo, by, by])  # fixed case 33077
ck(f(('bool', 'bytes1', 'bytes10'), D), T[bo, by, by])  # fixed case 33078
ck(f(('bool', 'bytes1', 'bytes11'), D), T[bo, by, by])  # fixed case 33079
ck(f(('bool', 'bytes1', 'bytes12'), D), T[bo, by, by])  # fixed case 33080
ck(f(('bool', 'bytes1', 'bytes13'), D), T[bo, by, by])  # fixed case 33081
ck(f(('bool', 'bytes1', 'bytes14'), D), T[bo, by, by])  # fixed case 33082
ck(f(('bool', 'bytes1', 'bytes15'), D), T[bo, by, by])  # fixed case 33083
ck(f(('bool', 'bytes1', 'bytes16'), D), T[bo, by, by])  # fixed case 33084
ck(f(('bool', 'bytes1', 'bytes17'), D), T[bo, by, by])  # fixed case 33085
ck(f(('bool', 'bytes1', 'bytes18'), D), T[bo, by, by])  # fixed case 33086
ck(f(('bool', 'bytes1', 'bytes19'), D), T[bo, by, by])  # fixed case 33087
ck(f(('bool', 'bytes1', 'bytes2'), D), T[bo, by, by])  # fixed case 33088
ck(f(('bool', 'bytes1', 'bytes20'), D), T[bo, by, by])  # fixed case 33089
ck(f(('bool', 'bytes1', 'bytes21'), D), T[bo, by, by])  # fixed case 33090
ck(f(('bool', 'bytes1', 'bytes22'), D), T[bo, by, by])  # fixed case 33091
ck(f(('bool', 'bytes1', 'bytes23'), D), T[bo, by, by])  # fixed case 33092
ck(f(('bool', 'bytes1', 'bytes24'), D), T[bo, by, by])  # fixed case 33093
ck(f(('bool', 'bytes1', 'bytes25'), D), T[bo, by, by])  # fixed case 33094
ck(f(('bool', 'bytes1', 'bytes26'), D), T[bo, by, by])  # fixed case 33095
ck(f(('bool', 'bytes1', 'bytes27'), D), T[bo, by, by])  # fixed case 33096
ck(f(('bool', 'bytes1', 'bytes28'), D), T[bo, by, by])  # fixed case 33097
ck(f(('bool', 'bytes1', 'bytes29'), D), T[bo, by, by])  # fixed case 33098
ck(f(('bool', 'bytes1', 'bytes3'), D), T[bo, by, by])  # fixed case 33099
ck(f(('bool', 'bytes1', 'bytes30'), D), T[bo, by, by])  # fixed case 33100
ck(f(('bool', 'bytes1', 'bytes31'), D), T[bo, by, by])  # fixed case 33101
ck(f(('bool', 'bytes1', 'bytes32'), D), T[bo, by, by])  # fixed case 33102
ck(f(('bool', 'bytes1', 'bytes4'), D), T[bo, by, by])  # fixed case 33103
ck(f(('bool', 'bytes1', 'bytes5'), D), T[bo, by, by])  # fixed case 33104
ck(f(('bool', 'bytes1', 'bytes6'), D), T[bo, by, by])  # fixed case 33105
ck(f(('bool', 'bytes1', 'bytes7'), D), T[bo, by, by])  # fixed case 33106
ck(f(('bool', 'bytes1', 'bytes8'), D), T[bo, by, by])  # fixed case 33107
ck(f(('bool', 'bytes1', 'bytes9'), D), T[bo, by, by])  # fixed case 33108
ck(f(('bool', 'bytes1', 'int104'), D), T[bo, by, i])  # fixed case 33109
ck(f(('bool', 'bytes1', 'int112'), D), T[bo, by, i])  # fixed case 33110
ck(f(('bool', 'bytes1', 'int120'), D), T[bo, by, i])  # fixed case 33111
ck(f(('bool', 'bytes1', 'int128'), D), T[bo, by, i])  # fixed case 33112
ck(f(('bool', 'bytes1', 'int136'), D), T[bo, by, i])  # fixed case 33113
ck(f(('bool', 'bytes1', 'int144'), D), T[bo, by, i])  # fixed case 33114
ck(f(('bool', 'bytes1', 'int152'), D), T[bo, by, i])  # fixed case 33115
ck(f(('bool', 'bytes1', 'int16'), D), T[bo, by, i])  # fixed case 33116
ck(f(('bool', 'bytes1', 'int160'), D), T[bo, by, i])  # fixed case 33117
ck(f(('bool', 'bytes1', 'int168'), D), T[bo, by, i])  # fixed case 33118
ck(f(('bool', 'bytes1', 'int176'), D), T[bo, by, i])  # fixed case 33119
ck(f(('bool', 'bytes1', 'int184'), D), T[bo, by, i])  # fixed case 33120
ck(f(('bool', 'bytes1', 'int192'), D), T[bo, by, i])  # fixed case 33121
ck(f(('bool', 'bytes1', 'int200'), D), T[bo, by, i])  # fixed case 33122
ck(f(('bool', 'bytes1', 'int208'), D), T[bo, by, i])  # fixed case 33123
ck(f(('bool', 'bytes1', 'int216'), D), T[bo, by, i])  # fixed case 33124
ck(f(('bool', 'bytes1', 'int224'), D), T[bo, by, i])  # fixed case 33125
ck(f(('bool', 'bytes1', 'int232'), D), T[bo, by, i])  # fixed case 33126
ck(f(('bool', 'bytes1', 'int24'), D), T[bo, by, i])  # fixed case 33127
ck(f(('bool', 'bytes1', 'int240'), D), T[bo, by, i])  # fixed case 33128
ck(f(('bool', 'bytes1', 'int248'), D), T[bo, by, i])  # fixed case 33129
ck(f(('bool', 'bytes1', 'int256'), D), T[bo, by, i])  # fixed case 33130
ck(f(('bool', 'bytes1', 'int32'), D), T[bo, by, i])  # fixed case 33131
ck(f(('bool', 'bytes1', 'int40'), D), T[bo, by, i])  # fixed case 33132
ck(f(('bool', 'bytes1', 'int48'), D), T[bo, by, i])  # fixed case 33133
ck(f(('bool', 'bytes1', 'int56'), D), T[bo, by, i])  # fixed case 33134
ck(f(('bool', 'bytes1', 'int64'), D), T[bo, by, i])  # fixed case 33135
ck(f(('bool', 'bytes1', 'int72'), D), T[bo, by, i])  # fixed case 33136
ck(f(('bool', 'bytes1', 'int8'), D), T[bo, by, i])  # fixed case 33137
ck(f(('bool', 'bytes1', 'int80'), D), T[bo, by, i])  # fixed case 33138
ck(f(('bool', 'bytes1', 'int88'), D), T[bo, by, i])  # fixed case 33139
ck(f(('bool', 'bytes1', 'int96'), D), T[bo, by, i])  # fixed case 33140
ck(f(('bool', 'bytes1', 'string'), D), T[bo, by, s])  # fixed case 33141
ck(f(('bool', 'bytes1', 'string[]'), D), T[bo, by, A])  # fixed case 33142
ck(f(('bool', 'bytes1', 'uint104'), D), T[bo, by, i])  # fixed case 33143
ck(f(('bool', 'bytes1', 'uint112'), D), T[bo, by, i])  # fixed case 33144
ck(f(('bool', 'bytes1', 'uint120'), D), T[bo, by, i])  # fixed case 33145
ck(f(('bool', 'bytes1', 'uint128'), D), T[bo, by, i])  # fixed case 33146
ck(f(('bool', 'bytes1', 'uint136'), D), T[bo, by, i])  # fixed case 33147
ck(f(('bool', 'bytes1', 'uint144'), D), T[bo, by, i])  # fixed case 33148
ck(f(('bool', 'bytes1', 'uint152'), D), T[bo, by, i])  # fixed case 33149
ck(f(('bool', 'bytes1', 'uint16'), D), T[bo, by, i])  # fixed case 33150
ck(f(('bool', 'bytes1', 'uint160'), D), T[bo, by, i])  # fixed case 33151
ck(f(('bool', 'bytes1', 'uint168'), D), T[bo, by, i])  # fixed case 33152
ck(f(('bool', 'bytes1', 'uint176'), D), T[bo, by, i])  # fixed case 33153
ck(f(('bool', 'bytes1', 'uint184'), D), T[bo, by, i])  # fixed case 33154
ck(f(('bool', 'bytes1', 'uint192'), D), T[bo, by, i])  # fixed case 33155
ck(f(('bool', 'bytes1', 'uint200'), D), T[bo, by, i])  # fixed case 33156
ck(f(('bool', 'bytes1', 'uint208'), D), T[bo, by, i])  # fixed case 33157
ck(f(('bool', 'bytes1', 'uint216'), D), T[bo, by, i])  # fixed case 33158
ck(f(('bool', 'bytes1', 'uint224'), D), T[bo, by, i])  # fixed case 33159
ck(f(('bool', 'bytes1', 'uint232'), D), T[bo, by, i])  # fixed case 33160
ck(f(('bool', 'bytes1', 'uint24'), D), T[bo, by, i])  # fixed case 33161
ck(f(('bool', 'bytes1', 'uint240'), D), T[bo, by, i])  # fixed case 33162
ck(f(('bool', 'bytes1', 'uint248'), D), T[bo, by, i])  # fixed case 33163
ck(f(('bool', 'bytes1', 'uint256'), D), T[bo, by, i])  # fixed case 33164
ck(f(('bool', 'bytes1', 'uint32'), D), T[bo, by, i])  # fixed case 33165
ck(f(('bool', 'bytes1', 'uint40'), D), T[bo, by, i])  # fixed case 33166
ck(f(('bool', 'bytes1', 'uint48'), D), T[bo, by, i])  # fixed case 33167
ck(f(('bool', 'bytes1', 'uint56'), D), T[bo, by, i])  # fixed case 33168
ck(f(('bool', 'bytes1', 'uint64'), D), T[bo, by, i])  # fixed case 33169
ck(f(('bool', 'bytes1', 'uint72'), D), T[bo, by, i])  # fixed case 33170
ck(f(('bool', 'bytes1', 'uint8'), D), T[bo, by, i])  # fixed case 33171
ck(f(('bool', 'bytes1', 'uint80'), D), T[bo, by, i])  # fixed case 33172
ck(f(('bool', 'bytes1', 'uint88'), D), T[bo, by, i])  # fixed case 33173
ck(f(('bool', 'bytes1', 'uint96'), D), T[bo, by, i])  # fixed case 33174
ck(f(('bool', 'bytes1', '?'), D), T[bo, by, A])  # fixed case 33175
ck(f(('bool', 'bytes10', 'address'), D), T[bo, by, HA])  # fixed case 33176
ck(f(('bool', 'bytes10', 'address[]'), D), T[bo, by, A])  # fixed case 33177
ck(f(('bool', 'bytes10', 'bool'), D), T[bo, by, bo])  # fixed case 33178
ck(f(('bool', 'bytes10', 'bool[]'), D), T[bo, by, A])  # fixed case 33179
ck(f(('bool', 'bytes10', 'bytes'), D), T[bo, by, by])  # fixed case 33180
ck(f(('bool', 'bytes10', 'bytes1'), D), T[bo, by, by])  # fixed case 33181
ck(f(('bool', 'bytes10', 'bytes10'), D), T[bo, by, by])  # fixed case 33182
ck(f(('bool', 'bytes10', 'bytes11'), D), T[bo, by, by])  # fixed case 33183
ck(f(('bool', 'bytes10', 'bytes12'), D), T[bo, by, by])  # fixed case 33184
ck(f(('bool', 'bytes10', 'bytes13'), D), T[bo, by, by])  # fixed case 33185
ck(f(('bool', 'bytes10', 'bytes14'), D), T[bo, by, by])  # fixed case 33186
ck(f(('bool', 'bytes10', 'bytes15'), D), T[bo, by, by])  # fixed case 33187
ck(f(('bool', 'bytes10', 'bytes16'), D), T[bo, by, by])  # fixed case 33188
ck(f(('bool', 'bytes10', 'bytes17'), D), T[bo, by, by])  # fixed case 33189
ck(f(('bool', 'bytes10', 'bytes18'), D), T[bo, by, by])  # fixed case 33190
ck(f(('bool', 'bytes10', 'bytes19'), D), T[bo, by, by])  # fixed case 33191
ck(f(('bool', 'bytes10', 'bytes2'), D), T[bo, by, by])  # fixed case 33192
ck(f(('bool', 'bytes10', 'bytes20'), D), T[bo, by, by])  # fixed case 33193
ck(f(('bool', 'bytes10', 'bytes21'), D), T[bo, by, by])  # fixed case 33194
ck(f(('bool', 'bytes10', 'bytes22'), D), T[bo, by, by])  # fixed case 33195
ck(f(('bool', 'bytes10', 'bytes23'), D), T[bo, by, by])  # fixed case 33196
ck(f(('bool', 'bytes10', 'bytes24'), D), T[bo, by, by])  # fixed case 33197
ck(f(('bool', 'bytes10', 'bytes25'), D), T[bo, by, by])  # fixed case 33198
ck(f(('bool', 'bytes10', 'bytes26'), D), T[bo, by, by])  # fixed case 33199
ck(f(('bool', 'bytes10', 'bytes27'), D), T[bo, by, by])  # fixed case 33200
ck(f(('bool', 'bytes10', 'bytes28'), D), T[bo, by, by])  # fixed case 33201
ck(f(('bool', 'bytes10', 'bytes29'), D), T[bo, by, by])  # fixed case 33202
ck(f(('bool', 'bytes10', 'bytes3'), D), T[bo, by, by])  # fixed case 33203
ck(f(('bool', 'bytes10', 'bytes30'), D), T[bo, by, by])  # fixed case 33204
ck(f(('bool', 'bytes10', 'bytes31'), D), T[bo, by, by])  # fixed case 33205
ck(f(('bool', 'bytes10', 'bytes32'), D), T[bo, by, by])  # fixed case 33206
ck(f(('bool', 'bytes10', 'bytes4'), D), T[bo, by, by])  # fixed case 33207
ck(f(('bool', 'bytes10', 'bytes5'), D), T[bo, by, by])  # fixed case 33208
ck(f(('bool', 'bytes10', 'bytes6'), D), T[bo, by, by])  # fixed case 33209
ck(f(('bool', 'bytes10', 'bytes7'), D), T[bo, by, by])  # fixed case 33210
ck(f(('bool', 'bytes10', 'bytes8'), D), T[bo, by, by])  # fixed case 33211
ck(f(('bool', 'bytes10', 'bytes9'), D), T[bo, by, by])  # fixed case 33212
ck(f(('bool', 'bytes10', 'int104'), D), T[bo, by, i])  # fixed case 33213
ck(f(('bool', 'bytes10', 'int112'), D), T[bo, by, i])  # fixed case 33214
ck(f(('bool', 'bytes10', 'int120'), D), T[bo, by, i])  # fixed case 33215
ck(f(('bool', 'bytes10', 'int128'), D), T[bo, by, i])  # fixed case 33216
ck(f(('bool', 'bytes10', 'int136'), D), T[bo, by, i])  # fixed case 33217
ck(f(('bool', 'bytes10', 'int144'), D), T[bo, by, i])  # fixed case 33218
ck(f(('bool', 'bytes10', 'int152'), D), T[bo, by, i])  # fixed case 33219
ck(f(('bool', 'bytes10', 'int16'), D), T[bo, by, i])  # fixed case 33220
ck(f(('bool', 'bytes10', 'int160'), D), T[bo, by, i])  # fixed case 33221
ck(f(('bool', 'bytes10', 'int168'), D), T[bo, by, i])  # fixed case 33222
ck(f(('bool', 'bytes10', 'int176'), D), T[bo, by, i])  # fixed case 33223
ck(f(('bool', 'bytes10', 'int184'), D), T[bo, by, i])  # fixed case 33224
ck(f(('bool', 'bytes10', 'int192'), D), T[bo, by, i])  # fixed case 33225
ck(f(('bool', 'bytes10', 'int200'), D), T[bo, by, i])  # fixed case 33226
ck(f(('bool', 'bytes10', 'int208'), D), T[bo, by, i])  # fixed case 33227
ck(f(('bool', 'bytes10', 'int216'), D), T[bo, by, i])  # fixed case 33228
ck(f(('bool', 'bytes10', 'int224'), D), T[bo, by, i])  # fixed case 33229
ck(f(('bool', 'bytes10', 'int232'), D), T[bo, by, i])  # fixed case 33230
ck(f(('bool', 'bytes10', 'int24'), D), T[bo, by, i])  # fixed case 33231
ck(f(('bool', 'bytes10', 'int240'), D), T[bo, by, i])  # fixed case 33232
ck(f(('bool', 'bytes10', 'int248'), D), T[bo, by, i])  # fixed case 33233
ck(f(('bool', 'bytes10', 'int256'), D), T[bo, by, i])  # fixed case 33234
ck(f(('bool', 'bytes10', 'int32'), D), T[bo, by, i])  # fixed case 33235
ck(f(('bool', 'bytes10', 'int40'), D), T[bo, by, i])  # fixed case 33236
ck(f(('bool', 'bytes10', 'int48'), D), T[bo, by, i])  # fixed case 33237
ck(f(('bool', 'bytes10', 'int56'), D), T[bo, by, i])  # fixed case 33238
ck(f(('bool', 'bytes10', 'int64'), D), T[bo, by, i])  # fixed case 33239
ck(f(('bool', 'bytes10', 'int72'), D), T[bo, by, i])  # fixed case 33240
ck(f(('bool', 'bytes10', 'int8'), D), T[bo, by, i])  # fixed case 33241
ck(f(('bool', 'bytes10', 'int80'), D), T[bo, by, i])  # fixed case 33242
ck(f(('bool', 'bytes10', 'int88'), D), T[bo, by, i])  # fixed case 33243
ck(f(('bool', 'bytes10', 'int96'), D), T[bo, by, i])  # fixed case 33244
ck(f(('bool', 'bytes10', 'string'), D), T[bo, by, s])  # fixed case 33245
ck(f(('bool', 'bytes10', 'string[]'), D), T[bo, by, A])  # fixed case 33246
ck(f(('bool', 'bytes10', 'uint104'), D), T[bo, by, i])  # fixed case 33247
ck(f(('bool', 'bytes10', 'uint112'), D), T[bo, by, i])  # fixed case 33248
ck(f(('bool', 'bytes10', 'uint120'), D), T[bo, by, i])  # fixed case 33249
ck(f(('bool', 'bytes10', 'uint128'), D), T[bo, by, i])  # fixed case 33250
ck(f(('bool', 'bytes10', 'uint136'), D), T[bo, by, i])  # fixed case 33251
ck(f(('bool', 'bytes10', 'uint144'), D), T[bo, by, i])  # fixed case 33252
ck(f(('bool', 'bytes10', 'uint152'), D), T[bo, by, i])  # fixed case 33253
ck(f(('bool', 'bytes10', 'uint16'), D), T[bo, by, i])  # fixed case 33254
ck(f(('bool', 'bytes10', 'uint160'), D), T[bo, by, i])  # fixed case 33255
ck(f(('bool', 'bytes10', 'uint168'), D), T[bo, by, i])  # fixed case 33256
ck(f(('bool', 'bytes10', 'uint176'), D), T[bo, by, i])  # fixed case 33257
ck(f(('bool', 'bytes10', 'uint184'), D), T[bo, by, i])  # fixed case 33258
ck(f(('bool', 'bytes10', 'uint192'), D), T[bo, by, i])  # fixed case 33259
ck(f(('bool', 'bytes10', 'uint200'), D), T[bo, by, i])  # fixed case 33260
ck(f(('bool', 'bytes10', 'uint208'), D), T[bo, by, i])  # fixed case 33261
ck(f(('bool', 'bytes10', 'uint216'), D), T[bo, by, i])  # fixed case 33262
ck(f(('bool', 'bytes10', 'uint224'), D), T[bo, by, i])  # fixed case 33263
ck(f(('bool', 'bytes10', 'uint232'), D), T[bo, by, i])  # fixed case 33264
ck(f(('bool', 'bytes10', 'uint24'), D), T[bo, by, i])  # fixed case 33265
ck(f(('bool', 'bytes10', 'uint240'), D), T[bo, by, i])  # fixed case 33266
ck(f(('bool', 'bytes10', 'uint248'), D), T[bo, by, i])  # fixed case 33267
ck(f(('bool', 'bytes10', 'uint256'), D), T[bo, by, i])  # fixed case 33268
ck(f(('bool', 'bytes10', 'uint32'), D), T[bo, by, i])  # fixed case 33269
ck(f(('bool', 'bytes10', 'uint40'), D), T[bo, by, i])  # fixed case 33270
ck(f(('bool', 'bytes10', 'uint48'), D), T[bo, by, i])  # fixed case 33271
ck(f(('bool', 'bytes10', 'uint56'), D), T[bo, by, i])  # fixed case 33272
ck(f(('bool', 'bytes10', 'uint64'), D), T[bo, by, i])  # fixed case 33273
ck(f(('bool', 'bytes10', 'uint72'), D), T[bo, by, i])  # fixed case 33274
ck(f(('bool', 'bytes10', 'uint8'), D), T[bo, by, i])  # fixed case 33275
ck(f(('bool', 'bytes10', 'uint80'), D), T[bo, by, i])  # fixed case 33276
ck(f(('bool', 'bytes10', 'uint88'), D), T[bo, by, i])  # fixed case 33277
ck(f(('bool', 'bytes10', 'uint96'), D), T[bo, by, i])  # fixed case 33278
ck(f(('bool', 'bytes10', '?'), D), T[bo, by, A])  # fixed case 33279
ck(f(('bool', 'bytes11', 'address'), D), T[bo, by, HA])  # fixed case 33280
ck(f(('bool', 'bytes11', 'address[]'), D), T[bo, by, A])  # fixed case 33281
ck(f(('bool', 'bytes11', 'bool'), D), T[bo, by, bo])  # fixed case 33282
ck(f(('bool', 'bytes11', 'bool[]'), D), T[bo, by, A])  # fixed case 33283
ck(f(('bool', 'bytes11', 'bytes'), D), T[bo, by, by])  # fixed case 33284
ck(f(('bool', 'bytes11', 'bytes1'), D), T[bo, by, by])  # fixed case 33285
ck(f(('bool', 'bytes11', 'bytes10'), D), T[bo, by, by])  # fixed case 33286
ck(f(('bool', 'bytes11', 'bytes11'), D), T[bo, by, by])  # fixed case 33287
ck(f(('bool', 'bytes11', 'bytes12'), D), T[bo, by, by])  # fixed case 33288
ck(f(('bool', 'bytes11', 'bytes13'), D), T[bo, by, by])  # fixed case 33289
ck(f(('bool', 'bytes11', 'bytes14'), D), T[bo, by, by])  # fixed case 33290
ck(f(('bool', 'bytes11', 'bytes15'), D), T[bo, by, by])  # fixed case 33291
ck(f(('bool', 'bytes11', 'bytes16'), D), T[bo, by, by])  # fixed case 33292
ck(f(('bool', 'bytes11', 'bytes17'), D), T[bo, by, by])  # fixed case 33293
ck(f(('bool', 'bytes11', 'bytes18'), D), T[bo, by, by])  # fixed case 33294
ck(f(('bool', 'bytes11', 'bytes19'), D), T[bo, by, by])  # fixed case 33295
ck(f(('bool', 'bytes11', 'bytes2'), D), T[bo, by, by])  # fixed case 33296
ck(f(('bool', 'bytes11', 'bytes20'), D), T[bo, by, by])  # fixed case 33297
ck(f(('bool', 'bytes11', 'bytes21'), D), T[bo, by, by])  # fixed case 33298
ck(f(('bool', 'bytes11', 'bytes22'), D), T[bo, by, by])  # fixed case 33299
ck(f(('bool', 'bytes11', 'bytes23'), D), T[bo, by, by])  # fixed case 33300
ck(f(('bool', 'bytes11', 'bytes24'), D), T[bo, by, by])  # fixed case 33301
ck(f(('bool', 'bytes11', 'bytes25'), D), T[bo, by, by])  # fixed case 33302
ck(f(('bool', 'bytes11', 'bytes26'), D), T[bo, by, by])  # fixed case 33303
ck(f(('bool', 'bytes11', 'bytes27'), D), T[bo, by, by])  # fixed case 33304
ck(f(('bool', 'bytes11', 'bytes28'), D), T[bo, by, by])  # fixed case 33305
ck(f(('bool', 'bytes11', 'bytes29'), D), T[bo, by, by])  # fixed case 33306
ck(f(('bool', 'bytes11', 'bytes3'), D), T[bo, by, by])  # fixed case 33307
ck(f(('bool', 'bytes11', 'bytes30'), D), T[bo, by, by])  # fixed case 33308
ck(f(('bool', 'bytes11', 'bytes31'), D), T[bo, by, by])  # fixed case 33309
ck(f(('bool', 'bytes11', 'bytes32'), D), T[bo, by, by])  # fixed case 33310
ck(f(('bool', 'bytes11', 'bytes4'), D), T[bo, by, by])  # fixed case 33311
ck(f(('bool', 'bytes11', 'bytes5'), D), T[bo, by, by])  # fixed case 33312
ck(f(('bool', 'bytes11', 'bytes6'), D), T[bo, by, by])  # fixed case 33313
ck(f(('bool', 'bytes11', 'bytes7'), D), T[bo, by, by])  # fixed case 33314
ck(f(('bool', 'bytes11', 'bytes8'), D), T[bo, by, by])  # fixed case 33315
ck(f(('bool', 'bytes11', 'bytes9'), D), T[bo, by, by])  # fixed case 33316
ck(f(('bool', 'bytes11', 'int104'), D), T[bo, by, i])  # fixed case 33317
ck(f(('bool', 'bytes11', 'int112'), D), T[bo, by, i])  # fixed case 33318
ck(f(('bool', 'bytes11', 'int120'), D), T[bo, by, i])  # fixed case 33319
ck(f(('bool', 'bytes11', 'int128'), D), T[bo, by, i])  # fixed case 33320
ck(f(('bool', 'bytes11', 'int136'), D), T[bo, by, i])  # fixed case 33321
ck(f(('bool', 'bytes11', 'int144'), D), T[bo, by, i])  # fixed case 33322
ck(f(('bool', 'bytes11', 'int152'), D), T[bo, by, i])  # fixed case 33323
ck(f(('bool', 'bytes11', 'int16'), D), T[bo, by, i])  # fixed case 33324
ck(f(('bool', 'bytes11', 'int160'), D), T[bo, by, i])  # fixed case 33325
ck(f(('bool', 'bytes11', 'int168'), D), T[bo, by, i])  # fixed case 33326
ck(f(('bool', 'bytes11', 'int176'), D), T[bo, by, i])  # fixed case 33327
ck(f(('bool', 'bytes11', 'int184'), D), T[bo, by, i])  # fixed case 33328
ck(f(('bool', 'bytes11', 'int192'), D), T[bo, by, i])  # fixed case 33329
ck(f(('bool', 'bytes11', 'int200'), D), T[bo, by, i])  # fixed case 33330
ck(f(('bool', 'bytes11', 'int208'), D), T[bo, by, i])  # fixed case 33331
ck(f(('bool', 'bytes11', 'int216'), D), T[bo, by, i])  # fixed case 33332
ck(f(('bool', 'bytes11', 'int224'), D), T[bo, by, i])  # fixed case 33333
ck(f(('bool', 'bytes11', 'int232'), D), T[bo, by, i])  # fixed case 33334
ck(f(('bool', 'bytes11', 'int24'), D), T[bo, by, i])  # fixed case 33335
ck(f(('bool', 'bytes11', 'int240'), D), T[bo, by, i])  # fixed case 33336
ck(f(('bool', 'bytes11', 'int248'), D), T[bo, by, i])  # fixed case 33337
ck(f(('bool', 'bytes11', 'int256'), D), T[bo, by, i])  # fixed case 33338
ck(f(('bool', 'bytes11', 'int32'), D), T[bo, by, i])  # fixed case 33339
ck(f(('bool', 'bytes11', 'int40'), D), T[bo, by, i])  # fixed case 33340
ck(f(('bool', 'bytes11', 'int48'), D), T[bo, by, i])  # fixed case 33341
ck(f(('bool', 'bytes11', 'int56'), D), T[bo, by, i])  # fixed case 33342
ck(f(('bool', 'bytes11', 'int64'), D), T[bo, by, i])  # fixed case 33343
ck(f(('bool', 'bytes11', 'int72'), D), T[bo, by, i])  # fixed case 33344
ck(f(('bool', 'bytes11', 'int8'), D), T[bo, by, i])  # fixed case 33345
ck(f(('bool', 'bytes11', 'int80'), D), T[bo, by, i])  # fixed case 33346
ck(f(('bool', 'bytes11', 'int88'), D), T[bo, by, i])  # fixed case 33347
ck(f(('bool', 'bytes11', 'int96'), D), T[bo, by, i])  # fixed case 33348
ck(f(('bool', 'bytes11', 'string'), D), T[bo, by, s])  # fixed case 33349
ck(f(('bool', 'bytes11', 'string[]'), D), T[bo, by, A])  # fixed case 33350
ck(f(('bool', 'bytes11', 'uint104'), D), T[bo, by, i])  # fixed case 33351
ck(f(('bool', 'bytes11', 'uint112'), D), T[bo, by, i])  # fixed case 33352
ck(f(('bool', 'bytes11', 'uint120'), D), T[bo, by, i])  # fixed case 33353
ck(f(('bool', 'bytes11', 'uint128'), D), T[bo, by, i])  # fixed case 33354
ck(f(('bool', 'bytes11', 'uint136'), D), T[bo, by, i])  # fixed case 33355
ck(f(('bool', 'bytes11', 'uint144'), D), T[bo, by, i])  # fixed case 33356
ck(f(('bool', 'bytes11', 'uint152'), D), T[bo, by, i])  # fixed case 33357
ck(f(('bool', 'bytes11', 'uint16'), D), T[bo, by, i])  # fixed case 33358
ck(f(('bool', 'bytes11', 'uint160'), D), T[bo, by, i])  # fixed case 33359
ck(f(('bool', 'bytes11', 'uint168'), D), T[bo, by, i])  # fixed case 33360
ck(f(('bool', 'bytes11', 'uint176'), D), T[bo, by, i])  # fixed case 33361
ck(f(('bool', 'bytes11', 'uint184'), D), T[bo, by, i])  # fixed case 33362
ck(f(('bool', 'bytes11', 'uint192'), D), T[bo, by, i])  # fixed case 33363
ck(f(('bool', 'bytes11', 'uint200'), D), T[bo, by, i])  # fixed case 33364
ck(f(('bool', 'bytes11', 'uint208'), D), T[bo, by, i])  # fixed case 33365
ck(f(('bool', 'bytes11', 'uint216'), D), T[bo, by, i])  # fixed case 33366
ck(f(('bool', 'bytes11', 'uint224'), D), T[bo, by, i])  # fixed case 33367
ck(f(('bool', 'bytes11', 'uint232'), D), T[bo, by, i])  # fixed case 33368
ck(f(('bool', 'bytes11', 'uint24'), D), T[bo, by, i])  # fixed case 33369
ck(f(('bool', 'bytes11', 'uint240'), D), T[bo, by, i])  # fixed case 33370
ck(f(('bool', 'bytes11', 'uint248'), D), T[bo, by, i])  # fixed case 33371
ck(f(('bool', 'bytes11', 'uint256'), D), T[bo, by, i])  # fixed case 33372
ck(f(('bool', 'bytes11', 'uint32'), D), T[bo, by, i])  # fixed case 33373
ck(f(('bool', 'bytes11', 'uint40'), D), T[bo, by, i])  # fixed case 33374
ck(f(('bool', 'bytes11', 'uint48'), D), T[bo, by, i])  # fixed case 33375
ck(f(('bool', 'bytes11', 'uint56'), D), T[bo, by, i])  # fixed case 33376
ck(f(('bool', 'bytes11', 'uint64'), D), T[bo, by, i])  # fixed case 33377
ck(f(('bool', 'bytes11', 'uint72'), D), T[bo, by, i])  # fixed case 33378
ck(f(('bool', 'bytes11', 'uint8'), D), T[bo, by, i])  # fixed case 33379
ck(f(('bool', 'bytes11', 'uint80'), D), T[bo, by, i])  # fixed case 33380
ck(f(('bool', 'bytes11', 'uint88'), D), T[bo, by, i])  # fixed case 33381
ck(f(('bool', 'bytes11', 'uint96'), D), T[bo, by, i])  # fixed case 33382
ck(f(('bool', 'bytes11', '?'), D), T[bo, by, A])  # fixed case 33383
ck(f(('bool', 'bytes12', 'address'), D), T[bo, by, HA])  # fixed case 33384
ck(f(('bool', 'bytes12', 'address[]'), D), T[bo, by, A])  # fixed case 33385
ck(f(('bool', 'bytes12', 'bool'), D), T[bo, by, bo])  # fixed case 33386
ck(f(('bool', 'bytes12', 'bool[]'), D), T[bo, by, A])  # fixed case 33387
ck(f(('bool', 'bytes12', 'bytes'), D), T[bo, by, by])  # fixed case 33388
ck(f(('bool', 'bytes12', 'bytes1'), D), T[bo, by, by])  # fixed case 33389
ck(f(('bool', 'bytes12', 'bytes10'), D), T[bo, by, by])  # fixed case 33390
ck(f(('bool', 'bytes12', 'bytes11'), D), T[bo, by, by])  # fixed case 33391
ck(f(('bool', 'bytes12', 'bytes12'), D), T[bo, by, by])  # fixed case 33392
ck(f(('bool', 'bytes12', 'bytes13'), D), T[bo, by, by])  # fixed case 33393
ck(f(('bool', 'bytes12', 'bytes14'), D), T[bo, by, by])  # fixed case 33394
ck(f(('bool', 'bytes12', 'bytes15'), D), T[bo, by, by])  # fixed case 33395
ck(f(('bool', 'bytes12', 'bytes16'), D), T[bo, by, by])  # fixed case 33396
ck(f(('bool', 'bytes12', 'bytes17'), D), T[bo, by, by])  # fixed case 33397
ck(f(('bool', 'bytes12', 'bytes18'), D), T[bo, by, by])  # fixed case 33398
ck(f(('bool', 'bytes12', 'bytes19'), D), T[bo, by, by])  # fixed case 33399
ck(f(('bool', 'bytes12', 'bytes2'), D), T[bo, by, by])  # fixed case 33400
ck(f(('bool', 'bytes12', 'bytes20'), D), T[bo, by, by])  # fixed case 33401
ck(f(('bool', 'bytes12', 'bytes21'), D), T[bo, by, by])  # fixed case 33402
ck(f(('bool', 'bytes12', 'bytes22'), D), T[bo, by, by])  # fixed case 33403
ck(f(('bool', 'bytes12', 'bytes23'), D), T[bo, by, by])  # fixed case 33404
ck(f(('bool', 'bytes12', 'bytes24'), D), T[bo, by, by])  # fixed case 33405
ck(f(('bool', 'bytes12', 'bytes25'), D), T[bo, by, by])  # fixed case 33406
ck(f(('bool', 'bytes12', 'bytes26'), D), T[bo, by, by])  # fixed case 33407
ck(f(('bool', 'bytes12', 'bytes27'), D), T[bo, by, by])  # fixed case 33408
ck(f(('bool', 'bytes12', 'bytes28'), D), T[bo, by, by])  # fixed case 33409
ck(f(('bool', 'bytes12', 'bytes29'), D), T[bo, by, by])  # fixed case 33410
ck(f(('bool', 'bytes12', 'bytes3'), D), T[bo, by, by])  # fixed case 33411
ck(f(('bool', 'bytes12', 'bytes30'), D), T[bo, by, by])  # fixed case 33412
ck(f(('bool', 'bytes12', 'bytes31'), D), T[bo, by, by])  # fixed case 33413
ck(f(('bool', 'bytes12', 'bytes32'), D), T[bo, by, by])  # fixed case 33414
ck(f(('bool', 'bytes12', 'bytes4'), D), T[bo, by, by])  # fixed case 33415
ck(f(('bool', 'bytes12', 'bytes5'), D), T[bo, by, by])  # fixed case 33416
ck(f(('bool', 'bytes12', 'bytes6'), D), T[bo, by, by])  # fixed case 33417
ck(f(('bool', 'bytes12', 'bytes7'), D), T[bo, by, by])  # fixed case 33418
ck(f(('bool', 'bytes12', 'bytes8'), D), T[bo, by, by])  # fixed case 33419
ck(f(('bool', 'bytes12', 'bytes9'), D), T[bo, by, by])  # fixed case 33420
ck(f(('bool', 'bytes12', 'int104'), D), T[bo, by, i])  # fixed case 33421
ck(f(('bool', 'bytes12', 'int112'), D), T[bo, by, i])  # fixed case 33422
ck(f(('bool', 'bytes12', 'int120'), D), T[bo, by, i])  # fixed case 33423
ck(f(('bool', 'bytes12', 'int128'), D), T[bo, by, i])  # fixed case 33424
ck(f(('bool', 'bytes12', 'int136'), D), T[bo, by, i])  # fixed case 33425
ck(f(('bool', 'bytes12', 'int144'), D), T[bo, by, i])  # fixed case 33426
ck(f(('bool', 'bytes12', 'int152'), D), T[bo, by, i])  # fixed case 33427
ck(f(('bool', 'bytes12', 'int16'), D), T[bo, by, i])  # fixed case 33428
ck(f(('bool', 'bytes12', 'int160'), D), T[bo, by, i])  # fixed case 33429
ck(f(('bool', 'bytes12', 'int168'), D), T[bo, by, i])  # fixed case 33430
ck(f(('bool', 'bytes12', 'int176'), D), T[bo, by, i])  # fixed case 33431
ck(f(('bool', 'bytes12', 'int184'), D), T[bo, by, i])  # fixed case 33432
ck(f(('bool', 'bytes12', 'int192'), D), T[bo, by, i])  # fixed case 33433
ck(f(('bool', 'bytes12', 'int200'), D), T[bo, by, i])  # fixed case 33434
ck(f(('bool', 'bytes12', 'int208'), D), T[bo, by, i])  # fixed case 33435
ck(f(('bool', 'bytes12', 'int216'), D), T[bo, by, i])  # fixed case 33436
ck(f(('bool', 'bytes12', 'int224'), D), T[bo, by, i])  # fixed case 33437
ck(f(('bool', 'bytes12', 'int232'), D), T[bo, by, i])  # fixed case 33438
ck(f(('bool', 'bytes12', 'int24'), D), T[bo, by, i])  # fixed case 33439
ck(f(('bool', 'bytes12', 'int240'), D), T[bo, by, i])  # fixed case 33440
ck(f(('bool', 'bytes12', 'int248'), D), T[bo, by, i])  # fixed case 33441
ck(f(('bool', 'bytes12', 'int256'), D), T[bo, by, i])  # fixed case 33442
ck(f(('bool', 'bytes12', 'int32'), D), T[bo, by, i])  # fixed case 33443
ck(f(('bool', 'bytes12', 'int40'), D), T[bo, by, i])  # fixed case 33444
ck(f(('bool', 'bytes12', 'int48'), D), T[bo, by, i])  # fixed case 33445
ck(f(('bool', 'bytes12', 'int56'), D), T[bo, by, i])  # fixed case 33446
ck(f(('bool', 'bytes12', 'int64'), D), T[bo, by, i])  # fixed case 33447
ck(f(('bool', 'bytes12', 'int72'), D), T[bo, by, i])  # fixed case 33448
ck(f(('bool', 'bytes12', 'int8'), D), T[bo, by, i])  # fixed case 33449
ck(f(('bool', 'bytes12', 'int80'), D), T[bo, by, i])  # fixed case 33450
ck(f(('bool', 'bytes12', 'int88'), D), T[bo, by, i])  # fixed case 33451
ck(f(('bool', 'bytes12', 'int96'), D), T[bo, by, i])  # fixed case 33452
ck(f(('bool', 'bytes12', 'string'), D), T[bo, by, s])  # fixed case 33453
ck(f(('bool', 'bytes12', 'string[]'), D), T[bo, by, A])  # fixed case 33454
ck(f(('bool', 'bytes12', 'uint104'), D), T[bo, by, i])  # fixed case 33455
ck(f(('bool', 'bytes12', 'uint112'), D), T[bo, by, i])  # fixed case 33456
ck(f(('bool', 'bytes12', 'uint120'), D), T[bo, by, i])  # fixed case 33457
ck(f(('bool', 'bytes12', 'uint128'), D), T[bo, by, i])  # fixed case 33458
ck(f(('bool', 'bytes12', 'uint136'), D), T[bo, by, i])  # fixed case 33459
ck(f(('bool', 'bytes12', 'uint144'), D), T[bo, by, i])  # fixed case 33460
ck(f(('bool', 'bytes12', 'uint152'), D), T[bo, by, i])  # fixed case 33461
ck(f(('bool', 'bytes12', 'uint16'), D), T[bo, by, i])  # fixed case 33462
ck(f(('bool', 'bytes12', 'uint160'), D), T[bo, by, i])  # fixed case 33463
ck(f(('bool', 'bytes12', 'uint168'), D), T[bo, by, i])  # fixed case 33464
ck(f(('bool', 'bytes12', 'uint176'), D), T[bo, by, i])  # fixed case 33465
ck(f(('bool', 'bytes12', 'uint184'), D), T[bo, by, i])  # fixed case 33466
ck(f(('bool', 'bytes12', 'uint192'), D), T[bo, by, i])  # fixed case 33467
ck(f(('bool', 'bytes12', 'uint200'), D), T[bo, by, i])  # fixed case 33468
ck(f(('bool', 'bytes12', 'uint208'), D), T[bo, by, i])  # fixed case 33469
ck(f(('bool', 'bytes12', 'uint216'), D), T[bo, by, i])  # fixed case 33470
ck(f(('bool', 'bytes12', 'uint224'), D), T[bo, by, i])  # fixed case 33471
ck(f(('bool', 'bytes12', 'uint232'), D), T[bo, by, i])  # fixed case 33472
ck(f(('bool', 'bytes12', 'uint24'), D), T[bo, by, i])  # fixed case 33473
ck(f(('bool', 'bytes12', 'uint240'), D), T[bo, by, i])  # fixed case 33474
ck(f(('bool', 'bytes12', 'uint248'), D), T[bo, by, i])  # fixed case 33475
ck(f(('bool', 'bytes12', 'uint256'), D), T[bo, by, i])  # fixed case 33476
ck(f(('bool', 'bytes12', 'uint32'), D), T[bo, by, i])  # fixed case 33477
ck(f(('bool', 'bytes12', 'uint40'), D), T[bo, by, i])  # fixed case 33478
ck(f(('bool', 'bytes12', 'uint48'), D), T[bo, by, i])  # fixed case 33479
ck(f(('bool', 'bytes12', 'uint56'), D), T[bo, by, i])  # fixed case 33480
ck(f(('bool', 'bytes12', 'uint64'), D), T[bo, by, i])  # fixed case 33481
ck(f(('bool', 'bytes12', 'uint72'), D), T[bo, by, i])  # fixed case 33482
ck(f(('bool', 'bytes12', 'uint8'), D), T[bo, by, i])  # fixed case 33483
ck(f(('bool', 'bytes12', 'uint80'), D), T[bo, by, i])  # fixed case 33484
ck(f(('bool', 'bytes12', 'uint88'), D), T[bo, by, i])  # fixed case 33485
ck(f(('bool', 'bytes12', 'uint96'), D), T[bo, by, i])  # fixed case 33486
ck(f(('bool', 'bytes12', '?'), D), T[bo, by, A])  # fixed case 33487
ck(f(('bool', 'bytes13', 'address'), D), T[bo, by, HA])  # fixed case 33488
ck(f(('bool', 'bytes13', 'address[]'), D), T[bo, by, A])  # fixed case 33489
ck(f(('bool', 'bytes13', 'bool'), D), T[bo, by, bo])  # fixed case 33490
ck(f(('bool', 'bytes13', 'bool[]'), D), T[bo, by, A])  # fixed case 33491
ck(f(('bool', 'bytes13', 'bytes'), D), T[bo, by, by])  # fixed case 33492
ck(f(('bool', 'bytes13', 'bytes1'), D), T[bo, by, by])  # fixed case 33493
ck(f(('bool', 'bytes13', 'bytes10'), D), T[bo, by, by])  # fixed case 33494
ck(f(('bool', 'bytes13', 'bytes11'), D), T[bo, by, by])  # fixed case 33495
ck(f(('bool', 'bytes13', 'bytes12'), D), T[bo, by, by])  # fixed case 33496
ck(f(('bool', 'bytes13', 'bytes13'), D), T[bo, by, by])  # fixed case 33497
ck(f(('bool', 'bytes13', 'bytes14'), D), T[bo, by, by])  # fixed case 33498
ck(f(('bool', 'bytes13', 'bytes15'), D), T[bo, by, by])  # fixed case 33499
ck(f(('bool', 'bytes13', 'bytes16'), D), T[bo, by, by])  # fixed case 33500
ck(f(('bool', 'bytes13', 'bytes17'), D), T[bo, by, by])  # fixed case 33501
ck(f(('bool', 'bytes13', 'bytes18'), D), T[bo, by, by])  # fixed case 33502
ck(f(('bool', 'bytes13', 'bytes19'), D), T[bo, by, by])  # fixed case 33503
ck(f(('bool', 'bytes13', 'bytes2'), D), T[bo, by, by])  # fixed case 33504
ck(f(('bool', 'bytes13', 'bytes20'), D), T[bo, by, by])  # fixed case 33505
ck(f(('bool', 'bytes13', 'bytes21'), D), T[bo, by, by])  # fixed case 33506
ck(f(('bool', 'bytes13', 'bytes22'), D), T[bo, by, by])  # fixed case 33507
ck(f(('bool', 'bytes13', 'bytes23'), D), T[bo, by, by])  # fixed case 33508
ck(f(('bool', 'bytes13', 'bytes24'), D), T[bo, by, by])  # fixed case 33509
ck(f(('bool', 'bytes13', 'bytes25'), D), T[bo, by, by])  # fixed case 33510
ck(f(('bool', 'bytes13', 'bytes26'), D), T[bo, by, by])  # fixed case 33511
ck(f(('bool', 'bytes13', 'bytes27'), D), T[bo, by, by])  # fixed case 33512
ck(f(('bool', 'bytes13', 'bytes28'), D), T[bo, by, by])  # fixed case 33513
ck(f(('bool', 'bytes13', 'bytes29'), D), T[bo, by, by])  # fixed case 33514
ck(f(('bool', 'bytes13', 'bytes3'), D), T[bo, by, by])  # fixed case 33515
ck(f(('bool', 'bytes13', 'bytes30'), D), T[bo, by, by])  # fixed case 33516
ck(f(('bool', 'bytes13', 'bytes31'), D), T[bo, by, by])  # fixed case 33517
ck(f(('bool', 'bytes13', 'bytes32'), D), T[bo, by, by])  # fixed case 33518
ck(f(('bool', 'bytes13', 'bytes4'), D), T[bo, by, by])  # fixed case 33519
ck(f(('bool', 'bytes13', 'bytes5'), D), T[bo, by, by])  # fixed case 33520
ck(f(('bool', 'bytes13', 'bytes6'), D), T[bo, by, by])  # fixed case 33521
ck(f(('bool', 'bytes13', 'bytes7'), D), T[bo, by, by])  # fixed case 33522
ck(f(('bool', 'bytes13', 'bytes8'), D), T[bo, by, by])  # fixed case 33523
ck(f(('bool', 'bytes13', 'bytes9'), D), T[bo, by, by])  # fixed case 33524
ck(f(('bool', 'bytes13', 'int104'), D), T[bo, by, i])  # fixed case 33525
ck(f(('bool', 'bytes13', 'int112'), D), T[bo, by, i])  # fixed case 33526
ck(f(('bool', 'bytes13', 'int120'), D), T[bo, by, i])  # fixed case 33527
ck(f(('bool', 'bytes13', 'int128'), D), T[bo, by, i])  # fixed case 33528
ck(f(('bool', 'bytes13', 'int136'), D), T[bo, by, i])  # fixed case 33529
ck(f(('bool', 'bytes13', 'int144'), D), T[bo, by, i])  # fixed case 33530
ck(f(('bool', 'bytes13', 'int152'), D), T[bo, by, i])  # fixed case 33531
ck(f(('bool', 'bytes13', 'int16'), D), T[bo, by, i])  # fixed case 33532
ck(f(('bool', 'bytes13', 'int160'), D), T[bo, by, i])  # fixed case 33533
ck(f(('bool', 'bytes13', 'int168'), D), T[bo, by, i])  # fixed case 33534
ck(f(('bool', 'bytes13', 'int176'), D), T[bo, by, i])  # fixed case 33535
ck(f(('bool', 'bytes13', 'int184'), D), T[bo, by, i])  # fixed case 33536
ck(f(('bool', 'bytes13', 'int192'), D), T[bo, by, i])  # fixed case 33537
ck(f(('bool', 'bytes13', 'int200'), D), T[bo, by, i])  # fixed case 33538
ck(f(('bool', 'bytes13', 'int208'), D), T[bo, by, i])  # fixed case 33539
ck(f(('bool', 'bytes13', 'int216'), D), T[bo, by, i])  # fixed case 33540
ck(f(('bool', 'bytes13', 'int224'), D), T[bo, by, i])  # fixed case 33541
ck(f(('bool', 'bytes13', 'int232'), D), T[bo, by, i])  # fixed case 33542
ck(f(('bool', 'bytes13', 'int24'), D), T[bo, by, i])  # fixed case 33543
ck(f(('bool', 'bytes13', 'int240'), D), T[bo, by, i])  # fixed case 33544
ck(f(('bool', 'bytes13', 'int248'), D), T[bo, by, i])  # fixed case 33545
ck(f(('bool', 'bytes13', 'int256'), D), T[bo, by, i])  # fixed case 33546
ck(f(('bool', 'bytes13', 'int32'), D), T[bo, by, i])  # fixed case 33547
ck(f(('bool', 'bytes13', 'int40'), D), T[bo, by, i])  # fixed case 33548
ck(f(('bool', 'bytes13', 'int48'), D), T[bo, by, i])  # fixed case 33549
ck(f(('bool', 'bytes13', 'int56'), D), T[bo, by, i])  # fixed case 33550
ck(f(('bool', 'bytes13', 'int64'), D), T[bo, by, i])  # fixed case 33551
ck(f(('bool', 'bytes13', 'int72'), D), T[bo, by, i])  # fixed case 33552
ck(f(('bool', 'bytes13', 'int8'), D), T[bo, by, i])  # fixed case 33553
ck(f(('bool', 'bytes13', 'int80'), D), T[bo, by, i])  # fixed case 33554
ck(f(('bool', 'bytes13', 'int88'), D), T[bo, by, i])  # fixed case 33555
ck(f(('bool', 'bytes13', 'int96'), D), T[bo, by, i])  # fixed case 33556
ck(f(('bool', 'bytes13', 'string'), D), T[bo, by, s])  # fixed case 33557
ck(f(('bool', 'bytes13', 'string[]'), D), T[bo, by, A])  # fixed case 33558
ck(f(('bool', 'bytes13', 'uint104'), D), T[bo, by, i])  # fixed case 33559
ck(f(('bool', 'bytes13', 'uint112'), D), T[bo, by, i])  # fixed case 33560
ck(f(('bool', 'bytes13', 'uint120'), D), T[bo, by, i])  # fixed case 33561
ck(f(('bool', 'bytes13', 'uint128'), D), T[bo, by, i])  # fixed case 33562
ck(f(('bool', 'bytes13', 'uint136'), D), T[bo, by, i])  # fixed case 33563
ck(f(('bool', 'bytes13', 'uint144'), D), T[bo, by, i])  # fixed case 33564
ck(f(('bool', 'bytes13', 'uint152'), D), T[bo, by, i])  # fixed case 33565
ck(f(('bool', 'bytes13', 'uint16'), D), T[bo, by, i])  # fixed case 33566
ck(f(('bool', 'bytes13', 'uint160'), D), T[bo, by, i])  # fixed case 33567
ck(f(('bool', 'bytes13', 'uint168'), D), T[bo, by, i])  # fixed case 33568
ck(f(('bool', 'bytes13', 'uint176'), D), T[bo, by, i])  # fixed case 33569
ck(f(('bool', 'bytes13', 'uint184'), D), T[bo, by, i])  # fixed case 33570
ck(f(('bool', 'bytes13', 'uint192'), D), T[bo, by, i])  # fixed case 33571
ck(f(('bool', 'bytes13', 'uint200'), D), T[bo, by, i])  # fixed case 33572
ck(f(('bool', 'bytes13', 'uint208'), D), T[bo, by, i])  # fixed case 33573
ck(f(('bool', 'bytes13', 'uint216'), D), T[bo, by, i])  # fixed case 33574
ck(f(('bool', 'bytes13', 'uint224'), D), T[bo, by, i])  # fixed case 33575
ck(f(('bool', 'bytes13', 'uint232'), D), T[bo, by, i])  # fixed case 33576
ck(f(('bool', 'bytes13', 'uint24'), D), T[bo, by, i])  # fixed case 33577
ck(f(('bool', 'bytes13', 'uint240'), D), T[bo, by, i])  # fixed case 33578
ck(f(('bool', 'bytes13', 'uint248'), D), T[bo, by, i])  # fixed case 33579
ck(f(('bool', 'bytes13', 'uint256'), D), T[bo, by, i])  # fixed case 33580
ck(f(('bool', 'bytes13', 'uint32'), D), T[bo, by, i])  # fixed case 33581
ck(f(('bool', 'bytes13', 'uint40'), D), T[bo, by, i])  # fixed case 33582
ck(f(('bool', 'bytes13', 'uint48'), D), T[bo, by, i])  # fixed case 33583
ck(f(('bool', 'bytes13', 'uint56'), D), T[bo, by, i])  # fixed case 33584
ck(f(('bool', 'bytes13', 'uint64'), D), T[bo, by, i])  # fixed case 33585
ck(f(('bool', 'bytes13', 'uint72'), D), T[bo, by, i])  # fixed case 33586
ck(f(('bool', 'bytes13', 'uint8'), D), T[bo, by, i])  # fixed case 33587
ck(f(('bool', 'bytes13', 'uint80'), D), T[bo, by, i])  # fixed case 33588
ck(f(('bool', 'bytes13', 'uint88'), D), T[bo, by, i])  # fixed case 33589
ck(f(('bool', 'bytes13', 'uint96'), D), T[bo, by, i])  # fixed case 33590
ck(f(('bool', 'bytes13', '?'), D), T[bo, by, A])  # fixed case 33591
ck(f(('bool', 'bytes14', 'address'), D), T[bo, by, HA])  # fixed case 33592
ck(f(('bool', 'bytes14', 'address[]'), D), T[bo, by, A])  # fixed case 33593
ck(f(('bool', 'bytes14', 'bool'), D), T[bo, by, bo])  # fixed case 33594
ck(f(('bool', 'bytes14', 'bool[]'), D), T[bo, by, A])  # fixed case 33595
ck(f(('bool', 'bytes14', 'bytes'), D), T[bo, by, by])  # fixed case 33596
ck(f(('bool', 'bytes14', 'bytes1'), D), T[bo, by, by])  # fixed case 33597
ck(f(('bool', 'bytes14', 'bytes10'), D), T[bo, by, by])  # fixed case 33598
ck(f(('bool', 'bytes14', 'bytes11'), D), T[bo, by, by])  # fixed case 33599
ck(f(('bool', 'bytes14', 'bytes12'), D), T[bo, by, by])  # fixed case 33600
ck(f(('bool', 'bytes14', 'bytes13'), D), T[bo, by, by])  # fixed case 33601
ck(f(('bool', 'bytes14', 'bytes14'), D), T[bo, by, by])  # fixed case 33602
ck(f(('bool', 'bytes14', 'bytes15'), D), T[bo, by, by])  # fixed case 33603
ck(f(('bool', 'bytes14', 'bytes16'), D), T[bo, by, by])  # fixed case 33604
ck(f(('bool', 'bytes14', 'bytes17'), D), T[bo, by, by])  # fixed case 33605
ck(f(('bool', 'bytes14', 'bytes18'), D), T[bo, by, by])  # fixed case 33606
ck(f(('bool', 'bytes14', 'bytes19'), D), T[bo, by, by])  # fixed case 33607
ck(f(('bool', 'bytes14', 'bytes2'), D), T[bo, by, by])  # fixed case 33608
ck(f(('bool', 'bytes14', 'bytes20'), D), T[bo, by, by])  # fixed case 33609
ck(f(('bool', 'bytes14', 'bytes21'), D), T[bo, by, by])  # fixed case 33610
ck(f(('bool', 'bytes14', 'bytes22'), D), T[bo, by, by])  # fixed case 33611
ck(f(('bool', 'bytes14', 'bytes23'), D), T[bo, by, by])  # fixed case 33612
ck(f(('bool', 'bytes14', 'bytes24'), D), T[bo, by, by])  # fixed case 33613
ck(f(('bool', 'bytes14', 'bytes25'), D), T[bo, by, by])  # fixed case 33614
ck(f(('bool', 'bytes14', 'bytes26'), D), T[bo, by, by])  # fixed case 33615
ck(f(('bool', 'bytes14', 'bytes27'), D), T[bo, by, by])  # fixed case 33616
ck(f(('bool', 'bytes14', 'bytes28'), D), T[bo, by, by])  # fixed case 33617
ck(f(('bool', 'bytes14', 'bytes29'), D), T[bo, by, by])  # fixed case 33618
ck(f(('bool', 'bytes14', 'bytes3'), D), T[bo, by, by])  # fixed case 33619
ck(f(('bool', 'bytes14', 'bytes30'), D), T[bo, by, by])  # fixed case 33620
ck(f(('bool', 'bytes14', 'bytes31'), D), T[bo, by, by])  # fixed case 33621
ck(f(('bool', 'bytes14', 'bytes32'), D), T[bo, by, by])  # fixed case 33622
ck(f(('bool', 'bytes14', 'bytes4'), D), T[bo, by, by])  # fixed case 33623
ck(f(('bool', 'bytes14', 'bytes5'), D), T[bo, by, by])  # fixed case 33624
ck(f(('bool', 'bytes14', 'bytes6'), D), T[bo, by, by])  # fixed case 33625
ck(f(('bool', 'bytes14', 'bytes7'), D), T[bo, by, by])  # fixed case 33626
ck(f(('bool', 'bytes14', 'bytes8'), D), T[bo, by, by])  # fixed case 33627
ck(f(('bool', 'bytes14', 'bytes9'), D), T[bo, by, by])  # fixed case 33628
ck(f(('bool', 'bytes14', 'int104'), D), T[bo, by, i])  # fixed case 33629
ck(f(('bool', 'bytes14', 'int112'), D), T[bo, by, i])  # fixed case 33630
ck(f(('bool', 'bytes14', 'int120'), D), T[bo, by, i])  # fixed case 33631
ck(f(('bool', 'bytes14', 'int128'), D), T[bo, by, i])  # fixed case 33632
ck(f(('bool', 'bytes14', 'int136'), D), T[bo, by, i])  # fixed case 33633
ck(f(('bool', 'bytes14', 'int144'), D), T[bo, by, i])  # fixed case 33634
ck(f(('bool', 'bytes14', 'int152'), D), T[bo, by, i])  # fixed case 33635
ck(f(('bool', 'bytes14', 'int16'), D), T[bo, by, i])  # fixed case 33636
ck(f(('bool', 'bytes14', 'int160'), D), T[bo, by, i])  # fixed case 33637
ck(f(('bool', 'bytes14', 'int168'), D), T[bo, by, i])  # fixed case 33638
ck(f(('bool', 'bytes14', 'int176'), D), T[bo, by, i])  # fixed case 33639
ck(f(('bool', 'bytes14', 'int184'), D), T[bo, by, i])  # fixed case 33640
ck(f(('bool', 'bytes14', 'int192'), D), T[bo, by, i])  # fixed case 33641
ck(f(('bool', 'bytes14', 'int200'), D), T[bo, by, i])  # fixed case 33642
ck(f(('bool', 'bytes14', 'int208'), D), T[bo, by, i])  # fixed case 33643
ck(f(('bool', 'bytes14', 'int216'), D), T[bo, by, i])  # fixed case 33644
ck(f(('bool', 'bytes14', 'int224'), D), T[bo, by, i])  # fixed case 33645
ck(f(('bool', 'bytes14', 'int232'), D), T[bo, by, i])  # fixed case 33646
ck(f(('bool', 'bytes14', 'int24'), D), T[bo, by, i])  # fixed case 33647
ck(f(('bool', 'bytes14', 'int240'), D), T[bo, by, i])  # fixed case 33648
ck(f(('bool', 'bytes14', 'int248'), D), T[bo, by, i])  # fixed case 33649
ck(f(('bool', 'bytes14', 'int256'), D), T[bo, by, i])  # fixed case 33650
ck(f(('bool', 'bytes14', 'int32'), D), T[bo, by, i])  # fixed case 33651
ck(f(('bool', 'bytes14', 'int40'), D), T[bo, by, i])  # fixed case 33652
ck(f(('bool', 'bytes14', 'int48'), D), T[bo, by, i])  # fixed case 33653
ck(f(('bool', 'bytes14', 'int56'), D), T[bo, by, i])  # fixed case 33654
ck(f(('bool', 'bytes14', 'int64'), D), T[bo, by, i])  # fixed case 33655
ck(f(('bool', 'bytes14', 'int72'), D), T[bo, by, i])  # fixed case 33656
ck(f(('bool', 'bytes14', 'int8'), D), T[bo, by, i])  # fixed case 33657
ck(f(('bool', 'bytes14', 'int80'), D), T[bo, by, i])  # fixed case 33658
ck(f(('bool', 'bytes14', 'int88'), D), T[bo, by, i])  # fixed case 33659
ck(f(('bool', 'bytes14', 'int96'), D), T[bo, by, i])  # fixed case 33660
ck(f(('bool', 'bytes14', 'string'), D), T[bo, by, s])  # fixed case 33661
ck(f(('bool', 'bytes14', 'string[]'), D), T[bo, by, A])  # fixed case 33662
ck(f(('bool', 'bytes14', 'uint104'), D), T[bo, by, i])  # fixed case 33663
ck(f(('bool', 'bytes14', 'uint112'), D), T[bo, by, i])  # fixed case 33664
ck(f(('bool', 'bytes14', 'uint120'), D), T[bo, by, i])  # fixed case 33665
ck(f(('bool', 'bytes14', 'uint128'), D), T[bo, by, i])  # fixed case 33666
ck(f(('bool', 'bytes14', 'uint136'), D), T[bo, by, i])  # fixed case 33667
ck(f(('bool', 'bytes14', 'uint144'), D), T[bo, by, i])  # fixed case 33668
ck(f(('bool', 'bytes14', 'uint152'), D), T[bo, by, i])  # fixed case 33669
ck(f(('bool', 'bytes14', 'uint16'), D), T[bo, by, i])  # fixed case 33670
ck(f(('bool', 'bytes14', 'uint160'), D), T[bo, by, i])  # fixed case 33671
ck(f(('bool', 'bytes14', 'uint168'), D), T[bo, by, i])  # fixed case 33672
ck(f(('bool', 'bytes14', 'uint176'), D), T[bo, by, i])  # fixed case 33673
ck(f(('bool', 'bytes14', 'uint184'), D), T[bo, by, i])  # fixed case 33674
ck(f(('bool', 'bytes14', 'uint192'), D), T[bo, by, i])  # fixed case 33675
ck(f(('bool', 'bytes14', 'uint200'), D), T[bo, by, i])  # fixed case 33676
ck(f(('bool', 'bytes14', 'uint208'), D), T[bo, by, i])  # fixed case 33677
ck(f(('bool', 'bytes14', 'uint216'), D), T[bo, by, i])  # fixed case 33678
ck(f(('bool', 'bytes14', 'uint224'), D), T[bo, by, i])  # fixed case 33679
ck(f(('bool', 'bytes14', 'uint232'), D), T[bo, by, i])  # fixed case 33680
ck(f(('bool', 'bytes14', 'uint24'), D), T[bo, by, i])  # fixed case 33681
ck(f(('bool', 'bytes14', 'uint240'), D), T[bo, by, i])  # fixed case 33682
ck(f(('bool', 'bytes14', 'uint248'), D), T[bo, by, i])  # fixed case 33683
ck(f(('bool', 'bytes14', 'uint256'), D), T[bo, by, i])  # fixed case 33684
ck(f(('bool', 'bytes14', 'uint32'), D), T[bo, by, i])  # fixed case 33685
ck(f(('bool', 'bytes14', 'uint40'), D), T[bo, by, i])  # fixed case 33686
ck(f(('bool', 'bytes14', 'uint48'), D), T[bo, by, i])  # fixed case 33687
ck(f(('bool', 'bytes14', 'uint56'), D), T[bo, by, i])  # fixed case 33688
ck(f(('bool', 'bytes14', 'uint64'), D), T[bo, by, i])  # fixed case 33689
ck(f(('bool', 'bytes14', 'uint72'), D), T[bo, by, i])  # fixed case 33690
ck(f(('bool', 'bytes14', 'uint8'), D), T[bo, by, i])  # fixed case 33691
ck(f(('bool', 'bytes14', 'uint80'), D), T[bo, by, i])  # fixed case 33692
ck(f(('bool', 'bytes14', 'uint88'), D), T[bo, by, i])  # fixed case 33693
ck(f(('bool', 'bytes14', 'uint96'), D), T[bo, by, i])  # fixed case 33694
ck(f(('bool', 'bytes14', '?'), D), T[bo, by, A])  # fixed case 33695
ck(f(('bool', 'bytes15', 'address'), D), T[bo, by, HA])  # fixed case 33696
ck(f(('bool', 'bytes15', 'address[]'), D), T[bo, by, A])  # fixed case 33697
ck(f(('bool', 'bytes15', 'bool'), D), T[bo, by, bo])  # fixed case 33698
ck(f(('bool', 'bytes15', 'bool[]'), D), T[bo, by, A])  # fixed case 33699
ck(f(('bool', 'bytes15', 'bytes'), D), T[bo, by, by])  # fixed case 33700
ck(f(('bool', 'bytes15', 'bytes1'), D), T[bo, by, by])  # fixed case 33701
ck(f(('bool', 'bytes15', 'bytes10'), D), T[bo, by, by])  # fixed case 33702
ck(f(('bool', 'bytes15', 'bytes11'), D), T[bo, by, by])  # fixed case 33703
ck(f(('bool', 'bytes15', 'bytes12'), D), T[bo, by, by])  # fixed case 33704
ck(f(('bool', 'bytes15', 'bytes13'), D), T[bo, by, by])  # fixed case 33705
ck(f(('bool', 'bytes15', 'bytes14'), D), T[bo, by, by])  # fixed case 33706
ck(f(('bool', 'bytes15', 'bytes15'), D), T[bo, by, by])  # fixed case 33707
ck(f(('bool', 'bytes15', 'bytes16'), D), T[bo, by, by])  # fixed case 33708
ck(f(('bool', 'bytes15', 'bytes17'), D), T[bo, by, by])  # fixed case 33709
ck(f(('bool', 'bytes15', 'bytes18'), D), T[bo, by, by])  # fixed case 33710
ck(f(('bool', 'bytes15', 'bytes19'), D), T[bo, by, by])  # fixed case 33711
ck(f(('bool', 'bytes15', 'bytes2'), D), T[bo, by, by])  # fixed case 33712
ck(f(('bool', 'bytes15', 'bytes20'), D), T[bo, by, by])  # fixed case 33713
ck(f(('bool', 'bytes15', 'bytes21'), D), T[bo, by, by])  # fixed case 33714
ck(f(('bool', 'bytes15', 'bytes22'), D), T[bo, by, by])  # fixed case 33715
ck(f(('bool', 'bytes15', 'bytes23'), D), T[bo, by, by])  # fixed case 33716
ck(f(('bool', 'bytes15', 'bytes24'), D), T[bo, by, by])  # fixed case 33717
ck(f(('bool', 'bytes15', 'bytes25'), D), T[bo, by, by])  # fixed case 33718
ck(f(('bool', 'bytes15', 'bytes26'), D), T[bo, by, by])  # fixed case 33719
ck(f(('bool', 'bytes15', 'bytes27'), D), T[bo, by, by])  # fixed case 33720
ck(f(('bool', 'bytes15', 'bytes28'), D), T[bo, by, by])  # fixed case 33721
ck(f(('bool', 'bytes15', 'bytes29'), D), T[bo, by, by])  # fixed case 33722
ck(f(('bool', 'bytes15', 'bytes3'), D), T[bo, by, by])  # fixed case 33723
ck(f(('bool', 'bytes15', 'bytes30'), D), T[bo, by, by])  # fixed case 33724
ck(f(('bool', 'bytes15', 'bytes31'), D), T[bo, by, by])  # fixed case 33725
ck(f(('bool', 'bytes15', 'bytes32'), D), T[bo, by, by])  # fixed case 33726
ck(f(('bool', 'bytes15', 'bytes4'), D), T[bo, by, by])  # fixed case 33727
ck(f(('bool', 'bytes15', 'bytes5'), D), T[bo, by, by])  # fixed case 33728
ck(f(('bool', 'bytes15', 'bytes6'), D), T[bo, by, by])  # fixed case 33729
ck(f(('bool', 'bytes15', 'bytes7'), D), T[bo, by, by])  # fixed case 33730
ck(f(('bool', 'bytes15', 'bytes8'), D), T[bo, by, by])  # fixed case 33731
ck(f(('bool', 'bytes15', 'bytes9'), D), T[bo, by, by])  # fixed case 33732
ck(f(('bool', 'bytes15', 'int104'), D), T[bo, by, i])  # fixed case 33733
ck(f(('bool', 'bytes15', 'int112'), D), T[bo, by, i])  # fixed case 33734
ck(f(('bool', 'bytes15', 'int120'), D), T[bo, by, i])  # fixed case 33735
ck(f(('bool', 'bytes15', 'int128'), D), T[bo, by, i])  # fixed case 33736
ck(f(('bool', 'bytes15', 'int136'), D), T[bo, by, i])  # fixed case 33737
ck(f(('bool', 'bytes15', 'int144'), D), T[bo, by, i])  # fixed case 33738
ck(f(('bool', 'bytes15', 'int152'), D), T[bo, by, i])  # fixed case 33739
ck(f(('bool', 'bytes15', 'int16'), D), T[bo, by, i])  # fixed case 33740
ck(f(('bool', 'bytes15', 'int160'), D), T[bo, by, i])  # fixed case 33741
ck(f(('bool', 'bytes15', 'int168'), D), T[bo, by, i])  # fixed case 33742
ck(f(('bool', 'bytes15', 'int176'), D), T[bo, by, i])  # fixed case 33743
ck(f(('bool', 'bytes15', 'int184'), D), T[bo, by, i])  # fixed case 33744
ck(f(('bool', 'bytes15', 'int192'), D), T[bo, by, i])  # fixed case 33745
ck(f(('bool', 'bytes15', 'int200'), D), T[bo, by, i])  # fixed case 33746
ck(f(('bool', 'bytes15', 'int208'), D), T[bo, by, i])  # fixed case 33747
ck(f(('bool', 'bytes15', 'int216'), D), T[bo, by, i])  # fixed case 33748
ck(f(('bool', 'bytes15', 'int224'), D), T[bo, by, i])  # fixed case 33749
ck(f(('bool', 'bytes15', 'int232'), D), T[bo, by, i])  # fixed case 33750
ck(f(('bool', 'bytes15', 'int24'), D), T[bo, by, i])  # fixed case 33751
ck(f(('bool', 'bytes15', 'int240'), D), T[bo, by, i])  # fixed case 33752
ck(f(('bool', 'bytes15', 'int248'), D), T[bo, by, i])  # fixed case 33753
ck(f(('bool', 'bytes15', 'int256'), D), T[bo, by, i])  # fixed case 33754
ck(f(('bool', 'bytes15', 'int32'), D), T[bo, by, i])  # fixed case 33755
ck(f(('bool', 'bytes15', 'int40'), D), T[bo, by, i])  # fixed case 33756
ck(f(('bool', 'bytes15', 'int48'), D), T[bo, by, i])  # fixed case 33757
ck(f(('bool', 'bytes15', 'int56'), D), T[bo, by, i])  # fixed case 33758
ck(f(('bool', 'bytes15', 'int64'), D), T[bo, by, i])  # fixed case 33759
ck(f(('bool', 'bytes15', 'int72'), D), T[bo, by, i])  # fixed case 33760
ck(f(('bool', 'bytes15', 'int8'), D), T[bo, by, i])  # fixed case 33761
ck(f(('bool', 'bytes15', 'int80'), D), T[bo, by, i])  # fixed case 33762
ck(f(('bool', 'bytes15', 'int88'), D), T[bo, by, i])  # fixed case 33763
ck(f(('bool', 'bytes15', 'int96'), D), T[bo, by, i])  # fixed case 33764
ck(f(('bool', 'bytes15', 'string'), D), T[bo, by, s])  # fixed case 33765
ck(f(('bool', 'bytes15', 'string[]'), D), T[bo, by, A])  # fixed case 33766
ck(f(('bool', 'bytes15', 'uint104'), D), T[bo, by, i])  # fixed case 33767
ck(f(('bool', 'bytes15', 'uint112'), D), T[bo, by, i])  # fixed case 33768
ck(f(('bool', 'bytes15', 'uint120'), D), T[bo, by, i])  # fixed case 33769
ck(f(('bool', 'bytes15', 'uint128'), D), T[bo, by, i])  # fixed case 33770
ck(f(('bool', 'bytes15', 'uint136'), D), T[bo, by, i])  # fixed case 33771
ck(f(('bool', 'bytes15', 'uint144'), D), T[bo, by, i])  # fixed case 33772
ck(f(('bool', 'bytes15', 'uint152'), D), T[bo, by, i])  # fixed case 33773
ck(f(('bool', 'bytes15', 'uint16'), D), T[bo, by, i])  # fixed case 33774
ck(f(('bool', 'bytes15', 'uint160'), D), T[bo, by, i])  # fixed case 33775
ck(f(('bool', 'bytes15', 'uint168'), D), T[bo, by, i])  # fixed case 33776
ck(f(('bool', 'bytes15', 'uint176'), D), T[bo, by, i])  # fixed case 33777
ck(f(('bool', 'bytes15', 'uint184'), D), T[bo, by, i])  # fixed case 33778
ck(f(('bool', 'bytes15', 'uint192'), D), T[bo, by, i])  # fixed case 33779
ck(f(('bool', 'bytes15', 'uint200'), D), T[bo, by, i])  # fixed case 33780
ck(f(('bool', 'bytes15', 'uint208'), D), T[bo, by, i])  # fixed case 33781
ck(f(('bool', 'bytes15', 'uint216'), D), T[bo, by, i])  # fixed case 33782
ck(f(('bool', 'bytes15', 'uint224'), D), T[bo, by, i])  # fixed case 33783
ck(f(('bool', 'bytes15', 'uint232'), D), T[bo, by, i])  # fixed case 33784
ck(f(('bool', 'bytes15', 'uint24'), D), T[bo, by, i])  # fixed case 33785
ck(f(('bool', 'bytes15', 'uint240'), D), T[bo, by, i])  # fixed case 33786
ck(f(('bool', 'bytes15', 'uint248'), D), T[bo, by, i])  # fixed case 33787
ck(f(('bool', 'bytes15', 'uint256'), D), T[bo, by, i])  # fixed case 33788
ck(f(('bool', 'bytes15', 'uint32'), D), T[bo, by, i])  # fixed case 33789
ck(f(('bool', 'bytes15', 'uint40'), D), T[bo, by, i])  # fixed case 33790
ck(f(('bool', 'bytes15', 'uint48'), D), T[bo, by, i])  # fixed case 33791
ck(f(('bool', 'bytes15', 'uint56'), D), T[bo, by, i])  # fixed case 33792
ck(f(('bool', 'bytes15', 'uint64'), D), T[bo, by, i])  # fixed case 33793
ck(f(('bool', 'bytes15', 'uint72'), D), T[bo, by, i])  # fixed case 33794
ck(f(('bool', 'bytes15', 'uint8'), D), T[bo, by, i])  # fixed case 33795
ck(f(('bool', 'bytes15', 'uint80'), D), T[bo, by, i])  # fixed case 33796
ck(f(('bool', 'bytes15', 'uint88'), D), T[bo, by, i])  # fixed case 33797
ck(f(('bool', 'bytes15', 'uint96'), D), T[bo, by, i])  # fixed case 33798
ck(f(('bool', 'bytes15', '?'), D), T[bo, by, A])  # fixed case 33799
ck(f(('bool', 'bytes16', 'address'), D), T[bo, by, HA])  # fixed case 33800
ck(f(('bool', 'bytes16', 'address[]'), D), T[bo, by, A])  # fixed case 33801
ck(f(('bool', 'bytes16', 'bool'), D), T[bo, by, bo])  # fixed case 33802
ck(f(('bool', 'bytes16', 'bool[]'), D), T[bo, by, A])  # fixed case 33803
ck(f(('bool', 'bytes16', 'bytes'), D), T[bo, by, by])  # fixed case 33804
ck(f(('bool', 'bytes16', 'bytes1'), D), T[bo, by, by])  # fixed case 33805
ck(f(('bool', 'bytes16', 'bytes10'), D), T[bo, by, by])  # fixed case 33806
ck(f(('bool', 'bytes16', 'bytes11'), D), T[bo, by, by])  # fixed case 33807
ck(f(('bool', 'bytes16', 'bytes12'), D), T[bo, by, by])  # fixed case 33808
ck(f(('bool', 'bytes16', 'bytes13'), D), T[bo, by, by])  # fixed case 33809
ck(f(('bool', 'bytes16', 'bytes14'), D), T[bo, by, by])  # fixed case 33810
ck(f(('bool', 'bytes16', 'bytes15'), D), T[bo, by, by])  # fixed case 33811
ck(f(('bool', 'bytes16', 'bytes16'), D), T[bo, by, by])  # fixed case 33812
ck(f(('bool', 'bytes16', 'bytes17'), D), T[bo, by, by])  # fixed case 33813
ck(f(('bool', 'bytes16', 'bytes18'), D), T[bo, by, by])  # fixed case 33814
ck(f(('bool', 'bytes16', 'bytes19'), D), T[bo, by, by])  # fixed case 33815
ck(f(('bool', 'bytes16', 'bytes2'), D), T[bo, by, by])  # fixed case 33816
ck(f(('bool', 'bytes16', 'bytes20'), D), T[bo, by, by])  # fixed case 33817
ck(f(('bool', 'bytes16', 'bytes21'), D), T[bo, by, by])  # fixed case 33818
ck(f(('bool', 'bytes16', 'bytes22'), D), T[bo, by, by])  # fixed case 33819
ck(f(('bool', 'bytes16', 'bytes23'), D), T[bo, by, by])  # fixed case 33820
ck(f(('bool', 'bytes16', 'bytes24'), D), T[bo, by, by])  # fixed case 33821
ck(f(('bool', 'bytes16', 'bytes25'), D), T[bo, by, by])  # fixed case 33822
ck(f(('bool', 'bytes16', 'bytes26'), D), T[bo, by, by])  # fixed case 33823
ck(f(('bool', 'bytes16', 'bytes27'), D), T[bo, by, by])  # fixed case 33824
ck(f(('bool', 'bytes16', 'bytes28'), D), T[bo, by, by])  # fixed case 33825
ck(f(('bool', 'bytes16', 'bytes29'), D), T[bo, by, by])  # fixed case 33826
ck(f(('bool', 'bytes16', 'bytes3'), D), T[bo, by, by])  # fixed case 33827
ck(f(('bool', 'bytes16', 'bytes30'), D), T[bo, by, by])  # fixed case 33828
ck(f(('bool', 'bytes16', 'bytes31'), D), T[bo, by, by])  # fixed case 33829
ck(f(('bool', 'bytes16', 'bytes32'), D), T[bo, by, by])  # fixed case 33830
ck(f(('bool', 'bytes16', 'bytes4'), D), T[bo, by, by])  # fixed case 33831
ck(f(('bool', 'bytes16', 'bytes5'), D), T[bo, by, by])  # fixed case 33832
ck(f(('bool', 'bytes16', 'bytes6'), D), T[bo, by, by])  # fixed case 33833
ck(f(('bool', 'bytes16', 'bytes7'), D), T[bo, by, by])  # fixed case 33834
ck(f(('bool', 'bytes16', 'bytes8'), D), T[bo, by, by])  # fixed case 33835
ck(f(('bool', 'bytes16', 'bytes9'), D), T[bo, by, by])  # fixed case 33836
ck(f(('bool', 'bytes16', 'int104'), D), T[bo, by, i])  # fixed case 33837
ck(f(('bool', 'bytes16', 'int112'), D), T[bo, by, i])  # fixed case 33838
ck(f(('bool', 'bytes16', 'int120'), D), T[bo, by, i])  # fixed case 33839
ck(f(('bool', 'bytes16', 'int128'), D), T[bo, by, i])  # fixed case 33840
ck(f(('bool', 'bytes16', 'int136'), D), T[bo, by, i])  # fixed case 33841
ck(f(('bool', 'bytes16', 'int144'), D), T[bo, by, i])  # fixed case 33842
ck(f(('bool', 'bytes16', 'int152'), D), T[bo, by, i])  # fixed case 33843
ck(f(('bool', 'bytes16', 'int16'), D), T[bo, by, i])  # fixed case 33844
ck(f(('bool', 'bytes16', 'int160'), D), T[bo, by, i])  # fixed case 33845
ck(f(('bool', 'bytes16', 'int168'), D), T[bo, by, i])  # fixed case 33846
ck(f(('bool', 'bytes16', 'int176'), D), T[bo, by, i])  # fixed case 33847
ck(f(('bool', 'bytes16', 'int184'), D), T[bo, by, i])  # fixed case 33848
ck(f(('bool', 'bytes16', 'int192'), D), T[bo, by, i])  # fixed case 33849
ck(f(('bool', 'bytes16', 'int200'), D), T[bo, by, i])  # fixed case 33850
ck(f(('bool', 'bytes16', 'int208'), D), T[bo, by, i])  # fixed case 33851
ck(f(('bool', 'bytes16', 'int216'), D), T[bo, by, i])  # fixed case 33852
ck(f(('bool', 'bytes16', 'int224'), D), T[bo, by, i])  # fixed case 33853
ck(f(('bool', 'bytes16', 'int232'), D), T[bo, by, i])  # fixed case 33854
ck(f(('bool', 'bytes16', 'int24'), D), T[bo, by, i])  # fixed case 33855
ck(f(('bool', 'bytes16', 'int240'), D), T[bo, by, i])  # fixed case 33856
ck(f(('bool', 'bytes16', 'int248'), D), T[bo, by, i])  # fixed case 33857
ck(f(('bool', 'bytes16', 'int256'), D), T[bo, by, i])  # fixed case 33858
ck(f(('bool', 'bytes16', 'int32'), D), T[bo, by, i])  # fixed case 33859
ck(f(('bool', 'bytes16', 'int40'), D), T[bo, by, i])  # fixed case 33860
ck(f(('bool', 'bytes16', 'int48'), D), T[bo, by, i])  # fixed case 33861
ck(f(('bool', 'bytes16', 'int56'), D), T[bo, by, i])  # fixed case 33862
ck(f(('bool', 'bytes16', 'int64'), D), T[bo, by, i])  # fixed case 33863
ck(f(('bool', 'bytes16', 'int72'), D), T[bo, by, i])  # fixed case 33864
ck(f(('bool', 'bytes16', 'int8'), D), T[bo, by, i])  # fixed case 33865
ck(f(('bool', 'bytes16', 'int80'), D), T[bo, by, i])  # fixed case 33866
ck(f(('bool', 'bytes16', 'int88'), D), T[bo, by, i])  # fixed case 33867
ck(f(('bool', 'bytes16', 'int96'), D), T[bo, by, i])  # fixed case 33868
ck(f(('bool', 'bytes16', 'string'), D), T[bo, by, s])  # fixed case 33869
ck(f(('bool', 'bytes16', 'string[]'), D), T[bo, by, A])  # fixed case 33870
ck(f(('bool', 'bytes16', 'uint104'), D), T[bo, by, i])  # fixed case 33871
ck(f(('bool', 'bytes16', 'uint112'), D), T[bo, by, i])  # fixed case 33872
ck(f(('bool', 'bytes16', 'uint120'), D), T[bo, by, i])  # fixed case 33873
ck(f(('bool', 'bytes16', 'uint128'), D), T[bo, by, i])  # fixed case 33874
ck(f(('bool', 'bytes16', 'uint136'), D), T[bo, by, i])  # fixed case 33875
ck(f(('bool', 'bytes16', 'uint144'), D), T[bo, by, i])  # fixed case 33876
ck(f(('bool', 'bytes16', 'uint152'), D), T[bo, by, i])  # fixed case 33877
ck(f(('bool', 'bytes16', 'uint16'), D), T[bo, by, i])  # fixed case 33878
ck(f(('bool', 'bytes16', 'uint160'), D), T[bo, by, i])  # fixed case 33879
ck(f(('bool', 'bytes16', 'uint168'), D), T[bo, by, i])  # fixed case 33880
ck(f(('bool', 'bytes16', 'uint176'), D), T[bo, by, i])  # fixed case 33881
ck(f(('bool', 'bytes16', 'uint184'), D), T[bo, by, i])  # fixed case 33882
ck(f(('bool', 'bytes16', 'uint192'), D), T[bo, by, i])  # fixed case 33883
ck(f(('bool', 'bytes16', 'uint200'), D), T[bo, by, i])  # fixed case 33884
ck(f(('bool', 'bytes16', 'uint208'), D), T[bo, by, i])  # fixed case 33885
ck(f(('bool', 'bytes16', 'uint216'), D), T[bo, by, i])  # fixed case 33886
ck(f(('bool', 'bytes16', 'uint224'), D), T[bo, by, i])  # fixed case 33887
ck(f(('bool', 'bytes16', 'uint232'), D), T[bo, by, i])  # fixed case 33888
ck(f(('bool', 'bytes16', 'uint24'), D), T[bo, by, i])  # fixed case 33889
ck(f(('bool', 'bytes16', 'uint240'), D), T[bo, by, i])  # fixed case 33890
ck(f(('bool', 'bytes16', 'uint248'), D), T[bo, by, i])  # fixed case 33891
ck(f(('bool', 'bytes16', 'uint256'), D), T[bo, by, i])  # fixed case 33892
ck(f(('bool', 'bytes16', 'uint32'), D), T[bo, by, i])  # fixed case 33893
ck(f(('bool', 'bytes16', 'uint40'), D), T[bo, by, i])  # fixed case 33894
ck(f(('bool', 'bytes16', 'uint48'), D), T[bo, by, i])  # fixed case 33895
ck(f(('bool', 'bytes16', 'uint56'), D), T[bo, by, i])  # fixed case 33896
ck(f(('bool', 'bytes16', 'uint64'), D), T[bo, by, i])  # fixed case 33897
ck(f(('bool', 'bytes16', 'uint72'), D), T[bo, by, i])  # fixed case 33898
ck(f(('bool', 'bytes16', 'uint8'), D), T[bo, by, i])  # fixed case 33899
ck(f(('bool', 'bytes16', 'uint80'), D), T[bo, by, i])  # fixed case 33900
ck(f(('bool', 'bytes16', 'uint88'), D), T[bo, by, i])  # fixed case 33901
ck(f(('bool', 'bytes16', 'uint96'), D), T[bo, by, i])  # fixed case 33902
ck(f(('bool', 'bytes16', '?'), D), T[bo, by, A])  # fixed case 33903
ck(f(('bool', 'bytes17', 'address'), D), T[bo, by, HA])  # fixed case 33904
ck(f(('bool', 'bytes17', 'address[]'), D), T[bo, by, A])  # fixed case 33905
ck(f(('bool', 'bytes17', 'bool'), D), T[bo, by, bo])  # fixed case 33906
ck(f(('bool', 'bytes17', 'bool[]'), D), T[bo, by, A])  # fixed case 33907
ck(f(('bool', 'bytes17', 'bytes'), D), T[bo, by, by])  # fixed case 33908
ck(f(('bool', 'bytes17', 'bytes1'), D), T[bo, by, by])  # fixed case 33909
ck(f(('bool', 'bytes17', 'bytes10'), D), T[bo, by, by])  # fixed case 33910
ck(f(('bool', 'bytes17', 'bytes11'), D), T[bo, by, by])  # fixed case 33911
ck(f(('bool', 'bytes17', 'bytes12'), D), T[bo, by, by])  # fixed case 33912
ck(f(('bool', 'bytes17', 'bytes13'), D), T[bo, by, by])  # fixed case 33913
ck(f(('bool', 'bytes17', 'bytes14'), D), T[bo, by, by])  # fixed case 33914
ck(f(('bool', 'bytes17', 'bytes15'), D), T[bo, by, by])  # fixed case 33915
ck(f(('bool', 'bytes17', 'bytes16'), D), T[bo, by, by])  # fixed case 33916
ck(f(('bool', 'bytes17', 'bytes17'), D), T[bo, by, by])  # fixed case 33917
ck(f(('bool', 'bytes17', 'bytes18'), D), T[bo, by, by])  # fixed case 33918
ck(f(('bool', 'bytes17', 'bytes19'), D), T[bo, by, by])  # fixed case 33919
ck(f(('bool', 'bytes17', 'bytes2'), D), T[bo, by, by])  # fixed case 33920
ck(f(('bool', 'bytes17', 'bytes20'), D), T[bo, by, by])  # fixed case 33921
ck(f(('bool', 'bytes17', 'bytes21'), D), T[bo, by, by])  # fixed case 33922
ck(f(('bool', 'bytes17', 'bytes22'), D), T[bo, by, by])  # fixed case 33923
ck(f(('bool', 'bytes17', 'bytes23'), D), T[bo, by, by])  # fixed case 33924
ck(f(('bool', 'bytes17', 'bytes24'), D), T[bo, by, by])  # fixed case 33925
ck(f(('bool', 'bytes17', 'bytes25'), D), T[bo, by, by])  # fixed case 33926
ck(f(('bool', 'bytes17', 'bytes26'), D), T[bo, by, by])  # fixed case 33927
ck(f(('bool', 'bytes17', 'bytes27'), D), T[bo, by, by])  # fixed case 33928
ck(f(('bool', 'bytes17', 'bytes28'), D), T[bo, by, by])  # fixed case 33929
ck(f(('bool', 'bytes17', 'bytes29'), D), T[bo, by, by])  # fixed case 33930
ck(f(('bool', 'bytes17', 'bytes3'), D), T[bo, by, by])  # fixed case 33931
ck(f(('bool', 'bytes17', 'bytes30'), D), T[bo, by, by])  # fixed case 33932
ck(f(('bool', 'bytes17', 'bytes31'), D), T[bo, by, by])  # fixed case 33933
ck(f(('bool', 'bytes17', 'bytes32'), D), T[bo, by, by])  # fixed case 33934
ck(f(('bool', 'bytes17', 'bytes4'), D), T[bo, by, by])  # fixed case 33935
ck(f(('bool', 'bytes17', 'bytes5'), D), T[bo, by, by])  # fixed case 33936
ck(f(('bool', 'bytes17', 'bytes6'), D), T[bo, by, by])  # fixed case 33937
ck(f(('bool', 'bytes17', 'bytes7'), D), T[bo, by, by])  # fixed case 33938
ck(f(('bool', 'bytes17', 'bytes8'), D), T[bo, by, by])  # fixed case 33939
ck(f(('bool', 'bytes17', 'bytes9'), D), T[bo, by, by])  # fixed case 33940
ck(f(('bool', 'bytes17', 'int104'), D), T[bo, by, i])  # fixed case 33941
ck(f(('bool', 'bytes17', 'int112'), D), T[bo, by, i])  # fixed case 33942
ck(f(('bool', 'bytes17', 'int120'), D), T[bo, by, i])  # fixed case 33943
ck(f(('bool', 'bytes17', 'int128'), D), T[bo, by, i])  # fixed case 33944
ck(f(('bool', 'bytes17', 'int136'), D), T[bo, by, i])  # fixed case 33945
ck(f(('bool', 'bytes17', 'int144'), D), T[bo, by, i])  # fixed case 33946
ck(f(('bool', 'bytes17', 'int152'), D), T[bo, by, i])  # fixed case 33947
ck(f(('bool', 'bytes17', 'int16'), D), T[bo, by, i])  # fixed case 33948
ck(f(('bool', 'bytes17', 'int160'), D), T[bo, by, i])  # fixed case 33949
ck(f(('bool', 'bytes17', 'int168'), D), T[bo, by, i])  # fixed case 33950
ck(f(('bool', 'bytes17', 'int176'), D), T[bo, by, i])  # fixed case 33951
ck(f(('bool', 'bytes17', 'int184'), D), T[bo, by, i])  # fixed case 33952
ck(f(('bool', 'bytes17', 'int192'), D), T[bo, by, i])  # fixed case 33953
ck(f(('bool', 'bytes17', 'int200'), D), T[bo, by, i])  # fixed case 33954
ck(f(('bool', 'bytes17', 'int208'), D), T[bo, by, i])  # fixed case 33955
ck(f(('bool', 'bytes17', 'int216'), D), T[bo, by, i])  # fixed case 33956
ck(f(('bool', 'bytes17', 'int224'), D), T[bo, by, i])  # fixed case 33957
ck(f(('bool', 'bytes17', 'int232'), D), T[bo, by, i])  # fixed case 33958
ck(f(('bool', 'bytes17', 'int24'), D), T[bo, by, i])  # fixed case 33959
ck(f(('bool', 'bytes17', 'int240'), D), T[bo, by, i])  # fixed case 33960
ck(f(('bool', 'bytes17', 'int248'), D), T[bo, by, i])  # fixed case 33961
ck(f(('bool', 'bytes17', 'int256'), D), T[bo, by, i])  # fixed case 33962
ck(f(('bool', 'bytes17', 'int32'), D), T[bo, by, i])  # fixed case 33963
ck(f(('bool', 'bytes17', 'int40'), D), T[bo, by, i])  # fixed case 33964
ck(f(('bool', 'bytes17', 'int48'), D), T[bo, by, i])  # fixed case 33965
ck(f(('bool', 'bytes17', 'int56'), D), T[bo, by, i])  # fixed case 33966
ck(f(('bool', 'bytes17', 'int64'), D), T[bo, by, i])  # fixed case 33967
ck(f(('bool', 'bytes17', 'int72'), D), T[bo, by, i])  # fixed case 33968
ck(f(('bool', 'bytes17', 'int8'), D), T[bo, by, i])  # fixed case 33969
ck(f(('bool', 'bytes17', 'int80'), D), T[bo, by, i])  # fixed case 33970
ck(f(('bool', 'bytes17', 'int88'), D), T[bo, by, i])  # fixed case 33971
ck(f(('bool', 'bytes17', 'int96'), D), T[bo, by, i])  # fixed case 33972
ck(f(('bool', 'bytes17', 'string'), D), T[bo, by, s])  # fixed case 33973
ck(f(('bool', 'bytes17', 'string[]'), D), T[bo, by, A])  # fixed case 33974
ck(f(('bool', 'bytes17', 'uint104'), D), T[bo, by, i])  # fixed case 33975
ck(f(('bool', 'bytes17', 'uint112'), D), T[bo, by, i])  # fixed case 33976
ck(f(('bool', 'bytes17', 'uint120'), D), T[bo, by, i])  # fixed case 33977
ck(f(('bool', 'bytes17', 'uint128'), D), T[bo, by, i])  # fixed case 33978
ck(f(('bool', 'bytes17', 'uint136'), D), T[bo, by, i])  # fixed case 33979
ck(f(('bool', 'bytes17', 'uint144'), D), T[bo, by, i])  # fixed case 33980
ck(f(('bool', 'bytes17', 'uint152'), D), T[bo, by, i])  # fixed case 33981
ck(f(('bool', 'bytes17', 'uint16'), D), T[bo, by, i])  # fixed case 33982
ck(f(('bool', 'bytes17', 'uint160'), D), T[bo, by, i])  # fixed case 33983
ck(f(('bool', 'bytes17', 'uint168'), D), T[bo, by, i])  # fixed case 33984
ck(f(('bool', 'bytes17', 'uint176'), D), T[bo, by, i])  # fixed case 33985
ck(f(('bool', 'bytes17', 'uint184'), D), T[bo, by, i])  # fixed case 33986
ck(f(('bool', 'bytes17', 'uint192'), D), T[bo, by, i])  # fixed case 33987
ck(f(('bool', 'bytes17', 'uint200'), D), T[bo, by, i])  # fixed case 33988
ck(f(('bool', 'bytes17', 'uint208'), D), T[bo, by, i])  # fixed case 33989
ck(f(('bool', 'bytes17', 'uint216'), D), T[bo, by, i])  # fixed case 33990
ck(f(('bool', 'bytes17', 'uint224'), D), T[bo, by, i])  # fixed case 33991
ck(f(('bool', 'bytes17', 'uint232'), D), T[bo, by, i])  # fixed case 33992
ck(f(('bool', 'bytes17', 'uint24'), D), T[bo, by, i])  # fixed case 33993
ck(f(('bool', 'bytes17', 'uint240'), D), T[bo, by, i])  # fixed case 33994
ck(f(('bool', 'bytes17', 'uint248'), D), T[bo, by, i])  # fixed case 33995
ck(f(('bool', 'bytes17', 'uint256'), D), T[bo, by, i])  # fixed case 33996
ck(f(('bool', 'bytes17', 'uint32'), D), T[bo, by, i])  # fixed case 33997
ck(f(('bool', 'bytes17', 'uint40'), D), T[bo, by, i])  # fixed case 33998
ck(f(('bool', 'bytes17', 'uint48'), D), T[bo, by, i])  # fixed case 33999
ck(f(('bool', 'bytes17', 'uint56'), D), T[bo, by, i])  # fixed case 34000
ck(f(('bool', 'bytes17', 'uint64'), D), T[bo, by, i])  # fixed case 34001
ck(f(('bool', 'bytes17', 'uint72'), D), T[bo, by, i])  # fixed case 34002
ck(f(('bool', 'bytes17', 'uint8'), D), T[bo, by, i])  # fixed case 34003
ck(f(('bool', 'bytes17', 'uint80'), D), T[bo, by, i])  # fixed case 34004
ck(f(('bool', 'bytes17', 'uint88'), D), T[bo, by, i])  # fixed case 34005
ck(f(('bool', 'bytes17', 'uint96'), D), T[bo, by, i])  # fixed case 34006
ck(f(('bool', 'bytes17', '?'), D), T[bo, by, A])  # fixed case 34007
ck(f(('bool', 'bytes18', 'address'), D), T[bo, by, HA])  # fixed case 34008
ck(f(('bool', 'bytes18', 'address[]'), D), T[bo, by, A])  # fixed case 34009
ck(f(('bool', 'bytes18', 'bool'), D), T[bo, by, bo])  # fixed case 34010
ck(f(('bool', 'bytes18', 'bool[]'), D), T[bo, by, A])  # fixed case 34011
ck(f(('bool', 'bytes18', 'bytes'), D), T[bo, by, by])  # fixed case 34012
ck(f(('bool', 'bytes18', 'bytes1'), D), T[bo, by, by])  # fixed case 34013
ck(f(('bool', 'bytes18', 'bytes10'), D), T[bo, by, by])  # fixed case 34014
ck(f(('bool', 'bytes18', 'bytes11'), D), T[bo, by, by])  # fixed case 34015
ck(f(('bool', 'bytes18', 'bytes12'), D), T[bo, by, by])  # fixed case 34016
ck(f(('bool', 'bytes18', 'bytes13'), D), T[bo, by, by])  # fixed case 34017
ck(f(('bool', 'bytes18', 'bytes14'), D), T[bo, by, by])  # fixed case 34018
ck(f(('bool', 'bytes18', 'bytes15'), D), T[bo, by, by])  # fixed case 34019
ck(f(('bool', 'bytes18', 'bytes16'), D), T[bo, by, by])  # fixed case 34020
ck(f(('bool', 'bytes18', 'bytes17'), D), T[bo, by, by])  # fixed case 34021
ck(f(('bool', 'bytes18', 'bytes18'), D), T[bo, by, by])  # fixed case 34022
ck(f(('bool', 'bytes18', 'bytes19'), D), T[bo, by, by])  # fixed case 34023
ck(f(('bool', 'bytes18', 'bytes2'), D), T[bo, by, by])  # fixed case 34024
ck(f(('bool', 'bytes18', 'bytes20'), D), T[bo, by, by])  # fixed case 34025
ck(f(('bool', 'bytes18', 'bytes21'), D), T[bo, by, by])  # fixed case 34026
ck(f(('bool', 'bytes18', 'bytes22'), D), T[bo, by, by])  # fixed case 34027
ck(f(('bool', 'bytes18', 'bytes23'), D), T[bo, by, by])  # fixed case 34028
ck(f(('bool', 'bytes18', 'bytes24'), D), T[bo, by, by])  # fixed case 34029
ck(f(('bool', 'bytes18', 'bytes25'), D), T[bo, by, by])  # fixed case 34030
ck(f(('bool', 'bytes18', 'bytes26'), D), T[bo, by, by])  # fixed case 34031
ck(f(('bool', 'bytes18', 'bytes27'), D), T[bo, by, by])  # fixed case 34032
ck(f(('bool', 'bytes18', 'bytes28'), D), T[bo, by, by])  # fixed case 34033
ck(f(('bool', 'bytes18', 'bytes29'), D), T[bo, by, by])  # fixed case 34034
ck(f(('bool', 'bytes18', 'bytes3'), D), T[bo, by, by])  # fixed case 34035
ck(f(('bool', 'bytes18', 'bytes30'), D), T[bo, by, by])  # fixed case 34036
ck(f(('bool', 'bytes18', 'bytes31'), D), T[bo, by, by])  # fixed case 34037
ck(f(('bool', 'bytes18', 'bytes32'), D), T[bo, by, by])  # fixed case 34038
ck(f(('bool', 'bytes18', 'bytes4'), D), T[bo, by, by])  # fixed case 34039
ck(f(('bool', 'bytes18', 'bytes5'), D), T[bo, by, by])  # fixed case 34040
ck(f(('bool', 'bytes18', 'bytes6'), D), T[bo, by, by])  # fixed case 34041
ck(f(('bool', 'bytes18', 'bytes7'), D), T[bo, by, by])  # fixed case 34042
ck(f(('bool', 'bytes18', 'bytes8'), D), T[bo, by, by])  # fixed case 34043
ck(f(('bool', 'bytes18', 'bytes9'), D), T[bo, by, by])  # fixed case 34044
ck(f(('bool', 'bytes18', 'int104'), D), T[bo, by, i])  # fixed case 34045
ck(f(('bool', 'bytes18', 'int112'), D), T[bo, by, i])  # fixed case 34046
ck(f(('bool', 'bytes18', 'int120'), D), T[bo, by, i])  # fixed case 34047
ck(f(('bool', 'bytes18', 'int128'), D), T[bo, by, i])  # fixed case 34048
ck(f(('bool', 'bytes18', 'int136'), D), T[bo, by, i])  # fixed case 34049
ck(f(('bool', 'bytes18', 'int144'), D), T[bo, by, i])  # fixed case 34050
ck(f(('bool', 'bytes18', 'int152'), D), T[bo, by, i])  # fixed case 34051
ck(f(('bool', 'bytes18', 'int16'), D), T[bo, by, i])  # fixed case 34052
ck(f(('bool', 'bytes18', 'int160'), D), T[bo, by, i])  # fixed case 34053
ck(f(('bool', 'bytes18', 'int168'), D), T[bo, by, i])  # fixed case 34054
ck(f(('bool', 'bytes18', 'int176'), D), T[bo, by, i])  # fixed case 34055
ck(f(('bool', 'bytes18', 'int184'), D), T[bo, by, i])  # fixed case 34056
ck(f(('bool', 'bytes18', 'int192'), D), T[bo, by, i])  # fixed case 34057
ck(f(('bool', 'bytes18', 'int200'), D), T[bo, by, i])  # fixed case 34058
ck(f(('bool', 'bytes18', 'int208'), D), T[bo, by, i])  # fixed case 34059
ck(f(('bool', 'bytes18', 'int216'), D), T[bo, by, i])  # fixed case 34060
ck(f(('bool', 'bytes18', 'int224'), D), T[bo, by, i])  # fixed case 34061
ck(f(('bool', 'bytes18', 'int232'), D), T[bo, by, i])  # fixed case 34062
ck(f(('bool', 'bytes18', 'int24'), D), T[bo, by, i])  # fixed case 34063
ck(f(('bool', 'bytes18', 'int240'), D), T[bo, by, i])  # fixed case 34064
ck(f(('bool', 'bytes18', 'int248'), D), T[bo, by, i])  # fixed case 34065
ck(f(('bool', 'bytes18', 'int256'), D), T[bo, by, i])  # fixed case 34066
ck(f(('bool', 'bytes18', 'int32'), D), T[bo, by, i])  # fixed case 34067
ck(f(('bool', 'bytes18', 'int40'), D), T[bo, by, i])  # fixed case 34068
ck(f(('bool', 'bytes18', 'int48'), D), T[bo, by, i])  # fixed case 34069
ck(f(('bool', 'bytes18', 'int56'), D), T[bo, by, i])  # fixed case 34070
ck(f(('bool', 'bytes18', 'int64'), D), T[bo, by, i])  # fixed case 34071
ck(f(('bool', 'bytes18', 'int72'), D), T[bo, by, i])  # fixed case 34072
ck(f(('bool', 'bytes18', 'int8'), D), T[bo, by, i])  # fixed case 34073
ck(f(('bool', 'bytes18', 'int80'), D), T[bo, by, i])  # fixed case 34074
ck(f(('bool', 'bytes18', 'int88'), D), T[bo, by, i])  # fixed case 34075
ck(f(('bool', 'bytes18', 'int96'), D), T[bo, by, i])  # fixed case 34076
ck(f(('bool', 'bytes18', 'string'), D), T[bo, by, s])  # fixed case 34077
ck(f(('bool', 'bytes18', 'string[]'), D), T[bo, by, A])  # fixed case 34078
ck(f(('bool', 'bytes18', 'uint104'), D), T[bo, by, i])  # fixed case 34079
ck(f(('bool', 'bytes18', 'uint112'), D), T[bo, by, i])  # fixed case 34080
ck(f(('bool', 'bytes18', 'uint120'), D), T[bo, by, i])  # fixed case 34081
ck(f(('bool', 'bytes18', 'uint128'), D), T[bo, by, i])  # fixed case 34082
ck(f(('bool', 'bytes18', 'uint136'), D), T[bo, by, i])  # fixed case 34083
ck(f(('bool', 'bytes18', 'uint144'), D), T[bo, by, i])  # fixed case 34084
ck(f(('bool', 'bytes18', 'uint152'), D), T[bo, by, i])  # fixed case 34085
ck(f(('bool', 'bytes18', 'uint16'), D), T[bo, by, i])  # fixed case 34086
ck(f(('bool', 'bytes18', 'uint160'), D), T[bo, by, i])  # fixed case 34087
ck(f(('bool', 'bytes18', 'uint168'), D), T[bo, by, i])  # fixed case 34088
ck(f(('bool', 'bytes18', 'uint176'), D), T[bo, by, i])  # fixed case 34089
ck(f(('bool', 'bytes18', 'uint184'), D), T[bo, by, i])  # fixed case 34090
ck(f(('bool', 'bytes18', 'uint192'), D), T[bo, by, i])  # fixed case 34091
ck(f(('bool', 'bytes18', 'uint200'), D), T[bo, by, i])  # fixed case 34092
ck(f(('bool', 'bytes18', 'uint208'), D), T[bo, by, i])  # fixed case 34093
ck(f(('bool', 'bytes18', 'uint216'), D), T[bo, by, i])  # fixed case 34094
ck(f(('bool', 'bytes18', 'uint224'), D), T[bo, by, i])  # fixed case 34095
ck(f(('bool', 'bytes18', 'uint232'), D), T[bo, by, i])  # fixed case 34096
ck(f(('bool', 'bytes18', 'uint24'), D), T[bo, by, i])  # fixed case 34097
ck(f(('bool', 'bytes18', 'uint240'), D), T[bo, by, i])  # fixed case 34098
ck(f(('bool', 'bytes18', 'uint248'), D), T[bo, by, i])  # fixed case 34099
ck(f(('bool', 'bytes18', 'uint256'), D), T[bo, by, i])  # fixed case 34100
ck(f(('bool', 'bytes18', 'uint32'), D), T[bo, by, i])  # fixed case 34101
ck(f(('bool', 'bytes18', 'uint40'), D), T[bo, by, i])  # fixed case 34102
ck(f(('bool', 'bytes18', 'uint48'), D), T[bo, by, i])  # fixed case 34103
ck(f(('bool', 'bytes18', 'uint56'), D), T[bo, by, i])  # fixed case 34104
ck(f(('bool', 'bytes18', 'uint64'), D), T[bo, by, i])  # fixed case 34105
ck(f(('bool', 'bytes18', 'uint72'), D), T[bo, by, i])  # fixed case 34106
ck(f(('bool', 'bytes18', 'uint8'), D), T[bo, by, i])  # fixed case 34107
ck(f(('bool', 'bytes18', 'uint80'), D), T[bo, by, i])  # fixed case 34108
ck(f(('bool', 'bytes18', 'uint88'), D), T[bo, by, i])  # fixed case 34109
ck(f(('bool', 'bytes18', 'uint96'), D), T[bo, by, i])  # fixed case 34110
ck(f(('bool', 'bytes18', '?'), D), T[bo, by, A])  # fixed case 34111
ck(f(('bool', 'bytes19', 'address'), D), T[bo, by, HA])  # fixed case 34112
ck(f(('bool', 'bytes19', 'address[]'), D), T[bo, by, A])  # fixed case 34113
ck(f(('bool', 'bytes19', 'bool'), D), T[bo, by, bo])  # fixed case 34114
ck(f(('bool', 'bytes19', 'bool[]'), D), T[bo, by, A])  # fixed case 34115
ck(f(('bool', 'bytes19', 'bytes'), D), T[bo, by, by])  # fixed case 34116
ck(f(('bool', 'bytes19', 'bytes1'), D), T[bo, by, by])  # fixed case 34117
ck(f(('bool', 'bytes19', 'bytes10'), D), T[bo, by, by])  # fixed case 34118
ck(f(('bool', 'bytes19', 'bytes11'), D), T[bo, by, by])  # fixed case 34119
ck(f(('bool', 'bytes19', 'bytes12'), D), T[bo, by, by])  # fixed case 34120
ck(f(('bool', 'bytes19', 'bytes13'), D), T[bo, by, by])  # fixed case 34121
ck(f(('bool', 'bytes19', 'bytes14'), D), T[bo, by, by])  # fixed case 34122
ck(f(('bool', 'bytes19', 'bytes15'), D), T[bo, by, by])  # fixed case 34123
ck(f(('bool', 'bytes19', 'bytes16'), D), T[bo, by, by])  # fixed case 34124
ck(f(('bool', 'bytes19', 'bytes17'), D), T[bo, by, by])  # fixed case 34125
ck(f(('bool', 'bytes19', 'bytes18'), D), T[bo, by, by])  # fixed case 34126
ck(f(('bool', 'bytes19', 'bytes19'), D), T[bo, by, by])  # fixed case 34127
ck(f(('bool', 'bytes19', 'bytes2'), D), T[bo, by, by])  # fixed case 34128
ck(f(('bool', 'bytes19', 'bytes20'), D), T[bo, by, by])  # fixed case 34129
ck(f(('bool', 'bytes19', 'bytes21'), D), T[bo, by, by])  # fixed case 34130
ck(f(('bool', 'bytes19', 'bytes22'), D), T[bo, by, by])  # fixed case 34131
ck(f(('bool', 'bytes19', 'bytes23'), D), T[bo, by, by])  # fixed case 34132
ck(f(('bool', 'bytes19', 'bytes24'), D), T[bo, by, by])  # fixed case 34133
ck(f(('bool', 'bytes19', 'bytes25'), D), T[bo, by, by])  # fixed case 34134
ck(f(('bool', 'bytes19', 'bytes26'), D), T[bo, by, by])  # fixed case 34135
ck(f(('bool', 'bytes19', 'bytes27'), D), T[bo, by, by])  # fixed case 34136
ck(f(('bool', 'bytes19', 'bytes28'), D), T[bo, by, by])  # fixed case 34137
ck(f(('bool', 'bytes19', 'bytes29'), D), T[bo, by, by])  # fixed case 34138
ck(f(('bool', 'bytes19', 'bytes3'), D), T[bo, by, by])  # fixed case 34139
ck(f(('bool', 'bytes19', 'bytes30'), D), T[bo, by, by])  # fixed case 34140
ck(f(('bool', 'bytes19', 'bytes31'), D), T[bo, by, by])  # fixed case 34141
ck(f(('bool', 'bytes19', 'bytes32'), D), T[bo, by, by])  # fixed case 34142
ck(f(('bool', 'bytes19', 'bytes4'), D), T[bo, by, by])  # fixed case 34143
ck(f(('bool', 'bytes19', 'bytes5'), D), T[bo, by, by])  # fixed case 34144
ck(f(('bool', 'bytes19', 'bytes6'), D), T[bo, by, by])  # fixed case 34145
ck(f(('bool', 'bytes19', 'bytes7'), D), T[bo, by, by])  # fixed case 34146
ck(f(('bool', 'bytes19', 'bytes8'), D), T[bo, by, by])  # fixed case 34147
ck(f(('bool', 'bytes19', 'bytes9'), D), T[bo, by, by])  # fixed case 34148
ck(f(('bool', 'bytes19', 'int104'), D), T[bo, by, i])  # fixed case 34149
ck(f(('bool', 'bytes19', 'int112'), D), T[bo, by, i])  # fixed case 34150
ck(f(('bool', 'bytes19', 'int120'), D), T[bo, by, i])  # fixed case 34151
ck(f(('bool', 'bytes19', 'int128'), D), T[bo, by, i])  # fixed case 34152
ck(f(('bool', 'bytes19', 'int136'), D), T[bo, by, i])  # fixed case 34153
ck(f(('bool', 'bytes19', 'int144'), D), T[bo, by, i])  # fixed case 34154
ck(f(('bool', 'bytes19', 'int152'), D), T[bo, by, i])  # fixed case 34155
ck(f(('bool', 'bytes19', 'int16'), D), T[bo, by, i])  # fixed case 34156
ck(f(('bool', 'bytes19', 'int160'), D), T[bo, by, i])  # fixed case 34157
ck(f(('bool', 'bytes19', 'int168'), D), T[bo, by, i])  # fixed case 34158
ck(f(('bool', 'bytes19', 'int176'), D), T[bo, by, i])  # fixed case 34159
ck(f(('bool', 'bytes19', 'int184'), D), T[bo, by, i])  # fixed case 34160
ck(f(('bool', 'bytes19', 'int192'), D), T[bo, by, i])  # fixed case 34161
ck(f(('bool', 'bytes19', 'int200'), D), T[bo, by, i])  # fixed case 34162
ck(f(('bool', 'bytes19', 'int208'), D), T[bo, by, i])  # fixed case 34163
ck(f(('bool', 'bytes19', 'int216'), D), T[bo, by, i])  # fixed case 34164
ck(f(('bool', 'bytes19', 'int224'), D), T[bo, by, i])  # fixed case 34165
ck(f(('bool', 'bytes19', 'int232'), D), T[bo, by, i])  # fixed case 34166
ck(f(('bool', 'bytes19', 'int24'), D), T[bo, by, i])  # fixed case 34167
ck(f(('bool', 'bytes19', 'int240'), D), T[bo, by, i])  # fixed case 34168
ck(f(('bool', 'bytes19', 'int248'), D), T[bo, by, i])  # fixed case 34169
ck(f(('bool', 'bytes19', 'int256'), D), T[bo, by, i])  # fixed case 34170
ck(f(('bool', 'bytes19', 'int32'), D), T[bo, by, i])  # fixed case 34171
ck(f(('bool', 'bytes19', 'int40'), D), T[bo, by, i])  # fixed case 34172
ck(f(('bool', 'bytes19', 'int48'), D), T[bo, by, i])  # fixed case 34173
ck(f(('bool', 'bytes19', 'int56'), D), T[bo, by, i])  # fixed case 34174
ck(f(('bool', 'bytes19', 'int64'), D), T[bo, by, i])  # fixed case 34175
ck(f(('bool', 'bytes19', 'int72'), D), T[bo, by, i])  # fixed case 34176
ck(f(('bool', 'bytes19', 'int8'), D), T[bo, by, i])  # fixed case 34177
ck(f(('bool', 'bytes19', 'int80'), D), T[bo, by, i])  # fixed case 34178
ck(f(('bool', 'bytes19', 'int88'), D), T[bo, by, i])  # fixed case 34179
ck(f(('bool', 'bytes19', 'int96'), D), T[bo, by, i])  # fixed case 34180
ck(f(('bool', 'bytes19', 'string'), D), T[bo, by, s])  # fixed case 34181
ck(f(('bool', 'bytes19', 'string[]'), D), T[bo, by, A])  # fixed case 34182
ck(f(('bool', 'bytes19', 'uint104'), D), T[bo, by, i])  # fixed case 34183
ck(f(('bool', 'bytes19', 'uint112'), D), T[bo, by, i])  # fixed case 34184
ck(f(('bool', 'bytes19', 'uint120'), D), T[bo, by, i])  # fixed case 34185
ck(f(('bool', 'bytes19', 'uint128'), D), T[bo, by, i])  # fixed case 34186
ck(f(('bool', 'bytes19', 'uint136'), D), T[bo, by, i])  # fixed case 34187
ck(f(('bool', 'bytes19', 'uint144'), D), T[bo, by, i])  # fixed case 34188
ck(f(('bool', 'bytes19', 'uint152'), D), T[bo, by, i])  # fixed case 34189
ck(f(('bool', 'bytes19', 'uint16'), D), T[bo, by, i])  # fixed case 34190
ck(f(('bool', 'bytes19', 'uint160'), D), T[bo, by, i])  # fixed case 34191
ck(f(('bool', 'bytes19', 'uint168'), D), T[bo, by, i])  # fixed case 34192
ck(f(('bool', 'bytes19', 'uint176'), D), T[bo, by, i])  # fixed case 34193
ck(f(('bool', 'bytes19', 'uint184'), D), T[bo, by, i])  # fixed case 34194
ck(f(('bool', 'bytes19', 'uint192'), D), T[bo, by, i])  # fixed case 34195
ck(f(('bool', 'bytes19', 'uint200'), D), T[bo, by, i])  # fixed case 34196
ck(f(('bool', 'bytes19', 'uint208'), D), T[bo, by, i])  # fixed case 34197
ck(f(('bool', 'bytes19', 'uint216'), D), T[bo, by, i])  # fixed case 34198
ck(f(('bool', 'bytes19', 'uint224'), D), T[bo, by, i])  # fixed case 34199
ck(f(('bool', 'bytes19', 'uint232'), D), T[bo, by, i])  # fixed case 34200
ck(f(('bool', 'bytes19', 'uint24'), D), T[bo, by, i])  # fixed case 34201
ck(f(('bool', 'bytes19', 'uint240'), D), T[bo, by, i])  # fixed case 34202
ck(f(('bool', 'bytes19', 'uint248'), D), T[bo, by, i])  # fixed case 34203
ck(f(('bool', 'bytes19', 'uint256'), D), T[bo, by, i])  # fixed case 34204
ck(f(('bool', 'bytes19', 'uint32'), D), T[bo, by, i])  # fixed case 34205
ck(f(('bool', 'bytes19', 'uint40'), D), T[bo, by, i])  # fixed case 34206
ck(f(('bool', 'bytes19', 'uint48'), D), T[bo, by, i])  # fixed case 34207
ck(f(('bool', 'bytes19', 'uint56'), D), T[bo, by, i])  # fixed case 34208
ck(f(('bool', 'bytes19', 'uint64'), D), T[bo, by, i])  # fixed case 34209
ck(f(('bool', 'bytes19', 'uint72'), D), T[bo, by, i])  # fixed case 34210
ck(f(('bool', 'bytes19', 'uint8'), D), T[bo, by, i])  # fixed case 34211
ck(f(('bool', 'bytes19', 'uint80'), D), T[bo, by, i])  # fixed case 34212
ck(f(('bool', 'bytes19', 'uint88'), D), T[bo, by, i])  # fixed case 34213
ck(f(('bool', 'bytes19', 'uint96'), D), T[bo, by, i])  # fixed case 34214
ck(f(('bool', 'bytes19', '?'), D), T[bo, by, A])  # fixed case 34215
ck(f(('bool', 'bytes2', 'address'), D), T[bo, by, HA])  # fixed case 34216
ck(f(('bool', 'bytes2', 'address[]'), D), T[bo, by, A])  # fixed case 34217
ck(f(('bool', 'bytes2', 'bool'), D), T[bo, by, bo])  # fixed case 34218
ck(f(('bool', 'bytes2', 'bool[]'), D), T[bo, by, A])  # fixed case 34219
ck(f(('bool', 'bytes2', 'bytes'), D), T[bo, by, by])  # fixed case 34220
ck(f(('bool', 'bytes2', 'bytes1'), D), T[bo, by, by])  # fixed case 34221
ck(f(('bool', 'bytes2', 'bytes10'), D), T[bo, by, by])  # fixed case 34222
ck(f(('bool', 'bytes2', 'bytes11'), D), T[bo, by, by])  # fixed case 34223
ck(f(('bool', 'bytes2', 'bytes12'), D), T[bo, by, by])  # fixed case 34224
ck(f(('bool', 'bytes2', 'bytes13'), D), T[bo, by, by])  # fixed case 34225
ck(f(('bool', 'bytes2', 'bytes14'), D), T[bo, by, by])  # fixed case 34226
ck(f(('bool', 'bytes2', 'bytes15'), D), T[bo, by, by])  # fixed case 34227
ck(f(('bool', 'bytes2', 'bytes16'), D), T[bo, by, by])  # fixed case 34228
ck(f(('bool', 'bytes2', 'bytes17'), D), T[bo, by, by])  # fixed case 34229
ck(f(('bool', 'bytes2', 'bytes18'), D), T[bo, by, by])  # fixed case 34230
ck(f(('bool', 'bytes2', 'bytes19'), D), T[bo, by, by])  # fixed case 34231
ck(f(('bool', 'bytes2', 'bytes2'), D), T[bo, by, by])  # fixed case 34232
ck(f(('bool', 'bytes2', 'bytes20'), D), T[bo, by, by])  # fixed case 34233
ck(f(('bool', 'bytes2', 'bytes21'), D), T[bo, by, by])  # fixed case 34234
ck(f(('bool', 'bytes2', 'bytes22'), D), T[bo, by, by])  # fixed case 34235
ck(f(('bool', 'bytes2', 'bytes23'), D), T[bo, by, by])  # fixed case 34236
ck(f(('bool', 'bytes2', 'bytes24'), D), T[bo, by, by])  # fixed case 34237
ck(f(('bool', 'bytes2', 'bytes25'), D), T[bo, by, by])  # fixed case 34238
ck(f(('bool', 'bytes2', 'bytes26'), D), T[bo, by, by])  # fixed case 34239
ck(f(('bool', 'bytes2', 'bytes27'), D), T[bo, by, by])  # fixed case 34240
ck(f(('bool', 'bytes2', 'bytes28'), D), T[bo, by, by])  # fixed case 34241
ck(f(('bool', 'bytes2', 'bytes29'), D), T[bo, by, by])  # fixed case 34242
ck(f(('bool', 'bytes2', 'bytes3'), D), T[bo, by, by])  # fixed case 34243
ck(f(('bool', 'bytes2', 'bytes30'), D), T[bo, by, by])  # fixed case 34244
ck(f(('bool', 'bytes2', 'bytes31'), D), T[bo, by, by])  # fixed case 34245
ck(f(('bool', 'bytes2', 'bytes32'), D), T[bo, by, by])  # fixed case 34246
ck(f(('bool', 'bytes2', 'bytes4'), D), T[bo, by, by])  # fixed case 34247
ck(f(('bool', 'bytes2', 'bytes5'), D), T[bo, by, by])  # fixed case 34248
ck(f(('bool', 'bytes2', 'bytes6'), D), T[bo, by, by])  # fixed case 34249
ck(f(('bool', 'bytes2', 'bytes7'), D), T[bo, by, by])  # fixed case 34250
ck(f(('bool', 'bytes2', 'bytes8'), D), T[bo, by, by])  # fixed case 34251
ck(f(('bool', 'bytes2', 'bytes9'), D), T[bo, by, by])  # fixed case 34252
ck(f(('bool', 'bytes2', 'int104'), D), T[bo, by, i])  # fixed case 34253
ck(f(('bool', 'bytes2', 'int112'), D), T[bo, by, i])  # fixed case 34254
ck(f(('bool', 'bytes2', 'int120'), D), T[bo, by, i])  # fixed case 34255
ck(f(('bool', 'bytes2', 'int128'), D), T[bo, by, i])  # fixed case 34256
ck(f(('bool', 'bytes2', 'int136'), D), T[bo, by, i])  # fixed case 34257
ck(f(('bool', 'bytes2', 'int144'), D), T[bo, by, i])  # fixed case 34258
ck(f(('bool', 'bytes2', 'int152'), D), T[bo, by, i])  # fixed case 34259
ck(f(('bool', 'bytes2', 'int16'), D), T[bo, by, i])  # fixed case 34260
ck(f(('bool', 'bytes2', 'int160'), D), T[bo, by, i])  # fixed case 34261
ck(f(('bool', 'bytes2', 'int168'), D), T[bo, by, i])  # fixed case 34262
ck(f(('bool', 'bytes2', 'int176'), D), T[bo, by, i])  # fixed case 34263
ck(f(('bool', 'bytes2', 'int184'), D), T[bo, by, i])  # fixed case 34264
ck(f(('bool', 'bytes2', 'int192'), D), T[bo, by, i])  # fixed case 34265
ck(f(('bool', 'bytes2', 'int200'), D), T[bo, by, i])  # fixed case 34266
ck(f(('bool', 'bytes2', 'int208'), D), T[bo, by, i])  # fixed case 34267
ck(f(('bool', 'bytes2', 'int216'), D), T[bo, by, i])  # fixed case 34268
ck(f(('bool', 'bytes2', 'int224'), D), T[bo, by, i])  # fixed case 34269
ck(f(('bool', 'bytes2', 'int232'), D), T[bo, by, i])  # fixed case 34270
ck(f(('bool', 'bytes2', 'int24'), D), T[bo, by, i])  # fixed case 34271
ck(f(('bool', 'bytes2', 'int240'), D), T[bo, by, i])  # fixed case 34272
ck(f(('bool', 'bytes2', 'int248'), D), T[bo, by, i])  # fixed case 34273
ck(f(('bool', 'bytes2', 'int256'), D), T[bo, by, i])  # fixed case 34274
ck(f(('bool', 'bytes2', 'int32'), D), T[bo, by, i])  # fixed case 34275
ck(f(('bool', 'bytes2', 'int40'), D), T[bo, by, i])  # fixed case 34276
ck(f(('bool', 'bytes2', 'int48'), D), T[bo, by, i])  # fixed case 34277
ck(f(('bool', 'bytes2', 'int56'), D), T[bo, by, i])  # fixed case 34278
ck(f(('bool', 'bytes2', 'int64'), D), T[bo, by, i])  # fixed case 34279
ck(f(('bool', 'bytes2', 'int72'), D), T[bo, by, i])  # fixed case 34280
ck(f(('bool', 'bytes2', 'int8'), D), T[bo, by, i])  # fixed case 34281
ck(f(('bool', 'bytes2', 'int80'), D), T[bo, by, i])  # fixed case 34282
ck(f(('bool', 'bytes2', 'int88'), D), T[bo, by, i])  # fixed case 34283
ck(f(('bool', 'bytes2', 'int96'), D), T[bo, by, i])  # fixed case 34284
ck(f(('bool', 'bytes2', 'string'), D), T[bo, by, s])  # fixed case 34285
ck(f(('bool', 'bytes2', 'string[]'), D), T[bo, by, A])  # fixed case 34286
ck(f(('bool', 'bytes2', 'uint104'), D), T[bo, by, i])  # fixed case 34287
ck(f(('bool', 'bytes2', 'uint112'), D), T[bo, by, i])  # fixed case 34288
ck(f(('bool', 'bytes2', 'uint120'), D), T[bo, by, i])  # fixed case 34289
ck(f(('bool', 'bytes2', 'uint128'), D), T[bo, by, i])  # fixed case 34290
ck(f(('bool', 'bytes2', 'uint136'), D), T[bo, by, i])  # fixed case 34291
ck(f(('bool', 'bytes2', 'uint144'), D), T[bo, by, i])  # fixed case 34292
ck(f(('bool', 'bytes2', 'uint152'), D), T[bo, by, i])  # fixed case 34293
ck(f(('bool', 'bytes2', 'uint16'), D), T[bo, by, i])  # fixed case 34294
ck(f(('bool', 'bytes2', 'uint160'), D), T[bo, by, i])  # fixed case 34295
ck(f(('bool', 'bytes2', 'uint168'), D), T[bo, by, i])  # fixed case 34296
ck(f(('bool', 'bytes2', 'uint176'), D), T[bo, by, i])  # fixed case 34297
ck(f(('bool', 'bytes2', 'uint184'), D), T[bo, by, i])  # fixed case 34298
ck(f(('bool', 'bytes2', 'uint192'), D), T[bo, by, i])  # fixed case 34299
ck(f(('bool', 'bytes2', 'uint200'), D), T[bo, by, i])  # fixed case 34300
ck(f(('bool', 'bytes2', 'uint208'), D), T[bo, by, i])  # fixed case 34301
ck(f(('bool', 'bytes2', 'uint216'), D), T[bo, by, i])  # fixed case 34302
ck(f(('bool', 'bytes2', 'uint224'), D), T[bo, by, i])  # fixed case 34303
ck(f(('bool', 'bytes2', 'uint232'), D), T[bo, by, i])  # fixed case 34304
ck(f(('bool', 'bytes2', 'uint24'), D), T[bo, by, i])  # fixed case 34305
ck(f(('bool', 'bytes2', 'uint240'), D), T[bo, by, i])  # fixed case 34306
ck(f(('bool', 'bytes2', 'uint248'), D), T[bo, by, i])  # fixed case 34307
ck(f(('bool', 'bytes2', 'uint256'), D), T[bo, by, i])  # fixed case 34308
ck(f(('bool', 'bytes2', 'uint32'), D), T[bo, by, i])  # fixed case 34309
ck(f(('bool', 'bytes2', 'uint40'), D), T[bo, by, i])  # fixed case 34310
ck(f(('bool', 'bytes2', 'uint48'), D), T[bo, by, i])  # fixed case 34311
ck(f(('bool', 'bytes2', 'uint56'), D), T[bo, by, i])  # fixed case 34312
ck(f(('bool', 'bytes2', 'uint64'), D), T[bo, by, i])  # fixed case 34313
ck(f(('bool', 'bytes2', 'uint72'), D), T[bo, by, i])  # fixed case 34314
ck(f(('bool', 'bytes2', 'uint8'), D), T[bo, by, i])  # fixed case 34315
ck(f(('bool', 'bytes2', 'uint80'), D), T[bo, by, i])  # fixed case 34316
ck(f(('bool', 'bytes2', 'uint88'), D), T[bo, by, i])  # fixed case 34317
ck(f(('bool', 'bytes2', 'uint96'), D), T[bo, by, i])  # fixed case 34318
ck(f(('bool', 'bytes2', '?'), D), T[bo, by, A])  # fixed case 34319
ck(f(('bool', 'bytes20', 'address'), D), T[bo, by, HA])  # fixed case 34320
ck(f(('bool', 'bytes20', 'address[]'), D), T[bo, by, A])  # fixed case 34321
ck(f(('bool', 'bytes20', 'bool'), D), T[bo, by, bo])  # fixed case 34322
ck(f(('bool', 'bytes20', 'bool[]'), D), T[bo, by, A])  # fixed case 34323
ck(f(('bool', 'bytes20', 'bytes'), D), T[bo, by, by])  # fixed case 34324
ck(f(('bool', 'bytes20', 'bytes1'), D), T[bo, by, by])  # fixed case 34325
ck(f(('bool', 'bytes20', 'bytes10'), D), T[bo, by, by])  # fixed case 34326
ck(f(('bool', 'bytes20', 'bytes11'), D), T[bo, by, by])  # fixed case 34327
ck(f(('bool', 'bytes20', 'bytes12'), D), T[bo, by, by])  # fixed case 34328
ck(f(('bool', 'bytes20', 'bytes13'), D), T[bo, by, by])  # fixed case 34329
ck(f(('bool', 'bytes20', 'bytes14'), D), T[bo, by, by])  # fixed case 34330
ck(f(('bool', 'bytes20', 'bytes15'), D), T[bo, by, by])  # fixed case 34331
ck(f(('bool', 'bytes20', 'bytes16'), D), T[bo, by, by])  # fixed case 34332
ck(f(('bool', 'bytes20', 'bytes17'), D), T[bo, by, by])  # fixed case 34333
ck(f(('bool', 'bytes20', 'bytes18'), D), T[bo, by, by])  # fixed case 34334
ck(f(('bool', 'bytes20', 'bytes19'), D), T[bo, by, by])  # fixed case 34335
ck(f(('bool', 'bytes20', 'bytes2'), D), T[bo, by, by])  # fixed case 34336
ck(f(('bool', 'bytes20', 'bytes20'), D), T[bo, by, by])  # fixed case 34337
ck(f(('bool', 'bytes20', 'bytes21'), D), T[bo, by, by])  # fixed case 34338
ck(f(('bool', 'bytes20', 'bytes22'), D), T[bo, by, by])  # fixed case 34339
ck(f(('bool', 'bytes20', 'bytes23'), D), T[bo, by, by])  # fixed case 34340
ck(f(('bool', 'bytes20', 'bytes24'), D), T[bo, by, by])  # fixed case 34341
ck(f(('bool', 'bytes20', 'bytes25'), D), T[bo, by, by])  # fixed case 34342
ck(f(('bool', 'bytes20', 'bytes26'), D), T[bo, by, by])  # fixed case 34343
ck(f(('bool', 'bytes20', 'bytes27'), D), T[bo, by, by])  # fixed case 34344
ck(f(('bool', 'bytes20', 'bytes28'), D), T[bo, by, by])  # fixed case 34345
ck(f(('bool', 'bytes20', 'bytes29'), D), T[bo, by, by])  # fixed case 34346
ck(f(('bool', 'bytes20', 'bytes3'), D), T[bo, by, by])  # fixed case 34347
ck(f(('bool', 'bytes20', 'bytes30'), D), T[bo, by, by])  # fixed case 34348
ck(f(('bool', 'bytes20', 'bytes31'), D), T[bo, by, by])  # fixed case 34349
ck(f(('bool', 'bytes20', 'bytes32'), D), T[bo, by, by])  # fixed case 34350
ck(f(('bool', 'bytes20', 'bytes4'), D), T[bo, by, by])  # fixed case 34351
ck(f(('bool', 'bytes20', 'bytes5'), D), T[bo, by, by])  # fixed case 34352
ck(f(('bool', 'bytes20', 'bytes6'), D), T[bo, by, by])  # fixed case 34353
ck(f(('bool', 'bytes20', 'bytes7'), D), T[bo, by, by])  # fixed case 34354
ck(f(('bool', 'bytes20', 'bytes8'), D), T[bo, by, by])  # fixed case 34355
ck(f(('bool', 'bytes20', 'bytes9'), D), T[bo, by, by])  # fixed case 34356
ck(f(('bool', 'bytes20', 'int104'), D), T[bo, by, i])  # fixed case 34357
ck(f(('bool', 'bytes20', 'int112'), D), T[bo, by, i])  # fixed case 34358
ck(f(('bool', 'bytes20', 'int120'), D), T[bo, by, i])  # fixed case 34359
ck(f(('bool', 'bytes20', 'int128'), D), T[bo, by, i])  # fixed case 34360
ck(f(('bool', 'bytes20', 'int136'), D), T[bo, by, i])  # fixed case 34361
ck(f(('bool', 'bytes20', 'int144'), D), T[bo, by, i])  # fixed case 34362
ck(f(('bool', 'bytes20', 'int152'), D), T[bo, by, i])  # fixed case 34363
ck(f(('bool', 'bytes20', 'int16'), D), T[bo, by, i])  # fixed case 34364
ck(f(('bool', 'bytes20', 'int160'), D), T[bo, by, i])  # fixed case 34365
ck(f(('bool', 'bytes20', 'int168'), D), T[bo, by, i])  # fixed case 34366
ck(f(('bool', 'bytes20', 'int176'), D), T[bo, by, i])  # fixed case 34367
ck(f(('bool', 'bytes20', 'int184'), D), T[bo, by, i])  # fixed case 34368
ck(f(('bool', 'bytes20', 'int192'), D), T[bo, by, i])  # fixed case 34369
ck(f(('bool', 'bytes20', 'int200'), D), T[bo, by, i])  # fixed case 34370
ck(f(('bool', 'bytes20', 'int208'), D), T[bo, by, i])  # fixed case 34371
ck(f(('bool', 'bytes20', 'int216'), D), T[bo, by, i])  # fixed case 34372
ck(f(('bool', 'bytes20', 'int224'), D), T[bo, by, i])  # fixed case 34373
ck(f(('bool', 'bytes20', 'int232'), D), T[bo, by, i])  # fixed case 34374
ck(f(('bool', 'bytes20', 'int24'), D), T[bo, by, i])  # fixed case 34375
ck(f(('bool', 'bytes20', 'int240'), D), T[bo, by, i])  # fixed case 34376
ck(f(('bool', 'bytes20', 'int248'), D), T[bo, by, i])  # fixed case 34377
ck(f(('bool', 'bytes20', 'int256'), D), T[bo, by, i])  # fixed case 34378
ck(f(('bool', 'bytes20', 'int32'), D), T[bo, by, i])  # fixed case 34379
ck(f(('bool', 'bytes20', 'int40'), D), T[bo, by, i])  # fixed case 34380
ck(f(('bool', 'bytes20', 'int48'), D), T[bo, by, i])  # fixed case 34381
ck(f(('bool', 'bytes20', 'int56'), D), T[bo, by, i])  # fixed case 34382
ck(f(('bool', 'bytes20', 'int64'), D), T[bo, by, i])  # fixed case 34383
ck(f(('bool', 'bytes20', 'int72'), D), T[bo, by, i])  # fixed case 34384
ck(f(('bool', 'bytes20', 'int8'), D), T[bo, by, i])  # fixed case 34385
ck(f(('bool', 'bytes20', 'int80'), D), T[bo, by, i])  # fixed case 34386
ck(f(('bool', 'bytes20', 'int88'), D), T[bo, by, i])  # fixed case 34387
ck(f(('bool', 'bytes20', 'int96'), D), T[bo, by, i])  # fixed case 34388
ck(f(('bool', 'bytes20', 'string'), D), T[bo, by, s])  # fixed case 34389
ck(f(('bool', 'bytes20', 'string[]'), D), T[bo, by, A])  # fixed case 34390
ck(f(('bool', 'bytes20', 'uint104'), D), T[bo, by, i])  # fixed case 34391
ck(f(('bool', 'bytes20', 'uint112'), D), T[bo, by, i])  # fixed case 34392
ck(f(('bool', 'bytes20', 'uint120'), D), T[bo, by, i])  # fixed case 34393
ck(f(('bool', 'bytes20', 'uint128'), D), T[bo, by, i])  # fixed case 34394
ck(f(('bool', 'bytes20', 'uint136'), D), T[bo, by, i])  # fixed case 34395
ck(f(('bool', 'bytes20', 'uint144'), D), T[bo, by, i])  # fixed case 34396
ck(f(('bool', 'bytes20', 'uint152'), D), T[bo, by, i])  # fixed case 34397
ck(f(('bool', 'bytes20', 'uint16'), D), T[bo, by, i])  # fixed case 34398
ck(f(('bool', 'bytes20', 'uint160'), D), T[bo, by, i])  # fixed case 34399
ck(f(('bool', 'bytes20', 'uint168'), D), T[bo, by, i])  # fixed case 34400
ck(f(('bool', 'bytes20', 'uint176'), D), T[bo, by, i])  # fixed case 34401
ck(f(('bool', 'bytes20', 'uint184'), D), T[bo, by, i])  # fixed case 34402
ck(f(('bool', 'bytes20', 'uint192'), D), T[bo, by, i])  # fixed case 34403
ck(f(('bool', 'bytes20', 'uint200'), D), T[bo, by, i])  # fixed case 34404
ck(f(('bool', 'bytes20', 'uint208'), D), T[bo, by, i])  # fixed case 34405
ck(f(('bool', 'bytes20', 'uint216'), D), T[bo, by, i])  # fixed case 34406
ck(f(('bool', 'bytes20', 'uint224'), D), T[bo, by, i])  # fixed case 34407
ck(f(('bool', 'bytes20', 'uint232'), D), T[bo, by, i])  # fixed case 34408
ck(f(('bool', 'bytes20', 'uint24'), D), T[bo, by, i])  # fixed case 34409
ck(f(('bool', 'bytes20', 'uint240'), D), T[bo, by, i])  # fixed case 34410
ck(f(('bool', 'bytes20', 'uint248'), D), T[bo, by, i])  # fixed case 34411
ck(f(('bool', 'bytes20', 'uint256'), D), T[bo, by, i])  # fixed case 34412
ck(f(('bool', 'bytes20', 'uint32'), D), T[bo, by, i])  # fixed case 34413
ck(f(('bool', 'bytes20', 'uint40'), D), T[bo, by, i])  # fixed case 34414
ck(f(('bool', 'bytes20', 'uint48'), D), T[bo, by, i])  # fixed case 34415
ck(f(('bool', 'bytes20', 'uint56'), D), T[bo, by, i])  # fixed case 34416
ck(f(('bool', 'bytes20', 'uint64'), D), T[bo, by, i])  # fixed case 34417
ck(f(('bool', 'bytes20', 'uint72'), D), T[bo, by, i])  # fixed case 34418
ck(f(('bool', 'bytes20', 'uint8'), D), T[bo, by, i])  # fixed case 34419
ck(f(('bool', 'bytes20', 'uint80'), D), T[bo, by, i])  # fixed case 34420
ck(f(('bool', 'bytes20', 'uint88'), D), T[bo, by, i])  # fixed case 34421
ck(f(('bool', 'bytes20', 'uint96'), D), T[bo, by, i])  # fixed case 34422
ck(f(('bool', 'bytes20', '?'), D), T[bo, by, A])  # fixed case 34423
ck(f(('bool', 'bytes21', 'address'), D), T[bo, by, HA])  # fixed case 34424
ck(f(('bool', 'bytes21', 'address[]'), D), T[bo, by, A])  # fixed case 34425
ck(f(('bool', 'bytes21', 'bool'), D), T[bo, by, bo])  # fixed case 34426
ck(f(('bool', 'bytes21', 'bool[]'), D), T[bo, by, A])  # fixed case 34427
ck(f(('bool', 'bytes21', 'bytes'), D), T[bo, by, by])  # fixed case 34428
ck(f(('bool', 'bytes21', 'bytes1'), D), T[bo, by, by])  # fixed case 34429
ck(f(('bool', 'bytes21', 'bytes10'), D), T[bo, by, by])  # fixed case 34430
ck(f(('bool', 'bytes21', 'bytes11'), D), T[bo, by, by])  # fixed case 34431
ck(f(('bool', 'bytes21', 'bytes12'), D), T[bo, by, by])  # fixed case 34432
ck(f(('bool', 'bytes21', 'bytes13'), D), T[bo, by, by])  # fixed case 34433
ck(f(('bool', 'bytes21', 'bytes14'), D), T[bo, by, by])  # fixed case 34434
ck(f(('bool', 'bytes21', 'bytes15'), D), T[bo, by, by])  # fixed case 34435
ck(f(('bool', 'bytes21', 'bytes16'), D), T[bo, by, by])  # fixed case 34436
ck(f(('bool', 'bytes21', 'bytes17'), D), T[bo, by, by])  # fixed case 34437
ck(f(('bool', 'bytes21', 'bytes18'), D), T[bo, by, by])  # fixed case 34438
ck(f(('bool', 'bytes21', 'bytes19'), D), T[bo, by, by])  # fixed case 34439
ck(f(('bool', 'bytes21', 'bytes2'), D), T[bo, by, by])  # fixed case 34440
ck(f(('bool', 'bytes21', 'bytes20'), D), T[bo, by, by])  # fixed case 34441
ck(f(('bool', 'bytes21', 'bytes21'), D), T[bo, by, by])  # fixed case 34442
ck(f(('bool', 'bytes21', 'bytes22'), D), T[bo, by, by])  # fixed case 34443
ck(f(('bool', 'bytes21', 'bytes23'), D), T[bo, by, by])  # fixed case 34444
ck(f(('bool', 'bytes21', 'bytes24'), D), T[bo, by, by])  # fixed case 34445
ck(f(('bool', 'bytes21', 'bytes25'), D), T[bo, by, by])  # fixed case 34446
ck(f(('bool', 'bytes21', 'bytes26'), D), T[bo, by, by])  # fixed case 34447
ck(f(('bool', 'bytes21', 'bytes27'), D), T[bo, by, by])  # fixed case 34448
ck(f(('bool', 'bytes21', 'bytes28'), D), T[bo, by, by])  # fixed case 34449
ck(f(('bool', 'bytes21', 'bytes29'), D), T[bo, by, by])  # fixed case 34450
ck(f(('bool', 'bytes21', 'bytes3'), D), T[bo, by, by])  # fixed case 34451
ck(f(('bool', 'bytes21', 'bytes30'), D), T[bo, by, by])  # fixed case 34452
ck(f(('bool', 'bytes21', 'bytes31'), D), T[bo, by, by])  # fixed case 34453
ck(f(('bool', 'bytes21', 'bytes32'), D), T[bo, by, by])  # fixed case 34454
ck(f(('bool', 'bytes21', 'bytes4'), D), T[bo, by, by])  # fixed case 34455
ck(f(('bool', 'bytes21', 'bytes5'), D), T[bo, by, by])  # fixed case 34456
ck(f(('bool', 'bytes21', 'bytes6'), D), T[bo, by, by])  # fixed case 34457
ck(f(('bool', 'bytes21', 'bytes7'), D), T[bo, by, by])  # fixed case 34458
ck(f(('bool', 'bytes21', 'bytes8'), D), T[bo, by, by])  # fixed case 34459
ck(f(('bool', 'bytes21', 'bytes9'), D), T[bo, by, by])  # fixed case 34460
ck(f(('bool', 'bytes21', 'int104'), D), T[bo, by, i])  # fixed case 34461
ck(f(('bool', 'bytes21', 'int112'), D), T[bo, by, i])  # fixed case 34462
ck(f(('bool', 'bytes21', 'int120'), D), T[bo, by, i])  # fixed case 34463
ck(f(('bool', 'bytes21', 'int128'), D), T[bo, by, i])  # fixed case 34464
ck(f(('bool', 'bytes21', 'int136'), D), T[bo, by, i])  # fixed case 34465
ck(f(('bool', 'bytes21', 'int144'), D), T[bo, by, i])  # fixed case 34466
ck(f(('bool', 'bytes21', 'int152'), D), T[bo, by, i])  # fixed case 34467
ck(f(('bool', 'bytes21', 'int16'), D), T[bo, by, i])  # fixed case 34468
ck(f(('bool', 'bytes21', 'int160'), D), T[bo, by, i])  # fixed case 34469
ck(f(('bool', 'bytes21', 'int168'), D), T[bo, by, i])  # fixed case 34470
ck(f(('bool', 'bytes21', 'int176'), D), T[bo, by, i])  # fixed case 34471
ck(f(('bool', 'bytes21', 'int184'), D), T[bo, by, i])  # fixed case 34472
ck(f(('bool', 'bytes21', 'int192'), D), T[bo, by, i])  # fixed case 34473
ck(f(('bool', 'bytes21', 'int200'), D), T[bo, by, i])  # fixed case 34474
ck(f(('bool', 'bytes21', 'int208'), D), T[bo, by, i])  # fixed case 34475
ck(f(('bool', 'bytes21', 'int216'), D), T[bo, by, i])  # fixed case 34476
ck(f(('bool', 'bytes21', 'int224'), D), T[bo, by, i])  # fixed case 34477
ck(f(('bool', 'bytes21', 'int232'), D), T[bo, by, i])  # fixed case 34478
ck(f(('bool', 'bytes21', 'int24'), D), T[bo, by, i])  # fixed case 34479
ck(f(('bool', 'bytes21', 'int240'), D), T[bo, by, i])  # fixed case 34480
ck(f(('bool', 'bytes21', 'int248'), D), T[bo, by, i])  # fixed case 34481
ck(f(('bool', 'bytes21', 'int256'), D), T[bo, by, i])  # fixed case 34482
ck(f(('bool', 'bytes21', 'int32'), D), T[bo, by, i])  # fixed case 34483
ck(f(('bool', 'bytes21', 'int40'), D), T[bo, by, i])  # fixed case 34484
ck(f(('bool', 'bytes21', 'int48'), D), T[bo, by, i])  # fixed case 34485
ck(f(('bool', 'bytes21', 'int56'), D), T[bo, by, i])  # fixed case 34486
ck(f(('bool', 'bytes21', 'int64'), D), T[bo, by, i])  # fixed case 34487
ck(f(('bool', 'bytes21', 'int72'), D), T[bo, by, i])  # fixed case 34488
ck(f(('bool', 'bytes21', 'int8'), D), T[bo, by, i])  # fixed case 34489
ck(f(('bool', 'bytes21', 'int80'), D), T[bo, by, i])  # fixed case 34490
ck(f(('bool', 'bytes21', 'int88'), D), T[bo, by, i])  # fixed case 34491
ck(f(('bool', 'bytes21', 'int96'), D), T[bo, by, i])  # fixed case 34492
ck(f(('bool', 'bytes21', 'string'), D), T[bo, by, s])  # fixed case 34493
ck(f(('bool', 'bytes21', 'string[]'), D), T[bo, by, A])  # fixed case 34494
ck(f(('bool', 'bytes21', 'uint104'), D), T[bo, by, i])  # fixed case 34495
ck(f(('bool', 'bytes21', 'uint112'), D), T[bo, by, i])  # fixed case 34496
ck(f(('bool', 'bytes21', 'uint120'), D), T[bo, by, i])  # fixed case 34497
ck(f(('bool', 'bytes21', 'uint128'), D), T[bo, by, i])  # fixed case 34498
ck(f(('bool', 'bytes21', 'uint136'), D), T[bo, by, i])  # fixed case 34499
ck(f(('bool', 'bytes21', 'uint144'), D), T[bo, by, i])  # fixed case 34500
ck(f(('bool', 'bytes21', 'uint152'), D), T[bo, by, i])  # fixed case 34501
ck(f(('bool', 'bytes21', 'uint16'), D), T[bo, by, i])  # fixed case 34502
ck(f(('bool', 'bytes21', 'uint160'), D), T[bo, by, i])  # fixed case 34503
ck(f(('bool', 'bytes21', 'uint168'), D), T[bo, by, i])  # fixed case 34504
ck(f(('bool', 'bytes21', 'uint176'), D), T[bo, by, i])  # fixed case 34505
ck(f(('bool', 'bytes21', 'uint184'), D), T[bo, by, i])  # fixed case 34506
ck(f(('bool', 'bytes21', 'uint192'), D), T[bo, by, i])  # fixed case 34507
ck(f(('bool', 'bytes21', 'uint200'), D), T[bo, by, i])  # fixed case 34508
ck(f(('bool', 'bytes21', 'uint208'), D), T[bo, by, i])  # fixed case 34509
ck(f(('bool', 'bytes21', 'uint216'), D), T[bo, by, i])  # fixed case 34510
ck(f(('bool', 'bytes21', 'uint224'), D), T[bo, by, i])  # fixed case 34511
ck(f(('bool', 'bytes21', 'uint232'), D), T[bo, by, i])  # fixed case 34512
ck(f(('bool', 'bytes21', 'uint24'), D), T[bo, by, i])  # fixed case 34513
ck(f(('bool', 'bytes21', 'uint240'), D), T[bo, by, i])  # fixed case 34514
ck(f(('bool', 'bytes21', 'uint248'), D), T[bo, by, i])  # fixed case 34515
ck(f(('bool', 'bytes21', 'uint256'), D), T[bo, by, i])  # fixed case 34516
ck(f(('bool', 'bytes21', 'uint32'), D), T[bo, by, i])  # fixed case 34517
ck(f(('bool', 'bytes21', 'uint40'), D), T[bo, by, i])  # fixed case 34518
ck(f(('bool', 'bytes21', 'uint48'), D), T[bo, by, i])  # fixed case 34519
ck(f(('bool', 'bytes21', 'uint56'), D), T[bo, by, i])  # fixed case 34520
ck(f(('bool', 'bytes21', 'uint64'), D), T[bo, by, i])  # fixed case 34521
ck(f(('bool', 'bytes21', 'uint72'), D), T[bo, by, i])  # fixed case 34522
ck(f(('bool', 'bytes21', 'uint8'), D), T[bo, by, i])  # fixed case 34523
ck(f(('bool', 'bytes21', 'uint80'), D), T[bo, by, i])  # fixed case 34524
ck(f(('bool', 'bytes21', 'uint88'), D), T[bo, by, i])  # fixed case 34525
ck(f(('bool', 'bytes21', 'uint96'), D), T[bo, by, i])  # fixed case 34526
ck(f(('bool', 'bytes21', '?'), D), T[bo, by, A])  # fixed case 34527
ck(f(('bool', 'bytes22', 'address'), D), T[bo, by, HA])  # fixed case 34528
ck(f(('bool', 'bytes22', 'address[]'), D), T[bo, by, A])  # fixed case 34529
ck(f(('bool', 'bytes22', 'bool'), D), T[bo, by, bo])  # fixed case 34530
ck(f(('bool', 'bytes22', 'bool[]'), D), T[bo, by, A])  # fixed case 34531
ck(f(('bool', 'bytes22', 'bytes'), D), T[bo, by, by])  # fixed case 34532
ck(f(('bool', 'bytes22', 'bytes1'), D), T[bo, by, by])  # fixed case 34533
ck(f(('bool', 'bytes22', 'bytes10'), D), T[bo, by, by])  # fixed case 34534
ck(f(('bool', 'bytes22', 'bytes11'), D), T[bo, by, by])  # fixed case 34535
ck(f(('bool', 'bytes22', 'bytes12'), D), T[bo, by, by])  # fixed case 34536
ck(f(('bool', 'bytes22', 'bytes13'), D), T[bo, by, by])  # fixed case 34537
ck(f(('bool', 'bytes22', 'bytes14'), D), T[bo, by, by])  # fixed case 34538
ck(f(('bool', 'bytes22', 'bytes15'), D), T[bo, by, by])  # fixed case 34539
ck(f(('bool', 'bytes22', 'bytes16'), D), T[bo, by, by])  # fixed case 34540
ck(f(('bool', 'bytes22', 'bytes17'), D), T[bo, by, by])  # fixed case 34541
ck(f(('bool', 'bytes22', 'bytes18'), D), T[bo, by, by])  # fixed case 34542
ck(f(('bool', 'bytes22', 'bytes19'), D), T[bo, by, by])  # fixed case 34543
ck(f(('bool', 'bytes22', 'bytes2'), D), T[bo, by, by])  # fixed case 34544
ck(f(('bool', 'bytes22', 'bytes20'), D), T[bo, by, by])  # fixed case 34545
ck(f(('bool', 'bytes22', 'bytes21'), D), T[bo, by, by])  # fixed case 34546
ck(f(('bool', 'bytes22', 'bytes22'), D), T[bo, by, by])  # fixed case 34547
ck(f(('bool', 'bytes22', 'bytes23'), D), T[bo, by, by])  # fixed case 34548
ck(f(('bool', 'bytes22', 'bytes24'), D), T[bo, by, by])  # fixed case 34549
ck(f(('bool', 'bytes22', 'bytes25'), D), T[bo, by, by])  # fixed case 34550
ck(f(('bool', 'bytes22', 'bytes26'), D), T[bo, by, by])  # fixed case 34551
ck(f(('bool', 'bytes22', 'bytes27'), D), T[bo, by, by])  # fixed case 34552
ck(f(('bool', 'bytes22', 'bytes28'), D), T[bo, by, by])  # fixed case 34553
ck(f(('bool', 'bytes22', 'bytes29'), D), T[bo, by, by])  # fixed case 34554
ck(f(('bool', 'bytes22', 'bytes3'), D), T[bo, by, by])  # fixed case 34555
ck(f(('bool', 'bytes22', 'bytes30'), D), T[bo, by, by])  # fixed case 34556
ck(f(('bool', 'bytes22', 'bytes31'), D), T[bo, by, by])  # fixed case 34557
ck(f(('bool', 'bytes22', 'bytes32'), D), T[bo, by, by])  # fixed case 34558
ck(f(('bool', 'bytes22', 'bytes4'), D), T[bo, by, by])  # fixed case 34559
ck(f(('bool', 'bytes22', 'bytes5'), D), T[bo, by, by])  # fixed case 34560
ck(f(('bool', 'bytes22', 'bytes6'), D), T[bo, by, by])  # fixed case 34561
ck(f(('bool', 'bytes22', 'bytes7'), D), T[bo, by, by])  # fixed case 34562
ck(f(('bool', 'bytes22', 'bytes8'), D), T[bo, by, by])  # fixed case 34563
ck(f(('bool', 'bytes22', 'bytes9'), D), T[bo, by, by])  # fixed case 34564
ck(f(('bool', 'bytes22', 'int104'), D), T[bo, by, i])  # fixed case 34565
ck(f(('bool', 'bytes22', 'int112'), D), T[bo, by, i])  # fixed case 34566
ck(f(('bool', 'bytes22', 'int120'), D), T[bo, by, i])  # fixed case 34567
ck(f(('bool', 'bytes22', 'int128'), D), T[bo, by, i])  # fixed case 34568
ck(f(('bool', 'bytes22', 'int136'), D), T[bo, by, i])  # fixed case 34569
ck(f(('bool', 'bytes22', 'int144'), D), T[bo, by, i])  # fixed case 34570
ck(f(('bool', 'bytes22', 'int152'), D), T[bo, by, i])  # fixed case 34571
ck(f(('bool', 'bytes22', 'int16'), D), T[bo, by, i])  # fixed case 34572
ck(f(('bool', 'bytes22', 'int160'), D), T[bo, by, i])  # fixed case 34573
ck(f(('bool', 'bytes22', 'int168'), D), T[bo, by, i])  # fixed case 34574
ck(f(('bool', 'bytes22', 'int176'), D), T[bo, by, i])  # fixed case 34575
ck(f(('bool', 'bytes22', 'int184'), D), T[bo, by, i])  # fixed case 34576
ck(f(('bool', 'bytes22', 'int192'), D), T[bo, by, i])  # fixed case 34577
ck(f(('bool', 'bytes22', 'int200'), D), T[bo, by, i])  # fixed case 34578
ck(f(('bool', 'bytes22', 'int208'), D), T[bo, by, i])  # fixed case 34579
ck(f(('bool', 'bytes22', 'int216'), D), T[bo, by, i])  # fixed case 34580
ck(f(('bool', 'bytes22', 'int224'), D), T[bo, by, i])  # fixed case 34581
ck(f(('bool', 'bytes22', 'int232'), D), T[bo, by, i])  # fixed case 34582
ck(f(('bool', 'bytes22', 'int24'), D), T[bo, by, i])  # fixed case 34583
ck(f(('bool', 'bytes22', 'int240'), D), T[bo, by, i])  # fixed case 34584
ck(f(('bool', 'bytes22', 'int248'), D), T[bo, by, i])  # fixed case 34585
ck(f(('bool', 'bytes22', 'int256'), D), T[bo, by, i])  # fixed case 34586
ck(f(('bool', 'bytes22', 'int32'), D), T[bo, by, i])  # fixed case 34587
ck(f(('bool', 'bytes22', 'int40'), D), T[bo, by, i])  # fixed case 34588
ck(f(('bool', 'bytes22', 'int48'), D), T[bo, by, i])  # fixed case 34589
ck(f(('bool', 'bytes22', 'int56'), D), T[bo, by, i])  # fixed case 34590
ck(f(('bool', 'bytes22', 'int64'), D), T[bo, by, i])  # fixed case 34591
ck(f(('bool', 'bytes22', 'int72'), D), T[bo, by, i])  # fixed case 34592
ck(f(('bool', 'bytes22', 'int8'), D), T[bo, by, i])  # fixed case 34593
ck(f(('bool', 'bytes22', 'int80'), D), T[bo, by, i])  # fixed case 34594
ck(f(('bool', 'bytes22', 'int88'), D), T[bo, by, i])  # fixed case 34595
ck(f(('bool', 'bytes22', 'int96'), D), T[bo, by, i])  # fixed case 34596
ck(f(('bool', 'bytes22', 'string'), D), T[bo, by, s])  # fixed case 34597
ck(f(('bool', 'bytes22', 'string[]'), D), T[bo, by, A])  # fixed case 34598
ck(f(('bool', 'bytes22', 'uint104'), D), T[bo, by, i])  # fixed case 34599
ck(f(('bool', 'bytes22', 'uint112'), D), T[bo, by, i])  # fixed case 34600
ck(f(('bool', 'bytes22', 'uint120'), D), T[bo, by, i])  # fixed case 34601
ck(f(('bool', 'bytes22', 'uint128'), D), T[bo, by, i])  # fixed case 34602
ck(f(('bool', 'bytes22', 'uint136'), D), T[bo, by, i])  # fixed case 34603
ck(f(('bool', 'bytes22', 'uint144'), D), T[bo, by, i])  # fixed case 34604
ck(f(('bool', 'bytes22', 'uint152'), D), T[bo, by, i])  # fixed case 34605
ck(f(('bool', 'bytes22', 'uint16'), D), T[bo, by, i])  # fixed case 34606
ck(f(('bool', 'bytes22', 'uint160'), D), T[bo, by, i])  # fixed case 34607
ck(f(('bool', 'bytes22', 'uint168'), D), T[bo, by, i])  # fixed case 34608
ck(f(('bool', 'bytes22', 'uint176'), D), T[bo, by, i])  # fixed case 34609
ck(f(('bool', 'bytes22', 'uint184'), D), T[bo, by, i])  # fixed case 34610
ck(f(('bool', 'bytes22', 'uint192'), D), T[bo, by, i])  # fixed case 34611
ck(f(('bool', 'bytes22', 'uint200'), D), T[bo, by, i])  # fixed case 34612
ck(f(('bool', 'bytes22', 'uint208'), D), T[bo, by, i])  # fixed case 34613
ck(f(('bool', 'bytes22', 'uint216'), D), T[bo, by, i])  # fixed case 34614
ck(f(('bool', 'bytes22', 'uint224'), D), T[bo, by, i])  # fixed case 34615
ck(f(('bool', 'bytes22', 'uint232'), D), T[bo, by, i])  # fixed case 34616
ck(f(('bool', 'bytes22', 'uint24'), D), T[bo, by, i])  # fixed case 34617
ck(f(('bool', 'bytes22', 'uint240'), D), T[bo, by, i])  # fixed case 34618
ck(f(('bool', 'bytes22', 'uint248'), D), T[bo, by, i])  # fixed case 34619
ck(f(('bool', 'bytes22', 'uint256'), D), T[bo, by, i])  # fixed case 34620
ck(f(('bool', 'bytes22', 'uint32'), D), T[bo, by, i])  # fixed case 34621
ck(f(('bool', 'bytes22', 'uint40'), D), T[bo, by, i])  # fixed case 34622
ck(f(('bool', 'bytes22', 'uint48'), D), T[bo, by, i])  # fixed case 34623
ck(f(('bool', 'bytes22', 'uint56'), D), T[bo, by, i])  # fixed case 34624
ck(f(('bool', 'bytes22', 'uint64'), D), T[bo, by, i])  # fixed case 34625
ck(f(('bool', 'bytes22', 'uint72'), D), T[bo, by, i])  # fixed case 34626
ck(f(('bool', 'bytes22', 'uint8'), D), T[bo, by, i])  # fixed case 34627
ck(f(('bool', 'bytes22', 'uint80'), D), T[bo, by, i])  # fixed case 34628
ck(f(('bool', 'bytes22', 'uint88'), D), T[bo, by, i])  # fixed case 34629
ck(f(('bool', 'bytes22', 'uint96'), D), T[bo, by, i])  # fixed case 34630
ck(f(('bool', 'bytes22', '?'), D), T[bo, by, A])  # fixed case 34631
ck(f(('bool', 'bytes23', 'address'), D), T[bo, by, HA])  # fixed case 34632
ck(f(('bool', 'bytes23', 'address[]'), D), T[bo, by, A])  # fixed case 34633
ck(f(('bool', 'bytes23', 'bool'), D), T[bo, by, bo])  # fixed case 34634
ck(f(('bool', 'bytes23', 'bool[]'), D), T[bo, by, A])  # fixed case 34635
ck(f(('bool', 'bytes23', 'bytes'), D), T[bo, by, by])  # fixed case 34636
ck(f(('bool', 'bytes23', 'bytes1'), D), T[bo, by, by])  # fixed case 34637
ck(f(('bool', 'bytes23', 'bytes10'), D), T[bo, by, by])  # fixed case 34638
ck(f(('bool', 'bytes23', 'bytes11'), D), T[bo, by, by])  # fixed case 34639
ck(f(('bool', 'bytes23', 'bytes12'), D), T[bo, by, by])  # fixed case 34640
ck(f(('bool', 'bytes23', 'bytes13'), D), T[bo, by, by])  # fixed case 34641
ck(f(('bool', 'bytes23', 'bytes14'), D), T[bo, by, by])  # fixed case 34642
ck(f(('bool', 'bytes23', 'bytes15'), D), T[bo, by, by])  # fixed case 34643
ck(f(('bool', 'bytes23', 'bytes16'), D), T[bo, by, by])  # fixed case 34644
ck(f(('bool', 'bytes23', 'bytes17'), D), T[bo, by, by])  # fixed case 34645
ck(f(('bool', 'bytes23', 'bytes18'), D), T[bo, by, by])  # fixed case 34646
ck(f(('bool', 'bytes23', 'bytes19'), D), T[bo, by, by])  # fixed case 34647
ck(f(('bool', 'bytes23', 'bytes2'), D), T[bo, by, by])  # fixed case 34648
ck(f(('bool', 'bytes23', 'bytes20'), D), T[bo, by, by])  # fixed case 34649
ck(f(('bool', 'bytes23', 'bytes21'), D), T[bo, by, by])  # fixed case 34650
ck(f(('bool', 'bytes23', 'bytes22'), D), T[bo, by, by])  # fixed case 34651
ck(f(('bool', 'bytes23', 'bytes23'), D), T[bo, by, by])  # fixed case 34652
ck(f(('bool', 'bytes23', 'bytes24'), D), T[bo, by, by])  # fixed case 34653
ck(f(('bool', 'bytes23', 'bytes25'), D), T[bo, by, by])  # fixed case 34654
ck(f(('bool', 'bytes23', 'bytes26'), D), T[bo, by, by])  # fixed case 34655
ck(f(('bool', 'bytes23', 'bytes27'), D), T[bo, by, by])  # fixed case 34656
ck(f(('bool', 'bytes23', 'bytes28'), D), T[bo, by, by])  # fixed case 34657
ck(f(('bool', 'bytes23', 'bytes29'), D), T[bo, by, by])  # fixed case 34658
ck(f(('bool', 'bytes23', 'bytes3'), D), T[bo, by, by])  # fixed case 34659
ck(f(('bool', 'bytes23', 'bytes30'), D), T[bo, by, by])  # fixed case 34660
ck(f(('bool', 'bytes23', 'bytes31'), D), T[bo, by, by])  # fixed case 34661
ck(f(('bool', 'bytes23', 'bytes32'), D), T[bo, by, by])  # fixed case 34662
ck(f(('bool', 'bytes23', 'bytes4'), D), T[bo, by, by])  # fixed case 34663
ck(f(('bool', 'bytes23', 'bytes5'), D), T[bo, by, by])  # fixed case 34664
ck(f(('bool', 'bytes23', 'bytes6'), D), T[bo, by, by])  # fixed case 34665
ck(f(('bool', 'bytes23', 'bytes7'), D), T[bo, by, by])  # fixed case 34666
ck(f(('bool', 'bytes23', 'bytes8'), D), T[bo, by, by])  # fixed case 34667
ck(f(('bool', 'bytes23', 'bytes9'), D), T[bo, by, by])  # fixed case 34668
ck(f(('bool', 'bytes23', 'int104'), D), T[bo, by, i])  # fixed case 34669
ck(f(('bool', 'bytes23', 'int112'), D), T[bo, by, i])  # fixed case 34670
ck(f(('bool', 'bytes23', 'int120'), D), T[bo, by, i])  # fixed case 34671
ck(f(('bool', 'bytes23', 'int128'), D), T[bo, by, i])  # fixed case 34672
ck(f(('bool', 'bytes23', 'int136'), D), T[bo, by, i])  # fixed case 34673
ck(f(('bool', 'bytes23', 'int144'), D), T[bo, by, i])  # fixed case 34674
ck(f(('bool', 'bytes23', 'int152'), D), T[bo, by, i])  # fixed case 34675
ck(f(('bool', 'bytes23', 'int16'), D), T[bo, by, i])  # fixed case 34676
ck(f(('bool', 'bytes23', 'int160'), D), T[bo, by, i])  # fixed case 34677
ck(f(('bool', 'bytes23', 'int168'), D), T[bo, by, i])  # fixed case 34678
ck(f(('bool', 'bytes23', 'int176'), D), T[bo, by, i])  # fixed case 34679
ck(f(('bool', 'bytes23', 'int184'), D), T[bo, by, i])  # fixed case 34680
ck(f(('bool', 'bytes23', 'int192'), D), T[bo, by, i])  # fixed case 34681
ck(f(('bool', 'bytes23', 'int200'), D), T[bo, by, i])  # fixed case 34682
ck(f(('bool', 'bytes23', 'int208'), D), T[bo, by, i])  # fixed case 34683
ck(f(('bool', 'bytes23', 'int216'), D), T[bo, by, i])  # fixed case 34684
ck(f(('bool', 'bytes23', 'int224'), D), T[bo, by, i])  # fixed case 34685
ck(f(('bool', 'bytes23', 'int232'), D), T[bo, by, i])  # fixed case 34686
ck(f(('bool', 'bytes23', 'int24'), D), T[bo, by, i])  # fixed case 34687
ck(f(('bool', 'bytes23', 'int240'), D), T[bo, by, i])  # fixed case 34688
ck(f(('bool', 'bytes23', 'int248'), D), T[bo, by, i])  # fixed case 34689
ck(f(('bool', 'bytes23', 'int256'), D), T[bo, by, i])  # fixed case 34690
ck(f(('bool', 'bytes23', 'int32'), D), T[bo, by, i])  # fixed case 34691
ck(f(('bool', 'bytes23', 'int40'), D), T[bo, by, i])  # fixed case 34692
ck(f(('bool', 'bytes23', 'int48'), D), T[bo, by, i])  # fixed case 34693
ck(f(('bool', 'bytes23', 'int56'), D), T[bo, by, i])  # fixed case 34694
ck(f(('bool', 'bytes23', 'int64'), D), T[bo, by, i])  # fixed case 34695
ck(f(('bool', 'bytes23', 'int72'), D), T[bo, by, i])  # fixed case 34696
ck(f(('bool', 'bytes23', 'int8'), D), T[bo, by, i])  # fixed case 34697
ck(f(('bool', 'bytes23', 'int80'), D), T[bo, by, i])  # fixed case 34698
ck(f(('bool', 'bytes23', 'int88'), D), T[bo, by, i])  # fixed case 34699
ck(f(('bool', 'bytes23', 'int96'), D), T[bo, by, i])  # fixed case 34700
ck(f(('bool', 'bytes23', 'string'), D), T[bo, by, s])  # fixed case 34701
ck(f(('bool', 'bytes23', 'string[]'), D), T[bo, by, A])  # fixed case 34702
ck(f(('bool', 'bytes23', 'uint104'), D), T[bo, by, i])  # fixed case 34703
ck(f(('bool', 'bytes23', 'uint112'), D), T[bo, by, i])  # fixed case 34704
ck(f(('bool', 'bytes23', 'uint120'), D), T[bo, by, i])  # fixed case 34705
ck(f(('bool', 'bytes23', 'uint128'), D), T[bo, by, i])  # fixed case 34706
ck(f(('bool', 'bytes23', 'uint136'), D), T[bo, by, i])  # fixed case 34707
ck(f(('bool', 'bytes23', 'uint144'), D), T[bo, by, i])  # fixed case 34708
ck(f(('bool', 'bytes23', 'uint152'), D), T[bo, by, i])  # fixed case 34709
ck(f(('bool', 'bytes23', 'uint16'), D), T[bo, by, i])  # fixed case 34710
ck(f(('bool', 'bytes23', 'uint160'), D), T[bo, by, i])  # fixed case 34711
ck(f(('bool', 'bytes23', 'uint168'), D), T[bo, by, i])  # fixed case 34712
ck(f(('bool', 'bytes23', 'uint176'), D), T[bo, by, i])  # fixed case 34713
ck(f(('bool', 'bytes23', 'uint184'), D), T[bo, by, i])  # fixed case 34714
ck(f(('bool', 'bytes23', 'uint192'), D), T[bo, by, i])  # fixed case 34715
ck(f(('bool', 'bytes23', 'uint200'), D), T[bo, by, i])  # fixed case 34716
ck(f(('bool', 'bytes23', 'uint208'), D), T[bo, by, i])  # fixed case 34717
ck(f(('bool', 'bytes23', 'uint216'), D), T[bo, by, i])  # fixed case 34718
ck(f(('bool', 'bytes23', 'uint224'), D), T[bo, by, i])  # fixed case 34719
ck(f(('bool', 'bytes23', 'uint232'), D), T[bo, by, i])  # fixed case 34720
ck(f(('bool', 'bytes23', 'uint24'), D), T[bo, by, i])  # fixed case 34721
ck(f(('bool', 'bytes23', 'uint240'), D), T[bo, by, i])  # fixed case 34722
ck(f(('bool', 'bytes23', 'uint248'), D), T[bo, by, i])  # fixed case 34723
ck(f(('bool', 'bytes23', 'uint256'), D), T[bo, by, i])  # fixed case 34724
ck(f(('bool', 'bytes23', 'uint32'), D), T[bo, by, i])  # fixed case 34725
ck(f(('bool', 'bytes23', 'uint40'), D), T[bo, by, i])  # fixed case 34726
ck(f(('bool', 'bytes23', 'uint48'), D), T[bo, by, i])  # fixed case 34727
ck(f(('bool', 'bytes23', 'uint56'), D), T[bo, by, i])  # fixed case 34728
ck(f(('bool', 'bytes23', 'uint64'), D), T[bo, by, i])  # fixed case 34729
ck(f(('bool', 'bytes23', 'uint72'), D), T[bo, by, i])  # fixed case 34730
ck(f(('bool', 'bytes23', 'uint8'), D), T[bo, by, i])  # fixed case 34731
ck(f(('bool', 'bytes23', 'uint80'), D), T[bo, by, i])  # fixed case 34732
ck(f(('bool', 'bytes23', 'uint88'), D), T[bo, by, i])  # fixed case 34733
ck(f(('bool', 'bytes23', 'uint96'), D), T[bo, by, i])  # fixed case 34734
ck(f(('bool', 'bytes23', '?'), D), T[bo, by, A])  # fixed case 34735
ck(f(('bool', 'bytes24', 'address'), D), T[bo, by, HA])  # fixed case 34736
ck(f(('bool', 'bytes24', 'address[]'), D), T[bo, by, A])  # fixed case 34737
ck(f(('bool', 'bytes24', 'bool'), D), T[bo, by, bo])  # fixed case 34738
ck(f(('bool', 'bytes24', 'bool[]'), D), T[bo, by, A])  # fixed case 34739
ck(f(('bool', 'bytes24', 'bytes'), D), T[bo, by, by])  # fixed case 34740
ck(f(('bool', 'bytes24', 'bytes1'), D), T[bo, by, by])  # fixed case 34741
ck(f(('bool', 'bytes24', 'bytes10'), D), T[bo, by, by])  # fixed case 34742
ck(f(('bool', 'bytes24', 'bytes11'), D), T[bo, by, by])  # fixed case 34743
ck(f(('bool', 'bytes24', 'bytes12'), D), T[bo, by, by])  # fixed case 34744
ck(f(('bool', 'bytes24', 'bytes13'), D), T[bo, by, by])  # fixed case 34745
ck(f(('bool', 'bytes24', 'bytes14'), D), T[bo, by, by])  # fixed case 34746
ck(f(('bool', 'bytes24', 'bytes15'), D), T[bo, by, by])  # fixed case 34747
ck(f(('bool', 'bytes24', 'bytes16'), D), T[bo, by, by])  # fixed case 34748
ck(f(('bool', 'bytes24', 'bytes17'), D), T[bo, by, by])  # fixed case 34749
ck(f(('bool', 'bytes24', 'bytes18'), D), T[bo, by, by])  # fixed case 34750
ck(f(('bool', 'bytes24', 'bytes19'), D), T[bo, by, by])  # fixed case 34751
ck(f(('bool', 'bytes24', 'bytes2'), D), T[bo, by, by])  # fixed case 34752
ck(f(('bool', 'bytes24', 'bytes20'), D), T[bo, by, by])  # fixed case 34753
ck(f(('bool', 'bytes24', 'bytes21'), D), T[bo, by, by])  # fixed case 34754
ck(f(('bool', 'bytes24', 'bytes22'), D), T[bo, by, by])  # fixed case 34755
ck(f(('bool', 'bytes24', 'bytes23'), D), T[bo, by, by])  # fixed case 34756
ck(f(('bool', 'bytes24', 'bytes24'), D), T[bo, by, by])  # fixed case 34757
ck(f(('bool', 'bytes24', 'bytes25'), D), T[bo, by, by])  # fixed case 34758
ck(f(('bool', 'bytes24', 'bytes26'), D), T[bo, by, by])  # fixed case 34759
ck(f(('bool', 'bytes24', 'bytes27'), D), T[bo, by, by])  # fixed case 34760
ck(f(('bool', 'bytes24', 'bytes28'), D), T[bo, by, by])  # fixed case 34761
ck(f(('bool', 'bytes24', 'bytes29'), D), T[bo, by, by])  # fixed case 34762
ck(f(('bool', 'bytes24', 'bytes3'), D), T[bo, by, by])  # fixed case 34763
ck(f(('bool', 'bytes24', 'bytes30'), D), T[bo, by, by])  # fixed case 34764
ck(f(('bool', 'bytes24', 'bytes31'), D), T[bo, by, by])  # fixed case 34765
ck(f(('bool', 'bytes24', 'bytes32'), D), T[bo, by, by])  # fixed case 34766
ck(f(('bool', 'bytes24', 'bytes4'), D), T[bo, by, by])  # fixed case 34767
ck(f(('bool', 'bytes24', 'bytes5'), D), T[bo, by, by])  # fixed case 34768
ck(f(('bool', 'bytes24', 'bytes6'), D), T[bo, by, by])  # fixed case 34769
ck(f(('bool', 'bytes24', 'bytes7'), D), T[bo, by, by])  # fixed case 34770
ck(f(('bool', 'bytes24', 'bytes8'), D), T[bo, by, by])  # fixed case 34771
ck(f(('bool', 'bytes24', 'bytes9'), D), T[bo, by, by])  # fixed case 34772
ck(f(('bool', 'bytes24', 'int104'), D), T[bo, by, i])  # fixed case 34773
ck(f(('bool', 'bytes24', 'int112'), D), T[bo, by, i])  # fixed case 34774
ck(f(('bool', 'bytes24', 'int120'), D), T[bo, by, i])  # fixed case 34775
ck(f(('bool', 'bytes24', 'int128'), D), T[bo, by, i])  # fixed case 34776
ck(f(('bool', 'bytes24', 'int136'), D), T[bo, by, i])  # fixed case 34777
ck(f(('bool', 'bytes24', 'int144'), D), T[bo, by, i])  # fixed case 34778
ck(f(('bool', 'bytes24', 'int152'), D), T[bo, by, i])  # fixed case 34779
ck(f(('bool', 'bytes24', 'int16'), D), T[bo, by, i])  # fixed case 34780
ck(f(('bool', 'bytes24', 'int160'), D), T[bo, by, i])  # fixed case 34781
ck(f(('bool', 'bytes24', 'int168'), D), T[bo, by, i])  # fixed case 34782
ck(f(('bool', 'bytes24', 'int176'), D), T[bo, by, i])  # fixed case 34783
ck(f(('bool', 'bytes24', 'int184'), D), T[bo, by, i])  # fixed case 34784
ck(f(('bool', 'bytes24', 'int192'), D), T[bo, by, i])  # fixed case 34785
ck(f(('bool', 'bytes24', 'int200'), D), T[bo, by, i])  # fixed case 34786
ck(f(('bool', 'bytes24', 'int208'), D), T[bo, by, i])  # fixed case 34787
ck(f(('bool', 'bytes24', 'int216'), D), T[bo, by, i])  # fixed case 34788
ck(f(('bool', 'bytes24', 'int224'), D), T[bo, by, i])  # fixed case 34789
ck(f(('bool', 'bytes24', 'int232'), D), T[bo, by, i])  # fixed case 34790
ck(f(('bool', 'bytes24', 'int24'), D), T[bo, by, i])  # fixed case 34791
ck(f(('bool', 'bytes24', 'int240'), D), T[bo, by, i])  # fixed case 34792
ck(f(('bool', 'bytes24', 'int248'), D), T[bo, by, i])  # fixed case 34793
ck(f(('bool', 'bytes24', 'int256'), D), T[bo, by, i])  # fixed case 34794
ck(f(('bool', 'bytes24', 'int32'), D), T[bo, by, i])  # fixed case 34795
ck(f(('bool', 'bytes24', 'int40'), D), T[bo, by, i])  # fixed case 34796
ck(f(('bool', 'bytes24', 'int48'), D), T[bo, by, i])  # fixed case 34797
ck(f(('bool', 'bytes24', 'int56'), D), T[bo, by, i])  # fixed case 34798
ck(f(('bool', 'bytes24', 'int64'), D), T[bo, by, i])  # fixed case 34799
ck(f(('bool', 'bytes24', 'int72'), D), T[bo, by, i])  # fixed case 34800
ck(f(('bool', 'bytes24', 'int8'), D), T[bo, by, i])  # fixed case 34801
ck(f(('bool', 'bytes24', 'int80'), D), T[bo, by, i])  # fixed case 34802
ck(f(('bool', 'bytes24', 'int88'), D), T[bo, by, i])  # fixed case 34803
ck(f(('bool', 'bytes24', 'int96'), D), T[bo, by, i])  # fixed case 34804
ck(f(('bool', 'bytes24', 'string'), D), T[bo, by, s])  # fixed case 34805
ck(f(('bool', 'bytes24', 'string[]'), D), T[bo, by, A])  # fixed case 34806
ck(f(('bool', 'bytes24', 'uint104'), D), T[bo, by, i])  # fixed case 34807
ck(f(('bool', 'bytes24', 'uint112'), D), T[bo, by, i])  # fixed case 34808
ck(f(('bool', 'bytes24', 'uint120'), D), T[bo, by, i])  # fixed case 34809
ck(f(('bool', 'bytes24', 'uint128'), D), T[bo, by, i])  # fixed case 34810
ck(f(('bool', 'bytes24', 'uint136'), D), T[bo, by, i])  # fixed case 34811
ck(f(('bool', 'bytes24', 'uint144'), D), T[bo, by, i])  # fixed case 34812
ck(f(('bool', 'bytes24', 'uint152'), D), T[bo, by, i])  # fixed case 34813
ck(f(('bool', 'bytes24', 'uint16'), D), T[bo, by, i])  # fixed case 34814
ck(f(('bool', 'bytes24', 'uint160'), D), T[bo, by, i])  # fixed case 34815
ck(f(('bool', 'bytes24', 'uint168'), D), T[bo, by, i])  # fixed case 34816
ck(f(('bool', 'bytes24', 'uint176'), D), T[bo, by, i])  # fixed case 34817
ck(f(('bool', 'bytes24', 'uint184'), D), T[bo, by, i])  # fixed case 34818
ck(f(('bool', 'bytes24', 'uint192'), D), T[bo, by, i])  # fixed case 34819
ck(f(('bool', 'bytes24', 'uint200'), D), T[bo, by, i])  # fixed case 34820
ck(f(('bool', 'bytes24', 'uint208'), D), T[bo, by, i])  # fixed case 34821
ck(f(('bool', 'bytes24', 'uint216'), D), T[bo, by, i])  # fixed case 34822
ck(f(('bool', 'bytes24', 'uint224'), D), T[bo, by, i])  # fixed case 34823
ck(f(('bool', 'bytes24', 'uint232'), D), T[bo, by, i])  # fixed case 34824
ck(f(('bool', 'bytes24', 'uint24'), D), T[bo, by, i])  # fixed case 34825
ck(f(('bool', 'bytes24', 'uint240'), D), T[bo, by, i])  # fixed case 34826
ck(f(('bool', 'bytes24', 'uint248'), D), T[bo, by, i])  # fixed case 34827
ck(f(('bool', 'bytes24', 'uint256'), D), T[bo, by, i])  # fixed case 34828
ck(f(('bool', 'bytes24', 'uint32'), D), T[bo, by, i])  # fixed case 34829
ck(f(('bool', 'bytes24', 'uint40'), D), T[bo, by, i])  # fixed case 34830
ck(f(('bool', 'bytes24', 'uint48'), D), T[bo, by, i])  # fixed case 34831
ck(f(('bool', 'bytes24', 'uint56'), D), T[bo, by, i])  # fixed case 34832
ck(f(('bool', 'bytes24', 'uint64'), D), T[bo, by, i])  # fixed case 34833
ck(f(('bool', 'bytes24', 'uint72'), D), T[bo, by, i])  # fixed case 34834
ck(f(('bool', 'bytes24', 'uint8'), D), T[bo, by, i])  # fixed case 34835
ck(f(('bool', 'bytes24', 'uint80'), D), T[bo, by, i])  # fixed case 34836
ck(f(('bool', 'bytes24', 'uint88'), D), T[bo, by, i])  # fixed case 34837
ck(f(('bool', 'bytes24', 'uint96'), D), T[bo, by, i])  # fixed case 34838
ck(f(('bool', 'bytes24', '?'), D), T[bo, by, A])  # fixed case 34839
ck(f(('bool', 'bytes25', 'address'), D), T[bo, by, HA])  # fixed case 34840
ck(f(('bool', 'bytes25', 'address[]'), D), T[bo, by, A])  # fixed case 34841
ck(f(('bool', 'bytes25', 'bool'), D), T[bo, by, bo])  # fixed case 34842
ck(f(('bool', 'bytes25', 'bool[]'), D), T[bo, by, A])  # fixed case 34843
ck(f(('bool', 'bytes25', 'bytes'), D), T[bo, by, by])  # fixed case 34844
ck(f(('bool', 'bytes25', 'bytes1'), D), T[bo, by, by])  # fixed case 34845
ck(f(('bool', 'bytes25', 'bytes10'), D), T[bo, by, by])  # fixed case 34846
ck(f(('bool', 'bytes25', 'bytes11'), D), T[bo, by, by])  # fixed case 34847
ck(f(('bool', 'bytes25', 'bytes12'), D), T[bo, by, by])  # fixed case 34848
ck(f(('bool', 'bytes25', 'bytes13'), D), T[bo, by, by])  # fixed case 34849
ck(f(('bool', 'bytes25', 'bytes14'), D), T[bo, by, by])  # fixed case 34850
ck(f(('bool', 'bytes25', 'bytes15'), D), T[bo, by, by])  # fixed case 34851
ck(f(('bool', 'bytes25', 'bytes16'), D), T[bo, by, by])  # fixed case 34852
ck(f(('bool', 'bytes25', 'bytes17'), D), T[bo, by, by])  # fixed case 34853
ck(f(('bool', 'bytes25', 'bytes18'), D), T[bo, by, by])  # fixed case 34854
ck(f(('bool', 'bytes25', 'bytes19'), D), T[bo, by, by])  # fixed case 34855
ck(f(('bool', 'bytes25', 'bytes2'), D), T[bo, by, by])  # fixed case 34856
ck(f(('bool', 'bytes25', 'bytes20'), D), T[bo, by, by])  # fixed case 34857
ck(f(('bool', 'bytes25', 'bytes21'), D), T[bo, by, by])  # fixed case 34858
ck(f(('bool', 'bytes25', 'bytes22'), D), T[bo, by, by])  # fixed case 34859
ck(f(('bool', 'bytes25', 'bytes23'), D), T[bo, by, by])  # fixed case 34860
ck(f(('bool', 'bytes25', 'bytes24'), D), T[bo, by, by])  # fixed case 34861
ck(f(('bool', 'bytes25', 'bytes25'), D), T[bo, by, by])  # fixed case 34862
ck(f(('bool', 'bytes25', 'bytes26'), D), T[bo, by, by])  # fixed case 34863
ck(f(('bool', 'bytes25', 'bytes27'), D), T[bo, by, by])  # fixed case 34864
ck(f(('bool', 'bytes25', 'bytes28'), D), T[bo, by, by])  # fixed case 34865
ck(f(('bool', 'bytes25', 'bytes29'), D), T[bo, by, by])  # fixed case 34866
ck(f(('bool', 'bytes25', 'bytes3'), D), T[bo, by, by])  # fixed case 34867
ck(f(('bool', 'bytes25', 'bytes30'), D), T[bo, by, by])  # fixed case 34868
ck(f(('bool', 'bytes25', 'bytes31'), D), T[bo, by, by])  # fixed case 34869
ck(f(('bool', 'bytes25', 'bytes32'), D), T[bo, by, by])  # fixed case 34870
ck(f(('bool', 'bytes25', 'bytes4'), D), T[bo, by, by])  # fixed case 34871
ck(f(('bool', 'bytes25', 'bytes5'), D), T[bo, by, by])  # fixed case 34872
ck(f(('bool', 'bytes25', 'bytes6'), D), T[bo, by, by])  # fixed case 34873
ck(f(('bool', 'bytes25', 'bytes7'), D), T[bo, by, by])  # fixed case 34874
ck(f(('bool', 'bytes25', 'bytes8'), D), T[bo, by, by])  # fixed case 34875
ck(f(('bool', 'bytes25', 'bytes9'), D), T[bo, by, by])  # fixed case 34876
ck(f(('bool', 'bytes25', 'int104'), D), T[bo, by, i])  # fixed case 34877
ck(f(('bool', 'bytes25', 'int112'), D), T[bo, by, i])  # fixed case 34878
ck(f(('bool', 'bytes25', 'int120'), D), T[bo, by, i])  # fixed case 34879
ck(f(('bool', 'bytes25', 'int128'), D), T[bo, by, i])  # fixed case 34880
ck(f(('bool', 'bytes25', 'int136'), D), T[bo, by, i])  # fixed case 34881
ck(f(('bool', 'bytes25', 'int144'), D), T[bo, by, i])  # fixed case 34882
ck(f(('bool', 'bytes25', 'int152'), D), T[bo, by, i])  # fixed case 34883
ck(f(('bool', 'bytes25', 'int16'), D), T[bo, by, i])  # fixed case 34884
ck(f(('bool', 'bytes25', 'int160'), D), T[bo, by, i])  # fixed case 34885
ck(f(('bool', 'bytes25', 'int168'), D), T[bo, by, i])  # fixed case 34886
ck(f(('bool', 'bytes25', 'int176'), D), T[bo, by, i])  # fixed case 34887
ck(f(('bool', 'bytes25', 'int184'), D), T[bo, by, i])  # fixed case 34888
ck(f(('bool', 'bytes25', 'int192'), D), T[bo, by, i])  # fixed case 34889
ck(f(('bool', 'bytes25', 'int200'), D), T[bo, by, i])  # fixed case 34890
ck(f(('bool', 'bytes25', 'int208'), D), T[bo, by, i])  # fixed case 34891
ck(f(('bool', 'bytes25', 'int216'), D), T[bo, by, i])  # fixed case 34892
ck(f(('bool', 'bytes25', 'int224'), D), T[bo, by, i])  # fixed case 34893
ck(f(('bool', 'bytes25', 'int232'), D), T[bo, by, i])  # fixed case 34894
ck(f(('bool', 'bytes25', 'int24'), D), T[bo, by, i])  # fixed case 34895
ck(f(('bool', 'bytes25', 'int240'), D), T[bo, by, i])  # fixed case 34896
ck(f(('bool', 'bytes25', 'int248'), D), T[bo, by, i])  # fixed case 34897
ck(f(('bool', 'bytes25', 'int256'), D), T[bo, by, i])  # fixed case 34898
ck(f(('bool', 'bytes25', 'int32'), D), T[bo, by, i])  # fixed case 34899
ck(f(('bool', 'bytes25', 'int40'), D), T[bo, by, i])  # fixed case 34900
ck(f(('bool', 'bytes25', 'int48'), D), T[bo, by, i])  # fixed case 34901
ck(f(('bool', 'bytes25', 'int56'), D), T[bo, by, i])  # fixed case 34902
ck(f(('bool', 'bytes25', 'int64'), D), T[bo, by, i])  # fixed case 34903
ck(f(('bool', 'bytes25', 'int72'), D), T[bo, by, i])  # fixed case 34904
ck(f(('bool', 'bytes25', 'int8'), D), T[bo, by, i])  # fixed case 34905
ck(f(('bool', 'bytes25', 'int80'), D), T[bo, by, i])  # fixed case 34906
ck(f(('bool', 'bytes25', 'int88'), D), T[bo, by, i])  # fixed case 34907
ck(f(('bool', 'bytes25', 'int96'), D), T[bo, by, i])  # fixed case 34908
ck(f(('bool', 'bytes25', 'string'), D), T[bo, by, s])  # fixed case 34909
ck(f(('bool', 'bytes25', 'string[]'), D), T[bo, by, A])  # fixed case 34910
ck(f(('bool', 'bytes25', 'uint104'), D), T[bo, by, i])  # fixed case 34911
ck(f(('bool', 'bytes25', 'uint112'), D), T[bo, by, i])  # fixed case 34912
ck(f(('bool', 'bytes25', 'uint120'), D), T[bo, by, i])  # fixed case 34913
ck(f(('bool', 'bytes25', 'uint128'), D), T[bo, by, i])  # fixed case 34914
ck(f(('bool', 'bytes25', 'uint136'), D), T[bo, by, i])  # fixed case 34915
ck(f(('bool', 'bytes25', 'uint144'), D), T[bo, by, i])  # fixed case 34916
ck(f(('bool', 'bytes25', 'uint152'), D), T[bo, by, i])  # fixed case 34917
ck(f(('bool', 'bytes25', 'uint16'), D), T[bo, by, i])  # fixed case 34918
ck(f(('bool', 'bytes25', 'uint160'), D), T[bo, by, i])  # fixed case 34919
ck(f(('bool', 'bytes25', 'uint168'), D), T[bo, by, i])  # fixed case 34920
ck(f(('bool', 'bytes25', 'uint176'), D), T[bo, by, i])  # fixed case 34921
ck(f(('bool', 'bytes25', 'uint184'), D), T[bo, by, i])  # fixed case 34922
ck(f(('bool', 'bytes25', 'uint192'), D), T[bo, by, i])  # fixed case 34923
ck(f(('bool', 'bytes25', 'uint200'), D), T[bo, by, i])  # fixed case 34924
ck(f(('bool', 'bytes25', 'uint208'), D), T[bo, by, i])  # fixed case 34925
ck(f(('bool', 'bytes25', 'uint216'), D), T[bo, by, i])  # fixed case 34926
ck(f(('bool', 'bytes25', 'uint224'), D), T[bo, by, i])  # fixed case 34927
ck(f(('bool', 'bytes25', 'uint232'), D), T[bo, by, i])  # fixed case 34928
ck(f(('bool', 'bytes25', 'uint24'), D), T[bo, by, i])  # fixed case 34929
ck(f(('bool', 'bytes25', 'uint240'), D), T[bo, by, i])  # fixed case 34930
ck(f(('bool', 'bytes25', 'uint248'), D), T[bo, by, i])  # fixed case 34931
ck(f(('bool', 'bytes25', 'uint256'), D), T[bo, by, i])  # fixed case 34932
ck(f(('bool', 'bytes25', 'uint32'), D), T[bo, by, i])  # fixed case 34933
ck(f(('bool', 'bytes25', 'uint40'), D), T[bo, by, i])  # fixed case 34934
ck(f(('bool', 'bytes25', 'uint48'), D), T[bo, by, i])  # fixed case 34935
ck(f(('bool', 'bytes25', 'uint56'), D), T[bo, by, i])  # fixed case 34936
ck(f(('bool', 'bytes25', 'uint64'), D), T[bo, by, i])  # fixed case 34937
ck(f(('bool', 'bytes25', 'uint72'), D), T[bo, by, i])  # fixed case 34938
ck(f(('bool', 'bytes25', 'uint8'), D), T[bo, by, i])  # fixed case 34939
ck(f(('bool', 'bytes25', 'uint80'), D), T[bo, by, i])  # fixed case 34940
ck(f(('bool', 'bytes25', 'uint88'), D), T[bo, by, i])  # fixed case 34941
ck(f(('bool', 'bytes25', 'uint96'), D), T[bo, by, i])  # fixed case 34942
ck(f(('bool', 'bytes25', '?'), D), T[bo, by, A])  # fixed case 34943
ck(f(('bool', 'bytes26', 'address'), D), T[bo, by, HA])  # fixed case 34944
ck(f(('bool', 'bytes26', 'address[]'), D), T[bo, by, A])  # fixed case 34945
ck(f(('bool', 'bytes26', 'bool'), D), T[bo, by, bo])  # fixed case 34946
ck(f(('bool', 'bytes26', 'bool[]'), D), T[bo, by, A])  # fixed case 34947
ck(f(('bool', 'bytes26', 'bytes'), D), T[bo, by, by])  # fixed case 34948
ck(f(('bool', 'bytes26', 'bytes1'), D), T[bo, by, by])  # fixed case 34949
ck(f(('bool', 'bytes26', 'bytes10'), D), T[bo, by, by])  # fixed case 34950
ck(f(('bool', 'bytes26', 'bytes11'), D), T[bo, by, by])  # fixed case 34951
ck(f(('bool', 'bytes26', 'bytes12'), D), T[bo, by, by])  # fixed case 34952
ck(f(('bool', 'bytes26', 'bytes13'), D), T[bo, by, by])  # fixed case 34953
ck(f(('bool', 'bytes26', 'bytes14'), D), T[bo, by, by])  # fixed case 34954
ck(f(('bool', 'bytes26', 'bytes15'), D), T[bo, by, by])  # fixed case 34955
ck(f(('bool', 'bytes26', 'bytes16'), D), T[bo, by, by])  # fixed case 34956
ck(f(('bool', 'bytes26', 'bytes17'), D), T[bo, by, by])  # fixed case 34957
ck(f(('bool', 'bytes26', 'bytes18'), D), T[bo, by, by])  # fixed case 34958
ck(f(('bool', 'bytes26', 'bytes19'), D), T[bo, by, by])  # fixed case 34959
ck(f(('bool', 'bytes26', 'bytes2'), D), T[bo, by, by])  # fixed case 34960
ck(f(('bool', 'bytes26', 'bytes20'), D), T[bo, by, by])  # fixed case 34961
ck(f(('bool', 'bytes26', 'bytes21'), D), T[bo, by, by])  # fixed case 34962
ck(f(('bool', 'bytes26', 'bytes22'), D), T[bo, by, by])  # fixed case 34963
ck(f(('bool', 'bytes26', 'bytes23'), D), T[bo, by, by])  # fixed case 34964
ck(f(('bool', 'bytes26', 'bytes24'), D), T[bo, by, by])  # fixed case 34965
ck(f(('bool', 'bytes26', 'bytes25'), D), T[bo, by, by])  # fixed case 34966
ck(f(('bool', 'bytes26', 'bytes26'), D), T[bo, by, by])  # fixed case 34967
ck(f(('bool', 'bytes26', 'bytes27'), D), T[bo, by, by])  # fixed case 34968
ck(f(('bool', 'bytes26', 'bytes28'), D), T[bo, by, by])  # fixed case 34969
ck(f(('bool', 'bytes26', 'bytes29'), D), T[bo, by, by])  # fixed case 34970
ck(f(('bool', 'bytes26', 'bytes3'), D), T[bo, by, by])  # fixed case 34971
ck(f(('bool', 'bytes26', 'bytes30'), D), T[bo, by, by])  # fixed case 34972
ck(f(('bool', 'bytes26', 'bytes31'), D), T[bo, by, by])  # fixed case 34973
ck(f(('bool', 'bytes26', 'bytes32'), D), T[bo, by, by])  # fixed case 34974
ck(f(('bool', 'bytes26', 'bytes4'), D), T[bo, by, by])  # fixed case 34975
ck(f(('bool', 'bytes26', 'bytes5'), D), T[bo, by, by])  # fixed case 34976
ck(f(('bool', 'bytes26', 'bytes6'), D), T[bo, by, by])  # fixed case 34977
ck(f(('bool', 'bytes26', 'bytes7'), D), T[bo, by, by])  # fixed case 34978
ck(f(('bool', 'bytes26', 'bytes8'), D), T[bo, by, by])  # fixed case 34979
ck(f(('bool', 'bytes26', 'bytes9'), D), T[bo, by, by])  # fixed case 34980
ck(f(('bool', 'bytes26', 'int104'), D), T[bo, by, i])  # fixed case 34981
ck(f(('bool', 'bytes26', 'int112'), D), T[bo, by, i])  # fixed case 34982
ck(f(('bool', 'bytes26', 'int120'), D), T[bo, by, i])  # fixed case 34983
ck(f(('bool', 'bytes26', 'int128'), D), T[bo, by, i])  # fixed case 34984
ck(f(('bool', 'bytes26', 'int136'), D), T[bo, by, i])  # fixed case 34985
ck(f(('bool', 'bytes26', 'int144'), D), T[bo, by, i])  # fixed case 34986
ck(f(('bool', 'bytes26', 'int152'), D), T[bo, by, i])  # fixed case 34987
ck(f(('bool', 'bytes26', 'int16'), D), T[bo, by, i])  # fixed case 34988
ck(f(('bool', 'bytes26', 'int160'), D), T[bo, by, i])  # fixed case 34989
ck(f(('bool', 'bytes26', 'int168'), D), T[bo, by, i])  # fixed case 34990
ck(f(('bool', 'bytes26', 'int176'), D), T[bo, by, i])  # fixed case 34991
ck(f(('bool', 'bytes26', 'int184'), D), T[bo, by, i])  # fixed case 34992
ck(f(('bool', 'bytes26', 'int192'), D), T[bo, by, i])  # fixed case 34993
ck(f(('bool', 'bytes26', 'int200'), D), T[bo, by, i])  # fixed case 34994
ck(f(('bool', 'bytes26', 'int208'), D), T[bo, by, i])  # fixed case 34995
ck(f(('bool', 'bytes26', 'int216'), D), T[bo, by, i])  # fixed case 34996
ck(f(('bool', 'bytes26', 'int224'), D), T[bo, by, i])  # fixed case 34997
ck(f(('bool', 'bytes26', 'int232'), D), T[bo, by, i])  # fixed case 34998
ck(f(('bool', 'bytes26', 'int24'), D), T[bo, by, i])  # fixed case 34999
ck(f(('bool', 'bytes26', 'int240'), D), T[bo, by, i])  # fixed case 35000
ck(f(('bool', 'bytes26', 'int248'), D), T[bo, by, i])  # fixed case 35001
ck(f(('bool', 'bytes26', 'int256'), D), T[bo, by, i])  # fixed case 35002
ck(f(('bool', 'bytes26', 'int32'), D), T[bo, by, i])  # fixed case 35003
ck(f(('bool', 'bytes26', 'int40'), D), T[bo, by, i])  # fixed case 35004
ck(f(('bool', 'bytes26', 'int48'), D), T[bo, by, i])  # fixed case 35005
ck(f(('bool', 'bytes26', 'int56'), D), T[bo, by, i])  # fixed case 35006
ck(f(('bool', 'bytes26', 'int64'), D), T[bo, by, i])  # fixed case 35007
ck(f(('bool', 'bytes26', 'int72'), D), T[bo, by, i])  # fixed case 35008
ck(f(('bool', 'bytes26', 'int8'), D), T[bo, by, i])  # fixed case 35009
ck(f(('bool', 'bytes26', 'int80'), D), T[bo, by, i])  # fixed case 35010
ck(f(('bool', 'bytes26', 'int88'), D), T[bo, by, i])  # fixed case 35011
ck(f(('bool', 'bytes26', 'int96'), D), T[bo, by, i])  # fixed case 35012
ck(f(('bool', 'bytes26', 'string'), D), T[bo, by, s])  # fixed case 35013
ck(f(('bool', 'bytes26', 'string[]'), D), T[bo, by, A])  # fixed case 35014
ck(f(('bool', 'bytes26', 'uint104'), D), T[bo, by, i])  # fixed case 35015
ck(f(('bool', 'bytes26', 'uint112'), D), T[bo, by, i])  # fixed case 35016
ck(f(('bool', 'bytes26', 'uint120'), D), T[bo, by, i])  # fixed case 35017
ck(f(('bool', 'bytes26', 'uint128'), D), T[bo, by, i])  # fixed case 35018
ck(f(('bool', 'bytes26', 'uint136'), D), T[bo, by, i])  # fixed case 35019
ck(f(('bool', 'bytes26', 'uint144'), D), T[bo, by, i])  # fixed case 35020
ck(f(('bool', 'bytes26', 'uint152'), D), T[bo, by, i])  # fixed case 35021
ck(f(('bool', 'bytes26', 'uint16'), D), T[bo, by, i])  # fixed case 35022
ck(f(('bool', 'bytes26', 'uint160'), D), T[bo, by, i])  # fixed case 35023
ck(f(('bool', 'bytes26', 'uint168'), D), T[bo, by, i])  # fixed case 35024
ck(f(('bool', 'bytes26', 'uint176'), D), T[bo, by, i])  # fixed case 35025
ck(f(('bool', 'bytes26', 'uint184'), D), T[bo, by, i])  # fixed case 35026
ck(f(('bool', 'bytes26', 'uint192'), D), T[bo, by, i])  # fixed case 35027
ck(f(('bool', 'bytes26', 'uint200'), D), T[bo, by, i])  # fixed case 35028
ck(f(('bool', 'bytes26', 'uint208'), D), T[bo, by, i])  # fixed case 35029
ck(f(('bool', 'bytes26', 'uint216'), D), T[bo, by, i])  # fixed case 35030
ck(f(('bool', 'bytes26', 'uint224'), D), T[bo, by, i])  # fixed case 35031
ck(f(('bool', 'bytes26', 'uint232'), D), T[bo, by, i])  # fixed case 35032
ck(f(('bool', 'bytes26', 'uint24'), D), T[bo, by, i])  # fixed case 35033
ck(f(('bool', 'bytes26', 'uint240'), D), T[bo, by, i])  # fixed case 35034
ck(f(('bool', 'bytes26', 'uint248'), D), T[bo, by, i])  # fixed case 35035
ck(f(('bool', 'bytes26', 'uint256'), D), T[bo, by, i])  # fixed case 35036
ck(f(('bool', 'bytes26', 'uint32'), D), T[bo, by, i])  # fixed case 35037
ck(f(('bool', 'bytes26', 'uint40'), D), T[bo, by, i])  # fixed case 35038
ck(f(('bool', 'bytes26', 'uint48'), D), T[bo, by, i])  # fixed case 35039
ck(f(('bool', 'bytes26', 'uint56'), D), T[bo, by, i])  # fixed case 35040
ck(f(('bool', 'bytes26', 'uint64'), D), T[bo, by, i])  # fixed case 35041
ck(f(('bool', 'bytes26', 'uint72'), D), T[bo, by, i])  # fixed case 35042
ck(f(('bool', 'bytes26', 'uint8'), D), T[bo, by, i])  # fixed case 35043
ck(f(('bool', 'bytes26', 'uint80'), D), T[bo, by, i])  # fixed case 35044
ck(f(('bool', 'bytes26', 'uint88'), D), T[bo, by, i])  # fixed case 35045
ck(f(('bool', 'bytes26', 'uint96'), D), T[bo, by, i])  # fixed case 35046
ck(f(('bool', 'bytes26', '?'), D), T[bo, by, A])  # fixed case 35047
ck(f(('bool', 'bytes27', 'address'), D), T[bo, by, HA])  # fixed case 35048
ck(f(('bool', 'bytes27', 'address[]'), D), T[bo, by, A])  # fixed case 35049
ck(f(('bool', 'bytes27', 'bool'), D), T[bo, by, bo])  # fixed case 35050
ck(f(('bool', 'bytes27', 'bool[]'), D), T[bo, by, A])  # fixed case 35051
ck(f(('bool', 'bytes27', 'bytes'), D), T[bo, by, by])  # fixed case 35052
ck(f(('bool', 'bytes27', 'bytes1'), D), T[bo, by, by])  # fixed case 35053
ck(f(('bool', 'bytes27', 'bytes10'), D), T[bo, by, by])  # fixed case 35054
ck(f(('bool', 'bytes27', 'bytes11'), D), T[bo, by, by])  # fixed case 35055
ck(f(('bool', 'bytes27', 'bytes12'), D), T[bo, by, by])  # fixed case 35056
ck(f(('bool', 'bytes27', 'bytes13'), D), T[bo, by, by])  # fixed case 35057
ck(f(('bool', 'bytes27', 'bytes14'), D), T[bo, by, by])  # fixed case 35058
ck(f(('bool', 'bytes27', 'bytes15'), D), T[bo, by, by])  # fixed case 35059
ck(f(('bool', 'bytes27', 'bytes16'), D), T[bo, by, by])  # fixed case 35060
ck(f(('bool', 'bytes27', 'bytes17'), D), T[bo, by, by])  # fixed case 35061
ck(f(('bool', 'bytes27', 'bytes18'), D), T[bo, by, by])  # fixed case 35062
ck(f(('bool', 'bytes27', 'bytes19'), D), T[bo, by, by])  # fixed case 35063
ck(f(('bool', 'bytes27', 'bytes2'), D), T[bo, by, by])  # fixed case 35064
ck(f(('bool', 'bytes27', 'bytes20'), D), T[bo, by, by])  # fixed case 35065
ck(f(('bool', 'bytes27', 'bytes21'), D), T[bo, by, by])  # fixed case 35066
ck(f(('bool', 'bytes27', 'bytes22'), D), T[bo, by, by])  # fixed case 35067
ck(f(('bool', 'bytes27', 'bytes23'), D), T[bo, by, by])  # fixed case 35068
ck(f(('bool', 'bytes27', 'bytes24'), D), T[bo, by, by])  # fixed case 35069
ck(f(('bool', 'bytes27', 'bytes25'), D), T[bo, by, by])  # fixed case 35070
ck(f(('bool', 'bytes27', 'bytes26'), D), T[bo, by, by])  # fixed case 35071
ck(f(('bool', 'bytes27', 'bytes27'), D), T[bo, by, by])  # fixed case 35072
ck(f(('bool', 'bytes27', 'bytes28'), D), T[bo, by, by])  # fixed case 35073
ck(f(('bool', 'bytes27', 'bytes29'), D), T[bo, by, by])  # fixed case 35074
ck(f(('bool', 'bytes27', 'bytes3'), D), T[bo, by, by])  # fixed case 35075
ck(f(('bool', 'bytes27', 'bytes30'), D), T[bo, by, by])  # fixed case 35076
ck(f(('bool', 'bytes27', 'bytes31'), D), T[bo, by, by])  # fixed case 35077
ck(f(('bool', 'bytes27', 'bytes32'), D), T[bo, by, by])  # fixed case 35078
ck(f(('bool', 'bytes27', 'bytes4'), D), T[bo, by, by])  # fixed case 35079
ck(f(('bool', 'bytes27', 'bytes5'), D), T[bo, by, by])  # fixed case 35080
ck(f(('bool', 'bytes27', 'bytes6'), D), T[bo, by, by])  # fixed case 35081
ck(f(('bool', 'bytes27', 'bytes7'), D), T[bo, by, by])  # fixed case 35082
ck(f(('bool', 'bytes27', 'bytes8'), D), T[bo, by, by])  # fixed case 35083
ck(f(('bool', 'bytes27', 'bytes9'), D), T[bo, by, by])  # fixed case 35084
ck(f(('bool', 'bytes27', 'int104'), D), T[bo, by, i])  # fixed case 35085
ck(f(('bool', 'bytes27', 'int112'), D), T[bo, by, i])  # fixed case 35086
ck(f(('bool', 'bytes27', 'int120'), D), T[bo, by, i])  # fixed case 35087
ck(f(('bool', 'bytes27', 'int128'), D), T[bo, by, i])  # fixed case 35088
ck(f(('bool', 'bytes27', 'int136'), D), T[bo, by, i])  # fixed case 35089
ck(f(('bool', 'bytes27', 'int144'), D), T[bo, by, i])  # fixed case 35090
ck(f(('bool', 'bytes27', 'int152'), D), T[bo, by, i])  # fixed case 35091
ck(f(('bool', 'bytes27', 'int16'), D), T[bo, by, i])  # fixed case 35092
ck(f(('bool', 'bytes27', 'int160'), D), T[bo, by, i])  # fixed case 35093
ck(f(('bool', 'bytes27', 'int168'), D), T[bo, by, i])  # fixed case 35094
ck(f(('bool', 'bytes27', 'int176'), D), T[bo, by, i])  # fixed case 35095
ck(f(('bool', 'bytes27', 'int184'), D), T[bo, by, i])  # fixed case 35096
ck(f(('bool', 'bytes27', 'int192'), D), T[bo, by, i])  # fixed case 35097
ck(f(('bool', 'bytes27', 'int200'), D), T[bo, by, i])  # fixed case 35098
ck(f(('bool', 'bytes27', 'int208'), D), T[bo, by, i])  # fixed case 35099
ck(f(('bool', 'bytes27', 'int216'), D), T[bo, by, i])  # fixed case 35100
ck(f(('bool', 'bytes27', 'int224'), D), T[bo, by, i])  # fixed case 35101
ck(f(('bool', 'bytes27', 'int232'), D), T[bo, by, i])  # fixed case 35102
ck(f(('bool', 'bytes27', 'int24'), D), T[bo, by, i])  # fixed case 35103
ck(f(('bool', 'bytes27', 'int240'), D), T[bo, by, i])  # fixed case 35104
ck(f(('bool', 'bytes27', 'int248'), D), T[bo, by, i])  # fixed case 35105
ck(f(('bool', 'bytes27', 'int256'), D), T[bo, by, i])  # fixed case 35106
ck(f(('bool', 'bytes27', 'int32'), D), T[bo, by, i])  # fixed case 35107
ck(f(('bool', 'bytes27', 'int40'), D), T[bo, by, i])  # fixed case 35108
ck(f(('bool', 'bytes27', 'int48'), D), T[bo, by, i])  # fixed case 35109
ck(f(('bool', 'bytes27', 'int56'), D), T[bo, by, i])  # fixed case 35110
ck(f(('bool', 'bytes27', 'int64'), D), T[bo, by, i])  # fixed case 35111
ck(f(('bool', 'bytes27', 'int72'), D), T[bo, by, i])  # fixed case 35112
ck(f(('bool', 'bytes27', 'int8'), D), T[bo, by, i])  # fixed case 35113
ck(f(('bool', 'bytes27', 'int80'), D), T[bo, by, i])  # fixed case 35114
ck(f(('bool', 'bytes27', 'int88'), D), T[bo, by, i])  # fixed case 35115
ck(f(('bool', 'bytes27', 'int96'), D), T[bo, by, i])  # fixed case 35116
ck(f(('bool', 'bytes27', 'string'), D), T[bo, by, s])  # fixed case 35117
ck(f(('bool', 'bytes27', 'string[]'), D), T[bo, by, A])  # fixed case 35118
ck(f(('bool', 'bytes27', 'uint104'), D), T[bo, by, i])  # fixed case 35119
ck(f(('bool', 'bytes27', 'uint112'), D), T[bo, by, i])  # fixed case 35120
ck(f(('bool', 'bytes27', 'uint120'), D), T[bo, by, i])  # fixed case 35121
ck(f(('bool', 'bytes27', 'uint128'), D), T[bo, by, i])  # fixed case 35122
ck(f(('bool', 'bytes27', 'uint136'), D), T[bo, by, i])  # fixed case 35123
ck(f(('bool', 'bytes27', 'uint144'), D), T[bo, by, i])  # fixed case 35124
ck(f(('bool', 'bytes27', 'uint152'), D), T[bo, by, i])  # fixed case 35125
ck(f(('bool', 'bytes27', 'uint16'), D), T[bo, by, i])  # fixed case 35126
ck(f(('bool', 'bytes27', 'uint160'), D), T[bo, by, i])  # fixed case 35127
ck(f(('bool', 'bytes27', 'uint168'), D), T[bo, by, i])  # fixed case 35128
ck(f(('bool', 'bytes27', 'uint176'), D), T[bo, by, i])  # fixed case 35129
ck(f(('bool', 'bytes27', 'uint184'), D), T[bo, by, i])  # fixed case 35130
ck(f(('bool', 'bytes27', 'uint192'), D), T[bo, by, i])  # fixed case 35131
ck(f(('bool', 'bytes27', 'uint200'), D), T[bo, by, i])  # fixed case 35132
ck(f(('bool', 'bytes27', 'uint208'), D), T[bo, by, i])  # fixed case 35133
ck(f(('bool', 'bytes27', 'uint216'), D), T[bo, by, i])  # fixed case 35134
ck(f(('bool', 'bytes27', 'uint224'), D), T[bo, by, i])  # fixed case 35135
ck(f(('bool', 'bytes27', 'uint232'), D), T[bo, by, i])  # fixed case 35136
ck(f(('bool', 'bytes27', 'uint24'), D), T[bo, by, i])  # fixed case 35137
ck(f(('bool', 'bytes27', 'uint240'), D), T[bo, by, i])  # fixed case 35138
ck(f(('bool', 'bytes27', 'uint248'), D), T[bo, by, i])  # fixed case 35139
ck(f(('bool', 'bytes27', 'uint256'), D), T[bo, by, i])  # fixed case 35140
ck(f(('bool', 'bytes27', 'uint32'), D), T[bo, by, i])  # fixed case 35141
ck(f(('bool', 'bytes27', 'uint40'), D), T[bo, by, i])  # fixed case 35142
ck(f(('bool', 'bytes27', 'uint48'), D), T[bo, by, i])  # fixed case 35143
ck(f(('bool', 'bytes27', 'uint56'), D), T[bo, by, i])  # fixed case 35144
ck(f(('bool', 'bytes27', 'uint64'), D), T[bo, by, i])  # fixed case 35145
ck(f(('bool', 'bytes27', 'uint72'), D), T[bo, by, i])  # fixed case 35146
ck(f(('bool', 'bytes27', 'uint8'), D), T[bo, by, i])  # fixed case 35147
ck(f(('bool', 'bytes27', 'uint80'), D), T[bo, by, i])  # fixed case 35148
ck(f(('bool', 'bytes27', 'uint88'), D), T[bo, by, i])  # fixed case 35149
ck(f(('bool', 'bytes27', 'uint96'), D), T[bo, by, i])  # fixed case 35150
ck(f(('bool', 'bytes27', '?'), D), T[bo, by, A])  # fixed case 35151
ck(f(('bool', 'bytes28', 'address'), D), T[bo, by, HA])  # fixed case 35152
ck(f(('bool', 'bytes28', 'address[]'), D), T[bo, by, A])  # fixed case 35153
ck(f(('bool', 'bytes28', 'bool'), D), T[bo, by, bo])  # fixed case 35154
ck(f(('bool', 'bytes28', 'bool[]'), D), T[bo, by, A])  # fixed case 35155
ck(f(('bool', 'bytes28', 'bytes'), D), T[bo, by, by])  # fixed case 35156
ck(f(('bool', 'bytes28', 'bytes1'), D), T[bo, by, by])  # fixed case 35157
ck(f(('bool', 'bytes28', 'bytes10'), D), T[bo, by, by])  # fixed case 35158
ck(f(('bool', 'bytes28', 'bytes11'), D), T[bo, by, by])  # fixed case 35159
ck(f(('bool', 'bytes28', 'bytes12'), D), T[bo, by, by])  # fixed case 35160
ck(f(('bool', 'bytes28', 'bytes13'), D), T[bo, by, by])  # fixed case 35161
ck(f(('bool', 'bytes28', 'bytes14'), D), T[bo, by, by])  # fixed case 35162
ck(f(('bool', 'bytes28', 'bytes15'), D), T[bo, by, by])  # fixed case 35163
ck(f(('bool', 'bytes28', 'bytes16'), D), T[bo, by, by])  # fixed case 35164
ck(f(('bool', 'bytes28', 'bytes17'), D), T[bo, by, by])  # fixed case 35165
ck(f(('bool', 'bytes28', 'bytes18'), D), T[bo, by, by])  # fixed case 35166
ck(f(('bool', 'bytes28', 'bytes19'), D), T[bo, by, by])  # fixed case 35167
ck(f(('bool', 'bytes28', 'bytes2'), D), T[bo, by, by])  # fixed case 35168
ck(f(('bool', 'bytes28', 'bytes20'), D), T[bo, by, by])  # fixed case 35169
ck(f(('bool', 'bytes28', 'bytes21'), D), T[bo, by, by])  # fixed case 35170
ck(f(('bool', 'bytes28', 'bytes22'), D), T[bo, by, by])  # fixed case 35171
ck(f(('bool', 'bytes28', 'bytes23'), D), T[bo, by, by])  # fixed case 35172
ck(f(('bool', 'bytes28', 'bytes24'), D), T[bo, by, by])  # fixed case 35173
ck(f(('bool', 'bytes28', 'bytes25'), D), T[bo, by, by])  # fixed case 35174
ck(f(('bool', 'bytes28', 'bytes26'), D), T[bo, by, by])  # fixed case 35175
ck(f(('bool', 'bytes28', 'bytes27'), D), T[bo, by, by])  # fixed case 35176
ck(f(('bool', 'bytes28', 'bytes28'), D), T[bo, by, by])  # fixed case 35177
ck(f(('bool', 'bytes28', 'bytes29'), D), T[bo, by, by])  # fixed case 35178
ck(f(('bool', 'bytes28', 'bytes3'), D), T[bo, by, by])  # fixed case 35179
ck(f(('bool', 'bytes28', 'bytes30'), D), T[bo, by, by])  # fixed case 35180
ck(f(('bool', 'bytes28', 'bytes31'), D), T[bo, by, by])  # fixed case 35181
ck(f(('bool', 'bytes28', 'bytes32'), D), T[bo, by, by])  # fixed case 35182
ck(f(('bool', 'bytes28', 'bytes4'), D), T[bo, by, by])  # fixed case 35183
ck(f(('bool', 'bytes28', 'bytes5'), D), T[bo, by, by])  # fixed case 35184
ck(f(('bool', 'bytes28', 'bytes6'), D), T[bo, by, by])  # fixed case 35185
ck(f(('bool', 'bytes28', 'bytes7'), D), T[bo, by, by])  # fixed case 35186
ck(f(('bool', 'bytes28', 'bytes8'), D), T[bo, by, by])  # fixed case 35187
ck(f(('bool', 'bytes28', 'bytes9'), D), T[bo, by, by])  # fixed case 35188
ck(f(('bool', 'bytes28', 'int104'), D), T[bo, by, i])  # fixed case 35189
ck(f(('bool', 'bytes28', 'int112'), D), T[bo, by, i])  # fixed case 35190
ck(f(('bool', 'bytes28', 'int120'), D), T[bo, by, i])  # fixed case 35191
ck(f(('bool', 'bytes28', 'int128'), D), T[bo, by, i])  # fixed case 35192
ck(f(('bool', 'bytes28', 'int136'), D), T[bo, by, i])  # fixed case 35193
ck(f(('bool', 'bytes28', 'int144'), D), T[bo, by, i])  # fixed case 35194
ck(f(('bool', 'bytes28', 'int152'), D), T[bo, by, i])  # fixed case 35195
ck(f(('bool', 'bytes28', 'int16'), D), T[bo, by, i])  # fixed case 35196
ck(f(('bool', 'bytes28', 'int160'), D), T[bo, by, i])  # fixed case 35197
ck(f(('bool', 'bytes28', 'int168'), D), T[bo, by, i])  # fixed case 35198
ck(f(('bool', 'bytes28', 'int176'), D), T[bo, by, i])  # fixed case 35199
ck(f(('bool', 'bytes28', 'int184'), D), T[bo, by, i])  # fixed case 35200
ck(f(('bool', 'bytes28', 'int192'), D), T[bo, by, i])  # fixed case 35201
ck(f(('bool', 'bytes28', 'int200'), D), T[bo, by, i])  # fixed case 35202
ck(f(('bool', 'bytes28', 'int208'), D), T[bo, by, i])  # fixed case 35203
ck(f(('bool', 'bytes28', 'int216'), D), T[bo, by, i])  # fixed case 35204
ck(f(('bool', 'bytes28', 'int224'), D), T[bo, by, i])  # fixed case 35205
ck(f(('bool', 'bytes28', 'int232'), D), T[bo, by, i])  # fixed case 35206
ck(f(('bool', 'bytes28', 'int24'), D), T[bo, by, i])  # fixed case 35207
ck(f(('bool', 'bytes28', 'int240'), D), T[bo, by, i])  # fixed case 35208
ck(f(('bool', 'bytes28', 'int248'), D), T[bo, by, i])  # fixed case 35209
ck(f(('bool', 'bytes28', 'int256'), D), T[bo, by, i])  # fixed case 35210
ck(f(('bool', 'bytes28', 'int32'), D), T[bo, by, i])  # fixed case 35211
ck(f(('bool', 'bytes28', 'int40'), D), T[bo, by, i])  # fixed case 35212
ck(f(('bool', 'bytes28', 'int48'), D), T[bo, by, i])  # fixed case 35213
ck(f(('bool', 'bytes28', 'int56'), D), T[bo, by, i])  # fixed case 35214
ck(f(('bool', 'bytes28', 'int64'), D), T[bo, by, i])  # fixed case 35215
ck(f(('bool', 'bytes28', 'int72'), D), T[bo, by, i])  # fixed case 35216
ck(f(('bool', 'bytes28', 'int8'), D), T[bo, by, i])  # fixed case 35217
ck(f(('bool', 'bytes28', 'int80'), D), T[bo, by, i])  # fixed case 35218
ck(f(('bool', 'bytes28', 'int88'), D), T[bo, by, i])  # fixed case 35219
ck(f(('bool', 'bytes28', 'int96'), D), T[bo, by, i])  # fixed case 35220
ck(f(('bool', 'bytes28', 'string'), D), T[bo, by, s])  # fixed case 35221
ck(f(('bool', 'bytes28', 'string[]'), D), T[bo, by, A])  # fixed case 35222
ck(f(('bool', 'bytes28', 'uint104'), D), T[bo, by, i])  # fixed case 35223
ck(f(('bool', 'bytes28', 'uint112'), D), T[bo, by, i])  # fixed case 35224
ck(f(('bool', 'bytes28', 'uint120'), D), T[bo, by, i])  # fixed case 35225
ck(f(('bool', 'bytes28', 'uint128'), D), T[bo, by, i])  # fixed case 35226
ck(f(('bool', 'bytes28', 'uint136'), D), T[bo, by, i])  # fixed case 35227
ck(f(('bool', 'bytes28', 'uint144'), D), T[bo, by, i])  # fixed case 35228
ck(f(('bool', 'bytes28', 'uint152'), D), T[bo, by, i])  # fixed case 35229
ck(f(('bool', 'bytes28', 'uint16'), D), T[bo, by, i])  # fixed case 35230
ck(f(('bool', 'bytes28', 'uint160'), D), T[bo, by, i])  # fixed case 35231
ck(f(('bool', 'bytes28', 'uint168'), D), T[bo, by, i])  # fixed case 35232
ck(f(('bool', 'bytes28', 'uint176'), D), T[bo, by, i])  # fixed case 35233
ck(f(('bool', 'bytes28', 'uint184'), D), T[bo, by, i])  # fixed case 35234
ck(f(('bool', 'bytes28', 'uint192'), D), T[bo, by, i])  # fixed case 35235
ck(f(('bool', 'bytes28', 'uint200'), D), T[bo, by, i])  # fixed case 35236
ck(f(('bool', 'bytes28', 'uint208'), D), T[bo, by, i])  # fixed case 35237
ck(f(('bool', 'bytes28', 'uint216'), D), T[bo, by, i])  # fixed case 35238
ck(f(('bool', 'bytes28', 'uint224'), D), T[bo, by, i])  # fixed case 35239
ck(f(('bool', 'bytes28', 'uint232'), D), T[bo, by, i])  # fixed case 35240
ck(f(('bool', 'bytes28', 'uint24'), D), T[bo, by, i])  # fixed case 35241
ck(f(('bool', 'bytes28', 'uint240'), D), T[bo, by, i])  # fixed case 35242
ck(f(('bool', 'bytes28', 'uint248'), D), T[bo, by, i])  # fixed case 35243
ck(f(('bool', 'bytes28', 'uint256'), D), T[bo, by, i])  # fixed case 35244
ck(f(('bool', 'bytes28', 'uint32'), D), T[bo, by, i])  # fixed case 35245
ck(f(('bool', 'bytes28', 'uint40'), D), T[bo, by, i])  # fixed case 35246
ck(f(('bool', 'bytes28', 'uint48'), D), T[bo, by, i])  # fixed case 35247
ck(f(('bool', 'bytes28', 'uint56'), D), T[bo, by, i])  # fixed case 35248
ck(f(('bool', 'bytes28', 'uint64'), D), T[bo, by, i])  # fixed case 35249
ck(f(('bool', 'bytes28', 'uint72'), D), T[bo, by, i])  # fixed case 35250
ck(f(('bool', 'bytes28', 'uint8'), D), T[bo, by, i])  # fixed case 35251
ck(f(('bool', 'bytes28', 'uint80'), D), T[bo, by, i])  # fixed case 35252
ck(f(('bool', 'bytes28', 'uint88'), D), T[bo, by, i])  # fixed case 35253
ck(f(('bool', 'bytes28', 'uint96'), D), T[bo, by, i])  # fixed case 35254
ck(f(('bool', 'bytes28', '?'), D), T[bo, by, A])  # fixed case 35255
ck(f(('bool', 'bytes29', 'address'), D), T[bo, by, HA])  # fixed case 35256
ck(f(('bool', 'bytes29', 'address[]'), D), T[bo, by, A])  # fixed case 35257
ck(f(('bool', 'bytes29', 'bool'), D), T[bo, by, bo])  # fixed case 35258
ck(f(('bool', 'bytes29', 'bool[]'), D), T[bo, by, A])  # fixed case 35259
ck(f(('bool', 'bytes29', 'bytes'), D), T[bo, by, by])  # fixed case 35260
ck(f(('bool', 'bytes29', 'bytes1'), D), T[bo, by, by])  # fixed case 35261
ck(f(('bool', 'bytes29', 'bytes10'), D), T[bo, by, by])  # fixed case 35262
ck(f(('bool', 'bytes29', 'bytes11'), D), T[bo, by, by])  # fixed case 35263
ck(f(('bool', 'bytes29', 'bytes12'), D), T[bo, by, by])  # fixed case 35264
ck(f(('bool', 'bytes29', 'bytes13'), D), T[bo, by, by])  # fixed case 35265
ck(f(('bool', 'bytes29', 'bytes14'), D), T[bo, by, by])  # fixed case 35266
ck(f(('bool', 'bytes29', 'bytes15'), D), T[bo, by, by])  # fixed case 35267
ck(f(('bool', 'bytes29', 'bytes16'), D), T[bo, by, by])  # fixed case 35268
ck(f(('bool', 'bytes29', 'bytes17'), D), T[bo, by, by])  # fixed case 35269
ck(f(('bool', 'bytes29', 'bytes18'), D), T[bo, by, by])  # fixed case 35270
ck(f(('bool', 'bytes29', 'bytes19'), D), T[bo, by, by])  # fixed case 35271
ck(f(('bool', 'bytes29', 'bytes2'), D), T[bo, by, by])  # fixed case 35272
ck(f(('bool', 'bytes29', 'bytes20'), D), T[bo, by, by])  # fixed case 35273
ck(f(('bool', 'bytes29', 'bytes21'), D), T[bo, by, by])  # fixed case 35274
ck(f(('bool', 'bytes29', 'bytes22'), D), T[bo, by, by])  # fixed case 35275
ck(f(('bool', 'bytes29', 'bytes23'), D), T[bo, by, by])  # fixed case 35276
ck(f(('bool', 'bytes29', 'bytes24'), D), T[bo, by, by])  # fixed case 35277
ck(f(('bool', 'bytes29', 'bytes25'), D), T[bo, by, by])  # fixed case 35278
ck(f(('bool', 'bytes29', 'bytes26'), D), T[bo, by, by])  # fixed case 35279
ck(f(('bool', 'bytes29', 'bytes27'), D), T[bo, by, by])  # fixed case 35280
ck(f(('bool', 'bytes29', 'bytes28'), D), T[bo, by, by])  # fixed case 35281
ck(f(('bool', 'bytes29', 'bytes29'), D), T[bo, by, by])  # fixed case 35282
ck(f(('bool', 'bytes29', 'bytes3'), D), T[bo, by, by])  # fixed case 35283
ck(f(('bool', 'bytes29', 'bytes30'), D), T[bo, by, by])  # fixed case 35284
ck(f(('bool', 'bytes29', 'bytes31'), D), T[bo, by, by])  # fixed case 35285
ck(f(('bool', 'bytes29', 'bytes32'), D), T[bo, by, by])  # fixed case 35286
ck(f(('bool', 'bytes29', 'bytes4'), D), T[bo, by, by])  # fixed case 35287
ck(f(('bool', 'bytes29', 'bytes5'), D), T[bo, by, by])  # fixed case 35288
ck(f(('bool', 'bytes29', 'bytes6'), D), T[bo, by, by])  # fixed case 35289
ck(f(('bool', 'bytes29', 'bytes7'), D), T[bo, by, by])  # fixed case 35290
ck(f(('bool', 'bytes29', 'bytes8'), D), T[bo, by, by])  # fixed case 35291
ck(f(('bool', 'bytes29', 'bytes9'), D), T[bo, by, by])  # fixed case 35292
ck(f(('bool', 'bytes29', 'int104'), D), T[bo, by, i])  # fixed case 35293
ck(f(('bool', 'bytes29', 'int112'), D), T[bo, by, i])  # fixed case 35294
ck(f(('bool', 'bytes29', 'int120'), D), T[bo, by, i])  # fixed case 35295
ck(f(('bool', 'bytes29', 'int128'), D), T[bo, by, i])  # fixed case 35296
ck(f(('bool', 'bytes29', 'int136'), D), T[bo, by, i])  # fixed case 35297
ck(f(('bool', 'bytes29', 'int144'), D), T[bo, by, i])  # fixed case 35298
ck(f(('bool', 'bytes29', 'int152'), D), T[bo, by, i])  # fixed case 35299
ck(f(('bool', 'bytes29', 'int16'), D), T[bo, by, i])  # fixed case 35300
ck(f(('bool', 'bytes29', 'int160'), D), T[bo, by, i])  # fixed case 35301
ck(f(('bool', 'bytes29', 'int168'), D), T[bo, by, i])  # fixed case 35302
ck(f(('bool', 'bytes29', 'int176'), D), T[bo, by, i])  # fixed case 35303
ck(f(('bool', 'bytes29', 'int184'), D), T[bo, by, i])  # fixed case 35304
ck(f(('bool', 'bytes29', 'int192'), D), T[bo, by, i])  # fixed case 35305
ck(f(('bool', 'bytes29', 'int200'), D), T[bo, by, i])  # fixed case 35306
ck(f(('bool', 'bytes29', 'int208'), D), T[bo, by, i])  # fixed case 35307
ck(f(('bool', 'bytes29', 'int216'), D), T[bo, by, i])  # fixed case 35308
ck(f(('bool', 'bytes29', 'int224'), D), T[bo, by, i])  # fixed case 35309
ck(f(('bool', 'bytes29', 'int232'), D), T[bo, by, i])  # fixed case 35310
ck(f(('bool', 'bytes29', 'int24'), D), T[bo, by, i])  # fixed case 35311
ck(f(('bool', 'bytes29', 'int240'), D), T[bo, by, i])  # fixed case 35312
ck(f(('bool', 'bytes29', 'int248'), D), T[bo, by, i])  # fixed case 35313
ck(f(('bool', 'bytes29', 'int256'), D), T[bo, by, i])  # fixed case 35314
ck(f(('bool', 'bytes29', 'int32'), D), T[bo, by, i])  # fixed case 35315
ck(f(('bool', 'bytes29', 'int40'), D), T[bo, by, i])  # fixed case 35316
ck(f(('bool', 'bytes29', 'int48'), D), T[bo, by, i])  # fixed case 35317
ck(f(('bool', 'bytes29', 'int56'), D), T[bo, by, i])  # fixed case 35318
ck(f(('bool', 'bytes29', 'int64'), D), T[bo, by, i])  # fixed case 35319
ck(f(('bool', 'bytes29', 'int72'), D), T[bo, by, i])  # fixed case 35320
ck(f(('bool', 'bytes29', 'int8'), D), T[bo, by, i])  # fixed case 35321
ck(f(('bool', 'bytes29', 'int80'), D), T[bo, by, i])  # fixed case 35322
ck(f(('bool', 'bytes29', 'int88'), D), T[bo, by, i])  # fixed case 35323
ck(f(('bool', 'bytes29', 'int96'), D), T[bo, by, i])  # fixed case 35324
ck(f(('bool', 'bytes29', 'string'), D), T[bo, by, s])  # fixed case 35325
ck(f(('bool', 'bytes29', 'string[]'), D), T[bo, by, A])  # fixed case 35326
ck(f(('bool', 'bytes29', 'uint104'), D), T[bo, by, i])  # fixed case 35327
ck(f(('bool', 'bytes29', 'uint112'), D), T[bo, by, i])  # fixed case 35328
ck(f(('bool', 'bytes29', 'uint120'), D), T[bo, by, i])  # fixed case 35329
ck(f(('bool', 'bytes29', 'uint128'), D), T[bo, by, i])  # fixed case 35330
ck(f(('bool', 'bytes29', 'uint136'), D), T[bo, by, i])  # fixed case 35331
ck(f(('bool', 'bytes29', 'uint144'), D), T[bo, by, i])  # fixed case 35332
ck(f(('bool', 'bytes29', 'uint152'), D), T[bo, by, i])  # fixed case 35333
ck(f(('bool', 'bytes29', 'uint16'), D), T[bo, by, i])  # fixed case 35334
ck(f(('bool', 'bytes29', 'uint160'), D), T[bo, by, i])  # fixed case 35335
ck(f(('bool', 'bytes29', 'uint168'), D), T[bo, by, i])  # fixed case 35336
ck(f(('bool', 'bytes29', 'uint176'), D), T[bo, by, i])  # fixed case 35337
ck(f(('bool', 'bytes29', 'uint184'), D), T[bo, by, i])  # fixed case 35338
ck(f(('bool', 'bytes29', 'uint192'), D), T[bo, by, i])  # fixed case 35339
ck(f(('bool', 'bytes29', 'uint200'), D), T[bo, by, i])  # fixed case 35340
ck(f(('bool', 'bytes29', 'uint208'), D), T[bo, by, i])  # fixed case 35341
ck(f(('bool', 'bytes29', 'uint216'), D), T[bo, by, i])  # fixed case 35342
ck(f(('bool', 'bytes29', 'uint224'), D), T[bo, by, i])  # fixed case 35343
ck(f(('bool', 'bytes29', 'uint232'), D), T[bo, by, i])  # fixed case 35344
ck(f(('bool', 'bytes29', 'uint24'), D), T[bo, by, i])  # fixed case 35345
ck(f(('bool', 'bytes29', 'uint240'), D), T[bo, by, i])  # fixed case 35346
ck(f(('bool', 'bytes29', 'uint248'), D), T[bo, by, i])  # fixed case 35347
ck(f(('bool', 'bytes29', 'uint256'), D), T[bo, by, i])  # fixed case 35348
ck(f(('bool', 'bytes29', 'uint32'), D), T[bo, by, i])  # fixed case 35349
ck(f(('bool', 'bytes29', 'uint40'), D), T[bo, by, i])  # fixed case 35350
ck(f(('bool', 'bytes29', 'uint48'), D), T[bo, by, i])  # fixed case 35351
ck(f(('bool', 'bytes29', 'uint56'), D), T[bo, by, i])  # fixed case 35352
ck(f(('bool', 'bytes29', 'uint64'), D), T[bo, by, i])  # fixed case 35353
ck(f(('bool', 'bytes29', 'uint72'), D), T[bo, by, i])  # fixed case 35354
ck(f(('bool', 'bytes29', 'uint8'), D), T[bo, by, i])  # fixed case 35355
ck(f(('bool', 'bytes29', 'uint80'), D), T[bo, by, i])  # fixed case 35356
ck(f(('bool', 'bytes29', 'uint88'), D), T[bo, by, i])  # fixed case 35357
ck(f(('bool', 'bytes29', 'uint96'), D), T[bo, by, i])  # fixed case 35358
ck(f(('bool', 'bytes29', '?'), D), T[bo, by, A])  # fixed case 35359
ck(f(('bool', 'bytes3', 'address'), D), T[bo, by, HA])  # fixed case 35360
ck(f(('bool', 'bytes3', 'address[]'), D), T[bo, by, A])  # fixed case 35361
ck(f(('bool', 'bytes3', 'bool'), D), T[bo, by, bo])  # fixed case 35362
ck(f(('bool', 'bytes3', 'bool[]'), D), T[bo, by, A])  # fixed case 35363
ck(f(('bool', 'bytes3', 'bytes'), D), T[bo, by, by])  # fixed case 35364
ck(f(('bool', 'bytes3', 'bytes1'), D), T[bo, by, by])  # fixed case 35365
ck(f(('bool', 'bytes3', 'bytes10'), D), T[bo, by, by])  # fixed case 35366
ck(f(('bool', 'bytes3', 'bytes11'), D), T[bo, by, by])  # fixed case 35367
ck(f(('bool', 'bytes3', 'bytes12'), D), T[bo, by, by])  # fixed case 35368
ck(f(('bool', 'bytes3', 'bytes13'), D), T[bo, by, by])  # fixed case 35369
ck(f(('bool', 'bytes3', 'bytes14'), D), T[bo, by, by])  # fixed case 35370
ck(f(('bool', 'bytes3', 'bytes15'), D), T[bo, by, by])  # fixed case 35371
ck(f(('bool', 'bytes3', 'bytes16'), D), T[bo, by, by])  # fixed case 35372
ck(f(('bool', 'bytes3', 'bytes17'), D), T[bo, by, by])  # fixed case 35373
ck(f(('bool', 'bytes3', 'bytes18'), D), T[bo, by, by])  # fixed case 35374
ck(f(('bool', 'bytes3', 'bytes19'), D), T[bo, by, by])  # fixed case 35375
ck(f(('bool', 'bytes3', 'bytes2'), D), T[bo, by, by])  # fixed case 35376
ck(f(('bool', 'bytes3', 'bytes20'), D), T[bo, by, by])  # fixed case 35377
ck(f(('bool', 'bytes3', 'bytes21'), D), T[bo, by, by])  # fixed case 35378
ck(f(('bool', 'bytes3', 'bytes22'), D), T[bo, by, by])  # fixed case 35379
ck(f(('bool', 'bytes3', 'bytes23'), D), T[bo, by, by])  # fixed case 35380
ck(f(('bool', 'bytes3', 'bytes24'), D), T[bo, by, by])  # fixed case 35381
ck(f(('bool', 'bytes3', 'bytes25'), D), T[bo, by, by])  # fixed case 35382
ck(f(('bool', 'bytes3', 'bytes26'), D), T[bo, by, by])  # fixed case 35383
ck(f(('bool', 'bytes3', 'bytes27'), D), T[bo, by, by])  # fixed case 35384
ck(f(('bool', 'bytes3', 'bytes28'), D), T[bo, by, by])  # fixed case 35385
ck(f(('bool', 'bytes3', 'bytes29'), D), T[bo, by, by])  # fixed case 35386
ck(f(('bool', 'bytes3', 'bytes3'), D), T[bo, by, by])  # fixed case 35387
ck(f(('bool', 'bytes3', 'bytes30'), D), T[bo, by, by])  # fixed case 35388
ck(f(('bool', 'bytes3', 'bytes31'), D), T[bo, by, by])  # fixed case 35389
ck(f(('bool', 'bytes3', 'bytes32'), D), T[bo, by, by])  # fixed case 35390
ck(f(('bool', 'bytes3', 'bytes4'), D), T[bo, by, by])  # fixed case 35391
ck(f(('bool', 'bytes3', 'bytes5'), D), T[bo, by, by])  # fixed case 35392
ck(f(('bool', 'bytes3', 'bytes6'), D), T[bo, by, by])  # fixed case 35393
ck(f(('bool', 'bytes3', 'bytes7'), D), T[bo, by, by])  # fixed case 35394
ck(f(('bool', 'bytes3', 'bytes8'), D), T[bo, by, by])  # fixed case 35395
ck(f(('bool', 'bytes3', 'bytes9'), D), T[bo, by, by])  # fixed case 35396
ck(f(('bool', 'bytes3', 'int104'), D), T[bo, by, i])  # fixed case 35397
ck(f(('bool', 'bytes3', 'int112'), D), T[bo, by, i])  # fixed case 35398
ck(f(('bool', 'bytes3', 'int120'), D), T[bo, by, i])  # fixed case 35399
ck(f(('bool', 'bytes3', 'int128'), D), T[bo, by, i])  # fixed case 35400
ck(f(('bool', 'bytes3', 'int136'), D), T[bo, by, i])  # fixed case 35401
ck(f(('bool', 'bytes3', 'int144'), D), T[bo, by, i])  # fixed case 35402
ck(f(('bool', 'bytes3', 'int152'), D), T[bo, by, i])  # fixed case 35403
ck(f(('bool', 'bytes3', 'int16'), D), T[bo, by, i])  # fixed case 35404
ck(f(('bool', 'bytes3', 'int160'), D), T[bo, by, i])  # fixed case 35405
ck(f(('bool', 'bytes3', 'int168'), D), T[bo, by, i])  # fixed case 35406
ck(f(('bool', 'bytes3', 'int176'), D), T[bo, by, i])  # fixed case 35407
ck(f(('bool', 'bytes3', 'int184'), D), T[bo, by, i])  # fixed case 35408
ck(f(('bool', 'bytes3', 'int192'), D), T[bo, by, i])  # fixed case 35409
ck(f(('bool', 'bytes3', 'int200'), D), T[bo, by, i])  # fixed case 35410
ck(f(('bool', 'bytes3', 'int208'), D), T[bo, by, i])  # fixed case 35411
ck(f(('bool', 'bytes3', 'int216'), D), T[bo, by, i])  # fixed case 35412
ck(f(('bool', 'bytes3', 'int224'), D), T[bo, by, i])  # fixed case 35413
ck(f(('bool', 'bytes3', 'int232'), D), T[bo, by, i])  # fixed case 35414
ck(f(('bool', 'bytes3', 'int24'), D), T[bo, by, i])  # fixed case 35415
ck(f(('bool', 'bytes3', 'int240'), D), T[bo, by, i])  # fixed case 35416
ck(f(('bool', 'bytes3', 'int248'), D), T[bo, by, i])  # fixed case 35417
ck(f(('bool', 'bytes3', 'int256'), D), T[bo, by, i])  # fixed case 35418
ck(f(('bool', 'bytes3', 'int32'), D), T[bo, by, i])  # fixed case 35419
ck(f(('bool', 'bytes3', 'int40'), D), T[bo, by, i])  # fixed case 35420
ck(f(('bool', 'bytes3', 'int48'), D), T[bo, by, i])  # fixed case 35421
ck(f(('bool', 'bytes3', 'int56'), D), T[bo, by, i])  # fixed case 35422
ck(f(('bool', 'bytes3', 'int64'), D), T[bo, by, i])  # fixed case 35423
ck(f(('bool', 'bytes3', 'int72'), D), T[bo, by, i])  # fixed case 35424
ck(f(('bool', 'bytes3', 'int8'), D), T[bo, by, i])  # fixed case 35425
ck(f(('bool', 'bytes3', 'int80'), D), T[bo, by, i])  # fixed case 35426
ck(f(('bool', 'bytes3', 'int88'), D), T[bo, by, i])  # fixed case 35427
ck(f(('bool', 'bytes3', 'int96'), D), T[bo, by, i])  # fixed case 35428
ck(f(('bool', 'bytes3', 'string'), D), T[bo, by, s])  # fixed case 35429
ck(f(('bool', 'bytes3', 'string[]'), D), T[bo, by, A])  # fixed case 35430
ck(f(('bool', 'bytes3', 'uint104'), D), T[bo, by, i])  # fixed case 35431
ck(f(('bool', 'bytes3', 'uint112'), D), T[bo, by, i])  # fixed case 35432
ck(f(('bool', 'bytes3', 'uint120'), D), T[bo, by, i])  # fixed case 35433
ck(f(('bool', 'bytes3', 'uint128'), D), T[bo, by, i])  # fixed case 35434
ck(f(('bool', 'bytes3', 'uint136'), D), T[bo, by, i])  # fixed case 35435
ck(f(('bool', 'bytes3', 'uint144'), D), T[bo, by, i])  # fixed case 35436
ck(f(('bool', 'bytes3', 'uint152'), D), T[bo, by, i])  # fixed case 35437
ck(f(('bool', 'bytes3', 'uint16'), D), T[bo, by, i])  # fixed case 35438
ck(f(('bool', 'bytes3', 'uint160'), D), T[bo, by, i])  # fixed case 35439
ck(f(('bool', 'bytes3', 'uint168'), D), T[bo, by, i])  # fixed case 35440
ck(f(('bool', 'bytes3', 'uint176'), D), T[bo, by, i])  # fixed case 35441
ck(f(('bool', 'bytes3', 'uint184'), D), T[bo, by, i])  # fixed case 35442
ck(f(('bool', 'bytes3', 'uint192'), D), T[bo, by, i])  # fixed case 35443
ck(f(('bool', 'bytes3', 'uint200'), D), T[bo, by, i])  # fixed case 35444
ck(f(('bool', 'bytes3', 'uint208'), D), T[bo, by, i])  # fixed case 35445
ck(f(('bool', 'bytes3', 'uint216'), D), T[bo, by, i])  # fixed case 35446
ck(f(('bool', 'bytes3', 'uint224'), D), T[bo, by, i])  # fixed case 35447
ck(f(('bool', 'bytes3', 'uint232'), D), T[bo, by, i])  # fixed case 35448
ck(f(('bool', 'bytes3', 'uint24'), D), T[bo, by, i])  # fixed case 35449
ck(f(('bool', 'bytes3', 'uint240'), D), T[bo, by, i])  # fixed case 35450
ck(f(('bool', 'bytes3', 'uint248'), D), T[bo, by, i])  # fixed case 35451
ck(f(('bool', 'bytes3', 'uint256'), D), T[bo, by, i])  # fixed case 35452
ck(f(('bool', 'bytes3', 'uint32'), D), T[bo, by, i])  # fixed case 35453
ck(f(('bool', 'bytes3', 'uint40'), D), T[bo, by, i])  # fixed case 35454
ck(f(('bool', 'bytes3', 'uint48'), D), T[bo, by, i])  # fixed case 35455
ck(f(('bool', 'bytes3', 'uint56'), D), T[bo, by, i])  # fixed case 35456
ck(f(('bool', 'bytes3', 'uint64'), D), T[bo, by, i])  # fixed case 35457
ck(f(('bool', 'bytes3', 'uint72'), D), T[bo, by, i])  # fixed case 35458
ck(f(('bool', 'bytes3', 'uint8'), D), T[bo, by, i])  # fixed case 35459
ck(f(('bool', 'bytes3', 'uint80'), D), T[bo, by, i])  # fixed case 35460
ck(f(('bool', 'bytes3', 'uint88'), D), T[bo, by, i])  # fixed case 35461
ck(f(('bool', 'bytes3', 'uint96'), D), T[bo, by, i])  # fixed case 35462
ck(f(('bool', 'bytes3', '?'), D), T[bo, by, A])  # fixed case 35463
ck(f(('bool', 'bytes30', 'address'), D), T[bo, by, HA])  # fixed case 35464
ck(f(('bool', 'bytes30', 'address[]'), D), T[bo, by, A])  # fixed case 35465
ck(f(('bool', 'bytes30', 'bool'), D), T[bo, by, bo])  # fixed case 35466
ck(f(('bool', 'bytes30', 'bool[]'), D), T[bo, by, A])  # fixed case 35467
ck(f(('bool', 'bytes30', 'bytes'), D), T[bo, by, by])  # fixed case 35468
ck(f(('bool', 'bytes30', 'bytes1'), D), T[bo, by, by])  # fixed case 35469
ck(f(('bool', 'bytes30', 'bytes10'), D), T[bo, by, by])  # fixed case 35470
ck(f(('bool', 'bytes30', 'bytes11'), D), T[bo, by, by])  # fixed case 35471
ck(f(('bool', 'bytes30', 'bytes12'), D), T[bo, by, by])  # fixed case 35472
ck(f(('bool', 'bytes30', 'bytes13'), D), T[bo, by, by])  # fixed case 35473
ck(f(('bool', 'bytes30', 'bytes14'), D), T[bo, by, by])  # fixed case 35474
ck(f(('bool', 'bytes30', 'bytes15'), D), T[bo, by, by])  # fixed case 35475
ck(f(('bool', 'bytes30', 'bytes16'), D), T[bo, by, by])  # fixed case 35476
ck(f(('bool', 'bytes30', 'bytes17'), D), T[bo, by, by])  # fixed case 35477
ck(f(('bool', 'bytes30', 'bytes18'), D), T[bo, by, by])  # fixed case 35478
ck(f(('bool', 'bytes30', 'bytes19'), D), T[bo, by, by])  # fixed case 35479
ck(f(('bool', 'bytes30', 'bytes2'), D), T[bo, by, by])  # fixed case 35480
ck(f(('bool', 'bytes30', 'bytes20'), D), T[bo, by, by])  # fixed case 35481
ck(f(('bool', 'bytes30', 'bytes21'), D), T[bo, by, by])  # fixed case 35482
ck(f(('bool', 'bytes30', 'bytes22'), D), T[bo, by, by])  # fixed case 35483
ck(f(('bool', 'bytes30', 'bytes23'), D), T[bo, by, by])  # fixed case 35484
ck(f(('bool', 'bytes30', 'bytes24'), D), T[bo, by, by])  # fixed case 35485
ck(f(('bool', 'bytes30', 'bytes25'), D), T[bo, by, by])  # fixed case 35486
ck(f(('bool', 'bytes30', 'bytes26'), D), T[bo, by, by])  # fixed case 35487
ck(f(('bool', 'bytes30', 'bytes27'), D), T[bo, by, by])  # fixed case 35488
ck(f(('bool', 'bytes30', 'bytes28'), D), T[bo, by, by])  # fixed case 35489
ck(f(('bool', 'bytes30', 'bytes29'), D), T[bo, by, by])  # fixed case 35490
ck(f(('bool', 'bytes30', 'bytes3'), D), T[bo, by, by])  # fixed case 35491
ck(f(('bool', 'bytes30', 'bytes30'), D), T[bo, by, by])  # fixed case 35492
ck(f(('bool', 'bytes30', 'bytes31'), D), T[bo, by, by])  # fixed case 35493
ck(f(('bool', 'bytes30', 'bytes32'), D), T[bo, by, by])  # fixed case 35494
ck(f(('bool', 'bytes30', 'bytes4'), D), T[bo, by, by])  # fixed case 35495
ck(f(('bool', 'bytes30', 'bytes5'), D), T[bo, by, by])  # fixed case 35496
ck(f(('bool', 'bytes30', 'bytes6'), D), T[bo, by, by])  # fixed case 35497
ck(f(('bool', 'bytes30', 'bytes7'), D), T[bo, by, by])  # fixed case 35498
ck(f(('bool', 'bytes30', 'bytes8'), D), T[bo, by, by])  # fixed case 35499
ck(f(('bool', 'bytes30', 'bytes9'), D), T[bo, by, by])  # fixed case 35500
ck(f(('bool', 'bytes30', 'int104'), D), T[bo, by, i])  # fixed case 35501
ck(f(('bool', 'bytes30', 'int112'), D), T[bo, by, i])  # fixed case 35502
ck(f(('bool', 'bytes30', 'int120'), D), T[bo, by, i])  # fixed case 35503
ck(f(('bool', 'bytes30', 'int128'), D), T[bo, by, i])  # fixed case 35504
ck(f(('bool', 'bytes30', 'int136'), D), T[bo, by, i])  # fixed case 35505
ck(f(('bool', 'bytes30', 'int144'), D), T[bo, by, i])  # fixed case 35506
ck(f(('bool', 'bytes30', 'int152'), D), T[bo, by, i])  # fixed case 35507
ck(f(('bool', 'bytes30', 'int16'), D), T[bo, by, i])  # fixed case 35508
ck(f(('bool', 'bytes30', 'int160'), D), T[bo, by, i])  # fixed case 35509
ck(f(('bool', 'bytes30', 'int168'), D), T[bo, by, i])  # fixed case 35510
ck(f(('bool', 'bytes30', 'int176'), D), T[bo, by, i])  # fixed case 35511
ck(f(('bool', 'bytes30', 'int184'), D), T[bo, by, i])  # fixed case 35512
ck(f(('bool', 'bytes30', 'int192'), D), T[bo, by, i])  # fixed case 35513
ck(f(('bool', 'bytes30', 'int200'), D), T[bo, by, i])  # fixed case 35514
ck(f(('bool', 'bytes30', 'int208'), D), T[bo, by, i])  # fixed case 35515
ck(f(('bool', 'bytes30', 'int216'), D), T[bo, by, i])  # fixed case 35516
ck(f(('bool', 'bytes30', 'int224'), D), T[bo, by, i])  # fixed case 35517
ck(f(('bool', 'bytes30', 'int232'), D), T[bo, by, i])  # fixed case 35518
ck(f(('bool', 'bytes30', 'int24'), D), T[bo, by, i])  # fixed case 35519
ck(f(('bool', 'bytes30', 'int240'), D), T[bo, by, i])  # fixed case 35520
ck(f(('bool', 'bytes30', 'int248'), D), T[bo, by, i])  # fixed case 35521
ck(f(('bool', 'bytes30', 'int256'), D), T[bo, by, i])  # fixed case 35522
ck(f(('bool', 'bytes30', 'int32'), D), T[bo, by, i])  # fixed case 35523
ck(f(('bool', 'bytes30', 'int40'), D), T[bo, by, i])  # fixed case 35524
ck(f(('bool', 'bytes30', 'int48'), D), T[bo, by, i])  # fixed case 35525
ck(f(('bool', 'bytes30', 'int56'), D), T[bo, by, i])  # fixed case 35526
ck(f(('bool', 'bytes30', 'int64'), D), T[bo, by, i])  # fixed case 35527
ck(f(('bool', 'bytes30', 'int72'), D), T[bo, by, i])  # fixed case 35528
ck(f(('bool', 'bytes30', 'int8'), D), T[bo, by, i])  # fixed case 35529
ck(f(('bool', 'bytes30', 'int80'), D), T[bo, by, i])  # fixed case 35530
ck(f(('bool', 'bytes30', 'int88'), D), T[bo, by, i])  # fixed case 35531
ck(f(('bool', 'bytes30', 'int96'), D), T[bo, by, i])  # fixed case 35532
ck(f(('bool', 'bytes30', 'string'), D), T[bo, by, s])  # fixed case 35533
ck(f(('bool', 'bytes30', 'string[]'), D), T[bo, by, A])  # fixed case 35534
ck(f(('bool', 'bytes30', 'uint104'), D), T[bo, by, i])  # fixed case 35535
ck(f(('bool', 'bytes30', 'uint112'), D), T[bo, by, i])  # fixed case 35536
ck(f(('bool', 'bytes30', 'uint120'), D), T[bo, by, i])  # fixed case 35537
ck(f(('bool', 'bytes30', 'uint128'), D), T[bo, by, i])  # fixed case 35538
ck(f(('bool', 'bytes30', 'uint136'), D), T[bo, by, i])  # fixed case 35539
ck(f(('bool', 'bytes30', 'uint144'), D), T[bo, by, i])  # fixed case 35540
ck(f(('bool', 'bytes30', 'uint152'), D), T[bo, by, i])  # fixed case 35541
ck(f(('bool', 'bytes30', 'uint16'), D), T[bo, by, i])  # fixed case 35542
ck(f(('bool', 'bytes30', 'uint160'), D), T[bo, by, i])  # fixed case 35543
ck(f(('bool', 'bytes30', 'uint168'), D), T[bo, by, i])  # fixed case 35544
ck(f(('bool', 'bytes30', 'uint176'), D), T[bo, by, i])  # fixed case 35545
ck(f(('bool', 'bytes30', 'uint184'), D), T[bo, by, i])  # fixed case 35546
ck(f(('bool', 'bytes30', 'uint192'), D), T[bo, by, i])  # fixed case 35547
ck(f(('bool', 'bytes30', 'uint200'), D), T[bo, by, i])  # fixed case 35548
ck(f(('bool', 'bytes30', 'uint208'), D), T[bo, by, i])  # fixed case 35549
ck(f(('bool', 'bytes30', 'uint216'), D), T[bo, by, i])  # fixed case 35550
ck(f(('bool', 'bytes30', 'uint224'), D), T[bo, by, i])  # fixed case 35551
ck(f(('bool', 'bytes30', 'uint232'), D), T[bo, by, i])  # fixed case 35552
ck(f(('bool', 'bytes30', 'uint24'), D), T[bo, by, i])  # fixed case 35553
ck(f(('bool', 'bytes30', 'uint240'), D), T[bo, by, i])  # fixed case 35554
ck(f(('bool', 'bytes30', 'uint248'), D), T[bo, by, i])  # fixed case 35555
ck(f(('bool', 'bytes30', 'uint256'), D), T[bo, by, i])  # fixed case 35556
ck(f(('bool', 'bytes30', 'uint32'), D), T[bo, by, i])  # fixed case 35557
ck(f(('bool', 'bytes30', 'uint40'), D), T[bo, by, i])  # fixed case 35558
ck(f(('bool', 'bytes30', 'uint48'), D), T[bo, by, i])  # fixed case 35559
ck(f(('bool', 'bytes30', 'uint56'), D), T[bo, by, i])  # fixed case 35560
ck(f(('bool', 'bytes30', 'uint64'), D), T[bo, by, i])  # fixed case 35561
ck(f(('bool', 'bytes30', 'uint72'), D), T[bo, by, i])  # fixed case 35562
ck(f(('bool', 'bytes30', 'uint8'), D), T[bo, by, i])  # fixed case 35563
ck(f(('bool', 'bytes30', 'uint80'), D), T[bo, by, i])  # fixed case 35564
ck(f(('bool', 'bytes30', 'uint88'), D), T[bo, by, i])  # fixed case 35565
ck(f(('bool', 'bytes30', 'uint96'), D), T[bo, by, i])  # fixed case 35566
ck(f(('bool', 'bytes30', '?'), D), T[bo, by, A])  # fixed case 35567
ck(f(('bool', 'bytes31', 'address'), D), T[bo, by, HA])  # fixed case 35568
ck(f(('bool', 'bytes31', 'address[]'), D), T[bo, by, A])  # fixed case 35569
ck(f(('bool', 'bytes31', 'bool'), D), T[bo, by, bo])  # fixed case 35570
ck(f(('bool', 'bytes31', 'bool[]'), D), T[bo, by, A])  # fixed case 35571
ck(f(('bool', 'bytes31', 'bytes'), D), T[bo, by, by])  # fixed case 35572
ck(f(('bool', 'bytes31', 'bytes1'), D), T[bo, by, by])  # fixed case 35573
ck(f(('bool', 'bytes31', 'bytes10'), D), T[bo, by, by])  # fixed case 35574
ck(f(('bool', 'bytes31', 'bytes11'), D), T[bo, by, by])  # fixed case 35575
ck(f(('bool', 'bytes31', 'bytes12'), D), T[bo, by, by])  # fixed case 35576
ck(f(('bool', 'bytes31', 'bytes13'), D), T[bo, by, by])  # fixed case 35577
ck(f(('bool', 'bytes31', 'bytes14'), D), T[bo, by, by])  # fixed case 35578
ck(f(('bool', 'bytes31', 'bytes15'), D), T[bo, by, by])  # fixed case 35579
ck(f(('bool', 'bytes31', 'bytes16'), D), T[bo, by, by])  # fixed case 35580
ck(f(('bool', 'bytes31', 'bytes17'), D), T[bo, by, by])  # fixed case 35581
ck(f(('bool', 'bytes31', 'bytes18'), D), T[bo, by, by])  # fixed case 35582
ck(f(('bool', 'bytes31', 'bytes19'), D), T[bo, by, by])  # fixed case 35583
ck(f(('bool', 'bytes31', 'bytes2'), D), T[bo, by, by])  # fixed case 35584
ck(f(('bool', 'bytes31', 'bytes20'), D), T[bo, by, by])  # fixed case 35585
ck(f(('bool', 'bytes31', 'bytes21'), D), T[bo, by, by])  # fixed case 35586
ck(f(('bool', 'bytes31', 'bytes22'), D), T[bo, by, by])  # fixed case 35587
ck(f(('bool', 'bytes31', 'bytes23'), D), T[bo, by, by])  # fixed case 35588
ck(f(('bool', 'bytes31', 'bytes24'), D), T[bo, by, by])  # fixed case 35589
ck(f(('bool', 'bytes31', 'bytes25'), D), T[bo, by, by])  # fixed case 35590
ck(f(('bool', 'bytes31', 'bytes26'), D), T[bo, by, by])  # fixed case 35591
ck(f(('bool', 'bytes31', 'bytes27'), D), T[bo, by, by])  # fixed case 35592
ck(f(('bool', 'bytes31', 'bytes28'), D), T[bo, by, by])  # fixed case 35593
ck(f(('bool', 'bytes31', 'bytes29'), D), T[bo, by, by])  # fixed case 35594
ck(f(('bool', 'bytes31', 'bytes3'), D), T[bo, by, by])  # fixed case 35595
ck(f(('bool', 'bytes31', 'bytes30'), D), T[bo, by, by])  # fixed case 35596
ck(f(('bool', 'bytes31', 'bytes31'), D), T[bo, by, by])  # fixed case 35597
ck(f(('bool', 'bytes31', 'bytes32'), D), T[bo, by, by])  # fixed case 35598
ck(f(('bool', 'bytes31', 'bytes4'), D), T[bo, by, by])  # fixed case 35599
ck(f(('bool', 'bytes31', 'bytes5'), D), T[bo, by, by])  # fixed case 35600
ck(f(('bool', 'bytes31', 'bytes6'), D), T[bo, by, by])  # fixed case 35601
ck(f(('bool', 'bytes31', 'bytes7'), D), T[bo, by, by])  # fixed case 35602
ck(f(('bool', 'bytes31', 'bytes8'), D), T[bo, by, by])  # fixed case 35603
ck(f(('bool', 'bytes31', 'bytes9'), D), T[bo, by, by])  # fixed case 35604
ck(f(('bool', 'bytes31', 'int104'), D), T[bo, by, i])  # fixed case 35605
ck(f(('bool', 'bytes31', 'int112'), D), T[bo, by, i])  # fixed case 35606
ck(f(('bool', 'bytes31', 'int120'), D), T[bo, by, i])  # fixed case 35607
ck(f(('bool', 'bytes31', 'int128'), D), T[bo, by, i])  # fixed case 35608
ck(f(('bool', 'bytes31', 'int136'), D), T[bo, by, i])  # fixed case 35609
ck(f(('bool', 'bytes31', 'int144'), D), T[bo, by, i])  # fixed case 35610
ck(f(('bool', 'bytes31', 'int152'), D), T[bo, by, i])  # fixed case 35611
ck(f(('bool', 'bytes31', 'int16'), D), T[bo, by, i])  # fixed case 35612
ck(f(('bool', 'bytes31', 'int160'), D), T[bo, by, i])  # fixed case 35613
ck(f(('bool', 'bytes31', 'int168'), D), T[bo, by, i])  # fixed case 35614
ck(f(('bool', 'bytes31', 'int176'), D), T[bo, by, i])  # fixed case 35615
ck(f(('bool', 'bytes31', 'int184'), D), T[bo, by, i])  # fixed case 35616
ck(f(('bool', 'bytes31', 'int192'), D), T[bo, by, i])  # fixed case 35617
ck(f(('bool', 'bytes31', 'int200'), D), T[bo, by, i])  # fixed case 35618
ck(f(('bool', 'bytes31', 'int208'), D), T[bo, by, i])  # fixed case 35619
ck(f(('bool', 'bytes31', 'int216'), D), T[bo, by, i])  # fixed case 35620
ck(f(('bool', 'bytes31', 'int224'), D), T[bo, by, i])  # fixed case 35621
ck(f(('bool', 'bytes31', 'int232'), D), T[bo, by, i])  # fixed case 35622
ck(f(('bool', 'bytes31', 'int24'), D), T[bo, by, i])  # fixed case 35623
ck(f(('bool', 'bytes31', 'int240'), D), T[bo, by, i])  # fixed case 35624
ck(f(('bool', 'bytes31', 'int248'), D), T[bo, by, i])  # fixed case 35625
ck(f(('bool', 'bytes31', 'int256'), D), T[bo, by, i])  # fixed case 35626
ck(f(('bool', 'bytes31', 'int32'), D), T[bo, by, i])  # fixed case 35627
ck(f(('bool', 'bytes31', 'int40'), D), T[bo, by, i])  # fixed case 35628
ck(f(('bool', 'bytes31', 'int48'), D), T[bo, by, i])  # fixed case 35629
ck(f(('bool', 'bytes31', 'int56'), D), T[bo, by, i])  # fixed case 35630
ck(f(('bool', 'bytes31', 'int64'), D), T[bo, by, i])  # fixed case 35631
ck(f(('bool', 'bytes31', 'int72'), D), T[bo, by, i])  # fixed case 35632
ck(f(('bool', 'bytes31', 'int8'), D), T[bo, by, i])  # fixed case 35633
ck(f(('bool', 'bytes31', 'int80'), D), T[bo, by, i])  # fixed case 35634
ck(f(('bool', 'bytes31', 'int88'), D), T[bo, by, i])  # fixed case 35635
ck(f(('bool', 'bytes31', 'int96'), D), T[bo, by, i])  # fixed case 35636
ck(f(('bool', 'bytes31', 'string'), D), T[bo, by, s])  # fixed case 35637
ck(f(('bool', 'bytes31', 'string[]'), D), T[bo, by, A])  # fixed case 35638
ck(f(('bool', 'bytes31', 'uint104'), D), T[bo, by, i])  # fixed case 35639
ck(f(('bool', 'bytes31', 'uint112'), D), T[bo, by, i])  # fixed case 35640
ck(f(('bool', 'bytes31', 'uint120'), D), T[bo, by, i])  # fixed case 35641
ck(f(('bool', 'bytes31', 'uint128'), D), T[bo, by, i])  # fixed case 35642
ck(f(('bool', 'bytes31', 'uint136'), D), T[bo, by, i])  # fixed case 35643
ck(f(('bool', 'bytes31', 'uint144'), D), T[bo, by, i])  # fixed case 35644
ck(f(('bool', 'bytes31', 'uint152'), D), T[bo, by, i])  # fixed case 35645
ck(f(('bool', 'bytes31', 'uint16'), D), T[bo, by, i])  # fixed case 35646
ck(f(('bool', 'bytes31', 'uint160'), D), T[bo, by, i])  # fixed case 35647
ck(f(('bool', 'bytes31', 'uint168'), D), T[bo, by, i])  # fixed case 35648
ck(f(('bool', 'bytes31', 'uint176'), D), T[bo, by, i])  # fixed case 35649
ck(f(('bool', 'bytes31', 'uint184'), D), T[bo, by, i])  # fixed case 35650
ck(f(('bool', 'bytes31', 'uint192'), D), T[bo, by, i])  # fixed case 35651
ck(f(('bool', 'bytes31', 'uint200'), D), T[bo, by, i])  # fixed case 35652
ck(f(('bool', 'bytes31', 'uint208'), D), T[bo, by, i])  # fixed case 35653
ck(f(('bool', 'bytes31', 'uint216'), D), T[bo, by, i])  # fixed case 35654
ck(f(('bool', 'bytes31', 'uint224'), D), T[bo, by, i])  # fixed case 35655
ck(f(('bool', 'bytes31', 'uint232'), D), T[bo, by, i])  # fixed case 35656
ck(f(('bool', 'bytes31', 'uint24'), D), T[bo, by, i])  # fixed case 35657
ck(f(('bool', 'bytes31', 'uint240'), D), T[bo, by, i])  # fixed case 35658
ck(f(('bool', 'bytes31', 'uint248'), D), T[bo, by, i])  # fixed case 35659
ck(f(('bool', 'bytes31', 'uint256'), D), T[bo, by, i])  # fixed case 35660
ck(f(('bool', 'bytes31', 'uint32'), D), T[bo, by, i])  # fixed case 35661
ck(f(('bool', 'bytes31', 'uint40'), D), T[bo, by, i])  # fixed case 35662
ck(f(('bool', 'bytes31', 'uint48'), D), T[bo, by, i])  # fixed case 35663
ck(f(('bool', 'bytes31', 'uint56'), D), T[bo, by, i])  # fixed case 35664
ck(f(('bool', 'bytes31', 'uint64'), D), T[bo, by, i])  # fixed case 35665
ck(f(('bool', 'bytes31', 'uint72'), D), T[bo, by, i])  # fixed case 35666
ck(f(('bool', 'bytes31', 'uint8'), D), T[bo, by, i])  # fixed case 35667
ck(f(('bool', 'bytes31', 'uint80'), D), T[bo, by, i])  # fixed case 35668
ck(f(('bool', 'bytes31', 'uint88'), D), T[bo, by, i])  # fixed case 35669
ck(f(('bool', 'bytes31', 'uint96'), D), T[bo, by, i])  # fixed case 35670
ck(f(('bool', 'bytes31', '?'), D), T[bo, by, A])  # fixed case 35671
ck(f(('bool', 'bytes32', 'address'), D), T[bo, by, HA])  # fixed case 35672
ck(f(('bool', 'bytes32', 'address[]'), D), T[bo, by, A])  # fixed case 35673
ck(f(('bool', 'bytes32', 'bool'), D), T[bo, by, bo])  # fixed case 35674
ck(f(('bool', 'bytes32', 'bool[]'), D), T[bo, by, A])  # fixed case 35675
ck(f(('bool', 'bytes32', 'bytes'), D), T[bo, by, by])  # fixed case 35676
ck(f(('bool', 'bytes32', 'bytes1'), D), T[bo, by, by])  # fixed case 35677
ck(f(('bool', 'bytes32', 'bytes10'), D), T[bo, by, by])  # fixed case 35678
ck(f(('bool', 'bytes32', 'bytes11'), D), T[bo, by, by])  # fixed case 35679
ck(f(('bool', 'bytes32', 'bytes12'), D), T[bo, by, by])  # fixed case 35680
ck(f(('bool', 'bytes32', 'bytes13'), D), T[bo, by, by])  # fixed case 35681
ck(f(('bool', 'bytes32', 'bytes14'), D), T[bo, by, by])  # fixed case 35682
ck(f(('bool', 'bytes32', 'bytes15'), D), T[bo, by, by])  # fixed case 35683
ck(f(('bool', 'bytes32', 'bytes16'), D), T[bo, by, by])  # fixed case 35684
ck(f(('bool', 'bytes32', 'bytes17'), D), T[bo, by, by])  # fixed case 35685
ck(f(('bool', 'bytes32', 'bytes18'), D), T[bo, by, by])  # fixed case 35686
ck(f(('bool', 'bytes32', 'bytes19'), D), T[bo, by, by])  # fixed case 35687
ck(f(('bool', 'bytes32', 'bytes2'), D), T[bo, by, by])  # fixed case 35688
ck(f(('bool', 'bytes32', 'bytes20'), D), T[bo, by, by])  # fixed case 35689
ck(f(('bool', 'bytes32', 'bytes21'), D), T[bo, by, by])  # fixed case 35690
ck(f(('bool', 'bytes32', 'bytes22'), D), T[bo, by, by])  # fixed case 35691
ck(f(('bool', 'bytes32', 'bytes23'), D), T[bo, by, by])  # fixed case 35692
ck(f(('bool', 'bytes32', 'bytes24'), D), T[bo, by, by])  # fixed case 35693
ck(f(('bool', 'bytes32', 'bytes25'), D), T[bo, by, by])  # fixed case 35694
ck(f(('bool', 'bytes32', 'bytes26'), D), T[bo, by, by])  # fixed case 35695
ck(f(('bool', 'bytes32', 'bytes27'), D), T[bo, by, by])  # fixed case 35696
ck(f(('bool', 'bytes32', 'bytes28'), D), T[bo, by, by])  # fixed case 35697
ck(f(('bool', 'bytes32', 'bytes29'), D), T[bo, by, by])  # fixed case 35698
ck(f(('bool', 'bytes32', 'bytes3'), D), T[bo, by, by])  # fixed case 35699
ck(f(('bool', 'bytes32', 'bytes30'), D), T[bo, by, by])  # fixed case 35700
ck(f(('bool', 'bytes32', 'bytes31'), D), T[bo, by, by])  # fixed case 35701
ck(f(('bool', 'bytes32', 'bytes32'), D), T[bo, by, by])  # fixed case 35702
ck(f(('bool', 'bytes32', 'bytes4'), D), T[bo, by, by])  # fixed case 35703
ck(f(('bool', 'bytes32', 'bytes5'), D), T[bo, by, by])  # fixed case 35704
ck(f(('bool', 'bytes32', 'bytes6'), D), T[bo, by, by])  # fixed case 35705
ck(f(('bool', 'bytes32', 'bytes7'), D), T[bo, by, by])  # fixed case 35706
ck(f(('bool', 'bytes32', 'bytes8'), D), T[bo, by, by])  # fixed case 35707
ck(f(('bool', 'bytes32', 'bytes9'), D), T[bo, by, by])  # fixed case 35708
ck(f(('bool', 'bytes32', 'int104'), D), T[bo, by, i])  # fixed case 35709
ck(f(('bool', 'bytes32', 'int112'), D), T[bo, by, i])  # fixed case 35710
ck(f(('bool', 'bytes32', 'int120'), D), T[bo, by, i])  # fixed case 35711
ck(f(('bool', 'bytes32', 'int128'), D), T[bo, by, i])  # fixed case 35712
ck(f(('bool', 'bytes32', 'int136'), D), T[bo, by, i])  # fixed case 35713
ck(f(('bool', 'bytes32', 'int144'), D), T[bo, by, i])  # fixed case 35714
ck(f(('bool', 'bytes32', 'int152'), D), T[bo, by, i])  # fixed case 35715
ck(f(('bool', 'bytes32', 'int16'), D), T[bo, by, i])  # fixed case 35716
ck(f(('bool', 'bytes32', 'int160'), D), T[bo, by, i])  # fixed case 35717
ck(f(('bool', 'bytes32', 'int168'), D), T[bo, by, i])  # fixed case 35718
ck(f(('bool', 'bytes32', 'int176'), D), T[bo, by, i])  # fixed case 35719
ck(f(('bool', 'bytes32', 'int184'), D), T[bo, by, i])  # fixed case 35720
ck(f(('bool', 'bytes32', 'int192'), D), T[bo, by, i])  # fixed case 35721
ck(f(('bool', 'bytes32', 'int200'), D), T[bo, by, i])  # fixed case 35722
ck(f(('bool', 'bytes32', 'int208'), D), T[bo, by, i])  # fixed case 35723
ck(f(('bool', 'bytes32', 'int216'), D), T[bo, by, i])  # fixed case 35724
ck(f(('bool', 'bytes32', 'int224'), D), T[bo, by, i])  # fixed case 35725
ck(f(('bool', 'bytes32', 'int232'), D), T[bo, by, i])  # fixed case 35726
ck(f(('bool', 'bytes32', 'int24'), D), T[bo, by, i])  # fixed case 35727
ck(f(('bool', 'bytes32', 'int240'), D), T[bo, by, i])  # fixed case 35728
ck(f(('bool', 'bytes32', 'int248'), D), T[bo, by, i])  # fixed case 35729
ck(f(('bool', 'bytes32', 'int256'), D), T[bo, by, i])  # fixed case 35730
ck(f(('bool', 'bytes32', 'int32'), D), T[bo, by, i])  # fixed case 35731
ck(f(('bool', 'bytes32', 'int40'), D), T[bo, by, i])  # fixed case 35732
ck(f(('bool', 'bytes32', 'int48'), D), T[bo, by, i])  # fixed case 35733
ck(f(('bool', 'bytes32', 'int56'), D), T[bo, by, i])  # fixed case 35734
ck(f(('bool', 'bytes32', 'int64'), D), T[bo, by, i])  # fixed case 35735
ck(f(('bool', 'bytes32', 'int72'), D), T[bo, by, i])  # fixed case 35736
ck(f(('bool', 'bytes32', 'int8'), D), T[bo, by, i])  # fixed case 35737
ck(f(('bool', 'bytes32', 'int80'), D), T[bo, by, i])  # fixed case 35738
ck(f(('bool', 'bytes32', 'int88'), D), T[bo, by, i])  # fixed case 35739
ck(f(('bool', 'bytes32', 'int96'), D), T[bo, by, i])  # fixed case 35740
ck(f(('bool', 'bytes32', 'string'), D), T[bo, by, s])  # fixed case 35741
ck(f(('bool', 'bytes32', 'string[]'), D), T[bo, by, A])  # fixed case 35742
ck(f(('bool', 'bytes32', 'uint104'), D), T[bo, by, i])  # fixed case 35743
ck(f(('bool', 'bytes32', 'uint112'), D), T[bo, by, i])  # fixed case 35744
ck(f(('bool', 'bytes32', 'uint120'), D), T[bo, by, i])  # fixed case 35745
ck(f(('bool', 'bytes32', 'uint128'), D), T[bo, by, i])  # fixed case 35746
ck(f(('bool', 'bytes32', 'uint136'), D), T[bo, by, i])  # fixed case 35747
ck(f(('bool', 'bytes32', 'uint144'), D), T[bo, by, i])  # fixed case 35748
ck(f(('bool', 'bytes32', 'uint152'), D), T[bo, by, i])  # fixed case 35749
ck(f(('bool', 'bytes32', 'uint16'), D), T[bo, by, i])  # fixed case 35750
ck(f(('bool', 'bytes32', 'uint160'), D), T[bo, by, i])  # fixed case 35751
ck(f(('bool', 'bytes32', 'uint168'), D), T[bo, by, i])  # fixed case 35752
ck(f(('bool', 'bytes32', 'uint176'), D), T[bo, by, i])  # fixed case 35753
ck(f(('bool', 'bytes32', 'uint184'), D), T[bo, by, i])  # fixed case 35754
ck(f(('bool', 'bytes32', 'uint192'), D), T[bo, by, i])  # fixed case 35755
ck(f(('bool', 'bytes32', 'uint200'), D), T[bo, by, i])  # fixed case 35756
ck(f(('bool', 'bytes32', 'uint208'), D), T[bo, by, i])  # fixed case 35757
ck(f(('bool', 'bytes32', 'uint216'), D), T[bo, by, i])  # fixed case 35758
ck(f(('bool', 'bytes32', 'uint224'), D), T[bo, by, i])  # fixed case 35759
ck(f(('bool', 'bytes32', 'uint232'), D), T[bo, by, i])  # fixed case 35760
ck(f(('bool', 'bytes32', 'uint24'), D), T[bo, by, i])  # fixed case 35761
ck(f(('bool', 'bytes32', 'uint240'), D), T[bo, by, i])  # fixed case 35762
ck(f(('bool', 'bytes32', 'uint248'), D), T[bo, by, i])  # fixed case 35763
ck(f(('bool', 'bytes32', 'uint256'), D), T[bo, by, i])  # fixed case 35764
ck(f(('bool', 'bytes32', 'uint32'), D), T[bo, by, i])  # fixed case 35765
ck(f(('bool', 'bytes32', 'uint40'), D), T[bo, by, i])  # fixed case 35766
ck(f(('bool', 'bytes32', 'uint48'), D), T[bo, by, i])  # fixed case 35767
ck(f(('bool', 'bytes32', 'uint56'), D), T[bo, by, i])  # fixed case 35768
ck(f(('bool', 'bytes32', 'uint64'), D), T[bo, by, i])  # fixed case 35769
ck(f(('bool', 'bytes32', 'uint72'), D), T[bo, by, i])  # fixed case 35770
ck(f(('bool', 'bytes32', 'uint8'), D), T[bo, by, i])  # fixed case 35771
ck(f(('bool', 'bytes32', 'uint80'), D), T[bo, by, i])  # fixed case 35772
ck(f(('bool', 'bytes32', 'uint88'), D), T[bo, by, i])  # fixed case 35773
ck(f(('bool', 'bytes32', 'uint96'), D), T[bo, by, i])  # fixed case 35774
ck(f(('bool', 'bytes32', '?'), D), T[bo, by, A])  # fixed case 35775
ck(f(('bool', 'bytes4', 'address'), D), T[bo, by, HA])  # fixed case 35776
ck(f(('bool', 'bytes4', 'address[]'), D), T[bo, by, A])  # fixed case 35777
ck(f(('bool', 'bytes4', 'bool'), D), T[bo, by, bo])  # fixed case 35778
ck(f(('bool', 'bytes4', 'bool[]'), D), T[bo, by, A])  # fixed case 35779
ck(f(('bool', 'bytes4', 'bytes'), D), T[bo, by, by])  # fixed case 35780
ck(f(('bool', 'bytes4', 'bytes1'), D), T[bo, by, by])  # fixed case 35781
ck(f(('bool', 'bytes4', 'bytes10'), D), T[bo, by, by])  # fixed case 35782
ck(f(('bool', 'bytes4', 'bytes11'), D), T[bo, by, by])  # fixed case 35783
ck(f(('bool', 'bytes4', 'bytes12'), D), T[bo, by, by])  # fixed case 35784
ck(f(('bool', 'bytes4', 'bytes13'), D), T[bo, by, by])  # fixed case 35785
ck(f(('bool', 'bytes4', 'bytes14'), D), T[bo, by, by])  # fixed case 35786
ck(f(('bool', 'bytes4', 'bytes15'), D), T[bo, by, by])  # fixed case 35787
ck(f(('bool', 'bytes4', 'bytes16'), D), T[bo, by, by])  # fixed case 35788
ck(f(('bool', 'bytes4', 'bytes17'), D), T[bo, by, by])  # fixed case 35789
ck(f(('bool', 'bytes4', 'bytes18'), D), T[bo, by, by])  # fixed case 35790
ck(f(('bool', 'bytes4', 'bytes19'), D), T[bo, by, by])  # fixed case 35791
ck(f(('bool', 'bytes4', 'bytes2'), D), T[bo, by, by])  # fixed case 35792
ck(f(('bool', 'bytes4', 'bytes20'), D), T[bo, by, by])  # fixed case 35793
ck(f(('bool', 'bytes4', 'bytes21'), D), T[bo, by, by])  # fixed case 35794
ck(f(('bool', 'bytes4', 'bytes22'), D), T[bo, by, by])  # fixed case 35795
ck(f(('bool', 'bytes4', 'bytes23'), D), T[bo, by, by])  # fixed case 35796
ck(f(('bool', 'bytes4', 'bytes24'), D), T[bo, by, by])  # fixed case 35797
ck(f(('bool', 'bytes4', 'bytes25'), D), T[bo, by, by])  # fixed case 35798
ck(f(('bool', 'bytes4', 'bytes26'), D), T[bo, by, by])  # fixed case 35799
ck(f(('bool', 'bytes4', 'bytes27'), D), T[bo, by, by])  # fixed case 35800
ck(f(('bool', 'bytes4', 'bytes28'), D), T[bo, by, by])  # fixed case 35801
ck(f(('bool', 'bytes4', 'bytes29'), D), T[bo, by, by])  # fixed case 35802
ck(f(('bool', 'bytes4', 'bytes3'), D), T[bo, by, by])  # fixed case 35803
ck(f(('bool', 'bytes4', 'bytes30'), D), T[bo, by, by])  # fixed case 35804
ck(f(('bool', 'bytes4', 'bytes31'), D), T[bo, by, by])  # fixed case 35805
ck(f(('bool', 'bytes4', 'bytes32'), D), T[bo, by, by])  # fixed case 35806
ck(f(('bool', 'bytes4', 'bytes4'), D), T[bo, by, by])  # fixed case 35807
ck(f(('bool', 'bytes4', 'bytes5'), D), T[bo, by, by])  # fixed case 35808
ck(f(('bool', 'bytes4', 'bytes6'), D), T[bo, by, by])  # fixed case 35809
ck(f(('bool', 'bytes4', 'bytes7'), D), T[bo, by, by])  # fixed case 35810
ck(f(('bool', 'bytes4', 'bytes8'), D), T[bo, by, by])  # fixed case 35811
ck(f(('bool', 'bytes4', 'bytes9'), D), T[bo, by, by])  # fixed case 35812
ck(f(('bool', 'bytes4', 'int104'), D), T[bo, by, i])  # fixed case 35813
ck(f(('bool', 'bytes4', 'int112'), D), T[bo, by, i])  # fixed case 35814
ck(f(('bool', 'bytes4', 'int120'), D), T[bo, by, i])  # fixed case 35815
ck(f(('bool', 'bytes4', 'int128'), D), T[bo, by, i])  # fixed case 35816
ck(f(('bool', 'bytes4', 'int136'), D), T[bo, by, i])  # fixed case 35817
ck(f(('bool', 'bytes4', 'int144'), D), T[bo, by, i])  # fixed case 35818
ck(f(('bool', 'bytes4', 'int152'), D), T[bo, by, i])  # fixed case 35819
ck(f(('bool', 'bytes4', 'int16'), D), T[bo, by, i])  # fixed case 35820
ck(f(('bool', 'bytes4', 'int160'), D), T[bo, by, i])  # fixed case 35821
ck(f(('bool', 'bytes4', 'int168'), D), T[bo, by, i])  # fixed case 35822
ck(f(('bool', 'bytes4', 'int176'), D), T[bo, by, i])  # fixed case 35823
ck(f(('bool', 'bytes4', 'int184'), D), T[bo, by, i])  # fixed case 35824
ck(f(('bool', 'bytes4', 'int192'), D), T[bo, by, i])  # fixed case 35825
ck(f(('bool', 'bytes4', 'int200'), D), T[bo, by, i])  # fixed case 35826
ck(f(('bool', 'bytes4', 'int208'), D), T[bo, by, i])  # fixed case 35827
ck(f(('bool', 'bytes4', 'int216'), D), T[bo, by, i])  # fixed case 35828
ck(f(('bool', 'bytes4', 'int224'), D), T[bo, by, i])  # fixed case 35829
ck(f(('bool', 'bytes4', 'int232'), D), T[bo, by, i])  # fixed case 35830
ck(f(('bool', 'bytes4', 'int24'), D), T[bo, by, i])  # fixed case 35831
ck(f(('bool', 'bytes4', 'int240'), D), T[bo, by, i])  # fixed case 35832
ck(f(('bool', 'bytes4', 'int248'), D), T[bo, by, i])  # fixed case 35833
ck(f(('bool', 'bytes4', 'int256'), D), T[bo, by, i])  # fixed case 35834
ck(f(('bool', 'bytes4', 'int32'), D), T[bo, by, i])  # fixed case 35835
ck(f(('bool', 'bytes4', 'int40'), D), T[bo, by, i])  # fixed case 35836
ck(f(('bool', 'bytes4', 'int48'), D), T[bo, by, i])  # fixed case 35837
ck(f(('bool', 'bytes4', 'int56'), D), T[bo, by, i])  # fixed case 35838
ck(f(('bool', 'bytes4', 'int64'), D), T[bo, by, i])  # fixed case 35839
ck(f(('bool', 'bytes4', 'int72'), D), T[bo, by, i])  # fixed case 35840
ck(f(('bool', 'bytes4', 'int8'), D), T[bo, by, i])  # fixed case 35841
ck(f(('bool', 'bytes4', 'int80'), D), T[bo, by, i])  # fixed case 35842
ck(f(('bool', 'bytes4', 'int88'), D), T[bo, by, i])  # fixed case 35843
ck(f(('bool', 'bytes4', 'int96'), D), T[bo, by, i])  # fixed case 35844
ck(f(('bool', 'bytes4', 'string'), D), T[bo, by, s])  # fixed case 35845
ck(f(('bool', 'bytes4', 'string[]'), D), T[bo, by, A])  # fixed case 35846
ck(f(('bool', 'bytes4', 'uint104'), D), T[bo, by, i])  # fixed case 35847
ck(f(('bool', 'bytes4', 'uint112'), D), T[bo, by, i])  # fixed case 35848
ck(f(('bool', 'bytes4', 'uint120'), D), T[bo, by, i])  # fixed case 35849
ck(f(('bool', 'bytes4', 'uint128'), D), T[bo, by, i])  # fixed case 35850
ck(f(('bool', 'bytes4', 'uint136'), D), T[bo, by, i])  # fixed case 35851
ck(f(('bool', 'bytes4', 'uint144'), D), T[bo, by, i])  # fixed case 35852
ck(f(('bool', 'bytes4', 'uint152'), D), T[bo, by, i])  # fixed case 35853
ck(f(('bool', 'bytes4', 'uint16'), D), T[bo, by, i])  # fixed case 35854
ck(f(('bool', 'bytes4', 'uint160'), D), T[bo, by, i])  # fixed case 35855
ck(f(('bool', 'bytes4', 'uint168'), D), T[bo, by, i])  # fixed case 35856
ck(f(('bool', 'bytes4', 'uint176'), D), T[bo, by, i])  # fixed case 35857
ck(f(('bool', 'bytes4', 'uint184'), D), T[bo, by, i])  # fixed case 35858
ck(f(('bool', 'bytes4', 'uint192'), D), T[bo, by, i])  # fixed case 35859
ck(f(('bool', 'bytes4', 'uint200'), D), T[bo, by, i])  # fixed case 35860
ck(f(('bool', 'bytes4', 'uint208'), D), T[bo, by, i])  # fixed case 35861
ck(f(('bool', 'bytes4', 'uint216'), D), T[bo, by, i])  # fixed case 35862
ck(f(('bool', 'bytes4', 'uint224'), D), T[bo, by, i])  # fixed case 35863
ck(f(('bool', 'bytes4', 'uint232'), D), T[bo, by, i])  # fixed case 35864
ck(f(('bool', 'bytes4', 'uint24'), D), T[bo, by, i])  # fixed case 35865
ck(f(('bool', 'bytes4', 'uint240'), D), T[bo, by, i])  # fixed case 35866
ck(f(('bool', 'bytes4', 'uint248'), D), T[bo, by, i])  # fixed case 35867
ck(f(('bool', 'bytes4', 'uint256'), D), T[bo, by, i])  # fixed case 35868
ck(f(('bool', 'bytes4', 'uint32'), D), T[bo, by, i])  # fixed case 35869
ck(f(('bool', 'bytes4', 'uint40'), D), T[bo, by, i])  # fixed case 35870
ck(f(('bool', 'bytes4', 'uint48'), D), T[bo, by, i])  # fixed case 35871
ck(f(('bool', 'bytes4', 'uint56'), D), T[bo, by, i])  # fixed case 35872
ck(f(('bool', 'bytes4', 'uint64'), D), T[bo, by, i])  # fixed case 35873
ck(f(('bool', 'bytes4', 'uint72'), D), T[bo, by, i])  # fixed case 35874
ck(f(('bool', 'bytes4', 'uint8'), D), T[bo, by, i])  # fixed case 35875
ck(f(('bool', 'bytes4', 'uint80'), D), T[bo, by, i])  # fixed case 35876
ck(f(('bool', 'bytes4', 'uint88'), D), T[bo, by, i])  # fixed case 35877
ck(f(('bool', 'bytes4', 'uint96'), D), T[bo, by, i])  # fixed case 35878
ck(f(('bool', 'bytes4', '?'), D), T[bo, by, A])  # fixed case 35879
ck(f(('bool', 'bytes5', 'address'), D), T[bo, by, HA])  # fixed case 35880
ck(f(('bool', 'bytes5', 'address[]'), D), T[bo, by, A])  # fixed case 35881
ck(f(('bool', 'bytes5', 'bool'), D), T[bo, by, bo])  # fixed case 35882
ck(f(('bool', 'bytes5', 'bool[]'), D), T[bo, by, A])  # fixed case 35883
ck(f(('bool', 'bytes5', 'bytes'), D), T[bo, by, by])  # fixed case 35884
ck(f(('bool', 'bytes5', 'bytes1'), D), T[bo, by, by])  # fixed case 35885
ck(f(('bool', 'bytes5', 'bytes10'), D), T[bo, by, by])  # fixed case 35886
ck(f(('bool', 'bytes5', 'bytes11'), D), T[bo, by, by])  # fixed case 35887
ck(f(('bool', 'bytes5', 'bytes12'), D), T[bo, by, by])  # fixed case 35888
ck(f(('bool', 'bytes5', 'bytes13'), D), T[bo, by, by])  # fixed case 35889
ck(f(('bool', 'bytes5', 'bytes14'), D), T[bo, by, by])  # fixed case 35890
ck(f(('bool', 'bytes5', 'bytes15'), D), T[bo, by, by])  # fixed case 35891
ck(f(('bool', 'bytes5', 'bytes16'), D), T[bo, by, by])  # fixed case 35892
ck(f(('bool', 'bytes5', 'bytes17'), D), T[bo, by, by])  # fixed case 35893
ck(f(('bool', 'bytes5', 'bytes18'), D), T[bo, by, by])  # fixed case 35894
ck(f(('bool', 'bytes5', 'bytes19'), D), T[bo, by, by])  # fixed case 35895
ck(f(('bool', 'bytes5', 'bytes2'), D), T[bo, by, by])  # fixed case 35896
ck(f(('bool', 'bytes5', 'bytes20'), D), T[bo, by, by])  # fixed case 35897
ck(f(('bool', 'bytes5', 'bytes21'), D), T[bo, by, by])  # fixed case 35898
ck(f(('bool', 'bytes5', 'bytes22'), D), T[bo, by, by])  # fixed case 35899
ck(f(('bool', 'bytes5', 'bytes23'), D), T[bo, by, by])  # fixed case 35900
ck(f(('bool', 'bytes5', 'bytes24'), D), T[bo, by, by])  # fixed case 35901
ck(f(('bool', 'bytes5', 'bytes25'), D), T[bo, by, by])  # fixed case 35902
ck(f(('bool', 'bytes5', 'bytes26'), D), T[bo, by, by])  # fixed case 35903
ck(f(('bool', 'bytes5', 'bytes27'), D), T[bo, by, by])  # fixed case 35904
ck(f(('bool', 'bytes5', 'bytes28'), D), T[bo, by, by])  # fixed case 35905
ck(f(('bool', 'bytes5', 'bytes29'), D), T[bo, by, by])  # fixed case 35906
ck(f(('bool', 'bytes5', 'bytes3'), D), T[bo, by, by])  # fixed case 35907
ck(f(('bool', 'bytes5', 'bytes30'), D), T[bo, by, by])  # fixed case 35908
ck(f(('bool', 'bytes5', 'bytes31'), D), T[bo, by, by])  # fixed case 35909
ck(f(('bool', 'bytes5', 'bytes32'), D), T[bo, by, by])  # fixed case 35910
ck(f(('bool', 'bytes5', 'bytes4'), D), T[bo, by, by])  # fixed case 35911
ck(f(('bool', 'bytes5', 'bytes5'), D), T[bo, by, by])  # fixed case 35912
ck(f(('bool', 'bytes5', 'bytes6'), D), T[bo, by, by])  # fixed case 35913
ck(f(('bool', 'bytes5', 'bytes7'), D), T[bo, by, by])  # fixed case 35914
ck(f(('bool', 'bytes5', 'bytes8'), D), T[bo, by, by])  # fixed case 35915
ck(f(('bool', 'bytes5', 'bytes9'), D), T[bo, by, by])  # fixed case 35916
ck(f(('bool', 'bytes5', 'int104'), D), T[bo, by, i])  # fixed case 35917
ck(f(('bool', 'bytes5', 'int112'), D), T[bo, by, i])  # fixed case 35918
ck(f(('bool', 'bytes5', 'int120'), D), T[bo, by, i])  # fixed case 35919
ck(f(('bool', 'bytes5', 'int128'), D), T[bo, by, i])  # fixed case 35920
ck(f(('bool', 'bytes5', 'int136'), D), T[bo, by, i])  # fixed case 35921
ck(f(('bool', 'bytes5', 'int144'), D), T[bo, by, i])  # fixed case 35922
ck(f(('bool', 'bytes5', 'int152'), D), T[bo, by, i])  # fixed case 35923
ck(f(('bool', 'bytes5', 'int16'), D), T[bo, by, i])  # fixed case 35924
ck(f(('bool', 'bytes5', 'int160'), D), T[bo, by, i])  # fixed case 35925
ck(f(('bool', 'bytes5', 'int168'), D), T[bo, by, i])  # fixed case 35926
ck(f(('bool', 'bytes5', 'int176'), D), T[bo, by, i])  # fixed case 35927
ck(f(('bool', 'bytes5', 'int184'), D), T[bo, by, i])  # fixed case 35928
ck(f(('bool', 'bytes5', 'int192'), D), T[bo, by, i])  # fixed case 35929
ck(f(('bool', 'bytes5', 'int200'), D), T[bo, by, i])  # fixed case 35930
ck(f(('bool', 'bytes5', 'int208'), D), T[bo, by, i])  # fixed case 35931
ck(f(('bool', 'bytes5', 'int216'), D), T[bo, by, i])  # fixed case 35932
ck(f(('bool', 'bytes5', 'int224'), D), T[bo, by, i])  # fixed case 35933
ck(f(('bool', 'bytes5', 'int232'), D), T[bo, by, i])  # fixed case 35934
ck(f(('bool', 'bytes5', 'int24'), D), T[bo, by, i])  # fixed case 35935
ck(f(('bool', 'bytes5', 'int240'), D), T[bo, by, i])  # fixed case 35936
ck(f(('bool', 'bytes5', 'int248'), D), T[bo, by, i])  # fixed case 35937
ck(f(('bool', 'bytes5', 'int256'), D), T[bo, by, i])  # fixed case 35938
ck(f(('bool', 'bytes5', 'int32'), D), T[bo, by, i])  # fixed case 35939
ck(f(('bool', 'bytes5', 'int40'), D), T[bo, by, i])  # fixed case 35940
ck(f(('bool', 'bytes5', 'int48'), D), T[bo, by, i])  # fixed case 35941
ck(f(('bool', 'bytes5', 'int56'), D), T[bo, by, i])  # fixed case 35942
ck(f(('bool', 'bytes5', 'int64'), D), T[bo, by, i])  # fixed case 35943
ck(f(('bool', 'bytes5', 'int72'), D), T[bo, by, i])  # fixed case 35944
ck(f(('bool', 'bytes5', 'int8'), D), T[bo, by, i])  # fixed case 35945
ck(f(('bool', 'bytes5', 'int80'), D), T[bo, by, i])  # fixed case 35946
ck(f(('bool', 'bytes5', 'int88'), D), T[bo, by, i])  # fixed case 35947
ck(f(('bool', 'bytes5', 'int96'), D), T[bo, by, i])  # fixed case 35948
ck(f(('bool', 'bytes5', 'string'), D), T[bo, by, s])  # fixed case 35949
ck(f(('bool', 'bytes5', 'string[]'), D), T[bo, by, A])  # fixed case 35950
ck(f(('bool', 'bytes5', 'uint104'), D), T[bo, by, i])  # fixed case 35951
ck(f(('bool', 'bytes5', 'uint112'), D), T[bo, by, i])  # fixed case 35952
ck(f(('bool', 'bytes5', 'uint120'), D), T[bo, by, i])  # fixed case 35953
ck(f(('bool', 'bytes5', 'uint128'), D), T[bo, by, i])  # fixed case 35954
ck(f(('bool', 'bytes5', 'uint136'), D), T[bo, by, i])  # fixed case 35955
ck(f(('bool', 'bytes5', 'uint144'), D), T[bo, by, i])  # fixed case 35956
ck(f(('bool', 'bytes5', 'uint152'), D), T[bo, by, i])  # fixed case 35957
ck(f(('bool', 'bytes5', 'uint16'), D), T[bo, by, i])  # fixed case 35958
ck(f(('bool', 'bytes5', 'uint160'), D), T[bo, by, i])  # fixed case 35959
ck(f(('bool', 'bytes5', 'uint168'), D), T[bo, by, i])  # fixed case 35960
ck(f(('bool', 'bytes5', 'uint176'), D), T[bo, by, i])  # fixed case 35961
ck(f(('bool', 'bytes5', 'uint184'), D), T[bo, by, i])  # fixed case 35962
ck(f(('bool', 'bytes5', 'uint192'), D), T[bo, by, i])  # fixed case 35963
ck(f(('bool', 'bytes5', 'uint200'), D), T[bo, by, i])  # fixed case 35964
ck(f(('bool', 'bytes5', 'uint208'), D), T[bo, by, i])  # fixed case 35965
ck(f(('bool', 'bytes5', 'uint216'), D), T[bo, by, i])  # fixed case 35966
ck(f(('bool', 'bytes5', 'uint224'), D), T[bo, by, i])  # fixed case 35967
ck(f(('bool', 'bytes5', 'uint232'), D), T[bo, by, i])  # fixed case 35968
ck(f(('bool', 'bytes5', 'uint24'), D), T[bo, by, i])  # fixed case 35969
ck(f(('bool', 'bytes5', 'uint240'), D), T[bo, by, i])  # fixed case 35970
ck(f(('bool', 'bytes5', 'uint248'), D), T[bo, by, i])  # fixed case 35971
ck(f(('bool', 'bytes5', 'uint256'), D), T[bo, by, i])  # fixed case 35972
ck(f(('bool', 'bytes5', 'uint32'), D), T[bo, by, i])  # fixed case 35973
ck(f(('bool', 'bytes5', 'uint40'), D), T[bo, by, i])  # fixed case 35974
ck(f(('bool', 'bytes5', 'uint48'), D), T[bo, by, i])  # fixed case 35975
ck(f(('bool', 'bytes5', 'uint56'), D), T[bo, by, i])  # fixed case 35976
ck(f(('bool', 'bytes5', 'uint64'), D), T[bo, by, i])  # fixed case 35977
ck(f(('bool', 'bytes5', 'uint72'), D), T[bo, by, i])  # fixed case 35978
ck(f(('bool', 'bytes5', 'uint8'), D), T[bo, by, i])  # fixed case 35979
ck(f(('bool', 'bytes5', 'uint80'), D), T[bo, by, i])  # fixed case 35980
ck(f(('bool', 'bytes5', 'uint88'), D), T[bo, by, i])  # fixed case 35981
ck(f(('bool', 'bytes5', 'uint96'), D), T[bo, by, i])  # fixed case 35982
ck(f(('bool', 'bytes5', '?'), D), T[bo, by, A])  # fixed case 35983
ck(f(('bool', 'bytes6', 'address'), D), T[bo, by, HA])  # fixed case 35984
ck(f(('bool', 'bytes6', 'address[]'), D), T[bo, by, A])  # fixed case 35985
ck(f(('bool', 'bytes6', 'bool'), D), T[bo, by, bo])  # fixed case 35986
ck(f(('bool', 'bytes6', 'bool[]'), D), T[bo, by, A])  # fixed case 35987
ck(f(('bool', 'bytes6', 'bytes'), D), T[bo, by, by])  # fixed case 35988
ck(f(('bool', 'bytes6', 'bytes1'), D), T[bo, by, by])  # fixed case 35989
ck(f(('bool', 'bytes6', 'bytes10'), D), T[bo, by, by])  # fixed case 35990
ck(f(('bool', 'bytes6', 'bytes11'), D), T[bo, by, by])  # fixed case 35991
ck(f(('bool', 'bytes6', 'bytes12'), D), T[bo, by, by])  # fixed case 35992
ck(f(('bool', 'bytes6', 'bytes13'), D), T[bo, by, by])  # fixed case 35993
ck(f(('bool', 'bytes6', 'bytes14'), D), T[bo, by, by])  # fixed case 35994
ck(f(('bool', 'bytes6', 'bytes15'), D), T[bo, by, by])  # fixed case 35995
ck(f(('bool', 'bytes6', 'bytes16'), D), T[bo, by, by])  # fixed case 35996
ck(f(('bool', 'bytes6', 'bytes17'), D), T[bo, by, by])  # fixed case 35997
ck(f(('bool', 'bytes6', 'bytes18'), D), T[bo, by, by])  # fixed case 35998
ck(f(('bool', 'bytes6', 'bytes19'), D), T[bo, by, by])  # fixed case 35999
ck(f(('bool', 'bytes6', 'bytes2'), D), T[bo, by, by])  # fixed case 36000
ck(f(('bool', 'bytes6', 'bytes20'), D), T[bo, by, by])  # fixed case 36001
ck(f(('bool', 'bytes6', 'bytes21'), D), T[bo, by, by])  # fixed case 36002
ck(f(('bool', 'bytes6', 'bytes22'), D), T[bo, by, by])  # fixed case 36003
ck(f(('bool', 'bytes6', 'bytes23'), D), T[bo, by, by])  # fixed case 36004
ck(f(('bool', 'bytes6', 'bytes24'), D), T[bo, by, by])  # fixed case 36005
ck(f(('bool', 'bytes6', 'bytes25'), D), T[bo, by, by])  # fixed case 36006
ck(f(('bool', 'bytes6', 'bytes26'), D), T[bo, by, by])  # fixed case 36007
ck(f(('bool', 'bytes6', 'bytes27'), D), T[bo, by, by])  # fixed case 36008
ck(f(('bool', 'bytes6', 'bytes28'), D), T[bo, by, by])  # fixed case 36009
ck(f(('bool', 'bytes6', 'bytes29'), D), T[bo, by, by])  # fixed case 36010
ck(f(('bool', 'bytes6', 'bytes3'), D), T[bo, by, by])  # fixed case 36011
ck(f(('bool', 'bytes6', 'bytes30'), D), T[bo, by, by])  # fixed case 36012
ck(f(('bool', 'bytes6', 'bytes31'), D), T[bo, by, by])  # fixed case 36013
ck(f(('bool', 'bytes6', 'bytes32'), D), T[bo, by, by])  # fixed case 36014
ck(f(('bool', 'bytes6', 'bytes4'), D), T[bo, by, by])  # fixed case 36015
ck(f(('bool', 'bytes6', 'bytes5'), D), T[bo, by, by])  # fixed case 36016
ck(f(('bool', 'bytes6', 'bytes6'), D), T[bo, by, by])  # fixed case 36017
ck(f(('bool', 'bytes6', 'bytes7'), D), T[bo, by, by])  # fixed case 36018
ck(f(('bool', 'bytes6', 'bytes8'), D), T[bo, by, by])  # fixed case 36019
ck(f(('bool', 'bytes6', 'bytes9'), D), T[bo, by, by])  # fixed case 36020
ck(f(('bool', 'bytes6', 'int104'), D), T[bo, by, i])  # fixed case 36021
ck(f(('bool', 'bytes6', 'int112'), D), T[bo, by, i])  # fixed case 36022
ck(f(('bool', 'bytes6', 'int120'), D), T[bo, by, i])  # fixed case 36023
ck(f(('bool', 'bytes6', 'int128'), D), T[bo, by, i])  # fixed case 36024
ck(f(('bool', 'bytes6', 'int136'), D), T[bo, by, i])  # fixed case 36025
ck(f(('bool', 'bytes6', 'int144'), D), T[bo, by, i])  # fixed case 36026
ck(f(('bool', 'bytes6', 'int152'), D), T[bo, by, i])  # fixed case 36027
ck(f(('bool', 'bytes6', 'int16'), D), T[bo, by, i])  # fixed case 36028
ck(f(('bool', 'bytes6', 'int160'), D), T[bo, by, i])  # fixed case 36029
ck(f(('bool', 'bytes6', 'int168'), D), T[bo, by, i])  # fixed case 36030
ck(f(('bool', 'bytes6', 'int176'), D), T[bo, by, i])  # fixed case 36031
ck(f(('bool', 'bytes6', 'int184'), D), T[bo, by, i])  # fixed case 36032
ck(f(('bool', 'bytes6', 'int192'), D), T[bo, by, i])  # fixed case 36033
ck(f(('bool', 'bytes6', 'int200'), D), T[bo, by, i])  # fixed case 36034
ck(f(('bool', 'bytes6', 'int208'), D), T[bo, by, i])  # fixed case 36035
ck(f(('bool', 'bytes6', 'int216'), D), T[bo, by, i])  # fixed case 36036
ck(f(('bool', 'bytes6', 'int224'), D), T[bo, by, i])  # fixed case 36037
ck(f(('bool', 'bytes6', 'int232'), D), T[bo, by, i])  # fixed case 36038
ck(f(('bool', 'bytes6', 'int24'), D), T[bo, by, i])  # fixed case 36039
ck(f(('bool', 'bytes6', 'int240'), D), T[bo, by, i])  # fixed case 36040
ck(f(('bool', 'bytes6', 'int248'), D), T[bo, by, i])  # fixed case 36041
ck(f(('bool', 'bytes6', 'int256'), D), T[bo, by, i])  # fixed case 36042
ck(f(('bool', 'bytes6', 'int32'), D), T[bo, by, i])  # fixed case 36043
ck(f(('bool', 'bytes6', 'int40'), D), T[bo, by, i])  # fixed case 36044
ck(f(('bool', 'bytes6', 'int48'), D), T[bo, by, i])  # fixed case 36045
ck(f(('bool', 'bytes6', 'int56'), D), T[bo, by, i])  # fixed case 36046
ck(f(('bool', 'bytes6', 'int64'), D), T[bo, by, i])  # fixed case 36047
ck(f(('bool', 'bytes6', 'int72'), D), T[bo, by, i])  # fixed case 36048
ck(f(('bool', 'bytes6', 'int8'), D), T[bo, by, i])  # fixed case 36049
ck(f(('bool', 'bytes6', 'int80'), D), T[bo, by, i])  # fixed case 36050
ck(f(('bool', 'bytes6', 'int88'), D), T[bo, by, i])  # fixed case 36051
ck(f(('bool', 'bytes6', 'int96'), D), T[bo, by, i])  # fixed case 36052
ck(f(('bool', 'bytes6', 'string'), D), T[bo, by, s])  # fixed case 36053
ck(f(('bool', 'bytes6', 'string[]'), D), T[bo, by, A])  # fixed case 36054
ck(f(('bool', 'bytes6', 'uint104'), D), T[bo, by, i])  # fixed case 36055
ck(f(('bool', 'bytes6', 'uint112'), D), T[bo, by, i])  # fixed case 36056
ck(f(('bool', 'bytes6', 'uint120'), D), T[bo, by, i])  # fixed case 36057
ck(f(('bool', 'bytes6', 'uint128'), D), T[bo, by, i])  # fixed case 36058
ck(f(('bool', 'bytes6', 'uint136'), D), T[bo, by, i])  # fixed case 36059
ck(f(('bool', 'bytes6', 'uint144'), D), T[bo, by, i])  # fixed case 36060
ck(f(('bool', 'bytes6', 'uint152'), D), T[bo, by, i])  # fixed case 36061
ck(f(('bool', 'bytes6', 'uint16'), D), T[bo, by, i])  # fixed case 36062
ck(f(('bool', 'bytes6', 'uint160'), D), T[bo, by, i])  # fixed case 36063
ck(f(('bool', 'bytes6', 'uint168'), D), T[bo, by, i])  # fixed case 36064
ck(f(('bool', 'bytes6', 'uint176'), D), T[bo, by, i])  # fixed case 36065
ck(f(('bool', 'bytes6', 'uint184'), D), T[bo, by, i])  # fixed case 36066
ck(f(('bool', 'bytes6', 'uint192'), D), T[bo, by, i])  # fixed case 36067
ck(f(('bool', 'bytes6', 'uint200'), D), T[bo, by, i])  # fixed case 36068
ck(f(('bool', 'bytes6', 'uint208'), D), T[bo, by, i])  # fixed case 36069
ck(f(('bool', 'bytes6', 'uint216'), D), T[bo, by, i])  # fixed case 36070
ck(f(('bool', 'bytes6', 'uint224'), D), T[bo, by, i])  # fixed case 36071
ck(f(('bool', 'bytes6', 'uint232'), D), T[bo, by, i])  # fixed case 36072
ck(f(('bool', 'bytes6', 'uint24'), D), T[bo, by, i])  # fixed case 36073
ck(f(('bool', 'bytes6', 'uint240'), D), T[bo, by, i])  # fixed case 36074
ck(f(('bool', 'bytes6', 'uint248'), D), T[bo, by, i])  # fixed case 36075
ck(f(('bool', 'bytes6', 'uint256'), D), T[bo, by, i])  # fixed case 36076
ck(f(('bool', 'bytes6', 'uint32'), D), T[bo, by, i])  # fixed case 36077
ck(f(('bool', 'bytes6', 'uint40'), D), T[bo, by, i])  # fixed case 36078
ck(f(('bool', 'bytes6', 'uint48'), D), T[bo, by, i])  # fixed case 36079
ck(f(('bool', 'bytes6', 'uint56'), D), T[bo, by, i])  # fixed case 36080
ck(f(('bool', 'bytes6', 'uint64'), D), T[bo, by, i])  # fixed case 36081
ck(f(('bool', 'bytes6', 'uint72'), D), T[bo, by, i])  # fixed case 36082
ck(f(('bool', 'bytes6', 'uint8'), D), T[bo, by, i])  # fixed case 36083
ck(f(('bool', 'bytes6', 'uint80'), D), T[bo, by, i])  # fixed case 36084
ck(f(('bool', 'bytes6', 'uint88'), D), T[bo, by, i])  # fixed case 36085
ck(f(('bool', 'bytes6', 'uint96'), D), T[bo, by, i])  # fixed case 36086
ck(f(('bool', 'bytes6', '?'), D), T[bo, by, A])  # fixed case 36087
ck(f(('bool', 'bytes7', 'address'), D), T[bo, by, HA])  # fixed case 36088
ck(f(('bool', 'bytes7', 'address[]'), D), T[bo, by, A])  # fixed case 36089
ck(f(('bool', 'bytes7', 'bool'), D), T[bo, by, bo])  # fixed case 36090
ck(f(('bool', 'bytes7', 'bool[]'), D), T[bo, by, A])  # fixed case 36091
ck(f(('bool', 'bytes7', 'bytes'), D), T[bo, by, by])  # fixed case 36092
ck(f(('bool', 'bytes7', 'bytes1'), D), T[bo, by, by])  # fixed case 36093
ck(f(('bool', 'bytes7', 'bytes10'), D), T[bo, by, by])  # fixed case 36094
ck(f(('bool', 'bytes7', 'bytes11'), D), T[bo, by, by])  # fixed case 36095
ck(f(('bool', 'bytes7', 'bytes12'), D), T[bo, by, by])  # fixed case 36096
ck(f(('bool', 'bytes7', 'bytes13'), D), T[bo, by, by])  # fixed case 36097
ck(f(('bool', 'bytes7', 'bytes14'), D), T[bo, by, by])  # fixed case 36098
ck(f(('bool', 'bytes7', 'bytes15'), D), T[bo, by, by])  # fixed case 36099
ck(f(('bool', 'bytes7', 'bytes16'), D), T[bo, by, by])  # fixed case 36100
ck(f(('bool', 'bytes7', 'bytes17'), D), T[bo, by, by])  # fixed case 36101
ck(f(('bool', 'bytes7', 'bytes18'), D), T[bo, by, by])  # fixed case 36102
ck(f(('bool', 'bytes7', 'bytes19'), D), T[bo, by, by])  # fixed case 36103
ck(f(('bool', 'bytes7', 'bytes2'), D), T[bo, by, by])  # fixed case 36104
ck(f(('bool', 'bytes7', 'bytes20'), D), T[bo, by, by])  # fixed case 36105
ck(f(('bool', 'bytes7', 'bytes21'), D), T[bo, by, by])  # fixed case 36106
ck(f(('bool', 'bytes7', 'bytes22'), D), T[bo, by, by])  # fixed case 36107
ck(f(('bool', 'bytes7', 'bytes23'), D), T[bo, by, by])  # fixed case 36108
ck(f(('bool', 'bytes7', 'bytes24'), D), T[bo, by, by])  # fixed case 36109
ck(f(('bool', 'bytes7', 'bytes25'), D), T[bo, by, by])  # fixed case 36110
ck(f(('bool', 'bytes7', 'bytes26'), D), T[bo, by, by])  # fixed case 36111
ck(f(('bool', 'bytes7', 'bytes27'), D), T[bo, by, by])  # fixed case 36112
ck(f(('bool', 'bytes7', 'bytes28'), D), T[bo, by, by])  # fixed case 36113
ck(f(('bool', 'bytes7', 'bytes29'), D), T[bo, by, by])  # fixed case 36114
ck(f(('bool', 'bytes7', 'bytes3'), D), T[bo, by, by])  # fixed case 36115
ck(f(('bool', 'bytes7', 'bytes30'), D), T[bo, by, by])  # fixed case 36116
ck(f(('bool', 'bytes7', 'bytes31'), D), T[bo, by, by])  # fixed case 36117
ck(f(('bool', 'bytes7', 'bytes32'), D), T[bo, by, by])  # fixed case 36118
ck(f(('bool', 'bytes7', 'bytes4'), D), T[bo, by, by])  # fixed case 36119
ck(f(('bool', 'bytes7', 'bytes5'), D), T[bo, by, by])  # fixed case 36120
ck(f(('bool', 'bytes7', 'bytes6'), D), T[bo, by, by])  # fixed case 36121
ck(f(('bool', 'bytes7', 'bytes7'), D), T[bo, by, by])  # fixed case 36122
ck(f(('bool', 'bytes7', 'bytes8'), D), T[bo, by, by])  # fixed case 36123
ck(f(('bool', 'bytes7', 'bytes9'), D), T[bo, by, by])  # fixed case 36124
ck(f(('bool', 'bytes7', 'int104'), D), T[bo, by, i])  # fixed case 36125
ck(f(('bool', 'bytes7', 'int112'), D), T[bo, by, i])  # fixed case 36126
ck(f(('bool', 'bytes7', 'int120'), D), T[bo, by, i])  # fixed case 36127
ck(f(('bool', 'bytes7', 'int128'), D), T[bo, by, i])  # fixed case 36128
ck(f(('bool', 'bytes7', 'int136'), D), T[bo, by, i])  # fixed case 36129
ck(f(('bool', 'bytes7', 'int144'), D), T[bo, by, i])  # fixed case 36130
ck(f(('bool', 'bytes7', 'int152'), D), T[bo, by, i])  # fixed case 36131
ck(f(('bool', 'bytes7', 'int16'), D), T[bo, by, i])  # fixed case 36132
ck(f(('bool', 'bytes7', 'int160'), D), T[bo, by, i])  # fixed case 36133
ck(f(('bool', 'bytes7', 'int168'), D), T[bo, by, i])  # fixed case 36134
ck(f(('bool', 'bytes7', 'int176'), D), T[bo, by, i])  # fixed case 36135
ck(f(('bool', 'bytes7', 'int184'), D), T[bo, by, i])  # fixed case 36136
ck(f(('bool', 'bytes7', 'int192'), D), T[bo, by, i])  # fixed case 36137
ck(f(('bool', 'bytes7', 'int200'), D), T[bo, by, i])  # fixed case 36138
ck(f(('bool', 'bytes7', 'int208'), D), T[bo, by, i])  # fixed case 36139
ck(f(('bool', 'bytes7', 'int216'), D), T[bo, by, i])  # fixed case 36140
ck(f(('bool', 'bytes7', 'int224'), D), T[bo, by, i])  # fixed case 36141
ck(f(('bool', 'bytes7', 'int232'), D), T[bo, by, i])  # fixed case 36142
ck(f(('bool', 'bytes7', 'int24'), D), T[bo, by, i])  # fixed case 36143
ck(f(('bool', 'bytes7', 'int240'), D), T[bo, by, i])  # fixed case 36144
ck(f(('bool', 'bytes7', 'int248'), D), T[bo, by, i])  # fixed case 36145
ck(f(('bool', 'bytes7', 'int256'), D), T[bo, by, i])  # fixed case 36146
ck(f(('bool', 'bytes7', 'int32'), D), T[bo, by, i])  # fixed case 36147
ck(f(('bool', 'bytes7', 'int40'), D), T[bo, by, i])  # fixed case 36148
ck(f(('bool', 'bytes7', 'int48'), D), T[bo, by, i])  # fixed case 36149
ck(f(('bool', 'bytes7', 'int56'), D), T[bo, by, i])  # fixed case 36150
ck(f(('bool', 'bytes7', 'int64'), D), T[bo, by, i])  # fixed case 36151
ck(f(('bool', 'bytes7', 'int72'), D), T[bo, by, i])  # fixed case 36152
ck(f(('bool', 'bytes7', 'int8'), D), T[bo, by, i])  # fixed case 36153
ck(f(('bool', 'bytes7', 'int80'), D), T[bo, by, i])  # fixed case 36154
ck(f(('bool', 'bytes7', 'int88'), D), T[bo, by, i])  # fixed case 36155
ck(f(('bool', 'bytes7', 'int96'), D), T[bo, by, i])  # fixed case 36156
ck(f(('bool', 'bytes7', 'string'), D), T[bo, by, s])  # fixed case 36157
ck(f(('bool', 'bytes7', 'string[]'), D), T[bo, by, A])  # fixed case 36158
ck(f(('bool', 'bytes7', 'uint104'), D), T[bo, by, i])  # fixed case 36159
ck(f(('bool', 'bytes7', 'uint112'), D), T[bo, by, i])  # fixed case 36160
ck(f(('bool', 'bytes7', 'uint120'), D), T[bo, by, i])  # fixed case 36161
ck(f(('bool', 'bytes7', 'uint128'), D), T[bo, by, i])  # fixed case 36162
ck(f(('bool', 'bytes7', 'uint136'), D), T[bo, by, i])  # fixed case 36163
ck(f(('bool', 'bytes7', 'uint144'), D), T[bo, by, i])  # fixed case 36164
ck(f(('bool', 'bytes7', 'uint152'), D), T[bo, by, i])  # fixed case 36165
ck(f(('bool', 'bytes7', 'uint16'), D), T[bo, by, i])  # fixed case 36166
ck(f(('bool', 'bytes7', 'uint160'), D), T[bo, by, i])  # fixed case 36167
ck(f(('bool', 'bytes7', 'uint168'), D), T[bo, by, i])  # fixed case 36168
ck(f(('bool', 'bytes7', 'uint176'), D), T[bo, by, i])  # fixed case 36169
ck(f(('bool', 'bytes7', 'uint184'), D), T[bo, by, i])  # fixed case 36170
ck(f(('bool', 'bytes7', 'uint192'), D), T[bo, by, i])  # fixed case 36171
ck(f(('bool', 'bytes7', 'uint200'), D), T[bo, by, i])  # fixed case 36172
ck(f(('bool', 'bytes7', 'uint208'), D), T[bo, by, i])  # fixed case 36173
ck(f(('bool', 'bytes7', 'uint216'), D), T[bo, by, i])  # fixed case 36174
ck(f(('bool', 'bytes7', 'uint224'), D), T[bo, by, i])  # fixed case 36175
ck(f(('bool', 'bytes7', 'uint232'), D), T[bo, by, i])  # fixed case 36176
ck(f(('bool', 'bytes7', 'uint24'), D), T[bo, by, i])  # fixed case 36177
ck(f(('bool', 'bytes7', 'uint240'), D), T[bo, by, i])  # fixed case 36178
ck(f(('bool', 'bytes7', 'uint248'), D), T[bo, by, i])  # fixed case 36179
ck(f(('bool', 'bytes7', 'uint256'), D), T[bo, by, i])  # fixed case 36180
ck(f(('bool', 'bytes7', 'uint32'), D), T[bo, by, i])  # fixed case 36181
ck(f(('bool', 'bytes7', 'uint40'), D), T[bo, by, i])  # fixed case 36182
ck(f(('bool', 'bytes7', 'uint48'), D), T[bo, by, i])  # fixed case 36183
ck(f(('bool', 'bytes7', 'uint56'), D), T[bo, by, i])  # fixed case 36184
ck(f(('bool', 'bytes7', 'uint64'), D), T[bo, by, i])  # fixed case 36185
ck(f(('bool', 'bytes7', 'uint72'), D), T[bo, by, i])  # fixed case 36186
ck(f(('bool', 'bytes7', 'uint8'), D), T[bo, by, i])  # fixed case 36187
ck(f(('bool', 'bytes7', 'uint80'), D), T[bo, by, i])  # fixed case 36188
ck(f(('bool', 'bytes7', 'uint88'), D), T[bo, by, i])  # fixed case 36189
ck(f(('bool', 'bytes7', 'uint96'), D), T[bo, by, i])  # fixed case 36190
ck(f(('bool', 'bytes7', '?'), D), T[bo, by, A])  # fixed case 36191
ck(f(('bool', 'bytes8', 'address'), D), T[bo, by, HA])  # fixed case 36192
ck(f(('bool', 'bytes8', 'address[]'), D), T[bo, by, A])  # fixed case 36193
ck(f(('bool', 'bytes8', 'bool'), D), T[bo, by, bo])  # fixed case 36194
ck(f(('bool', 'bytes8', 'bool[]'), D), T[bo, by, A])  # fixed case 36195
ck(f(('bool', 'bytes8', 'bytes'), D), T[bo, by, by])  # fixed case 36196
ck(f(('bool', 'bytes8', 'bytes1'), D), T[bo, by, by])  # fixed case 36197
ck(f(('bool', 'bytes8', 'bytes10'), D), T[bo, by, by])  # fixed case 36198
ck(f(('bool', 'bytes8', 'bytes11'), D), T[bo, by, by])  # fixed case 36199
ck(f(('bool', 'bytes8', 'bytes12'), D), T[bo, by, by])  # fixed case 36200
ck(f(('bool', 'bytes8', 'bytes13'), D), T[bo, by, by])  # fixed case 36201
ck(f(('bool', 'bytes8', 'bytes14'), D), T[bo, by, by])  # fixed case 36202
ck(f(('bool', 'bytes8', 'bytes15'), D), T[bo, by, by])  # fixed case 36203
ck(f(('bool', 'bytes8', 'bytes16'), D), T[bo, by, by])  # fixed case 36204
ck(f(('bool', 'bytes8', 'bytes17'), D), T[bo, by, by])  # fixed case 36205
ck(f(('bool', 'bytes8', 'bytes18'), D), T[bo, by, by])  # fixed case 36206
ck(f(('bool', 'bytes8', 'bytes19'), D), T[bo, by, by])  # fixed case 36207
ck(f(('bool', 'bytes8', 'bytes2'), D), T[bo, by, by])  # fixed case 36208
ck(f(('bool', 'bytes8', 'bytes20'), D), T[bo, by, by])  # fixed case 36209
ck(f(('bool', 'bytes8', 'bytes21'), D), T[bo, by, by])  # fixed case 36210
ck(f(('bool', 'bytes8', 'bytes22'), D), T[bo, by, by])  # fixed case 36211
ck(f(('bool', 'bytes8', 'bytes23'), D), T[bo, by, by])  # fixed case 36212
ck(f(('bool', 'bytes8', 'bytes24'), D), T[bo, by, by])  # fixed case 36213
ck(f(('bool', 'bytes8', 'bytes25'), D), T[bo, by, by])  # fixed case 36214
ck(f(('bool', 'bytes8', 'bytes26'), D), T[bo, by, by])  # fixed case 36215
ck(f(('bool', 'bytes8', 'bytes27'), D), T[bo, by, by])  # fixed case 36216
ck(f(('bool', 'bytes8', 'bytes28'), D), T[bo, by, by])  # fixed case 36217
ck(f(('bool', 'bytes8', 'bytes29'), D), T[bo, by, by])  # fixed case 36218
ck(f(('bool', 'bytes8', 'bytes3'), D), T[bo, by, by])  # fixed case 36219
ck(f(('bool', 'bytes8', 'bytes30'), D), T[bo, by, by])  # fixed case 36220
ck(f(('bool', 'bytes8', 'bytes31'), D), T[bo, by, by])  # fixed case 36221
ck(f(('bool', 'bytes8', 'bytes32'), D), T[bo, by, by])  # fixed case 36222
ck(f(('bool', 'bytes8', 'bytes4'), D), T[bo, by, by])  # fixed case 36223
ck(f(('bool', 'bytes8', 'bytes5'), D), T[bo, by, by])  # fixed case 36224
ck(f(('bool', 'bytes8', 'bytes6'), D), T[bo, by, by])  # fixed case 36225
ck(f(('bool', 'bytes8', 'bytes7'), D), T[bo, by, by])  # fixed case 36226
ck(f(('bool', 'bytes8', 'bytes8'), D), T[bo, by, by])  # fixed case 36227
ck(f(('bool', 'bytes8', 'bytes9'), D), T[bo, by, by])  # fixed case 36228
ck(f(('bool', 'bytes8', 'int104'), D), T[bo, by, i])  # fixed case 36229
ck(f(('bool', 'bytes8', 'int112'), D), T[bo, by, i])  # fixed case 36230
ck(f(('bool', 'bytes8', 'int120'), D), T[bo, by, i])  # fixed case 36231
ck(f(('bool', 'bytes8', 'int128'), D), T[bo, by, i])  # fixed case 36232
ck(f(('bool', 'bytes8', 'int136'), D), T[bo, by, i])  # fixed case 36233
ck(f(('bool', 'bytes8', 'int144'), D), T[bo, by, i])  # fixed case 36234
ck(f(('bool', 'bytes8', 'int152'), D), T[bo, by, i])  # fixed case 36235
ck(f(('bool', 'bytes8', 'int16'), D), T[bo, by, i])  # fixed case 36236
ck(f(('bool', 'bytes8', 'int160'), D), T[bo, by, i])  # fixed case 36237
ck(f(('bool', 'bytes8', 'int168'), D), T[bo, by, i])  # fixed case 36238
ck(f(('bool', 'bytes8', 'int176'), D), T[bo, by, i])  # fixed case 36239
ck(f(('bool', 'bytes8', 'int184'), D), T[bo, by, i])  # fixed case 36240
ck(f(('bool', 'bytes8', 'int192'), D), T[bo, by, i])  # fixed case 36241
ck(f(('bool', 'bytes8', 'int200'), D), T[bo, by, i])  # fixed case 36242
ck(f(('bool', 'bytes8', 'int208'), D), T[bo, by, i])  # fixed case 36243
ck(f(('bool', 'bytes8', 'int216'), D), T[bo, by, i])  # fixed case 36244
ck(f(('bool', 'bytes8', 'int224'), D), T[bo, by, i])  # fixed case 36245
ck(f(('bool', 'bytes8', 'int232'), D), T[bo, by, i])  # fixed case 36246
ck(f(('bool', 'bytes8', 'int24'), D), T[bo, by, i])  # fixed case 36247
ck(f(('bool', 'bytes8', 'int240'), D), T[bo, by, i])  # fixed case 36248
ck(f(('bool', 'bytes8', 'int248'), D), T[bo, by, i])  # fixed case 36249
ck(f(('bool', 'bytes8', 'int256'), D), T[bo, by, i])  # fixed case 36250
ck(f(('bool', 'bytes8', 'int32'), D), T[bo, by, i])  # fixed case 36251
ck(f(('bool', 'bytes8', 'int40'), D), T[bo, by, i])  # fixed case 36252
ck(f(('bool', 'bytes8', 'int48'), D), T[bo, by, i])  # fixed case 36253
ck(f(('bool', 'bytes8', 'int56'), D), T[bo, by, i])  # fixed case 36254
ck(f(('bool', 'bytes8', 'int64'), D), T[bo, by, i])  # fixed case 36255
ck(f(('bool', 'bytes8', 'int72'), D), T[bo, by, i])  # fixed case 36256
ck(f(('bool', 'bytes8', 'int8'), D), T[bo, by, i])  # fixed case 36257
ck(f(('bool', 'bytes8', 'int80'), D), T[bo, by, i])  # fixed case 36258
ck(f(('bool', 'bytes8', 'int88'), D), T[bo, by, i])  # fixed case 36259
ck(f(('bool', 'bytes8', 'int96'), D), T[bo, by, i])  # fixed case 36260
ck(f(('bool', 'bytes8', 'string'), D), T[bo, by, s])  # fixed case 36261
ck(f(('bool', 'bytes8', 'string[]'), D), T[bo, by, A])  # fixed case 36262
ck(f(('bool', 'bytes8', 'uint104'), D), T[bo, by, i])  # fixed case 36263
ck(f(('bool', 'bytes8', 'uint112'), D), T[bo, by, i])  # fixed case 36264
ck(f(('bool', 'bytes8', 'uint120'), D), T[bo, by, i])  # fixed case 36265
ck(f(('bool', 'bytes8', 'uint128'), D), T[bo, by, i])  # fixed case 36266
ck(f(('bool', 'bytes8', 'uint136'), D), T[bo, by, i])  # fixed case 36267
ck(f(('bool', 'bytes8', 'uint144'), D), T[bo, by, i])  # fixed case 36268
ck(f(('bool', 'bytes8', 'uint152'), D), T[bo, by, i])  # fixed case 36269
ck(f(('bool', 'bytes8', 'uint16'), D), T[bo, by, i])  # fixed case 36270
ck(f(('bool', 'bytes8', 'uint160'), D), T[bo, by, i])  # fixed case 36271
ck(f(('bool', 'bytes8', 'uint168'), D), T[bo, by, i])  # fixed case 36272
ck(f(('bool', 'bytes8', 'uint176'), D), T[bo, by, i])  # fixed case 36273
ck(f(('bool', 'bytes8', 'uint184'), D), T[bo, by, i])  # fixed case 36274
ck(f(('bool', 'bytes8', 'uint192'), D), T[bo, by, i])  # fixed case 36275
ck(f(('bool', 'bytes8', 'uint200'), D), T[bo, by, i])  # fixed case 36276
ck(f(('bool', 'bytes8', 'uint208'), D), T[bo, by, i])  # fixed case 36277
ck(f(('bool', 'bytes8', 'uint216'), D), T[bo, by, i])  # fixed case 36278
ck(f(('bool', 'bytes8', 'uint224'), D), T[bo, by, i])  # fixed case 36279
ck(f(('bool', 'bytes8', 'uint232'), D), T[bo, by, i])  # fixed case 36280
ck(f(('bool', 'bytes8', 'uint24'), D), T[bo, by, i])  # fixed case 36281
ck(f(('bool', 'bytes8', 'uint240'), D), T[bo, by, i])  # fixed case 36282
ck(f(('bool', 'bytes8', 'uint248'), D), T[bo, by, i])  # fixed case 36283
ck(f(('bool', 'bytes8', 'uint256'), D), T[bo, by, i])  # fixed case 36284
ck(f(('bool', 'bytes8', 'uint32'), D), T[bo, by, i])  # fixed case 36285
ck(f(('bool', 'bytes8', 'uint40'), D), T[bo, by, i])  # fixed case 36286
ck(f(('bool', 'bytes8', 'uint48'), D), T[bo, by, i])  # fixed case 36287
ck(f(('bool', 'bytes8', 'uint56'), D), T[bo, by, i])  # fixed case 36288
ck(f(('bool', 'bytes8', 'uint64'), D), T[bo, by, i])  # fixed case 36289
ck(f(('bool', 'bytes8', 'uint72'), D), T[bo, by, i])  # fixed case 36290
ck(f(('bool', 'bytes8', 'uint8'), D), T[bo, by, i])  # fixed case 36291
ck(f(('bool', 'bytes8', 'uint80'), D), T[bo, by, i])  # fixed case 36292
ck(f(('bool', 'bytes8', 'uint88'), D), T[bo, by, i])  # fixed case 36293
ck(f(('bool', 'bytes8', 'uint96'), D), T[bo, by, i])  # fixed case 36294
ck(f(('bool', 'bytes8', '?'), D), T[bo, by, A])  # fixed case 36295
ck(f(('bool', 'bytes9', 'address'), D), T[bo, by, HA])  # fixed case 36296
ck(f(('bool', 'bytes9', 'address[]'), D), T[bo, by, A])  # fixed case 36297
ck(f(('bool', 'bytes9', 'bool'), D), T[bo, by, bo])  # fixed case 36298
ck(f(('bool', 'bytes9', 'bool[]'), D), T[bo, by, A])  # fixed case 36299
ck(f(('bool', 'bytes9', 'bytes'), D), T[bo, by, by])  # fixed case 36300
ck(f(('bool', 'bytes9', 'bytes1'), D), T[bo, by, by])  # fixed case 36301
ck(f(('bool', 'bytes9', 'bytes10'), D), T[bo, by, by])  # fixed case 36302
ck(f(('bool', 'bytes9', 'bytes11'), D), T[bo, by, by])  # fixed case 36303
ck(f(('bool', 'bytes9', 'bytes12'), D), T[bo, by, by])  # fixed case 36304
ck(f(('bool', 'bytes9', 'bytes13'), D), T[bo, by, by])  # fixed case 36305
ck(f(('bool', 'bytes9', 'bytes14'), D), T[bo, by, by])  # fixed case 36306
ck(f(('bool', 'bytes9', 'bytes15'), D), T[bo, by, by])  # fixed case 36307
ck(f(('bool', 'bytes9', 'bytes16'), D), T[bo, by, by])  # fixed case 36308
ck(f(('bool', 'bytes9', 'bytes17'), D), T[bo, by, by])  # fixed case 36309
ck(f(('bool', 'bytes9', 'bytes18'), D), T[bo, by, by])  # fixed case 36310
ck(f(('bool', 'bytes9', 'bytes19'), D), T[bo, by, by])  # fixed case 36311
ck(f(('bool', 'bytes9', 'bytes2'), D), T[bo, by, by])  # fixed case 36312
ck(f(('bool', 'bytes9', 'bytes20'), D), T[bo, by, by])  # fixed case 36313
ck(f(('bool', 'bytes9', 'bytes21'), D), T[bo, by, by])  # fixed case 36314
ck(f(('bool', 'bytes9', 'bytes22'), D), T[bo, by, by])  # fixed case 36315
ck(f(('bool', 'bytes9', 'bytes23'), D), T[bo, by, by])  # fixed case 36316
ck(f(('bool', 'bytes9', 'bytes24'), D), T[bo, by, by])  # fixed case 36317
ck(f(('bool', 'bytes9', 'bytes25'), D), T[bo, by, by])  # fixed case 36318
ck(f(('bool', 'bytes9', 'bytes26'), D), T[bo, by, by])  # fixed case 36319
ck(f(('bool', 'bytes9', 'bytes27'), D), T[bo, by, by])  # fixed case 36320
ck(f(('bool', 'bytes9', 'bytes28'), D), T[bo, by, by])  # fixed case 36321
ck(f(('bool', 'bytes9', 'bytes29'), D), T[bo, by, by])  # fixed case 36322
ck(f(('bool', 'bytes9', 'bytes3'), D), T[bo, by, by])  # fixed case 36323
ck(f(('bool', 'bytes9', 'bytes30'), D), T[bo, by, by])  # fixed case 36324
ck(f(('bool', 'bytes9', 'bytes31'), D), T[bo, by, by])  # fixed case 36325
ck(f(('bool', 'bytes9', 'bytes32'), D), T[bo, by, by])  # fixed case 36326
ck(f(('bool', 'bytes9', 'bytes4'), D), T[bo, by, by])  # fixed case 36327
ck(f(('bool', 'bytes9', 'bytes5'), D), T[bo, by, by])  # fixed case 36328
ck(f(('bool', 'bytes9', 'bytes6'), D), T[bo, by, by])  # fixed case 36329
ck(f(('bool', 'bytes9', 'bytes7'), D), T[bo, by, by])  # fixed case 36330
ck(f(('bool', 'bytes9', 'bytes8'), D), T[bo, by, by])  # fixed case 36331
ck(f(('bool', 'bytes9', 'bytes9'), D), T[bo, by, by])  # fixed case 36332
ck(f(('bool', 'bytes9', 'int104'), D), T[bo, by, i])  # fixed case 36333
ck(f(('bool', 'bytes9', 'int112'), D), T[bo, by, i])  # fixed case 36334
ck(f(('bool', 'bytes9', 'int120'), D), T[bo, by, i])  # fixed case 36335
ck(f(('bool', 'bytes9', 'int128'), D), T[bo, by, i])  # fixed case 36336
ck(f(('bool', 'bytes9', 'int136'), D), T[bo, by, i])  # fixed case 36337
ck(f(('bool', 'bytes9', 'int144'), D), T[bo, by, i])  # fixed case 36338
ck(f(('bool', 'bytes9', 'int152'), D), T[bo, by, i])  # fixed case 36339
ck(f(('bool', 'bytes9', 'int16'), D), T[bo, by, i])  # fixed case 36340
ck(f(('bool', 'bytes9', 'int160'), D), T[bo, by, i])  # fixed case 36341
ck(f(('bool', 'bytes9', 'int168'), D), T[bo, by, i])  # fixed case 36342
ck(f(('bool', 'bytes9', 'int176'), D), T[bo, by, i])  # fixed case 36343
ck(f(('bool', 'bytes9', 'int184'), D), T[bo, by, i])  # fixed case 36344
ck(f(('bool', 'bytes9', 'int192'), D), T[bo, by, i])  # fixed case 36345
ck(f(('bool', 'bytes9', 'int200'), D), T[bo, by, i])  # fixed case 36346
ck(f(('bool', 'bytes9', 'int208'), D), T[bo, by, i])  # fixed case 36347
ck(f(('bool', 'bytes9', 'int216'), D), T[bo, by, i])  # fixed case 36348
ck(f(('bool', 'bytes9', 'int224'), D), T[bo, by, i])  # fixed case 36349
ck(f(('bool', 'bytes9', 'int232'), D), T[bo, by, i])  # fixed case 36350
ck(f(('bool', 'bytes9', 'int24'), D), T[bo, by, i])  # fixed case 36351
ck(f(('bool', 'bytes9', 'int240'), D), T[bo, by, i])  # fixed case 36352
ck(f(('bool', 'bytes9', 'int248'), D), T[bo, by, i])  # fixed case 36353
ck(f(('bool', 'bytes9', 'int256'), D), T[bo, by, i])  # fixed case 36354
ck(f(('bool', 'bytes9', 'int32'), D), T[bo, by, i])  # fixed case 36355
ck(f(('bool', 'bytes9', 'int40'), D), T[bo, by, i])  # fixed case 36356
ck(f(('bool', 'bytes9', 'int48'), D), T[bo, by, i])  # fixed case 36357
ck(f(('bool', 'bytes9', 'int56'), D), T[bo, by, i])  # fixed case 36358
ck(f(('bool', 'bytes9', 'int64'), D), T[bo, by, i])  # fixed case 36359
ck(f(('bool', 'bytes9', 'int72'), D), T[bo, by, i])  # fixed case 36360
ck(f(('bool', 'bytes9', 'int8'), D), T[bo, by, i])  # fixed case 36361
ck(f(('bool', 'bytes9', 'int80'), D), T[bo, by, i])  # fixed case 36362
ck(f(('bool', 'bytes9', 'int88'), D), T[bo, by, i])  # fixed case 36363
ck(f(('bool', 'bytes9', 'int96'), D), T[bo, by, i])  # fixed case 36364
ck(f(('bool', 'bytes9', 'string'), D), T[bo, by, s])  # fixed case 36365
ck(f(('bool', 'bytes9', 'string[]'), D), T[bo, by, A])  # fixed case 36366
ck(f(('bool', 'bytes9', 'uint104'), D), T[bo, by, i])  # fixed case 36367
ck(f(('bool', 'bytes9', 'uint112'), D), T[bo, by, i])  # fixed case 36368
ck(f(('bool', 'bytes9', 'uint120'), D), T[bo, by, i])  # fixed case 36369
ck(f(('bool', 'bytes9', 'uint128'), D), T[bo, by, i])  # fixed case 36370
ck(f(('bool', 'bytes9', 'uint136'), D), T[bo, by, i])  # fixed case 36371
ck(f(('bool', 'bytes9', 'uint144'), D), T[bo, by, i])  # fixed case 36372
ck(f(('bool', 'bytes9', 'uint152'), D), T[bo, by, i])  # fixed case 36373
ck(f(('bool', 'bytes9', 'uint16'), D), T[bo, by, i])  # fixed case 36374
ck(f(('bool', 'bytes9', 'uint160'), D), T[bo, by, i])  # fixed case 36375
ck(f(('bool', 'bytes9', 'uint168'), D), T[bo, by, i])  # fixed case 36376
ck(f(('bool', 'bytes9', 'uint176'), D), T[bo, by, i])  # fixed case 36377
ck(f(('bool', 'bytes9', 'uint184'), D), T[bo, by, i])  # fixed case 36378
ck(f(('bool', 'bytes9', 'uint192'), D), T[bo, by, i])  # fixed case 36379
ck(f(('bool', 'bytes9', 'uint200'), D), T[bo, by, i])  # fixed case 36380
ck(f(('bool', 'bytes9', 'uint208'), D), T[bo, by, i])  # fixed case 36381
ck(f(('bool', 'bytes9', 'uint216'), D), T[bo, by, i])  # fixed case 36382
ck(f(('bool', 'bytes9', 'uint224'), D), T[bo, by, i])  # fixed case 36383
ck(f(('bool', 'bytes9', 'uint232'), D), T[bo, by, i])  # fixed case 36384
ck(f(('bool', 'bytes9', 'uint24'), D), T[bo, by, i])  # fixed case 36385
ck(f(('bool', 'bytes9', 'uint240'), D), T[bo, by, i])  # fixed case 36386
ck(f(('bool', 'bytes9', 'uint248'), D), T[bo, by, i])  # fixed case 36387
ck(f(('bool', 'bytes9', 'uint256'), D), T[bo, by, i])  # fixed case 36388
ck(f(('bool', 'bytes9', 'uint32'), D), T[bo, by, i])  # fixed case 36389
ck(f(('bool', 'bytes9', 'uint40'), D), T[bo, by, i])  # fixed case 36390
ck(f(('bool', 'bytes9', 'uint48'), D), T[bo, by, i])  # fixed case 36391
ck(f(('bool', 'bytes9', 'uint56'), D), T[bo, by, i])  # fixed case 36392
ck(f(('bool', 'bytes9', 'uint64'), D), T[bo, by, i])  # fixed case 36393
ck(f(('bool', 'bytes9', 'uint72'), D), T[bo, by, i])  # fixed case 36394
ck(f(('bool', 'bytes9', 'uint8'), D), T[bo, by, i])  # fixed case 36395
ck(f(('bool', 'bytes9', 'uint80'), D), T[bo, by, i])  # fixed case 36396
ck(f(('bool', 'bytes9', 'uint88'), D), T[bo, by, i])  # fixed case 36397
ck(f(('bool', 'bytes9', 'uint96'), D), T[bo, by, i])  # fixed case 36398
ck(f(('bool', 'bytes9', '?'), D), T[bo, by, A])  # fixed case 36399
ck(f(('bool', 'int104', 'address'), D), T[bo, i, HA])  # fixed case 36400
ck(f(('bool', 'int104', 'address[]'), D), T[bo, i, A])  # fixed case 36401
ck(f(('bool', 'int104', 'bool'), D), T[bo, i, bo])  # fixed case 36402
ck(f(('bool', 'int104', 'bool[]'), D), T[bo, i, A])  # fixed case 36403
ck(f(('bool', 'int104', 'bytes'), D), T[bo, i, by])  # fixed case 36404
ck(f(('bool', 'int104', 'bytes1'), D), T[bo, i, by])  # fixed case 36405
ck(f(('bool', 'int104', 'bytes10'), D), T[bo, i, by])  # fixed case 36406
ck(f(('bool', 'int104', 'bytes11'), D), T[bo, i, by])  # fixed case 36407
ck(f(('bool', 'int104', 'bytes12'), D), T[bo, i, by])  # fixed case 36408
ck(f(('bool', 'int104', 'bytes13'), D), T[bo, i, by])  # fixed case 36409
ck(f(('bool', 'int104', 'bytes14'), D), T[bo, i, by])  # fixed case 36410
ck(f(('bool', 'int104', 'bytes15'), D), T[bo, i, by])  # fixed case 36411
ck(f(('bool', 'int104', 'bytes16'), D), T[bo, i, by])  # fixed case 36412
ck(f(('bool', 'int104', 'bytes17'), D), T[bo, i, by])  # fixed case 36413
ck(f(('bool', 'int104', 'bytes18'), D), T[bo, i, by])  # fixed case 36414
ck(f(('bool', 'int104', 'bytes19'), D), T[bo, i, by])  # fixed case 36415
ck(f(('bool', 'int104', 'bytes2'), D), T[bo, i, by])  # fixed case 36416
ck(f(('bool', 'int104', 'bytes20'), D), T[bo, i, by])  # fixed case 36417
ck(f(('bool', 'int104', 'bytes21'), D), T[bo, i, by])  # fixed case 36418
ck(f(('bool', 'int104', 'bytes22'), D), T[bo, i, by])  # fixed case 36419
ck(f(('bool', 'int104', 'bytes23'), D), T[bo, i, by])  # fixed case 36420
ck(f(('bool', 'int104', 'bytes24'), D), T[bo, i, by])  # fixed case 36421
ck(f(('bool', 'int104', 'bytes25'), D), T[bo, i, by])  # fixed case 36422
ck(f(('bool', 'int104', 'bytes26'), D), T[bo, i, by])  # fixed case 36423
ck(f(('bool', 'int104', 'bytes27'), D), T[bo, i, by])  # fixed case 36424
ck(f(('bool', 'int104', 'bytes28'), D), T[bo, i, by])  # fixed case 36425
ck(f(('bool', 'int104', 'bytes29'), D), T[bo, i, by])  # fixed case 36426
ck(f(('bool', 'int104', 'bytes3'), D), T[bo, i, by])  # fixed case 36427
ck(f(('bool', 'int104', 'bytes30'), D), T[bo, i, by])  # fixed case 36428
ck(f(('bool', 'int104', 'bytes31'), D), T[bo, i, by])  # fixed case 36429
ck(f(('bool', 'int104', 'bytes32'), D), T[bo, i, by])  # fixed case 36430
ck(f(('bool', 'int104', 'bytes4'), D), T[bo, i, by])  # fixed case 36431
ck(f(('bool', 'int104', 'bytes5'), D), T[bo, i, by])  # fixed case 36432
ck(f(('bool', 'int104', 'bytes6'), D), T[bo, i, by])  # fixed case 36433
ck(f(('bool', 'int104', 'bytes7'), D), T[bo, i, by])  # fixed case 36434
ck(f(('bool', 'int104', 'bytes8'), D), T[bo, i, by])  # fixed case 36435
ck(f(('bool', 'int104', 'bytes9'), D), T[bo, i, by])  # fixed case 36436
ck(f(('bool', 'int104', 'int104'), D), T[bo, i, i])  # fixed case 36437
ck(f(('bool', 'int104', 'int112'), D), T[bo, i, i])  # fixed case 36438
ck(f(('bool', 'int104', 'int120'), D), T[bo, i, i])  # fixed case 36439
ck(f(('bool', 'int104', 'int128'), D), T[bo, i, i])  # fixed case 36440
ck(f(('bool', 'int104', 'int136'), D), T[bo, i, i])  # fixed case 36441
ck(f(('bool', 'int104', 'int144'), D), T[bo, i, i])  # fixed case 36442
ck(f(('bool', 'int104', 'int152'), D), T[bo, i, i])  # fixed case 36443
ck(f(('bool', 'int104', 'int16'), D), T[bo, i, i])  # fixed case 36444
ck(f(('bool', 'int104', 'int160'), D), T[bo, i, i])  # fixed case 36445
ck(f(('bool', 'int104', 'int168'), D), T[bo, i, i])  # fixed case 36446
ck(f(('bool', 'int104', 'int176'), D), T[bo, i, i])  # fixed case 36447
ck(f(('bool', 'int104', 'int184'), D), T[bo, i, i])  # fixed case 36448
ck(f(('bool', 'int104', 'int192'), D), T[bo, i, i])  # fixed case 36449
ck(f(('bool', 'int104', 'int200'), D), T[bo, i, i])  # fixed case 36450
ck(f(('bool', 'int104', 'int208'), D), T[bo, i, i])  # fixed case 36451
ck(f(('bool', 'int104', 'int216'), D), T[bo, i, i])  # fixed case 36452
ck(f(('bool', 'int104', 'int224'), D), T[bo, i, i])  # fixed case 36453
ck(f(('bool', 'int104', 'int232'), D), T[bo, i, i])  # fixed case 36454
ck(f(('bool', 'int104', 'int24'), D), T[bo, i, i])  # fixed case 36455
ck(f(('bool', 'int104', 'int240'), D), T[bo, i, i])  # fixed case 36456
ck(f(('bool', 'int104', 'int248'), D), T[bo, i, i])  # fixed case 36457
ck(f(('bool', 'int104', 'int256'), D), T[bo, i, i])  # fixed case 36458
ck(f(('bool', 'int104', 'int32'), D), T[bo, i, i])  # fixed case 36459
ck(f(('bool', 'int104', 'int40'), D), T[bo, i, i])  # fixed case 36460
ck(f(('bool', 'int104', 'int48'), D), T[bo, i, i])  # fixed case 36461
ck(f(('bool', 'int104', 'int56'), D), T[bo, i, i])  # fixed case 36462
ck(f(('bool', 'int104', 'int64'), D), T[bo, i, i])  # fixed case 36463
ck(f(('bool', 'int104', 'int72'), D), T[bo, i, i])  # fixed case 36464
ck(f(('bool', 'int104', 'int8'), D), T[bo, i, i])  # fixed case 36465
ck(f(('bool', 'int104', 'int80'), D), T[bo, i, i])  # fixed case 36466
ck(f(('bool', 'int104', 'int88'), D), T[bo, i, i])  # fixed case 36467
ck(f(('bool', 'int104', 'int96'), D), T[bo, i, i])  # fixed case 36468
ck(f(('bool', 'int104', 'string'), D), T[bo, i, s])  # fixed case 36469
ck(f(('bool', 'int104', 'string[]'), D), T[bo, i, A])  # fixed case 36470
ck(f(('bool', 'int104', 'uint104'), D), T[bo, i, i])  # fixed case 36471
ck(f(('bool', 'int104', 'uint112'), D), T[bo, i, i])  # fixed case 36472
ck(f(('bool', 'int104', 'uint120'), D), T[bo, i, i])  # fixed case 36473
ck(f(('bool', 'int104', 'uint128'), D), T[bo, i, i])  # fixed case 36474
ck(f(('bool', 'int104', 'uint136'), D), T[bo, i, i])  # fixed case 36475
ck(f(('bool', 'int104', 'uint144'), D), T[bo, i, i])  # fixed case 36476
ck(f(('bool', 'int104', 'uint152'), D), T[bo, i, i])  # fixed case 36477
ck(f(('bool', 'int104', 'uint16'), D), T[bo, i, i])  # fixed case 36478
ck(f(('bool', 'int104', 'uint160'), D), T[bo, i, i])  # fixed case 36479
ck(f(('bool', 'int104', 'uint168'), D), T[bo, i, i])  # fixed case 36480
ck(f(('bool', 'int104', 'uint176'), D), T[bo, i, i])  # fixed case 36481
ck(f(('bool', 'int104', 'uint184'), D), T[bo, i, i])  # fixed case 36482
ck(f(('bool', 'int104', 'uint192'), D), T[bo, i, i])  # fixed case 36483
ck(f(('bool', 'int104', 'uint200'), D), T[bo, i, i])  # fixed case 36484
ck(f(('bool', 'int104', 'uint208'), D), T[bo, i, i])  # fixed case 36485
ck(f(('bool', 'int104', 'uint216'), D), T[bo, i, i])  # fixed case 36486
ck(f(('bool', 'int104', 'uint224'), D), T[bo, i, i])  # fixed case 36487
ck(f(('bool', 'int104', 'uint232'), D), T[bo, i, i])  # fixed case 36488
ck(f(('bool', 'int104', 'uint24'), D), T[bo, i, i])  # fixed case 36489
ck(f(('bool', 'int104', 'uint240'), D), T[bo, i, i])  # fixed case 36490
ck(f(('bool', 'int104', 'uint248'), D), T[bo, i, i])  # fixed case 36491
ck(f(('bool', 'int104', 'uint256'), D), T[bo, i, i])  # fixed case 36492
ck(f(('bool', 'int104', 'uint32'), D), T[bo, i, i])  # fixed case 36493
ck(f(('bool', 'int104', 'uint40'), D), T[bo, i, i])  # fixed case 36494
ck(f(('bool', 'int104', 'uint48'), D), T[bo, i, i])  # fixed case 36495
ck(f(('bool', 'int104', 'uint56'), D), T[bo, i, i])  # fixed case 36496
ck(f(('bool', 'int104', 'uint64'), D), T[bo, i, i])  # fixed case 36497
ck(f(('bool', 'int104', 'uint72'), D), T[bo, i, i])  # fixed case 36498
ck(f(('bool', 'int104', 'uint8'), D), T[bo, i, i])  # fixed case 36499
ck(f(('bool', 'int104', 'uint80'), D), T[bo, i, i])  # fixed case 36500
ck(f(('bool', 'int104', 'uint88'), D), T[bo, i, i])  # fixed case 36501
ck(f(('bool', 'int104', 'uint96'), D), T[bo, i, i])  # fixed case 36502
ck(f(('bool', 'int104', '?'), D), T[bo, i, A])  # fixed case 36503
ck(f(('bool', 'int112', 'address'), D), T[bo, i, HA])  # fixed case 36504
ck(f(('bool', 'int112', 'address[]'), D), T[bo, i, A])  # fixed case 36505
ck(f(('bool', 'int112', 'bool'), D), T[bo, i, bo])  # fixed case 36506
ck(f(('bool', 'int112', 'bool[]'), D), T[bo, i, A])  # fixed case 36507
ck(f(('bool', 'int112', 'bytes'), D), T[bo, i, by])  # fixed case 36508
ck(f(('bool', 'int112', 'bytes1'), D), T[bo, i, by])  # fixed case 36509
ck(f(('bool', 'int112', 'bytes10'), D), T[bo, i, by])  # fixed case 36510
ck(f(('bool', 'int112', 'bytes11'), D), T[bo, i, by])  # fixed case 36511
ck(f(('bool', 'int112', 'bytes12'), D), T[bo, i, by])  # fixed case 36512
ck(f(('bool', 'int112', 'bytes13'), D), T[bo, i, by])  # fixed case 36513
ck(f(('bool', 'int112', 'bytes14'), D), T[bo, i, by])  # fixed case 36514
ck(f(('bool', 'int112', 'bytes15'), D), T[bo, i, by])  # fixed case 36515
ck(f(('bool', 'int112', 'bytes16'), D), T[bo, i, by])  # fixed case 36516
ck(f(('bool', 'int112', 'bytes17'), D), T[bo, i, by])  # fixed case 36517
ck(f(('bool', 'int112', 'bytes18'), D), T[bo, i, by])  # fixed case 36518
ck(f(('bool', 'int112', 'bytes19'), D), T[bo, i, by])  # fixed case 36519
ck(f(('bool', 'int112', 'bytes2'), D), T[bo, i, by])  # fixed case 36520
ck(f(('bool', 'int112', 'bytes20'), D), T[bo, i, by])  # fixed case 36521
ck(f(('bool', 'int112', 'bytes21'), D), T[bo, i, by])  # fixed case 36522
ck(f(('bool', 'int112', 'bytes22'), D), T[bo, i, by])  # fixed case 36523
ck(f(('bool', 'int112', 'bytes23'), D), T[bo, i, by])  # fixed case 36524
ck(f(('bool', 'int112', 'bytes24'), D), T[bo, i, by])  # fixed case 36525
ck(f(('bool', 'int112', 'bytes25'), D), T[bo, i, by])  # fixed case 36526
ck(f(('bool', 'int112', 'bytes26'), D), T[bo, i, by])  # fixed case 36527
ck(f(('bool', 'int112', 'bytes27'), D), T[bo, i, by])  # fixed case 36528
ck(f(('bool', 'int112', 'bytes28'), D), T[bo, i, by])  # fixed case 36529
ck(f(('bool', 'int112', 'bytes29'), D), T[bo, i, by])  # fixed case 36530
ck(f(('bool', 'int112', 'bytes3'), D), T[bo, i, by])  # fixed case 36531
ck(f(('bool', 'int112', 'bytes30'), D), T[bo, i, by])  # fixed case 36532
ck(f(('bool', 'int112', 'bytes31'), D), T[bo, i, by])  # fixed case 36533
ck(f(('bool', 'int112', 'bytes32'), D), T[bo, i, by])  # fixed case 36534
ck(f(('bool', 'int112', 'bytes4'), D), T[bo, i, by])  # fixed case 36535
ck(f(('bool', 'int112', 'bytes5'), D), T[bo, i, by])  # fixed case 36536
ck(f(('bool', 'int112', 'bytes6'), D), T[bo, i, by])  # fixed case 36537
ck(f(('bool', 'int112', 'bytes7'), D), T[bo, i, by])  # fixed case 36538
ck(f(('bool', 'int112', 'bytes8'), D), T[bo, i, by])  # fixed case 36539
ck(f(('bool', 'int112', 'bytes9'), D), T[bo, i, by])  # fixed case 36540
ck(f(('bool', 'int112', 'int104'), D), T[bo, i, i])  # fixed case 36541
ck(f(('bool', 'int112', 'int112'), D), T[bo, i, i])  # fixed case 36542
ck(f(('bool', 'int112', 'int120'), D), T[bo, i, i])  # fixed case 36543
ck(f(('bool', 'int112', 'int128'), D), T[bo, i, i])  # fixed case 36544
ck(f(('bool', 'int112', 'int136'), D), T[bo, i, i])  # fixed case 36545
ck(f(('bool', 'int112', 'int144'), D), T[bo, i, i])  # fixed case 36546
ck(f(('bool', 'int112', 'int152'), D), T[bo, i, i])  # fixed case 36547
ck(f(('bool', 'int112', 'int16'), D), T[bo, i, i])  # fixed case 36548
ck(f(('bool', 'int112', 'int160'), D), T[bo, i, i])  # fixed case 36549
ck(f(('bool', 'int112', 'int168'), D), T[bo, i, i])  # fixed case 36550
ck(f(('bool', 'int112', 'int176'), D), T[bo, i, i])  # fixed case 36551
ck(f(('bool', 'int112', 'int184'), D), T[bo, i, i])  # fixed case 36552
ck(f(('bool', 'int112', 'int192'), D), T[bo, i, i])  # fixed case 36553
ck(f(('bool', 'int112', 'int200'), D), T[bo, i, i])  # fixed case 36554
ck(f(('bool', 'int112', 'int208'), D), T[bo, i, i])  # fixed case 36555
ck(f(('bool', 'int112', 'int216'), D), T[bo, i, i])  # fixed case 36556
ck(f(('bool', 'int112', 'int224'), D), T[bo, i, i])  # fixed case 36557
ck(f(('bool', 'int112', 'int232'), D), T[bo, i, i])  # fixed case 36558
ck(f(('bool', 'int112', 'int24'), D), T[bo, i, i])  # fixed case 36559
ck(f(('bool', 'int112', 'int240'), D), T[bo, i, i])  # fixed case 36560
ck(f(('bool', 'int112', 'int248'), D), T[bo, i, i])  # fixed case 36561
ck(f(('bool', 'int112', 'int256'), D), T[bo, i, i])  # fixed case 36562
ck(f(('bool', 'int112', 'int32'), D), T[bo, i, i])  # fixed case 36563
ck(f(('bool', 'int112', 'int40'), D), T[bo, i, i])  # fixed case 36564
ck(f(('bool', 'int112', 'int48'), D), T[bo, i, i])  # fixed case 36565
ck(f(('bool', 'int112', 'int56'), D), T[bo, i, i])  # fixed case 36566
ck(f(('bool', 'int112', 'int64'), D), T[bo, i, i])  # fixed case 36567
ck(f(('bool', 'int112', 'int72'), D), T[bo, i, i])  # fixed case 36568
ck(f(('bool', 'int112', 'int8'), D), T[bo, i, i])  # fixed case 36569
ck(f(('bool', 'int112', 'int80'), D), T[bo, i, i])  # fixed case 36570
ck(f(('bool', 'int112', 'int88'), D), T[bo, i, i])  # fixed case 36571
ck(f(('bool', 'int112', 'int96'), D), T[bo, i, i])  # fixed case 36572
ck(f(('bool', 'int112', 'string'), D), T[bo, i, s])  # fixed case 36573
ck(f(('bool', 'int112', 'string[]'), D), T[bo, i, A])  # fixed case 36574
ck(f(('bool', 'int112', 'uint104'), D), T[bo, i, i])  # fixed case 36575
ck(f(('bool', 'int112', 'uint112'), D), T[bo, i, i])  # fixed case 36576
ck(f(('bool', 'int112', 'uint120'), D), T[bo, i, i])  # fixed case 36577
ck(f(('bool', 'int112', 'uint128'), D), T[bo, i, i])  # fixed case 36578
ck(f(('bool', 'int112', 'uint136'), D), T[bo, i, i])  # fixed case 36579
ck(f(('bool', 'int112', 'uint144'), D), T[bo, i, i])  # fixed case 36580
ck(f(('bool', 'int112', 'uint152'), D), T[bo, i, i])  # fixed case 36581
ck(f(('bool', 'int112', 'uint16'), D), T[bo, i, i])  # fixed case 36582
ck(f(('bool', 'int112', 'uint160'), D), T[bo, i, i])  # fixed case 36583
ck(f(('bool', 'int112', 'uint168'), D), T[bo, i, i])  # fixed case 36584
ck(f(('bool', 'int112', 'uint176'), D), T[bo, i, i])  # fixed case 36585
ck(f(('bool', 'int112', 'uint184'), D), T[bo, i, i])  # fixed case 36586
ck(f(('bool', 'int112', 'uint192'), D), T[bo, i, i])  # fixed case 36587
ck(f(('bool', 'int112', 'uint200'), D), T[bo, i, i])  # fixed case 36588
ck(f(('bool', 'int112', 'uint208'), D), T[bo, i, i])  # fixed case 36589
ck(f(('bool', 'int112', 'uint216'), D), T[bo, i, i])  # fixed case 36590
ck(f(('bool', 'int112', 'uint224'), D), T[bo, i, i])  # fixed case 36591
ck(f(('bool', 'int112', 'uint232'), D), T[bo, i, i])  # fixed case 36592
ck(f(('bool', 'int112', 'uint24'), D), T[bo, i, i])  # fixed case 36593
ck(f(('bool', 'int112', 'uint240'), D), T[bo, i, i])  # fixed case 36594
ck(f(('bool', 'int112', 'uint248'), D), T[bo, i, i])  # fixed case 36595
ck(f(('bool', 'int112', 'uint256'), D), T[bo, i, i])  # fixed case 36596
ck(f(('bool', 'int112', 'uint32'), D), T[bo, i, i])  # fixed case 36597
ck(f(('bool', 'int112', 'uint40'), D), T[bo, i, i])  # fixed case 36598
ck(f(('bool', 'int112', 'uint48'), D), T[bo, i, i])  # fixed case 36599
ck(f(('bool', 'int112', 'uint56'), D), T[bo, i, i])  # fixed case 36600
ck(f(('bool', 'int112', 'uint64'), D), T[bo, i, i])  # fixed case 36601
ck(f(('bool', 'int112', 'uint72'), D), T[bo, i, i])  # fixed case 36602
ck(f(('bool', 'int112', 'uint8'), D), T[bo, i, i])  # fixed case 36603
ck(f(('bool', 'int112', 'uint80'), D), T[bo, i, i])  # fixed case 36604
ck(f(('bool', 'int112', 'uint88'), D), T[bo, i, i])  # fixed case 36605
ck(f(('bool', 'int112', 'uint96'), D), T[bo, i, i])  # fixed case 36606
ck(f(('bool', 'int112', '?'), D), T[bo, i, A])  # fixed case 36607
ck(f(('bool', 'int120', 'address'), D), T[bo, i, HA])  # fixed case 36608
ck(f(('bool', 'int120', 'address[]'), D), T[bo, i, A])  # fixed case 36609
ck(f(('bool', 'int120', 'bool'), D), T[bo, i, bo])  # fixed case 36610
ck(f(('bool', 'int120', 'bool[]'), D), T[bo, i, A])  # fixed case 36611
ck(f(('bool', 'int120', 'bytes'), D), T[bo, i, by])  # fixed case 36612
ck(f(('bool', 'int120', 'bytes1'), D), T[bo, i, by])  # fixed case 36613
ck(f(('bool', 'int120', 'bytes10'), D), T[bo, i, by])  # fixed case 36614
ck(f(('bool', 'int120', 'bytes11'), D), T[bo, i, by])  # fixed case 36615
ck(f(('bool', 'int120', 'bytes12'), D), T[bo, i, by])  # fixed case 36616
ck(f(('bool', 'int120', 'bytes13'), D), T[bo, i, by])  # fixed case 36617
ck(f(('bool', 'int120', 'bytes14'), D), T[bo, i, by])  # fixed case 36618
ck(f(('bool', 'int120', 'bytes15'), D), T[bo, i, by])  # fixed case 36619
ck(f(('bool', 'int120', 'bytes16'), D), T[bo, i, by])  # fixed case 36620
ck(f(('bool', 'int120', 'bytes17'), D), T[bo, i, by])  # fixed case 36621
ck(f(('bool', 'int120', 'bytes18'), D), T[bo, i, by])  # fixed case 36622
ck(f(('bool', 'int120', 'bytes19'), D), T[bo, i, by])  # fixed case 36623
ck(f(('bool', 'int120', 'bytes2'), D), T[bo, i, by])  # fixed case 36624
ck(f(('bool', 'int120', 'bytes20'), D), T[bo, i, by])  # fixed case 36625
ck(f(('bool', 'int120', 'bytes21'), D), T[bo, i, by])  # fixed case 36626
ck(f(('bool', 'int120', 'bytes22'), D), T[bo, i, by])  # fixed case 36627
ck(f(('bool', 'int120', 'bytes23'), D), T[bo, i, by])  # fixed case 36628
ck(f(('bool', 'int120', 'bytes24'), D), T[bo, i, by])  # fixed case 36629
ck(f(('bool', 'int120', 'bytes25'), D), T[bo, i, by])  # fixed case 36630
ck(f(('bool', 'int120', 'bytes26'), D), T[bo, i, by])  # fixed case 36631
ck(f(('bool', 'int120', 'bytes27'), D), T[bo, i, by])  # fixed case 36632
ck(f(('bool', 'int120', 'bytes28'), D), T[bo, i, by])  # fixed case 36633
ck(f(('bool', 'int120', 'bytes29'), D), T[bo, i, by])  # fixed case 36634
ck(f(('bool', 'int120', 'bytes3'), D), T[bo, i, by])  # fixed case 36635
ck(f(('bool', 'int120', 'bytes30'), D), T[bo, i, by])  # fixed case 36636
ck(f(('bool', 'int120', 'bytes31'), D), T[bo, i, by])  # fixed case 36637
ck(f(('bool', 'int120', 'bytes32'), D), T[bo, i, by])  # fixed case 36638
ck(f(('bool', 'int120', 'bytes4'), D), T[bo, i, by])  # fixed case 36639
ck(f(('bool', 'int120', 'bytes5'), D), T[bo, i, by])  # fixed case 36640
ck(f(('bool', 'int120', 'bytes6'), D), T[bo, i, by])  # fixed case 36641
ck(f(('bool', 'int120', 'bytes7'), D), T[bo, i, by])  # fixed case 36642
ck(f(('bool', 'int120', 'bytes8'), D), T[bo, i, by])  # fixed case 36643
ck(f(('bool', 'int120', 'bytes9'), D), T[bo, i, by])  # fixed case 36644
ck(f(('bool', 'int120', 'int104'), D), T[bo, i, i])  # fixed case 36645
ck(f(('bool', 'int120', 'int112'), D), T[bo, i, i])  # fixed case 36646
ck(f(('bool', 'int120', 'int120'), D), T[bo, i, i])  # fixed case 36647
ck(f(('bool', 'int120', 'int128'), D), T[bo, i, i])  # fixed case 36648
ck(f(('bool', 'int120', 'int136'), D), T[bo, i, i])  # fixed case 36649
ck(f(('bool', 'int120', 'int144'), D), T[bo, i, i])  # fixed case 36650
ck(f(('bool', 'int120', 'int152'), D), T[bo, i, i])  # fixed case 36651
ck(f(('bool', 'int120', 'int16'), D), T[bo, i, i])  # fixed case 36652
ck(f(('bool', 'int120', 'int160'), D), T[bo, i, i])  # fixed case 36653
ck(f(('bool', 'int120', 'int168'), D), T[bo, i, i])  # fixed case 36654
ck(f(('bool', 'int120', 'int176'), D), T[bo, i, i])  # fixed case 36655
ck(f(('bool', 'int120', 'int184'), D), T[bo, i, i])  # fixed case 36656
ck(f(('bool', 'int120', 'int192'), D), T[bo, i, i])  # fixed case 36657
ck(f(('bool', 'int120', 'int200'), D), T[bo, i, i])  # fixed case 36658
ck(f(('bool', 'int120', 'int208'), D), T[bo, i, i])  # fixed case 36659
ck(f(('bool', 'int120', 'int216'), D), T[bo, i, i])  # fixed case 36660
ck(f(('bool', 'int120', 'int224'), D), T[bo, i, i])  # fixed case 36661
ck(f(('bool', 'int120', 'int232'), D), T[bo, i, i])  # fixed case 36662
ck(f(('bool', 'int120', 'int24'), D), T[bo, i, i])  # fixed case 36663
ck(f(('bool', 'int120', 'int240'), D), T[bo, i, i])  # fixed case 36664
ck(f(('bool', 'int120', 'int248'), D), T[bo, i, i])  # fixed case 36665
ck(f(('bool', 'int120', 'int256'), D), T[bo, i, i])  # fixed case 36666
ck(f(('bool', 'int120', 'int32'), D), T[bo, i, i])  # fixed case 36667
ck(f(('bool', 'int120', 'int40'), D), T[bo, i, i])  # fixed case 36668
ck(f(('bool', 'int120', 'int48'), D), T[bo, i, i])  # fixed case 36669
ck(f(('bool', 'int120', 'int56'), D), T[bo, i, i])  # fixed case 36670
ck(f(('bool', 'int120', 'int64'), D), T[bo, i, i])  # fixed case 36671
ck(f(('bool', 'int120', 'int72'), D), T[bo, i, i])  # fixed case 36672
ck(f(('bool', 'int120', 'int8'), D), T[bo, i, i])  # fixed case 36673
ck(f(('bool', 'int120', 'int80'), D), T[bo, i, i])  # fixed case 36674
ck(f(('bool', 'int120', 'int88'), D), T[bo, i, i])  # fixed case 36675
ck(f(('bool', 'int120', 'int96'), D), T[bo, i, i])  # fixed case 36676
ck(f(('bool', 'int120', 'string'), D), T[bo, i, s])  # fixed case 36677
ck(f(('bool', 'int120', 'string[]'), D), T[bo, i, A])  # fixed case 36678
ck(f(('bool', 'int120', 'uint104'), D), T[bo, i, i])  # fixed case 36679
ck(f(('bool', 'int120', 'uint112'), D), T[bo, i, i])  # fixed case 36680
ck(f(('bool', 'int120', 'uint120'), D), T[bo, i, i])  # fixed case 36681
ck(f(('bool', 'int120', 'uint128'), D), T[bo, i, i])  # fixed case 36682
ck(f(('bool', 'int120', 'uint136'), D), T[bo, i, i])  # fixed case 36683
ck(f(('bool', 'int120', 'uint144'), D), T[bo, i, i])  # fixed case 36684
ck(f(('bool', 'int120', 'uint152'), D), T[bo, i, i])  # fixed case 36685
ck(f(('bool', 'int120', 'uint16'), D), T[bo, i, i])  # fixed case 36686
ck(f(('bool', 'int120', 'uint160'), D), T[bo, i, i])  # fixed case 36687
ck(f(('bool', 'int120', 'uint168'), D), T[bo, i, i])  # fixed case 36688
ck(f(('bool', 'int120', 'uint176'), D), T[bo, i, i])  # fixed case 36689
ck(f(('bool', 'int120', 'uint184'), D), T[bo, i, i])  # fixed case 36690
ck(f(('bool', 'int120', 'uint192'), D), T[bo, i, i])  # fixed case 36691
ck(f(('bool', 'int120', 'uint200'), D), T[bo, i, i])  # fixed case 36692
ck(f(('bool', 'int120', 'uint208'), D), T[bo, i, i])  # fixed case 36693
ck(f(('bool', 'int120', 'uint216'), D), T[bo, i, i])  # fixed case 36694
ck(f(('bool', 'int120', 'uint224'), D), T[bo, i, i])  # fixed case 36695
ck(f(('bool', 'int120', 'uint232'), D), T[bo, i, i])  # fixed case 36696
ck(f(('bool', 'int120', 'uint24'), D), T[bo, i, i])  # fixed case 36697
ck(f(('bool', 'int120', 'uint240'), D), T[bo, i, i])  # fixed case 36698
ck(f(('bool', 'int120', 'uint248'), D), T[bo, i, i])  # fixed case 36699
ck(f(('bool', 'int120', 'uint256'), D), T[bo, i, i])  # fixed case 36700
ck(f(('bool', 'int120', 'uint32'), D), T[bo, i, i])  # fixed case 36701
ck(f(('bool', 'int120', 'uint40'), D), T[bo, i, i])  # fixed case 36702
ck(f(('bool', 'int120', 'uint48'), D), T[bo, i, i])  # fixed case 36703
ck(f(('bool', 'int120', 'uint56'), D), T[bo, i, i])  # fixed case 36704
ck(f(('bool', 'int120', 'uint64'), D), T[bo, i, i])  # fixed case 36705
ck(f(('bool', 'int120', 'uint72'), D), T[bo, i, i])  # fixed case 36706
ck(f(('bool', 'int120', 'uint8'), D), T[bo, i, i])  # fixed case 36707
ck(f(('bool', 'int120', 'uint80'), D), T[bo, i, i])  # fixed case 36708
ck(f(('bool', 'int120', 'uint88'), D), T[bo, i, i])  # fixed case 36709
ck(f(('bool', 'int120', 'uint96'), D), T[bo, i, i])  # fixed case 36710
ck(f(('bool', 'int120', '?'), D), T[bo, i, A])  # fixed case 36711
ck(f(('bool', 'int128', 'address'), D), T[bo, i, HA])  # fixed case 36712
ck(f(('bool', 'int128', 'address[]'), D), T[bo, i, A])  # fixed case 36713
ck(f(('bool', 'int128', 'bool'), D), T[bo, i, bo])  # fixed case 36714
ck(f(('bool', 'int128', 'bool[]'), D), T[bo, i, A])  # fixed case 36715
ck(f(('bool', 'int128', 'bytes'), D), T[bo, i, by])  # fixed case 36716
ck(f(('bool', 'int128', 'bytes1'), D), T[bo, i, by])  # fixed case 36717
ck(f(('bool', 'int128', 'bytes10'), D), T[bo, i, by])  # fixed case 36718
ck(f(('bool', 'int128', 'bytes11'), D), T[bo, i, by])  # fixed case 36719
ck(f(('bool', 'int128', 'bytes12'), D), T[bo, i, by])  # fixed case 36720
ck(f(('bool', 'int128', 'bytes13'), D), T[bo, i, by])  # fixed case 36721
ck(f(('bool', 'int128', 'bytes14'), D), T[bo, i, by])  # fixed case 36722
ck(f(('bool', 'int128', 'bytes15'), D), T[bo, i, by])  # fixed case 36723
ck(f(('bool', 'int128', 'bytes16'), D), T[bo, i, by])  # fixed case 36724
ck(f(('bool', 'int128', 'bytes17'), D), T[bo, i, by])  # fixed case 36725
ck(f(('bool', 'int128', 'bytes18'), D), T[bo, i, by])  # fixed case 36726
ck(f(('bool', 'int128', 'bytes19'), D), T[bo, i, by])  # fixed case 36727
ck(f(('bool', 'int128', 'bytes2'), D), T[bo, i, by])  # fixed case 36728
ck(f(('bool', 'int128', 'bytes20'), D), T[bo, i, by])  # fixed case 36729
ck(f(('bool', 'int128', 'bytes21'), D), T[bo, i, by])  # fixed case 36730
ck(f(('bool', 'int128', 'bytes22'), D), T[bo, i, by])  # fixed case 36731
ck(f(('bool', 'int128', 'bytes23'), D), T[bo, i, by])  # fixed case 36732
ck(f(('bool', 'int128', 'bytes24'), D), T[bo, i, by])  # fixed case 36733
ck(f(('bool', 'int128', 'bytes25'), D), T[bo, i, by])  # fixed case 36734
ck(f(('bool', 'int128', 'bytes26'), D), T[bo, i, by])  # fixed case 36735
ck(f(('bool', 'int128', 'bytes27'), D), T[bo, i, by])  # fixed case 36736
ck(f(('bool', 'int128', 'bytes28'), D), T[bo, i, by])  # fixed case 36737
ck(f(('bool', 'int128', 'bytes29'), D), T[bo, i, by])  # fixed case 36738
ck(f(('bool', 'int128', 'bytes3'), D), T[bo, i, by])  # fixed case 36739
ck(f(('bool', 'int128', 'bytes30'), D), T[bo, i, by])  # fixed case 36740
ck(f(('bool', 'int128', 'bytes31'), D), T[bo, i, by])  # fixed case 36741
ck(f(('bool', 'int128', 'bytes32'), D), T[bo, i, by])  # fixed case 36742
ck(f(('bool', 'int128', 'bytes4'), D), T[bo, i, by])  # fixed case 36743
ck(f(('bool', 'int128', 'bytes5'), D), T[bo, i, by])  # fixed case 36744
ck(f(('bool', 'int128', 'bytes6'), D), T[bo, i, by])  # fixed case 36745
ck(f(('bool', 'int128', 'bytes7'), D), T[bo, i, by])  # fixed case 36746
ck(f(('bool', 'int128', 'bytes8'), D), T[bo, i, by])  # fixed case 36747
ck(f(('bool', 'int128', 'bytes9'), D), T[bo, i, by])  # fixed case 36748
ck(f(('bool', 'int128', 'int104'), D), T[bo, i, i])  # fixed case 36749
ck(f(('bool', 'int128', 'int112'), D), T[bo, i, i])  # fixed case 36750
ck(f(('bool', 'int128', 'int120'), D), T[bo, i, i])  # fixed case 36751
ck(f(('bool', 'int128', 'int128'), D), T[bo, i, i])  # fixed case 36752
ck(f(('bool', 'int128', 'int136'), D), T[bo, i, i])  # fixed case 36753
ck(f(('bool', 'int128', 'int144'), D), T[bo, i, i])  # fixed case 36754
ck(f(('bool', 'int128', 'int152'), D), T[bo, i, i])  # fixed case 36755
ck(f(('bool', 'int128', 'int16'), D), T[bo, i, i])  # fixed case 36756
ck(f(('bool', 'int128', 'int160'), D), T[bo, i, i])  # fixed case 36757
ck(f(('bool', 'int128', 'int168'), D), T[bo, i, i])  # fixed case 36758
ck(f(('bool', 'int128', 'int176'), D), T[bo, i, i])  # fixed case 36759
ck(f(('bool', 'int128', 'int184'), D), T[bo, i, i])  # fixed case 36760
ck(f(('bool', 'int128', 'int192'), D), T[bo, i, i])  # fixed case 36761
ck(f(('bool', 'int128', 'int200'), D), T[bo, i, i])  # fixed case 36762
ck(f(('bool', 'int128', 'int208'), D), T[bo, i, i])  # fixed case 36763
ck(f(('bool', 'int128', 'int216'), D), T[bo, i, i])  # fixed case 36764
ck(f(('bool', 'int128', 'int224'), D), T[bo, i, i])  # fixed case 36765
ck(f(('bool', 'int128', 'int232'), D), T[bo, i, i])  # fixed case 36766
ck(f(('bool', 'int128', 'int24'), D), T[bo, i, i])  # fixed case 36767
ck(f(('bool', 'int128', 'int240'), D), T[bo, i, i])  # fixed case 36768
ck(f(('bool', 'int128', 'int248'), D), T[bo, i, i])  # fixed case 36769
ck(f(('bool', 'int128', 'int256'), D), T[bo, i, i])  # fixed case 36770
ck(f(('bool', 'int128', 'int32'), D), T[bo, i, i])  # fixed case 36771
ck(f(('bool', 'int128', 'int40'), D), T[bo, i, i])  # fixed case 36772
ck(f(('bool', 'int128', 'int48'), D), T[bo, i, i])  # fixed case 36773
ck(f(('bool', 'int128', 'int56'), D), T[bo, i, i])  # fixed case 36774
ck(f(('bool', 'int128', 'int64'), D), T[bo, i, i])  # fixed case 36775
ck(f(('bool', 'int128', 'int72'), D), T[bo, i, i])  # fixed case 36776
ck(f(('bool', 'int128', 'int8'), D), T[bo, i, i])  # fixed case 36777
ck(f(('bool', 'int128', 'int80'), D), T[bo, i, i])  # fixed case 36778
ck(f(('bool', 'int128', 'int88'), D), T[bo, i, i])  # fixed case 36779
ck(f(('bool', 'int128', 'int96'), D), T[bo, i, i])  # fixed case 36780
ck(f(('bool', 'int128', 'string'), D), T[bo, i, s])  # fixed case 36781
ck(f(('bool', 'int128', 'string[]'), D), T[bo, i, A])  # fixed case 36782
ck(f(('bool', 'int128', 'uint104'), D), T[bo, i, i])  # fixed case 36783
ck(f(('bool', 'int128', 'uint112'), D), T[bo, i, i])  # fixed case 36784
ck(f(('bool', 'int128', 'uint120'), D), T[bo, i, i])  # fixed case 36785
ck(f(('bool', 'int128', 'uint128'), D), T[bo, i, i])  # fixed case 36786
ck(f(('bool', 'int128', 'uint136'), D), T[bo, i, i])  # fixed case 36787
ck(f(('bool', 'int128', 'uint144'), D), T[bo, i, i])  # fixed case 36788
ck(f(('bool', 'int128', 'uint152'), D), T[bo, i, i])  # fixed case 36789
ck(f(('bool', 'int128', 'uint16'), D), T[bo, i, i])  # fixed case 36790
ck(f(('bool', 'int128', 'uint160'), D), T[bo, i, i])  # fixed case 36791
ck(f(('bool', 'int128', 'uint168'), D), T[bo, i, i])  # fixed case 36792
ck(f(('bool', 'int128', 'uint176'), D), T[bo, i, i])  # fixed case 36793
ck(f(('bool', 'int128', 'uint184'), D), T[bo, i, i])  # fixed case 36794
ck(f(('bool', 'int128', 'uint192'), D), T[bo, i, i])  # fixed case 36795
ck(f(('bool', 'int128', 'uint200'), D), T[bo, i, i])  # fixed case 36796
ck(f(('bool', 'int128', 'uint208'), D), T[bo, i, i])  # fixed case 36797
ck(f(('bool', 'int128', 'uint216'), D), T[bo, i, i])  # fixed case 36798
ck(f(('bool', 'int128', 'uint224'), D), T[bo, i, i])  # fixed case 36799
ck(f(('bool', 'int128', 'uint232'), D), T[bo, i, i])  # fixed case 36800
ck(f(('bool', 'int128', 'uint24'), D), T[bo, i, i])  # fixed case 36801
ck(f(('bool', 'int128', 'uint240'), D), T[bo, i, i])  # fixed case 36802
ck(f(('bool', 'int128', 'uint248'), D), T[bo, i, i])  # fixed case 36803
ck(f(('bool', 'int128', 'uint256'), D), T[bo, i, i])  # fixed case 36804
ck(f(('bool', 'int128', 'uint32'), D), T[bo, i, i])  # fixed case 36805
ck(f(('bool', 'int128', 'uint40'), D), T[bo, i, i])  # fixed case 36806
ck(f(('bool', 'int128', 'uint48'), D), T[bo, i, i])  # fixed case 36807
ck(f(('bool', 'int128', 'uint56'), D), T[bo, i, i])  # fixed case 36808
ck(f(('bool', 'int128', 'uint64'), D), T[bo, i, i])  # fixed case 36809
ck(f(('bool', 'int128', 'uint72'), D), T[bo, i, i])  # fixed case 36810
ck(f(('bool', 'int128', 'uint8'), D), T[bo, i, i])  # fixed case 36811
ck(f(('bool', 'int128', 'uint80'), D), T[bo, i, i])  # fixed case 36812
ck(f(('bool', 'int128', 'uint88'), D), T[bo, i, i])  # fixed case 36813
ck(f(('bool', 'int128', 'uint96'), D), T[bo, i, i])  # fixed case 36814
ck(f(('bool', 'int128', '?'), D), T[bo, i, A])  # fixed case 36815
ck(f(('bool', 'int136', 'address'), D), T[bo, i, HA])  # fixed case 36816
ck(f(('bool', 'int136', 'address[]'), D), T[bo, i, A])  # fixed case 36817
ck(f(('bool', 'int136', 'bool'), D), T[bo, i, bo])  # fixed case 36818
ck(f(('bool', 'int136', 'bool[]'), D), T[bo, i, A])  # fixed case 36819
ck(f(('bool', 'int136', 'bytes'), D), T[bo, i, by])  # fixed case 36820
ck(f(('bool', 'int136', 'bytes1'), D), T[bo, i, by])  # fixed case 36821
ck(f(('bool', 'int136', 'bytes10'), D), T[bo, i, by])  # fixed case 36822
ck(f(('bool', 'int136', 'bytes11'), D), T[bo, i, by])  # fixed case 36823
ck(f(('bool', 'int136', 'bytes12'), D), T[bo, i, by])  # fixed case 36824
ck(f(('bool', 'int136', 'bytes13'), D), T[bo, i, by])  # fixed case 36825
ck(f(('bool', 'int136', 'bytes14'), D), T[bo, i, by])  # fixed case 36826
ck(f(('bool', 'int136', 'bytes15'), D), T[bo, i, by])  # fixed case 36827
ck(f(('bool', 'int136', 'bytes16'), D), T[bo, i, by])  # fixed case 36828
ck(f(('bool', 'int136', 'bytes17'), D), T[bo, i, by])  # fixed case 36829
ck(f(('bool', 'int136', 'bytes18'), D), T[bo, i, by])  # fixed case 36830
ck(f(('bool', 'int136', 'bytes19'), D), T[bo, i, by])  # fixed case 36831
ck(f(('bool', 'int136', 'bytes2'), D), T[bo, i, by])  # fixed case 36832
ck(f(('bool', 'int136', 'bytes20'), D), T[bo, i, by])  # fixed case 36833
ck(f(('bool', 'int136', 'bytes21'), D), T[bo, i, by])  # fixed case 36834
ck(f(('bool', 'int136', 'bytes22'), D), T[bo, i, by])  # fixed case 36835
ck(f(('bool', 'int136', 'bytes23'), D), T[bo, i, by])  # fixed case 36836
ck(f(('bool', 'int136', 'bytes24'), D), T[bo, i, by])  # fixed case 36837
ck(f(('bool', 'int136', 'bytes25'), D), T[bo, i, by])  # fixed case 36838
ck(f(('bool', 'int136', 'bytes26'), D), T[bo, i, by])  # fixed case 36839
ck(f(('bool', 'int136', 'bytes27'), D), T[bo, i, by])  # fixed case 36840
ck(f(('bool', 'int136', 'bytes28'), D), T[bo, i, by])  # fixed case 36841
ck(f(('bool', 'int136', 'bytes29'), D), T[bo, i, by])  # fixed case 36842
ck(f(('bool', 'int136', 'bytes3'), D), T[bo, i, by])  # fixed case 36843
ck(f(('bool', 'int136', 'bytes30'), D), T[bo, i, by])  # fixed case 36844
ck(f(('bool', 'int136', 'bytes31'), D), T[bo, i, by])  # fixed case 36845
ck(f(('bool', 'int136', 'bytes32'), D), T[bo, i, by])  # fixed case 36846
ck(f(('bool', 'int136', 'bytes4'), D), T[bo, i, by])  # fixed case 36847
ck(f(('bool', 'int136', 'bytes5'), D), T[bo, i, by])  # fixed case 36848
ck(f(('bool', 'int136', 'bytes6'), D), T[bo, i, by])  # fixed case 36849
ck(f(('bool', 'int136', 'bytes7'), D), T[bo, i, by])  # fixed case 36850
ck(f(('bool', 'int136', 'bytes8'), D), T[bo, i, by])  # fixed case 36851
ck(f(('bool', 'int136', 'bytes9'), D), T[bo, i, by])  # fixed case 36852
ck(f(('bool', 'int136', 'int104'), D), T[bo, i, i])  # fixed case 36853
ck(f(('bool', 'int136', 'int112'), D), T[bo, i, i])  # fixed case 36854
ck(f(('bool', 'int136', 'int120'), D), T[bo, i, i])  # fixed case 36855
ck(f(('bool', 'int136', 'int128'), D), T[bo, i, i])  # fixed case 36856
ck(f(('bool', 'int136', 'int136'), D), T[bo, i, i])  # fixed case 36857
ck(f(('bool', 'int136', 'int144'), D), T[bo, i, i])  # fixed case 36858
ck(f(('bool', 'int136', 'int152'), D), T[bo, i, i])  # fixed case 36859
ck(f(('bool', 'int136', 'int16'), D), T[bo, i, i])  # fixed case 36860
ck(f(('bool', 'int136', 'int160'), D), T[bo, i, i])  # fixed case 36861
ck(f(('bool', 'int136', 'int168'), D), T[bo, i, i])  # fixed case 36862
ck(f(('bool', 'int136', 'int176'), D), T[bo, i, i])  # fixed case 36863
ck(f(('bool', 'int136', 'int184'), D), T[bo, i, i])  # fixed case 36864
ck(f(('bool', 'int136', 'int192'), D), T[bo, i, i])  # fixed case 36865
ck(f(('bool', 'int136', 'int200'), D), T[bo, i, i])  # fixed case 36866
ck(f(('bool', 'int136', 'int208'), D), T[bo, i, i])  # fixed case 36867
ck(f(('bool', 'int136', 'int216'), D), T[bo, i, i])  # fixed case 36868
ck(f(('bool', 'int136', 'int224'), D), T[bo, i, i])  # fixed case 36869
ck(f(('bool', 'int136', 'int232'), D), T[bo, i, i])  # fixed case 36870
ck(f(('bool', 'int136', 'int24'), D), T[bo, i, i])  # fixed case 36871
ck(f(('bool', 'int136', 'int240'), D), T[bo, i, i])  # fixed case 36872
ck(f(('bool', 'int136', 'int248'), D), T[bo, i, i])  # fixed case 36873
ck(f(('bool', 'int136', 'int256'), D), T[bo, i, i])  # fixed case 36874
ck(f(('bool', 'int136', 'int32'), D), T[bo, i, i])  # fixed case 36875
ck(f(('bool', 'int136', 'int40'), D), T[bo, i, i])  # fixed case 36876
ck(f(('bool', 'int136', 'int48'), D), T[bo, i, i])  # fixed case 36877
ck(f(('bool', 'int136', 'int56'), D), T[bo, i, i])  # fixed case 36878
ck(f(('bool', 'int136', 'int64'), D), T[bo, i, i])  # fixed case 36879
ck(f(('bool', 'int136', 'int72'), D), T[bo, i, i])  # fixed case 36880
ck(f(('bool', 'int136', 'int8'), D), T[bo, i, i])  # fixed case 36881
ck(f(('bool', 'int136', 'int80'), D), T[bo, i, i])  # fixed case 36882
ck(f(('bool', 'int136', 'int88'), D), T[bo, i, i])  # fixed case 36883
ck(f(('bool', 'int136', 'int96'), D), T[bo, i, i])  # fixed case 36884
ck(f(('bool', 'int136', 'string'), D), T[bo, i, s])  # fixed case 36885
ck(f(('bool', 'int136', 'string[]'), D), T[bo, i, A])  # fixed case 36886
ck(f(('bool', 'int136', 'uint104'), D), T[bo, i, i])  # fixed case 36887
ck(f(('bool', 'int136', 'uint112'), D), T[bo, i, i])  # fixed case 36888
ck(f(('bool', 'int136', 'uint120'), D), T[bo, i, i])  # fixed case 36889
ck(f(('bool', 'int136', 'uint128'), D), T[bo, i, i])  # fixed case 36890
ck(f(('bool', 'int136', 'uint136'), D), T[bo, i, i])  # fixed case 36891
ck(f(('bool', 'int136', 'uint144'), D), T[bo, i, i])  # fixed case 36892
ck(f(('bool', 'int136', 'uint152'), D), T[bo, i, i])  # fixed case 36893
ck(f(('bool', 'int136', 'uint16'), D), T[bo, i, i])  # fixed case 36894
ck(f(('bool', 'int136', 'uint160'), D), T[bo, i, i])  # fixed case 36895
ck(f(('bool', 'int136', 'uint168'), D), T[bo, i, i])  # fixed case 36896
ck(f(('bool', 'int136', 'uint176'), D), T[bo, i, i])  # fixed case 36897
ck(f(('bool', 'int136', 'uint184'), D), T[bo, i, i])  # fixed case 36898
ck(f(('bool', 'int136', 'uint192'), D), T[bo, i, i])  # fixed case 36899
ck(f(('bool', 'int136', 'uint200'), D), T[bo, i, i])  # fixed case 36900
ck(f(('bool', 'int136', 'uint208'), D), T[bo, i, i])  # fixed case 36901
ck(f(('bool', 'int136', 'uint216'), D), T[bo, i, i])  # fixed case 36902
ck(f(('bool', 'int136', 'uint224'), D), T[bo, i, i])  # fixed case 36903
ck(f(('bool', 'int136', 'uint232'), D), T[bo, i, i])  # fixed case 36904
ck(f(('bool', 'int136', 'uint24'), D), T[bo, i, i])  # fixed case 36905
ck(f(('bool', 'int136', 'uint240'), D), T[bo, i, i])  # fixed case 36906
ck(f(('bool', 'int136', 'uint248'), D), T[bo, i, i])  # fixed case 36907
ck(f(('bool', 'int136', 'uint256'), D), T[bo, i, i])  # fixed case 36908
ck(f(('bool', 'int136', 'uint32'), D), T[bo, i, i])  # fixed case 36909
ck(f(('bool', 'int136', 'uint40'), D), T[bo, i, i])  # fixed case 36910
ck(f(('bool', 'int136', 'uint48'), D), T[bo, i, i])  # fixed case 36911
ck(f(('bool', 'int136', 'uint56'), D), T[bo, i, i])  # fixed case 36912
ck(f(('bool', 'int136', 'uint64'), D), T[bo, i, i])  # fixed case 36913
ck(f(('bool', 'int136', 'uint72'), D), T[bo, i, i])  # fixed case 36914
ck(f(('bool', 'int136', 'uint8'), D), T[bo, i, i])  # fixed case 36915
ck(f(('bool', 'int136', 'uint80'), D), T[bo, i, i])  # fixed case 36916
ck(f(('bool', 'int136', 'uint88'), D), T[bo, i, i])  # fixed case 36917
ck(f(('bool', 'int136', 'uint96'), D), T[bo, i, i])  # fixed case 36918
ck(f(('bool', 'int136', '?'), D), T[bo, i, A])  # fixed case 36919
ck(f(('bool', 'int144', 'address'), D), T[bo, i, HA])  # fixed case 36920
ck(f(('bool', 'int144', 'address[]'), D), T[bo, i, A])  # fixed case 36921
ck(f(('bool', 'int144', 'bool'), D), T[bo, i, bo])  # fixed case 36922
ck(f(('bool', 'int144', 'bool[]'), D), T[bo, i, A])  # fixed case 36923
ck(f(('bool', 'int144', 'bytes'), D), T[bo, i, by])  # fixed case 36924
ck(f(('bool', 'int144', 'bytes1'), D), T[bo, i, by])  # fixed case 36925
ck(f(('bool', 'int144', 'bytes10'), D), T[bo, i, by])  # fixed case 36926
ck(f(('bool', 'int144', 'bytes11'), D), T[bo, i, by])  # fixed case 36927
ck(f(('bool', 'int144', 'bytes12'), D), T[bo, i, by])  # fixed case 36928
ck(f(('bool', 'int144', 'bytes13'), D), T[bo, i, by])  # fixed case 36929
ck(f(('bool', 'int144', 'bytes14'), D), T[bo, i, by])  # fixed case 36930
ck(f(('bool', 'int144', 'bytes15'), D), T[bo, i, by])  # fixed case 36931
ck(f(('bool', 'int144', 'bytes16'), D), T[bo, i, by])  # fixed case 36932
ck(f(('bool', 'int144', 'bytes17'), D), T[bo, i, by])  # fixed case 36933
ck(f(('bool', 'int144', 'bytes18'), D), T[bo, i, by])  # fixed case 36934
ck(f(('bool', 'int144', 'bytes19'), D), T[bo, i, by])  # fixed case 36935
ck(f(('bool', 'int144', 'bytes2'), D), T[bo, i, by])  # fixed case 36936
ck(f(('bool', 'int144', 'bytes20'), D), T[bo, i, by])  # fixed case 36937
ck(f(('bool', 'int144', 'bytes21'), D), T[bo, i, by])  # fixed case 36938
ck(f(('bool', 'int144', 'bytes22'), D), T[bo, i, by])  # fixed case 36939
ck(f(('bool', 'int144', 'bytes23'), D), T[bo, i, by])  # fixed case 36940
ck(f(('bool', 'int144', 'bytes24'), D), T[bo, i, by])  # fixed case 36941
ck(f(('bool', 'int144', 'bytes25'), D), T[bo, i, by])  # fixed case 36942
ck(f(('bool', 'int144', 'bytes26'), D), T[bo, i, by])  # fixed case 36943
ck(f(('bool', 'int144', 'bytes27'), D), T[bo, i, by])  # fixed case 36944
ck(f(('bool', 'int144', 'bytes28'), D), T[bo, i, by])  # fixed case 36945
ck(f(('bool', 'int144', 'bytes29'), D), T[bo, i, by])  # fixed case 36946
ck(f(('bool', 'int144', 'bytes3'), D), T[bo, i, by])  # fixed case 36947
ck(f(('bool', 'int144', 'bytes30'), D), T[bo, i, by])  # fixed case 36948
ck(f(('bool', 'int144', 'bytes31'), D), T[bo, i, by])  # fixed case 36949
ck(f(('bool', 'int144', 'bytes32'), D), T[bo, i, by])  # fixed case 36950
ck(f(('bool', 'int144', 'bytes4'), D), T[bo, i, by])  # fixed case 36951
ck(f(('bool', 'int144', 'bytes5'), D), T[bo, i, by])  # fixed case 36952
ck(f(('bool', 'int144', 'bytes6'), D), T[bo, i, by])  # fixed case 36953
ck(f(('bool', 'int144', 'bytes7'), D), T[bo, i, by])  # fixed case 36954
ck(f(('bool', 'int144', 'bytes8'), D), T[bo, i, by])  # fixed case 36955
ck(f(('bool', 'int144', 'bytes9'), D), T[bo, i, by])  # fixed case 36956
ck(f(('bool', 'int144', 'int104'), D), T[bo, i, i])  # fixed case 36957
ck(f(('bool', 'int144', 'int112'), D), T[bo, i, i])  # fixed case 36958
ck(f(('bool', 'int144', 'int120'), D), T[bo, i, i])  # fixed case 36959
ck(f(('bool', 'int144', 'int128'), D), T[bo, i, i])  # fixed case 36960
ck(f(('bool', 'int144', 'int136'), D), T[bo, i, i])  # fixed case 36961
ck(f(('bool', 'int144', 'int144'), D), T[bo, i, i])  # fixed case 36962
ck(f(('bool', 'int144', 'int152'), D), T[bo, i, i])  # fixed case 36963
ck(f(('bool', 'int144', 'int16'), D), T[bo, i, i])  # fixed case 36964
ck(f(('bool', 'int144', 'int160'), D), T[bo, i, i])  # fixed case 36965
ck(f(('bool', 'int144', 'int168'), D), T[bo, i, i])  # fixed case 36966
ck(f(('bool', 'int144', 'int176'), D), T[bo, i, i])  # fixed case 36967
ck(f(('bool', 'int144', 'int184'), D), T[bo, i, i])  # fixed case 36968
ck(f(('bool', 'int144', 'int192'), D), T[bo, i, i])  # fixed case 36969
ck(f(('bool', 'int144', 'int200'), D), T[bo, i, i])  # fixed case 36970
ck(f(('bool', 'int144', 'int208'), D), T[bo, i, i])  # fixed case 36971
ck(f(('bool', 'int144', 'int216'), D), T[bo, i, i])  # fixed case 36972
ck(f(('bool', 'int144', 'int224'), D), T[bo, i, i])  # fixed case 36973
ck(f(('bool', 'int144', 'int232'), D), T[bo, i, i])  # fixed case 36974
ck(f(('bool', 'int144', 'int24'), D), T[bo, i, i])  # fixed case 36975
ck(f(('bool', 'int144', 'int240'), D), T[bo, i, i])  # fixed case 36976
ck(f(('bool', 'int144', 'int248'), D), T[bo, i, i])  # fixed case 36977
ck(f(('bool', 'int144', 'int256'), D), T[bo, i, i])  # fixed case 36978
ck(f(('bool', 'int144', 'int32'), D), T[bo, i, i])  # fixed case 36979
ck(f(('bool', 'int144', 'int40'), D), T[bo, i, i])  # fixed case 36980
ck(f(('bool', 'int144', 'int48'), D), T[bo, i, i])  # fixed case 36981
ck(f(('bool', 'int144', 'int56'), D), T[bo, i, i])  # fixed case 36982
ck(f(('bool', 'int144', 'int64'), D), T[bo, i, i])  # fixed case 36983
ck(f(('bool', 'int144', 'int72'), D), T[bo, i, i])  # fixed case 36984
ck(f(('bool', 'int144', 'int8'), D), T[bo, i, i])  # fixed case 36985
ck(f(('bool', 'int144', 'int80'), D), T[bo, i, i])  # fixed case 36986
ck(f(('bool', 'int144', 'int88'), D), T[bo, i, i])  # fixed case 36987
ck(f(('bool', 'int144', 'int96'), D), T[bo, i, i])  # fixed case 36988
ck(f(('bool', 'int144', 'string'), D), T[bo, i, s])  # fixed case 36989
ck(f(('bool', 'int144', 'string[]'), D), T[bo, i, A])  # fixed case 36990
ck(f(('bool', 'int144', 'uint104'), D), T[bo, i, i])  # fixed case 36991
ck(f(('bool', 'int144', 'uint112'), D), T[bo, i, i])  # fixed case 36992
ck(f(('bool', 'int144', 'uint120'), D), T[bo, i, i])  # fixed case 36993
ck(f(('bool', 'int144', 'uint128'), D), T[bo, i, i])  # fixed case 36994
ck(f(('bool', 'int144', 'uint136'), D), T[bo, i, i])  # fixed case 36995
ck(f(('bool', 'int144', 'uint144'), D), T[bo, i, i])  # fixed case 36996
ck(f(('bool', 'int144', 'uint152'), D), T[bo, i, i])  # fixed case 36997
ck(f(('bool', 'int144', 'uint16'), D), T[bo, i, i])  # fixed case 36998
ck(f(('bool', 'int144', 'uint160'), D), T[bo, i, i])  # fixed case 36999
ck(f(('bool', 'int144', 'uint168'), D), T[bo, i, i])  # fixed case 37000
ck(f(('bool', 'int144', 'uint176'), D), T[bo, i, i])  # fixed case 37001
ck(f(('bool', 'int144', 'uint184'), D), T[bo, i, i])  # fixed case 37002
ck(f(('bool', 'int144', 'uint192'), D), T[bo, i, i])  # fixed case 37003
ck(f(('bool', 'int144', 'uint200'), D), T[bo, i, i])  # fixed case 37004
ck(f(('bool', 'int144', 'uint208'), D), T[bo, i, i])  # fixed case 37005
ck(f(('bool', 'int144', 'uint216'), D), T[bo, i, i])  # fixed case 37006
ck(f(('bool', 'int144', 'uint224'), D), T[bo, i, i])  # fixed case 37007
ck(f(('bool', 'int144', 'uint232'), D), T[bo, i, i])  # fixed case 37008
ck(f(('bool', 'int144', 'uint24'), D), T[bo, i, i])  # fixed case 37009
ck(f(('bool', 'int144', 'uint240'), D), T[bo, i, i])  # fixed case 37010
ck(f(('bool', 'int144', 'uint248'), D), T[bo, i, i])  # fixed case 37011
ck(f(('bool', 'int144', 'uint256'), D), T[bo, i, i])  # fixed case 37012
ck(f(('bool', 'int144', 'uint32'), D), T[bo, i, i])  # fixed case 37013
ck(f(('bool', 'int144', 'uint40'), D), T[bo, i, i])  # fixed case 37014
ck(f(('bool', 'int144', 'uint48'), D), T[bo, i, i])  # fixed case 37015
ck(f(('bool', 'int144', 'uint56'), D), T[bo, i, i])  # fixed case 37016
ck(f(('bool', 'int144', 'uint64'), D), T[bo, i, i])  # fixed case 37017
ck(f(('bool', 'int144', 'uint72'), D), T[bo, i, i])  # fixed case 37018
ck(f(('bool', 'int144', 'uint8'), D), T[bo, i, i])  # fixed case 37019
ck(f(('bool', 'int144', 'uint80'), D), T[bo, i, i])  # fixed case 37020
ck(f(('bool', 'int144', 'uint88'), D), T[bo, i, i])  # fixed case 37021
ck(f(('bool', 'int144', 'uint96'), D), T[bo, i, i])  # fixed case 37022
ck(f(('bool', 'int144', '?'), D), T[bo, i, A])  # fixed case 37023
ck(f(('bool', 'int152', 'address'), D), T[bo, i, HA])  # fixed case 37024
ck(f(('bool', 'int152', 'address[]'), D), T[bo, i, A])  # fixed case 37025
ck(f(('bool', 'int152', 'bool'), D), T[bo, i, bo])  # fixed case 37026
ck(f(('bool', 'int152', 'bool[]'), D), T[bo, i, A])  # fixed case 37027
ck(f(('bool', 'int152', 'bytes'), D), T[bo, i, by])  # fixed case 37028
ck(f(('bool', 'int152', 'bytes1'), D), T[bo, i, by])  # fixed case 37029
ck(f(('bool', 'int152', 'bytes10'), D), T[bo, i, by])  # fixed case 37030
ck(f(('bool', 'int152', 'bytes11'), D), T[bo, i, by])  # fixed case 37031
ck(f(('bool', 'int152', 'bytes12'), D), T[bo, i, by])  # fixed case 37032
ck(f(('bool', 'int152', 'bytes13'), D), T[bo, i, by])  # fixed case 37033
ck(f(('bool', 'int152', 'bytes14'), D), T[bo, i, by])  # fixed case 37034
ck(f(('bool', 'int152', 'bytes15'), D), T[bo, i, by])  # fixed case 37035
ck(f(('bool', 'int152', 'bytes16'), D), T[bo, i, by])  # fixed case 37036
ck(f(('bool', 'int152', 'bytes17'), D), T[bo, i, by])  # fixed case 37037
ck(f(('bool', 'int152', 'bytes18'), D), T[bo, i, by])  # fixed case 37038
ck(f(('bool', 'int152', 'bytes19'), D), T[bo, i, by])  # fixed case 37039
ck(f(('bool', 'int152', 'bytes2'), D), T[bo, i, by])  # fixed case 37040
ck(f(('bool', 'int152', 'bytes20'), D), T[bo, i, by])  # fixed case 37041
ck(f(('bool', 'int152', 'bytes21'), D), T[bo, i, by])  # fixed case 37042
ck(f(('bool', 'int152', 'bytes22'), D), T[bo, i, by])  # fixed case 37043
ck(f(('bool', 'int152', 'bytes23'), D), T[bo, i, by])  # fixed case 37044
ck(f(('bool', 'int152', 'bytes24'), D), T[bo, i, by])  # fixed case 37045
ck(f(('bool', 'int152', 'bytes25'), D), T[bo, i, by])  # fixed case 37046
ck(f(('bool', 'int152', 'bytes26'), D), T[bo, i, by])  # fixed case 37047
ck(f(('bool', 'int152', 'bytes27'), D), T[bo, i, by])  # fixed case 37048
ck(f(('bool', 'int152', 'bytes28'), D), T[bo, i, by])  # fixed case 37049
ck(f(('bool', 'int152', 'bytes29'), D), T[bo, i, by])  # fixed case 37050
ck(f(('bool', 'int152', 'bytes3'), D), T[bo, i, by])  # fixed case 37051
ck(f(('bool', 'int152', 'bytes30'), D), T[bo, i, by])  # fixed case 37052
ck(f(('bool', 'int152', 'bytes31'), D), T[bo, i, by])  # fixed case 37053
ck(f(('bool', 'int152', 'bytes32'), D), T[bo, i, by])  # fixed case 37054
ck(f(('bool', 'int152', 'bytes4'), D), T[bo, i, by])  # fixed case 37055
ck(f(('bool', 'int152', 'bytes5'), D), T[bo, i, by])  # fixed case 37056
ck(f(('bool', 'int152', 'bytes6'), D), T[bo, i, by])  # fixed case 37057
ck(f(('bool', 'int152', 'bytes7'), D), T[bo, i, by])  # fixed case 37058
ck(f(('bool', 'int152', 'bytes8'), D), T[bo, i, by])  # fixed case 37059
ck(f(('bool', 'int152', 'bytes9'), D), T[bo, i, by])  # fixed case 37060
ck(f(('bool', 'int152', 'int104'), D), T[bo, i, i])  # fixed case 37061
ck(f(('bool', 'int152', 'int112'), D), T[bo, i, i])  # fixed case 37062
ck(f(('bool', 'int152', 'int120'), D), T[bo, i, i])  # fixed case 37063
ck(f(('bool', 'int152', 'int128'), D), T[bo, i, i])  # fixed case 37064
ck(f(('bool', 'int152', 'int136'), D), T[bo, i, i])  # fixed case 37065
ck(f(('bool', 'int152', 'int144'), D), T[bo, i, i])  # fixed case 37066
ck(f(('bool', 'int152', 'int152'), D), T[bo, i, i])  # fixed case 37067
ck(f(('bool', 'int152', 'int16'), D), T[bo, i, i])  # fixed case 37068
ck(f(('bool', 'int152', 'int160'), D), T[bo, i, i])  # fixed case 37069
ck(f(('bool', 'int152', 'int168'), D), T[bo, i, i])  # fixed case 37070
ck(f(('bool', 'int152', 'int176'), D), T[bo, i, i])  # fixed case 37071
ck(f(('bool', 'int152', 'int184'), D), T[bo, i, i])  # fixed case 37072
ck(f(('bool', 'int152', 'int192'), D), T[bo, i, i])  # fixed case 37073
ck(f(('bool', 'int152', 'int200'), D), T[bo, i, i])  # fixed case 37074
ck(f(('bool', 'int152', 'int208'), D), T[bo, i, i])  # fixed case 37075
ck(f(('bool', 'int152', 'int216'), D), T[bo, i, i])  # fixed case 37076
ck(f(('bool', 'int152', 'int224'), D), T[bo, i, i])  # fixed case 37077
ck(f(('bool', 'int152', 'int232'), D), T[bo, i, i])  # fixed case 37078
ck(f(('bool', 'int152', 'int24'), D), T[bo, i, i])  # fixed case 37079
ck(f(('bool', 'int152', 'int240'), D), T[bo, i, i])  # fixed case 37080
ck(f(('bool', 'int152', 'int248'), D), T[bo, i, i])  # fixed case 37081
ck(f(('bool', 'int152', 'int256'), D), T[bo, i, i])  # fixed case 37082
ck(f(('bool', 'int152', 'int32'), D), T[bo, i, i])  # fixed case 37083
ck(f(('bool', 'int152', 'int40'), D), T[bo, i, i])  # fixed case 37084
ck(f(('bool', 'int152', 'int48'), D), T[bo, i, i])  # fixed case 37085
ck(f(('bool', 'int152', 'int56'), D), T[bo, i, i])  # fixed case 37086
ck(f(('bool', 'int152', 'int64'), D), T[bo, i, i])  # fixed case 37087
ck(f(('bool', 'int152', 'int72'), D), T[bo, i, i])  # fixed case 37088
ck(f(('bool', 'int152', 'int8'), D), T[bo, i, i])  # fixed case 37089
ck(f(('bool', 'int152', 'int80'), D), T[bo, i, i])  # fixed case 37090
ck(f(('bool', 'int152', 'int88'), D), T[bo, i, i])  # fixed case 37091
ck(f(('bool', 'int152', 'int96'), D), T[bo, i, i])  # fixed case 37092
ck(f(('bool', 'int152', 'string'), D), T[bo, i, s])  # fixed case 37093
ck(f(('bool', 'int152', 'string[]'), D), T[bo, i, A])  # fixed case 37094
ck(f(('bool', 'int152', 'uint104'), D), T[bo, i, i])  # fixed case 37095
ck(f(('bool', 'int152', 'uint112'), D), T[bo, i, i])  # fixed case 37096
ck(f(('bool', 'int152', 'uint120'), D), T[bo, i, i])  # fixed case 37097
ck(f(('bool', 'int152', 'uint128'), D), T[bo, i, i])  # fixed case 37098
ck(f(('bool', 'int152', 'uint136'), D), T[bo, i, i])  # fixed case 37099
ck(f(('bool', 'int152', 'uint144'), D), T[bo, i, i])  # fixed case 37100
ck(f(('bool', 'int152', 'uint152'), D), T[bo, i, i])  # fixed case 37101
ck(f(('bool', 'int152', 'uint16'), D), T[bo, i, i])  # fixed case 37102
ck(f(('bool', 'int152', 'uint160'), D), T[bo, i, i])  # fixed case 37103
ck(f(('bool', 'int152', 'uint168'), D), T[bo, i, i])  # fixed case 37104
ck(f(('bool', 'int152', 'uint176'), D), T[bo, i, i])  # fixed case 37105
ck(f(('bool', 'int152', 'uint184'), D), T[bo, i, i])  # fixed case 37106
ck(f(('bool', 'int152', 'uint192'), D), T[bo, i, i])  # fixed case 37107
ck(f(('bool', 'int152', 'uint200'), D), T[bo, i, i])  # fixed case 37108
ck(f(('bool', 'int152', 'uint208'), D), T[bo, i, i])  # fixed case 37109
ck(f(('bool', 'int152', 'uint216'), D), T[bo, i, i])  # fixed case 37110
ck(f(('bool', 'int152', 'uint224'), D), T[bo, i, i])  # fixed case 37111
ck(f(('bool', 'int152', 'uint232'), D), T[bo, i, i])  # fixed case 37112
ck(f(('bool', 'int152', 'uint24'), D), T[bo, i, i])  # fixed case 37113
ck(f(('bool', 'int152', 'uint240'), D), T[bo, i, i])  # fixed case 37114
ck(f(('bool', 'int152', 'uint248'), D), T[bo, i, i])  # fixed case 37115
ck(f(('bool', 'int152', 'uint256'), D), T[bo, i, i])  # fixed case 37116
ck(f(('bool', 'int152', 'uint32'), D), T[bo, i, i])  # fixed case 37117
ck(f(('bool', 'int152', 'uint40'), D), T[bo, i, i])  # fixed case 37118
ck(f(('bool', 'int152', 'uint48'), D), T[bo, i, i])  # fixed case 37119
ck(f(('bool', 'int152', 'uint56'), D), T[bo, i, i])  # fixed case 37120
ck(f(('bool', 'int152', 'uint64'), D), T[bo, i, i])  # fixed case 37121
ck(f(('bool', 'int152', 'uint72'), D), T[bo, i, i])  # fixed case 37122
ck(f(('bool', 'int152', 'uint8'), D), T[bo, i, i])  # fixed case 37123
ck(f(('bool', 'int152', 'uint80'), D), T[bo, i, i])  # fixed case 37124
ck(f(('bool', 'int152', 'uint88'), D), T[bo, i, i])  # fixed case 37125
ck(f(('bool', 'int152', 'uint96'), D), T[bo, i, i])  # fixed case 37126
ck(f(('bool', 'int152', '?'), D), T[bo, i, A])  # fixed case 37127
ck(f(('bool', 'int16', 'address'), D), T[bo, i, HA])  # fixed case 37128
ck(f(('bool', 'int16', 'address[]'), D), T[bo, i, A])  # fixed case 37129
ck(f(('bool', 'int16', 'bool'), D), T[bo, i, bo])  # fixed case 37130
ck(f(('bool', 'int16', 'bool[]'), D), T[bo, i, A])  # fixed case 37131
ck(f(('bool', 'int16', 'bytes'), D), T[bo, i, by])  # fixed case 37132
ck(f(('bool', 'int16', 'bytes1'), D), T[bo, i, by])  # fixed case 37133
ck(f(('bool', 'int16', 'bytes10'), D), T[bo, i, by])  # fixed case 37134
ck(f(('bool', 'int16', 'bytes11'), D), T[bo, i, by])  # fixed case 37135
ck(f(('bool', 'int16', 'bytes12'), D), T[bo, i, by])  # fixed case 37136
ck(f(('bool', 'int16', 'bytes13'), D), T[bo, i, by])  # fixed case 37137
ck(f(('bool', 'int16', 'bytes14'), D), T[bo, i, by])  # fixed case 37138
ck(f(('bool', 'int16', 'bytes15'), D), T[bo, i, by])  # fixed case 37139
ck(f(('bool', 'int16', 'bytes16'), D), T[bo, i, by])  # fixed case 37140
ck(f(('bool', 'int16', 'bytes17'), D), T[bo, i, by])  # fixed case 37141
ck(f(('bool', 'int16', 'bytes18'), D), T[bo, i, by])  # fixed case 37142
ck(f(('bool', 'int16', 'bytes19'), D), T[bo, i, by])  # fixed case 37143
ck(f(('bool', 'int16', 'bytes2'), D), T[bo, i, by])  # fixed case 37144
ck(f(('bool', 'int16', 'bytes20'), D), T[bo, i, by])  # fixed case 37145
ck(f(('bool', 'int16', 'bytes21'), D), T[bo, i, by])  # fixed case 37146
ck(f(('bool', 'int16', 'bytes22'), D), T[bo, i, by])  # fixed case 37147
ck(f(('bool', 'int16', 'bytes23'), D), T[bo, i, by])  # fixed case 37148
ck(f(('bool', 'int16', 'bytes24'), D), T[bo, i, by])  # fixed case 37149
ck(f(('bool', 'int16', 'bytes25'), D), T[bo, i, by])  # fixed case 37150
ck(f(('bool', 'int16', 'bytes26'), D), T[bo, i, by])  # fixed case 37151
ck(f(('bool', 'int16', 'bytes27'), D), T[bo, i, by])  # fixed case 37152
ck(f(('bool', 'int16', 'bytes28'), D), T[bo, i, by])  # fixed case 37153
ck(f(('bool', 'int16', 'bytes29'), D), T[bo, i, by])  # fixed case 37154
ck(f(('bool', 'int16', 'bytes3'), D), T[bo, i, by])  # fixed case 37155
ck(f(('bool', 'int16', 'bytes30'), D), T[bo, i, by])  # fixed case 37156
ck(f(('bool', 'int16', 'bytes31'), D), T[bo, i, by])  # fixed case 37157
ck(f(('bool', 'int16', 'bytes32'), D), T[bo, i, by])  # fixed case 37158
ck(f(('bool', 'int16', 'bytes4'), D), T[bo, i, by])  # fixed case 37159
ck(f(('bool', 'int16', 'bytes5'), D), T[bo, i, by])  # fixed case 37160
ck(f(('bool', 'int16', 'bytes6'), D), T[bo, i, by])  # fixed case 37161
ck(f(('bool', 'int16', 'bytes7'), D), T[bo, i, by])  # fixed case 37162
ck(f(('bool', 'int16', 'bytes8'), D), T[bo, i, by])  # fixed case 37163
ck(f(('bool', 'int16', 'bytes9'), D), T[bo, i, by])  # fixed case 37164
ck(f(('bool', 'int16', 'int104'), D), T[bo, i, i])  # fixed case 37165
ck(f(('bool', 'int16', 'int112'), D), T[bo, i, i])  # fixed case 37166
ck(f(('bool', 'int16', 'int120'), D), T[bo, i, i])  # fixed case 37167
ck(f(('bool', 'int16', 'int128'), D), T[bo, i, i])  # fixed case 37168
ck(f(('bool', 'int16', 'int136'), D), T[bo, i, i])  # fixed case 37169
ck(f(('bool', 'int16', 'int144'), D), T[bo, i, i])  # fixed case 37170
ck(f(('bool', 'int16', 'int152'), D), T[bo, i, i])  # fixed case 37171
ck(f(('bool', 'int16', 'int16'), D), T[bo, i, i])  # fixed case 37172
ck(f(('bool', 'int16', 'int160'), D), T[bo, i, i])  # fixed case 37173
ck(f(('bool', 'int16', 'int168'), D), T[bo, i, i])  # fixed case 37174
ck(f(('bool', 'int16', 'int176'), D), T[bo, i, i])  # fixed case 37175
ck(f(('bool', 'int16', 'int184'), D), T[bo, i, i])  # fixed case 37176
ck(f(('bool', 'int16', 'int192'), D), T[bo, i, i])  # fixed case 37177
ck(f(('bool', 'int16', 'int200'), D), T[bo, i, i])  # fixed case 37178
ck(f(('bool', 'int16', 'int208'), D), T[bo, i, i])  # fixed case 37179
ck(f(('bool', 'int16', 'int216'), D), T[bo, i, i])  # fixed case 37180
ck(f(('bool', 'int16', 'int224'), D), T[bo, i, i])  # fixed case 37181
ck(f(('bool', 'int16', 'int232'), D), T[bo, i, i])  # fixed case 37182
ck(f(('bool', 'int16', 'int24'), D), T[bo, i, i])  # fixed case 37183
ck(f(('bool', 'int16', 'int240'), D), T[bo, i, i])  # fixed case 37184
ck(f(('bool', 'int16', 'int248'), D), T[bo, i, i])  # fixed case 37185
ck(f(('bool', 'int16', 'int256'), D), T[bo, i, i])  # fixed case 37186
ck(f(('bool', 'int16', 'int32'), D), T[bo, i, i])  # fixed case 37187
ck(f(('bool', 'int16', 'int40'), D), T[bo, i, i])  # fixed case 37188
ck(f(('bool', 'int16', 'int48'), D), T[bo, i, i])  # fixed case 37189
ck(f(('bool', 'int16', 'int56'), D), T[bo, i, i])  # fixed case 37190
ck(f(('bool', 'int16', 'int64'), D), T[bo, i, i])  # fixed case 37191
ck(f(('bool', 'int16', 'int72'), D), T[bo, i, i])  # fixed case 37192
ck(f(('bool', 'int16', 'int8'), D), T[bo, i, i])  # fixed case 37193
ck(f(('bool', 'int16', 'int80'), D), T[bo, i, i])  # fixed case 37194
ck(f(('bool', 'int16', 'int88'), D), T[bo, i, i])  # fixed case 37195
ck(f(('bool', 'int16', 'int96'), D), T[bo, i, i])  # fixed case 37196
ck(f(('bool', 'int16', 'string'), D), T[bo, i, s])  # fixed case 37197
ck(f(('bool', 'int16', 'string[]'), D), T[bo, i, A])  # fixed case 37198
ck(f(('bool', 'int16', 'uint104'), D), T[bo, i, i])  # fixed case 37199
ck(f(('bool', 'int16', 'uint112'), D), T[bo, i, i])  # fixed case 37200
ck(f(('bool', 'int16', 'uint120'), D), T[bo, i, i])  # fixed case 37201
ck(f(('bool', 'int16', 'uint128'), D), T[bo, i, i])  # fixed case 37202
ck(f(('bool', 'int16', 'uint136'), D), T[bo, i, i])  # fixed case 37203
ck(f(('bool', 'int16', 'uint144'), D), T[bo, i, i])  # fixed case 37204
ck(f(('bool', 'int16', 'uint152'), D), T[bo, i, i])  # fixed case 37205
ck(f(('bool', 'int16', 'uint16'), D), T[bo, i, i])  # fixed case 37206
ck(f(('bool', 'int16', 'uint160'), D), T[bo, i, i])  # fixed case 37207
ck(f(('bool', 'int16', 'uint168'), D), T[bo, i, i])  # fixed case 37208
ck(f(('bool', 'int16', 'uint176'), D), T[bo, i, i])  # fixed case 37209
ck(f(('bool', 'int16', 'uint184'), D), T[bo, i, i])  # fixed case 37210
ck(f(('bool', 'int16', 'uint192'), D), T[bo, i, i])  # fixed case 37211
ck(f(('bool', 'int16', 'uint200'), D), T[bo, i, i])  # fixed case 37212
ck(f(('bool', 'int16', 'uint208'), D), T[bo, i, i])  # fixed case 37213
ck(f(('bool', 'int16', 'uint216'), D), T[bo, i, i])  # fixed case 37214
ck(f(('bool', 'int16', 'uint224'), D), T[bo, i, i])  # fixed case 37215
ck(f(('bool', 'int16', 'uint232'), D), T[bo, i, i])  # fixed case 37216
ck(f(('bool', 'int16', 'uint24'), D), T[bo, i, i])  # fixed case 37217
ck(f(('bool', 'int16', 'uint240'), D), T[bo, i, i])  # fixed case 37218
ck(f(('bool', 'int16', 'uint248'), D), T[bo, i, i])  # fixed case 37219
ck(f(('bool', 'int16', 'uint256'), D), T[bo, i, i])  # fixed case 37220
ck(f(('bool', 'int16', 'uint32'), D), T[bo, i, i])  # fixed case 37221
ck(f(('bool', 'int16', 'uint40'), D), T[bo, i, i])  # fixed case 37222
ck(f(('bool', 'int16', 'uint48'), D), T[bo, i, i])  # fixed case 37223
ck(f(('bool', 'int16', 'uint56'), D), T[bo, i, i])  # fixed case 37224
ck(f(('bool', 'int16', 'uint64'), D), T[bo, i, i])  # fixed case 37225
ck(f(('bool', 'int16', 'uint72'), D), T[bo, i, i])  # fixed case 37226
ck(f(('bool', 'int16', 'uint8'), D), T[bo, i, i])  # fixed case 37227
ck(f(('bool', 'int16', 'uint80'), D), T[bo, i, i])  # fixed case 37228
ck(f(('bool', 'int16', 'uint88'), D), T[bo, i, i])  # fixed case 37229
ck(f(('bool', 'int16', 'uint96'), D), T[bo, i, i])  # fixed case 37230
ck(f(('bool', 'int16', '?'), D), T[bo, i, A])  # fixed case 37231
ck(f(('bool', 'int160', 'address'), D), T[bo, i, HA])  # fixed case 37232
ck(f(('bool', 'int160', 'address[]'), D), T[bo, i, A])  # fixed case 37233
ck(f(('bool', 'int160', 'bool'), D), T[bo, i, bo])  # fixed case 37234
ck(f(('bool', 'int160', 'bool[]'), D), T[bo, i, A])  # fixed case 37235
ck(f(('bool', 'int160', 'bytes'), D), T[bo, i, by])  # fixed case 37236
ck(f(('bool', 'int160', 'bytes1'), D), T[bo, i, by])  # fixed case 37237
ck(f(('bool', 'int160', 'bytes10'), D), T[bo, i, by])  # fixed case 37238
ck(f(('bool', 'int160', 'bytes11'), D), T[bo, i, by])  # fixed case 37239
ck(f(('bool', 'int160', 'bytes12'), D), T[bo, i, by])  # fixed case 37240
ck(f(('bool', 'int160', 'bytes13'), D), T[bo, i, by])  # fixed case 37241
ck(f(('bool', 'int160', 'bytes14'), D), T[bo, i, by])  # fixed case 37242
ck(f(('bool', 'int160', 'bytes15'), D), T[bo, i, by])  # fixed case 37243
ck(f(('bool', 'int160', 'bytes16'), D), T[bo, i, by])  # fixed case 37244
ck(f(('bool', 'int160', 'bytes17'), D), T[bo, i, by])  # fixed case 37245
ck(f(('bool', 'int160', 'bytes18'), D), T[bo, i, by])  # fixed case 37246
ck(f(('bool', 'int160', 'bytes19'), D), T[bo, i, by])  # fixed case 37247
ck(f(('bool', 'int160', 'bytes2'), D), T[bo, i, by])  # fixed case 37248
ck(f(('bool', 'int160', 'bytes20'), D), T[bo, i, by])  # fixed case 37249
ck(f(('bool', 'int160', 'bytes21'), D), T[bo, i, by])  # fixed case 37250
ck(f(('bool', 'int160', 'bytes22'), D), T[bo, i, by])  # fixed case 37251
ck(f(('bool', 'int160', 'bytes23'), D), T[bo, i, by])  # fixed case 37252
ck(f(('bool', 'int160', 'bytes24'), D), T[bo, i, by])  # fixed case 37253
ck(f(('bool', 'int160', 'bytes25'), D), T[bo, i, by])  # fixed case 37254
ck(f(('bool', 'int160', 'bytes26'), D), T[bo, i, by])  # fixed case 37255
ck(f(('bool', 'int160', 'bytes27'), D), T[bo, i, by])  # fixed case 37256
ck(f(('bool', 'int160', 'bytes28'), D), T[bo, i, by])  # fixed case 37257
ck(f(('bool', 'int160', 'bytes29'), D), T[bo, i, by])  # fixed case 37258
ck(f(('bool', 'int160', 'bytes3'), D), T[bo, i, by])  # fixed case 37259
ck(f(('bool', 'int160', 'bytes30'), D), T[bo, i, by])  # fixed case 37260
ck(f(('bool', 'int160', 'bytes31'), D), T[bo, i, by])  # fixed case 37261
ck(f(('bool', 'int160', 'bytes32'), D), T[bo, i, by])  # fixed case 37262
ck(f(('bool', 'int160', 'bytes4'), D), T[bo, i, by])  # fixed case 37263
ck(f(('bool', 'int160', 'bytes5'), D), T[bo, i, by])  # fixed case 37264
ck(f(('bool', 'int160', 'bytes6'), D), T[bo, i, by])  # fixed case 37265
ck(f(('bool', 'int160', 'bytes7'), D), T[bo, i, by])  # fixed case 37266
ck(f(('bool', 'int160', 'bytes8'), D), T[bo, i, by])  # fixed case 37267
ck(f(('bool', 'int160', 'bytes9'), D), T[bo, i, by])  # fixed case 37268
ck(f(('bool', 'int160', 'int104'), D), T[bo, i, i])  # fixed case 37269
ck(f(('bool', 'int160', 'int112'), D), T[bo, i, i])  # fixed case 37270
ck(f(('bool', 'int160', 'int120'), D), T[bo, i, i])  # fixed case 37271
ck(f(('bool', 'int160', 'int128'), D), T[bo, i, i])  # fixed case 37272
ck(f(('bool', 'int160', 'int136'), D), T[bo, i, i])  # fixed case 37273
ck(f(('bool', 'int160', 'int144'), D), T[bo, i, i])  # fixed case 37274
ck(f(('bool', 'int160', 'int152'), D), T[bo, i, i])  # fixed case 37275
ck(f(('bool', 'int160', 'int16'), D), T[bo, i, i])  # fixed case 37276
ck(f(('bool', 'int160', 'int160'), D), T[bo, i, i])  # fixed case 37277
ck(f(('bool', 'int160', 'int168'), D), T[bo, i, i])  # fixed case 37278
ck(f(('bool', 'int160', 'int176'), D), T[bo, i, i])  # fixed case 37279
ck(f(('bool', 'int160', 'int184'), D), T[bo, i, i])  # fixed case 37280
ck(f(('bool', 'int160', 'int192'), D), T[bo, i, i])  # fixed case 37281
ck(f(('bool', 'int160', 'int200'), D), T[bo, i, i])  # fixed case 37282
ck(f(('bool', 'int160', 'int208'), D), T[bo, i, i])  # fixed case 37283
ck(f(('bool', 'int160', 'int216'), D), T[bo, i, i])  # fixed case 37284
ck(f(('bool', 'int160', 'int224'), D), T[bo, i, i])  # fixed case 37285
ck(f(('bool', 'int160', 'int232'), D), T[bo, i, i])  # fixed case 37286
ck(f(('bool', 'int160', 'int24'), D), T[bo, i, i])  # fixed case 37287
ck(f(('bool', 'int160', 'int240'), D), T[bo, i, i])  # fixed case 37288
ck(f(('bool', 'int160', 'int248'), D), T[bo, i, i])  # fixed case 37289
ck(f(('bool', 'int160', 'int256'), D), T[bo, i, i])  # fixed case 37290
ck(f(('bool', 'int160', 'int32'), D), T[bo, i, i])  # fixed case 37291
ck(f(('bool', 'int160', 'int40'), D), T[bo, i, i])  # fixed case 37292
ck(f(('bool', 'int160', 'int48'), D), T[bo, i, i])  # fixed case 37293
ck(f(('bool', 'int160', 'int56'), D), T[bo, i, i])  # fixed case 37294
ck(f(('bool', 'int160', 'int64'), D), T[bo, i, i])  # fixed case 37295
ck(f(('bool', 'int160', 'int72'), D), T[bo, i, i])  # fixed case 37296
ck(f(('bool', 'int160', 'int8'), D), T[bo, i, i])  # fixed case 37297
ck(f(('bool', 'int160', 'int80'), D), T[bo, i, i])  # fixed case 37298
ck(f(('bool', 'int160', 'int88'), D), T[bo, i, i])  # fixed case 37299
ck(f(('bool', 'int160', 'int96'), D), T[bo, i, i])  # fixed case 37300
ck(f(('bool', 'int160', 'string'), D), T[bo, i, s])  # fixed case 37301
ck(f(('bool', 'int160', 'string[]'), D), T[bo, i, A])  # fixed case 37302
ck(f(('bool', 'int160', 'uint104'), D), T[bo, i, i])  # fixed case 37303
ck(f(('bool', 'int160', 'uint112'), D), T[bo, i, i])  # fixed case 37304
ck(f(('bool', 'int160', 'uint120'), D), T[bo, i, i])  # fixed case 37305
ck(f(('bool', 'int160', 'uint128'), D), T[bo, i, i])  # fixed case 37306
ck(f(('bool', 'int160', 'uint136'), D), T[bo, i, i])  # fixed case 37307
ck(f(('bool', 'int160', 'uint144'), D), T[bo, i, i])  # fixed case 37308
ck(f(('bool', 'int160', 'uint152'), D), T[bo, i, i])  # fixed case 37309
ck(f(('bool', 'int160', 'uint16'), D), T[bo, i, i])  # fixed case 37310
ck(f(('bool', 'int160', 'uint160'), D), T[bo, i, i])  # fixed case 37311
ck(f(('bool', 'int160', 'uint168'), D), T[bo, i, i])  # fixed case 37312
ck(f(('bool', 'int160', 'uint176'), D), T[bo, i, i])  # fixed case 37313
ck(f(('bool', 'int160', 'uint184'), D), T[bo, i, i])  # fixed case 37314
ck(f(('bool', 'int160', 'uint192'), D), T[bo, i, i])  # fixed case 37315
ck(f(('bool', 'int160', 'uint200'), D), T[bo, i, i])  # fixed case 37316
ck(f(('bool', 'int160', 'uint208'), D), T[bo, i, i])  # fixed case 37317
ck(f(('bool', 'int160', 'uint216'), D), T[bo, i, i])  # fixed case 37318
ck(f(('bool', 'int160', 'uint224'), D), T[bo, i, i])  # fixed case 37319
ck(f(('bool', 'int160', 'uint232'), D), T[bo, i, i])  # fixed case 37320
ck(f(('bool', 'int160', 'uint24'), D), T[bo, i, i])  # fixed case 37321
ck(f(('bool', 'int160', 'uint240'), D), T[bo, i, i])  # fixed case 37322
ck(f(('bool', 'int160', 'uint248'), D), T[bo, i, i])  # fixed case 37323
ck(f(('bool', 'int160', 'uint256'), D), T[bo, i, i])  # fixed case 37324
ck(f(('bool', 'int160', 'uint32'), D), T[bo, i, i])  # fixed case 37325
ck(f(('bool', 'int160', 'uint40'), D), T[bo, i, i])  # fixed case 37326
ck(f(('bool', 'int160', 'uint48'), D), T[bo, i, i])  # fixed case 37327
ck(f(('bool', 'int160', 'uint56'), D), T[bo, i, i])  # fixed case 37328
ck(f(('bool', 'int160', 'uint64'), D), T[bo, i, i])  # fixed case 37329
ck(f(('bool', 'int160', 'uint72'), D), T[bo, i, i])  # fixed case 37330
ck(f(('bool', 'int160', 'uint8'), D), T[bo, i, i])  # fixed case 37331
ck(f(('bool', 'int160', 'uint80'), D), T[bo, i, i])  # fixed case 37332
ck(f(('bool', 'int160', 'uint88'), D), T[bo, i, i])  # fixed case 37333
ck(f(('bool', 'int160', 'uint96'), D), T[bo, i, i])  # fixed case 37334
ck(f(('bool', 'int160', '?'), D), T[bo, i, A])  # fixed case 37335
ck(f(('bool', 'int168', 'address'), D), T[bo, i, HA])  # fixed case 37336
ck(f(('bool', 'int168', 'address[]'), D), T[bo, i, A])  # fixed case 37337
ck(f(('bool', 'int168', 'bool'), D), T[bo, i, bo])  # fixed case 37338
ck(f(('bool', 'int168', 'bool[]'), D), T[bo, i, A])  # fixed case 37339
ck(f(('bool', 'int168', 'bytes'), D), T[bo, i, by])  # fixed case 37340
ck(f(('bool', 'int168', 'bytes1'), D), T[bo, i, by])  # fixed case 37341
ck(f(('bool', 'int168', 'bytes10'), D), T[bo, i, by])  # fixed case 37342
ck(f(('bool', 'int168', 'bytes11'), D), T[bo, i, by])  # fixed case 37343
ck(f(('bool', 'int168', 'bytes12'), D), T[bo, i, by])  # fixed case 37344
ck(f(('bool', 'int168', 'bytes13'), D), T[bo, i, by])  # fixed case 37345
ck(f(('bool', 'int168', 'bytes14'), D), T[bo, i, by])  # fixed case 37346
ck(f(('bool', 'int168', 'bytes15'), D), T[bo, i, by])  # fixed case 37347
ck(f(('bool', 'int168', 'bytes16'), D), T[bo, i, by])  # fixed case 37348
ck(f(('bool', 'int168', 'bytes17'), D), T[bo, i, by])  # fixed case 37349
ck(f(('bool', 'int168', 'bytes18'), D), T[bo, i, by])  # fixed case 37350
ck(f(('bool', 'int168', 'bytes19'), D), T[bo, i, by])  # fixed case 37351
ck(f(('bool', 'int168', 'bytes2'), D), T[bo, i, by])  # fixed case 37352
ck(f(('bool', 'int168', 'bytes20'), D), T[bo, i, by])  # fixed case 37353
ck(f(('bool', 'int168', 'bytes21'), D), T[bo, i, by])  # fixed case 37354
ck(f(('bool', 'int168', 'bytes22'), D), T[bo, i, by])  # fixed case 37355
ck(f(('bool', 'int168', 'bytes23'), D), T[bo, i, by])  # fixed case 37356
ck(f(('bool', 'int168', 'bytes24'), D), T[bo, i, by])  # fixed case 37357
ck(f(('bool', 'int168', 'bytes25'), D), T[bo, i, by])  # fixed case 37358
ck(f(('bool', 'int168', 'bytes26'), D), T[bo, i, by])  # fixed case 37359
ck(f(('bool', 'int168', 'bytes27'), D), T[bo, i, by])  # fixed case 37360
ck(f(('bool', 'int168', 'bytes28'), D), T[bo, i, by])  # fixed case 37361
ck(f(('bool', 'int168', 'bytes29'), D), T[bo, i, by])  # fixed case 37362
ck(f(('bool', 'int168', 'bytes3'), D), T[bo, i, by])  # fixed case 37363
ck(f(('bool', 'int168', 'bytes30'), D), T[bo, i, by])  # fixed case 37364
ck(f(('bool', 'int168', 'bytes31'), D), T[bo, i, by])  # fixed case 37365
ck(f(('bool', 'int168', 'bytes32'), D), T[bo, i, by])  # fixed case 37366
ck(f(('bool', 'int168', 'bytes4'), D), T[bo, i, by])  # fixed case 37367
ck(f(('bool', 'int168', 'bytes5'), D), T[bo, i, by])  # fixed case 37368
ck(f(('bool', 'int168', 'bytes6'), D), T[bo, i, by])  # fixed case 37369
ck(f(('bool', 'int168', 'bytes7'), D), T[bo, i, by])  # fixed case 37370
ck(f(('bool', 'int168', 'bytes8'), D), T[bo, i, by])  # fixed case 37371
ck(f(('bool', 'int168', 'bytes9'), D), T[bo, i, by])  # fixed case 37372
ck(f(('bool', 'int168', 'int104'), D), T[bo, i, i])  # fixed case 37373
ck(f(('bool', 'int168', 'int112'), D), T[bo, i, i])  # fixed case 37374
ck(f(('bool', 'int168', 'int120'), D), T[bo, i, i])  # fixed case 37375
ck(f(('bool', 'int168', 'int128'), D), T[bo, i, i])  # fixed case 37376
ck(f(('bool', 'int168', 'int136'), D), T[bo, i, i])  # fixed case 37377
ck(f(('bool', 'int168', 'int144'), D), T[bo, i, i])  # fixed case 37378
ck(f(('bool', 'int168', 'int152'), D), T[bo, i, i])  # fixed case 37379
ck(f(('bool', 'int168', 'int16'), D), T[bo, i, i])  # fixed case 37380
ck(f(('bool', 'int168', 'int160'), D), T[bo, i, i])  # fixed case 37381
ck(f(('bool', 'int168', 'int168'), D), T[bo, i, i])  # fixed case 37382
ck(f(('bool', 'int168', 'int176'), D), T[bo, i, i])  # fixed case 37383
ck(f(('bool', 'int168', 'int184'), D), T[bo, i, i])  # fixed case 37384
ck(f(('bool', 'int168', 'int192'), D), T[bo, i, i])  # fixed case 37385
ck(f(('bool', 'int168', 'int200'), D), T[bo, i, i])  # fixed case 37386
ck(f(('bool', 'int168', 'int208'), D), T[bo, i, i])  # fixed case 37387
ck(f(('bool', 'int168', 'int216'), D), T[bo, i, i])  # fixed case 37388
ck(f(('bool', 'int168', 'int224'), D), T[bo, i, i])  # fixed case 37389
ck(f(('bool', 'int168', 'int232'), D), T[bo, i, i])  # fixed case 37390
ck(f(('bool', 'int168', 'int24'), D), T[bo, i, i])  # fixed case 37391
ck(f(('bool', 'int168', 'int240'), D), T[bo, i, i])  # fixed case 37392
ck(f(('bool', 'int168', 'int248'), D), T[bo, i, i])  # fixed case 37393
ck(f(('bool', 'int168', 'int256'), D), T[bo, i, i])  # fixed case 37394
ck(f(('bool', 'int168', 'int32'), D), T[bo, i, i])  # fixed case 37395
ck(f(('bool', 'int168', 'int40'), D), T[bo, i, i])  # fixed case 37396
ck(f(('bool', 'int168', 'int48'), D), T[bo, i, i])  # fixed case 37397
ck(f(('bool', 'int168', 'int56'), D), T[bo, i, i])  # fixed case 37398
ck(f(('bool', 'int168', 'int64'), D), T[bo, i, i])  # fixed case 37399
ck(f(('bool', 'int168', 'int72'), D), T[bo, i, i])  # fixed case 37400
ck(f(('bool', 'int168', 'int8'), D), T[bo, i, i])  # fixed case 37401
ck(f(('bool', 'int168', 'int80'), D), T[bo, i, i])  # fixed case 37402
ck(f(('bool', 'int168', 'int88'), D), T[bo, i, i])  # fixed case 37403
ck(f(('bool', 'int168', 'int96'), D), T[bo, i, i])  # fixed case 37404
ck(f(('bool', 'int168', 'string'), D), T[bo, i, s])  # fixed case 37405
ck(f(('bool', 'int168', 'string[]'), D), T[bo, i, A])  # fixed case 37406
ck(f(('bool', 'int168', 'uint104'), D), T[bo, i, i])  # fixed case 37407
ck(f(('bool', 'int168', 'uint112'), D), T[bo, i, i])  # fixed case 37408
ck(f(('bool', 'int168', 'uint120'), D), T[bo, i, i])  # fixed case 37409
ck(f(('bool', 'int168', 'uint128'), D), T[bo, i, i])  # fixed case 37410
ck(f(('bool', 'int168', 'uint136'), D), T[bo, i, i])  # fixed case 37411
ck(f(('bool', 'int168', 'uint144'), D), T[bo, i, i])  # fixed case 37412
ck(f(('bool', 'int168', 'uint152'), D), T[bo, i, i])  # fixed case 37413
ck(f(('bool', 'int168', 'uint16'), D), T[bo, i, i])  # fixed case 37414
ck(f(('bool', 'int168', 'uint160'), D), T[bo, i, i])  # fixed case 37415
ck(f(('bool', 'int168', 'uint168'), D), T[bo, i, i])  # fixed case 37416
ck(f(('bool', 'int168', 'uint176'), D), T[bo, i, i])  # fixed case 37417
ck(f(('bool', 'int168', 'uint184'), D), T[bo, i, i])  # fixed case 37418
ck(f(('bool', 'int168', 'uint192'), D), T[bo, i, i])  # fixed case 37419
ck(f(('bool', 'int168', 'uint200'), D), T[bo, i, i])  # fixed case 37420
ck(f(('bool', 'int168', 'uint208'), D), T[bo, i, i])  # fixed case 37421
ck(f(('bool', 'int168', 'uint216'), D), T[bo, i, i])  # fixed case 37422
ck(f(('bool', 'int168', 'uint224'), D), T[bo, i, i])  # fixed case 37423
ck(f(('bool', 'int168', 'uint232'), D), T[bo, i, i])  # fixed case 37424
ck(f(('bool', 'int168', 'uint24'), D), T[bo, i, i])  # fixed case 37425
ck(f(('bool', 'int168', 'uint240'), D), T[bo, i, i])  # fixed case 37426
ck(f(('bool', 'int168', 'uint248'), D), T[bo, i, i])  # fixed case 37427
ck(f(('bool', 'int168', 'uint256'), D), T[bo, i, i])  # fixed case 37428
ck(f(('bool', 'int168', 'uint32'), D), T[bo, i, i])  # fixed case 37429
ck(f(('bool', 'int168', 'uint40'), D), T[bo, i, i])  # fixed case 37430
ck(f(('bool', 'int168', 'uint48'), D), T[bo, i, i])  # fixed case 37431
ck(f(('bool', 'int168', 'uint56'), D), T[bo, i, i])  # fixed case 37432
ck(f(('bool', 'int168', 'uint64'), D), T[bo, i, i])  # fixed case 37433
ck(f(('bool', 'int168', 'uint72'), D), T[bo, i, i])  # fixed case 37434
ck(f(('bool', 'int168', 'uint8'), D), T[bo, i, i])  # fixed case 37435
ck(f(('bool', 'int168', 'uint80'), D), T[bo, i, i])  # fixed case 37436
ck(f(('bool', 'int168', 'uint88'), D), T[bo, i, i])  # fixed case 37437
ck(f(('bool', 'int168', 'uint96'), D), T[bo, i, i])  # fixed case 37438
ck(f(('bool', 'int168', '?'), D), T[bo, i, A])  # fixed case 37439
ck(f(('bool', 'int176', 'address'), D), T[bo, i, HA])  # fixed case 37440
ck(f(('bool', 'int176', 'address[]'), D), T[bo, i, A])  # fixed case 37441
ck(f(('bool', 'int176', 'bool'), D), T[bo, i, bo])  # fixed case 37442
ck(f(('bool', 'int176', 'bool[]'), D), T[bo, i, A])  # fixed case 37443
ck(f(('bool', 'int176', 'bytes'), D), T[bo, i, by])  # fixed case 37444
ck(f(('bool', 'int176', 'bytes1'), D), T[bo, i, by])  # fixed case 37445
ck(f(('bool', 'int176', 'bytes10'), D), T[bo, i, by])  # fixed case 37446
ck(f(('bool', 'int176', 'bytes11'), D), T[bo, i, by])  # fixed case 37447
ck(f(('bool', 'int176', 'bytes12'), D), T[bo, i, by])  # fixed case 37448
ck(f(('bool', 'int176', 'bytes13'), D), T[bo, i, by])  # fixed case 37449
ck(f(('bool', 'int176', 'bytes14'), D), T[bo, i, by])  # fixed case 37450
ck(f(('bool', 'int176', 'bytes15'), D), T[bo, i, by])  # fixed case 37451
ck(f(('bool', 'int176', 'bytes16'), D), T[bo, i, by])  # fixed case 37452
ck(f(('bool', 'int176', 'bytes17'), D), T[bo, i, by])  # fixed case 37453
ck(f(('bool', 'int176', 'bytes18'), D), T[bo, i, by])  # fixed case 37454
ck(f(('bool', 'int176', 'bytes19'), D), T[bo, i, by])  # fixed case 37455
ck(f(('bool', 'int176', 'bytes2'), D), T[bo, i, by])  # fixed case 37456
ck(f(('bool', 'int176', 'bytes20'), D), T[bo, i, by])  # fixed case 37457
ck(f(('bool', 'int176', 'bytes21'), D), T[bo, i, by])  # fixed case 37458
ck(f(('bool', 'int176', 'bytes22'), D), T[bo, i, by])  # fixed case 37459
ck(f(('bool', 'int176', 'bytes23'), D), T[bo, i, by])  # fixed case 37460
ck(f(('bool', 'int176', 'bytes24'), D), T[bo, i, by])  # fixed case 37461
ck(f(('bool', 'int176', 'bytes25'), D), T[bo, i, by])  # fixed case 37462
ck(f(('bool', 'int176', 'bytes26'), D), T[bo, i, by])  # fixed case 37463
ck(f(('bool', 'int176', 'bytes27'), D), T[bo, i, by])  # fixed case 37464
ck(f(('bool', 'int176', 'bytes28'), D), T[bo, i, by])  # fixed case 37465
ck(f(('bool', 'int176', 'bytes29'), D), T[bo, i, by])  # fixed case 37466
ck(f(('bool', 'int176', 'bytes3'), D), T[bo, i, by])  # fixed case 37467
ck(f(('bool', 'int176', 'bytes30'), D), T[bo, i, by])  # fixed case 37468
ck(f(('bool', 'int176', 'bytes31'), D), T[bo, i, by])  # fixed case 37469
ck(f(('bool', 'int176', 'bytes32'), D), T[bo, i, by])  # fixed case 37470
ck(f(('bool', 'int176', 'bytes4'), D), T[bo, i, by])  # fixed case 37471
ck(f(('bool', 'int176', 'bytes5'), D), T[bo, i, by])  # fixed case 37472
ck(f(('bool', 'int176', 'bytes6'), D), T[bo, i, by])  # fixed case 37473
ck(f(('bool', 'int176', 'bytes7'), D), T[bo, i, by])  # fixed case 37474
ck(f(('bool', 'int176', 'bytes8'), D), T[bo, i, by])  # fixed case 37475
ck(f(('bool', 'int176', 'bytes9'), D), T[bo, i, by])  # fixed case 37476
ck(f(('bool', 'int176', 'int104'), D), T[bo, i, i])  # fixed case 37477
ck(f(('bool', 'int176', 'int112'), D), T[bo, i, i])  # fixed case 37478
ck(f(('bool', 'int176', 'int120'), D), T[bo, i, i])  # fixed case 37479
ck(f(('bool', 'int176', 'int128'), D), T[bo, i, i])  # fixed case 37480
ck(f(('bool', 'int176', 'int136'), D), T[bo, i, i])  # fixed case 37481
ck(f(('bool', 'int176', 'int144'), D), T[bo, i, i])  # fixed case 37482
ck(f(('bool', 'int176', 'int152'), D), T[bo, i, i])  # fixed case 37483
ck(f(('bool', 'int176', 'int16'), D), T[bo, i, i])  # fixed case 37484
ck(f(('bool', 'int176', 'int160'), D), T[bo, i, i])  # fixed case 37485
ck(f(('bool', 'int176', 'int168'), D), T[bo, i, i])  # fixed case 37486
ck(f(('bool', 'int176', 'int176'), D), T[bo, i, i])  # fixed case 37487
ck(f(('bool', 'int176', 'int184'), D), T[bo, i, i])  # fixed case 37488
ck(f(('bool', 'int176', 'int192'), D), T[bo, i, i])  # fixed case 37489
ck(f(('bool', 'int176', 'int200'), D), T[bo, i, i])  # fixed case 37490
ck(f(('bool', 'int176', 'int208'), D), T[bo, i, i])  # fixed case 37491
ck(f(('bool', 'int176', 'int216'), D), T[bo, i, i])  # fixed case 37492
ck(f(('bool', 'int176', 'int224'), D), T[bo, i, i])  # fixed case 37493
ck(f(('bool', 'int176', 'int232'), D), T[bo, i, i])  # fixed case 37494
ck(f(('bool', 'int176', 'int24'), D), T[bo, i, i])  # fixed case 37495
ck(f(('bool', 'int176', 'int240'), D), T[bo, i, i])  # fixed case 37496
ck(f(('bool', 'int176', 'int248'), D), T[bo, i, i])  # fixed case 37497
ck(f(('bool', 'int176', 'int256'), D), T[bo, i, i])  # fixed case 37498
ck(f(('bool', 'int176', 'int32'), D), T[bo, i, i])  # fixed case 37499
ck(f(('bool', 'int176', 'int40'), D), T[bo, i, i])  # fixed case 37500
ck(f(('bool', 'int176', 'int48'), D), T[bo, i, i])  # fixed case 37501
ck(f(('bool', 'int176', 'int56'), D), T[bo, i, i])  # fixed case 37502
ck(f(('bool', 'int176', 'int64'), D), T[bo, i, i])  # fixed case 37503
ck(f(('bool', 'int176', 'int72'), D), T[bo, i, i])  # fixed case 37504
ck(f(('bool', 'int176', 'int8'), D), T[bo, i, i])  # fixed case 37505
ck(f(('bool', 'int176', 'int80'), D), T[bo, i, i])  # fixed case 37506
ck(f(('bool', 'int176', 'int88'), D), T[bo, i, i])  # fixed case 37507
ck(f(('bool', 'int176', 'int96'), D), T[bo, i, i])  # fixed case 37508
ck(f(('bool', 'int176', 'string'), D), T[bo, i, s])  # fixed case 37509
ck(f(('bool', 'int176', 'string[]'), D), T[bo, i, A])  # fixed case 37510
ck(f(('bool', 'int176', 'uint104'), D), T[bo, i, i])  # fixed case 37511
ck(f(('bool', 'int176', 'uint112'), D), T[bo, i, i])  # fixed case 37512
ck(f(('bool', 'int176', 'uint120'), D), T[bo, i, i])  # fixed case 37513
ck(f(('bool', 'int176', 'uint128'), D), T[bo, i, i])  # fixed case 37514
ck(f(('bool', 'int176', 'uint136'), D), T[bo, i, i])  # fixed case 37515
ck(f(('bool', 'int176', 'uint144'), D), T[bo, i, i])  # fixed case 37516
ck(f(('bool', 'int176', 'uint152'), D), T[bo, i, i])  # fixed case 37517
ck(f(('bool', 'int176', 'uint16'), D), T[bo, i, i])  # fixed case 37518
ck(f(('bool', 'int176', 'uint160'), D), T[bo, i, i])  # fixed case 37519
ck(f(('bool', 'int176', 'uint168'), D), T[bo, i, i])  # fixed case 37520
ck(f(('bool', 'int176', 'uint176'), D), T[bo, i, i])  # fixed case 37521
ck(f(('bool', 'int176', 'uint184'), D), T[bo, i, i])  # fixed case 37522
ck(f(('bool', 'int176', 'uint192'), D), T[bo, i, i])  # fixed case 37523
ck(f(('bool', 'int176', 'uint200'), D), T[bo, i, i])  # fixed case 37524
ck(f(('bool', 'int176', 'uint208'), D), T[bo, i, i])  # fixed case 37525
ck(f(('bool', 'int176', 'uint216'), D), T[bo, i, i])  # fixed case 37526
ck(f(('bool', 'int176', 'uint224'), D), T[bo, i, i])  # fixed case 37527
ck(f(('bool', 'int176', 'uint232'), D), T[bo, i, i])  # fixed case 37528
ck(f(('bool', 'int176', 'uint24'), D), T[bo, i, i])  # fixed case 37529
ck(f(('bool', 'int176', 'uint240'), D), T[bo, i, i])  # fixed case 37530
ck(f(('bool', 'int176', 'uint248'), D), T[bo, i, i])  # fixed case 37531
ck(f(('bool', 'int176', 'uint256'), D), T[bo, i, i])  # fixed case 37532
ck(f(('bool', 'int176', 'uint32'), D), T[bo, i, i])  # fixed case 37533
ck(f(('bool', 'int176', 'uint40'), D), T[bo, i, i])  # fixed case 37534
ck(f(('bool', 'int176', 'uint48'), D), T[bo, i, i])  # fixed case 37535
ck(f(('bool', 'int176', 'uint56'), D), T[bo, i, i])  # fixed case 37536
ck(f(('bool', 'int176', 'uint64'), D), T[bo, i, i])  # fixed case 37537
ck(f(('bool', 'int176', 'uint72'), D), T[bo, i, i])  # fixed case 37538
ck(f(('bool', 'int176', 'uint8'), D), T[bo, i, i])  # fixed case 37539
ck(f(('bool', 'int176', 'uint80'), D), T[bo, i, i])  # fixed case 37540
ck(f(('bool', 'int176', 'uint88'), D), T[bo, i, i])  # fixed case 37541
ck(f(('bool', 'int176', 'uint96'), D), T[bo, i, i])  # fixed case 37542
ck(f(('bool', 'int176', '?'), D), T[bo, i, A])  # fixed case 37543
ck(f(('bool', 'int184', 'address'), D), T[bo, i, HA])  # fixed case 37544
ck(f(('bool', 'int184', 'address[]'), D), T[bo, i, A])  # fixed case 37545
ck(f(('bool', 'int184', 'bool'), D), T[bo, i, bo])  # fixed case 37546
ck(f(('bool', 'int184', 'bool[]'), D), T[bo, i, A])  # fixed case 37547
ck(f(('bool', 'int184', 'bytes'), D), T[bo, i, by])  # fixed case 37548
ck(f(('bool', 'int184', 'bytes1'), D), T[bo, i, by])  # fixed case 37549
ck(f(('bool', 'int184', 'bytes10'), D), T[bo, i, by])  # fixed case 37550
ck(f(('bool', 'int184', 'bytes11'), D), T[bo, i, by])  # fixed case 37551
ck(f(('bool', 'int184', 'bytes12'), D), T[bo, i, by])  # fixed case 37552
ck(f(('bool', 'int184', 'bytes13'), D), T[bo, i, by])  # fixed case 37553
ck(f(('bool', 'int184', 'bytes14'), D), T[bo, i, by])  # fixed case 37554
ck(f(('bool', 'int184', 'bytes15'), D), T[bo, i, by])  # fixed case 37555
ck(f(('bool', 'int184', 'bytes16'), D), T[bo, i, by])  # fixed case 37556
ck(f(('bool', 'int184', 'bytes17'), D), T[bo, i, by])  # fixed case 37557
ck(f(('bool', 'int184', 'bytes18'), D), T[bo, i, by])  # fixed case 37558
ck(f(('bool', 'int184', 'bytes19'), D), T[bo, i, by])  # fixed case 37559
ck(f(('bool', 'int184', 'bytes2'), D), T[bo, i, by])  # fixed case 37560
ck(f(('bool', 'int184', 'bytes20'), D), T[bo, i, by])  # fixed case 37561
ck(f(('bool', 'int184', 'bytes21'), D), T[bo, i, by])  # fixed case 37562
ck(f(('bool', 'int184', 'bytes22'), D), T[bo, i, by])  # fixed case 37563
ck(f(('bool', 'int184', 'bytes23'), D), T[bo, i, by])  # fixed case 37564
ck(f(('bool', 'int184', 'bytes24'), D), T[bo, i, by])  # fixed case 37565
ck(f(('bool', 'int184', 'bytes25'), D), T[bo, i, by])  # fixed case 37566
ck(f(('bool', 'int184', 'bytes26'), D), T[bo, i, by])  # fixed case 37567
ck(f(('bool', 'int184', 'bytes27'), D), T[bo, i, by])  # fixed case 37568
ck(f(('bool', 'int184', 'bytes28'), D), T[bo, i, by])  # fixed case 37569
ck(f(('bool', 'int184', 'bytes29'), D), T[bo, i, by])  # fixed case 37570
ck(f(('bool', 'int184', 'bytes3'), D), T[bo, i, by])  # fixed case 37571
ck(f(('bool', 'int184', 'bytes30'), D), T[bo, i, by])  # fixed case 37572
ck(f(('bool', 'int184', 'bytes31'), D), T[bo, i, by])  # fixed case 37573
ck(f(('bool', 'int184', 'bytes32'), D), T[bo, i, by])  # fixed case 37574
ck(f(('bool', 'int184', 'bytes4'), D), T[bo, i, by])  # fixed case 37575
ck(f(('bool', 'int184', 'bytes5'), D), T[bo, i, by])  # fixed case 37576
ck(f(('bool', 'int184', 'bytes6'), D), T[bo, i, by])  # fixed case 37577
ck(f(('bool', 'int184', 'bytes7'), D), T[bo, i, by])  # fixed case 37578
ck(f(('bool', 'int184', 'bytes8'), D), T[bo, i, by])  # fixed case 37579
ck(f(('bool', 'int184', 'bytes9'), D), T[bo, i, by])  # fixed case 37580
ck(f(('bool', 'int184', 'int104'), D), T[bo, i, i])  # fixed case 37581
ck(f(('bool', 'int184', 'int112'), D), T[bo, i, i])  # fixed case 37582
ck(f(('bool', 'int184', 'int120'), D), T[bo, i, i])  # fixed case 37583
ck(f(('bool', 'int184', 'int128'), D), T[bo, i, i])  # fixed case 37584
ck(f(('bool', 'int184', 'int136'), D), T[bo, i, i])  # fixed case 37585
ck(f(('bool', 'int184', 'int144'), D), T[bo, i, i])  # fixed case 37586
ck(f(('bool', 'int184', 'int152'), D), T[bo, i, i])  # fixed case 37587
ck(f(('bool', 'int184', 'int16'), D), T[bo, i, i])  # fixed case 37588
ck(f(('bool', 'int184', 'int160'), D), T[bo, i, i])  # fixed case 37589
ck(f(('bool', 'int184', 'int168'), D), T[bo, i, i])  # fixed case 37590
ck(f(('bool', 'int184', 'int176'), D), T[bo, i, i])  # fixed case 37591
ck(f(('bool', 'int184', 'int184'), D), T[bo, i, i])  # fixed case 37592
ck(f(('bool', 'int184', 'int192'), D), T[bo, i, i])  # fixed case 37593
ck(f(('bool', 'int184', 'int200'), D), T[bo, i, i])  # fixed case 37594
ck(f(('bool', 'int184', 'int208'), D), T[bo, i, i])  # fixed case 37595
ck(f(('bool', 'int184', 'int216'), D), T[bo, i, i])  # fixed case 37596
ck(f(('bool', 'int184', 'int224'), D), T[bo, i, i])  # fixed case 37597
ck(f(('bool', 'int184', 'int232'), D), T[bo, i, i])  # fixed case 37598
ck(f(('bool', 'int184', 'int24'), D), T[bo, i, i])  # fixed case 37599
ck(f(('bool', 'int184', 'int240'), D), T[bo, i, i])  # fixed case 37600
ck(f(('bool', 'int184', 'int248'), D), T[bo, i, i])  # fixed case 37601
ck(f(('bool', 'int184', 'int256'), D), T[bo, i, i])  # fixed case 37602
ck(f(('bool', 'int184', 'int32'), D), T[bo, i, i])  # fixed case 37603
ck(f(('bool', 'int184', 'int40'), D), T[bo, i, i])  # fixed case 37604
ck(f(('bool', 'int184', 'int48'), D), T[bo, i, i])  # fixed case 37605
ck(f(('bool', 'int184', 'int56'), D), T[bo, i, i])  # fixed case 37606
ck(f(('bool', 'int184', 'int64'), D), T[bo, i, i])  # fixed case 37607
ck(f(('bool', 'int184', 'int72'), D), T[bo, i, i])  # fixed case 37608
ck(f(('bool', 'int184', 'int8'), D), T[bo, i, i])  # fixed case 37609
ck(f(('bool', 'int184', 'int80'), D), T[bo, i, i])  # fixed case 37610
ck(f(('bool', 'int184', 'int88'), D), T[bo, i, i])  # fixed case 37611
ck(f(('bool', 'int184', 'int96'), D), T[bo, i, i])  # fixed case 37612
ck(f(('bool', 'int184', 'string'), D), T[bo, i, s])  # fixed case 37613
ck(f(('bool', 'int184', 'string[]'), D), T[bo, i, A])  # fixed case 37614
ck(f(('bool', 'int184', 'uint104'), D), T[bo, i, i])  # fixed case 37615
ck(f(('bool', 'int184', 'uint112'), D), T[bo, i, i])  # fixed case 37616
ck(f(('bool', 'int184', 'uint120'), D), T[bo, i, i])  # fixed case 37617
ck(f(('bool', 'int184', 'uint128'), D), T[bo, i, i])  # fixed case 37618
ck(f(('bool', 'int184', 'uint136'), D), T[bo, i, i])  # fixed case 37619
ck(f(('bool', 'int184', 'uint144'), D), T[bo, i, i])  # fixed case 37620
ck(f(('bool', 'int184', 'uint152'), D), T[bo, i, i])  # fixed case 37621
ck(f(('bool', 'int184', 'uint16'), D), T[bo, i, i])  # fixed case 37622
ck(f(('bool', 'int184', 'uint160'), D), T[bo, i, i])  # fixed case 37623
ck(f(('bool', 'int184', 'uint168'), D), T[bo, i, i])  # fixed case 37624
ck(f(('bool', 'int184', 'uint176'), D), T[bo, i, i])  # fixed case 37625
ck(f(('bool', 'int184', 'uint184'), D), T[bo, i, i])  # fixed case 37626
ck(f(('bool', 'int184', 'uint192'), D), T[bo, i, i])  # fixed case 37627
ck(f(('bool', 'int184', 'uint200'), D), T[bo, i, i])  # fixed case 37628
ck(f(('bool', 'int184', 'uint208'), D), T[bo, i, i])  # fixed case 37629
ck(f(('bool', 'int184', 'uint216'), D), T[bo, i, i])  # fixed case 37630
ck(f(('bool', 'int184', 'uint224'), D), T[bo, i, i])  # fixed case 37631
ck(f(('bool', 'int184', 'uint232'), D), T[bo, i, i])  # fixed case 37632
ck(f(('bool', 'int184', 'uint24'), D), T[bo, i, i])  # fixed case 37633
ck(f(('bool', 'int184', 'uint240'), D), T[bo, i, i])  # fixed case 37634
ck(f(('bool', 'int184', 'uint248'), D), T[bo, i, i])  # fixed case 37635
ck(f(('bool', 'int184', 'uint256'), D), T[bo, i, i])  # fixed case 37636
ck(f(('bool', 'int184', 'uint32'), D), T[bo, i, i])  # fixed case 37637
ck(f(('bool', 'int184', 'uint40'), D), T[bo, i, i])  # fixed case 37638
ck(f(('bool', 'int184', 'uint48'), D), T[bo, i, i])  # fixed case 37639
ck(f(('bool', 'int184', 'uint56'), D), T[bo, i, i])  # fixed case 37640
ck(f(('bool', 'int184', 'uint64'), D), T[bo, i, i])  # fixed case 37641
ck(f(('bool', 'int184', 'uint72'), D), T[bo, i, i])  # fixed case 37642
ck(f(('bool', 'int184', 'uint8'), D), T[bo, i, i])  # fixed case 37643
ck(f(('bool', 'int184', 'uint80'), D), T[bo, i, i])  # fixed case 37644
ck(f(('bool', 'int184', 'uint88'), D), T[bo, i, i])  # fixed case 37645
ck(f(('bool', 'int184', 'uint96'), D), T[bo, i, i])  # fixed case 37646
ck(f(('bool', 'int184', '?'), D), T[bo, i, A])  # fixed case 37647
ck(f(('bool', 'int192', 'address'), D), T[bo, i, HA])  # fixed case 37648
ck(f(('bool', 'int192', 'address[]'), D), T[bo, i, A])  # fixed case 37649
ck(f(('bool', 'int192', 'bool'), D), T[bo, i, bo])  # fixed case 37650
ck(f(('bool', 'int192', 'bool[]'), D), T[bo, i, A])  # fixed case 37651
ck(f(('bool', 'int192', 'bytes'), D), T[bo, i, by])  # fixed case 37652
ck(f(('bool', 'int192', 'bytes1'), D), T[bo, i, by])  # fixed case 37653
ck(f(('bool', 'int192', 'bytes10'), D), T[bo, i, by])  # fixed case 37654
ck(f(('bool', 'int192', 'bytes11'), D), T[bo, i, by])  # fixed case 37655
ck(f(('bool', 'int192', 'bytes12'), D), T[bo, i, by])  # fixed case 37656
ck(f(('bool', 'int192', 'bytes13'), D), T[bo, i, by])  # fixed case 37657
ck(f(('bool', 'int192', 'bytes14'), D), T[bo, i, by])  # fixed case 37658
ck(f(('bool', 'int192', 'bytes15'), D), T[bo, i, by])  # fixed case 37659
ck(f(('bool', 'int192', 'bytes16'), D), T[bo, i, by])  # fixed case 37660
ck(f(('bool', 'int192', 'bytes17'), D), T[bo, i, by])  # fixed case 37661
ck(f(('bool', 'int192', 'bytes18'), D), T[bo, i, by])  # fixed case 37662
ck(f(('bool', 'int192', 'bytes19'), D), T[bo, i, by])  # fixed case 37663
ck(f(('bool', 'int192', 'bytes2'), D), T[bo, i, by])  # fixed case 37664
ck(f(('bool', 'int192', 'bytes20'), D), T[bo, i, by])  # fixed case 37665
ck(f(('bool', 'int192', 'bytes21'), D), T[bo, i, by])  # fixed case 37666
ck(f(('bool', 'int192', 'bytes22'), D), T[bo, i, by])  # fixed case 37667
ck(f(('bool', 'int192', 'bytes23'), D), T[bo, i, by])  # fixed case 37668
ck(f(('bool', 'int192', 'bytes24'), D), T[bo, i, by])  # fixed case 37669
ck(f(('bool', 'int192', 'bytes25'), D), T[bo, i, by])  # fixed case 37670
ck(f(('bool', 'int192', 'bytes26'), D), T[bo, i, by])  # fixed case 37671
ck(f(('bool', 'int192', 'bytes27'), D), T[bo, i, by])  # fixed case 37672
ck(f(('bool', 'int192', 'bytes28'), D), T[bo, i, by])  # fixed case 37673
ck(f(('bool', 'int192', 'bytes29'), D), T[bo, i, by])  # fixed case 37674
ck(f(('bool', 'int192', 'bytes3'), D), T[bo, i, by])  # fixed case 37675
ck(f(('bool', 'int192', 'bytes30'), D), T[bo, i, by])  # fixed case 37676
ck(f(('bool', 'int192', 'bytes31'), D), T[bo, i, by])  # fixed case 37677
ck(f(('bool', 'int192', 'bytes32'), D), T[bo, i, by])  # fixed case 37678
ck(f(('bool', 'int192', 'bytes4'), D), T[bo, i, by])  # fixed case 37679
ck(f(('bool', 'int192', 'bytes5'), D), T[bo, i, by])  # fixed case 37680
ck(f(('bool', 'int192', 'bytes6'), D), T[bo, i, by])  # fixed case 37681
ck(f(('bool', 'int192', 'bytes7'), D), T[bo, i, by])  # fixed case 37682
ck(f(('bool', 'int192', 'bytes8'), D), T[bo, i, by])  # fixed case 37683
ck(f(('bool', 'int192', 'bytes9'), D), T[bo, i, by])  # fixed case 37684
ck(f(('bool', 'int192', 'int104'), D), T[bo, i, i])  # fixed case 37685
ck(f(('bool', 'int192', 'int112'), D), T[bo, i, i])  # fixed case 37686
ck(f(('bool', 'int192', 'int120'), D), T[bo, i, i])  # fixed case 37687
ck(f(('bool', 'int192', 'int128'), D), T[bo, i, i])  # fixed case 37688
ck(f(('bool', 'int192', 'int136'), D), T[bo, i, i])  # fixed case 37689
ck(f(('bool', 'int192', 'int144'), D), T[bo, i, i])  # fixed case 37690
ck(f(('bool', 'int192', 'int152'), D), T[bo, i, i])  # fixed case 37691
ck(f(('bool', 'int192', 'int16'), D), T[bo, i, i])  # fixed case 37692
ck(f(('bool', 'int192', 'int160'), D), T[bo, i, i])  # fixed case 37693
ck(f(('bool', 'int192', 'int168'), D), T[bo, i, i])  # fixed case 37694
ck(f(('bool', 'int192', 'int176'), D), T[bo, i, i])  # fixed case 37695
ck(f(('bool', 'int192', 'int184'), D), T[bo, i, i])  # fixed case 37696
ck(f(('bool', 'int192', 'int192'), D), T[bo, i, i])  # fixed case 37697
ck(f(('bool', 'int192', 'int200'), D), T[bo, i, i])  # fixed case 37698
ck(f(('bool', 'int192', 'int208'), D), T[bo, i, i])  # fixed case 37699
ck(f(('bool', 'int192', 'int216'), D), T[bo, i, i])  # fixed case 37700
ck(f(('bool', 'int192', 'int224'), D), T[bo, i, i])  # fixed case 37701
ck(f(('bool', 'int192', 'int232'), D), T[bo, i, i])  # fixed case 37702
ck(f(('bool', 'int192', 'int24'), D), T[bo, i, i])  # fixed case 37703
ck(f(('bool', 'int192', 'int240'), D), T[bo, i, i])  # fixed case 37704
ck(f(('bool', 'int192', 'int248'), D), T[bo, i, i])  # fixed case 37705
ck(f(('bool', 'int192', 'int256'), D), T[bo, i, i])  # fixed case 37706
ck(f(('bool', 'int192', 'int32'), D), T[bo, i, i])  # fixed case 37707
ck(f(('bool', 'int192', 'int40'), D), T[bo, i, i])  # fixed case 37708
ck(f(('bool', 'int192', 'int48'), D), T[bo, i, i])  # fixed case 37709
ck(f(('bool', 'int192', 'int56'), D), T[bo, i, i])  # fixed case 37710
ck(f(('bool', 'int192', 'int64'), D), T[bo, i, i])  # fixed case 37711
ck(f(('bool', 'int192', 'int72'), D), T[bo, i, i])  # fixed case 37712
ck(f(('bool', 'int192', 'int8'), D), T[bo, i, i])  # fixed case 37713
ck(f(('bool', 'int192', 'int80'), D), T[bo, i, i])  # fixed case 37714
ck(f(('bool', 'int192', 'int88'), D), T[bo, i, i])  # fixed case 37715
ck(f(('bool', 'int192', 'int96'), D), T[bo, i, i])  # fixed case 37716
ck(f(('bool', 'int192', 'string'), D), T[bo, i, s])  # fixed case 37717
ck(f(('bool', 'int192', 'string[]'), D), T[bo, i, A])  # fixed case 37718
ck(f(('bool', 'int192', 'uint104'), D), T[bo, i, i])  # fixed case 37719
ck(f(('bool', 'int192', 'uint112'), D), T[bo, i, i])  # fixed case 37720
ck(f(('bool', 'int192', 'uint120'), D), T[bo, i, i])  # fixed case 37721
ck(f(('bool', 'int192', 'uint128'), D), T[bo, i, i])  # fixed case 37722
ck(f(('bool', 'int192', 'uint136'), D), T[bo, i, i])  # fixed case 37723
ck(f(('bool', 'int192', 'uint144'), D), T[bo, i, i])  # fixed case 37724
ck(f(('bool', 'int192', 'uint152'), D), T[bo, i, i])  # fixed case 37725
ck(f(('bool', 'int192', 'uint16'), D), T[bo, i, i])  # fixed case 37726
ck(f(('bool', 'int192', 'uint160'), D), T[bo, i, i])  # fixed case 37727
ck(f(('bool', 'int192', 'uint168'), D), T[bo, i, i])  # fixed case 37728
ck(f(('bool', 'int192', 'uint176'), D), T[bo, i, i])  # fixed case 37729
ck(f(('bool', 'int192', 'uint184'), D), T[bo, i, i])  # fixed case 37730
ck(f(('bool', 'int192', 'uint192'), D), T[bo, i, i])  # fixed case 37731
ck(f(('bool', 'int192', 'uint200'), D), T[bo, i, i])  # fixed case 37732
ck(f(('bool', 'int192', 'uint208'), D), T[bo, i, i])  # fixed case 37733
ck(f(('bool', 'int192', 'uint216'), D), T[bo, i, i])  # fixed case 37734
ck(f(('bool', 'int192', 'uint224'), D), T[bo, i, i])  # fixed case 37735
ck(f(('bool', 'int192', 'uint232'), D), T[bo, i, i])  # fixed case 37736
ck(f(('bool', 'int192', 'uint24'), D), T[bo, i, i])  # fixed case 37737
ck(f(('bool', 'int192', 'uint240'), D), T[bo, i, i])  # fixed case 37738
ck(f(('bool', 'int192', 'uint248'), D), T[bo, i, i])  # fixed case 37739
ck(f(('bool', 'int192', 'uint256'), D), T[bo, i, i])  # fixed case 37740
ck(f(('bool', 'int192', 'uint32'), D), T[bo, i, i])  # fixed case 37741
ck(f(('bool', 'int192', 'uint40'), D), T[bo, i, i])  # fixed case 37742
ck(f(('bool', 'int192', 'uint48'), D), T[bo, i, i])  # fixed case 37743
ck(f(('bool', 'int192', 'uint56'), D), T[bo, i, i])  # fixed case 37744
ck(f(('bool', 'int192', 'uint64'), D), T[bo, i, i])  # fixed case 37745
ck(f(('bool', 'int192', 'uint72'), D), T[bo, i, i])  # fixed case 37746
ck(f(('bool', 'int192', 'uint8'), D), T[bo, i, i])  # fixed case 37747
ck(f(('bool', 'int192', 'uint80'), D), T[bo, i, i])  # fixed case 37748
ck(f(('bool', 'int192', 'uint88'), D), T[bo, i, i])  # fixed case 37749
ck(f(('bool', 'int192', 'uint96'), D), T[bo, i, i])  # fixed case 37750
ck(f(('bool', 'int192', '?'), D), T[bo, i, A])  # fixed case 37751
ck(f(('bool', 'int200', 'address'), D), T[bo, i, HA])  # fixed case 37752
ck(f(('bool', 'int200', 'address[]'), D), T[bo, i, A])  # fixed case 37753
ck(f(('bool', 'int200', 'bool'), D), T[bo, i, bo])  # fixed case 37754
ck(f(('bool', 'int200', 'bool[]'), D), T[bo, i, A])  # fixed case 37755
ck(f(('bool', 'int200', 'bytes'), D), T[bo, i, by])  # fixed case 37756
ck(f(('bool', 'int200', 'bytes1'), D), T[bo, i, by])  # fixed case 37757
ck(f(('bool', 'int200', 'bytes10'), D), T[bo, i, by])  # fixed case 37758
ck(f(('bool', 'int200', 'bytes11'), D), T[bo, i, by])  # fixed case 37759
ck(f(('bool', 'int200', 'bytes12'), D), T[bo, i, by])  # fixed case 37760
ck(f(('bool', 'int200', 'bytes13'), D), T[bo, i, by])  # fixed case 37761
ck(f(('bool', 'int200', 'bytes14'), D), T[bo, i, by])  # fixed case 37762
ck(f(('bool', 'int200', 'bytes15'), D), T[bo, i, by])  # fixed case 37763
ck(f(('bool', 'int200', 'bytes16'), D), T[bo, i, by])  # fixed case 37764
ck(f(('bool', 'int200', 'bytes17'), D), T[bo, i, by])  # fixed case 37765
ck(f(('bool', 'int200', 'bytes18'), D), T[bo, i, by])  # fixed case 37766
ck(f(('bool', 'int200', 'bytes19'), D), T[bo, i, by])  # fixed case 37767
ck(f(('bool', 'int200', 'bytes2'), D), T[bo, i, by])  # fixed case 37768
ck(f(('bool', 'int200', 'bytes20'), D), T[bo, i, by])  # fixed case 37769
ck(f(('bool', 'int200', 'bytes21'), D), T[bo, i, by])  # fixed case 37770
ck(f(('bool', 'int200', 'bytes22'), D), T[bo, i, by])  # fixed case 37771
ck(f(('bool', 'int200', 'bytes23'), D), T[bo, i, by])  # fixed case 37772
ck(f(('bool', 'int200', 'bytes24'), D), T[bo, i, by])  # fixed case 37773
ck(f(('bool', 'int200', 'bytes25'), D), T[bo, i, by])  # fixed case 37774
ck(f(('bool', 'int200', 'bytes26'), D), T[bo, i, by])  # fixed case 37775
ck(f(('bool', 'int200', 'bytes27'), D), T[bo, i, by])  # fixed case 37776
ck(f(('bool', 'int200', 'bytes28'), D), T[bo, i, by])  # fixed case 37777
ck(f(('bool', 'int200', 'bytes29'), D), T[bo, i, by])  # fixed case 37778
ck(f(('bool', 'int200', 'bytes3'), D), T[bo, i, by])  # fixed case 37779
ck(f(('bool', 'int200', 'bytes30'), D), T[bo, i, by])  # fixed case 37780
ck(f(('bool', 'int200', 'bytes31'), D), T[bo, i, by])  # fixed case 37781
ck(f(('bool', 'int200', 'bytes32'), D), T[bo, i, by])  # fixed case 37782
ck(f(('bool', 'int200', 'bytes4'), D), T[bo, i, by])  # fixed case 37783
ck(f(('bool', 'int200', 'bytes5'), D), T[bo, i, by])  # fixed case 37784
ck(f(('bool', 'int200', 'bytes6'), D), T[bo, i, by])  # fixed case 37785
ck(f(('bool', 'int200', 'bytes7'), D), T[bo, i, by])  # fixed case 37786
ck(f(('bool', 'int200', 'bytes8'), D), T[bo, i, by])  # fixed case 37787
ck(f(('bool', 'int200', 'bytes9'), D), T[bo, i, by])  # fixed case 37788
ck(f(('bool', 'int200', 'int104'), D), T[bo, i, i])  # fixed case 37789
ck(f(('bool', 'int200', 'int112'), D), T[bo, i, i])  # fixed case 37790
ck(f(('bool', 'int200', 'int120'), D), T[bo, i, i])  # fixed case 37791
ck(f(('bool', 'int200', 'int128'), D), T[bo, i, i])  # fixed case 37792
ck(f(('bool', 'int200', 'int136'), D), T[bo, i, i])  # fixed case 37793
ck(f(('bool', 'int200', 'int144'), D), T[bo, i, i])  # fixed case 37794
ck(f(('bool', 'int200', 'int152'), D), T[bo, i, i])  # fixed case 37795
ck(f(('bool', 'int200', 'int16'), D), T[bo, i, i])  # fixed case 37796
ck(f(('bool', 'int200', 'int160'), D), T[bo, i, i])  # fixed case 37797
ck(f(('bool', 'int200', 'int168'), D), T[bo, i, i])  # fixed case 37798
ck(f(('bool', 'int200', 'int176'), D), T[bo, i, i])  # fixed case 37799
ck(f(('bool', 'int200', 'int184'), D), T[bo, i, i])  # fixed case 37800
ck(f(('bool', 'int200', 'int192'), D), T[bo, i, i])  # fixed case 37801
ck(f(('bool', 'int200', 'int200'), D), T[bo, i, i])  # fixed case 37802
ck(f(('bool', 'int200', 'int208'), D), T[bo, i, i])  # fixed case 37803
ck(f(('bool', 'int200', 'int216'), D), T[bo, i, i])  # fixed case 37804
ck(f(('bool', 'int200', 'int224'), D), T[bo, i, i])  # fixed case 37805
ck(f(('bool', 'int200', 'int232'), D), T[bo, i, i])  # fixed case 37806
ck(f(('bool', 'int200', 'int24'), D), T[bo, i, i])  # fixed case 37807
ck(f(('bool', 'int200', 'int240'), D), T[bo, i, i])  # fixed case 37808
ck(f(('bool', 'int200', 'int248'), D), T[bo, i, i])  # fixed case 37809
ck(f(('bool', 'int200', 'int256'), D), T[bo, i, i])  # fixed case 37810
ck(f(('bool', 'int200', 'int32'), D), T[bo, i, i])  # fixed case 37811
ck(f(('bool', 'int200', 'int40'), D), T[bo, i, i])  # fixed case 37812
ck(f(('bool', 'int200', 'int48'), D), T[bo, i, i])  # fixed case 37813
ck(f(('bool', 'int200', 'int56'), D), T[bo, i, i])  # fixed case 37814
ck(f(('bool', 'int200', 'int64'), D), T[bo, i, i])  # fixed case 37815
ck(f(('bool', 'int200', 'int72'), D), T[bo, i, i])  # fixed case 37816
ck(f(('bool', 'int200', 'int8'), D), T[bo, i, i])  # fixed case 37817
ck(f(('bool', 'int200', 'int80'), D), T[bo, i, i])  # fixed case 37818
ck(f(('bool', 'int200', 'int88'), D), T[bo, i, i])  # fixed case 37819
ck(f(('bool', 'int200', 'int96'), D), T[bo, i, i])  # fixed case 37820
ck(f(('bool', 'int200', 'string'), D), T[bo, i, s])  # fixed case 37821
ck(f(('bool', 'int200', 'string[]'), D), T[bo, i, A])  # fixed case 37822
ck(f(('bool', 'int200', 'uint104'), D), T[bo, i, i])  # fixed case 37823
ck(f(('bool', 'int200', 'uint112'), D), T[bo, i, i])  # fixed case 37824
ck(f(('bool', 'int200', 'uint120'), D), T[bo, i, i])  # fixed case 37825
ck(f(('bool', 'int200', 'uint128'), D), T[bo, i, i])  # fixed case 37826
ck(f(('bool', 'int200', 'uint136'), D), T[bo, i, i])  # fixed case 37827
ck(f(('bool', 'int200', 'uint144'), D), T[bo, i, i])  # fixed case 37828
ck(f(('bool', 'int200', 'uint152'), D), T[bo, i, i])  # fixed case 37829
ck(f(('bool', 'int200', 'uint16'), D), T[bo, i, i])  # fixed case 37830
ck(f(('bool', 'int200', 'uint160'), D), T[bo, i, i])  # fixed case 37831
ck(f(('bool', 'int200', 'uint168'), D), T[bo, i, i])  # fixed case 37832
ck(f(('bool', 'int200', 'uint176'), D), T[bo, i, i])  # fixed case 37833
ck(f(('bool', 'int200', 'uint184'), D), T[bo, i, i])  # fixed case 37834
ck(f(('bool', 'int200', 'uint192'), D), T[bo, i, i])  # fixed case 37835
ck(f(('bool', 'int200', 'uint200'), D), T[bo, i, i])  # fixed case 37836
ck(f(('bool', 'int200', 'uint208'), D), T[bo, i, i])  # fixed case 37837
ck(f(('bool', 'int200', 'uint216'), D), T[bo, i, i])  # fixed case 37838
ck(f(('bool', 'int200', 'uint224'), D), T[bo, i, i])  # fixed case 37839
ck(f(('bool', 'int200', 'uint232'), D), T[bo, i, i])  # fixed case 37840
ck(f(('bool', 'int200', 'uint24'), D), T[bo, i, i])  # fixed case 37841
ck(f(('bool', 'int200', 'uint240'), D), T[bo, i, i])  # fixed case 37842
ck(f(('bool', 'int200', 'uint248'), D), T[bo, i, i])  # fixed case 37843
ck(f(('bool', 'int200', 'uint256'), D), T[bo, i, i])  # fixed case 37844
ck(f(('bool', 'int200', 'uint32'), D), T[bo, i, i])  # fixed case 37845
ck(f(('bool', 'int200', 'uint40'), D), T[bo, i, i])  # fixed case 37846
ck(f(('bool', 'int200', 'uint48'), D), T[bo, i, i])  # fixed case 37847
ck(f(('bool', 'int200', 'uint56'), D), T[bo, i, i])  # fixed case 37848
ck(f(('bool', 'int200', 'uint64'), D), T[bo, i, i])  # fixed case 37849
ck(f(('bool', 'int200', 'uint72'), D), T[bo, i, i])  # fixed case 37850
ck(f(('bool', 'int200', 'uint8'), D), T[bo, i, i])  # fixed case 37851
ck(f(('bool', 'int200', 'uint80'), D), T[bo, i, i])  # fixed case 37852
ck(f(('bool', 'int200', 'uint88'), D), T[bo, i, i])  # fixed case 37853
ck(f(('bool', 'int200', 'uint96'), D), T[bo, i, i])  # fixed case 37854
ck(f(('bool', 'int200', '?'), D), T[bo, i, A])  # fixed case 37855
ck(f(('bool', 'int208', 'address'), D), T[bo, i, HA])  # fixed case 37856
ck(f(('bool', 'int208', 'address[]'), D), T[bo, i, A])  # fixed case 37857
ck(f(('bool', 'int208', 'bool'), D), T[bo, i, bo])  # fixed case 37858
ck(f(('bool', 'int208', 'bool[]'), D), T[bo, i, A])  # fixed case 37859
ck(f(('bool', 'int208', 'bytes'), D), T[bo, i, by])  # fixed case 37860
ck(f(('bool', 'int208', 'bytes1'), D), T[bo, i, by])  # fixed case 37861
ck(f(('bool', 'int208', 'bytes10'), D), T[bo, i, by])  # fixed case 37862
ck(f(('bool', 'int208', 'bytes11'), D), T[bo, i, by])  # fixed case 37863
ck(f(('bool', 'int208', 'bytes12'), D), T[bo, i, by])  # fixed case 37864
ck(f(('bool', 'int208', 'bytes13'), D), T[bo, i, by])  # fixed case 37865
ck(f(('bool', 'int208', 'bytes14'), D), T[bo, i, by])  # fixed case 37866
ck(f(('bool', 'int208', 'bytes15'), D), T[bo, i, by])  # fixed case 37867
ck(f(('bool', 'int208', 'bytes16'), D), T[bo, i, by])  # fixed case 37868
ck(f(('bool', 'int208', 'bytes17'), D), T[bo, i, by])  # fixed case 37869
ck(f(('bool', 'int208', 'bytes18'), D), T[bo, i, by])  # fixed case 37870
ck(f(('bool', 'int208', 'bytes19'), D), T[bo, i, by])  # fixed case 37871
ck(f(('bool', 'int208', 'bytes2'), D), T[bo, i, by])  # fixed case 37872
ck(f(('bool', 'int208', 'bytes20'), D), T[bo, i, by])  # fixed case 37873
ck(f(('bool', 'int208', 'bytes21'), D), T[bo, i, by])  # fixed case 37874
ck(f(('bool', 'int208', 'bytes22'), D), T[bo, i, by])  # fixed case 37875
ck(f(('bool', 'int208', 'bytes23'), D), T[bo, i, by])  # fixed case 37876
ck(f(('bool', 'int208', 'bytes24'), D), T[bo, i, by])  # fixed case 37877
ck(f(('bool', 'int208', 'bytes25'), D), T[bo, i, by])  # fixed case 37878
ck(f(('bool', 'int208', 'bytes26'), D), T[bo, i, by])  # fixed case 37879
ck(f(('bool', 'int208', 'bytes27'), D), T[bo, i, by])  # fixed case 37880
ck(f(('bool', 'int208', 'bytes28'), D), T[bo, i, by])  # fixed case 37881
ck(f(('bool', 'int208', 'bytes29'), D), T[bo, i, by])  # fixed case 37882
ck(f(('bool', 'int208', 'bytes3'), D), T[bo, i, by])  # fixed case 37883
ck(f(('bool', 'int208', 'bytes30'), D), T[bo, i, by])  # fixed case 37884
ck(f(('bool', 'int208', 'bytes31'), D), T[bo, i, by])  # fixed case 37885
ck(f(('bool', 'int208', 'bytes32'), D), T[bo, i, by])  # fixed case 37886
ck(f(('bool', 'int208', 'bytes4'), D), T[bo, i, by])  # fixed case 37887
ck(f(('bool', 'int208', 'bytes5'), D), T[bo, i, by])  # fixed case 37888
ck(f(('bool', 'int208', 'bytes6'), D), T[bo, i, by])  # fixed case 37889
ck(f(('bool', 'int208', 'bytes7'), D), T[bo, i, by])  # fixed case 37890
ck(f(('bool', 'int208', 'bytes8'), D), T[bo, i, by])  # fixed case 37891
ck(f(('bool', 'int208', 'bytes9'), D), T[bo, i, by])  # fixed case 37892
ck(f(('bool', 'int208', 'int104'), D), T[bo, i, i])  # fixed case 37893
ck(f(('bool', 'int208', 'int112'), D), T[bo, i, i])  # fixed case 37894
ck(f(('bool', 'int208', 'int120'), D), T[bo, i, i])  # fixed case 37895
ck(f(('bool', 'int208', 'int128'), D), T[bo, i, i])  # fixed case 37896
ck(f(('bool', 'int208', 'int136'), D), T[bo, i, i])  # fixed case 37897
ck(f(('bool', 'int208', 'int144'), D), T[bo, i, i])  # fixed case 37898
ck(f(('bool', 'int208', 'int152'), D), T[bo, i, i])  # fixed case 37899
ck(f(('bool', 'int208', 'int16'), D), T[bo, i, i])  # fixed case 37900
ck(f(('bool', 'int208', 'int160'), D), T[bo, i, i])  # fixed case 37901
ck(f(('bool', 'int208', 'int168'), D), T[bo, i, i])  # fixed case 37902
ck(f(('bool', 'int208', 'int176'), D), T[bo, i, i])  # fixed case 37903
ck(f(('bool', 'int208', 'int184'), D), T[bo, i, i])  # fixed case 37904
ck(f(('bool', 'int208', 'int192'), D), T[bo, i, i])  # fixed case 37905
ck(f(('bool', 'int208', 'int200'), D), T[bo, i, i])  # fixed case 37906
ck(f(('bool', 'int208', 'int208'), D), T[bo, i, i])  # fixed case 37907
ck(f(('bool', 'int208', 'int216'), D), T[bo, i, i])  # fixed case 37908
ck(f(('bool', 'int208', 'int224'), D), T[bo, i, i])  # fixed case 37909
ck(f(('bool', 'int208', 'int232'), D), T[bo, i, i])  # fixed case 37910
ck(f(('bool', 'int208', 'int24'), D), T[bo, i, i])  # fixed case 37911
ck(f(('bool', 'int208', 'int240'), D), T[bo, i, i])  # fixed case 37912
ck(f(('bool', 'int208', 'int248'), D), T[bo, i, i])  # fixed case 37913
ck(f(('bool', 'int208', 'int256'), D), T[bo, i, i])  # fixed case 37914
ck(f(('bool', 'int208', 'int32'), D), T[bo, i, i])  # fixed case 37915
ck(f(('bool', 'int208', 'int40'), D), T[bo, i, i])  # fixed case 37916
ck(f(('bool', 'int208', 'int48'), D), T[bo, i, i])  # fixed case 37917
ck(f(('bool', 'int208', 'int56'), D), T[bo, i, i])  # fixed case 37918
ck(f(('bool', 'int208', 'int64'), D), T[bo, i, i])  # fixed case 37919
ck(f(('bool', 'int208', 'int72'), D), T[bo, i, i])  # fixed case 37920
ck(f(('bool', 'int208', 'int8'), D), T[bo, i, i])  # fixed case 37921
ck(f(('bool', 'int208', 'int80'), D), T[bo, i, i])  # fixed case 37922
ck(f(('bool', 'int208', 'int88'), D), T[bo, i, i])  # fixed case 37923
ck(f(('bool', 'int208', 'int96'), D), T[bo, i, i])  # fixed case 37924
ck(f(('bool', 'int208', 'string'), D), T[bo, i, s])  # fixed case 37925
ck(f(('bool', 'int208', 'string[]'), D), T[bo, i, A])  # fixed case 37926
ck(f(('bool', 'int208', 'uint104'), D), T[bo, i, i])  # fixed case 37927
ck(f(('bool', 'int208', 'uint112'), D), T[bo, i, i])  # fixed case 37928
ck(f(('bool', 'int208', 'uint120'), D), T[bo, i, i])  # fixed case 37929
ck(f(('bool', 'int208', 'uint128'), D), T[bo, i, i])  # fixed case 37930
ck(f(('bool', 'int208', 'uint136'), D), T[bo, i, i])  # fixed case 37931
ck(f(('bool', 'int208', 'uint144'), D), T[bo, i, i])  # fixed case 37932
ck(f(('bool', 'int208', 'uint152'), D), T[bo, i, i])  # fixed case 37933
ck(f(('bool', 'int208', 'uint16'), D), T[bo, i, i])  # fixed case 37934
ck(f(('bool', 'int208', 'uint160'), D), T[bo, i, i])  # fixed case 37935
ck(f(('bool', 'int208', 'uint168'), D), T[bo, i, i])  # fixed case 37936
ck(f(('bool', 'int208', 'uint176'), D), T[bo, i, i])  # fixed case 37937
ck(f(('bool', 'int208', 'uint184'), D), T[bo, i, i])  # fixed case 37938
ck(f(('bool', 'int208', 'uint192'), D), T[bo, i, i])  # fixed case 37939
ck(f(('bool', 'int208', 'uint200'), D), T[bo, i, i])  # fixed case 37940
ck(f(('bool', 'int208', 'uint208'), D), T[bo, i, i])  # fixed case 37941
ck(f(('bool', 'int208', 'uint216'), D), T[bo, i, i])  # fixed case 37942
ck(f(('bool', 'int208', 'uint224'), D), T[bo, i, i])  # fixed case 37943
ck(f(('bool', 'int208', 'uint232'), D), T[bo, i, i])  # fixed case 37944
ck(f(('bool', 'int208', 'uint24'), D), T[bo, i, i])  # fixed case 37945
ck(f(('bool', 'int208', 'uint240'), D), T[bo, i, i])  # fixed case 37946
ck(f(('bool', 'int208', 'uint248'), D), T[bo, i, i])  # fixed case 37947
ck(f(('bool', 'int208', 'uint256'), D), T[bo, i, i])  # fixed case 37948
ck(f(('bool', 'int208', 'uint32'), D), T[bo, i, i])  # fixed case 37949
ck(f(('bool', 'int208', 'uint40'), D), T[bo, i, i])  # fixed case 37950
ck(f(('bool', 'int208', 'uint48'), D), T[bo, i, i])  # fixed case 37951
ck(f(('bool', 'int208', 'uint56'), D), T[bo, i, i])  # fixed case 37952
ck(f(('bool', 'int208', 'uint64'), D), T[bo, i, i])  # fixed case 37953
ck(f(('bool', 'int208', 'uint72'), D), T[bo, i, i])  # fixed case 37954
ck(f(('bool', 'int208', 'uint8'), D), T[bo, i, i])  # fixed case 37955
ck(f(('bool', 'int208', 'uint80'), D), T[bo, i, i])  # fixed case 37956
ck(f(('bool', 'int208', 'uint88'), D), T[bo, i, i])  # fixed case 37957
ck(f(('bool', 'int208', 'uint96'), D), T[bo, i, i])  # fixed case 37958
ck(f(('bool', 'int208', '?'), D), T[bo, i, A])  # fixed case 37959
ck(f(('bool', 'int216', 'address'), D), T[bo, i, HA])  # fixed case 37960
ck(f(('bool', 'int216', 'address[]'), D), T[bo, i, A])  # fixed case 37961
ck(f(('bool', 'int216', 'bool'), D), T[bo, i, bo])  # fixed case 37962
ck(f(('bool', 'int216', 'bool[]'), D), T[bo, i, A])  # fixed case 37963
ck(f(('bool', 'int216', 'bytes'), D), T[bo, i, by])  # fixed case 37964
ck(f(('bool', 'int216', 'bytes1'), D), T[bo, i, by])  # fixed case 37965
ck(f(('bool', 'int216', 'bytes10'), D), T[bo, i, by])  # fixed case 37966
ck(f(('bool', 'int216', 'bytes11'), D), T[bo, i, by])  # fixed case 37967
ck(f(('bool', 'int216', 'bytes12'), D), T[bo, i, by])  # fixed case 37968
ck(f(('bool', 'int216', 'bytes13'), D), T[bo, i, by])  # fixed case 37969
ck(f(('bool', 'int216', 'bytes14'), D), T[bo, i, by])  # fixed case 37970
ck(f(('bool', 'int216', 'bytes15'), D), T[bo, i, by])  # fixed case 37971
ck(f(('bool', 'int216', 'bytes16'), D), T[bo, i, by])  # fixed case 37972
ck(f(('bool', 'int216', 'bytes17'), D), T[bo, i, by])  # fixed case 37973
ck(f(('bool', 'int216', 'bytes18'), D), T[bo, i, by])  # fixed case 37974
ck(f(('bool', 'int216', 'bytes19'), D), T[bo, i, by])  # fixed case 37975
ck(f(('bool', 'int216', 'bytes2'), D), T[bo, i, by])  # fixed case 37976
ck(f(('bool', 'int216', 'bytes20'), D), T[bo, i, by])  # fixed case 37977
ck(f(('bool', 'int216', 'bytes21'), D), T[bo, i, by])  # fixed case 37978
ck(f(('bool', 'int216', 'bytes22'), D), T[bo, i, by])  # fixed case 37979
ck(f(('bool', 'int216', 'bytes23'), D), T[bo, i, by])  # fixed case 37980
ck(f(('bool', 'int216', 'bytes24'), D), T[bo, i, by])  # fixed case 37981
ck(f(('bool', 'int216', 'bytes25'), D), T[bo, i, by])  # fixed case 37982
ck(f(('bool', 'int216', 'bytes26'), D), T[bo, i, by])  # fixed case 37983
ck(f(('bool', 'int216', 'bytes27'), D), T[bo, i, by])  # fixed case 37984
ck(f(('bool', 'int216', 'bytes28'), D), T[bo, i, by])  # fixed case 37985
ck(f(('bool', 'int216', 'bytes29'), D), T[bo, i, by])  # fixed case 37986
ck(f(('bool', 'int216', 'bytes3'), D), T[bo, i, by])  # fixed case 37987
ck(f(('bool', 'int216', 'bytes30'), D), T[bo, i, by])  # fixed case 37988
ck(f(('bool', 'int216', 'bytes31'), D), T[bo, i, by])  # fixed case 37989
ck(f(('bool', 'int216', 'bytes32'), D), T[bo, i, by])  # fixed case 37990
ck(f(('bool', 'int216', 'bytes4'), D), T[bo, i, by])  # fixed case 37991
ck(f(('bool', 'int216', 'bytes5'), D), T[bo, i, by])  # fixed case 37992
ck(f(('bool', 'int216', 'bytes6'), D), T[bo, i, by])  # fixed case 37993
ck(f(('bool', 'int216', 'bytes7'), D), T[bo, i, by])  # fixed case 37994
ck(f(('bool', 'int216', 'bytes8'), D), T[bo, i, by])  # fixed case 37995
ck(f(('bool', 'int216', 'bytes9'), D), T[bo, i, by])  # fixed case 37996
ck(f(('bool', 'int216', 'int104'), D), T[bo, i, i])  # fixed case 37997
ck(f(('bool', 'int216', 'int112'), D), T[bo, i, i])  # fixed case 37998
ck(f(('bool', 'int216', 'int120'), D), T[bo, i, i])  # fixed case 37999
ck(f(('bool', 'int216', 'int128'), D), T[bo, i, i])  # fixed case 38000
ck(f(('bool', 'int216', 'int136'), D), T[bo, i, i])  # fixed case 38001
ck(f(('bool', 'int216', 'int144'), D), T[bo, i, i])  # fixed case 38002
ck(f(('bool', 'int216', 'int152'), D), T[bo, i, i])  # fixed case 38003
ck(f(('bool', 'int216', 'int16'), D), T[bo, i, i])  # fixed case 38004
ck(f(('bool', 'int216', 'int160'), D), T[bo, i, i])  # fixed case 38005
ck(f(('bool', 'int216', 'int168'), D), T[bo, i, i])  # fixed case 38006
ck(f(('bool', 'int216', 'int176'), D), T[bo, i, i])  # fixed case 38007
ck(f(('bool', 'int216', 'int184'), D), T[bo, i, i])  # fixed case 38008
ck(f(('bool', 'int216', 'int192'), D), T[bo, i, i])  # fixed case 38009
ck(f(('bool', 'int216', 'int200'), D), T[bo, i, i])  # fixed case 38010
ck(f(('bool', 'int216', 'int208'), D), T[bo, i, i])  # fixed case 38011
ck(f(('bool', 'int216', 'int216'), D), T[bo, i, i])  # fixed case 38012
ck(f(('bool', 'int216', 'int224'), D), T[bo, i, i])  # fixed case 38013
ck(f(('bool', 'int216', 'int232'), D), T[bo, i, i])  # fixed case 38014
ck(f(('bool', 'int216', 'int24'), D), T[bo, i, i])  # fixed case 38015
ck(f(('bool', 'int216', 'int240'), D), T[bo, i, i])  # fixed case 38016
ck(f(('bool', 'int216', 'int248'), D), T[bo, i, i])  # fixed case 38017
ck(f(('bool', 'int216', 'int256'), D), T[bo, i, i])  # fixed case 38018
ck(f(('bool', 'int216', 'int32'), D), T[bo, i, i])  # fixed case 38019
ck(f(('bool', 'int216', 'int40'), D), T[bo, i, i])  # fixed case 38020
ck(f(('bool', 'int216', 'int48'), D), T[bo, i, i])  # fixed case 38021
ck(f(('bool', 'int216', 'int56'), D), T[bo, i, i])  # fixed case 38022
ck(f(('bool', 'int216', 'int64'), D), T[bo, i, i])  # fixed case 38023
ck(f(('bool', 'int216', 'int72'), D), T[bo, i, i])  # fixed case 38024
ck(f(('bool', 'int216', 'int8'), D), T[bo, i, i])  # fixed case 38025
ck(f(('bool', 'int216', 'int80'), D), T[bo, i, i])  # fixed case 38026
ck(f(('bool', 'int216', 'int88'), D), T[bo, i, i])  # fixed case 38027
ck(f(('bool', 'int216', 'int96'), D), T[bo, i, i])  # fixed case 38028
ck(f(('bool', 'int216', 'string'), D), T[bo, i, s])  # fixed case 38029
ck(f(('bool', 'int216', 'string[]'), D), T[bo, i, A])  # fixed case 38030
ck(f(('bool', 'int216', 'uint104'), D), T[bo, i, i])  # fixed case 38031
ck(f(('bool', 'int216', 'uint112'), D), T[bo, i, i])  # fixed case 38032
ck(f(('bool', 'int216', 'uint120'), D), T[bo, i, i])  # fixed case 38033
ck(f(('bool', 'int216', 'uint128'), D), T[bo, i, i])  # fixed case 38034
ck(f(('bool', 'int216', 'uint136'), D), T[bo, i, i])  # fixed case 38035
ck(f(('bool', 'int216', 'uint144'), D), T[bo, i, i])  # fixed case 38036
ck(f(('bool', 'int216', 'uint152'), D), T[bo, i, i])  # fixed case 38037
ck(f(('bool', 'int216', 'uint16'), D), T[bo, i, i])  # fixed case 38038
ck(f(('bool', 'int216', 'uint160'), D), T[bo, i, i])  # fixed case 38039
ck(f(('bool', 'int216', 'uint168'), D), T[bo, i, i])  # fixed case 38040
ck(f(('bool', 'int216', 'uint176'), D), T[bo, i, i])  # fixed case 38041
ck(f(('bool', 'int216', 'uint184'), D), T[bo, i, i])  # fixed case 38042
ck(f(('bool', 'int216', 'uint192'), D), T[bo, i, i])  # fixed case 38043
ck(f(('bool', 'int216', 'uint200'), D), T[bo, i, i])  # fixed case 38044
ck(f(('bool', 'int216', 'uint208'), D), T[bo, i, i])  # fixed case 38045
ck(f(('bool', 'int216', 'uint216'), D), T[bo, i, i])  # fixed case 38046
ck(f(('bool', 'int216', 'uint224'), D), T[bo, i, i])  # fixed case 38047
ck(f(('bool', 'int216', 'uint232'), D), T[bo, i, i])  # fixed case 38048
ck(f(('bool', 'int216', 'uint24'), D), T[bo, i, i])  # fixed case 38049
ck(f(('bool', 'int216', 'uint240'), D), T[bo, i, i])  # fixed case 38050
ck(f(('bool', 'int216', 'uint248'), D), T[bo, i, i])  # fixed case 38051
ck(f(('bool', 'int216', 'uint256'), D), T[bo, i, i])  # fixed case 38052
ck(f(('bool', 'int216', 'uint32'), D), T[bo, i, i])  # fixed case 38053
ck(f(('bool', 'int216', 'uint40'), D), T[bo, i, i])  # fixed case 38054
ck(f(('bool', 'int216', 'uint48'), D), T[bo, i, i])  # fixed case 38055
ck(f(('bool', 'int216', 'uint56'), D), T[bo, i, i])  # fixed case 38056
ck(f(('bool', 'int216', 'uint64'), D), T[bo, i, i])  # fixed case 38057
ck(f(('bool', 'int216', 'uint72'), D), T[bo, i, i])  # fixed case 38058
ck(f(('bool', 'int216', 'uint8'), D), T[bo, i, i])  # fixed case 38059
ck(f(('bool', 'int216', 'uint80'), D), T[bo, i, i])  # fixed case 38060
ck(f(('bool', 'int216', 'uint88'), D), T[bo, i, i])  # fixed case 38061
ck(f(('bool', 'int216', 'uint96'), D), T[bo, i, i])  # fixed case 38062
ck(f(('bool', 'int216', '?'), D), T[bo, i, A])  # fixed case 38063
ck(f(('bool', 'int224', 'address'), D), T[bo, i, HA])  # fixed case 38064
ck(f(('bool', 'int224', 'address[]'), D), T[bo, i, A])  # fixed case 38065
ck(f(('bool', 'int224', 'bool'), D), T[bo, i, bo])  # fixed case 38066
ck(f(('bool', 'int224', 'bool[]'), D), T[bo, i, A])  # fixed case 38067
ck(f(('bool', 'int224', 'bytes'), D), T[bo, i, by])  # fixed case 38068
ck(f(('bool', 'int224', 'bytes1'), D), T[bo, i, by])  # fixed case 38069
ck(f(('bool', 'int224', 'bytes10'), D), T[bo, i, by])  # fixed case 38070
ck(f(('bool', 'int224', 'bytes11'), D), T[bo, i, by])  # fixed case 38071
ck(f(('bool', 'int224', 'bytes12'), D), T[bo, i, by])  # fixed case 38072
ck(f(('bool', 'int224', 'bytes13'), D), T[bo, i, by])  # fixed case 38073
ck(f(('bool', 'int224', 'bytes14'), D), T[bo, i, by])  # fixed case 38074
ck(f(('bool', 'int224', 'bytes15'), D), T[bo, i, by])  # fixed case 38075
ck(f(('bool', 'int224', 'bytes16'), D), T[bo, i, by])  # fixed case 38076
ck(f(('bool', 'int224', 'bytes17'), D), T[bo, i, by])  # fixed case 38077
ck(f(('bool', 'int224', 'bytes18'), D), T[bo, i, by])  # fixed case 38078
ck(f(('bool', 'int224', 'bytes19'), D), T[bo, i, by])  # fixed case 38079
ck(f(('bool', 'int224', 'bytes2'), D), T[bo, i, by])  # fixed case 38080
ck(f(('bool', 'int224', 'bytes20'), D), T[bo, i, by])  # fixed case 38081
ck(f(('bool', 'int224', 'bytes21'), D), T[bo, i, by])  # fixed case 38082
ck(f(('bool', 'int224', 'bytes22'), D), T[bo, i, by])  # fixed case 38083
ck(f(('bool', 'int224', 'bytes23'), D), T[bo, i, by])  # fixed case 38084
ck(f(('bool', 'int224', 'bytes24'), D), T[bo, i, by])  # fixed case 38085
ck(f(('bool', 'int224', 'bytes25'), D), T[bo, i, by])  # fixed case 38086
ck(f(('bool', 'int224', 'bytes26'), D), T[bo, i, by])  # fixed case 38087
ck(f(('bool', 'int224', 'bytes27'), D), T[bo, i, by])  # fixed case 38088
ck(f(('bool', 'int224', 'bytes28'), D), T[bo, i, by])  # fixed case 38089
ck(f(('bool', 'int224', 'bytes29'), D), T[bo, i, by])  # fixed case 38090
ck(f(('bool', 'int224', 'bytes3'), D), T[bo, i, by])  # fixed case 38091
ck(f(('bool', 'int224', 'bytes30'), D), T[bo, i, by])  # fixed case 38092
ck(f(('bool', 'int224', 'bytes31'), D), T[bo, i, by])  # fixed case 38093
ck(f(('bool', 'int224', 'bytes32'), D), T[bo, i, by])  # fixed case 38094
ck(f(('bool', 'int224', 'bytes4'), D), T[bo, i, by])  # fixed case 38095
ck(f(('bool', 'int224', 'bytes5'), D), T[bo, i, by])  # fixed case 38096
ck(f(('bool', 'int224', 'bytes6'), D), T[bo, i, by])  # fixed case 38097
ck(f(('bool', 'int224', 'bytes7'), D), T[bo, i, by])  # fixed case 38098
ck(f(('bool', 'int224', 'bytes8'), D), T[bo, i, by])  # fixed case 38099
ck(f(('bool', 'int224', 'bytes9'), D), T[bo, i, by])  # fixed case 38100
ck(f(('bool', 'int224', 'int104'), D), T[bo, i, i])  # fixed case 38101
ck(f(('bool', 'int224', 'int112'), D), T[bo, i, i])  # fixed case 38102
ck(f(('bool', 'int224', 'int120'), D), T[bo, i, i])  # fixed case 38103
ck(f(('bool', 'int224', 'int128'), D), T[bo, i, i])  # fixed case 38104
ck(f(('bool', 'int224', 'int136'), D), T[bo, i, i])  # fixed case 38105
ck(f(('bool', 'int224', 'int144'), D), T[bo, i, i])  # fixed case 38106
ck(f(('bool', 'int224', 'int152'), D), T[bo, i, i])  # fixed case 38107
ck(f(('bool', 'int224', 'int16'), D), T[bo, i, i])  # fixed case 38108
ck(f(('bool', 'int224', 'int160'), D), T[bo, i, i])  # fixed case 38109
ck(f(('bool', 'int224', 'int168'), D), T[bo, i, i])  # fixed case 38110
ck(f(('bool', 'int224', 'int176'), D), T[bo, i, i])  # fixed case 38111
ck(f(('bool', 'int224', 'int184'), D), T[bo, i, i])  # fixed case 38112
ck(f(('bool', 'int224', 'int192'), D), T[bo, i, i])  # fixed case 38113
ck(f(('bool', 'int224', 'int200'), D), T[bo, i, i])  # fixed case 38114
ck(f(('bool', 'int224', 'int208'), D), T[bo, i, i])  # fixed case 38115
ck(f(('bool', 'int224', 'int216'), D), T[bo, i, i])  # fixed case 38116
ck(f(('bool', 'int224', 'int224'), D), T[bo, i, i])  # fixed case 38117
ck(f(('bool', 'int224', 'int232'), D), T[bo, i, i])  # fixed case 38118
ck(f(('bool', 'int224', 'int24'), D), T[bo, i, i])  # fixed case 38119
ck(f(('bool', 'int224', 'int240'), D), T[bo, i, i])  # fixed case 38120
ck(f(('bool', 'int224', 'int248'), D), T[bo, i, i])  # fixed case 38121
ck(f(('bool', 'int224', 'int256'), D), T[bo, i, i])  # fixed case 38122
ck(f(('bool', 'int224', 'int32'), D), T[bo, i, i])  # fixed case 38123
ck(f(('bool', 'int224', 'int40'), D), T[bo, i, i])  # fixed case 38124
ck(f(('bool', 'int224', 'int48'), D), T[bo, i, i])  # fixed case 38125
ck(f(('bool', 'int224', 'int56'), D), T[bo, i, i])  # fixed case 38126
ck(f(('bool', 'int224', 'int64'), D), T[bo, i, i])  # fixed case 38127
ck(f(('bool', 'int224', 'int72'), D), T[bo, i, i])  # fixed case 38128
ck(f(('bool', 'int224', 'int8'), D), T[bo, i, i])  # fixed case 38129
ck(f(('bool', 'int224', 'int80'), D), T[bo, i, i])  # fixed case 38130
ck(f(('bool', 'int224', 'int88'), D), T[bo, i, i])  # fixed case 38131
ck(f(('bool', 'int224', 'int96'), D), T[bo, i, i])  # fixed case 38132
ck(f(('bool', 'int224', 'string'), D), T[bo, i, s])  # fixed case 38133
ck(f(('bool', 'int224', 'string[]'), D), T[bo, i, A])  # fixed case 38134
ck(f(('bool', 'int224', 'uint104'), D), T[bo, i, i])  # fixed case 38135
ck(f(('bool', 'int224', 'uint112'), D), T[bo, i, i])  # fixed case 38136
ck(f(('bool', 'int224', 'uint120'), D), T[bo, i, i])  # fixed case 38137
ck(f(('bool', 'int224', 'uint128'), D), T[bo, i, i])  # fixed case 38138
ck(f(('bool', 'int224', 'uint136'), D), T[bo, i, i])  # fixed case 38139
ck(f(('bool', 'int224', 'uint144'), D), T[bo, i, i])  # fixed case 38140
ck(f(('bool', 'int224', 'uint152'), D), T[bo, i, i])  # fixed case 38141
ck(f(('bool', 'int224', 'uint16'), D), T[bo, i, i])  # fixed case 38142
ck(f(('bool', 'int224', 'uint160'), D), T[bo, i, i])  # fixed case 38143
ck(f(('bool', 'int224', 'uint168'), D), T[bo, i, i])  # fixed case 38144
ck(f(('bool', 'int224', 'uint176'), D), T[bo, i, i])  # fixed case 38145
ck(f(('bool', 'int224', 'uint184'), D), T[bo, i, i])  # fixed case 38146
ck(f(('bool', 'int224', 'uint192'), D), T[bo, i, i])  # fixed case 38147
ck(f(('bool', 'int224', 'uint200'), D), T[bo, i, i])  # fixed case 38148
ck(f(('bool', 'int224', 'uint208'), D), T[bo, i, i])  # fixed case 38149
ck(f(('bool', 'int224', 'uint216'), D), T[bo, i, i])  # fixed case 38150
ck(f(('bool', 'int224', 'uint224'), D), T[bo, i, i])  # fixed case 38151
ck(f(('bool', 'int224', 'uint232'), D), T[bo, i, i])  # fixed case 38152
ck(f(('bool', 'int224', 'uint24'), D), T[bo, i, i])  # fixed case 38153
ck(f(('bool', 'int224', 'uint240'), D), T[bo, i, i])  # fixed case 38154
ck(f(('bool', 'int224', 'uint248'), D), T[bo, i, i])  # fixed case 38155
ck(f(('bool', 'int224', 'uint256'), D), T[bo, i, i])  # fixed case 38156
ck(f(('bool', 'int224', 'uint32'), D), T[bo, i, i])  # fixed case 38157
ck(f(('bool', 'int224', 'uint40'), D), T[bo, i, i])  # fixed case 38158
ck(f(('bool', 'int224', 'uint48'), D), T[bo, i, i])  # fixed case 38159
ck(f(('bool', 'int224', 'uint56'), D), T[bo, i, i])  # fixed case 38160
ck(f(('bool', 'int224', 'uint64'), D), T[bo, i, i])  # fixed case 38161
ck(f(('bool', 'int224', 'uint72'), D), T[bo, i, i])  # fixed case 38162
ck(f(('bool', 'int224', 'uint8'), D), T[bo, i, i])  # fixed case 38163
ck(f(('bool', 'int224', 'uint80'), D), T[bo, i, i])  # fixed case 38164
ck(f(('bool', 'int224', 'uint88'), D), T[bo, i, i])  # fixed case 38165
ck(f(('bool', 'int224', 'uint96'), D), T[bo, i, i])  # fixed case 38166
ck(f(('bool', 'int224', '?'), D), T[bo, i, A])  # fixed case 38167
ck(f(('bool', 'int232', 'address'), D), T[bo, i, HA])  # fixed case 38168
ck(f(('bool', 'int232', 'address[]'), D), T[bo, i, A])  # fixed case 38169
ck(f(('bool', 'int232', 'bool'), D), T[bo, i, bo])  # fixed case 38170
ck(f(('bool', 'int232', 'bool[]'), D), T[bo, i, A])  # fixed case 38171
ck(f(('bool', 'int232', 'bytes'), D), T[bo, i, by])  # fixed case 38172
ck(f(('bool', 'int232', 'bytes1'), D), T[bo, i, by])  # fixed case 38173
ck(f(('bool', 'int232', 'bytes10'), D), T[bo, i, by])  # fixed case 38174
ck(f(('bool', 'int232', 'bytes11'), D), T[bo, i, by])  # fixed case 38175
ck(f(('bool', 'int232', 'bytes12'), D), T[bo, i, by])  # fixed case 38176
ck(f(('bool', 'int232', 'bytes13'), D), T[bo, i, by])  # fixed case 38177
ck(f(('bool', 'int232', 'bytes14'), D), T[bo, i, by])  # fixed case 38178
ck(f(('bool', 'int232', 'bytes15'), D), T[bo, i, by])  # fixed case 38179
ck(f(('bool', 'int232', 'bytes16'), D), T[bo, i, by])  # fixed case 38180
ck(f(('bool', 'int232', 'bytes17'), D), T[bo, i, by])  # fixed case 38181
ck(f(('bool', 'int232', 'bytes18'), D), T[bo, i, by])  # fixed case 38182
ck(f(('bool', 'int232', 'bytes19'), D), T[bo, i, by])  # fixed case 38183
ck(f(('bool', 'int232', 'bytes2'), D), T[bo, i, by])  # fixed case 38184
ck(f(('bool', 'int232', 'bytes20'), D), T[bo, i, by])  # fixed case 38185
ck(f(('bool', 'int232', 'bytes21'), D), T[bo, i, by])  # fixed case 38186
ck(f(('bool', 'int232', 'bytes22'), D), T[bo, i, by])  # fixed case 38187
ck(f(('bool', 'int232', 'bytes23'), D), T[bo, i, by])  # fixed case 38188
ck(f(('bool', 'int232', 'bytes24'), D), T[bo, i, by])  # fixed case 38189
ck(f(('bool', 'int232', 'bytes25'), D), T[bo, i, by])  # fixed case 38190
ck(f(('bool', 'int232', 'bytes26'), D), T[bo, i, by])  # fixed case 38191
ck(f(('bool', 'int232', 'bytes27'), D), T[bo, i, by])  # fixed case 38192
ck(f(('bool', 'int232', 'bytes28'), D), T[bo, i, by])  # fixed case 38193
ck(f(('bool', 'int232', 'bytes29'), D), T[bo, i, by])  # fixed case 38194
ck(f(('bool', 'int232', 'bytes3'), D), T[bo, i, by])  # fixed case 38195
ck(f(('bool', 'int232', 'bytes30'), D), T[bo, i, by])  # fixed case 38196
ck(f(('bool', 'int232', 'bytes31'), D), T[bo, i, by])  # fixed case 38197
ck(f(('bool', 'int232', 'bytes32'), D), T[bo, i, by])  # fixed case 38198
ck(f(('bool', 'int232', 'bytes4'), D), T[bo, i, by])  # fixed case 38199
ck(f(('bool', 'int232', 'bytes5'), D), T[bo, i, by])  # fixed case 38200
ck(f(('bool', 'int232', 'bytes6'), D), T[bo, i, by])  # fixed case 38201
ck(f(('bool', 'int232', 'bytes7'), D), T[bo, i, by])  # fixed case 38202
ck(f(('bool', 'int232', 'bytes8'), D), T[bo, i, by])  # fixed case 38203
ck(f(('bool', 'int232', 'bytes9'), D), T[bo, i, by])  # fixed case 38204
ck(f(('bool', 'int232', 'int104'), D), T[bo, i, i])  # fixed case 38205
ck(f(('bool', 'int232', 'int112'), D), T[bo, i, i])  # fixed case 38206
ck(f(('bool', 'int232', 'int120'), D), T[bo, i, i])  # fixed case 38207
ck(f(('bool', 'int232', 'int128'), D), T[bo, i, i])  # fixed case 38208
ck(f(('bool', 'int232', 'int136'), D), T[bo, i, i])  # fixed case 38209
ck(f(('bool', 'int232', 'int144'), D), T[bo, i, i])  # fixed case 38210
ck(f(('bool', 'int232', 'int152'), D), T[bo, i, i])  # fixed case 38211
ck(f(('bool', 'int232', 'int16'), D), T[bo, i, i])  # fixed case 38212
ck(f(('bool', 'int232', 'int160'), D), T[bo, i, i])  # fixed case 38213
ck(f(('bool', 'int232', 'int168'), D), T[bo, i, i])  # fixed case 38214
ck(f(('bool', 'int232', 'int176'), D), T[bo, i, i])  # fixed case 38215
ck(f(('bool', 'int232', 'int184'), D), T[bo, i, i])  # fixed case 38216
ck(f(('bool', 'int232', 'int192'), D), T[bo, i, i])  # fixed case 38217
ck(f(('bool', 'int232', 'int200'), D), T[bo, i, i])  # fixed case 38218
ck(f(('bool', 'int232', 'int208'), D), T[bo, i, i])  # fixed case 38219
ck(f(('bool', 'int232', 'int216'), D), T[bo, i, i])  # fixed case 38220
ck(f(('bool', 'int232', 'int224'), D), T[bo, i, i])  # fixed case 38221
ck(f(('bool', 'int232', 'int232'), D), T[bo, i, i])  # fixed case 38222
ck(f(('bool', 'int232', 'int24'), D), T[bo, i, i])  # fixed case 38223
ck(f(('bool', 'int232', 'int240'), D), T[bo, i, i])  # fixed case 38224
ck(f(('bool', 'int232', 'int248'), D), T[bo, i, i])  # fixed case 38225
ck(f(('bool', 'int232', 'int256'), D), T[bo, i, i])  # fixed case 38226
ck(f(('bool', 'int232', 'int32'), D), T[bo, i, i])  # fixed case 38227
ck(f(('bool', 'int232', 'int40'), D), T[bo, i, i])  # fixed case 38228
ck(f(('bool', 'int232', 'int48'), D), T[bo, i, i])  # fixed case 38229
ck(f(('bool', 'int232', 'int56'), D), T[bo, i, i])  # fixed case 38230
ck(f(('bool', 'int232', 'int64'), D), T[bo, i, i])  # fixed case 38231
ck(f(('bool', 'int232', 'int72'), D), T[bo, i, i])  # fixed case 38232
ck(f(('bool', 'int232', 'int8'), D), T[bo, i, i])  # fixed case 38233
ck(f(('bool', 'int232', 'int80'), D), T[bo, i, i])  # fixed case 38234
ck(f(('bool', 'int232', 'int88'), D), T[bo, i, i])  # fixed case 38235
ck(f(('bool', 'int232', 'int96'), D), T[bo, i, i])  # fixed case 38236
ck(f(('bool', 'int232', 'string'), D), T[bo, i, s])  # fixed case 38237
ck(f(('bool', 'int232', 'string[]'), D), T[bo, i, A])  # fixed case 38238
ck(f(('bool', 'int232', 'uint104'), D), T[bo, i, i])  # fixed case 38239
ck(f(('bool', 'int232', 'uint112'), D), T[bo, i, i])  # fixed case 38240
ck(f(('bool', 'int232', 'uint120'), D), T[bo, i, i])  # fixed case 38241
ck(f(('bool', 'int232', 'uint128'), D), T[bo, i, i])  # fixed case 38242
ck(f(('bool', 'int232', 'uint136'), D), T[bo, i, i])  # fixed case 38243
ck(f(('bool', 'int232', 'uint144'), D), T[bo, i, i])  # fixed case 38244
ck(f(('bool', 'int232', 'uint152'), D), T[bo, i, i])  # fixed case 38245
ck(f(('bool', 'int232', 'uint16'), D), T[bo, i, i])  # fixed case 38246
ck(f(('bool', 'int232', 'uint160'), D), T[bo, i, i])  # fixed case 38247
ck(f(('bool', 'int232', 'uint168'), D), T[bo, i, i])  # fixed case 38248
ck(f(('bool', 'int232', 'uint176'), D), T[bo, i, i])  # fixed case 38249
ck(f(('bool', 'int232', 'uint184'), D), T[bo, i, i])  # fixed case 38250
ck(f(('bool', 'int232', 'uint192'), D), T[bo, i, i])  # fixed case 38251
ck(f(('bool', 'int232', 'uint200'), D), T[bo, i, i])  # fixed case 38252
ck(f(('bool', 'int232', 'uint208'), D), T[bo, i, i])  # fixed case 38253
ck(f(('bool', 'int232', 'uint216'), D), T[bo, i, i])  # fixed case 38254
ck(f(('bool', 'int232', 'uint224'), D), T[bo, i, i])  # fixed case 38255
ck(f(('bool', 'int232', 'uint232'), D), T[bo, i, i])  # fixed case 38256
ck(f(('bool', 'int232', 'uint24'), D), T[bo, i, i])  # fixed case 38257
ck(f(('bool', 'int232', 'uint240'), D), T[bo, i, i])  # fixed case 38258
ck(f(('bool', 'int232', 'uint248'), D), T[bo, i, i])  # fixed case 38259
ck(f(('bool', 'int232', 'uint256'), D), T[bo, i, i])  # fixed case 38260
ck(f(('bool', 'int232', 'uint32'), D), T[bo, i, i])  # fixed case 38261
ck(f(('bool', 'int232', 'uint40'), D), T[bo, i, i])  # fixed case 38262
ck(f(('bool', 'int232', 'uint48'), D), T[bo, i, i])  # fixed case 38263
ck(f(('bool', 'int232', 'uint56'), D), T[bo, i, i])  # fixed case 38264
ck(f(('bool', 'int232', 'uint64'), D), T[bo, i, i])  # fixed case 38265
ck(f(('bool', 'int232', 'uint72'), D), T[bo, i, i])  # fixed case 38266
ck(f(('bool', 'int232', 'uint8'), D), T[bo, i, i])  # fixed case 38267
ck(f(('bool', 'int232', 'uint80'), D), T[bo, i, i])  # fixed case 38268
ck(f(('bool', 'int232', 'uint88'), D), T[bo, i, i])  # fixed case 38269
ck(f(('bool', 'int232', 'uint96'), D), T[bo, i, i])  # fixed case 38270
ck(f(('bool', 'int232', '?'), D), T[bo, i, A])  # fixed case 38271
ck(f(('bool', 'int24', 'address'), D), T[bo, i, HA])  # fixed case 38272
ck(f(('bool', 'int24', 'address[]'), D), T[bo, i, A])  # fixed case 38273
ck(f(('bool', 'int24', 'bool'), D), T[bo, i, bo])  # fixed case 38274
ck(f(('bool', 'int24', 'bool[]'), D), T[bo, i, A])  # fixed case 38275
ck(f(('bool', 'int24', 'bytes'), D), T[bo, i, by])  # fixed case 38276
ck(f(('bool', 'int24', 'bytes1'), D), T[bo, i, by])  # fixed case 38277
ck(f(('bool', 'int24', 'bytes10'), D), T[bo, i, by])  # fixed case 38278
ck(f(('bool', 'int24', 'bytes11'), D), T[bo, i, by])  # fixed case 38279
ck(f(('bool', 'int24', 'bytes12'), D), T[bo, i, by])  # fixed case 38280
ck(f(('bool', 'int24', 'bytes13'), D), T[bo, i, by])  # fixed case 38281
ck(f(('bool', 'int24', 'bytes14'), D), T[bo, i, by])  # fixed case 38282
ck(f(('bool', 'int24', 'bytes15'), D), T[bo, i, by])  # fixed case 38283
ck(f(('bool', 'int24', 'bytes16'), D), T[bo, i, by])  # fixed case 38284
ck(f(('bool', 'int24', 'bytes17'), D), T[bo, i, by])  # fixed case 38285
ck(f(('bool', 'int24', 'bytes18'), D), T[bo, i, by])  # fixed case 38286
ck(f(('bool', 'int24', 'bytes19'), D), T[bo, i, by])  # fixed case 38287
ck(f(('bool', 'int24', 'bytes2'), D), T[bo, i, by])  # fixed case 38288
ck(f(('bool', 'int24', 'bytes20'), D), T[bo, i, by])  # fixed case 38289
ck(f(('bool', 'int24', 'bytes21'), D), T[bo, i, by])  # fixed case 38290
ck(f(('bool', 'int24', 'bytes22'), D), T[bo, i, by])  # fixed case 38291
ck(f(('bool', 'int24', 'bytes23'), D), T[bo, i, by])  # fixed case 38292
ck(f(('bool', 'int24', 'bytes24'), D), T[bo, i, by])  # fixed case 38293
ck(f(('bool', 'int24', 'bytes25'), D), T[bo, i, by])  # fixed case 38294
ck(f(('bool', 'int24', 'bytes26'), D), T[bo, i, by])  # fixed case 38295
ck(f(('bool', 'int24', 'bytes27'), D), T[bo, i, by])  # fixed case 38296
ck(f(('bool', 'int24', 'bytes28'), D), T[bo, i, by])  # fixed case 38297
ck(f(('bool', 'int24', 'bytes29'), D), T[bo, i, by])  # fixed case 38298
ck(f(('bool', 'int24', 'bytes3'), D), T[bo, i, by])  # fixed case 38299
ck(f(('bool', 'int24', 'bytes30'), D), T[bo, i, by])  # fixed case 38300
ck(f(('bool', 'int24', 'bytes31'), D), T[bo, i, by])  # fixed case 38301
ck(f(('bool', 'int24', 'bytes32'), D), T[bo, i, by])  # fixed case 38302
ck(f(('bool', 'int24', 'bytes4'), D), T[bo, i, by])  # fixed case 38303
ck(f(('bool', 'int24', 'bytes5'), D), T[bo, i, by])  # fixed case 38304
ck(f(('bool', 'int24', 'bytes6'), D), T[bo, i, by])  # fixed case 38305
ck(f(('bool', 'int24', 'bytes7'), D), T[bo, i, by])  # fixed case 38306
ck(f(('bool', 'int24', 'bytes8'), D), T[bo, i, by])  # fixed case 38307
ck(f(('bool', 'int24', 'bytes9'), D), T[bo, i, by])  # fixed case 38308
ck(f(('bool', 'int24', 'int104'), D), T[bo, i, i])  # fixed case 38309
ck(f(('bool', 'int24', 'int112'), D), T[bo, i, i])  # fixed case 38310
ck(f(('bool', 'int24', 'int120'), D), T[bo, i, i])  # fixed case 38311
ck(f(('bool', 'int24', 'int128'), D), T[bo, i, i])  # fixed case 38312
ck(f(('bool', 'int24', 'int136'), D), T[bo, i, i])  # fixed case 38313
ck(f(('bool', 'int24', 'int144'), D), T[bo, i, i])  # fixed case 38314
ck(f(('bool', 'int24', 'int152'), D), T[bo, i, i])  # fixed case 38315
ck(f(('bool', 'int24', 'int16'), D), T[bo, i, i])  # fixed case 38316
ck(f(('bool', 'int24', 'int160'), D), T[bo, i, i])  # fixed case 38317
ck(f(('bool', 'int24', 'int168'), D), T[bo, i, i])  # fixed case 38318
ck(f(('bool', 'int24', 'int176'), D), T[bo, i, i])  # fixed case 38319
ck(f(('bool', 'int24', 'int184'), D), T[bo, i, i])  # fixed case 38320
ck(f(('bool', 'int24', 'int192'), D), T[bo, i, i])  # fixed case 38321
ck(f(('bool', 'int24', 'int200'), D), T[bo, i, i])  # fixed case 38322
ck(f(('bool', 'int24', 'int208'), D), T[bo, i, i])  # fixed case 38323
ck(f(('bool', 'int24', 'int216'), D), T[bo, i, i])  # fixed case 38324
ck(f(('bool', 'int24', 'int224'), D), T[bo, i, i])  # fixed case 38325
ck(f(('bool', 'int24', 'int232'), D), T[bo, i, i])  # fixed case 38326
ck(f(('bool', 'int24', 'int24'), D), T[bo, i, i])  # fixed case 38327
ck(f(('bool', 'int24', 'int240'), D), T[bo, i, i])  # fixed case 38328
ck(f(('bool', 'int24', 'int248'), D), T[bo, i, i])  # fixed case 38329
ck(f(('bool', 'int24', 'int256'), D), T[bo, i, i])  # fixed case 38330
ck(f(('bool', 'int24', 'int32'), D), T[bo, i, i])  # fixed case 38331
ck(f(('bool', 'int24', 'int40'), D), T[bo, i, i])  # fixed case 38332
ck(f(('bool', 'int24', 'int48'), D), T[bo, i, i])  # fixed case 38333
ck(f(('bool', 'int24', 'int56'), D), T[bo, i, i])  # fixed case 38334
ck(f(('bool', 'int24', 'int64'), D), T[bo, i, i])  # fixed case 38335
ck(f(('bool', 'int24', 'int72'), D), T[bo, i, i])  # fixed case 38336
ck(f(('bool', 'int24', 'int8'), D), T[bo, i, i])  # fixed case 38337
ck(f(('bool', 'int24', 'int80'), D), T[bo, i, i])  # fixed case 38338
ck(f(('bool', 'int24', 'int88'), D), T[bo, i, i])  # fixed case 38339
ck(f(('bool', 'int24', 'int96'), D), T[bo, i, i])  # fixed case 38340
ck(f(('bool', 'int24', 'string'), D), T[bo, i, s])  # fixed case 38341
ck(f(('bool', 'int24', 'string[]'), D), T[bo, i, A])  # fixed case 38342
ck(f(('bool', 'int24', 'uint104'), D), T[bo, i, i])  # fixed case 38343
ck(f(('bool', 'int24', 'uint112'), D), T[bo, i, i])  # fixed case 38344
ck(f(('bool', 'int24', 'uint120'), D), T[bo, i, i])  # fixed case 38345
ck(f(('bool', 'int24', 'uint128'), D), T[bo, i, i])  # fixed case 38346
ck(f(('bool', 'int24', 'uint136'), D), T[bo, i, i])  # fixed case 38347
ck(f(('bool', 'int24', 'uint144'), D), T[bo, i, i])  # fixed case 38348
ck(f(('bool', 'int24', 'uint152'), D), T[bo, i, i])  # fixed case 38349
ck(f(('bool', 'int24', 'uint16'), D), T[bo, i, i])  # fixed case 38350
ck(f(('bool', 'int24', 'uint160'), D), T[bo, i, i])  # fixed case 38351
ck(f(('bool', 'int24', 'uint168'), D), T[bo, i, i])  # fixed case 38352
ck(f(('bool', 'int24', 'uint176'), D), T[bo, i, i])  # fixed case 38353
ck(f(('bool', 'int24', 'uint184'), D), T[bo, i, i])  # fixed case 38354
ck(f(('bool', 'int24', 'uint192'), D), T[bo, i, i])  # fixed case 38355
ck(f(('bool', 'int24', 'uint200'), D), T[bo, i, i])  # fixed case 38356
ck(f(('bool', 'int24', 'uint208'), D), T[bo, i, i])  # fixed case 38357
ck(f(('bool', 'int24', 'uint216'), D), T[bo, i, i])  # fixed case 38358
ck(f(('bool', 'int24', 'uint224'), D), T[bo, i, i])  # fixed case 38359
ck(f(('bool', 'int24', 'uint232'), D), T[bo, i, i])  # fixed case 38360
ck(f(('bool', 'int24', 'uint24'), D), T[bo, i, i])  # fixed case 38361
ck(f(('bool', 'int24', 'uint240'), D), T[bo, i, i])  # fixed case 38362
ck(f(('bool', 'int24', 'uint248'), D), T[bo, i, i])  # fixed case 38363
ck(f(('bool', 'int24', 'uint256'), D), T[bo, i, i])  # fixed case 38364
ck(f(('bool', 'int24', 'uint32'), D), T[bo, i, i])  # fixed case 38365
ck(f(('bool', 'int24', 'uint40'), D), T[bo, i, i])  # fixed case 38366
ck(f(('bool', 'int24', 'uint48'), D), T[bo, i, i])  # fixed case 38367
ck(f(('bool', 'int24', 'uint56'), D), T[bo, i, i])  # fixed case 38368
ck(f(('bool', 'int24', 'uint64'), D), T[bo, i, i])  # fixed case 38369
ck(f(('bool', 'int24', 'uint72'), D), T[bo, i, i])  # fixed case 38370
ck(f(('bool', 'int24', 'uint8'), D), T[bo, i, i])  # fixed case 38371
ck(f(('bool', 'int24', 'uint80'), D), T[bo, i, i])  # fixed case 38372
ck(f(('bool', 'int24', 'uint88'), D), T[bo, i, i])  # fixed case 38373
ck(f(('bool', 'int24', 'uint96'), D), T[bo, i, i])  # fixed case 38374
ck(f(('bool', 'int24', '?'), D), T[bo, i, A])  # fixed case 38375
ck(f(('bool', 'int240', 'address'), D), T[bo, i, HA])  # fixed case 38376
ck(f(('bool', 'int240', 'address[]'), D), T[bo, i, A])  # fixed case 38377
ck(f(('bool', 'int240', 'bool'), D), T[bo, i, bo])  # fixed case 38378
ck(f(('bool', 'int240', 'bool[]'), D), T[bo, i, A])  # fixed case 38379
ck(f(('bool', 'int240', 'bytes'), D), T[bo, i, by])  # fixed case 38380
ck(f(('bool', 'int240', 'bytes1'), D), T[bo, i, by])  # fixed case 38381
ck(f(('bool', 'int240', 'bytes10'), D), T[bo, i, by])  # fixed case 38382
ck(f(('bool', 'int240', 'bytes11'), D), T[bo, i, by])  # fixed case 38383
ck(f(('bool', 'int240', 'bytes12'), D), T[bo, i, by])  # fixed case 38384
ck(f(('bool', 'int240', 'bytes13'), D), T[bo, i, by])  # fixed case 38385
ck(f(('bool', 'int240', 'bytes14'), D), T[bo, i, by])  # fixed case 38386
ck(f(('bool', 'int240', 'bytes15'), D), T[bo, i, by])  # fixed case 38387
ck(f(('bool', 'int240', 'bytes16'), D), T[bo, i, by])  # fixed case 38388
ck(f(('bool', 'int240', 'bytes17'), D), T[bo, i, by])  # fixed case 38389
ck(f(('bool', 'int240', 'bytes18'), D), T[bo, i, by])  # fixed case 38390
ck(f(('bool', 'int240', 'bytes19'), D), T[bo, i, by])  # fixed case 38391
ck(f(('bool', 'int240', 'bytes2'), D), T[bo, i, by])  # fixed case 38392
ck(f(('bool', 'int240', 'bytes20'), D), T[bo, i, by])  # fixed case 38393
ck(f(('bool', 'int240', 'bytes21'), D), T[bo, i, by])  # fixed case 38394
ck(f(('bool', 'int240', 'bytes22'), D), T[bo, i, by])  # fixed case 38395
ck(f(('bool', 'int240', 'bytes23'), D), T[bo, i, by])  # fixed case 38396
ck(f(('bool', 'int240', 'bytes24'), D), T[bo, i, by])  # fixed case 38397
ck(f(('bool', 'int240', 'bytes25'), D), T[bo, i, by])  # fixed case 38398
ck(f(('bool', 'int240', 'bytes26'), D), T[bo, i, by])  # fixed case 38399
ck(f(('bool', 'int240', 'bytes27'), D), T[bo, i, by])  # fixed case 38400
ck(f(('bool', 'int240', 'bytes28'), D), T[bo, i, by])  # fixed case 38401
ck(f(('bool', 'int240', 'bytes29'), D), T[bo, i, by])  # fixed case 38402
ck(f(('bool', 'int240', 'bytes3'), D), T[bo, i, by])  # fixed case 38403
ck(f(('bool', 'int240', 'bytes30'), D), T[bo, i, by])  # fixed case 38404
ck(f(('bool', 'int240', 'bytes31'), D), T[bo, i, by])  # fixed case 38405
ck(f(('bool', 'int240', 'bytes32'), D), T[bo, i, by])  # fixed case 38406
ck(f(('bool', 'int240', 'bytes4'), D), T[bo, i, by])  # fixed case 38407
ck(f(('bool', 'int240', 'bytes5'), D), T[bo, i, by])  # fixed case 38408
ck(f(('bool', 'int240', 'bytes6'), D), T[bo, i, by])  # fixed case 38409
ck(f(('bool', 'int240', 'bytes7'), D), T[bo, i, by])  # fixed case 38410
ck(f(('bool', 'int240', 'bytes8'), D), T[bo, i, by])  # fixed case 38411
ck(f(('bool', 'int240', 'bytes9'), D), T[bo, i, by])  # fixed case 38412
ck(f(('bool', 'int240', 'int104'), D), T[bo, i, i])  # fixed case 38413
ck(f(('bool', 'int240', 'int112'), D), T[bo, i, i])  # fixed case 38414
ck(f(('bool', 'int240', 'int120'), D), T[bo, i, i])  # fixed case 38415
ck(f(('bool', 'int240', 'int128'), D), T[bo, i, i])  # fixed case 38416
ck(f(('bool', 'int240', 'int136'), D), T[bo, i, i])  # fixed case 38417
ck(f(('bool', 'int240', 'int144'), D), T[bo, i, i])  # fixed case 38418
ck(f(('bool', 'int240', 'int152'), D), T[bo, i, i])  # fixed case 38419
ck(f(('bool', 'int240', 'int16'), D), T[bo, i, i])  # fixed case 38420
ck(f(('bool', 'int240', 'int160'), D), T[bo, i, i])  # fixed case 38421
ck(f(('bool', 'int240', 'int168'), D), T[bo, i, i])  # fixed case 38422
ck(f(('bool', 'int240', 'int176'), D), T[bo, i, i])  # fixed case 38423
ck(f(('bool', 'int240', 'int184'), D), T[bo, i, i])  # fixed case 38424
ck(f(('bool', 'int240', 'int192'), D), T[bo, i, i])  # fixed case 38425
ck(f(('bool', 'int240', 'int200'), D), T[bo, i, i])  # fixed case 38426
ck(f(('bool', 'int240', 'int208'), D), T[bo, i, i])  # fixed case 38427
ck(f(('bool', 'int240', 'int216'), D), T[bo, i, i])  # fixed case 38428
ck(f(('bool', 'int240', 'int224'), D), T[bo, i, i])  # fixed case 38429
ck(f(('bool', 'int240', 'int232'), D), T[bo, i, i])  # fixed case 38430
ck(f(('bool', 'int240', 'int24'), D), T[bo, i, i])  # fixed case 38431
ck(f(('bool', 'int240', 'int240'), D), T[bo, i, i])  # fixed case 38432
ck(f(('bool', 'int240', 'int248'), D), T[bo, i, i])  # fixed case 38433
ck(f(('bool', 'int240', 'int256'), D), T[bo, i, i])  # fixed case 38434
ck(f(('bool', 'int240', 'int32'), D), T[bo, i, i])  # fixed case 38435
ck(f(('bool', 'int240', 'int40'), D), T[bo, i, i])  # fixed case 38436
ck(f(('bool', 'int240', 'int48'), D), T[bo, i, i])  # fixed case 38437
ck(f(('bool', 'int240', 'int56'), D), T[bo, i, i])  # fixed case 38438
ck(f(('bool', 'int240', 'int64'), D), T[bo, i, i])  # fixed case 38439
ck(f(('bool', 'int240', 'int72'), D), T[bo, i, i])  # fixed case 38440
ck(f(('bool', 'int240', 'int8'), D), T[bo, i, i])  # fixed case 38441
ck(f(('bool', 'int240', 'int80'), D), T[bo, i, i])  # fixed case 38442
ck(f(('bool', 'int240', 'int88'), D), T[bo, i, i])  # fixed case 38443
ck(f(('bool', 'int240', 'int96'), D), T[bo, i, i])  # fixed case 38444
ck(f(('bool', 'int240', 'string'), D), T[bo, i, s])  # fixed case 38445
ck(f(('bool', 'int240', 'string[]'), D), T[bo, i, A])  # fixed case 38446
ck(f(('bool', 'int240', 'uint104'), D), T[bo, i, i])  # fixed case 38447
ck(f(('bool', 'int240', 'uint112'), D), T[bo, i, i])  # fixed case 38448
ck(f(('bool', 'int240', 'uint120'), D), T[bo, i, i])  # fixed case 38449
ck(f(('bool', 'int240', 'uint128'), D), T[bo, i, i])  # fixed case 38450
ck(f(('bool', 'int240', 'uint136'), D), T[bo, i, i])  # fixed case 38451
ck(f(('bool', 'int240', 'uint144'), D), T[bo, i, i])  # fixed case 38452
ck(f(('bool', 'int240', 'uint152'), D), T[bo, i, i])  # fixed case 38453
ck(f(('bool', 'int240', 'uint16'), D), T[bo, i, i])  # fixed case 38454
ck(f(('bool', 'int240', 'uint160'), D), T[bo, i, i])  # fixed case 38455
ck(f(('bool', 'int240', 'uint168'), D), T[bo, i, i])  # fixed case 38456
ck(f(('bool', 'int240', 'uint176'), D), T[bo, i, i])  # fixed case 38457
ck(f(('bool', 'int240', 'uint184'), D), T[bo, i, i])  # fixed case 38458
ck(f(('bool', 'int240', 'uint192'), D), T[bo, i, i])  # fixed case 38459
ck(f(('bool', 'int240', 'uint200'), D), T[bo, i, i])  # fixed case 38460
ck(f(('bool', 'int240', 'uint208'), D), T[bo, i, i])  # fixed case 38461
ck(f(('bool', 'int240', 'uint216'), D), T[bo, i, i])  # fixed case 38462
ck(f(('bool', 'int240', 'uint224'), D), T[bo, i, i])  # fixed case 38463
ck(f(('bool', 'int240', 'uint232'), D), T[bo, i, i])  # fixed case 38464
ck(f(('bool', 'int240', 'uint24'), D), T[bo, i, i])  # fixed case 38465
ck(f(('bool', 'int240', 'uint240'), D), T[bo, i, i])  # fixed case 38466
ck(f(('bool', 'int240', 'uint248'), D), T[bo, i, i])  # fixed case 38467
ck(f(('bool', 'int240', 'uint256'), D), T[bo, i, i])  # fixed case 38468
ck(f(('bool', 'int240', 'uint32'), D), T[bo, i, i])  # fixed case 38469
ck(f(('bool', 'int240', 'uint40'), D), T[bo, i, i])  # fixed case 38470
ck(f(('bool', 'int240', 'uint48'), D), T[bo, i, i])  # fixed case 38471
ck(f(('bool', 'int240', 'uint56'), D), T[bo, i, i])  # fixed case 38472
ck(f(('bool', 'int240', 'uint64'), D), T[bo, i, i])  # fixed case 38473
ck(f(('bool', 'int240', 'uint72'), D), T[bo, i, i])  # fixed case 38474
ck(f(('bool', 'int240', 'uint8'), D), T[bo, i, i])  # fixed case 38475
ck(f(('bool', 'int240', 'uint80'), D), T[bo, i, i])  # fixed case 38476
ck(f(('bool', 'int240', 'uint88'), D), T[bo, i, i])  # fixed case 38477
ck(f(('bool', 'int240', 'uint96'), D), T[bo, i, i])  # fixed case 38478
ck(f(('bool', 'int240', '?'), D), T[bo, i, A])  # fixed case 38479
ck(f(('bool', 'int248', 'address'), D), T[bo, i, HA])  # fixed case 38480
ck(f(('bool', 'int248', 'address[]'), D), T[bo, i, A])  # fixed case 38481
ck(f(('bool', 'int248', 'bool'), D), T[bo, i, bo])  # fixed case 38482
ck(f(('bool', 'int248', 'bool[]'), D), T[bo, i, A])  # fixed case 38483
ck(f(('bool', 'int248', 'bytes'), D), T[bo, i, by])  # fixed case 38484
ck(f(('bool', 'int248', 'bytes1'), D), T[bo, i, by])  # fixed case 38485
ck(f(('bool', 'int248', 'bytes10'), D), T[bo, i, by])  # fixed case 38486
ck(f(('bool', 'int248', 'bytes11'), D), T[bo, i, by])  # fixed case 38487
ck(f(('bool', 'int248', 'bytes12'), D), T[bo, i, by])  # fixed case 38488
ck(f(('bool', 'int248', 'bytes13'), D), T[bo, i, by])  # fixed case 38489
ck(f(('bool', 'int248', 'bytes14'), D), T[bo, i, by])  # fixed case 38490
ck(f(('bool', 'int248', 'bytes15'), D), T[bo, i, by])  # fixed case 38491
ck(f(('bool', 'int248', 'bytes16'), D), T[bo, i, by])  # fixed case 38492
ck(f(('bool', 'int248', 'bytes17'), D), T[bo, i, by])  # fixed case 38493
ck(f(('bool', 'int248', 'bytes18'), D), T[bo, i, by])  # fixed case 38494
ck(f(('bool', 'int248', 'bytes19'), D), T[bo, i, by])  # fixed case 38495
ck(f(('bool', 'int248', 'bytes2'), D), T[bo, i, by])  # fixed case 38496
ck(f(('bool', 'int248', 'bytes20'), D), T[bo, i, by])  # fixed case 38497
ck(f(('bool', 'int248', 'bytes21'), D), T[bo, i, by])  # fixed case 38498
ck(f(('bool', 'int248', 'bytes22'), D), T[bo, i, by])  # fixed case 38499
ck(f(('bool', 'int248', 'bytes23'), D), T[bo, i, by])  # fixed case 38500
ck(f(('bool', 'int248', 'bytes24'), D), T[bo, i, by])  # fixed case 38501
ck(f(('bool', 'int248', 'bytes25'), D), T[bo, i, by])  # fixed case 38502
ck(f(('bool', 'int248', 'bytes26'), D), T[bo, i, by])  # fixed case 38503
ck(f(('bool', 'int248', 'bytes27'), D), T[bo, i, by])  # fixed case 38504
ck(f(('bool', 'int248', 'bytes28'), D), T[bo, i, by])  # fixed case 38505
ck(f(('bool', 'int248', 'bytes29'), D), T[bo, i, by])  # fixed case 38506
ck(f(('bool', 'int248', 'bytes3'), D), T[bo, i, by])  # fixed case 38507
ck(f(('bool', 'int248', 'bytes30'), D), T[bo, i, by])  # fixed case 38508
ck(f(('bool', 'int248', 'bytes31'), D), T[bo, i, by])  # fixed case 38509
ck(f(('bool', 'int248', 'bytes32'), D), T[bo, i, by])  # fixed case 38510
ck(f(('bool', 'int248', 'bytes4'), D), T[bo, i, by])  # fixed case 38511
ck(f(('bool', 'int248', 'bytes5'), D), T[bo, i, by])  # fixed case 38512
ck(f(('bool', 'int248', 'bytes6'), D), T[bo, i, by])  # fixed case 38513
ck(f(('bool', 'int248', 'bytes7'), D), T[bo, i, by])  # fixed case 38514
ck(f(('bool', 'int248', 'bytes8'), D), T[bo, i, by])  # fixed case 38515
ck(f(('bool', 'int248', 'bytes9'), D), T[bo, i, by])  # fixed case 38516
ck(f(('bool', 'int248', 'int104'), D), T[bo, i, i])  # fixed case 38517
ck(f(('bool', 'int248', 'int112'), D), T[bo, i, i])  # fixed case 38518
ck(f(('bool', 'int248', 'int120'), D), T[bo, i, i])  # fixed case 38519
ck(f(('bool', 'int248', 'int128'), D), T[bo, i, i])  # fixed case 38520
ck(f(('bool', 'int248', 'int136'), D), T[bo, i, i])  # fixed case 38521
ck(f(('bool', 'int248', 'int144'), D), T[bo, i, i])  # fixed case 38522
ck(f(('bool', 'int248', 'int152'), D), T[bo, i, i])  # fixed case 38523
ck(f(('bool', 'int248', 'int16'), D), T[bo, i, i])  # fixed case 38524
ck(f(('bool', 'int248', 'int160'), D), T[bo, i, i])  # fixed case 38525
ck(f(('bool', 'int248', 'int168'), D), T[bo, i, i])  # fixed case 38526
ck(f(('bool', 'int248', 'int176'), D), T[bo, i, i])  # fixed case 38527
ck(f(('bool', 'int248', 'int184'), D), T[bo, i, i])  # fixed case 38528
ck(f(('bool', 'int248', 'int192'), D), T[bo, i, i])  # fixed case 38529
ck(f(('bool', 'int248', 'int200'), D), T[bo, i, i])  # fixed case 38530
ck(f(('bool', 'int248', 'int208'), D), T[bo, i, i])  # fixed case 38531
ck(f(('bool', 'int248', 'int216'), D), T[bo, i, i])  # fixed case 38532
ck(f(('bool', 'int248', 'int224'), D), T[bo, i, i])  # fixed case 38533
ck(f(('bool', 'int248', 'int232'), D), T[bo, i, i])  # fixed case 38534
ck(f(('bool', 'int248', 'int24'), D), T[bo, i, i])  # fixed case 38535
ck(f(('bool', 'int248', 'int240'), D), T[bo, i, i])  # fixed case 38536
ck(f(('bool', 'int248', 'int248'), D), T[bo, i, i])  # fixed case 38537
ck(f(('bool', 'int248', 'int256'), D), T[bo, i, i])  # fixed case 38538
ck(f(('bool', 'int248', 'int32'), D), T[bo, i, i])  # fixed case 38539
ck(f(('bool', 'int248', 'int40'), D), T[bo, i, i])  # fixed case 38540
ck(f(('bool', 'int248', 'int48'), D), T[bo, i, i])  # fixed case 38541
ck(f(('bool', 'int248', 'int56'), D), T[bo, i, i])  # fixed case 38542
ck(f(('bool', 'int248', 'int64'), D), T[bo, i, i])  # fixed case 38543
ck(f(('bool', 'int248', 'int72'), D), T[bo, i, i])  # fixed case 38544
ck(f(('bool', 'int248', 'int8'), D), T[bo, i, i])  # fixed case 38545
ck(f(('bool', 'int248', 'int80'), D), T[bo, i, i])  # fixed case 38546
ck(f(('bool', 'int248', 'int88'), D), T[bo, i, i])  # fixed case 38547
ck(f(('bool', 'int248', 'int96'), D), T[bo, i, i])  # fixed case 38548
ck(f(('bool', 'int248', 'string'), D), T[bo, i, s])  # fixed case 38549
ck(f(('bool', 'int248', 'string[]'), D), T[bo, i, A])  # fixed case 38550
ck(f(('bool', 'int248', 'uint104'), D), T[bo, i, i])  # fixed case 38551
ck(f(('bool', 'int248', 'uint112'), D), T[bo, i, i])  # fixed case 38552
ck(f(('bool', 'int248', 'uint120'), D), T[bo, i, i])  # fixed case 38553
ck(f(('bool', 'int248', 'uint128'), D), T[bo, i, i])  # fixed case 38554
ck(f(('bool', 'int248', 'uint136'), D), T[bo, i, i])  # fixed case 38555
ck(f(('bool', 'int248', 'uint144'), D), T[bo, i, i])  # fixed case 38556
ck(f(('bool', 'int248', 'uint152'), D), T[bo, i, i])  # fixed case 38557
ck(f(('bool', 'int248', 'uint16'), D), T[bo, i, i])  # fixed case 38558
ck(f(('bool', 'int248', 'uint160'), D), T[bo, i, i])  # fixed case 38559
ck(f(('bool', 'int248', 'uint168'), D), T[bo, i, i])  # fixed case 38560
ck(f(('bool', 'int248', 'uint176'), D), T[bo, i, i])  # fixed case 38561
ck(f(('bool', 'int248', 'uint184'), D), T[bo, i, i])  # fixed case 38562
ck(f(('bool', 'int248', 'uint192'), D), T[bo, i, i])  # fixed case 38563
ck(f(('bool', 'int248', 'uint200'), D), T[bo, i, i])  # fixed case 38564
ck(f(('bool', 'int248', 'uint208'), D), T[bo, i, i])  # fixed case 38565
ck(f(('bool', 'int248', 'uint216'), D), T[bo, i, i])  # fixed case 38566
ck(f(('bool', 'int248', 'uint224'), D), T[bo, i, i])  # fixed case 38567
ck(f(('bool', 'int248', 'uint232'), D), T[bo, i, i])  # fixed case 38568
ck(f(('bool', 'int248', 'uint24'), D), T[bo, i, i])  # fixed case 38569
ck(f(('bool', 'int248', 'uint240'), D), T[bo, i, i])  # fixed case 38570
ck(f(('bool', 'int248', 'uint248'), D), T[bo, i, i])  # fixed case 38571
ck(f(('bool', 'int248', 'uint256'), D), T[bo, i, i])  # fixed case 38572
ck(f(('bool', 'int248', 'uint32'), D), T[bo, i, i])  # fixed case 38573
ck(f(('bool', 'int248', 'uint40'), D), T[bo, i, i])  # fixed case 38574
ck(f(('bool', 'int248', 'uint48'), D), T[bo, i, i])  # fixed case 38575
ck(f(('bool', 'int248', 'uint56'), D), T[bo, i, i])  # fixed case 38576
ck(f(('bool', 'int248', 'uint64'), D), T[bo, i, i])  # fixed case 38577
ck(f(('bool', 'int248', 'uint72'), D), T[bo, i, i])  # fixed case 38578
ck(f(('bool', 'int248', 'uint8'), D), T[bo, i, i])  # fixed case 38579
ck(f(('bool', 'int248', 'uint80'), D), T[bo, i, i])  # fixed case 38580
ck(f(('bool', 'int248', 'uint88'), D), T[bo, i, i])  # fixed case 38581
ck(f(('bool', 'int248', 'uint96'), D), T[bo, i, i])  # fixed case 38582
ck(f(('bool', 'int248', '?'), D), T[bo, i, A])  # fixed case 38583
ck(f(('bool', 'int256', 'address'), D), T[bo, i, HA])  # fixed case 38584
ck(f(('bool', 'int256', 'address[]'), D), T[bo, i, A])  # fixed case 38585
ck(f(('bool', 'int256', 'bool'), D), T[bo, i, bo])  # fixed case 38586
ck(f(('bool', 'int256', 'bool[]'), D), T[bo, i, A])  # fixed case 38587
ck(f(('bool', 'int256', 'bytes'), D), T[bo, i, by])  # fixed case 38588
ck(f(('bool', 'int256', 'bytes1'), D), T[bo, i, by])  # fixed case 38589
ck(f(('bool', 'int256', 'bytes10'), D), T[bo, i, by])  # fixed case 38590
ck(f(('bool', 'int256', 'bytes11'), D), T[bo, i, by])  # fixed case 38591
ck(f(('bool', 'int256', 'bytes12'), D), T[bo, i, by])  # fixed case 38592
ck(f(('bool', 'int256', 'bytes13'), D), T[bo, i, by])  # fixed case 38593
ck(f(('bool', 'int256', 'bytes14'), D), T[bo, i, by])  # fixed case 38594
ck(f(('bool', 'int256', 'bytes15'), D), T[bo, i, by])  # fixed case 38595
ck(f(('bool', 'int256', 'bytes16'), D), T[bo, i, by])  # fixed case 38596
ck(f(('bool', 'int256', 'bytes17'), D), T[bo, i, by])  # fixed case 38597
ck(f(('bool', 'int256', 'bytes18'), D), T[bo, i, by])  # fixed case 38598
ck(f(('bool', 'int256', 'bytes19'), D), T[bo, i, by])  # fixed case 38599
ck(f(('bool', 'int256', 'bytes2'), D), T[bo, i, by])  # fixed case 38600
ck(f(('bool', 'int256', 'bytes20'), D), T[bo, i, by])  # fixed case 38601
ck(f(('bool', 'int256', 'bytes21'), D), T[bo, i, by])  # fixed case 38602
ck(f(('bool', 'int256', 'bytes22'), D), T[bo, i, by])  # fixed case 38603
ck(f(('bool', 'int256', 'bytes23'), D), T[bo, i, by])  # fixed case 38604
ck(f(('bool', 'int256', 'bytes24'), D), T[bo, i, by])  # fixed case 38605
ck(f(('bool', 'int256', 'bytes25'), D), T[bo, i, by])  # fixed case 38606
ck(f(('bool', 'int256', 'bytes26'), D), T[bo, i, by])  # fixed case 38607
ck(f(('bool', 'int256', 'bytes27'), D), T[bo, i, by])  # fixed case 38608
ck(f(('bool', 'int256', 'bytes28'), D), T[bo, i, by])  # fixed case 38609
ck(f(('bool', 'int256', 'bytes29'), D), T[bo, i, by])  # fixed case 38610
ck(f(('bool', 'int256', 'bytes3'), D), T[bo, i, by])  # fixed case 38611
ck(f(('bool', 'int256', 'bytes30'), D), T[bo, i, by])  # fixed case 38612
ck(f(('bool', 'int256', 'bytes31'), D), T[bo, i, by])  # fixed case 38613
ck(f(('bool', 'int256', 'bytes32'), D), T[bo, i, by])  # fixed case 38614
ck(f(('bool', 'int256', 'bytes4'), D), T[bo, i, by])  # fixed case 38615
ck(f(('bool', 'int256', 'bytes5'), D), T[bo, i, by])  # fixed case 38616
ck(f(('bool', 'int256', 'bytes6'), D), T[bo, i, by])  # fixed case 38617
ck(f(('bool', 'int256', 'bytes7'), D), T[bo, i, by])  # fixed case 38618
ck(f(('bool', 'int256', 'bytes8'), D), T[bo, i, by])  # fixed case 38619
ck(f(('bool', 'int256', 'bytes9'), D), T[bo, i, by])  # fixed case 38620
ck(f(('bool', 'int256', 'int104'), D), T[bo, i, i])  # fixed case 38621
ck(f(('bool', 'int256', 'int112'), D), T[bo, i, i])  # fixed case 38622
ck(f(('bool', 'int256', 'int120'), D), T[bo, i, i])  # fixed case 38623
ck(f(('bool', 'int256', 'int128'), D), T[bo, i, i])  # fixed case 38624
ck(f(('bool', 'int256', 'int136'), D), T[bo, i, i])  # fixed case 38625
ck(f(('bool', 'int256', 'int144'), D), T[bo, i, i])  # fixed case 38626
ck(f(('bool', 'int256', 'int152'), D), T[bo, i, i])  # fixed case 38627
ck(f(('bool', 'int256', 'int16'), D), T[bo, i, i])  # fixed case 38628
ck(f(('bool', 'int256', 'int160'), D), T[bo, i, i])  # fixed case 38629
ck(f(('bool', 'int256', 'int168'), D), T[bo, i, i])  # fixed case 38630
ck(f(('bool', 'int256', 'int176'), D), T[bo, i, i])  # fixed case 38631
ck(f(('bool', 'int256', 'int184'), D), T[bo, i, i])  # fixed case 38632
ck(f(('bool', 'int256', 'int192'), D), T[bo, i, i])  # fixed case 38633
ck(f(('bool', 'int256', 'int200'), D), T[bo, i, i])  # fixed case 38634
ck(f(('bool', 'int256', 'int208'), D), T[bo, i, i])  # fixed case 38635
ck(f(('bool', 'int256', 'int216'), D), T[bo, i, i])  # fixed case 38636
ck(f(('bool', 'int256', 'int224'), D), T[bo, i, i])  # fixed case 38637
ck(f(('bool', 'int256', 'int232'), D), T[bo, i, i])  # fixed case 38638
ck(f(('bool', 'int256', 'int24'), D), T[bo, i, i])  # fixed case 38639
ck(f(('bool', 'int256', 'int240'), D), T[bo, i, i])  # fixed case 38640
ck(f(('bool', 'int256', 'int248'), D), T[bo, i, i])  # fixed case 38641
ck(f(('bool', 'int256', 'int256'), D), T[bo, i, i])  # fixed case 38642
ck(f(('bool', 'int256', 'int32'), D), T[bo, i, i])  # fixed case 38643
ck(f(('bool', 'int256', 'int40'), D), T[bo, i, i])  # fixed case 38644
ck(f(('bool', 'int256', 'int48'), D), T[bo, i, i])  # fixed case 38645
ck(f(('bool', 'int256', 'int56'), D), T[bo, i, i])  # fixed case 38646
ck(f(('bool', 'int256', 'int64'), D), T[bo, i, i])  # fixed case 38647
ck(f(('bool', 'int256', 'int72'), D), T[bo, i, i])  # fixed case 38648
ck(f(('bool', 'int256', 'int8'), D), T[bo, i, i])  # fixed case 38649
ck(f(('bool', 'int256', 'int80'), D), T[bo, i, i])  # fixed case 38650
ck(f(('bool', 'int256', 'int88'), D), T[bo, i, i])  # fixed case 38651
ck(f(('bool', 'int256', 'int96'), D), T[bo, i, i])  # fixed case 38652
ck(f(('bool', 'int256', 'string'), D), T[bo, i, s])  # fixed case 38653
ck(f(('bool', 'int256', 'string[]'), D), T[bo, i, A])  # fixed case 38654
ck(f(('bool', 'int256', 'uint104'), D), T[bo, i, i])  # fixed case 38655
ck(f(('bool', 'int256', 'uint112'), D), T[bo, i, i])  # fixed case 38656
ck(f(('bool', 'int256', 'uint120'), D), T[bo, i, i])  # fixed case 38657
ck(f(('bool', 'int256', 'uint128'), D), T[bo, i, i])  # fixed case 38658
ck(f(('bool', 'int256', 'uint136'), D), T[bo, i, i])  # fixed case 38659
ck(f(('bool', 'int256', 'uint144'), D), T[bo, i, i])  # fixed case 38660
ck(f(('bool', 'int256', 'uint152'), D), T[bo, i, i])  # fixed case 38661
ck(f(('bool', 'int256', 'uint16'), D), T[bo, i, i])  # fixed case 38662
ck(f(('bool', 'int256', 'uint160'), D), T[bo, i, i])  # fixed case 38663
ck(f(('bool', 'int256', 'uint168'), D), T[bo, i, i])  # fixed case 38664
ck(f(('bool', 'int256', 'uint176'), D), T[bo, i, i])  # fixed case 38665
ck(f(('bool', 'int256', 'uint184'), D), T[bo, i, i])  # fixed case 38666
ck(f(('bool', 'int256', 'uint192'), D), T[bo, i, i])  # fixed case 38667
ck(f(('bool', 'int256', 'uint200'), D), T[bo, i, i])  # fixed case 38668
ck(f(('bool', 'int256', 'uint208'), D), T[bo, i, i])  # fixed case 38669
ck(f(('bool', 'int256', 'uint216'), D), T[bo, i, i])  # fixed case 38670
ck(f(('bool', 'int256', 'uint224'), D), T[bo, i, i])  # fixed case 38671
ck(f(('bool', 'int256', 'uint232'), D), T[bo, i, i])  # fixed case 38672
ck(f(('bool', 'int256', 'uint24'), D), T[bo, i, i])  # fixed case 38673
ck(f(('bool', 'int256', 'uint240'), D), T[bo, i, i])  # fixed case 38674
ck(f(('bool', 'int256', 'uint248'), D), T[bo, i, i])  # fixed case 38675
ck(f(('bool', 'int256', 'uint256'), D), T[bo, i, i])  # fixed case 38676
ck(f(('bool', 'int256', 'uint32'), D), T[bo, i, i])  # fixed case 38677
ck(f(('bool', 'int256', 'uint40'), D), T[bo, i, i])  # fixed case 38678
ck(f(('bool', 'int256', 'uint48'), D), T[bo, i, i])  # fixed case 38679
ck(f(('bool', 'int256', 'uint56'), D), T[bo, i, i])  # fixed case 38680
ck(f(('bool', 'int256', 'uint64'), D), T[bo, i, i])  # fixed case 38681
ck(f(('bool', 'int256', 'uint72'), D), T[bo, i, i])  # fixed case 38682
ck(f(('bool', 'int256', 'uint8'), D), T[bo, i, i])  # fixed case 38683
ck(f(('bool', 'int256', 'uint80'), D), T[bo, i, i])  # fixed case 38684
ck(f(('bool', 'int256', 'uint88'), D), T[bo, i, i])  # fixed case 38685
ck(f(('bool', 'int256', 'uint96'), D), T[bo, i, i])  # fixed case 38686
ck(f(('bool', 'int256', '?'), D), T[bo, i, A])  # fixed case 38687
ck(f(('bool', 'int32', 'address'), D), T[bo, i, HA])  # fixed case 38688
ck(f(('bool', 'int32', 'address[]'), D), T[bo, i, A])  # fixed case 38689
ck(f(('bool', 'int32', 'bool'), D), T[bo, i, bo])  # fixed case 38690
ck(f(('bool', 'int32', 'bool[]'), D), T[bo, i, A])  # fixed case 38691
ck(f(('bool', 'int32', 'bytes'), D), T[bo, i, by])  # fixed case 38692
ck(f(('bool', 'int32', 'bytes1'), D), T[bo, i, by])  # fixed case 38693
ck(f(('bool', 'int32', 'bytes10'), D), T[bo, i, by])  # fixed case 38694
ck(f(('bool', 'int32', 'bytes11'), D), T[bo, i, by])  # fixed case 38695
ck(f(('bool', 'int32', 'bytes12'), D), T[bo, i, by])  # fixed case 38696
ck(f(('bool', 'int32', 'bytes13'), D), T[bo, i, by])  # fixed case 38697
ck(f(('bool', 'int32', 'bytes14'), D), T[bo, i, by])  # fixed case 38698
ck(f(('bool', 'int32', 'bytes15'), D), T[bo, i, by])  # fixed case 38699
ck(f(('bool', 'int32', 'bytes16'), D), T[bo, i, by])  # fixed case 38700
ck(f(('bool', 'int32', 'bytes17'), D), T[bo, i, by])  # fixed case 38701
ck(f(('bool', 'int32', 'bytes18'), D), T[bo, i, by])  # fixed case 38702
ck(f(('bool', 'int32', 'bytes19'), D), T[bo, i, by])  # fixed case 38703
ck(f(('bool', 'int32', 'bytes2'), D), T[bo, i, by])  # fixed case 38704
ck(f(('bool', 'int32', 'bytes20'), D), T[bo, i, by])  # fixed case 38705
ck(f(('bool', 'int32', 'bytes21'), D), T[bo, i, by])  # fixed case 38706
ck(f(('bool', 'int32', 'bytes22'), D), T[bo, i, by])  # fixed case 38707
ck(f(('bool', 'int32', 'bytes23'), D), T[bo, i, by])  # fixed case 38708
ck(f(('bool', 'int32', 'bytes24'), D), T[bo, i, by])  # fixed case 38709
ck(f(('bool', 'int32', 'bytes25'), D), T[bo, i, by])  # fixed case 38710
ck(f(('bool', 'int32', 'bytes26'), D), T[bo, i, by])  # fixed case 38711
ck(f(('bool', 'int32', 'bytes27'), D), T[bo, i, by])  # fixed case 38712
ck(f(('bool', 'int32', 'bytes28'), D), T[bo, i, by])  # fixed case 38713
ck(f(('bool', 'int32', 'bytes29'), D), T[bo, i, by])  # fixed case 38714
ck(f(('bool', 'int32', 'bytes3'), D), T[bo, i, by])  # fixed case 38715
ck(f(('bool', 'int32', 'bytes30'), D), T[bo, i, by])  # fixed case 38716
ck(f(('bool', 'int32', 'bytes31'), D), T[bo, i, by])  # fixed case 38717
ck(f(('bool', 'int32', 'bytes32'), D), T[bo, i, by])  # fixed case 38718
ck(f(('bool', 'int32', 'bytes4'), D), T[bo, i, by])  # fixed case 38719
ck(f(('bool', 'int32', 'bytes5'), D), T[bo, i, by])  # fixed case 38720
ck(f(('bool', 'int32', 'bytes6'), D), T[bo, i, by])  # fixed case 38721
ck(f(('bool', 'int32', 'bytes7'), D), T[bo, i, by])  # fixed case 38722
ck(f(('bool', 'int32', 'bytes8'), D), T[bo, i, by])  # fixed case 38723
ck(f(('bool', 'int32', 'bytes9'), D), T[bo, i, by])  # fixed case 38724
ck(f(('bool', 'int32', 'int104'), D), T[bo, i, i])  # fixed case 38725
ck(f(('bool', 'int32', 'int112'), D), T[bo, i, i])  # fixed case 38726
ck(f(('bool', 'int32', 'int120'), D), T[bo, i, i])  # fixed case 38727
ck(f(('bool', 'int32', 'int128'), D), T[bo, i, i])  # fixed case 38728
ck(f(('bool', 'int32', 'int136'), D), T[bo, i, i])  # fixed case 38729
ck(f(('bool', 'int32', 'int144'), D), T[bo, i, i])  # fixed case 38730
ck(f(('bool', 'int32', 'int152'), D), T[bo, i, i])  # fixed case 38731
ck(f(('bool', 'int32', 'int16'), D), T[bo, i, i])  # fixed case 38732
ck(f(('bool', 'int32', 'int160'), D), T[bo, i, i])  # fixed case 38733
ck(f(('bool', 'int32', 'int168'), D), T[bo, i, i])  # fixed case 38734
ck(f(('bool', 'int32', 'int176'), D), T[bo, i, i])  # fixed case 38735
ck(f(('bool', 'int32', 'int184'), D), T[bo, i, i])  # fixed case 38736
ck(f(('bool', 'int32', 'int192'), D), T[bo, i, i])  # fixed case 38737
ck(f(('bool', 'int32', 'int200'), D), T[bo, i, i])  # fixed case 38738
ck(f(('bool', 'int32', 'int208'), D), T[bo, i, i])  # fixed case 38739
ck(f(('bool', 'int32', 'int216'), D), T[bo, i, i])  # fixed case 38740
ck(f(('bool', 'int32', 'int224'), D), T[bo, i, i])  # fixed case 38741
ck(f(('bool', 'int32', 'int232'), D), T[bo, i, i])  # fixed case 38742
ck(f(('bool', 'int32', 'int24'), D), T[bo, i, i])  # fixed case 38743
ck(f(('bool', 'int32', 'int240'), D), T[bo, i, i])  # fixed case 38744
ck(f(('bool', 'int32', 'int248'), D), T[bo, i, i])  # fixed case 38745
ck(f(('bool', 'int32', 'int256'), D), T[bo, i, i])  # fixed case 38746
ck(f(('bool', 'int32', 'int32'), D), T[bo, i, i])  # fixed case 38747
ck(f(('bool', 'int32', 'int40'), D), T[bo, i, i])  # fixed case 38748
ck(f(('bool', 'int32', 'int48'), D), T[bo, i, i])  # fixed case 38749
ck(f(('bool', 'int32', 'int56'), D), T[bo, i, i])  # fixed case 38750
ck(f(('bool', 'int32', 'int64'), D), T[bo, i, i])  # fixed case 38751
ck(f(('bool', 'int32', 'int72'), D), T[bo, i, i])  # fixed case 38752
ck(f(('bool', 'int32', 'int8'), D), T[bo, i, i])  # fixed case 38753
ck(f(('bool', 'int32', 'int80'), D), T[bo, i, i])  # fixed case 38754
ck(f(('bool', 'int32', 'int88'), D), T[bo, i, i])  # fixed case 38755
ck(f(('bool', 'int32', 'int96'), D), T[bo, i, i])  # fixed case 38756
ck(f(('bool', 'int32', 'string'), D), T[bo, i, s])  # fixed case 38757
ck(f(('bool', 'int32', 'string[]'), D), T[bo, i, A])  # fixed case 38758
ck(f(('bool', 'int32', 'uint104'), D), T[bo, i, i])  # fixed case 38759
ck(f(('bool', 'int32', 'uint112'), D), T[bo, i, i])  # fixed case 38760
ck(f(('bool', 'int32', 'uint120'), D), T[bo, i, i])  # fixed case 38761
ck(f(('bool', 'int32', 'uint128'), D), T[bo, i, i])  # fixed case 38762
ck(f(('bool', 'int32', 'uint136'), D), T[bo, i, i])  # fixed case 38763
ck(f(('bool', 'int32', 'uint144'), D), T[bo, i, i])  # fixed case 38764
ck(f(('bool', 'int32', 'uint152'), D), T[bo, i, i])  # fixed case 38765
ck(f(('bool', 'int32', 'uint16'), D), T[bo, i, i])  # fixed case 38766
ck(f(('bool', 'int32', 'uint160'), D), T[bo, i, i])  # fixed case 38767
ck(f(('bool', 'int32', 'uint168'), D), T[bo, i, i])  # fixed case 38768
ck(f(('bool', 'int32', 'uint176'), D), T[bo, i, i])  # fixed case 38769
ck(f(('bool', 'int32', 'uint184'), D), T[bo, i, i])  # fixed case 38770
ck(f(('bool', 'int32', 'uint192'), D), T[bo, i, i])  # fixed case 38771
ck(f(('bool', 'int32', 'uint200'), D), T[bo, i, i])  # fixed case 38772
ck(f(('bool', 'int32', 'uint208'), D), T[bo, i, i])  # fixed case 38773
ck(f(('bool', 'int32', 'uint216'), D), T[bo, i, i])  # fixed case 38774
ck(f(('bool', 'int32', 'uint224'), D), T[bo, i, i])  # fixed case 38775
ck(f(('bool', 'int32', 'uint232'), D), T[bo, i, i])  # fixed case 38776
ck(f(('bool', 'int32', 'uint24'), D), T[bo, i, i])  # fixed case 38777
ck(f(('bool', 'int32', 'uint240'), D), T[bo, i, i])  # fixed case 38778
ck(f(('bool', 'int32', 'uint248'), D), T[bo, i, i])  # fixed case 38779
ck(f(('bool', 'int32', 'uint256'), D), T[bo, i, i])  # fixed case 38780
ck(f(('bool', 'int32', 'uint32'), D), T[bo, i, i])  # fixed case 38781
ck(f(('bool', 'int32', 'uint40'), D), T[bo, i, i])  # fixed case 38782
ck(f(('bool', 'int32', 'uint48'), D), T[bo, i, i])  # fixed case 38783
ck(f(('bool', 'int32', 'uint56'), D), T[bo, i, i])  # fixed case 38784
ck(f(('bool', 'int32', 'uint64'), D), T[bo, i, i])  # fixed case 38785
ck(f(('bool', 'int32', 'uint72'), D), T[bo, i, i])  # fixed case 38786
ck(f(('bool', 'int32', 'uint8'), D), T[bo, i, i])  # fixed case 38787
ck(f(('bool', 'int32', 'uint80'), D), T[bo, i, i])  # fixed case 38788
ck(f(('bool', 'int32', 'uint88'), D), T[bo, i, i])  # fixed case 38789
ck(f(('bool', 'int32', 'uint96'), D), T[bo, i, i])  # fixed case 38790
ck(f(('bool', 'int32', '?'), D), T[bo, i, A])  # fixed case 38791
ck(f(('bool', 'int40', 'address'), D), T[bo, i, HA])  # fixed case 38792
ck(f(('bool', 'int40', 'address[]'), D), T[bo, i, A])  # fixed case 38793
ck(f(('bool', 'int40', 'bool'), D), T[bo, i, bo])  # fixed case 38794
ck(f(('bool', 'int40', 'bool[]'), D), T[bo, i, A])  # fixed case 38795
ck(f(('bool', 'int40', 'bytes'), D), T[bo, i, by])  # fixed case 38796
ck(f(('bool', 'int40', 'bytes1'), D), T[bo, i, by])  # fixed case 38797
ck(f(('bool', 'int40', 'bytes10'), D), T[bo, i, by])  # fixed case 38798
ck(f(('bool', 'int40', 'bytes11'), D), T[bo, i, by])  # fixed case 38799
ck(f(('bool', 'int40', 'bytes12'), D), T[bo, i, by])  # fixed case 38800
ck(f(('bool', 'int40', 'bytes13'), D), T[bo, i, by])  # fixed case 38801
ck(f(('bool', 'int40', 'bytes14'), D), T[bo, i, by])  # fixed case 38802
ck(f(('bool', 'int40', 'bytes15'), D), T[bo, i, by])  # fixed case 38803
ck(f(('bool', 'int40', 'bytes16'), D), T[bo, i, by])  # fixed case 38804
ck(f(('bool', 'int40', 'bytes17'), D), T[bo, i, by])  # fixed case 38805
ck(f(('bool', 'int40', 'bytes18'), D), T[bo, i, by])  # fixed case 38806
ck(f(('bool', 'int40', 'bytes19'), D), T[bo, i, by])  # fixed case 38807
ck(f(('bool', 'int40', 'bytes2'), D), T[bo, i, by])  # fixed case 38808
ck(f(('bool', 'int40', 'bytes20'), D), T[bo, i, by])  # fixed case 38809
ck(f(('bool', 'int40', 'bytes21'), D), T[bo, i, by])  # fixed case 38810
ck(f(('bool', 'int40', 'bytes22'), D), T[bo, i, by])  # fixed case 38811
ck(f(('bool', 'int40', 'bytes23'), D), T[bo, i, by])  # fixed case 38812
ck(f(('bool', 'int40', 'bytes24'), D), T[bo, i, by])  # fixed case 38813
ck(f(('bool', 'int40', 'bytes25'), D), T[bo, i, by])  # fixed case 38814
ck(f(('bool', 'int40', 'bytes26'), D), T[bo, i, by])  # fixed case 38815
ck(f(('bool', 'int40', 'bytes27'), D), T[bo, i, by])  # fixed case 38816
ck(f(('bool', 'int40', 'bytes28'), D), T[bo, i, by])  # fixed case 38817
ck(f(('bool', 'int40', 'bytes29'), D), T[bo, i, by])  # fixed case 38818
ck(f(('bool', 'int40', 'bytes3'), D), T[bo, i, by])  # fixed case 38819
ck(f(('bool', 'int40', 'bytes30'), D), T[bo, i, by])  # fixed case 38820
ck(f(('bool', 'int40', 'bytes31'), D), T[bo, i, by])  # fixed case 38821
ck(f(('bool', 'int40', 'bytes32'), D), T[bo, i, by])  # fixed case 38822
ck(f(('bool', 'int40', 'bytes4'), D), T[bo, i, by])  # fixed case 38823
ck(f(('bool', 'int40', 'bytes5'), D), T[bo, i, by])  # fixed case 38824
ck(f(('bool', 'int40', 'bytes6'), D), T[bo, i, by])  # fixed case 38825
ck(f(('bool', 'int40', 'bytes7'), D), T[bo, i, by])  # fixed case 38826
ck(f(('bool', 'int40', 'bytes8'), D), T[bo, i, by])  # fixed case 38827
ck(f(('bool', 'int40', 'bytes9'), D), T[bo, i, by])  # fixed case 38828
ck(f(('bool', 'int40', 'int104'), D), T[bo, i, i])  # fixed case 38829
ck(f(('bool', 'int40', 'int112'), D), T[bo, i, i])  # fixed case 38830
ck(f(('bool', 'int40', 'int120'), D), T[bo, i, i])  # fixed case 38831
ck(f(('bool', 'int40', 'int128'), D), T[bo, i, i])  # fixed case 38832
ck(f(('bool', 'int40', 'int136'), D), T[bo, i, i])  # fixed case 38833
ck(f(('bool', 'int40', 'int144'), D), T[bo, i, i])  # fixed case 38834
ck(f(('bool', 'int40', 'int152'), D), T[bo, i, i])  # fixed case 38835
ck(f(('bool', 'int40', 'int16'), D), T[bo, i, i])  # fixed case 38836
ck(f(('bool', 'int40', 'int160'), D), T[bo, i, i])  # fixed case 38837
ck(f(('bool', 'int40', 'int168'), D), T[bo, i, i])  # fixed case 38838
ck(f(('bool', 'int40', 'int176'), D), T[bo, i, i])  # fixed case 38839
ck(f(('bool', 'int40', 'int184'), D), T[bo, i, i])  # fixed case 38840
ck(f(('bool', 'int40', 'int192'), D), T[bo, i, i])  # fixed case 38841
ck(f(('bool', 'int40', 'int200'), D), T[bo, i, i])  # fixed case 38842
ck(f(('bool', 'int40', 'int208'), D), T[bo, i, i])  # fixed case 38843
ck(f(('bool', 'int40', 'int216'), D), T[bo, i, i])  # fixed case 38844
ck(f(('bool', 'int40', 'int224'), D), T[bo, i, i])  # fixed case 38845
ck(f(('bool', 'int40', 'int232'), D), T[bo, i, i])  # fixed case 38846
ck(f(('bool', 'int40', 'int24'), D), T[bo, i, i])  # fixed case 38847
ck(f(('bool', 'int40', 'int240'), D), T[bo, i, i])  # fixed case 38848
ck(f(('bool', 'int40', 'int248'), D), T[bo, i, i])  # fixed case 38849
ck(f(('bool', 'int40', 'int256'), D), T[bo, i, i])  # fixed case 38850
ck(f(('bool', 'int40', 'int32'), D), T[bo, i, i])  # fixed case 38851
ck(f(('bool', 'int40', 'int40'), D), T[bo, i, i])  # fixed case 38852
ck(f(('bool', 'int40', 'int48'), D), T[bo, i, i])  # fixed case 38853
ck(f(('bool', 'int40', 'int56'), D), T[bo, i, i])  # fixed case 38854
ck(f(('bool', 'int40', 'int64'), D), T[bo, i, i])  # fixed case 38855
ck(f(('bool', 'int40', 'int72'), D), T[bo, i, i])  # fixed case 38856
ck(f(('bool', 'int40', 'int8'), D), T[bo, i, i])  # fixed case 38857
ck(f(('bool', 'int40', 'int80'), D), T[bo, i, i])  # fixed case 38858
ck(f(('bool', 'int40', 'int88'), D), T[bo, i, i])  # fixed case 38859
ck(f(('bool', 'int40', 'int96'), D), T[bo, i, i])  # fixed case 38860
ck(f(('bool', 'int40', 'string'), D), T[bo, i, s])  # fixed case 38861
ck(f(('bool', 'int40', 'string[]'), D), T[bo, i, A])  # fixed case 38862
ck(f(('bool', 'int40', 'uint104'), D), T[bo, i, i])  # fixed case 38863
ck(f(('bool', 'int40', 'uint112'), D), T[bo, i, i])  # fixed case 38864
ck(f(('bool', 'int40', 'uint120'), D), T[bo, i, i])  # fixed case 38865
ck(f(('bool', 'int40', 'uint128'), D), T[bo, i, i])  # fixed case 38866
ck(f(('bool', 'int40', 'uint136'), D), T[bo, i, i])  # fixed case 38867
ck(f(('bool', 'int40', 'uint144'), D), T[bo, i, i])  # fixed case 38868
ck(f(('bool', 'int40', 'uint152'), D), T[bo, i, i])  # fixed case 38869
ck(f(('bool', 'int40', 'uint16'), D), T[bo, i, i])  # fixed case 38870
ck(f(('bool', 'int40', 'uint160'), D), T[bo, i, i])  # fixed case 38871
ck(f(('bool', 'int40', 'uint168'), D), T[bo, i, i])  # fixed case 38872
ck(f(('bool', 'int40', 'uint176'), D), T[bo, i, i])  # fixed case 38873
ck(f(('bool', 'int40', 'uint184'), D), T[bo, i, i])  # fixed case 38874
ck(f(('bool', 'int40', 'uint192'), D), T[bo, i, i])  # fixed case 38875
ck(f(('bool', 'int40', 'uint200'), D), T[bo, i, i])  # fixed case 38876
ck(f(('bool', 'int40', 'uint208'), D), T[bo, i, i])  # fixed case 38877
ck(f(('bool', 'int40', 'uint216'), D), T[bo, i, i])  # fixed case 38878
ck(f(('bool', 'int40', 'uint224'), D), T[bo, i, i])  # fixed case 38879
ck(f(('bool', 'int40', 'uint232'), D), T[bo, i, i])  # fixed case 38880
ck(f(('bool', 'int40', 'uint24'), D), T[bo, i, i])  # fixed case 38881
ck(f(('bool', 'int40', 'uint240'), D), T[bo, i, i])  # fixed case 38882
ck(f(('bool', 'int40', 'uint248'), D), T[bo, i, i])  # fixed case 38883
ck(f(('bool', 'int40', 'uint256'), D), T[bo, i, i])  # fixed case 38884
ck(f(('bool', 'int40', 'uint32'), D), T[bo, i, i])  # fixed case 38885
ck(f(('bool', 'int40', 'uint40'), D), T[bo, i, i])  # fixed case 38886
ck(f(('bool', 'int40', 'uint48'), D), T[bo, i, i])  # fixed case 38887
ck(f(('bool', 'int40', 'uint56'), D), T[bo, i, i])  # fixed case 38888
ck(f(('bool', 'int40', 'uint64'), D), T[bo, i, i])  # fixed case 38889
ck(f(('bool', 'int40', 'uint72'), D), T[bo, i, i])  # fixed case 38890
ck(f(('bool', 'int40', 'uint8'), D), T[bo, i, i])  # fixed case 38891
ck(f(('bool', 'int40', 'uint80'), D), T[bo, i, i])  # fixed case 38892
ck(f(('bool', 'int40', 'uint88'), D), T[bo, i, i])  # fixed case 38893
ck(f(('bool', 'int40', 'uint96'), D), T[bo, i, i])  # fixed case 38894
ck(f(('bool', 'int40', '?'), D), T[bo, i, A])  # fixed case 38895
ck(f(('bool', 'int48', 'address'), D), T[bo, i, HA])  # fixed case 38896
ck(f(('bool', 'int48', 'address[]'), D), T[bo, i, A])  # fixed case 38897
ck(f(('bool', 'int48', 'bool'), D), T[bo, i, bo])  # fixed case 38898
ck(f(('bool', 'int48', 'bool[]'), D), T[bo, i, A])  # fixed case 38899
ck(f(('bool', 'int48', 'bytes'), D), T[bo, i, by])  # fixed case 38900
ck(f(('bool', 'int48', 'bytes1'), D), T[bo, i, by])  # fixed case 38901
ck(f(('bool', 'int48', 'bytes10'), D), T[bo, i, by])  # fixed case 38902
ck(f(('bool', 'int48', 'bytes11'), D), T[bo, i, by])  # fixed case 38903
ck(f(('bool', 'int48', 'bytes12'), D), T[bo, i, by])  # fixed case 38904
ck(f(('bool', 'int48', 'bytes13'), D), T[bo, i, by])  # fixed case 38905
ck(f(('bool', 'int48', 'bytes14'), D), T[bo, i, by])  # fixed case 38906
ck(f(('bool', 'int48', 'bytes15'), D), T[bo, i, by])  # fixed case 38907
ck(f(('bool', 'int48', 'bytes16'), D), T[bo, i, by])  # fixed case 38908
ck(f(('bool', 'int48', 'bytes17'), D), T[bo, i, by])  # fixed case 38909
ck(f(('bool', 'int48', 'bytes18'), D), T[bo, i, by])  # fixed case 38910
ck(f(('bool', 'int48', 'bytes19'), D), T[bo, i, by])  # fixed case 38911
ck(f(('bool', 'int48', 'bytes2'), D), T[bo, i, by])  # fixed case 38912
ck(f(('bool', 'int48', 'bytes20'), D), T[bo, i, by])  # fixed case 38913
ck(f(('bool', 'int48', 'bytes21'), D), T[bo, i, by])  # fixed case 38914
ck(f(('bool', 'int48', 'bytes22'), D), T[bo, i, by])  # fixed case 38915
ck(f(('bool', 'int48', 'bytes23'), D), T[bo, i, by])  # fixed case 38916
ck(f(('bool', 'int48', 'bytes24'), D), T[bo, i, by])  # fixed case 38917
ck(f(('bool', 'int48', 'bytes25'), D), T[bo, i, by])  # fixed case 38918
ck(f(('bool', 'int48', 'bytes26'), D), T[bo, i, by])  # fixed case 38919
ck(f(('bool', 'int48', 'bytes27'), D), T[bo, i, by])  # fixed case 38920
ck(f(('bool', 'int48', 'bytes28'), D), T[bo, i, by])  # fixed case 38921
ck(f(('bool', 'int48', 'bytes29'), D), T[bo, i, by])  # fixed case 38922
ck(f(('bool', 'int48', 'bytes3'), D), T[bo, i, by])  # fixed case 38923
ck(f(('bool', 'int48', 'bytes30'), D), T[bo, i, by])  # fixed case 38924
ck(f(('bool', 'int48', 'bytes31'), D), T[bo, i, by])  # fixed case 38925
ck(f(('bool', 'int48', 'bytes32'), D), T[bo, i, by])  # fixed case 38926
ck(f(('bool', 'int48', 'bytes4'), D), T[bo, i, by])  # fixed case 38927
ck(f(('bool', 'int48', 'bytes5'), D), T[bo, i, by])  # fixed case 38928
ck(f(('bool', 'int48', 'bytes6'), D), T[bo, i, by])  # fixed case 38929
ck(f(('bool', 'int48', 'bytes7'), D), T[bo, i, by])  # fixed case 38930
ck(f(('bool', 'int48', 'bytes8'), D), T[bo, i, by])  # fixed case 38931
ck(f(('bool', 'int48', 'bytes9'), D), T[bo, i, by])  # fixed case 38932
ck(f(('bool', 'int48', 'int104'), D), T[bo, i, i])  # fixed case 38933
ck(f(('bool', 'int48', 'int112'), D), T[bo, i, i])  # fixed case 38934
ck(f(('bool', 'int48', 'int120'), D), T[bo, i, i])  # fixed case 38935
ck(f(('bool', 'int48', 'int128'), D), T[bo, i, i])  # fixed case 38936
ck(f(('bool', 'int48', 'int136'), D), T[bo, i, i])  # fixed case 38937
ck(f(('bool', 'int48', 'int144'), D), T[bo, i, i])  # fixed case 38938
ck(f(('bool', 'int48', 'int152'), D), T[bo, i, i])  # fixed case 38939
ck(f(('bool', 'int48', 'int16'), D), T[bo, i, i])  # fixed case 38940
ck(f(('bool', 'int48', 'int160'), D), T[bo, i, i])  # fixed case 38941
ck(f(('bool', 'int48', 'int168'), D), T[bo, i, i])  # fixed case 38942
ck(f(('bool', 'int48', 'int176'), D), T[bo, i, i])  # fixed case 38943
ck(f(('bool', 'int48', 'int184'), D), T[bo, i, i])  # fixed case 38944
ck(f(('bool', 'int48', 'int192'), D), T[bo, i, i])  # fixed case 38945
ck(f(('bool', 'int48', 'int200'), D), T[bo, i, i])  # fixed case 38946
ck(f(('bool', 'int48', 'int208'), D), T[bo, i, i])  # fixed case 38947
ck(f(('bool', 'int48', 'int216'), D), T[bo, i, i])  # fixed case 38948
ck(f(('bool', 'int48', 'int224'), D), T[bo, i, i])  # fixed case 38949
ck(f(('bool', 'int48', 'int232'), D), T[bo, i, i])  # fixed case 38950
ck(f(('bool', 'int48', 'int24'), D), T[bo, i, i])  # fixed case 38951
ck(f(('bool', 'int48', 'int240'), D), T[bo, i, i])  # fixed case 38952
ck(f(('bool', 'int48', 'int248'), D), T[bo, i, i])  # fixed case 38953
ck(f(('bool', 'int48', 'int256'), D), T[bo, i, i])  # fixed case 38954
ck(f(('bool', 'int48', 'int32'), D), T[bo, i, i])  # fixed case 38955
ck(f(('bool', 'int48', 'int40'), D), T[bo, i, i])  # fixed case 38956
ck(f(('bool', 'int48', 'int48'), D), T[bo, i, i])  # fixed case 38957
ck(f(('bool', 'int48', 'int56'), D), T[bo, i, i])  # fixed case 38958
ck(f(('bool', 'int48', 'int64'), D), T[bo, i, i])  # fixed case 38959
ck(f(('bool', 'int48', 'int72'), D), T[bo, i, i])  # fixed case 38960
ck(f(('bool', 'int48', 'int8'), D), T[bo, i, i])  # fixed case 38961
ck(f(('bool', 'int48', 'int80'), D), T[bo, i, i])  # fixed case 38962
ck(f(('bool', 'int48', 'int88'), D), T[bo, i, i])  # fixed case 38963
ck(f(('bool', 'int48', 'int96'), D), T[bo, i, i])  # fixed case 38964
ck(f(('bool', 'int48', 'string'), D), T[bo, i, s])  # fixed case 38965
ck(f(('bool', 'int48', 'string[]'), D), T[bo, i, A])  # fixed case 38966
ck(f(('bool', 'int48', 'uint104'), D), T[bo, i, i])  # fixed case 38967
ck(f(('bool', 'int48', 'uint112'), D), T[bo, i, i])  # fixed case 38968
ck(f(('bool', 'int48', 'uint120'), D), T[bo, i, i])  # fixed case 38969
ck(f(('bool', 'int48', 'uint128'), D), T[bo, i, i])  # fixed case 38970
ck(f(('bool', 'int48', 'uint136'), D), T[bo, i, i])  # fixed case 38971
ck(f(('bool', 'int48', 'uint144'), D), T[bo, i, i])  # fixed case 38972
ck(f(('bool', 'int48', 'uint152'), D), T[bo, i, i])  # fixed case 38973
ck(f(('bool', 'int48', 'uint16'), D), T[bo, i, i])  # fixed case 38974
ck(f(('bool', 'int48', 'uint160'), D), T[bo, i, i])  # fixed case 38975
ck(f(('bool', 'int48', 'uint168'), D), T[bo, i, i])  # fixed case 38976
ck(f(('bool', 'int48', 'uint176'), D), T[bo, i, i])  # fixed case 38977
ck(f(('bool', 'int48', 'uint184'), D), T[bo, i, i])  # fixed case 38978
ck(f(('bool', 'int48', 'uint192'), D), T[bo, i, i])  # fixed case 38979
ck(f(('bool', 'int48', 'uint200'), D), T[bo, i, i])  # fixed case 38980
ck(f(('bool', 'int48', 'uint208'), D), T[bo, i, i])  # fixed case 38981
ck(f(('bool', 'int48', 'uint216'), D), T[bo, i, i])  # fixed case 38982
ck(f(('bool', 'int48', 'uint224'), D), T[bo, i, i])  # fixed case 38983
ck(f(('bool', 'int48', 'uint232'), D), T[bo, i, i])  # fixed case 38984
ck(f(('bool', 'int48', 'uint24'), D), T[bo, i, i])  # fixed case 38985
ck(f(('bool', 'int48', 'uint240'), D), T[bo, i, i])  # fixed case 38986
ck(f(('bool', 'int48', 'uint248'), D), T[bo, i, i])  # fixed case 38987
ck(f(('bool', 'int48', 'uint256'), D), T[bo, i, i])  # fixed case 38988
ck(f(('bool', 'int48', 'uint32'), D), T[bo, i, i])  # fixed case 38989
ck(f(('bool', 'int48', 'uint40'), D), T[bo, i, i])  # fixed case 38990
ck(f(('bool', 'int48', 'uint48'), D), T[bo, i, i])  # fixed case 38991
ck(f(('bool', 'int48', 'uint56'), D), T[bo, i, i])  # fixed case 38992
ck(f(('bool', 'int48', 'uint64'), D), T[bo, i, i])  # fixed case 38993
ck(f(('bool', 'int48', 'uint72'), D), T[bo, i, i])  # fixed case 38994
ck(f(('bool', 'int48', 'uint8'), D), T[bo, i, i])  # fixed case 38995
ck(f(('bool', 'int48', 'uint80'), D), T[bo, i, i])  # fixed case 38996
ck(f(('bool', 'int48', 'uint88'), D), T[bo, i, i])  # fixed case 38997
ck(f(('bool', 'int48', 'uint96'), D), T[bo, i, i])  # fixed case 38998
ck(f(('bool', 'int48', '?'), D), T[bo, i, A])  # fixed case 38999
ck(f(('bool', 'int56', 'address'), D), T[bo, i, HA])  # fixed case 39000
ck(f(('bool', 'int56', 'address[]'), D), T[bo, i, A])  # fixed case 39001
ck(f(('bool', 'int56', 'bool'), D), T[bo, i, bo])  # fixed case 39002
ck(f(('bool', 'int56', 'bool[]'), D), T[bo, i, A])  # fixed case 39003
ck(f(('bool', 'int56', 'bytes'), D), T[bo, i, by])  # fixed case 39004
ck(f(('bool', 'int56', 'bytes1'), D), T[bo, i, by])  # fixed case 39005
ck(f(('bool', 'int56', 'bytes10'), D), T[bo, i, by])  # fixed case 39006
ck(f(('bool', 'int56', 'bytes11'), D), T[bo, i, by])  # fixed case 39007
ck(f(('bool', 'int56', 'bytes12'), D), T[bo, i, by])  # fixed case 39008
ck(f(('bool', 'int56', 'bytes13'), D), T[bo, i, by])  # fixed case 39009
ck(f(('bool', 'int56', 'bytes14'), D), T[bo, i, by])  # fixed case 39010
ck(f(('bool', 'int56', 'bytes15'), D), T[bo, i, by])  # fixed case 39011
ck(f(('bool', 'int56', 'bytes16'), D), T[bo, i, by])  # fixed case 39012
ck(f(('bool', 'int56', 'bytes17'), D), T[bo, i, by])  # fixed case 39013
ck(f(('bool', 'int56', 'bytes18'), D), T[bo, i, by])  # fixed case 39014
ck(f(('bool', 'int56', 'bytes19'), D), T[bo, i, by])  # fixed case 39015
ck(f(('bool', 'int56', 'bytes2'), D), T[bo, i, by])  # fixed case 39016
ck(f(('bool', 'int56', 'bytes20'), D), T[bo, i, by])  # fixed case 39017
ck(f(('bool', 'int56', 'bytes21'), D), T[bo, i, by])  # fixed case 39018
ck(f(('bool', 'int56', 'bytes22'), D), T[bo, i, by])  # fixed case 39019
ck(f(('bool', 'int56', 'bytes23'), D), T[bo, i, by])  # fixed case 39020
ck(f(('bool', 'int56', 'bytes24'), D), T[bo, i, by])  # fixed case 39021
ck(f(('bool', 'int56', 'bytes25'), D), T[bo, i, by])  # fixed case 39022
ck(f(('bool', 'int56', 'bytes26'), D), T[bo, i, by])  # fixed case 39023
ck(f(('bool', 'int56', 'bytes27'), D), T[bo, i, by])  # fixed case 39024
ck(f(('bool', 'int56', 'bytes28'), D), T[bo, i, by])  # fixed case 39025
ck(f(('bool', 'int56', 'bytes29'), D), T[bo, i, by])  # fixed case 39026
ck(f(('bool', 'int56', 'bytes3'), D), T[bo, i, by])  # fixed case 39027
ck(f(('bool', 'int56', 'bytes30'), D), T[bo, i, by])  # fixed case 39028
ck(f(('bool', 'int56', 'bytes31'), D), T[bo, i, by])  # fixed case 39029
ck(f(('bool', 'int56', 'bytes32'), D), T[bo, i, by])  # fixed case 39030
ck(f(('bool', 'int56', 'bytes4'), D), T[bo, i, by])  # fixed case 39031
ck(f(('bool', 'int56', 'bytes5'), D), T[bo, i, by])  # fixed case 39032
ck(f(('bool', 'int56', 'bytes6'), D), T[bo, i, by])  # fixed case 39033
ck(f(('bool', 'int56', 'bytes7'), D), T[bo, i, by])  # fixed case 39034
ck(f(('bool', 'int56', 'bytes8'), D), T[bo, i, by])  # fixed case 39035
ck(f(('bool', 'int56', 'bytes9'), D), T[bo, i, by])  # fixed case 39036
ck(f(('bool', 'int56', 'int104'), D), T[bo, i, i])  # fixed case 39037
ck(f(('bool', 'int56', 'int112'), D), T[bo, i, i])  # fixed case 39038
ck(f(('bool', 'int56', 'int120'), D), T[bo, i, i])  # fixed case 39039
ck(f(('bool', 'int56', 'int128'), D), T[bo, i, i])  # fixed case 39040
ck(f(('bool', 'int56', 'int136'), D), T[bo, i, i])  # fixed case 39041
ck(f(('bool', 'int56', 'int144'), D), T[bo, i, i])  # fixed case 39042
ck(f(('bool', 'int56', 'int152'), D), T[bo, i, i])  # fixed case 39043
ck(f(('bool', 'int56', 'int16'), D), T[bo, i, i])  # fixed case 39044
ck(f(('bool', 'int56', 'int160'), D), T[bo, i, i])  # fixed case 39045
ck(f(('bool', 'int56', 'int168'), D), T[bo, i, i])  # fixed case 39046
ck(f(('bool', 'int56', 'int176'), D), T[bo, i, i])  # fixed case 39047
ck(f(('bool', 'int56', 'int184'), D), T[bo, i, i])  # fixed case 39048
ck(f(('bool', 'int56', 'int192'), D), T[bo, i, i])  # fixed case 39049
ck(f(('bool', 'int56', 'int200'), D), T[bo, i, i])  # fixed case 39050
ck(f(('bool', 'int56', 'int208'), D), T[bo, i, i])  # fixed case 39051
ck(f(('bool', 'int56', 'int216'), D), T[bo, i, i])  # fixed case 39052
ck(f(('bool', 'int56', 'int224'), D), T[bo, i, i])  # fixed case 39053
ck(f(('bool', 'int56', 'int232'), D), T[bo, i, i])  # fixed case 39054
ck(f(('bool', 'int56', 'int24'), D), T[bo, i, i])  # fixed case 39055
ck(f(('bool', 'int56', 'int240'), D), T[bo, i, i])  # fixed case 39056
ck(f(('bool', 'int56', 'int248'), D), T[bo, i, i])  # fixed case 39057
ck(f(('bool', 'int56', 'int256'), D), T[bo, i, i])  # fixed case 39058
ck(f(('bool', 'int56', 'int32'), D), T[bo, i, i])  # fixed case 39059
ck(f(('bool', 'int56', 'int40'), D), T[bo, i, i])  # fixed case 39060
ck(f(('bool', 'int56', 'int48'), D), T[bo, i, i])  # fixed case 39061
ck(f(('bool', 'int56', 'int56'), D), T[bo, i, i])  # fixed case 39062
ck(f(('bool', 'int56', 'int64'), D), T[bo, i, i])  # fixed case 39063
ck(f(('bool', 'int56', 'int72'), D), T[bo, i, i])  # fixed case 39064
ck(f(('bool', 'int56', 'int8'), D), T[bo, i, i])  # fixed case 39065
ck(f(('bool', 'int56', 'int80'), D), T[bo, i, i])  # fixed case 39066
ck(f(('bool', 'int56', 'int88'), D), T[bo, i, i])  # fixed case 39067
ck(f(('bool', 'int56', 'int96'), D), T[bo, i, i])  # fixed case 39068
ck(f(('bool', 'int56', 'string'), D), T[bo, i, s])  # fixed case 39069
ck(f(('bool', 'int56', 'string[]'), D), T[bo, i, A])  # fixed case 39070
ck(f(('bool', 'int56', 'uint104'), D), T[bo, i, i])  # fixed case 39071
ck(f(('bool', 'int56', 'uint112'), D), T[bo, i, i])  # fixed case 39072
ck(f(('bool', 'int56', 'uint120'), D), T[bo, i, i])  # fixed case 39073
ck(f(('bool', 'int56', 'uint128'), D), T[bo, i, i])  # fixed case 39074
ck(f(('bool', 'int56', 'uint136'), D), T[bo, i, i])  # fixed case 39075
ck(f(('bool', 'int56', 'uint144'), D), T[bo, i, i])  # fixed case 39076
ck(f(('bool', 'int56', 'uint152'), D), T[bo, i, i])  # fixed case 39077
ck(f(('bool', 'int56', 'uint16'), D), T[bo, i, i])  # fixed case 39078
ck(f(('bool', 'int56', 'uint160'), D), T[bo, i, i])  # fixed case 39079
ck(f(('bool', 'int56', 'uint168'), D), T[bo, i, i])  # fixed case 39080
ck(f(('bool', 'int56', 'uint176'), D), T[bo, i, i])  # fixed case 39081
ck(f(('bool', 'int56', 'uint184'), D), T[bo, i, i])  # fixed case 39082
ck(f(('bool', 'int56', 'uint192'), D), T[bo, i, i])  # fixed case 39083
ck(f(('bool', 'int56', 'uint200'), D), T[bo, i, i])  # fixed case 39084
ck(f(('bool', 'int56', 'uint208'), D), T[bo, i, i])  # fixed case 39085
ck(f(('bool', 'int56', 'uint216'), D), T[bo, i, i])  # fixed case 39086
ck(f(('bool', 'int56', 'uint224'), D), T[bo, i, i])  # fixed case 39087
ck(f(('bool', 'int56', 'uint232'), D), T[bo, i, i])  # fixed case 39088
ck(f(('bool', 'int56', 'uint24'), D), T[bo, i, i])  # fixed case 39089
ck(f(('bool', 'int56', 'uint240'), D), T[bo, i, i])  # fixed case 39090
ck(f(('bool', 'int56', 'uint248'), D), T[bo, i, i])  # fixed case 39091
ck(f(('bool', 'int56', 'uint256'), D), T[bo, i, i])  # fixed case 39092
ck(f(('bool', 'int56', 'uint32'), D), T[bo, i, i])  # fixed case 39093
ck(f(('bool', 'int56', 'uint40'), D), T[bo, i, i])  # fixed case 39094
ck(f(('bool', 'int56', 'uint48'), D), T[bo, i, i])  # fixed case 39095
ck(f(('bool', 'int56', 'uint56'), D), T[bo, i, i])  # fixed case 39096
ck(f(('bool', 'int56', 'uint64'), D), T[bo, i, i])  # fixed case 39097
ck(f(('bool', 'int56', 'uint72'), D), T[bo, i, i])  # fixed case 39098
ck(f(('bool', 'int56', 'uint8'), D), T[bo, i, i])  # fixed case 39099
ck(f(('bool', 'int56', 'uint80'), D), T[bo, i, i])  # fixed case 39100
ck(f(('bool', 'int56', 'uint88'), D), T[bo, i, i])  # fixed case 39101
ck(f(('bool', 'int56', 'uint96'), D), T[bo, i, i])  # fixed case 39102
ck(f(('bool', 'int56', '?'), D), T[bo, i, A])  # fixed case 39103
ck(f(('bool', 'int64', 'address'), D), T[bo, i, HA])  # fixed case 39104
ck(f(('bool', 'int64', 'address[]'), D), T[bo, i, A])  # fixed case 39105
ck(f(('bool', 'int64', 'bool'), D), T[bo, i, bo])  # fixed case 39106
ck(f(('bool', 'int64', 'bool[]'), D), T[bo, i, A])  # fixed case 39107
ck(f(('bool', 'int64', 'bytes'), D), T[bo, i, by])  # fixed case 39108
ck(f(('bool', 'int64', 'bytes1'), D), T[bo, i, by])  # fixed case 39109
ck(f(('bool', 'int64', 'bytes10'), D), T[bo, i, by])  # fixed case 39110
ck(f(('bool', 'int64', 'bytes11'), D), T[bo, i, by])  # fixed case 39111
ck(f(('bool', 'int64', 'bytes12'), D), T[bo, i, by])  # fixed case 39112
ck(f(('bool', 'int64', 'bytes13'), D), T[bo, i, by])  # fixed case 39113
ck(f(('bool', 'int64', 'bytes14'), D), T[bo, i, by])  # fixed case 39114
ck(f(('bool', 'int64', 'bytes15'), D), T[bo, i, by])  # fixed case 39115
ck(f(('bool', 'int64', 'bytes16'), D), T[bo, i, by])  # fixed case 39116
ck(f(('bool', 'int64', 'bytes17'), D), T[bo, i, by])  # fixed case 39117
ck(f(('bool', 'int64', 'bytes18'), D), T[bo, i, by])  # fixed case 39118
ck(f(('bool', 'int64', 'bytes19'), D), T[bo, i, by])  # fixed case 39119
ck(f(('bool', 'int64', 'bytes2'), D), T[bo, i, by])  # fixed case 39120
ck(f(('bool', 'int64', 'bytes20'), D), T[bo, i, by])  # fixed case 39121
ck(f(('bool', 'int64', 'bytes21'), D), T[bo, i, by])  # fixed case 39122
ck(f(('bool', 'int64', 'bytes22'), D), T[bo, i, by])  # fixed case 39123
ck(f(('bool', 'int64', 'bytes23'), D), T[bo, i, by])  # fixed case 39124
ck(f(('bool', 'int64', 'bytes24'), D), T[bo, i, by])  # fixed case 39125
ck(f(('bool', 'int64', 'bytes25'), D), T[bo, i, by])  # fixed case 39126
ck(f(('bool', 'int64', 'bytes26'), D), T[bo, i, by])  # fixed case 39127
ck(f(('bool', 'int64', 'bytes27'), D), T[bo, i, by])  # fixed case 39128
ck(f(('bool', 'int64', 'bytes28'), D), T[bo, i, by])  # fixed case 39129
ck(f(('bool', 'int64', 'bytes29'), D), T[bo, i, by])  # fixed case 39130
ck(f(('bool', 'int64', 'bytes3'), D), T[bo, i, by])  # fixed case 39131
ck(f(('bool', 'int64', 'bytes30'), D), T[bo, i, by])  # fixed case 39132
ck(f(('bool', 'int64', 'bytes31'), D), T[bo, i, by])  # fixed case 39133
ck(f(('bool', 'int64', 'bytes32'), D), T[bo, i, by])  # fixed case 39134
ck(f(('bool', 'int64', 'bytes4'), D), T[bo, i, by])  # fixed case 39135
ck(f(('bool', 'int64', 'bytes5'), D), T[bo, i, by])  # fixed case 39136
ck(f(('bool', 'int64', 'bytes6'), D), T[bo, i, by])  # fixed case 39137
ck(f(('bool', 'int64', 'bytes7'), D), T[bo, i, by])  # fixed case 39138
ck(f(('bool', 'int64', 'bytes8'), D), T[bo, i, by])  # fixed case 39139
ck(f(('bool', 'int64', 'bytes9'), D), T[bo, i, by])  # fixed case 39140
ck(f(('bool', 'int64', 'int104'), D), T[bo, i, i])  # fixed case 39141
ck(f(('bool', 'int64', 'int112'), D), T[bo, i, i])  # fixed case 39142
ck(f(('bool', 'int64', 'int120'), D), T[bo, i, i])  # fixed case 39143
ck(f(('bool', 'int64', 'int128'), D), T[bo, i, i])  # fixed case 39144
ck(f(('bool', 'int64', 'int136'), D), T[bo, i, i])  # fixed case 39145
ck(f(('bool', 'int64', 'int144'), D), T[bo, i, i])  # fixed case 39146
ck(f(('bool', 'int64', 'int152'), D), T[bo, i, i])  # fixed case 39147
ck(f(('bool', 'int64', 'int16'), D), T[bo, i, i])  # fixed case 39148
ck(f(('bool', 'int64', 'int160'), D), T[bo, i, i])  # fixed case 39149
ck(f(('bool', 'int64', 'int168'), D), T[bo, i, i])  # fixed case 39150
ck(f(('bool', 'int64', 'int176'), D), T[bo, i, i])  # fixed case 39151
ck(f(('bool', 'int64', 'int184'), D), T[bo, i, i])  # fixed case 39152
ck(f(('bool', 'int64', 'int192'), D), T[bo, i, i])  # fixed case 39153
ck(f(('bool', 'int64', 'int200'), D), T[bo, i, i])  # fixed case 39154
ck(f(('bool', 'int64', 'int208'), D), T[bo, i, i])  # fixed case 39155
ck(f(('bool', 'int64', 'int216'), D), T[bo, i, i])  # fixed case 39156
ck(f(('bool', 'int64', 'int224'), D), T[bo, i, i])  # fixed case 39157
ck(f(('bool', 'int64', 'int232'), D), T[bo, i, i])  # fixed case 39158
ck(f(('bool', 'int64', 'int24'), D), T[bo, i, i])  # fixed case 39159
ck(f(('bool', 'int64', 'int240'), D), T[bo, i, i])  # fixed case 39160
ck(f(('bool', 'int64', 'int248'), D), T[bo, i, i])  # fixed case 39161
ck(f(('bool', 'int64', 'int256'), D), T[bo, i, i])  # fixed case 39162
ck(f(('bool', 'int64', 'int32'), D), T[bo, i, i])  # fixed case 39163
ck(f(('bool', 'int64', 'int40'), D), T[bo, i, i])  # fixed case 39164
ck(f(('bool', 'int64', 'int48'), D), T[bo, i, i])  # fixed case 39165
ck(f(('bool', 'int64', 'int56'), D), T[bo, i, i])  # fixed case 39166
ck(f(('bool', 'int64', 'int64'), D), T[bo, i, i])  # fixed case 39167
ck(f(('bool', 'int64', 'int72'), D), T[bo, i, i])  # fixed case 39168
ck(f(('bool', 'int64', 'int8'), D), T[bo, i, i])  # fixed case 39169
ck(f(('bool', 'int64', 'int80'), D), T[bo, i, i])  # fixed case 39170
ck(f(('bool', 'int64', 'int88'), D), T[bo, i, i])  # fixed case 39171
ck(f(('bool', 'int64', 'int96'), D), T[bo, i, i])  # fixed case 39172
ck(f(('bool', 'int64', 'string'), D), T[bo, i, s])  # fixed case 39173
ck(f(('bool', 'int64', 'string[]'), D), T[bo, i, A])  # fixed case 39174
ck(f(('bool', 'int64', 'uint104'), D), T[bo, i, i])  # fixed case 39175
ck(f(('bool', 'int64', 'uint112'), D), T[bo, i, i])  # fixed case 39176
ck(f(('bool', 'int64', 'uint120'), D), T[bo, i, i])  # fixed case 39177
ck(f(('bool', 'int64', 'uint128'), D), T[bo, i, i])  # fixed case 39178
ck(f(('bool', 'int64', 'uint136'), D), T[bo, i, i])  # fixed case 39179
ck(f(('bool', 'int64', 'uint144'), D), T[bo, i, i])  # fixed case 39180
ck(f(('bool', 'int64', 'uint152'), D), T[bo, i, i])  # fixed case 39181
ck(f(('bool', 'int64', 'uint16'), D), T[bo, i, i])  # fixed case 39182
ck(f(('bool', 'int64', 'uint160'), D), T[bo, i, i])  # fixed case 39183
ck(f(('bool', 'int64', 'uint168'), D), T[bo, i, i])  # fixed case 39184
ck(f(('bool', 'int64', 'uint176'), D), T[bo, i, i])  # fixed case 39185
ck(f(('bool', 'int64', 'uint184'), D), T[bo, i, i])  # fixed case 39186
ck(f(('bool', 'int64', 'uint192'), D), T[bo, i, i])  # fixed case 39187
ck(f(('bool', 'int64', 'uint200'), D), T[bo, i, i])  # fixed case 39188
ck(f(('bool', 'int64', 'uint208'), D), T[bo, i, i])  # fixed case 39189
ck(f(('bool', 'int64', 'uint216'), D), T[bo, i, i])  # fixed case 39190
ck(f(('bool', 'int64', 'uint224'), D), T[bo, i, i])  # fixed case 39191
ck(f(('bool', 'int64', 'uint232'), D), T[bo, i, i])  # fixed case 39192
ck(f(('bool', 'int64', 'uint24'), D), T[bo, i, i])  # fixed case 39193
ck(f(('bool', 'int64', 'uint240'), D), T[bo, i, i])  # fixed case 39194
ck(f(('bool', 'int64', 'uint248'), D), T[bo, i, i])  # fixed case 39195
ck(f(('bool', 'int64', 'uint256'), D), T[bo, i, i])  # fixed case 39196
ck(f(('bool', 'int64', 'uint32'), D), T[bo, i, i])  # fixed case 39197
ck(f(('bool', 'int64', 'uint40'), D), T[bo, i, i])  # fixed case 39198
ck(f(('bool', 'int64', 'uint48'), D), T[bo, i, i])  # fixed case 39199
ck(f(('bool', 'int64', 'uint56'), D), T[bo, i, i])  # fixed case 39200
ck(f(('bool', 'int64', 'uint64'), D), T[bo, i, i])  # fixed case 39201
ck(f(('bool', 'int64', 'uint72'), D), T[bo, i, i])  # fixed case 39202
ck(f(('bool', 'int64', 'uint8'), D), T[bo, i, i])  # fixed case 39203
ck(f(('bool', 'int64', 'uint80'), D), T[bo, i, i])  # fixed case 39204
ck(f(('bool', 'int64', 'uint88'), D), T[bo, i, i])  # fixed case 39205
ck(f(('bool', 'int64', 'uint96'), D), T[bo, i, i])  # fixed case 39206
ck(f(('bool', 'int64', '?'), D), T[bo, i, A])  # fixed case 39207
ck(f(('bool', 'int72', 'address'), D), T[bo, i, HA])  # fixed case 39208
ck(f(('bool', 'int72', 'address[]'), D), T[bo, i, A])  # fixed case 39209
ck(f(('bool', 'int72', 'bool'), D), T[bo, i, bo])  # fixed case 39210
ck(f(('bool', 'int72', 'bool[]'), D), T[bo, i, A])  # fixed case 39211
ck(f(('bool', 'int72', 'bytes'), D), T[bo, i, by])  # fixed case 39212
ck(f(('bool', 'int72', 'bytes1'), D), T[bo, i, by])  # fixed case 39213
ck(f(('bool', 'int72', 'bytes10'), D), T[bo, i, by])  # fixed case 39214
ck(f(('bool', 'int72', 'bytes11'), D), T[bo, i, by])  # fixed case 39215
ck(f(('bool', 'int72', 'bytes12'), D), T[bo, i, by])  # fixed case 39216
ck(f(('bool', 'int72', 'bytes13'), D), T[bo, i, by])  # fixed case 39217
ck(f(('bool', 'int72', 'bytes14'), D), T[bo, i, by])  # fixed case 39218
ck(f(('bool', 'int72', 'bytes15'), D), T[bo, i, by])  # fixed case 39219
ck(f(('bool', 'int72', 'bytes16'), D), T[bo, i, by])  # fixed case 39220
ck(f(('bool', 'int72', 'bytes17'), D), T[bo, i, by])  # fixed case 39221
ck(f(('bool', 'int72', 'bytes18'), D), T[bo, i, by])  # fixed case 39222
ck(f(('bool', 'int72', 'bytes19'), D), T[bo, i, by])  # fixed case 39223
ck(f(('bool', 'int72', 'bytes2'), D), T[bo, i, by])  # fixed case 39224
ck(f(('bool', 'int72', 'bytes20'), D), T[bo, i, by])  # fixed case 39225
ck(f(('bool', 'int72', 'bytes21'), D), T[bo, i, by])  # fixed case 39226
ck(f(('bool', 'int72', 'bytes22'), D), T[bo, i, by])  # fixed case 39227
ck(f(('bool', 'int72', 'bytes23'), D), T[bo, i, by])  # fixed case 39228
ck(f(('bool', 'int72', 'bytes24'), D), T[bo, i, by])  # fixed case 39229
ck(f(('bool', 'int72', 'bytes25'), D), T[bo, i, by])  # fixed case 39230
ck(f(('bool', 'int72', 'bytes26'), D), T[bo, i, by])  # fixed case 39231
ck(f(('bool', 'int72', 'bytes27'), D), T[bo, i, by])  # fixed case 39232
ck(f(('bool', 'int72', 'bytes28'), D), T[bo, i, by])  # fixed case 39233
ck(f(('bool', 'int72', 'bytes29'), D), T[bo, i, by])  # fixed case 39234
ck(f(('bool', 'int72', 'bytes3'), D), T[bo, i, by])  # fixed case 39235
ck(f(('bool', 'int72', 'bytes30'), D), T[bo, i, by])  # fixed case 39236
ck(f(('bool', 'int72', 'bytes31'), D), T[bo, i, by])  # fixed case 39237
ck(f(('bool', 'int72', 'bytes32'), D), T[bo, i, by])  # fixed case 39238
ck(f(('bool', 'int72', 'bytes4'), D), T[bo, i, by])  # fixed case 39239
ck(f(('bool', 'int72', 'bytes5'), D), T[bo, i, by])  # fixed case 39240
ck(f(('bool', 'int72', 'bytes6'), D), T[bo, i, by])  # fixed case 39241
ck(f(('bool', 'int72', 'bytes7'), D), T[bo, i, by])  # fixed case 39242
ck(f(('bool', 'int72', 'bytes8'), D), T[bo, i, by])  # fixed case 39243
ck(f(('bool', 'int72', 'bytes9'), D), T[bo, i, by])  # fixed case 39244
ck(f(('bool', 'int72', 'int104'), D), T[bo, i, i])  # fixed case 39245
ck(f(('bool', 'int72', 'int112'), D), T[bo, i, i])  # fixed case 39246
ck(f(('bool', 'int72', 'int120'), D), T[bo, i, i])  # fixed case 39247
ck(f(('bool', 'int72', 'int128'), D), T[bo, i, i])  # fixed case 39248
ck(f(('bool', 'int72', 'int136'), D), T[bo, i, i])  # fixed case 39249
ck(f(('bool', 'int72', 'int144'), D), T[bo, i, i])  # fixed case 39250
ck(f(('bool', 'int72', 'int152'), D), T[bo, i, i])  # fixed case 39251
ck(f(('bool', 'int72', 'int16'), D), T[bo, i, i])  # fixed case 39252
ck(f(('bool', 'int72', 'int160'), D), T[bo, i, i])  # fixed case 39253
ck(f(('bool', 'int72', 'int168'), D), T[bo, i, i])  # fixed case 39254
ck(f(('bool', 'int72', 'int176'), D), T[bo, i, i])  # fixed case 39255
ck(f(('bool', 'int72', 'int184'), D), T[bo, i, i])  # fixed case 39256
ck(f(('bool', 'int72', 'int192'), D), T[bo, i, i])  # fixed case 39257
ck(f(('bool', 'int72', 'int200'), D), T[bo, i, i])  # fixed case 39258
ck(f(('bool', 'int72', 'int208'), D), T[bo, i, i])  # fixed case 39259
ck(f(('bool', 'int72', 'int216'), D), T[bo, i, i])  # fixed case 39260
ck(f(('bool', 'int72', 'int224'), D), T[bo, i, i])  # fixed case 39261
ck(f(('bool', 'int72', 'int232'), D), T[bo, i, i])  # fixed case 39262
ck(f(('bool', 'int72', 'int24'), D), T[bo, i, i])  # fixed case 39263
ck(f(('bool', 'int72', 'int240'), D), T[bo, i, i])  # fixed case 39264
ck(f(('bool', 'int72', 'int248'), D), T[bo, i, i])  # fixed case 39265
ck(f(('bool', 'int72', 'int256'), D), T[bo, i, i])  # fixed case 39266
ck(f(('bool', 'int72', 'int32'), D), T[bo, i, i])  # fixed case 39267
ck(f(('bool', 'int72', 'int40'), D), T[bo, i, i])  # fixed case 39268
ck(f(('bool', 'int72', 'int48'), D), T[bo, i, i])  # fixed case 39269
ck(f(('bool', 'int72', 'int56'), D), T[bo, i, i])  # fixed case 39270
ck(f(('bool', 'int72', 'int64'), D), T[bo, i, i])  # fixed case 39271
ck(f(('bool', 'int72', 'int72'), D), T[bo, i, i])  # fixed case 39272
ck(f(('bool', 'int72', 'int8'), D), T[bo, i, i])  # fixed case 39273
ck(f(('bool', 'int72', 'int80'), D), T[bo, i, i])  # fixed case 39274
ck(f(('bool', 'int72', 'int88'), D), T[bo, i, i])  # fixed case 39275
ck(f(('bool', 'int72', 'int96'), D), T[bo, i, i])  # fixed case 39276
ck(f(('bool', 'int72', 'string'), D), T[bo, i, s])  # fixed case 39277
ck(f(('bool', 'int72', 'string[]'), D), T[bo, i, A])  # fixed case 39278
ck(f(('bool', 'int72', 'uint104'), D), T[bo, i, i])  # fixed case 39279
ck(f(('bool', 'int72', 'uint112'), D), T[bo, i, i])  # fixed case 39280
ck(f(('bool', 'int72', 'uint120'), D), T[bo, i, i])  # fixed case 39281
ck(f(('bool', 'int72', 'uint128'), D), T[bo, i, i])  # fixed case 39282
ck(f(('bool', 'int72', 'uint136'), D), T[bo, i, i])  # fixed case 39283
ck(f(('bool', 'int72', 'uint144'), D), T[bo, i, i])  # fixed case 39284
ck(f(('bool', 'int72', 'uint152'), D), T[bo, i, i])  # fixed case 39285
ck(f(('bool', 'int72', 'uint16'), D), T[bo, i, i])  # fixed case 39286
ck(f(('bool', 'int72', 'uint160'), D), T[bo, i, i])  # fixed case 39287
ck(f(('bool', 'int72', 'uint168'), D), T[bo, i, i])  # fixed case 39288
ck(f(('bool', 'int72', 'uint176'), D), T[bo, i, i])  # fixed case 39289
ck(f(('bool', 'int72', 'uint184'), D), T[bo, i, i])  # fixed case 39290
ck(f(('bool', 'int72', 'uint192'), D), T[bo, i, i])  # fixed case 39291
ck(f(('bool', 'int72', 'uint200'), D), T[bo, i, i])  # fixed case 39292
ck(f(('bool', 'int72', 'uint208'), D), T[bo, i, i])  # fixed case 39293
ck(f(('bool', 'int72', 'uint216'), D), T[bo, i, i])  # fixed case 39294
ck(f(('bool', 'int72', 'uint224'), D), T[bo, i, i])  # fixed case 39295
ck(f(('bool', 'int72', 'uint232'), D), T[bo, i, i])  # fixed case 39296
ck(f(('bool', 'int72', 'uint24'), D), T[bo, i, i])  # fixed case 39297
ck(f(('bool', 'int72', 'uint240'), D), T[bo, i, i])  # fixed case 39298
ck(f(('bool', 'int72', 'uint248'), D), T[bo, i, i])  # fixed case 39299
ck(f(('bool', 'int72', 'uint256'), D), T[bo, i, i])  # fixed case 39300
ck(f(('bool', 'int72', 'uint32'), D), T[bo, i, i])  # fixed case 39301
ck(f(('bool', 'int72', 'uint40'), D), T[bo, i, i])  # fixed case 39302
ck(f(('bool', 'int72', 'uint48'), D), T[bo, i, i])  # fixed case 39303
ck(f(('bool', 'int72', 'uint56'), D), T[bo, i, i])  # fixed case 39304
ck(f(('bool', 'int72', 'uint64'), D), T[bo, i, i])  # fixed case 39305
ck(f(('bool', 'int72', 'uint72'), D), T[bo, i, i])  # fixed case 39306
ck(f(('bool', 'int72', 'uint8'), D), T[bo, i, i])  # fixed case 39307
ck(f(('bool', 'int72', 'uint80'), D), T[bo, i, i])  # fixed case 39308
ck(f(('bool', 'int72', 'uint88'), D), T[bo, i, i])  # fixed case 39309
ck(f(('bool', 'int72', 'uint96'), D), T[bo, i, i])  # fixed case 39310
ck(f(('bool', 'int72', '?'), D), T[bo, i, A])  # fixed case 39311
ck(f(('bool', 'int8', 'address'), D), T[bo, i, HA])  # fixed case 39312
ck(f(('bool', 'int8', 'address[]'), D), T[bo, i, A])  # fixed case 39313
ck(f(('bool', 'int8', 'bool'), D), T[bo, i, bo])  # fixed case 39314
ck(f(('bool', 'int8', 'bool[]'), D), T[bo, i, A])  # fixed case 39315
ck(f(('bool', 'int8', 'bytes'), D), T[bo, i, by])  # fixed case 39316
ck(f(('bool', 'int8', 'bytes1'), D), T[bo, i, by])  # fixed case 39317
ck(f(('bool', 'int8', 'bytes10'), D), T[bo, i, by])  # fixed case 39318
ck(f(('bool', 'int8', 'bytes11'), D), T[bo, i, by])  # fixed case 39319
ck(f(('bool', 'int8', 'bytes12'), D), T[bo, i, by])  # fixed case 39320
ck(f(('bool', 'int8', 'bytes13'), D), T[bo, i, by])  # fixed case 39321
ck(f(('bool', 'int8', 'bytes14'), D), T[bo, i, by])  # fixed case 39322
ck(f(('bool', 'int8', 'bytes15'), D), T[bo, i, by])  # fixed case 39323
ck(f(('bool', 'int8', 'bytes16'), D), T[bo, i, by])  # fixed case 39324
ck(f(('bool', 'int8', 'bytes17'), D), T[bo, i, by])  # fixed case 39325
ck(f(('bool', 'int8', 'bytes18'), D), T[bo, i, by])  # fixed case 39326
ck(f(('bool', 'int8', 'bytes19'), D), T[bo, i, by])  # fixed case 39327
ck(f(('bool', 'int8', 'bytes2'), D), T[bo, i, by])  # fixed case 39328
ck(f(('bool', 'int8', 'bytes20'), D), T[bo, i, by])  # fixed case 39329
ck(f(('bool', 'int8', 'bytes21'), D), T[bo, i, by])  # fixed case 39330
ck(f(('bool', 'int8', 'bytes22'), D), T[bo, i, by])  # fixed case 39331
ck(f(('bool', 'int8', 'bytes23'), D), T[bo, i, by])  # fixed case 39332
ck(f(('bool', 'int8', 'bytes24'), D), T[bo, i, by])  # fixed case 39333
ck(f(('bool', 'int8', 'bytes25'), D), T[bo, i, by])  # fixed case 39334
ck(f(('bool', 'int8', 'bytes26'), D), T[bo, i, by])  # fixed case 39335
ck(f(('bool', 'int8', 'bytes27'), D), T[bo, i, by])  # fixed case 39336
ck(f(('bool', 'int8', 'bytes28'), D), T[bo, i, by])  # fixed case 39337
ck(f(('bool', 'int8', 'bytes29'), D), T[bo, i, by])  # fixed case 39338
ck(f(('bool', 'int8', 'bytes3'), D), T[bo, i, by])  # fixed case 39339
ck(f(('bool', 'int8', 'bytes30'), D), T[bo, i, by])  # fixed case 39340
ck(f(('bool', 'int8', 'bytes31'), D), T[bo, i, by])  # fixed case 39341
ck(f(('bool', 'int8', 'bytes32'), D), T[bo, i, by])  # fixed case 39342
ck(f(('bool', 'int8', 'bytes4'), D), T[bo, i, by])  # fixed case 39343
ck(f(('bool', 'int8', 'bytes5'), D), T[bo, i, by])  # fixed case 39344
ck(f(('bool', 'int8', 'bytes6'), D), T[bo, i, by])  # fixed case 39345
ck(f(('bool', 'int8', 'bytes7'), D), T[bo, i, by])  # fixed case 39346
ck(f(('bool', 'int8', 'bytes8'), D), T[bo, i, by])  # fixed case 39347
ck(f(('bool', 'int8', 'bytes9'), D), T[bo, i, by])  # fixed case 39348
ck(f(('bool', 'int8', 'int104'), D), T[bo, i, i])  # fixed case 39349
ck(f(('bool', 'int8', 'int112'), D), T[bo, i, i])  # fixed case 39350
ck(f(('bool', 'int8', 'int120'), D), T[bo, i, i])  # fixed case 39351
ck(f(('bool', 'int8', 'int128'), D), T[bo, i, i])  # fixed case 39352
ck(f(('bool', 'int8', 'int136'), D), T[bo, i, i])  # fixed case 39353
ck(f(('bool', 'int8', 'int144'), D), T[bo, i, i])  # fixed case 39354
ck(f(('bool', 'int8', 'int152'), D), T[bo, i, i])  # fixed case 39355
ck(f(('bool', 'int8', 'int16'), D), T[bo, i, i])  # fixed case 39356
ck(f(('bool', 'int8', 'int160'), D), T[bo, i, i])  # fixed case 39357
ck(f(('bool', 'int8', 'int168'), D), T[bo, i, i])  # fixed case 39358
ck(f(('bool', 'int8', 'int176'), D), T[bo, i, i])  # fixed case 39359
ck(f(('bool', 'int8', 'int184'), D), T[bo, i, i])  # fixed case 39360
ck(f(('bool', 'int8', 'int192'), D), T[bo, i, i])  # fixed case 39361
ck(f(('bool', 'int8', 'int200'), D), T[bo, i, i])  # fixed case 39362
ck(f(('bool', 'int8', 'int208'), D), T[bo, i, i])  # fixed case 39363
ck(f(('bool', 'int8', 'int216'), D), T[bo, i, i])  # fixed case 39364
ck(f(('bool', 'int8', 'int224'), D), T[bo, i, i])  # fixed case 39365
ck(f(('bool', 'int8', 'int232'), D), T[bo, i, i])  # fixed case 39366
ck(f(('bool', 'int8', 'int24'), D), T[bo, i, i])  # fixed case 39367
ck(f(('bool', 'int8', 'int240'), D), T[bo, i, i])  # fixed case 39368
ck(f(('bool', 'int8', 'int248'), D), T[bo, i, i])  # fixed case 39369
ck(f(('bool', 'int8', 'int256'), D), T[bo, i, i])  # fixed case 39370
ck(f(('bool', 'int8', 'int32'), D), T[bo, i, i])  # fixed case 39371
ck(f(('bool', 'int8', 'int40'), D), T[bo, i, i])  # fixed case 39372
ck(f(('bool', 'int8', 'int48'), D), T[bo, i, i])  # fixed case 39373
ck(f(('bool', 'int8', 'int56'), D), T[bo, i, i])  # fixed case 39374
ck(f(('bool', 'int8', 'int64'), D), T[bo, i, i])  # fixed case 39375
ck(f(('bool', 'int8', 'int72'), D), T[bo, i, i])  # fixed case 39376
ck(f(('bool', 'int8', 'int8'), D), T[bo, i, i])  # fixed case 39377
ck(f(('bool', 'int8', 'int80'), D), T[bo, i, i])  # fixed case 39378
ck(f(('bool', 'int8', 'int88'), D), T[bo, i, i])  # fixed case 39379
ck(f(('bool', 'int8', 'int96'), D), T[bo, i, i])  # fixed case 39380
ck(f(('bool', 'int8', 'string'), D), T[bo, i, s])  # fixed case 39381
ck(f(('bool', 'int8', 'string[]'), D), T[bo, i, A])  # fixed case 39382
ck(f(('bool', 'int8', 'uint104'), D), T[bo, i, i])  # fixed case 39383
ck(f(('bool', 'int8', 'uint112'), D), T[bo, i, i])  # fixed case 39384
ck(f(('bool', 'int8', 'uint120'), D), T[bo, i, i])  # fixed case 39385
ck(f(('bool', 'int8', 'uint128'), D), T[bo, i, i])  # fixed case 39386
ck(f(('bool', 'int8', 'uint136'), D), T[bo, i, i])  # fixed case 39387
ck(f(('bool', 'int8', 'uint144'), D), T[bo, i, i])  # fixed case 39388
ck(f(('bool', 'int8', 'uint152'), D), T[bo, i, i])  # fixed case 39389
ck(f(('bool', 'int8', 'uint16'), D), T[bo, i, i])  # fixed case 39390
ck(f(('bool', 'int8', 'uint160'), D), T[bo, i, i])  # fixed case 39391
ck(f(('bool', 'int8', 'uint168'), D), T[bo, i, i])  # fixed case 39392
ck(f(('bool', 'int8', 'uint176'), D), T[bo, i, i])  # fixed case 39393
ck(f(('bool', 'int8', 'uint184'), D), T[bo, i, i])  # fixed case 39394
ck(f(('bool', 'int8', 'uint192'), D), T[bo, i, i])  # fixed case 39395
ck(f(('bool', 'int8', 'uint200'), D), T[bo, i, i])  # fixed case 39396
ck(f(('bool', 'int8', 'uint208'), D), T[bo, i, i])  # fixed case 39397
ck(f(('bool', 'int8', 'uint216'), D), T[bo, i, i])  # fixed case 39398
ck(f(('bool', 'int8', 'uint224'), D), T[bo, i, i])  # fixed case 39399
ck(f(('bool', 'int8', 'uint232'), D), T[bo, i, i])  # fixed case 39400
ck(f(('bool', 'int8', 'uint24'), D), T[bo, i, i])  # fixed case 39401
ck(f(('bool', 'int8', 'uint240'), D), T[bo, i, i])  # fixed case 39402
ck(f(('bool', 'int8', 'uint248'), D), T[bo, i, i])  # fixed case 39403
ck(f(('bool', 'int8', 'uint256'), D), T[bo, i, i])  # fixed case 39404
ck(f(('bool', 'int8', 'uint32'), D), T[bo, i, i])  # fixed case 39405
ck(f(('bool', 'int8', 'uint40'), D), T[bo, i, i])  # fixed case 39406
ck(f(('bool', 'int8', 'uint48'), D), T[bo, i, i])  # fixed case 39407
ck(f(('bool', 'int8', 'uint56'), D), T[bo, i, i])  # fixed case 39408
ck(f(('bool', 'int8', 'uint64'), D), T[bo, i, i])  # fixed case 39409
ck(f(('bool', 'int8', 'uint72'), D), T[bo, i, i])  # fixed case 39410
ck(f(('bool', 'int8', 'uint8'), D), T[bo, i, i])  # fixed case 39411
ck(f(('bool', 'int8', 'uint80'), D), T[bo, i, i])  # fixed case 39412
ck(f(('bool', 'int8', 'uint88'), D), T[bo, i, i])  # fixed case 39413
ck(f(('bool', 'int8', 'uint96'), D), T[bo, i, i])  # fixed case 39414
ck(f(('bool', 'int8', '?'), D), T[bo, i, A])  # fixed case 39415
ck(f(('bool', 'int80', 'address'), D), T[bo, i, HA])  # fixed case 39416
ck(f(('bool', 'int80', 'address[]'), D), T[bo, i, A])  # fixed case 39417
ck(f(('bool', 'int80', 'bool'), D), T[bo, i, bo])  # fixed case 39418
ck(f(('bool', 'int80', 'bool[]'), D), T[bo, i, A])  # fixed case 39419
ck(f(('bool', 'int80', 'bytes'), D), T[bo, i, by])  # fixed case 39420
ck(f(('bool', 'int80', 'bytes1'), D), T[bo, i, by])  # fixed case 39421
ck(f(('bool', 'int80', 'bytes10'), D), T[bo, i, by])  # fixed case 39422
ck(f(('bool', 'int80', 'bytes11'), D), T[bo, i, by])  # fixed case 39423
ck(f(('bool', 'int80', 'bytes12'), D), T[bo, i, by])  # fixed case 39424
ck(f(('bool', 'int80', 'bytes13'), D), T[bo, i, by])  # fixed case 39425
ck(f(('bool', 'int80', 'bytes14'), D), T[bo, i, by])  # fixed case 39426
ck(f(('bool', 'int80', 'bytes15'), D), T[bo, i, by])  # fixed case 39427
ck(f(('bool', 'int80', 'bytes16'), D), T[bo, i, by])  # fixed case 39428
ck(f(('bool', 'int80', 'bytes17'), D), T[bo, i, by])  # fixed case 39429
ck(f(('bool', 'int80', 'bytes18'), D), T[bo, i, by])  # fixed case 39430
ck(f(('bool', 'int80', 'bytes19'), D), T[bo, i, by])  # fixed case 39431
ck(f(('bool', 'int80', 'bytes2'), D), T[bo, i, by])  # fixed case 39432
ck(f(('bool', 'int80', 'bytes20'), D), T[bo, i, by])  # fixed case 39433
ck(f(('bool', 'int80', 'bytes21'), D), T[bo, i, by])  # fixed case 39434
ck(f(('bool', 'int80', 'bytes22'), D), T[bo, i, by])  # fixed case 39435
ck(f(('bool', 'int80', 'bytes23'), D), T[bo, i, by])  # fixed case 39436
ck(f(('bool', 'int80', 'bytes24'), D), T[bo, i, by])  # fixed case 39437
ck(f(('bool', 'int80', 'bytes25'), D), T[bo, i, by])  # fixed case 39438
ck(f(('bool', 'int80', 'bytes26'), D), T[bo, i, by])  # fixed case 39439
ck(f(('bool', 'int80', 'bytes27'), D), T[bo, i, by])  # fixed case 39440
ck(f(('bool', 'int80', 'bytes28'), D), T[bo, i, by])  # fixed case 39441
ck(f(('bool', 'int80', 'bytes29'), D), T[bo, i, by])  # fixed case 39442
ck(f(('bool', 'int80', 'bytes3'), D), T[bo, i, by])  # fixed case 39443
ck(f(('bool', 'int80', 'bytes30'), D), T[bo, i, by])  # fixed case 39444
ck(f(('bool', 'int80', 'bytes31'), D), T[bo, i, by])  # fixed case 39445
ck(f(('bool', 'int80', 'bytes32'), D), T[bo, i, by])  # fixed case 39446
ck(f(('bool', 'int80', 'bytes4'), D), T[bo, i, by])  # fixed case 39447
ck(f(('bool', 'int80', 'bytes5'), D), T[bo, i, by])  # fixed case 39448
ck(f(('bool', 'int80', 'bytes6'), D), T[bo, i, by])  # fixed case 39449
ck(f(('bool', 'int80', 'bytes7'), D), T[bo, i, by])  # fixed case 39450
ck(f(('bool', 'int80', 'bytes8'), D), T[bo, i, by])  # fixed case 39451
ck(f(('bool', 'int80', 'bytes9'), D), T[bo, i, by])  # fixed case 39452
ck(f(('bool', 'int80', 'int104'), D), T[bo, i, i])  # fixed case 39453
ck(f(('bool', 'int80', 'int112'), D), T[bo, i, i])  # fixed case 39454
ck(f(('bool', 'int80', 'int120'), D), T[bo, i, i])  # fixed case 39455
ck(f(('bool', 'int80', 'int128'), D), T[bo, i, i])  # fixed case 39456
ck(f(('bool', 'int80', 'int136'), D), T[bo, i, i])  # fixed case 39457
ck(f(('bool', 'int80', 'int144'), D), T[bo, i, i])  # fixed case 39458
ck(f(('bool', 'int80', 'int152'), D), T[bo, i, i])  # fixed case 39459
ck(f(('bool', 'int80', 'int16'), D), T[bo, i, i])  # fixed case 39460
ck(f(('bool', 'int80', 'int160'), D), T[bo, i, i])  # fixed case 39461
ck(f(('bool', 'int80', 'int168'), D), T[bo, i, i])  # fixed case 39462
ck(f(('bool', 'int80', 'int176'), D), T[bo, i, i])  # fixed case 39463
ck(f(('bool', 'int80', 'int184'), D), T[bo, i, i])  # fixed case 39464
ck(f(('bool', 'int80', 'int192'), D), T[bo, i, i])  # fixed case 39465
ck(f(('bool', 'int80', 'int200'), D), T[bo, i, i])  # fixed case 39466
ck(f(('bool', 'int80', 'int208'), D), T[bo, i, i])  # fixed case 39467
ck(f(('bool', 'int80', 'int216'), D), T[bo, i, i])  # fixed case 39468
ck(f(('bool', 'int80', 'int224'), D), T[bo, i, i])  # fixed case 39469
ck(f(('bool', 'int80', 'int232'), D), T[bo, i, i])  # fixed case 39470
ck(f(('bool', 'int80', 'int24'), D), T[bo, i, i])  # fixed case 39471
ck(f(('bool', 'int80', 'int240'), D), T[bo, i, i])  # fixed case 39472
ck(f(('bool', 'int80', 'int248'), D), T[bo, i, i])  # fixed case 39473
ck(f(('bool', 'int80', 'int256'), D), T[bo, i, i])  # fixed case 39474
ck(f(('bool', 'int80', 'int32'), D), T[bo, i, i])  # fixed case 39475
ck(f(('bool', 'int80', 'int40'), D), T[bo, i, i])  # fixed case 39476
ck(f(('bool', 'int80', 'int48'), D), T[bo, i, i])  # fixed case 39477
ck(f(('bool', 'int80', 'int56'), D), T[bo, i, i])  # fixed case 39478
ck(f(('bool', 'int80', 'int64'), D), T[bo, i, i])  # fixed case 39479
ck(f(('bool', 'int80', 'int72'), D), T[bo, i, i])  # fixed case 39480
ck(f(('bool', 'int80', 'int8'), D), T[bo, i, i])  # fixed case 39481
ck(f(('bool', 'int80', 'int80'), D), T[bo, i, i])  # fixed case 39482
ck(f(('bool', 'int80', 'int88'), D), T[bo, i, i])  # fixed case 39483
ck(f(('bool', 'int80', 'int96'), D), T[bo, i, i])  # fixed case 39484
ck(f(('bool', 'int80', 'string'), D), T[bo, i, s])  # fixed case 39485
ck(f(('bool', 'int80', 'string[]'), D), T[bo, i, A])  # fixed case 39486
ck(f(('bool', 'int80', 'uint104'), D), T[bo, i, i])  # fixed case 39487
ck(f(('bool', 'int80', 'uint112'), D), T[bo, i, i])  # fixed case 39488
ck(f(('bool', 'int80', 'uint120'), D), T[bo, i, i])  # fixed case 39489
ck(f(('bool', 'int80', 'uint128'), D), T[bo, i, i])  # fixed case 39490
ck(f(('bool', 'int80', 'uint136'), D), T[bo, i, i])  # fixed case 39491
ck(f(('bool', 'int80', 'uint144'), D), T[bo, i, i])  # fixed case 39492
ck(f(('bool', 'int80', 'uint152'), D), T[bo, i, i])  # fixed case 39493
ck(f(('bool', 'int80', 'uint16'), D), T[bo, i, i])  # fixed case 39494
ck(f(('bool', 'int80', 'uint160'), D), T[bo, i, i])  # fixed case 39495
ck(f(('bool', 'int80', 'uint168'), D), T[bo, i, i])  # fixed case 39496
ck(f(('bool', 'int80', 'uint176'), D), T[bo, i, i])  # fixed case 39497
ck(f(('bool', 'int80', 'uint184'), D), T[bo, i, i])  # fixed case 39498
ck(f(('bool', 'int80', 'uint192'), D), T[bo, i, i])  # fixed case 39499
ck(f(('bool', 'int80', 'uint200'), D), T[bo, i, i])  # fixed case 39500
ck(f(('bool', 'int80', 'uint208'), D), T[bo, i, i])  # fixed case 39501
ck(f(('bool', 'int80', 'uint216'), D), T[bo, i, i])  # fixed case 39502
ck(f(('bool', 'int80', 'uint224'), D), T[bo, i, i])  # fixed case 39503
ck(f(('bool', 'int80', 'uint232'), D), T[bo, i, i])  # fixed case 39504
ck(f(('bool', 'int80', 'uint24'), D), T[bo, i, i])  # fixed case 39505
ck(f(('bool', 'int80', 'uint240'), D), T[bo, i, i])  # fixed case 39506
ck(f(('bool', 'int80', 'uint248'), D), T[bo, i, i])  # fixed case 39507
ck(f(('bool', 'int80', 'uint256'), D), T[bo, i, i])  # fixed case 39508
ck(f(('bool', 'int80', 'uint32'), D), T[bo, i, i])  # fixed case 39509
ck(f(('bool', 'int80', 'uint40'), D), T[bo, i, i])  # fixed case 39510
ck(f(('bool', 'int80', 'uint48'), D), T[bo, i, i])  # fixed case 39511
ck(f(('bool', 'int80', 'uint56'), D), T[bo, i, i])  # fixed case 39512
ck(f(('bool', 'int80', 'uint64'), D), T[bo, i, i])  # fixed case 39513
ck(f(('bool', 'int80', 'uint72'), D), T[bo, i, i])  # fixed case 39514
ck(f(('bool', 'int80', 'uint8'), D), T[bo, i, i])  # fixed case 39515
ck(f(('bool', 'int80', 'uint80'), D), T[bo, i, i])  # fixed case 39516
ck(f(('bool', 'int80', 'uint88'), D), T[bo, i, i])  # fixed case 39517
ck(f(('bool', 'int80', 'uint96'), D), T[bo, i, i])  # fixed case 39518
ck(f(('bool', 'int80', '?'), D), T[bo, i, A])  # fixed case 39519
ck(f(('bool', 'int88', 'address'), D), T[bo, i, HA])  # fixed case 39520
ck(f(('bool', 'int88', 'address[]'), D), T[bo, i, A])  # fixed case 39521
ck(f(('bool', 'int88', 'bool'), D), T[bo, i, bo])  # fixed case 39522
ck(f(('bool', 'int88', 'bool[]'), D), T[bo, i, A])  # fixed case 39523
ck(f(('bool', 'int88', 'bytes'), D), T[bo, i, by])  # fixed case 39524
ck(f(('bool', 'int88', 'bytes1'), D), T[bo, i, by])  # fixed case 39525
ck(f(('bool', 'int88', 'bytes10'), D), T[bo, i, by])  # fixed case 39526
ck(f(('bool', 'int88', 'bytes11'), D), T[bo, i, by])  # fixed case 39527
ck(f(('bool', 'int88', 'bytes12'), D), T[bo, i, by])  # fixed case 39528
ck(f(('bool', 'int88', 'bytes13'), D), T[bo, i, by])  # fixed case 39529
ck(f(('bool', 'int88', 'bytes14'), D), T[bo, i, by])  # fixed case 39530
ck(f(('bool', 'int88', 'bytes15'), D), T[bo, i, by])  # fixed case 39531
ck(f(('bool', 'int88', 'bytes16'), D), T[bo, i, by])  # fixed case 39532
ck(f(('bool', 'int88', 'bytes17'), D), T[bo, i, by])  # fixed case 39533
ck(f(('bool', 'int88', 'bytes18'), D), T[bo, i, by])  # fixed case 39534
ck(f(('bool', 'int88', 'bytes19'), D), T[bo, i, by])  # fixed case 39535
ck(f(('bool', 'int88', 'bytes2'), D), T[bo, i, by])  # fixed case 39536
ck(f(('bool', 'int88', 'bytes20'), D), T[bo, i, by])  # fixed case 39537
ck(f(('bool', 'int88', 'bytes21'), D), T[bo, i, by])  # fixed case 39538
ck(f(('bool', 'int88', 'bytes22'), D), T[bo, i, by])  # fixed case 39539
ck(f(('bool', 'int88', 'bytes23'), D), T[bo, i, by])  # fixed case 39540
ck(f(('bool', 'int88', 'bytes24'), D), T[bo, i, by])  # fixed case 39541
ck(f(('bool', 'int88', 'bytes25'), D), T[bo, i, by])  # fixed case 39542
ck(f(('bool', 'int88', 'bytes26'), D), T[bo, i, by])  # fixed case 39543
ck(f(('bool', 'int88', 'bytes27'), D), T[bo, i, by])  # fixed case 39544
ck(f(('bool', 'int88', 'bytes28'), D), T[bo, i, by])  # fixed case 39545
ck(f(('bool', 'int88', 'bytes29'), D), T[bo, i, by])  # fixed case 39546
ck(f(('bool', 'int88', 'bytes3'), D), T[bo, i, by])  # fixed case 39547
ck(f(('bool', 'int88', 'bytes30'), D), T[bo, i, by])  # fixed case 39548
ck(f(('bool', 'int88', 'bytes31'), D), T[bo, i, by])  # fixed case 39549
ck(f(('bool', 'int88', 'bytes32'), D), T[bo, i, by])  # fixed case 39550
ck(f(('bool', 'int88', 'bytes4'), D), T[bo, i, by])  # fixed case 39551
ck(f(('bool', 'int88', 'bytes5'), D), T[bo, i, by])  # fixed case 39552
ck(f(('bool', 'int88', 'bytes6'), D), T[bo, i, by])  # fixed case 39553
ck(f(('bool', 'int88', 'bytes7'), D), T[bo, i, by])  # fixed case 39554
ck(f(('bool', 'int88', 'bytes8'), D), T[bo, i, by])  # fixed case 39555
ck(f(('bool', 'int88', 'bytes9'), D), T[bo, i, by])  # fixed case 39556
ck(f(('bool', 'int88', 'int104'), D), T[bo, i, i])  # fixed case 39557
ck(f(('bool', 'int88', 'int112'), D), T[bo, i, i])  # fixed case 39558
ck(f(('bool', 'int88', 'int120'), D), T[bo, i, i])  # fixed case 39559
ck(f(('bool', 'int88', 'int128'), D), T[bo, i, i])  # fixed case 39560
ck(f(('bool', 'int88', 'int136'), D), T[bo, i, i])  # fixed case 39561
ck(f(('bool', 'int88', 'int144'), D), T[bo, i, i])  # fixed case 39562
ck(f(('bool', 'int88', 'int152'), D), T[bo, i, i])  # fixed case 39563
ck(f(('bool', 'int88', 'int16'), D), T[bo, i, i])  # fixed case 39564
ck(f(('bool', 'int88', 'int160'), D), T[bo, i, i])  # fixed case 39565
ck(f(('bool', 'int88', 'int168'), D), T[bo, i, i])  # fixed case 39566
ck(f(('bool', 'int88', 'int176'), D), T[bo, i, i])  # fixed case 39567
ck(f(('bool', 'int88', 'int184'), D), T[bo, i, i])  # fixed case 39568
ck(f(('bool', 'int88', 'int192'), D), T[bo, i, i])  # fixed case 39569
ck(f(('bool', 'int88', 'int200'), D), T[bo, i, i])  # fixed case 39570
ck(f(('bool', 'int88', 'int208'), D), T[bo, i, i])  # fixed case 39571
ck(f(('bool', 'int88', 'int216'), D), T[bo, i, i])  # fixed case 39572
ck(f(('bool', 'int88', 'int224'), D), T[bo, i, i])  # fixed case 39573
ck(f(('bool', 'int88', 'int232'), D), T[bo, i, i])  # fixed case 39574
ck(f(('bool', 'int88', 'int24'), D), T[bo, i, i])  # fixed case 39575
ck(f(('bool', 'int88', 'int240'), D), T[bo, i, i])  # fixed case 39576
ck(f(('bool', 'int88', 'int248'), D), T[bo, i, i])  # fixed case 39577
ck(f(('bool', 'int88', 'int256'), D), T[bo, i, i])  # fixed case 39578
ck(f(('bool', 'int88', 'int32'), D), T[bo, i, i])  # fixed case 39579
ck(f(('bool', 'int88', 'int40'), D), T[bo, i, i])  # fixed case 39580
ck(f(('bool', 'int88', 'int48'), D), T[bo, i, i])  # fixed case 39581
ck(f(('bool', 'int88', 'int56'), D), T[bo, i, i])  # fixed case 39582
ck(f(('bool', 'int88', 'int64'), D), T[bo, i, i])  # fixed case 39583
ck(f(('bool', 'int88', 'int72'), D), T[bo, i, i])  # fixed case 39584
ck(f(('bool', 'int88', 'int8'), D), T[bo, i, i])  # fixed case 39585
ck(f(('bool', 'int88', 'int80'), D), T[bo, i, i])  # fixed case 39586
ck(f(('bool', 'int88', 'int88'), D), T[bo, i, i])  # fixed case 39587
ck(f(('bool', 'int88', 'int96'), D), T[bo, i, i])  # fixed case 39588
ck(f(('bool', 'int88', 'string'), D), T[bo, i, s])  # fixed case 39589
ck(f(('bool', 'int88', 'string[]'), D), T[bo, i, A])  # fixed case 39590
ck(f(('bool', 'int88', 'uint104'), D), T[bo, i, i])  # fixed case 39591
ck(f(('bool', 'int88', 'uint112'), D), T[bo, i, i])  # fixed case 39592
ck(f(('bool', 'int88', 'uint120'), D), T[bo, i, i])  # fixed case 39593
ck(f(('bool', 'int88', 'uint128'), D), T[bo, i, i])  # fixed case 39594
ck(f(('bool', 'int88', 'uint136'), D), T[bo, i, i])  # fixed case 39595
ck(f(('bool', 'int88', 'uint144'), D), T[bo, i, i])  # fixed case 39596
ck(f(('bool', 'int88', 'uint152'), D), T[bo, i, i])  # fixed case 39597
ck(f(('bool', 'int88', 'uint16'), D), T[bo, i, i])  # fixed case 39598
ck(f(('bool', 'int88', 'uint160'), D), T[bo, i, i])  # fixed case 39599
ck(f(('bool', 'int88', 'uint168'), D), T[bo, i, i])  # fixed case 39600
ck(f(('bool', 'int88', 'uint176'), D), T[bo, i, i])  # fixed case 39601
ck(f(('bool', 'int88', 'uint184'), D), T[bo, i, i])  # fixed case 39602
ck(f(('bool', 'int88', 'uint192'), D), T[bo, i, i])  # fixed case 39603
ck(f(('bool', 'int88', 'uint200'), D), T[bo, i, i])  # fixed case 39604
ck(f(('bool', 'int88', 'uint208'), D), T[bo, i, i])  # fixed case 39605
ck(f(('bool', 'int88', 'uint216'), D), T[bo, i, i])  # fixed case 39606
ck(f(('bool', 'int88', 'uint224'), D), T[bo, i, i])  # fixed case 39607
ck(f(('bool', 'int88', 'uint232'), D), T[bo, i, i])  # fixed case 39608
ck(f(('bool', 'int88', 'uint24'), D), T[bo, i, i])  # fixed case 39609
ck(f(('bool', 'int88', 'uint240'), D), T[bo, i, i])  # fixed case 39610
ck(f(('bool', 'int88', 'uint248'), D), T[bo, i, i])  # fixed case 39611
ck(f(('bool', 'int88', 'uint256'), D), T[bo, i, i])  # fixed case 39612
ck(f(('bool', 'int88', 'uint32'), D), T[bo, i, i])  # fixed case 39613
ck(f(('bool', 'int88', 'uint40'), D), T[bo, i, i])  # fixed case 39614
ck(f(('bool', 'int88', 'uint48'), D), T[bo, i, i])  # fixed case 39615
ck(f(('bool', 'int88', 'uint56'), D), T[bo, i, i])  # fixed case 39616
ck(f(('bool', 'int88', 'uint64'), D), T[bo, i, i])  # fixed case 39617
ck(f(('bool', 'int88', 'uint72'), D), T[bo, i, i])  # fixed case 39618
ck(f(('bool', 'int88', 'uint8'), D), T[bo, i, i])  # fixed case 39619
ck(f(('bool', 'int88', 'uint80'), D), T[bo, i, i])  # fixed case 39620
ck(f(('bool', 'int88', 'uint88'), D), T[bo, i, i])  # fixed case 39621
ck(f(('bool', 'int88', 'uint96'), D), T[bo, i, i])  # fixed case 39622
ck(f(('bool', 'int88', '?'), D), T[bo, i, A])  # fixed case 39623
ck(f(('bool', 'int96', 'address'), D), T[bo, i, HA])  # fixed case 39624
ck(f(('bool', 'int96', 'address[]'), D), T[bo, i, A])  # fixed case 39625
ck(f(('bool', 'int96', 'bool'), D), T[bo, i, bo])  # fixed case 39626
ck(f(('bool', 'int96', 'bool[]'), D), T[bo, i, A])  # fixed case 39627
ck(f(('bool', 'int96', 'bytes'), D), T[bo, i, by])  # fixed case 39628
ck(f(('bool', 'int96', 'bytes1'), D), T[bo, i, by])  # fixed case 39629
ck(f(('bool', 'int96', 'bytes10'), D), T[bo, i, by])  # fixed case 39630
ck(f(('bool', 'int96', 'bytes11'), D), T[bo, i, by])  # fixed case 39631
ck(f(('bool', 'int96', 'bytes12'), D), T[bo, i, by])  # fixed case 39632
ck(f(('bool', 'int96', 'bytes13'), D), T[bo, i, by])  # fixed case 39633
ck(f(('bool', 'int96', 'bytes14'), D), T[bo, i, by])  # fixed case 39634
ck(f(('bool', 'int96', 'bytes15'), D), T[bo, i, by])  # fixed case 39635
ck(f(('bool', 'int96', 'bytes16'), D), T[bo, i, by])  # fixed case 39636
ck(f(('bool', 'int96', 'bytes17'), D), T[bo, i, by])  # fixed case 39637
ck(f(('bool', 'int96', 'bytes18'), D), T[bo, i, by])  # fixed case 39638
ck(f(('bool', 'int96', 'bytes19'), D), T[bo, i, by])  # fixed case 39639
ck(f(('bool', 'int96', 'bytes2'), D), T[bo, i, by])  # fixed case 39640
ck(f(('bool', 'int96', 'bytes20'), D), T[bo, i, by])  # fixed case 39641
ck(f(('bool', 'int96', 'bytes21'), D), T[bo, i, by])  # fixed case 39642
ck(f(('bool', 'int96', 'bytes22'), D), T[bo, i, by])  # fixed case 39643
ck(f(('bool', 'int96', 'bytes23'), D), T[bo, i, by])  # fixed case 39644
ck(f(('bool', 'int96', 'bytes24'), D), T[bo, i, by])  # fixed case 39645
ck(f(('bool', 'int96', 'bytes25'), D), T[bo, i, by])  # fixed case 39646
ck(f(('bool', 'int96', 'bytes26'), D), T[bo, i, by])  # fixed case 39647
ck(f(('bool', 'int96', 'bytes27'), D), T[bo, i, by])  # fixed case 39648
ck(f(('bool', 'int96', 'bytes28'), D), T[bo, i, by])  # fixed case 39649
ck(f(('bool', 'int96', 'bytes29'), D), T[bo, i, by])  # fixed case 39650
ck(f(('bool', 'int96', 'bytes3'), D), T[bo, i, by])  # fixed case 39651
ck(f(('bool', 'int96', 'bytes30'), D), T[bo, i, by])  # fixed case 39652
ck(f(('bool', 'int96', 'bytes31'), D), T[bo, i, by])  # fixed case 39653
ck(f(('bool', 'int96', 'bytes32'), D), T[bo, i, by])  # fixed case 39654
ck(f(('bool', 'int96', 'bytes4'), D), T[bo, i, by])  # fixed case 39655
ck(f(('bool', 'int96', 'bytes5'), D), T[bo, i, by])  # fixed case 39656
ck(f(('bool', 'int96', 'bytes6'), D), T[bo, i, by])  # fixed case 39657
ck(f(('bool', 'int96', 'bytes7'), D), T[bo, i, by])  # fixed case 39658
ck(f(('bool', 'int96', 'bytes8'), D), T[bo, i, by])  # fixed case 39659
ck(f(('bool', 'int96', 'bytes9'), D), T[bo, i, by])  # fixed case 39660
ck(f(('bool', 'int96', 'int104'), D), T[bo, i, i])  # fixed case 39661
ck(f(('bool', 'int96', 'int112'), D), T[bo, i, i])  # fixed case 39662
ck(f(('bool', 'int96', 'int120'), D), T[bo, i, i])  # fixed case 39663
ck(f(('bool', 'int96', 'int128'), D), T[bo, i, i])  # fixed case 39664
ck(f(('bool', 'int96', 'int136'), D), T[bo, i, i])  # fixed case 39665
ck(f(('bool', 'int96', 'int144'), D), T[bo, i, i])  # fixed case 39666
ck(f(('bool', 'int96', 'int152'), D), T[bo, i, i])  # fixed case 39667
ck(f(('bool', 'int96', 'int16'), D), T[bo, i, i])  # fixed case 39668
ck(f(('bool', 'int96', 'int160'), D), T[bo, i, i])  # fixed case 39669
ck(f(('bool', 'int96', 'int168'), D), T[bo, i, i])  # fixed case 39670
ck(f(('bool', 'int96', 'int176'), D), T[bo, i, i])  # fixed case 39671
ck(f(('bool', 'int96', 'int184'), D), T[bo, i, i])  # fixed case 39672
ck(f(('bool', 'int96', 'int192'), D), T[bo, i, i])  # fixed case 39673
ck(f(('bool', 'int96', 'int200'), D), T[bo, i, i])  # fixed case 39674
ck(f(('bool', 'int96', 'int208'), D), T[bo, i, i])  # fixed case 39675
ck(f(('bool', 'int96', 'int216'), D), T[bo, i, i])  # fixed case 39676
ck(f(('bool', 'int96', 'int224'), D), T[bo, i, i])  # fixed case 39677
ck(f(('bool', 'int96', 'int232'), D), T[bo, i, i])  # fixed case 39678
ck(f(('bool', 'int96', 'int24'), D), T[bo, i, i])  # fixed case 39679
ck(f(('bool', 'int96', 'int240'), D), T[bo, i, i])  # fixed case 39680
ck(f(('bool', 'int96', 'int248'), D), T[bo, i, i])  # fixed case 39681
ck(f(('bool', 'int96', 'int256'), D), T[bo, i, i])  # fixed case 39682
ck(f(('bool', 'int96', 'int32'), D), T[bo, i, i])  # fixed case 39683
ck(f(('bool', 'int96', 'int40'), D), T[bo, i, i])  # fixed case 39684
ck(f(('bool', 'int96', 'int48'), D), T[bo, i, i])  # fixed case 39685
ck(f(('bool', 'int96', 'int56'), D), T[bo, i, i])  # fixed case 39686
ck(f(('bool', 'int96', 'int64'), D), T[bo, i, i])  # fixed case 39687
ck(f(('bool', 'int96', 'int72'), D), T[bo, i, i])  # fixed case 39688
ck(f(('bool', 'int96', 'int8'), D), T[bo, i, i])  # fixed case 39689
ck(f(('bool', 'int96', 'int80'), D), T[bo, i, i])  # fixed case 39690
ck(f(('bool', 'int96', 'int88'), D), T[bo, i, i])  # fixed case 39691
ck(f(('bool', 'int96', 'int96'), D), T[bo, i, i])  # fixed case 39692
ck(f(('bool', 'int96', 'string'), D), T[bo, i, s])  # fixed case 39693
ck(f(('bool', 'int96', 'string[]'), D), T[bo, i, A])  # fixed case 39694
ck(f(('bool', 'int96', 'uint104'), D), T[bo, i, i])  # fixed case 39695
ck(f(('bool', 'int96', 'uint112'), D), T[bo, i, i])  # fixed case 39696
ck(f(('bool', 'int96', 'uint120'), D), T[bo, i, i])  # fixed case 39697
ck(f(('bool', 'int96', 'uint128'), D), T[bo, i, i])  # fixed case 39698
ck(f(('bool', 'int96', 'uint136'), D), T[bo, i, i])  # fixed case 39699
ck(f(('bool', 'int96', 'uint144'), D), T[bo, i, i])  # fixed case 39700
ck(f(('bool', 'int96', 'uint152'), D), T[bo, i, i])  # fixed case 39701
ck(f(('bool', 'int96', 'uint16'), D), T[bo, i, i])  # fixed case 39702
ck(f(('bool', 'int96', 'uint160'), D), T[bo, i, i])  # fixed case 39703
ck(f(('bool', 'int96', 'uint168'), D), T[bo, i, i])  # fixed case 39704
ck(f(('bool', 'int96', 'uint176'), D), T[bo, i, i])  # fixed case 39705
ck(f(('bool', 'int96', 'uint184'), D), T[bo, i, i])  # fixed case 39706
ck(f(('bool', 'int96', 'uint192'), D), T[bo, i, i])  # fixed case 39707
ck(f(('bool', 'int96', 'uint200'), D), T[bo, i, i])  # fixed case 39708
ck(f(('bool', 'int96', 'uint208'), D), T[bo, i, i])  # fixed case 39709
ck(f(('bool', 'int96', 'uint216'), D), T[bo, i, i])  # fixed case 39710
ck(f(('bool', 'int96', 'uint224'), D), T[bo, i, i])  # fixed case 39711
ck(f(('bool', 'int96', 'uint232'), D), T[bo, i, i])  # fixed case 39712
ck(f(('bool', 'int96', 'uint24'), D), T[bo, i, i])  # fixed case 39713
ck(f(('bool', 'int96', 'uint240'), D), T[bo, i, i])  # fixed case 39714
ck(f(('bool', 'int96', 'uint248'), D), T[bo, i, i])  # fixed case 39715
ck(f(('bool', 'int96', 'uint256'), D), T[bo, i, i])  # fixed case 39716
ck(f(('bool', 'int96', 'uint32'), D), T[bo, i, i])  # fixed case 39717
ck(f(('bool', 'int96', 'uint40'), D), T[bo, i, i])  # fixed case 39718
ck(f(('bool', 'int96', 'uint48'), D), T[bo, i, i])  # fixed case 39719
ck(f(('bool', 'int96', 'uint56'), D), T[bo, i, i])  # fixed case 39720
ck(f(('bool', 'int96', 'uint64'), D), T[bo, i, i])  # fixed case 39721
ck(f(('bool', 'int96', 'uint72'), D), T[bo, i, i])  # fixed case 39722
ck(f(('bool', 'int96', 'uint8'), D), T[bo, i, i])  # fixed case 39723
ck(f(('bool', 'int96', 'uint80'), D), T[bo, i, i])  # fixed case 39724
ck(f(('bool', 'int96', 'uint88'), D), T[bo, i, i])  # fixed case 39725
ck(f(('bool', 'int96', 'uint96'), D), T[bo, i, i])  # fixed case 39726
ck(f(('bool', 'int96', '?'), D), T[bo, i, A])  # fixed case 39727
ck(f(('bool', 'string', 'address'), D), T[bo, s, HA])  # fixed case 39728
ck(f(('bool', 'string', 'address[]'), D), T[bo, s, A])  # fixed case 39729
ck(f(('bool', 'string', 'bool'), D), T[bo, s, bo])  # fixed case 39730
ck(f(('bool', 'string', 'bool[]'), D), T[bo, s, A])  # fixed case 39731
ck(f(('bool', 'string', 'bytes'), D), T[bo, s, by])  # fixed case 39732
ck(f(('bool', 'string', 'bytes1'), D), T[bo, s, by])  # fixed case 39733
ck(f(('bool', 'string', 'bytes10'), D), T[bo, s, by])  # fixed case 39734
ck(f(('bool', 'string', 'bytes11'), D), T[bo, s, by])  # fixed case 39735
ck(f(('bool', 'string', 'bytes12'), D), T[bo, s, by])  # fixed case 39736
ck(f(('bool', 'string', 'bytes13'), D), T[bo, s, by])  # fixed case 39737
ck(f(('bool', 'string', 'bytes14'), D), T[bo, s, by])  # fixed case 39738
ck(f(('bool', 'string', 'bytes15'), D), T[bo, s, by])  # fixed case 39739
ck(f(('bool', 'string', 'bytes16'), D), T[bo, s, by])  # fixed case 39740
ck(f(('bool', 'string', 'bytes17'), D), T[bo, s, by])  # fixed case 39741
ck(f(('bool', 'string', 'bytes18'), D), T[bo, s, by])  # fixed case 39742
ck(f(('bool', 'string', 'bytes19'), D), T[bo, s, by])  # fixed case 39743
ck(f(('bool', 'string', 'bytes2'), D), T[bo, s, by])  # fixed case 39744
ck(f(('bool', 'string', 'bytes20'), D), T[bo, s, by])  # fixed case 39745
ck(f(('bool', 'string', 'bytes21'), D), T[bo, s, by])  # fixed case 39746
ck(f(('bool', 'string', 'bytes22'), D), T[bo, s, by])  # fixed case 39747
ck(f(('bool', 'string', 'bytes23'), D), T[bo, s, by])  # fixed case 39748
ck(f(('bool', 'string', 'bytes24'), D), T[bo, s, by])  # fixed case 39749
ck(f(('bool', 'string', 'bytes25'), D), T[bo, s, by])  # fixed case 39750
ck(f(('bool', 'string', 'bytes26'), D), T[bo, s, by])  # fixed case 39751
ck(f(('bool', 'string', 'bytes27'), D), T[bo, s, by])  # fixed case 39752
ck(f(('bool', 'string', 'bytes28'), D), T[bo, s, by])  # fixed case 39753
ck(f(('bool', 'string', 'bytes29'), D), T[bo, s, by])  # fixed case 39754
ck(f(('bool', 'string', 'bytes3'), D), T[bo, s, by])  # fixed case 39755
ck(f(('bool', 'string', 'bytes30'), D), T[bo, s, by])  # fixed case 39756
ck(f(('bool', 'string', 'bytes31'), D), T[bo, s, by])  # fixed case 39757
ck(f(('bool', 'string', 'bytes32'), D), T[bo, s, by])  # fixed case 39758
ck(f(('bool', 'string', 'bytes4'), D), T[bo, s, by])  # fixed case 39759
ck(f(('bool', 'string', 'bytes5'), D), T[bo, s, by])  # fixed case 39760
ck(f(('bool', 'string', 'bytes6'), D), T[bo, s, by])  # fixed case 39761
ck(f(('bool', 'string', 'bytes7'), D), T[bo, s, by])  # fixed case 39762
ck(f(('bool', 'string', 'bytes8'), D), T[bo, s, by])  # fixed case 39763
ck(f(('bool', 'string', 'bytes9'), D), T[bo, s, by])  # fixed case 39764
ck(f(('bool', 'string', 'int104'), D), T[bo, s, i])  # fixed case 39765
ck(f(('bool', 'string', 'int112'), D), T[bo, s, i])  # fixed case 39766
ck(f(('bool', 'string', 'int120'), D), T[bo, s, i])  # fixed case 39767
ck(f(('bool', 'string', 'int128'), D), T[bo, s, i])  # fixed case 39768
ck(f(('bool', 'string', 'int136'), D), T[bo, s, i])  # fixed case 39769
ck(f(('bool', 'string', 'int144'), D), T[bo, s, i])  # fixed case 39770
ck(f(('bool', 'string', 'int152'), D), T[bo, s, i])  # fixed case 39771
ck(f(('bool', 'string', 'int16'), D), T[bo, s, i])  # fixed case 39772
ck(f(('bool', 'string', 'int160'), D), T[bo, s, i])  # fixed case 39773
ck(f(('bool', 'string', 'int168'), D), T[bo, s, i])  # fixed case 39774
ck(f(('bool', 'string', 'int176'), D), T[bo, s, i])  # fixed case 39775
ck(f(('bool', 'string', 'int184'), D), T[bo, s, i])  # fixed case 39776
ck(f(('bool', 'string', 'int192'), D), T[bo, s, i])  # fixed case 39777
ck(f(('bool', 'string', 'int200'), D), T[bo, s, i])  # fixed case 39778
ck(f(('bool', 'string', 'int208'), D), T[bo, s, i])  # fixed case 39779
ck(f(('bool', 'string', 'int216'), D), T[bo, s, i])  # fixed case 39780
ck(f(('bool', 'string', 'int224'), D), T[bo, s, i])  # fixed case 39781
ck(f(('bool', 'string', 'int232'), D), T[bo, s, i])  # fixed case 39782
ck(f(('bool', 'string', 'int24'), D), T[bo, s, i])  # fixed case 39783
ck(f(('bool', 'string', 'int240'), D), T[bo, s, i])  # fixed case 39784
ck(f(('bool', 'string', 'int248'), D), T[bo, s, i])  # fixed case 39785
ck(f(('bool', 'string', 'int256'), D), T[bo, s, i])  # fixed case 39786
ck(f(('bool', 'string', 'int32'), D), T[bo, s, i])  # fixed case 39787
ck(f(('bool', 'string', 'int40'), D), T[bo, s, i])  # fixed case 39788
ck(f(('bool', 'string', 'int48'), D), T[bo, s, i])  # fixed case 39789
ck(f(('bool', 'string', 'int56'), D), T[bo, s, i])  # fixed case 39790
ck(f(('bool', 'string', 'int64'), D), T[bo, s, i])  # fixed case 39791
ck(f(('bool', 'string', 'int72'), D), T[bo, s, i])  # fixed case 39792
ck(f(('bool', 'string', 'int8'), D), T[bo, s, i])  # fixed case 39793
ck(f(('bool', 'string', 'int80'), D), T[bo, s, i])  # fixed case 39794
ck(f(('bool', 'string', 'int88'), D), T[bo, s, i])  # fixed case 39795
ck(f(('bool', 'string', 'int96'), D), T[bo, s, i])  # fixed case 39796
ck(f(('bool', 'string', 'string'), D), T[bo, s, s])  # fixed case 39797
ck(f(('bool', 'string', 'string[]'), D), T[bo, s, A])  # fixed case 39798
ck(f(('bool', 'string', 'uint104'), D), T[bo, s, i])  # fixed case 39799
ck(f(('bool', 'string', 'uint112'), D), T[bo, s, i])  # fixed case 39800
ck(f(('bool', 'string', 'uint120'), D), T[bo, s, i])  # fixed case 39801
ck(f(('bool', 'string', 'uint128'), D), T[bo, s, i])  # fixed case 39802
ck(f(('bool', 'string', 'uint136'), D), T[bo, s, i])  # fixed case 39803
ck(f(('bool', 'string', 'uint144'), D), T[bo, s, i])  # fixed case 39804
ck(f(('bool', 'string', 'uint152'), D), T[bo, s, i])  # fixed case 39805
ck(f(('bool', 'string', 'uint16'), D), T[bo, s, i])  # fixed case 39806
ck(f(('bool', 'string', 'uint160'), D), T[bo, s, i])  # fixed case 39807
ck(f(('bool', 'string', 'uint168'), D), T[bo, s, i])  # fixed case 39808
ck(f(('bool', 'string', 'uint176'), D), T[bo, s, i])  # fixed case 39809
ck(f(('bool', 'string', 'uint184'), D), T[bo, s, i])  # fixed case 39810
ck(f(('bool', 'string', 'uint192'), D), T[bo, s, i])  # fixed case 39811
ck(f(('bool', 'string', 'uint200'), D), T[bo, s, i])  # fixed case 39812
ck(f(('bool', 'string', 'uint208'), D), T[bo, s, i])  # fixed case 39813
ck(f(('bool', 'string', 'uint216'), D), T[bo, s, i])  # fixed case 39814
ck(f(('bool', 'string', 'uint224'), D), T[bo, s, i])  # fixed case 39815
ck(f(('bool', 'string', 'uint232'), D), T[bo, s, i])  # fixed case 39816
ck(f(('bool', 'string', 'uint24'), D), T[bo, s, i])  # fixed case 39817
ck(f(('bool', 'string', 'uint240'), D), T[bo, s, i])  # fixed case 39818
ck(f(('bool', 'string', 'uint248'), D), T[bo, s, i])  # fixed case 39819
ck(f(('bool', 'string', 'uint256'), D), T[bo, s, i])  # fixed case 39820
ck(f(('bool', 'string', 'uint32'), D), T[bo, s, i])  # fixed case 39821
ck(f(('bool', 'string', 'uint40'), D), T[bo, s, i])  # fixed case 39822
ck(f(('bool', 'string', 'uint48'), D), T[bo, s, i])  # fixed case 39823
ck(f(('bool', 'string', 'uint56'), D), T[bo, s, i])  # fixed case 39824
ck(f(('bool', 'string', 'uint64'), D), T[bo, s, i])  # fixed case 39825
ck(f(('bool', 'string', 'uint72'), D), T[bo, s, i])  # fixed case 39826
ck(f(('bool', 'string', 'uint8'), D), T[bo, s, i])  # fixed case 39827
ck(f(('bool', 'string', 'uint80'), D), T[bo, s, i])  # fixed case 39828
ck(f(('bool', 'string', 'uint88'), D), T[bo, s, i])  # fixed case 39829
ck(f(('bool', 'string', 'uint96'), D), T[bo, s, i])  # fixed case 39830
ck(f(('bool', 'string', '?'), D), T[bo, s, A])  # fixed case 39831
ck(f(('bool', 'string[]', 'address'), D), T[bo, A, HA])  # fixed case 39832
ck(f(('bool', 'string[]', 'address[]'), D), T[bo, A, A])  # fixed case 39833
ck(f(('bool', 'string[]', 'bool'), D), T[bo, A, bo])  # fixed case 39834
ck(f(('bool', 'string[]', 'bool[]'), D), T[bo, A, A])  # fixed case 39835
ck(f(('bool', 'string[]', 'bytes'), D), T[bo, A, by])  # fixed case 39836
ck(f(('bool', 'string[]', 'bytes1'), D), T[bo, A, by])  # fixed case 39837
ck(f(('bool', 'string[]', 'bytes10'), D), T[bo, A, by])  # fixed case 39838
ck(f(('bool', 'string[]', 'bytes11'), D), T[bo, A, by])  # fixed case 39839
ck(f(('bool', 'string[]', 'bytes12'), D), T[bo, A, by])  # fixed case 39840
ck(f(('bool', 'string[]', 'bytes13'), D), T[bo, A, by])  # fixed case 39841
ck(f(('bool', 'string[]', 'bytes14'), D), T[bo, A, by])  # fixed case 39842
ck(f(('bool', 'string[]', 'bytes15'), D), T[bo, A, by])  # fixed case 39843
ck(f(('bool', 'string[]', 'bytes16'), D), T[bo, A, by])  # fixed case 39844
ck(f(('bool', 'string[]', 'bytes17'), D), T[bo, A, by])  # fixed case 39845
ck(f(('bool', 'string[]', 'bytes18'), D), T[bo, A, by])  # fixed case 39846
ck(f(('bool', 'string[]', 'bytes19'), D), T[bo, A, by])  # fixed case 39847
ck(f(('bool', 'string[]', 'bytes2'), D), T[bo, A, by])  # fixed case 39848
ck(f(('bool', 'string[]', 'bytes20'), D), T[bo, A, by])  # fixed case 39849
ck(f(('bool', 'string[]', 'bytes21'), D), T[bo, A, by])  # fixed case 39850
ck(f(('bool', 'string[]', 'bytes22'), D), T[bo, A, by])  # fixed case 39851
ck(f(('bool', 'string[]', 'bytes23'), D), T[bo, A, by])  # fixed case 39852
ck(f(('bool', 'string[]', 'bytes24'), D), T[bo, A, by])  # fixed case 39853
ck(f(('bool', 'string[]', 'bytes25'), D), T[bo, A, by])  # fixed case 39854
ck(f(('bool', 'string[]', 'bytes26'), D), T[bo, A, by])  # fixed case 39855
ck(f(('bool', 'string[]', 'bytes27'), D), T[bo, A, by])  # fixed case 39856
ck(f(('bool', 'string[]', 'bytes28'), D), T[bo, A, by])  # fixed case 39857
ck(f(('bool', 'string[]', 'bytes29'), D), T[bo, A, by])  # fixed case 39858
ck(f(('bool', 'string[]', 'bytes3'), D), T[bo, A, by])  # fixed case 39859
ck(f(('bool', 'string[]', 'bytes30'), D), T[bo, A, by])  # fixed case 39860
ck(f(('bool', 'string[]', 'bytes31'), D), T[bo, A, by])  # fixed case 39861
ck(f(('bool', 'string[]', 'bytes32'), D), T[bo, A, by])  # fixed case 39862
ck(f(('bool', 'string[]', 'bytes4'), D), T[bo, A, by])  # fixed case 39863
ck(f(('bool', 'string[]', 'bytes5'), D), T[bo, A, by])  # fixed case 39864
ck(f(('bool', 'string[]', 'bytes6'), D), T[bo, A, by])  # fixed case 39865
ck(f(('bool', 'string[]', 'bytes7'), D), T[bo, A, by])  # fixed case 39866
ck(f(('bool', 'string[]', 'bytes8'), D), T[bo, A, by])  # fixed case 39867
ck(f(('bool', 'string[]', 'bytes9'), D), T[bo, A, by])  # fixed case 39868
ck(f(('bool', 'string[]', 'int104'), D), T[bo, A, i])  # fixed case 39869
ck(f(('bool', 'string[]', 'int112'), D), T[bo, A, i])  # fixed case 39870
ck(f(('bool', 'string[]', 'int120'), D), T[bo, A, i])  # fixed case 39871
ck(f(('bool', 'string[]', 'int128'), D), T[bo, A, i])  # fixed case 39872
ck(f(('bool', 'string[]', 'int136'), D), T[bo, A, i])  # fixed case 39873
ck(f(('bool', 'string[]', 'int144'), D), T[bo, A, i])  # fixed case 39874
ck(f(('bool', 'string[]', 'int152'), D), T[bo, A, i])  # fixed case 39875
ck(f(('bool', 'string[]', 'int16'), D), T[bo, A, i])  # fixed case 39876
ck(f(('bool', 'string[]', 'int160'), D), T[bo, A, i])  # fixed case 39877
ck(f(('bool', 'string[]', 'int168'), D), T[bo, A, i])  # fixed case 39878
ck(f(('bool', 'string[]', 'int176'), D), T[bo, A, i])  # fixed case 39879
ck(f(('bool', 'string[]', 'int184'), D), T[bo, A, i])  # fixed case 39880
ck(f(('bool', 'string[]', 'int192'), D), T[bo, A, i])  # fixed case 39881
ck(f(('bool', 'string[]', 'int200'), D), T[bo, A, i])  # fixed case 39882
ck(f(('bool', 'string[]', 'int208'), D), T[bo, A, i])  # fixed case 39883
ck(f(('bool', 'string[]', 'int216'), D), T[bo, A, i])  # fixed case 39884
ck(f(('bool', 'string[]', 'int224'), D), T[bo, A, i])  # fixed case 39885
ck(f(('bool', 'string[]', 'int232'), D), T[bo, A, i])  # fixed case 39886
ck(f(('bool', 'string[]', 'int24'), D), T[bo, A, i])  # fixed case 39887
ck(f(('bool', 'string[]', 'int240'), D), T[bo, A, i])  # fixed case 39888
ck(f(('bool', 'string[]', 'int248'), D), T[bo, A, i])  # fixed case 39889
ck(f(('bool', 'string[]', 'int256'), D), T[bo, A, i])  # fixed case 39890
ck(f(('bool', 'string[]', 'int32'), D), T[bo, A, i])  # fixed case 39891
ck(f(('bool', 'string[]', 'int40'), D), T[bo, A, i])  # fixed case 39892
ck(f(('bool', 'string[]', 'int48'), D), T[bo, A, i])  # fixed case 39893
ck(f(('bool', 'string[]', 'int56'), D), T[bo, A, i])  # fixed case 39894
ck(f(('bool', 'string[]', 'int64'), D), T[bo, A, i])  # fixed case 39895
ck(f(('bool', 'string[]', 'int72'), D), T[bo, A, i])  # fixed case 39896
ck(f(('bool', 'string[]', 'int8'), D), T[bo, A, i])  # fixed case 39897
ck(f(('bool', 'string[]', 'int80'), D), T[bo, A, i])  # fixed case 39898
ck(f(('bool', 'string[]', 'int88'), D), T[bo, A, i])  # fixed case 39899
ck(f(('bool', 'string[]', 'int96'), D), T[bo, A, i])  # fixed case 39900
ck(f(('bool', 'string[]', 'string'), D), T[bo, A, s])  # fixed case 39901
ck(f(('bool', 'string[]', 'string[]'), D), T[bo, A, A])  # fixed case 39902
ck(f(('bool', 'string[]', 'uint104'), D), T[bo, A, i])  # fixed case 39903
ck(f(('bool', 'string[]', 'uint112'), D), T[bo, A, i])  # fixed case 39904
ck(f(('bool', 'string[]', 'uint120'), D), T[bo, A, i])  # fixed case 39905
ck(f(('bool', 'string[]', 'uint128'), D), T[bo, A, i])  # fixed case 39906
ck(f(('bool', 'string[]', 'uint136'), D), T[bo, A, i])  # fixed case 39907
ck(f(('bool', 'string[]', 'uint144'), D), T[bo, A, i])  # fixed case 39908
ck(f(('bool', 'string[]', 'uint152'), D), T[bo, A, i])  # fixed case 39909
ck(f(('bool', 'string[]', 'uint16'), D), T[bo, A, i])  # fixed case 39910
ck(f(('bool', 'string[]', 'uint160'), D), T[bo, A, i])  # fixed case 39911
ck(f(('bool', 'string[]', 'uint168'), D), T[bo, A, i])  # fixed case 39912
ck(f(('bool', 'string[]', 'uint176'), D), T[bo, A, i])  # fixed case 39913
ck(f(('bool', 'string[]', 'uint184'), D), T[bo, A, i])  # fixed case 39914
ck(f(('bool', 'string[]', 'uint192'), D), T[bo, A, i])  # fixed case 39915
ck(f(('bool', 'string[]', 'uint200'), D), T[bo, A, i])  # fixed case 39916
ck(f(('bool', 'string[]', 'uint208'), D), T[bo, A, i])  # fixed case 39917
ck(f(('bool', 'string[]', 'uint216'), D), T[bo, A, i])  # fixed case 39918
ck(f(('bool', 'string[]', 'uint224'), D), T[bo, A, i])  # fixed case 39919
ck(f(('bool', 'string[]', 'uint232'), D), T[bo, A, i])  # fixed case 39920
ck(f(('bool', 'string[]', 'uint24'), D), T[bo, A, i])  # fixed case 39921
ck(f(('bool', 'string[]', 'uint240'), D), T[bo, A, i])  # fixed case 39922
ck(f(('bool', 'string[]', 'uint248'), D), T[bo, A, i])  # fixed case 39923
ck(f(('bool', 'string[]', 'uint256'), D), T[bo, A, i])  # fixed case 39924
ck(f(('bool', 'string[]', 'uint32'), D), T[bo, A, i])  # fixed case 39925
ck(f(('bool', 'string[]', 'uint40'), D), T[bo, A, i])  # fixed case 39926
ck(f(('bool', 'string[]', 'uint48'), D), T[bo, A, i])  # fixed case 39927
ck(f(('bool', 'string[]', 'uint56'), D), T[bo, A, i])  # fixed case 39928
ck(f(('bool', 'string[]', 'uint64'), D), T[bo, A, i])  # fixed case 39929
ck(f(('bool', 'string[]', 'uint72'), D), T[bo, A, i])  # fixed case 39930
ck(f(('bool', 'string[]', 'uint8'), D), T[bo, A, i])  # fixed case 39931
ck(f(('bool', 'string[]', 'uint80'), D), T[bo, A, i])  # fixed case 39932
ck(f(('bool', 'string[]', 'uint88'), D), T[bo, A, i])  # fixed case 39933
ck(f(('bool', 'string[]', 'uint96'), D), T[bo, A, i])  # fixed case 39934
ck(f(('bool', 'string[]', '?'), D), T[bo, A, A])  # fixed case 39935
ck(f(('bool', 'uint104', 'address'), D), T[bo, i, HA])  # fixed case 39936
ck(f(('bool', 'uint104', 'address[]'), D), T[bo, i, A])  # fixed case 39937
ck(f(('bool', 'uint104', 'bool'), D), T[bo, i, bo])  # fixed case 39938
ck(f(('bool', 'uint104', 'bool[]'), D), T[bo, i, A])  # fixed case 39939
ck(f(('bool', 'uint104', 'bytes'), D), T[bo, i, by])  # fixed case 39940
ck(f(('bool', 'uint104', 'bytes1'), D), T[bo, i, by])  # fixed case 39941
ck(f(('bool', 'uint104', 'bytes10'), D), T[bo, i, by])  # fixed case 39942
ck(f(('bool', 'uint104', 'bytes11'), D), T[bo, i, by])  # fixed case 39943
ck(f(('bool', 'uint104', 'bytes12'), D), T[bo, i, by])  # fixed case 39944
ck(f(('bool', 'uint104', 'bytes13'), D), T[bo, i, by])  # fixed case 39945
ck(f(('bool', 'uint104', 'bytes14'), D), T[bo, i, by])  # fixed case 39946
ck(f(('bool', 'uint104', 'bytes15'), D), T[bo, i, by])  # fixed case 39947
ck(f(('bool', 'uint104', 'bytes16'), D), T[bo, i, by])  # fixed case 39948
ck(f(('bool', 'uint104', 'bytes17'), D), T[bo, i, by])  # fixed case 39949
ck(f(('bool', 'uint104', 'bytes18'), D), T[bo, i, by])  # fixed case 39950
ck(f(('bool', 'uint104', 'bytes19'), D), T[bo, i, by])  # fixed case 39951
ck(f(('bool', 'uint104', 'bytes2'), D), T[bo, i, by])  # fixed case 39952
ck(f(('bool', 'uint104', 'bytes20'), D), T[bo, i, by])  # fixed case 39953
ck(f(('bool', 'uint104', 'bytes21'), D), T[bo, i, by])  # fixed case 39954
ck(f(('bool', 'uint104', 'bytes22'), D), T[bo, i, by])  # fixed case 39955
ck(f(('bool', 'uint104', 'bytes23'), D), T[bo, i, by])  # fixed case 39956
ck(f(('bool', 'uint104', 'bytes24'), D), T[bo, i, by])  # fixed case 39957
ck(f(('bool', 'uint104', 'bytes25'), D), T[bo, i, by])  # fixed case 39958
ck(f(('bool', 'uint104', 'bytes26'), D), T[bo, i, by])  # fixed case 39959
ck(f(('bool', 'uint104', 'bytes27'), D), T[bo, i, by])  # fixed case 39960
ck(f(('bool', 'uint104', 'bytes28'), D), T[bo, i, by])  # fixed case 39961
ck(f(('bool', 'uint104', 'bytes29'), D), T[bo, i, by])  # fixed case 39962
ck(f(('bool', 'uint104', 'bytes3'), D), T[bo, i, by])  # fixed case 39963
ck(f(('bool', 'uint104', 'bytes30'), D), T[bo, i, by])  # fixed case 39964
ck(f(('bool', 'uint104', 'bytes31'), D), T[bo, i, by])  # fixed case 39965
ck(f(('bool', 'uint104', 'bytes32'), D), T[bo, i, by])  # fixed case 39966
ck(f(('bool', 'uint104', 'bytes4'), D), T[bo, i, by])  # fixed case 39967
ck(f(('bool', 'uint104', 'bytes5'), D), T[bo, i, by])  # fixed case 39968
ck(f(('bool', 'uint104', 'bytes6'), D), T[bo, i, by])  # fixed case 39969
ck(f(('bool', 'uint104', 'bytes7'), D), T[bo, i, by])  # fixed case 39970
ck(f(('bool', 'uint104', 'bytes8'), D), T[bo, i, by])  # fixed case 39971
ck(f(('bool', 'uint104', 'bytes9'), D), T[bo, i, by])  # fixed case 39972
ck(f(('bool', 'uint104', 'int104'), D), T[bo, i, i])  # fixed case 39973
ck(f(('bool', 'uint104', 'int112'), D), T[bo, i, i])  # fixed case 39974
ck(f(('bool', 'uint104', 'int120'), D), T[bo, i, i])  # fixed case 39975
ck(f(('bool', 'uint104', 'int128'), D), T[bo, i, i])  # fixed case 39976
ck(f(('bool', 'uint104', 'int136'), D), T[bo, i, i])  # fixed case 39977
ck(f(('bool', 'uint104', 'int144'), D), T[bo, i, i])  # fixed case 39978
ck(f(('bool', 'uint104', 'int152'), D), T[bo, i, i])  # fixed case 39979
ck(f(('bool', 'uint104', 'int16'), D), T[bo, i, i])  # fixed case 39980
ck(f(('bool', 'uint104', 'int160'), D), T[bo, i, i])  # fixed case 39981
ck(f(('bool', 'uint104', 'int168'), D), T[bo, i, i])  # fixed case 39982
ck(f(('bool', 'uint104', 'int176'), D), T[bo, i, i])  # fixed case 39983
ck(f(('bool', 'uint104', 'int184'), D), T[bo, i, i])  # fixed case 39984
ck(f(('bool', 'uint104', 'int192'), D), T[bo, i, i])  # fixed case 39985
ck(f(('bool', 'uint104', 'int200'), D), T[bo, i, i])  # fixed case 39986
ck(f(('bool', 'uint104', 'int208'), D), T[bo, i, i])  # fixed case 39987
ck(f(('bool', 'uint104', 'int216'), D), T[bo, i, i])  # fixed case 39988
ck(f(('bool', 'uint104', 'int224'), D), T[bo, i, i])  # fixed case 39989
ck(f(('bool', 'uint104', 'int232'), D), T[bo, i, i])  # fixed case 39990
ck(f(('bool', 'uint104', 'int24'), D), T[bo, i, i])  # fixed case 39991
ck(f(('bool', 'uint104', 'int240'), D), T[bo, i, i])  # fixed case 39992
ck(f(('bool', 'uint104', 'int248'), D), T[bo, i, i])  # fixed case 39993
ck(f(('bool', 'uint104', 'int256'), D), T[bo, i, i])  # fixed case 39994
ck(f(('bool', 'uint104', 'int32'), D), T[bo, i, i])  # fixed case 39995
ck(f(('bool', 'uint104', 'int40'), D), T[bo, i, i])  # fixed case 39996
ck(f(('bool', 'uint104', 'int48'), D), T[bo, i, i])  # fixed case 39997
ck(f(('bool', 'uint104', 'int56'), D), T[bo, i, i])  # fixed case 39998
ck(f(('bool', 'uint104', 'int64'), D), T[bo, i, i])  # fixed case 39999
ck(f(('bool', 'uint104', 'int72'), D), T[bo, i, i])  # fixed case 40000
ck(f(('bool', 'uint104', 'int8'), D), T[bo, i, i])  # fixed case 40001
ck(f(('bool', 'uint104', 'int80'), D), T[bo, i, i])  # fixed case 40002
ck(f(('bool', 'uint104', 'int88'), D), T[bo, i, i])  # fixed case 40003
ck(f(('bool', 'uint104', 'int96'), D), T[bo, i, i])  # fixed case 40004
ck(f(('bool', 'uint104', 'string'), D), T[bo, i, s])  # fixed case 40005
ck(f(('bool', 'uint104', 'string[]'), D), T[bo, i, A])  # fixed case 40006
ck(f(('bool', 'uint104', 'uint104'), D), T[bo, i, i])  # fixed case 40007
ck(f(('bool', 'uint104', 'uint112'), D), T[bo, i, i])  # fixed case 40008
ck(f(('bool', 'uint104', 'uint120'), D), T[bo, i, i])  # fixed case 40009
ck(f(('bool', 'uint104', 'uint128'), D), T[bo, i, i])  # fixed case 40010
ck(f(('bool', 'uint104', 'uint136'), D), T[bo, i, i])  # fixed case 40011
ck(f(('bool', 'uint104', 'uint144'), D), T[bo, i, i])  # fixed case 40012
ck(f(('bool', 'uint104', 'uint152'), D), T[bo, i, i])  # fixed case 40013
ck(f(('bool', 'uint104', 'uint16'), D), T[bo, i, i])  # fixed case 40014
ck(f(('bool', 'uint104', 'uint160'), D), T[bo, i, i])  # fixed case 40015
ck(f(('bool', 'uint104', 'uint168'), D), T[bo, i, i])  # fixed case 40016
ck(f(('bool', 'uint104', 'uint176'), D), T[bo, i, i])  # fixed case 40017
ck(f(('bool', 'uint104', 'uint184'), D), T[bo, i, i])  # fixed case 40018
ck(f(('bool', 'uint104', 'uint192'), D), T[bo, i, i])  # fixed case 40019
ck(f(('bool', 'uint104', 'uint200'), D), T[bo, i, i])  # fixed case 40020
ck(f(('bool', 'uint104', 'uint208'), D), T[bo, i, i])  # fixed case 40021
ck(f(('bool', 'uint104', 'uint216'), D), T[bo, i, i])  # fixed case 40022
ck(f(('bool', 'uint104', 'uint224'), D), T[bo, i, i])  # fixed case 40023
ck(f(('bool', 'uint104', 'uint232'), D), T[bo, i, i])  # fixed case 40024
ck(f(('bool', 'uint104', 'uint24'), D), T[bo, i, i])  # fixed case 40025
ck(f(('bool', 'uint104', 'uint240'), D), T[bo, i, i])  # fixed case 40026
ck(f(('bool', 'uint104', 'uint248'), D), T[bo, i, i])  # fixed case 40027
ck(f(('bool', 'uint104', 'uint256'), D), T[bo, i, i])  # fixed case 40028
ck(f(('bool', 'uint104', 'uint32'), D), T[bo, i, i])  # fixed case 40029
ck(f(('bool', 'uint104', 'uint40'), D), T[bo, i, i])  # fixed case 40030
ck(f(('bool', 'uint104', 'uint48'), D), T[bo, i, i])  # fixed case 40031
ck(f(('bool', 'uint104', 'uint56'), D), T[bo, i, i])  # fixed case 40032
ck(f(('bool', 'uint104', 'uint64'), D), T[bo, i, i])  # fixed case 40033
ck(f(('bool', 'uint104', 'uint72'), D), T[bo, i, i])  # fixed case 40034
ck(f(('bool', 'uint104', 'uint8'), D), T[bo, i, i])  # fixed case 40035
ck(f(('bool', 'uint104', 'uint80'), D), T[bo, i, i])  # fixed case 40036
ck(f(('bool', 'uint104', 'uint88'), D), T[bo, i, i])  # fixed case 40037
ck(f(('bool', 'uint104', 'uint96'), D), T[bo, i, i])  # fixed case 40038
ck(f(('bool', 'uint104', '?'), D), T[bo, i, A])  # fixed case 40039
ck(f(('bool', 'uint112', 'address'), D), T[bo, i, HA])  # fixed case 40040
ck(f(('bool', 'uint112', 'address[]'), D), T[bo, i, A])  # fixed case 40041
ck(f(('bool', 'uint112', 'bool'), D), T[bo, i, bo])  # fixed case 40042
ck(f(('bool', 'uint112', 'bool[]'), D), T[bo, i, A])  # fixed case 40043
ck(f(('bool', 'uint112', 'bytes'), D), T[bo, i, by])  # fixed case 40044
ck(f(('bool', 'uint112', 'bytes1'), D), T[bo, i, by])  # fixed case 40045
ck(f(('bool', 'uint112', 'bytes10'), D), T[bo, i, by])  # fixed case 40046
ck(f(('bool', 'uint112', 'bytes11'), D), T[bo, i, by])  # fixed case 40047
ck(f(('bool', 'uint112', 'bytes12'), D), T[bo, i, by])  # fixed case 40048
ck(f(('bool', 'uint112', 'bytes13'), D), T[bo, i, by])  # fixed case 40049
ck(f(('bool', 'uint112', 'bytes14'), D), T[bo, i, by])  # fixed case 40050
ck(f(('bool', 'uint112', 'bytes15'), D), T[bo, i, by])  # fixed case 40051
ck(f(('bool', 'uint112', 'bytes16'), D), T[bo, i, by])  # fixed case 40052
ck(f(('bool', 'uint112', 'bytes17'), D), T[bo, i, by])  # fixed case 40053
ck(f(('bool', 'uint112', 'bytes18'), D), T[bo, i, by])  # fixed case 40054
ck(f(('bool', 'uint112', 'bytes19'), D), T[bo, i, by])  # fixed case 40055
ck(f(('bool', 'uint112', 'bytes2'), D), T[bo, i, by])  # fixed case 40056
ck(f(('bool', 'uint112', 'bytes20'), D), T[bo, i, by])  # fixed case 40057
ck(f(('bool', 'uint112', 'bytes21'), D), T[bo, i, by])  # fixed case 40058
ck(f(('bool', 'uint112', 'bytes22'), D), T[bo, i, by])  # fixed case 40059
ck(f(('bool', 'uint112', 'bytes23'), D), T[bo, i, by])  # fixed case 40060
ck(f(('bool', 'uint112', 'bytes24'), D), T[bo, i, by])  # fixed case 40061
ck(f(('bool', 'uint112', 'bytes25'), D), T[bo, i, by])  # fixed case 40062
ck(f(('bool', 'uint112', 'bytes26'), D), T[bo, i, by])  # fixed case 40063
ck(f(('bool', 'uint112', 'bytes27'), D), T[bo, i, by])  # fixed case 40064
ck(f(('bool', 'uint112', 'bytes28'), D), T[bo, i, by])  # fixed case 40065
ck(f(('bool', 'uint112', 'bytes29'), D), T[bo, i, by])  # fixed case 40066
ck(f(('bool', 'uint112', 'bytes3'), D), T[bo, i, by])  # fixed case 40067
ck(f(('bool', 'uint112', 'bytes30'), D), T[bo, i, by])  # fixed case 40068
ck(f(('bool', 'uint112', 'bytes31'), D), T[bo, i, by])  # fixed case 40069
ck(f(('bool', 'uint112', 'bytes32'), D), T[bo, i, by])  # fixed case 40070
ck(f(('bool', 'uint112', 'bytes4'), D), T[bo, i, by])  # fixed case 40071
ck(f(('bool', 'uint112', 'bytes5'), D), T[bo, i, by])  # fixed case 40072
ck(f(('bool', 'uint112', 'bytes6'), D), T[bo, i, by])  # fixed case 40073
ck(f(('bool', 'uint112', 'bytes7'), D), T[bo, i, by])  # fixed case 40074
ck(f(('bool', 'uint112', 'bytes8'), D), T[bo, i, by])  # fixed case 40075
ck(f(('bool', 'uint112', 'bytes9'), D), T[bo, i, by])  # fixed case 40076
ck(f(('bool', 'uint112', 'int104'), D), T[bo, i, i])  # fixed case 40077
ck(f(('bool', 'uint112', 'int112'), D), T[bo, i, i])  # fixed case 40078
ck(f(('bool', 'uint112', 'int120'), D), T[bo, i, i])  # fixed case 40079
ck(f(('bool', 'uint112', 'int128'), D), T[bo, i, i])  # fixed case 40080
ck(f(('bool', 'uint112', 'int136'), D), T[bo, i, i])  # fixed case 40081
ck(f(('bool', 'uint112', 'int144'), D), T[bo, i, i])  # fixed case 40082
ck(f(('bool', 'uint112', 'int152'), D), T[bo, i, i])  # fixed case 40083
ck(f(('bool', 'uint112', 'int16'), D), T[bo, i, i])  # fixed case 40084
ck(f(('bool', 'uint112', 'int160'), D), T[bo, i, i])  # fixed case 40085
ck(f(('bool', 'uint112', 'int168'), D), T[bo, i, i])  # fixed case 40086
ck(f(('bool', 'uint112', 'int176'), D), T[bo, i, i])  # fixed case 40087
ck(f(('bool', 'uint112', 'int184'), D), T[bo, i, i])  # fixed case 40088
ck(f(('bool', 'uint112', 'int192'), D), T[bo, i, i])  # fixed case 40089
ck(f(('bool', 'uint112', 'int200'), D), T[bo, i, i])  # fixed case 40090
ck(f(('bool', 'uint112', 'int208'), D), T[bo, i, i])  # fixed case 40091
ck(f(('bool', 'uint112', 'int216'), D), T[bo, i, i])  # fixed case 40092
ck(f(('bool', 'uint112', 'int224'), D), T[bo, i, i])  # fixed case 40093
ck(f(('bool', 'uint112', 'int232'), D), T[bo, i, i])  # fixed case 40094
ck(f(('bool', 'uint112', 'int24'), D), T[bo, i, i])  # fixed case 40095
ck(f(('bool', 'uint112', 'int240'), D), T[bo, i, i])  # fixed case 40096
ck(f(('bool', 'uint112', 'int248'), D), T[bo, i, i])  # fixed case 40097
ck(f(('bool', 'uint112', 'int256'), D), T[bo, i, i])  # fixed case 40098
ck(f(('bool', 'uint112', 'int32'), D), T[bo, i, i])  # fixed case 40099
ck(f(('bool', 'uint112', 'int40'), D), T[bo, i, i])  # fixed case 40100
ck(f(('bool', 'uint112', 'int48'), D), T[bo, i, i])  # fixed case 40101
ck(f(('bool', 'uint112', 'int56'), D), T[bo, i, i])  # fixed case 40102
ck(f(('bool', 'uint112', 'int64'), D), T[bo, i, i])  # fixed case 40103
ck(f(('bool', 'uint112', 'int72'), D), T[bo, i, i])  # fixed case 40104
ck(f(('bool', 'uint112', 'int8'), D), T[bo, i, i])  # fixed case 40105
ck(f(('bool', 'uint112', 'int80'), D), T[bo, i, i])  # fixed case 40106
ck(f(('bool', 'uint112', 'int88'), D), T[bo, i, i])  # fixed case 40107
ck(f(('bool', 'uint112', 'int96'), D), T[bo, i, i])  # fixed case 40108
ck(f(('bool', 'uint112', 'string'), D), T[bo, i, s])  # fixed case 40109
ck(f(('bool', 'uint112', 'string[]'), D), T[bo, i, A])  # fixed case 40110
ck(f(('bool', 'uint112', 'uint104'), D), T[bo, i, i])  # fixed case 40111
ck(f(('bool', 'uint112', 'uint112'), D), T[bo, i, i])  # fixed case 40112
ck(f(('bool', 'uint112', 'uint120'), D), T[bo, i, i])  # fixed case 40113
ck(f(('bool', 'uint112', 'uint128'), D), T[bo, i, i])  # fixed case 40114
ck(f(('bool', 'uint112', 'uint136'), D), T[bo, i, i])  # fixed case 40115
ck(f(('bool', 'uint112', 'uint144'), D), T[bo, i, i])  # fixed case 40116
ck(f(('bool', 'uint112', 'uint152'), D), T[bo, i, i])  # fixed case 40117
ck(f(('bool', 'uint112', 'uint16'), D), T[bo, i, i])  # fixed case 40118
ck(f(('bool', 'uint112', 'uint160'), D), T[bo, i, i])  # fixed case 40119
ck(f(('bool', 'uint112', 'uint168'), D), T[bo, i, i])  # fixed case 40120
ck(f(('bool', 'uint112', 'uint176'), D), T[bo, i, i])  # fixed case 40121
ck(f(('bool', 'uint112', 'uint184'), D), T[bo, i, i])  # fixed case 40122
ck(f(('bool', 'uint112', 'uint192'), D), T[bo, i, i])  # fixed case 40123
ck(f(('bool', 'uint112', 'uint200'), D), T[bo, i, i])  # fixed case 40124
ck(f(('bool', 'uint112', 'uint208'), D), T[bo, i, i])  # fixed case 40125
ck(f(('bool', 'uint112', 'uint216'), D), T[bo, i, i])  # fixed case 40126
ck(f(('bool', 'uint112', 'uint224'), D), T[bo, i, i])  # fixed case 40127
ck(f(('bool', 'uint112', 'uint232'), D), T[bo, i, i])  # fixed case 40128
ck(f(('bool', 'uint112', 'uint24'), D), T[bo, i, i])  # fixed case 40129
ck(f(('bool', 'uint112', 'uint240'), D), T[bo, i, i])  # fixed case 40130
ck(f(('bool', 'uint112', 'uint248'), D), T[bo, i, i])  # fixed case 40131
ck(f(('bool', 'uint112', 'uint256'), D), T[bo, i, i])  # fixed case 40132
ck(f(('bool', 'uint112', 'uint32'), D), T[bo, i, i])  # fixed case 40133
ck(f(('bool', 'uint112', 'uint40'), D), T[bo, i, i])  # fixed case 40134
ck(f(('bool', 'uint112', 'uint48'), D), T[bo, i, i])  # fixed case 40135
ck(f(('bool', 'uint112', 'uint56'), D), T[bo, i, i])  # fixed case 40136
ck(f(('bool', 'uint112', 'uint64'), D), T[bo, i, i])  # fixed case 40137
ck(f(('bool', 'uint112', 'uint72'), D), T[bo, i, i])  # fixed case 40138
ck(f(('bool', 'uint112', 'uint8'), D), T[bo, i, i])  # fixed case 40139
ck(f(('bool', 'uint112', 'uint80'), D), T[bo, i, i])  # fixed case 40140
ck(f(('bool', 'uint112', 'uint88'), D), T[bo, i, i])  # fixed case 40141
ck(f(('bool', 'uint112', 'uint96'), D), T[bo, i, i])  # fixed case 40142
ck(f(('bool', 'uint112', '?'), D), T[bo, i, A])  # fixed case 40143
ck(f(('bool', 'uint120', 'address'), D), T[bo, i, HA])  # fixed case 40144
ck(f(('bool', 'uint120', 'address[]'), D), T[bo, i, A])  # fixed case 40145
ck(f(('bool', 'uint120', 'bool'), D), T[bo, i, bo])  # fixed case 40146
ck(f(('bool', 'uint120', 'bool[]'), D), T[bo, i, A])  # fixed case 40147
ck(f(('bool', 'uint120', 'bytes'), D), T[bo, i, by])  # fixed case 40148
ck(f(('bool', 'uint120', 'bytes1'), D), T[bo, i, by])  # fixed case 40149
ck(f(('bool', 'uint120', 'bytes10'), D), T[bo, i, by])  # fixed case 40150
ck(f(('bool', 'uint120', 'bytes11'), D), T[bo, i, by])  # fixed case 40151
ck(f(('bool', 'uint120', 'bytes12'), D), T[bo, i, by])  # fixed case 40152
ck(f(('bool', 'uint120', 'bytes13'), D), T[bo, i, by])  # fixed case 40153
ck(f(('bool', 'uint120', 'bytes14'), D), T[bo, i, by])  # fixed case 40154
ck(f(('bool', 'uint120', 'bytes15'), D), T[bo, i, by])  # fixed case 40155
ck(f(('bool', 'uint120', 'bytes16'), D), T[bo, i, by])  # fixed case 40156
ck(f(('bool', 'uint120', 'bytes17'), D), T[bo, i, by])  # fixed case 40157
ck(f(('bool', 'uint120', 'bytes18'), D), T[bo, i, by])  # fixed case 40158
ck(f(('bool', 'uint120', 'bytes19'), D), T[bo, i, by])  # fixed case 40159
ck(f(('bool', 'uint120', 'bytes2'), D), T[bo, i, by])  # fixed case 40160
ck(f(('bool', 'uint120', 'bytes20'), D), T[bo, i, by])  # fixed case 40161
ck(f(('bool', 'uint120', 'bytes21'), D), T[bo, i, by])  # fixed case 40162
ck(f(('bool', 'uint120', 'bytes22'), D), T[bo, i, by])  # fixed case 40163
ck(f(('bool', 'uint120', 'bytes23'), D), T[bo, i, by])  # fixed case 40164
ck(f(('bool', 'uint120', 'bytes24'), D), T[bo, i, by])  # fixed case 40165
ck(f(('bool', 'uint120', 'bytes25'), D), T[bo, i, by])  # fixed case 40166
ck(f(('bool', 'uint120', 'bytes26'), D), T[bo, i, by])  # fixed case 40167
ck(f(('bool', 'uint120', 'bytes27'), D), T[bo, i, by])  # fixed case 40168
ck(f(('bool', 'uint120', 'bytes28'), D), T[bo, i, by])  # fixed case 40169
ck(f(('bool', 'uint120', 'bytes29'), D), T[bo, i, by])  # fixed case 40170
ck(f(('bool', 'uint120', 'bytes3'), D), T[bo, i, by])  # fixed case 40171
ck(f(('bool', 'uint120', 'bytes30'), D), T[bo, i, by])  # fixed case 40172
ck(f(('bool', 'uint120', 'bytes31'), D), T[bo, i, by])  # fixed case 40173
ck(f(('bool', 'uint120', 'bytes32'), D), T[bo, i, by])  # fixed case 40174
ck(f(('bool', 'uint120', 'bytes4'), D), T[bo, i, by])  # fixed case 40175
ck(f(('bool', 'uint120', 'bytes5'), D), T[bo, i, by])  # fixed case 40176
ck(f(('bool', 'uint120', 'bytes6'), D), T[bo, i, by])  # fixed case 40177
ck(f(('bool', 'uint120', 'bytes7'), D), T[bo, i, by])  # fixed case 40178
ck(f(('bool', 'uint120', 'bytes8'), D), T[bo, i, by])  # fixed case 40179
ck(f(('bool', 'uint120', 'bytes9'), D), T[bo, i, by])  # fixed case 40180
ck(f(('bool', 'uint120', 'int104'), D), T[bo, i, i])  # fixed case 40181
ck(f(('bool', 'uint120', 'int112'), D), T[bo, i, i])  # fixed case 40182
ck(f(('bool', 'uint120', 'int120'), D), T[bo, i, i])  # fixed case 40183
ck(f(('bool', 'uint120', 'int128'), D), T[bo, i, i])  # fixed case 40184
ck(f(('bool', 'uint120', 'int136'), D), T[bo, i, i])  # fixed case 40185
ck(f(('bool', 'uint120', 'int144'), D), T[bo, i, i])  # fixed case 40186
ck(f(('bool', 'uint120', 'int152'), D), T[bo, i, i])  # fixed case 40187
ck(f(('bool', 'uint120', 'int16'), D), T[bo, i, i])  # fixed case 40188
ck(f(('bool', 'uint120', 'int160'), D), T[bo, i, i])  # fixed case 40189
ck(f(('bool', 'uint120', 'int168'), D), T[bo, i, i])  # fixed case 40190
ck(f(('bool', 'uint120', 'int176'), D), T[bo, i, i])  # fixed case 40191
ck(f(('bool', 'uint120', 'int184'), D), T[bo, i, i])  # fixed case 40192
ck(f(('bool', 'uint120', 'int192'), D), T[bo, i, i])  # fixed case 40193
ck(f(('bool', 'uint120', 'int200'), D), T[bo, i, i])  # fixed case 40194
ck(f(('bool', 'uint120', 'int208'), D), T[bo, i, i])  # fixed case 40195
ck(f(('bool', 'uint120', 'int216'), D), T[bo, i, i])  # fixed case 40196
ck(f(('bool', 'uint120', 'int224'), D), T[bo, i, i])  # fixed case 40197
ck(f(('bool', 'uint120', 'int232'), D), T[bo, i, i])  # fixed case 40198
ck(f(('bool', 'uint120', 'int24'), D), T[bo, i, i])  # fixed case 40199
ck(f(('bool', 'uint120', 'int240'), D), T[bo, i, i])  # fixed case 40200
ck(f(('bool', 'uint120', 'int248'), D), T[bo, i, i])  # fixed case 40201
ck(f(('bool', 'uint120', 'int256'), D), T[bo, i, i])  # fixed case 40202
ck(f(('bool', 'uint120', 'int32'), D), T[bo, i, i])  # fixed case 40203
ck(f(('bool', 'uint120', 'int40'), D), T[bo, i, i])  # fixed case 40204
ck(f(('bool', 'uint120', 'int48'), D), T[bo, i, i])  # fixed case 40205
ck(f(('bool', 'uint120', 'int56'), D), T[bo, i, i])  # fixed case 40206
ck(f(('bool', 'uint120', 'int64'), D), T[bo, i, i])  # fixed case 40207
ck(f(('bool', 'uint120', 'int72'), D), T[bo, i, i])  # fixed case 40208
ck(f(('bool', 'uint120', 'int8'), D), T[bo, i, i])  # fixed case 40209
ck(f(('bool', 'uint120', 'int80'), D), T[bo, i, i])  # fixed case 40210
ck(f(('bool', 'uint120', 'int88'), D), T[bo, i, i])  # fixed case 40211
ck(f(('bool', 'uint120', 'int96'), D), T[bo, i, i])  # fixed case 40212
ck(f(('bool', 'uint120', 'string'), D), T[bo, i, s])  # fixed case 40213
ck(f(('bool', 'uint120', 'string[]'), D), T[bo, i, A])  # fixed case 40214
ck(f(('bool', 'uint120', 'uint104'), D), T[bo, i, i])  # fixed case 40215
ck(f(('bool', 'uint120', 'uint112'), D), T[bo, i, i])  # fixed case 40216
ck(f(('bool', 'uint120', 'uint120'), D), T[bo, i, i])  # fixed case 40217
ck(f(('bool', 'uint120', 'uint128'), D), T[bo, i, i])  # fixed case 40218
ck(f(('bool', 'uint120', 'uint136'), D), T[bo, i, i])  # fixed case 40219
ck(f(('bool', 'uint120', 'uint144'), D), T[bo, i, i])  # fixed case 40220
ck(f(('bool', 'uint120', 'uint152'), D), T[bo, i, i])  # fixed case 40221
ck(f(('bool', 'uint120', 'uint16'), D), T[bo, i, i])  # fixed case 40222
ck(f(('bool', 'uint120', 'uint160'), D), T[bo, i, i])  # fixed case 40223
ck(f(('bool', 'uint120', 'uint168'), D), T[bo, i, i])  # fixed case 40224
ck(f(('bool', 'uint120', 'uint176'), D), T[bo, i, i])  # fixed case 40225
ck(f(('bool', 'uint120', 'uint184'), D), T[bo, i, i])  # fixed case 40226
ck(f(('bool', 'uint120', 'uint192'), D), T[bo, i, i])  # fixed case 40227
ck(f(('bool', 'uint120', 'uint200'), D), T[bo, i, i])  # fixed case 40228
ck(f(('bool', 'uint120', 'uint208'), D), T[bo, i, i])  # fixed case 40229
ck(f(('bool', 'uint120', 'uint216'), D), T[bo, i, i])  # fixed case 40230
ck(f(('bool', 'uint120', 'uint224'), D), T[bo, i, i])  # fixed case 40231
ck(f(('bool', 'uint120', 'uint232'), D), T[bo, i, i])  # fixed case 40232
ck(f(('bool', 'uint120', 'uint24'), D), T[bo, i, i])  # fixed case 40233
ck(f(('bool', 'uint120', 'uint240'), D), T[bo, i, i])  # fixed case 40234
ck(f(('bool', 'uint120', 'uint248'), D), T[bo, i, i])  # fixed case 40235
ck(f(('bool', 'uint120', 'uint256'), D), T[bo, i, i])  # fixed case 40236
ck(f(('bool', 'uint120', 'uint32'), D), T[bo, i, i])  # fixed case 40237
ck(f(('bool', 'uint120', 'uint40'), D), T[bo, i, i])  # fixed case 40238
ck(f(('bool', 'uint120', 'uint48'), D), T[bo, i, i])  # fixed case 40239
ck(f(('bool', 'uint120', 'uint56'), D), T[bo, i, i])  # fixed case 40240
ck(f(('bool', 'uint120', 'uint64'), D), T[bo, i, i])  # fixed case 40241
ck(f(('bool', 'uint120', 'uint72'), D), T[bo, i, i])  # fixed case 40242
ck(f(('bool', 'uint120', 'uint8'), D), T[bo, i, i])  # fixed case 40243
ck(f(('bool', 'uint120', 'uint80'), D), T[bo, i, i])  # fixed case 40244
ck(f(('bool', 'uint120', 'uint88'), D), T[bo, i, i])  # fixed case 40245
ck(f(('bool', 'uint120', 'uint96'), D), T[bo, i, i])  # fixed case 40246
ck(f(('bool', 'uint120', '?'), D), T[bo, i, A])  # fixed case 40247
ck(f(('bool', 'uint128', 'address'), D), T[bo, i, HA])  # fixed case 40248
ck(f(('bool', 'uint128', 'address[]'), D), T[bo, i, A])  # fixed case 40249
ck(f(('bool', 'uint128', 'bool'), D), T[bo, i, bo])  # fixed case 40250
ck(f(('bool', 'uint128', 'bool[]'), D), T[bo, i, A])  # fixed case 40251
ck(f(('bool', 'uint128', 'bytes'), D), T[bo, i, by])  # fixed case 40252
ck(f(('bool', 'uint128', 'bytes1'), D), T[bo, i, by])  # fixed case 40253
ck(f(('bool', 'uint128', 'bytes10'), D), T[bo, i, by])  # fixed case 40254
ck(f(('bool', 'uint128', 'bytes11'), D), T[bo, i, by])  # fixed case 40255
ck(f(('bool', 'uint128', 'bytes12'), D), T[bo, i, by])  # fixed case 40256
ck(f(('bool', 'uint128', 'bytes13'), D), T[bo, i, by])  # fixed case 40257
ck(f(('bool', 'uint128', 'bytes14'), D), T[bo, i, by])  # fixed case 40258
ck(f(('bool', 'uint128', 'bytes15'), D), T[bo, i, by])  # fixed case 40259
ck(f(('bool', 'uint128', 'bytes16'), D), T[bo, i, by])  # fixed case 40260
ck(f(('bool', 'uint128', 'bytes17'), D), T[bo, i, by])  # fixed case 40261
ck(f(('bool', 'uint128', 'bytes18'), D), T[bo, i, by])  # fixed case 40262
ck(f(('bool', 'uint128', 'bytes19'), D), T[bo, i, by])  # fixed case 40263
ck(f(('bool', 'uint128', 'bytes2'), D), T[bo, i, by])  # fixed case 40264
ck(f(('bool', 'uint128', 'bytes20'), D), T[bo, i, by])  # fixed case 40265
ck(f(('bool', 'uint128', 'bytes21'), D), T[bo, i, by])  # fixed case 40266
ck(f(('bool', 'uint128', 'bytes22'), D), T[bo, i, by])  # fixed case 40267
ck(f(('bool', 'uint128', 'bytes23'), D), T[bo, i, by])  # fixed case 40268
ck(f(('bool', 'uint128', 'bytes24'), D), T[bo, i, by])  # fixed case 40269
ck(f(('bool', 'uint128', 'bytes25'), D), T[bo, i, by])  # fixed case 40270
ck(f(('bool', 'uint128', 'bytes26'), D), T[bo, i, by])  # fixed case 40271
ck(f(('bool', 'uint128', 'bytes27'), D), T[bo, i, by])  # fixed case 40272
ck(f(('bool', 'uint128', 'bytes28'), D), T[bo, i, by])  # fixed case 40273
ck(f(('bool', 'uint128', 'bytes29'), D), T[bo, i, by])  # fixed case 40274
ck(f(('bool', 'uint128', 'bytes3'), D), T[bo, i, by])  # fixed case 40275
ck(f(('bool', 'uint128', 'bytes30'), D), T[bo, i, by])  # fixed case 40276
ck(f(('bool', 'uint128', 'bytes31'), D), T[bo, i, by])  # fixed case 40277
ck(f(('bool', 'uint128', 'bytes32'), D), T[bo, i, by])  # fixed case 40278
ck(f(('bool', 'uint128', 'bytes4'), D), T[bo, i, by])  # fixed case 40279
ck(f(('bool', 'uint128', 'bytes5'), D), T[bo, i, by])  # fixed case 40280
ck(f(('bool', 'uint128', 'bytes6'), D), T[bo, i, by])  # fixed case 40281
ck(f(('bool', 'uint128', 'bytes7'), D), T[bo, i, by])  # fixed case 40282
ck(f(('bool', 'uint128', 'bytes8'), D), T[bo, i, by])  # fixed case 40283
ck(f(('bool', 'uint128', 'bytes9'), D), T[bo, i, by])  # fixed case 40284
ck(f(('bool', 'uint128', 'int104'), D), T[bo, i, i])  # fixed case 40285
ck(f(('bool', 'uint128', 'int112'), D), T[bo, i, i])  # fixed case 40286
ck(f(('bool', 'uint128', 'int120'), D), T[bo, i, i])  # fixed case 40287
ck(f(('bool', 'uint128', 'int128'), D), T[bo, i, i])  # fixed case 40288
ck(f(('bool', 'uint128', 'int136'), D), T[bo, i, i])  # fixed case 40289
ck(f(('bool', 'uint128', 'int144'), D), T[bo, i, i])  # fixed case 40290
ck(f(('bool', 'uint128', 'int152'), D), T[bo, i, i])  # fixed case 40291
ck(f(('bool', 'uint128', 'int16'), D), T[bo, i, i])  # fixed case 40292
ck(f(('bool', 'uint128', 'int160'), D), T[bo, i, i])  # fixed case 40293
ck(f(('bool', 'uint128', 'int168'), D), T[bo, i, i])  # fixed case 40294
ck(f(('bool', 'uint128', 'int176'), D), T[bo, i, i])  # fixed case 40295
ck(f(('bool', 'uint128', 'int184'), D), T[bo, i, i])  # fixed case 40296
ck(f(('bool', 'uint128', 'int192'), D), T[bo, i, i])  # fixed case 40297
ck(f(('bool', 'uint128', 'int200'), D), T[bo, i, i])  # fixed case 40298
ck(f(('bool', 'uint128', 'int208'), D), T[bo, i, i])  # fixed case 40299
ck(f(('bool', 'uint128', 'int216'), D), T[bo, i, i])  # fixed case 40300
ck(f(('bool', 'uint128', 'int224'), D), T[bo, i, i])  # fixed case 40301
ck(f(('bool', 'uint128', 'int232'), D), T[bo, i, i])  # fixed case 40302
ck(f(('bool', 'uint128', 'int24'), D), T[bo, i, i])  # fixed case 40303
ck(f(('bool', 'uint128', 'int240'), D), T[bo, i, i])  # fixed case 40304
ck(f(('bool', 'uint128', 'int248'), D), T[bo, i, i])  # fixed case 40305
ck(f(('bool', 'uint128', 'int256'), D), T[bo, i, i])  # fixed case 40306
ck(f(('bool', 'uint128', 'int32'), D), T[bo, i, i])  # fixed case 40307
ck(f(('bool', 'uint128', 'int40'), D), T[bo, i, i])  # fixed case 40308
ck(f(('bool', 'uint128', 'int48'), D), T[bo, i, i])  # fixed case 40309
ck(f(('bool', 'uint128', 'int56'), D), T[bo, i, i])  # fixed case 40310
ck(f(('bool', 'uint128', 'int64'), D), T[bo, i, i])  # fixed case 40311
ck(f(('bool', 'uint128', 'int72'), D), T[bo, i, i])  # fixed case 40312
ck(f(('bool', 'uint128', 'int8'), D), T[bo, i, i])  # fixed case 40313
ck(f(('bool', 'uint128', 'int80'), D), T[bo, i, i])  # fixed case 40314
ck(f(('bool', 'uint128', 'int88'), D), T[bo, i, i])  # fixed case 40315
ck(f(('bool', 'uint128', 'int96'), D), T[bo, i, i])  # fixed case 40316
ck(f(('bool', 'uint128', 'string'), D), T[bo, i, s])  # fixed case 40317
ck(f(('bool', 'uint128', 'string[]'), D), T[bo, i, A])  # fixed case 40318
ck(f(('bool', 'uint128', 'uint104'), D), T[bo, i, i])  # fixed case 40319
ck(f(('bool', 'uint128', 'uint112'), D), T[bo, i, i])  # fixed case 40320
ck(f(('bool', 'uint128', 'uint120'), D), T[bo, i, i])  # fixed case 40321
ck(f(('bool', 'uint128', 'uint128'), D), T[bo, i, i])  # fixed case 40322
ck(f(('bool', 'uint128', 'uint136'), D), T[bo, i, i])  # fixed case 40323
ck(f(('bool', 'uint128', 'uint144'), D), T[bo, i, i])  # fixed case 40324
ck(f(('bool', 'uint128', 'uint152'), D), T[bo, i, i])  # fixed case 40325
ck(f(('bool', 'uint128', 'uint16'), D), T[bo, i, i])  # fixed case 40326
ck(f(('bool', 'uint128', 'uint160'), D), T[bo, i, i])  # fixed case 40327
ck(f(('bool', 'uint128', 'uint168'), D), T[bo, i, i])  # fixed case 40328
ck(f(('bool', 'uint128', 'uint176'), D), T[bo, i, i])  # fixed case 40329
ck(f(('bool', 'uint128', 'uint184'), D), T[bo, i, i])  # fixed case 40330
ck(f(('bool', 'uint128', 'uint192'), D), T[bo, i, i])  # fixed case 40331
ck(f(('bool', 'uint128', 'uint200'), D), T[bo, i, i])  # fixed case 40332
ck(f(('bool', 'uint128', 'uint208'), D), T[bo, i, i])  # fixed case 40333
ck(f(('bool', 'uint128', 'uint216'), D), T[bo, i, i])  # fixed case 40334
ck(f(('bool', 'uint128', 'uint224'), D), T[bo, i, i])  # fixed case 40335
ck(f(('bool', 'uint128', 'uint232'), D), T[bo, i, i])  # fixed case 40336
ck(f(('bool', 'uint128', 'uint24'), D), T[bo, i, i])  # fixed case 40337
ck(f(('bool', 'uint128', 'uint240'), D), T[bo, i, i])  # fixed case 40338
ck(f(('bool', 'uint128', 'uint248'), D), T[bo, i, i])  # fixed case 40339
ck(f(('bool', 'uint128', 'uint256'), D), T[bo, i, i])  # fixed case 40340
ck(f(('bool', 'uint128', 'uint32'), D), T[bo, i, i])  # fixed case 40341
ck(f(('bool', 'uint128', 'uint40'), D), T[bo, i, i])  # fixed case 40342
ck(f(('bool', 'uint128', 'uint48'), D), T[bo, i, i])  # fixed case 40343
ck(f(('bool', 'uint128', 'uint56'), D), T[bo, i, i])  # fixed case 40344
ck(f(('bool', 'uint128', 'uint64'), D), T[bo, i, i])  # fixed case 40345
ck(f(('bool', 'uint128', 'uint72'), D), T[bo, i, i])  # fixed case 40346
ck(f(('bool', 'uint128', 'uint8'), D), T[bo, i, i])  # fixed case 40347
ck(f(('bool', 'uint128', 'uint80'), D), T[bo, i, i])  # fixed case 40348
ck(f(('bool', 'uint128', 'uint88'), D), T[bo, i, i])  # fixed case 40349
ck(f(('bool', 'uint128', 'uint96'), D), T[bo, i, i])  # fixed case 40350
ck(f(('bool', 'uint128', '?'), D), T[bo, i, A])  # fixed case 40351
ck(f(('bool', 'uint136', 'address'), D), T[bo, i, HA])  # fixed case 40352
ck(f(('bool', 'uint136', 'address[]'), D), T[bo, i, A])  # fixed case 40353
ck(f(('bool', 'uint136', 'bool'), D), T[bo, i, bo])  # fixed case 40354
ck(f(('bool', 'uint136', 'bool[]'), D), T[bo, i, A])  # fixed case 40355
ck(f(('bool', 'uint136', 'bytes'), D), T[bo, i, by])  # fixed case 40356
ck(f(('bool', 'uint136', 'bytes1'), D), T[bo, i, by])  # fixed case 40357
ck(f(('bool', 'uint136', 'bytes10'), D), T[bo, i, by])  # fixed case 40358
ck(f(('bool', 'uint136', 'bytes11'), D), T[bo, i, by])  # fixed case 40359
ck(f(('bool', 'uint136', 'bytes12'), D), T[bo, i, by])  # fixed case 40360
ck(f(('bool', 'uint136', 'bytes13'), D), T[bo, i, by])  # fixed case 40361
ck(f(('bool', 'uint136', 'bytes14'), D), T[bo, i, by])  # fixed case 40362
ck(f(('bool', 'uint136', 'bytes15'), D), T[bo, i, by])  # fixed case 40363
ck(f(('bool', 'uint136', 'bytes16'), D), T[bo, i, by])  # fixed case 40364
ck(f(('bool', 'uint136', 'bytes17'), D), T[bo, i, by])  # fixed case 40365
ck(f(('bool', 'uint136', 'bytes18'), D), T[bo, i, by])  # fixed case 40366
ck(f(('bool', 'uint136', 'bytes19'), D), T[bo, i, by])  # fixed case 40367
ck(f(('bool', 'uint136', 'bytes2'), D), T[bo, i, by])  # fixed case 40368
ck(f(('bool', 'uint136', 'bytes20'), D), T[bo, i, by])  # fixed case 40369
ck(f(('bool', 'uint136', 'bytes21'), D), T[bo, i, by])  # fixed case 40370
ck(f(('bool', 'uint136', 'bytes22'), D), T[bo, i, by])  # fixed case 40371
ck(f(('bool', 'uint136', 'bytes23'), D), T[bo, i, by])  # fixed case 40372
ck(f(('bool', 'uint136', 'bytes24'), D), T[bo, i, by])  # fixed case 40373
ck(f(('bool', 'uint136', 'bytes25'), D), T[bo, i, by])  # fixed case 40374
ck(f(('bool', 'uint136', 'bytes26'), D), T[bo, i, by])  # fixed case 40375
ck(f(('bool', 'uint136', 'bytes27'), D), T[bo, i, by])  # fixed case 40376
ck(f(('bool', 'uint136', 'bytes28'), D), T[bo, i, by])  # fixed case 40377
ck(f(('bool', 'uint136', 'bytes29'), D), T[bo, i, by])  # fixed case 40378
ck(f(('bool', 'uint136', 'bytes3'), D), T[bo, i, by])  # fixed case 40379
ck(f(('bool', 'uint136', 'bytes30'), D), T[bo, i, by])  # fixed case 40380
ck(f(('bool', 'uint136', 'bytes31'), D), T[bo, i, by])  # fixed case 40381
ck(f(('bool', 'uint136', 'bytes32'), D), T[bo, i, by])  # fixed case 40382
ck(f(('bool', 'uint136', 'bytes4'), D), T[bo, i, by])  # fixed case 40383
ck(f(('bool', 'uint136', 'bytes5'), D), T[bo, i, by])  # fixed case 40384
ck(f(('bool', 'uint136', 'bytes6'), D), T[bo, i, by])  # fixed case 40385
ck(f(('bool', 'uint136', 'bytes7'), D), T[bo, i, by])  # fixed case 40386
ck(f(('bool', 'uint136', 'bytes8'), D), T[bo, i, by])  # fixed case 40387
ck(f(('bool', 'uint136', 'bytes9'), D), T[bo, i, by])  # fixed case 40388
ck(f(('bool', 'uint136', 'int104'), D), T[bo, i, i])  # fixed case 40389
ck(f(('bool', 'uint136', 'int112'), D), T[bo, i, i])  # fixed case 40390
ck(f(('bool', 'uint136', 'int120'), D), T[bo, i, i])  # fixed case 40391
ck(f(('bool', 'uint136', 'int128'), D), T[bo, i, i])  # fixed case 40392
ck(f(('bool', 'uint136', 'int136'), D), T[bo, i, i])  # fixed case 40393
ck(f(('bool', 'uint136', 'int144'), D), T[bo, i, i])  # fixed case 40394
ck(f(('bool', 'uint136', 'int152'), D), T[bo, i, i])  # fixed case 40395
ck(f(('bool', 'uint136', 'int16'), D), T[bo, i, i])  # fixed case 40396
ck(f(('bool', 'uint136', 'int160'), D), T[bo, i, i])  # fixed case 40397
ck(f(('bool', 'uint136', 'int168'), D), T[bo, i, i])  # fixed case 40398
ck(f(('bool', 'uint136', 'int176'), D), T[bo, i, i])  # fixed case 40399
ck(f(('bool', 'uint136', 'int184'), D), T[bo, i, i])  # fixed case 40400
ck(f(('bool', 'uint136', 'int192'), D), T[bo, i, i])  # fixed case 40401
ck(f(('bool', 'uint136', 'int200'), D), T[bo, i, i])  # fixed case 40402
ck(f(('bool', 'uint136', 'int208'), D), T[bo, i, i])  # fixed case 40403
ck(f(('bool', 'uint136', 'int216'), D), T[bo, i, i])  # fixed case 40404
ck(f(('bool', 'uint136', 'int224'), D), T[bo, i, i])  # fixed case 40405
ck(f(('bool', 'uint136', 'int232'), D), T[bo, i, i])  # fixed case 40406
ck(f(('bool', 'uint136', 'int24'), D), T[bo, i, i])  # fixed case 40407
ck(f(('bool', 'uint136', 'int240'), D), T[bo, i, i])  # fixed case 40408
ck(f(('bool', 'uint136', 'int248'), D), T[bo, i, i])  # fixed case 40409
ck(f(('bool', 'uint136', 'int256'), D), T[bo, i, i])  # fixed case 40410
ck(f(('bool', 'uint136', 'int32'), D), T[bo, i, i])  # fixed case 40411
ck(f(('bool', 'uint136', 'int40'), D), T[bo, i, i])  # fixed case 40412
ck(f(('bool', 'uint136', 'int48'), D), T[bo, i, i])  # fixed case 40413
ck(f(('bool', 'uint136', 'int56'), D), T[bo, i, i])  # fixed case 40414
ck(f(('bool', 'uint136', 'int64'), D), T[bo, i, i])  # fixed case 40415
ck(f(('bool', 'uint136', 'int72'), D), T[bo, i, i])  # fixed case 40416
ck(f(('bool', 'uint136', 'int8'), D), T[bo, i, i])  # fixed case 40417
ck(f(('bool', 'uint136', 'int80'), D), T[bo, i, i])  # fixed case 40418
ck(f(('bool', 'uint136', 'int88'), D), T[bo, i, i])  # fixed case 40419
ck(f(('bool', 'uint136', 'int96'), D), T[bo, i, i])  # fixed case 40420
ck(f(('bool', 'uint136', 'string'), D), T[bo, i, s])  # fixed case 40421
ck(f(('bool', 'uint136', 'string[]'), D), T[bo, i, A])  # fixed case 40422
ck(f(('bool', 'uint136', 'uint104'), D), T[bo, i, i])  # fixed case 40423
ck(f(('bool', 'uint136', 'uint112'), D), T[bo, i, i])  # fixed case 40424
ck(f(('bool', 'uint136', 'uint120'), D), T[bo, i, i])  # fixed case 40425
ck(f(('bool', 'uint136', 'uint128'), D), T[bo, i, i])  # fixed case 40426
ck(f(('bool', 'uint136', 'uint136'), D), T[bo, i, i])  # fixed case 40427
ck(f(('bool', 'uint136', 'uint144'), D), T[bo, i, i])  # fixed case 40428
ck(f(('bool', 'uint136', 'uint152'), D), T[bo, i, i])  # fixed case 40429
ck(f(('bool', 'uint136', 'uint16'), D), T[bo, i, i])  # fixed case 40430
ck(f(('bool', 'uint136', 'uint160'), D), T[bo, i, i])  # fixed case 40431
ck(f(('bool', 'uint136', 'uint168'), D), T[bo, i, i])  # fixed case 40432
ck(f(('bool', 'uint136', 'uint176'), D), T[bo, i, i])  # fixed case 40433
ck(f(('bool', 'uint136', 'uint184'), D), T[bo, i, i])  # fixed case 40434
ck(f(('bool', 'uint136', 'uint192'), D), T[bo, i, i])  # fixed case 40435
ck(f(('bool', 'uint136', 'uint200'), D), T[bo, i, i])  # fixed case 40436
ck(f(('bool', 'uint136', 'uint208'), D), T[bo, i, i])  # fixed case 40437
ck(f(('bool', 'uint136', 'uint216'), D), T[bo, i, i])  # fixed case 40438
ck(f(('bool', 'uint136', 'uint224'), D), T[bo, i, i])  # fixed case 40439
ck(f(('bool', 'uint136', 'uint232'), D), T[bo, i, i])  # fixed case 40440
ck(f(('bool', 'uint136', 'uint24'), D), T[bo, i, i])  # fixed case 40441
ck(f(('bool', 'uint136', 'uint240'), D), T[bo, i, i])  # fixed case 40442
ck(f(('bool', 'uint136', 'uint248'), D), T[bo, i, i])  # fixed case 40443
ck(f(('bool', 'uint136', 'uint256'), D), T[bo, i, i])  # fixed case 40444
ck(f(('bool', 'uint136', 'uint32'), D), T[bo, i, i])  # fixed case 40445
ck(f(('bool', 'uint136', 'uint40'), D), T[bo, i, i])  # fixed case 40446
ck(f(('bool', 'uint136', 'uint48'), D), T[bo, i, i])  # fixed case 40447
ck(f(('bool', 'uint136', 'uint56'), D), T[bo, i, i])  # fixed case 40448
ck(f(('bool', 'uint136', 'uint64'), D), T[bo, i, i])  # fixed case 40449
ck(f(('bool', 'uint136', 'uint72'), D), T[bo, i, i])  # fixed case 40450
ck(f(('bool', 'uint136', 'uint8'), D), T[bo, i, i])  # fixed case 40451
ck(f(('bool', 'uint136', 'uint80'), D), T[bo, i, i])  # fixed case 40452
ck(f(('bool', 'uint136', 'uint88'), D), T[bo, i, i])  # fixed case 40453
ck(f(('bool', 'uint136', 'uint96'), D), T[bo, i, i])  # fixed case 40454
ck(f(('bool', 'uint136', '?'), D), T[bo, i, A])  # fixed case 40455
ck(f(('bool', 'uint144', 'address'), D), T[bo, i, HA])  # fixed case 40456
ck(f(('bool', 'uint144', 'address[]'), D), T[bo, i, A])  # fixed case 40457
ck(f(('bool', 'uint144', 'bool'), D), T[bo, i, bo])  # fixed case 40458
ck(f(('bool', 'uint144', 'bool[]'), D), T[bo, i, A])  # fixed case 40459
ck(f(('bool', 'uint144', 'bytes'), D), T[bo, i, by])  # fixed case 40460
ck(f(('bool', 'uint144', 'bytes1'), D), T[bo, i, by])  # fixed case 40461
ck(f(('bool', 'uint144', 'bytes10'), D), T[bo, i, by])  # fixed case 40462
ck(f(('bool', 'uint144', 'bytes11'), D), T[bo, i, by])  # fixed case 40463
ck(f(('bool', 'uint144', 'bytes12'), D), T[bo, i, by])  # fixed case 40464
ck(f(('bool', 'uint144', 'bytes13'), D), T[bo, i, by])  # fixed case 40465
ck(f(('bool', 'uint144', 'bytes14'), D), T[bo, i, by])  # fixed case 40466
ck(f(('bool', 'uint144', 'bytes15'), D), T[bo, i, by])  # fixed case 40467
ck(f(('bool', 'uint144', 'bytes16'), D), T[bo, i, by])  # fixed case 40468
ck(f(('bool', 'uint144', 'bytes17'), D), T[bo, i, by])  # fixed case 40469
ck(f(('bool', 'uint144', 'bytes18'), D), T[bo, i, by])  # fixed case 40470
ck(f(('bool', 'uint144', 'bytes19'), D), T[bo, i, by])  # fixed case 40471
ck(f(('bool', 'uint144', 'bytes2'), D), T[bo, i, by])  # fixed case 40472
ck(f(('bool', 'uint144', 'bytes20'), D), T[bo, i, by])  # fixed case 40473
ck(f(('bool', 'uint144', 'bytes21'), D), T[bo, i, by])  # fixed case 40474
ck(f(('bool', 'uint144', 'bytes22'), D), T[bo, i, by])  # fixed case 40475
ck(f(('bool', 'uint144', 'bytes23'), D), T[bo, i, by])  # fixed case 40476
ck(f(('bool', 'uint144', 'bytes24'), D), T[bo, i, by])  # fixed case 40477
ck(f(('bool', 'uint144', 'bytes25'), D), T[bo, i, by])  # fixed case 40478
ck(f(('bool', 'uint144', 'bytes26'), D), T[bo, i, by])  # fixed case 40479
ck(f(('bool', 'uint144', 'bytes27'), D), T[bo, i, by])  # fixed case 40480
ck(f(('bool', 'uint144', 'bytes28'), D), T[bo, i, by])  # fixed case 40481
ck(f(('bool', 'uint144', 'bytes29'), D), T[bo, i, by])  # fixed case 40482
ck(f(('bool', 'uint144', 'bytes3'), D), T[bo, i, by])  # fixed case 40483
ck(f(('bool', 'uint144', 'bytes30'), D), T[bo, i, by])  # fixed case 40484
ck(f(('bool', 'uint144', 'bytes31'), D), T[bo, i, by])  # fixed case 40485
ck(f(('bool', 'uint144', 'bytes32'), D), T[bo, i, by])  # fixed case 40486
ck(f(('bool', 'uint144', 'bytes4'), D), T[bo, i, by])  # fixed case 40487
ck(f(('bool', 'uint144', 'bytes5'), D), T[bo, i, by])  # fixed case 40488
ck(f(('bool', 'uint144', 'bytes6'), D), T[bo, i, by])  # fixed case 40489
ck(f(('bool', 'uint144', 'bytes7'), D), T[bo, i, by])  # fixed case 40490
ck(f(('bool', 'uint144', 'bytes8'), D), T[bo, i, by])  # fixed case 40491
ck(f(('bool', 'uint144', 'bytes9'), D), T[bo, i, by])  # fixed case 40492
ck(f(('bool', 'uint144', 'int104'), D), T[bo, i, i])  # fixed case 40493
ck(f(('bool', 'uint144', 'int112'), D), T[bo, i, i])  # fixed case 40494
ck(f(('bool', 'uint144', 'int120'), D), T[bo, i, i])  # fixed case 40495
ck(f(('bool', 'uint144', 'int128'), D), T[bo, i, i])  # fixed case 40496
ck(f(('bool', 'uint144', 'int136'), D), T[bo, i, i])  # fixed case 40497
ck(f(('bool', 'uint144', 'int144'), D), T[bo, i, i])  # fixed case 40498
ck(f(('bool', 'uint144', 'int152'), D), T[bo, i, i])  # fixed case 40499
ck(f(('bool', 'uint144', 'int16'), D), T[bo, i, i])  # fixed case 40500
ck(f(('bool', 'uint144', 'int160'), D), T[bo, i, i])  # fixed case 40501
ck(f(('bool', 'uint144', 'int168'), D), T[bo, i, i])  # fixed case 40502
ck(f(('bool', 'uint144', 'int176'), D), T[bo, i, i])  # fixed case 40503
ck(f(('bool', 'uint144', 'int184'), D), T[bo, i, i])  # fixed case 40504
ck(f(('bool', 'uint144', 'int192'), D), T[bo, i, i])  # fixed case 40505
ck(f(('bool', 'uint144', 'int200'), D), T[bo, i, i])  # fixed case 40506
ck(f(('bool', 'uint144', 'int208'), D), T[bo, i, i])  # fixed case 40507
ck(f(('bool', 'uint144', 'int216'), D), T[bo, i, i])  # fixed case 40508
ck(f(('bool', 'uint144', 'int224'), D), T[bo, i, i])  # fixed case 40509
ck(f(('bool', 'uint144', 'int232'), D), T[bo, i, i])  # fixed case 40510
ck(f(('bool', 'uint144', 'int24'), D), T[bo, i, i])  # fixed case 40511
ck(f(('bool', 'uint144', 'int240'), D), T[bo, i, i])  # fixed case 40512
ck(f(('bool', 'uint144', 'int248'), D), T[bo, i, i])  # fixed case 40513
ck(f(('bool', 'uint144', 'int256'), D), T[bo, i, i])  # fixed case 40514
ck(f(('bool', 'uint144', 'int32'), D), T[bo, i, i])  # fixed case 40515
ck(f(('bool', 'uint144', 'int40'), D), T[bo, i, i])  # fixed case 40516
ck(f(('bool', 'uint144', 'int48'), D), T[bo, i, i])  # fixed case 40517
ck(f(('bool', 'uint144', 'int56'), D), T[bo, i, i])  # fixed case 40518
ck(f(('bool', 'uint144', 'int64'), D), T[bo, i, i])  # fixed case 40519
ck(f(('bool', 'uint144', 'int72'), D), T[bo, i, i])  # fixed case 40520
ck(f(('bool', 'uint144', 'int8'), D), T[bo, i, i])  # fixed case 40521
ck(f(('bool', 'uint144', 'int80'), D), T[bo, i, i])  # fixed case 40522
ck(f(('bool', 'uint144', 'int88'), D), T[bo, i, i])  # fixed case 40523
ck(f(('bool', 'uint144', 'int96'), D), T[bo, i, i])  # fixed case 40524
ck(f(('bool', 'uint144', 'string'), D), T[bo, i, s])  # fixed case 40525
ck(f(('bool', 'uint144', 'string[]'), D), T[bo, i, A])  # fixed case 40526
ck(f(('bool', 'uint144', 'uint104'), D), T[bo, i, i])  # fixed case 40527
ck(f(('bool', 'uint144', 'uint112'), D), T[bo, i, i])  # fixed case 40528
ck(f(('bool', 'uint144', 'uint120'), D), T[bo, i, i])  # fixed case 40529
ck(f(('bool', 'uint144', 'uint128'), D), T[bo, i, i])  # fixed case 40530
ck(f(('bool', 'uint144', 'uint136'), D), T[bo, i, i])  # fixed case 40531
ck(f(('bool', 'uint144', 'uint144'), D), T[bo, i, i])  # fixed case 40532
ck(f(('bool', 'uint144', 'uint152'), D), T[bo, i, i])  # fixed case 40533
ck(f(('bool', 'uint144', 'uint16'), D), T[bo, i, i])  # fixed case 40534
ck(f(('bool', 'uint144', 'uint160'), D), T[bo, i, i])  # fixed case 40535
ck(f(('bool', 'uint144', 'uint168'), D), T[bo, i, i])  # fixed case 40536
ck(f(('bool', 'uint144', 'uint176'), D), T[bo, i, i])  # fixed case 40537
ck(f(('bool', 'uint144', 'uint184'), D), T[bo, i, i])  # fixed case 40538
ck(f(('bool', 'uint144', 'uint192'), D), T[bo, i, i])  # fixed case 40539
ck(f(('bool', 'uint144', 'uint200'), D), T[bo, i, i])  # fixed case 40540
ck(f(('bool', 'uint144', 'uint208'), D), T[bo, i, i])  # fixed case 40541
ck(f(('bool', 'uint144', 'uint216'), D), T[bo, i, i])  # fixed case 40542
ck(f(('bool', 'uint144', 'uint224'), D), T[bo, i, i])  # fixed case 40543
ck(f(('bool', 'uint144', 'uint232'), D), T[bo, i, i])  # fixed case 40544
ck(f(('bool', 'uint144', 'uint24'), D), T[bo, i, i])  # fixed case 40545
ck(f(('bool', 'uint144', 'uint240'), D), T[bo, i, i])  # fixed case 40546
ck(f(('bool', 'uint144', 'uint248'), D), T[bo, i, i])  # fixed case 40547
ck(f(('bool', 'uint144', 'uint256'), D), T[bo, i, i])  # fixed case 40548
ck(f(('bool', 'uint144', 'uint32'), D), T[bo, i, i])  # fixed case 40549
ck(f(('bool', 'uint144', 'uint40'), D), T[bo, i, i])  # fixed case 40550
ck(f(('bool', 'uint144', 'uint48'), D), T[bo, i, i])  # fixed case 40551
ck(f(('bool', 'uint144', 'uint56'), D), T[bo, i, i])  # fixed case 40552
ck(f(('bool', 'uint144', 'uint64'), D), T[bo, i, i])  # fixed case 40553
ck(f(('bool', 'uint144', 'uint72'), D), T[bo, i, i])  # fixed case 40554
ck(f(('bool', 'uint144', 'uint8'), D), T[bo, i, i])  # fixed case 40555
ck(f(('bool', 'uint144', 'uint80'), D), T[bo, i, i])  # fixed case 40556
ck(f(('bool', 'uint144', 'uint88'), D), T[bo, i, i])  # fixed case 40557
ck(f(('bool', 'uint144', 'uint96'), D), T[bo, i, i])  # fixed case 40558
ck(f(('bool', 'uint144', '?'), D), T[bo, i, A])  # fixed case 40559
ck(f(('bool', 'uint152', 'address'), D), T[bo, i, HA])  # fixed case 40560
ck(f(('bool', 'uint152', 'address[]'), D), T[bo, i, A])  # fixed case 40561
ck(f(('bool', 'uint152', 'bool'), D), T[bo, i, bo])  # fixed case 40562
ck(f(('bool', 'uint152', 'bool[]'), D), T[bo, i, A])  # fixed case 40563
ck(f(('bool', 'uint152', 'bytes'), D), T[bo, i, by])  # fixed case 40564
ck(f(('bool', 'uint152', 'bytes1'), D), T[bo, i, by])  # fixed case 40565
ck(f(('bool', 'uint152', 'bytes10'), D), T[bo, i, by])  # fixed case 40566
ck(f(('bool', 'uint152', 'bytes11'), D), T[bo, i, by])  # fixed case 40567
ck(f(('bool', 'uint152', 'bytes12'), D), T[bo, i, by])  # fixed case 40568
ck(f(('bool', 'uint152', 'bytes13'), D), T[bo, i, by])  # fixed case 40569
ck(f(('bool', 'uint152', 'bytes14'), D), T[bo, i, by])  # fixed case 40570
ck(f(('bool', 'uint152', 'bytes15'), D), T[bo, i, by])  # fixed case 40571
ck(f(('bool', 'uint152', 'bytes16'), D), T[bo, i, by])  # fixed case 40572
ck(f(('bool', 'uint152', 'bytes17'), D), T[bo, i, by])  # fixed case 40573
ck(f(('bool', 'uint152', 'bytes18'), D), T[bo, i, by])  # fixed case 40574
ck(f(('bool', 'uint152', 'bytes19'), D), T[bo, i, by])  # fixed case 40575
ck(f(('bool', 'uint152', 'bytes2'), D), T[bo, i, by])  # fixed case 40576
ck(f(('bool', 'uint152', 'bytes20'), D), T[bo, i, by])  # fixed case 40577
ck(f(('bool', 'uint152', 'bytes21'), D), T[bo, i, by])  # fixed case 40578
ck(f(('bool', 'uint152', 'bytes22'), D), T[bo, i, by])  # fixed case 40579
ck(f(('bool', 'uint152', 'bytes23'), D), T[bo, i, by])  # fixed case 40580
ck(f(('bool', 'uint152', 'bytes24'), D), T[bo, i, by])  # fixed case 40581
ck(f(('bool', 'uint152', 'bytes25'), D), T[bo, i, by])  # fixed case 40582
ck(f(('bool', 'uint152', 'bytes26'), D), T[bo, i, by])  # fixed case 40583
ck(f(('bool', 'uint152', 'bytes27'), D), T[bo, i, by])  # fixed case 40584
ck(f(('bool', 'uint152', 'bytes28'), D), T[bo, i, by])  # fixed case 40585
ck(f(('bool', 'uint152', 'bytes29'), D), T[bo, i, by])  # fixed case 40586
ck(f(('bool', 'uint152', 'bytes3'), D), T[bo, i, by])  # fixed case 40587
ck(f(('bool', 'uint152', 'bytes30'), D), T[bo, i, by])  # fixed case 40588
ck(f(('bool', 'uint152', 'bytes31'), D), T[bo, i, by])  # fixed case 40589
ck(f(('bool', 'uint152', 'bytes32'), D), T[bo, i, by])  # fixed case 40590
ck(f(('bool', 'uint152', 'bytes4'), D), T[bo, i, by])  # fixed case 40591
ck(f(('bool', 'uint152', 'bytes5'), D), T[bo, i, by])  # fixed case 40592
ck(f(('bool', 'uint152', 'bytes6'), D), T[bo, i, by])  # fixed case 40593
ck(f(('bool', 'uint152', 'bytes7'), D), T[bo, i, by])  # fixed case 40594
ck(f(('bool', 'uint152', 'bytes8'), D), T[bo, i, by])  # fixed case 40595
ck(f(('bool', 'uint152', 'bytes9'), D), T[bo, i, by])  # fixed case 40596
ck(f(('bool', 'uint152', 'int104'), D), T[bo, i, i])  # fixed case 40597
ck(f(('bool', 'uint152', 'int112'), D), T[bo, i, i])  # fixed case 40598
ck(f(('bool', 'uint152', 'int120'), D), T[bo, i, i])  # fixed case 40599
ck(f(('bool', 'uint152', 'int128'), D), T[bo, i, i])  # fixed case 40600
ck(f(('bool', 'uint152', 'int136'), D), T[bo, i, i])  # fixed case 40601
ck(f(('bool', 'uint152', 'int144'), D), T[bo, i, i])  # fixed case 40602
ck(f(('bool', 'uint152', 'int152'), D), T[bo, i, i])  # fixed case 40603
ck(f(('bool', 'uint152', 'int16'), D), T[bo, i, i])  # fixed case 40604
ck(f(('bool', 'uint152', 'int160'), D), T[bo, i, i])  # fixed case 40605
ck(f(('bool', 'uint152', 'int168'), D), T[bo, i, i])  # fixed case 40606
ck(f(('bool', 'uint152', 'int176'), D), T[bo, i, i])  # fixed case 40607
ck(f(('bool', 'uint152', 'int184'), D), T[bo, i, i])  # fixed case 40608
ck(f(('bool', 'uint152', 'int192'), D), T[bo, i, i])  # fixed case 40609
ck(f(('bool', 'uint152', 'int200'), D), T[bo, i, i])  # fixed case 40610
ck(f(('bool', 'uint152', 'int208'), D), T[bo, i, i])  # fixed case 40611
ck(f(('bool', 'uint152', 'int216'), D), T[bo, i, i])  # fixed case 40612
ck(f(('bool', 'uint152', 'int224'), D), T[bo, i, i])  # fixed case 40613
ck(f(('bool', 'uint152', 'int232'), D), T[bo, i, i])  # fixed case 40614
ck(f(('bool', 'uint152', 'int24'), D), T[bo, i, i])  # fixed case 40615
ck(f(('bool', 'uint152', 'int240'), D), T[bo, i, i])  # fixed case 40616
ck(f(('bool', 'uint152', 'int248'), D), T[bo, i, i])  # fixed case 40617
ck(f(('bool', 'uint152', 'int256'), D), T[bo, i, i])  # fixed case 40618
ck(f(('bool', 'uint152', 'int32'), D), T[bo, i, i])  # fixed case 40619
ck(f(('bool', 'uint152', 'int40'), D), T[bo, i, i])  # fixed case 40620
ck(f(('bool', 'uint152', 'int48'), D), T[bo, i, i])  # fixed case 40621
ck(f(('bool', 'uint152', 'int56'), D), T[bo, i, i])  # fixed case 40622
ck(f(('bool', 'uint152', 'int64'), D), T[bo, i, i])  # fixed case 40623
ck(f(('bool', 'uint152', 'int72'), D), T[bo, i, i])  # fixed case 40624
ck(f(('bool', 'uint152', 'int8'), D), T[bo, i, i])  # fixed case 40625
ck(f(('bool', 'uint152', 'int80'), D), T[bo, i, i])  # fixed case 40626
ck(f(('bool', 'uint152', 'int88'), D), T[bo, i, i])  # fixed case 40627
ck(f(('bool', 'uint152', 'int96'), D), T[bo, i, i])  # fixed case 40628
ck(f(('bool', 'uint152', 'string'), D), T[bo, i, s])  # fixed case 40629
ck(f(('bool', 'uint152', 'string[]'), D), T[bo, i, A])  # fixed case 40630
ck(f(('bool', 'uint152', 'uint104'), D), T[bo, i, i])  # fixed case 40631
ck(f(('bool', 'uint152', 'uint112'), D), T[bo, i, i])  # fixed case 40632
ck(f(('bool', 'uint152', 'uint120'), D), T[bo, i, i])  # fixed case 40633
ck(f(('bool', 'uint152', 'uint128'), D), T[bo, i, i])  # fixed case 40634
ck(f(('bool', 'uint152', 'uint136'), D), T[bo, i, i])  # fixed case 40635
ck(f(('bool', 'uint152', 'uint144'), D), T[bo, i, i])  # fixed case 40636
ck(f(('bool', 'uint152', 'uint152'), D), T[bo, i, i])  # fixed case 40637
ck(f(('bool', 'uint152', 'uint16'), D), T[bo, i, i])  # fixed case 40638
ck(f(('bool', 'uint152', 'uint160'), D), T[bo, i, i])  # fixed case 40639
ck(f(('bool', 'uint152', 'uint168'), D), T[bo, i, i])  # fixed case 40640
ck(f(('bool', 'uint152', 'uint176'), D), T[bo, i, i])  # fixed case 40641
ck(f(('bool', 'uint152', 'uint184'), D), T[bo, i, i])  # fixed case 40642
ck(f(('bool', 'uint152', 'uint192'), D), T[bo, i, i])  # fixed case 40643
ck(f(('bool', 'uint152', 'uint200'), D), T[bo, i, i])  # fixed case 40644
ck(f(('bool', 'uint152', 'uint208'), D), T[bo, i, i])  # fixed case 40645
ck(f(('bool', 'uint152', 'uint216'), D), T[bo, i, i])  # fixed case 40646
ck(f(('bool', 'uint152', 'uint224'), D), T[bo, i, i])  # fixed case 40647
ck(f(('bool', 'uint152', 'uint232'), D), T[bo, i, i])  # fixed case 40648
ck(f(('bool', 'uint152', 'uint24'), D), T[bo, i, i])  # fixed case 40649
ck(f(('bool', 'uint152', 'uint240'), D), T[bo, i, i])  # fixed case 40650
ck(f(('bool', 'uint152', 'uint248'), D), T[bo, i, i])  # fixed case 40651
ck(f(('bool', 'uint152', 'uint256'), D), T[bo, i, i])  # fixed case 40652
ck(f(('bool', 'uint152', 'uint32'), D), T[bo, i, i])  # fixed case 40653
ck(f(('bool', 'uint152', 'uint40'), D), T[bo, i, i])  # fixed case 40654
ck(f(('bool', 'uint152', 'uint48'), D), T[bo, i, i])  # fixed case 40655
ck(f(('bool', 'uint152', 'uint56'), D), T[bo, i, i])  # fixed case 40656
ck(f(('bool', 'uint152', 'uint64'), D), T[bo, i, i])  # fixed case 40657
ck(f(('bool', 'uint152', 'uint72'), D), T[bo, i, i])  # fixed case 40658
ck(f(('bool', 'uint152', 'uint8'), D), T[bo, i, i])  # fixed case 40659
ck(f(('bool', 'uint152', 'uint80'), D), T[bo, i, i])  # fixed case 40660
ck(f(('bool', 'uint152', 'uint88'), D), T[bo, i, i])  # fixed case 40661
ck(f(('bool', 'uint152', 'uint96'), D), T[bo, i, i])  # fixed case 40662
ck(f(('bool', 'uint152', '?'), D), T[bo, i, A])  # fixed case 40663
ck(f(('bool', 'uint16', 'address'), D), T[bo, i, HA])  # fixed case 40664
ck(f(('bool', 'uint16', 'address[]'), D), T[bo, i, A])  # fixed case 40665
ck(f(('bool', 'uint16', 'bool'), D), T[bo, i, bo])  # fixed case 40666
ck(f(('bool', 'uint16', 'bool[]'), D), T[bo, i, A])  # fixed case 40667
ck(f(('bool', 'uint16', 'bytes'), D), T[bo, i, by])  # fixed case 40668
ck(f(('bool', 'uint16', 'bytes1'), D), T[bo, i, by])  # fixed case 40669
ck(f(('bool', 'uint16', 'bytes10'), D), T[bo, i, by])  # fixed case 40670
ck(f(('bool', 'uint16', 'bytes11'), D), T[bo, i, by])  # fixed case 40671
ck(f(('bool', 'uint16', 'bytes12'), D), T[bo, i, by])  # fixed case 40672
ck(f(('bool', 'uint16', 'bytes13'), D), T[bo, i, by])  # fixed case 40673
ck(f(('bool', 'uint16', 'bytes14'), D), T[bo, i, by])  # fixed case 40674
ck(f(('bool', 'uint16', 'bytes15'), D), T[bo, i, by])  # fixed case 40675
ck(f(('bool', 'uint16', 'bytes16'), D), T[bo, i, by])  # fixed case 40676
ck(f(('bool', 'uint16', 'bytes17'), D), T[bo, i, by])  # fixed case 40677
ck(f(('bool', 'uint16', 'bytes18'), D), T[bo, i, by])  # fixed case 40678
ck(f(('bool', 'uint16', 'bytes19'), D), T[bo, i, by])  # fixed case 40679
ck(f(('bool', 'uint16', 'bytes2'), D), T[bo, i, by])  # fixed case 40680
ck(f(('bool', 'uint16', 'bytes20'), D), T[bo, i, by])  # fixed case 40681
ck(f(('bool', 'uint16', 'bytes21'), D), T[bo, i, by])  # fixed case 40682
ck(f(('bool', 'uint16', 'bytes22'), D), T[bo, i, by])  # fixed case 40683
ck(f(('bool', 'uint16', 'bytes23'), D), T[bo, i, by])  # fixed case 40684
ck(f(('bool', 'uint16', 'bytes24'), D), T[bo, i, by])  # fixed case 40685
ck(f(('bool', 'uint16', 'bytes25'), D), T[bo, i, by])  # fixed case 40686
ck(f(('bool', 'uint16', 'bytes26'), D), T[bo, i, by])  # fixed case 40687
ck(f(('bool', 'uint16', 'bytes27'), D), T[bo, i, by])  # fixed case 40688
ck(f(('bool', 'uint16', 'bytes28'), D), T[bo, i, by])  # fixed case 40689
ck(f(('bool', 'uint16', 'bytes29'), D), T[bo, i, by])  # fixed case 40690
ck(f(('bool', 'uint16', 'bytes3'), D), T[bo, i, by])  # fixed case 40691
ck(f(('bool', 'uint16', 'bytes30'), D), T[bo, i, by])  # fixed case 40692
ck(f(('bool', 'uint16', 'bytes31'), D), T[bo, i, by])  # fixed case 40693
ck(f(('bool', 'uint16', 'bytes32'), D), T[bo, i, by])  # fixed case 40694
ck(f(('bool', 'uint16', 'bytes4'), D), T[bo, i, by])  # fixed case 40695
ck(f(('bool', 'uint16', 'bytes5'), D), T[bo, i, by])  # fixed case 40696
ck(f(('bool', 'uint16', 'bytes6'), D), T[bo, i, by])  # fixed case 40697
ck(f(('bool', 'uint16', 'bytes7'), D), T[bo, i, by])  # fixed case 40698
ck(f(('bool', 'uint16', 'bytes8'), D), T[bo, i, by])  # fixed case 40699
ck(f(('bool', 'uint16', 'bytes9'), D), T[bo, i, by])  # fixed case 40700
ck(f(('bool', 'uint16', 'int104'), D), T[bo, i, i])  # fixed case 40701
ck(f(('bool', 'uint16', 'int112'), D), T[bo, i, i])  # fixed case 40702
ck(f(('bool', 'uint16', 'int120'), D), T[bo, i, i])  # fixed case 40703
ck(f(('bool', 'uint16', 'int128'), D), T[bo, i, i])  # fixed case 40704
ck(f(('bool', 'uint16', 'int136'), D), T[bo, i, i])  # fixed case 40705
ck(f(('bool', 'uint16', 'int144'), D), T[bo, i, i])  # fixed case 40706
ck(f(('bool', 'uint16', 'int152'), D), T[bo, i, i])  # fixed case 40707
ck(f(('bool', 'uint16', 'int16'), D), T[bo, i, i])  # fixed case 40708
ck(f(('bool', 'uint16', 'int160'), D), T[bo, i, i])  # fixed case 40709
ck(f(('bool', 'uint16', 'int168'), D), T[bo, i, i])  # fixed case 40710
ck(f(('bool', 'uint16', 'int176'), D), T[bo, i, i])  # fixed case 40711
ck(f(('bool', 'uint16', 'int184'), D), T[bo, i, i])  # fixed case 40712
ck(f(('bool', 'uint16', 'int192'), D), T[bo, i, i])  # fixed case 40713
ck(f(('bool', 'uint16', 'int200'), D), T[bo, i, i])  # fixed case 40714
ck(f(('bool', 'uint16', 'int208'), D), T[bo, i, i])  # fixed case 40715
ck(f(('bool', 'uint16', 'int216'), D), T[bo, i, i])  # fixed case 40716
ck(f(('bool', 'uint16', 'int224'), D), T[bo, i, i])  # fixed case 40717
ck(f(('bool', 'uint16', 'int232'), D), T[bo, i, i])  # fixed case 40718
ck(f(('bool', 'uint16', 'int24'), D), T[bo, i, i])  # fixed case 40719
ck(f(('bool', 'uint16', 'int240'), D), T[bo, i, i])  # fixed case 40720
ck(f(('bool', 'uint16', 'int248'), D), T[bo, i, i])  # fixed case 40721
ck(f(('bool', 'uint16', 'int256'), D), T[bo, i, i])  # fixed case 40722
ck(f(('bool', 'uint16', 'int32'), D), T[bo, i, i])  # fixed case 40723
ck(f(('bool', 'uint16', 'int40'), D), T[bo, i, i])  # fixed case 40724
ck(f(('bool', 'uint16', 'int48'), D), T[bo, i, i])  # fixed case 40725
ck(f(('bool', 'uint16', 'int56'), D), T[bo, i, i])  # fixed case 40726
ck(f(('bool', 'uint16', 'int64'), D), T[bo, i, i])  # fixed case 40727
ck(f(('bool', 'uint16', 'int72'), D), T[bo, i, i])  # fixed case 40728
ck(f(('bool', 'uint16', 'int8'), D), T[bo, i, i])  # fixed case 40729
ck(f(('bool', 'uint16', 'int80'), D), T[bo, i, i])  # fixed case 40730
ck(f(('bool', 'uint16', 'int88'), D), T[bo, i, i])  # fixed case 40731
ck(f(('bool', 'uint16', 'int96'), D), T[bo, i, i])  # fixed case 40732
ck(f(('bool', 'uint16', 'string'), D), T[bo, i, s])  # fixed case 40733
ck(f(('bool', 'uint16', 'string[]'), D), T[bo, i, A])  # fixed case 40734
ck(f(('bool', 'uint16', 'uint104'), D), T[bo, i, i])  # fixed case 40735
ck(f(('bool', 'uint16', 'uint112'), D), T[bo, i, i])  # fixed case 40736
ck(f(('bool', 'uint16', 'uint120'), D), T[bo, i, i])  # fixed case 40737
ck(f(('bool', 'uint16', 'uint128'), D), T[bo, i, i])  # fixed case 40738
ck(f(('bool', 'uint16', 'uint136'), D), T[bo, i, i])  # fixed case 40739
ck(f(('bool', 'uint16', 'uint144'), D), T[bo, i, i])  # fixed case 40740
ck(f(('bool', 'uint16', 'uint152'), D), T[bo, i, i])  # fixed case 40741
ck(f(('bool', 'uint16', 'uint16'), D), T[bo, i, i])  # fixed case 40742
ck(f(('bool', 'uint16', 'uint160'), D), T[bo, i, i])  # fixed case 40743
ck(f(('bool', 'uint16', 'uint168'), D), T[bo, i, i])  # fixed case 40744
ck(f(('bool', 'uint16', 'uint176'), D), T[bo, i, i])  # fixed case 40745
ck(f(('bool', 'uint16', 'uint184'), D), T[bo, i, i])  # fixed case 40746
ck(f(('bool', 'uint16', 'uint192'), D), T[bo, i, i])  # fixed case 40747
ck(f(('bool', 'uint16', 'uint200'), D), T[bo, i, i])  # fixed case 40748
ck(f(('bool', 'uint16', 'uint208'), D), T[bo, i, i])  # fixed case 40749
ck(f(('bool', 'uint16', 'uint216'), D), T[bo, i, i])  # fixed case 40750
ck(f(('bool', 'uint16', 'uint224'), D), T[bo, i, i])  # fixed case 40751
ck(f(('bool', 'uint16', 'uint232'), D), T[bo, i, i])  # fixed case 40752
ck(f(('bool', 'uint16', 'uint24'), D), T[bo, i, i])  # fixed case 40753
ck(f(('bool', 'uint16', 'uint240'), D), T[bo, i, i])  # fixed case 40754
ck(f(('bool', 'uint16', 'uint248'), D), T[bo, i, i])  # fixed case 40755
ck(f(('bool', 'uint16', 'uint256'), D), T[bo, i, i])  # fixed case 40756
ck(f(('bool', 'uint16', 'uint32'), D), T[bo, i, i])  # fixed case 40757
ck(f(('bool', 'uint16', 'uint40'), D), T[bo, i, i])  # fixed case 40758
ck(f(('bool', 'uint16', 'uint48'), D), T[bo, i, i])  # fixed case 40759
ck(f(('bool', 'uint16', 'uint56'), D), T[bo, i, i])  # fixed case 40760
ck(f(('bool', 'uint16', 'uint64'), D), T[bo, i, i])  # fixed case 40761
ck(f(('bool', 'uint16', 'uint72'), D), T[bo, i, i])  # fixed case 40762
ck(f(('bool', 'uint16', 'uint8'), D), T[bo, i, i])  # fixed case 40763
ck(f(('bool', 'uint16', 'uint80'), D), T[bo, i, i])  # fixed case 40764
ck(f(('bool', 'uint16', 'uint88'), D), T[bo, i, i])  # fixed case 40765
ck(f(('bool', 'uint16', 'uint96'), D), T[bo, i, i])  # fixed case 40766
ck(f(('bool', 'uint16', '?'), D), T[bo, i, A])  # fixed case 40767
ck(f(('bool', 'uint160', 'address'), D), T[bo, i, HA])  # fixed case 40768
ck(f(('bool', 'uint160', 'address[]'), D), T[bo, i, A])  # fixed case 40769
ck(f(('bool', 'uint160', 'bool'), D), T[bo, i, bo])  # fixed case 40770
ck(f(('bool', 'uint160', 'bool[]'), D), T[bo, i, A])  # fixed case 40771
ck(f(('bool', 'uint160', 'bytes'), D), T[bo, i, by])  # fixed case 40772
ck(f(('bool', 'uint160', 'bytes1'), D), T[bo, i, by])  # fixed case 40773
ck(f(('bool', 'uint160', 'bytes10'), D), T[bo, i, by])  # fixed case 40774
ck(f(('bool', 'uint160', 'bytes11'), D), T[bo, i, by])  # fixed case 40775
ck(f(('bool', 'uint160', 'bytes12'), D), T[bo, i, by])  # fixed case 40776
ck(f(('bool', 'uint160', 'bytes13'), D), T[bo, i, by])  # fixed case 40777
ck(f(('bool', 'uint160', 'bytes14'), D), T[bo, i, by])  # fixed case 40778
ck(f(('bool', 'uint160', 'bytes15'), D), T[bo, i, by])  # fixed case 40779
ck(f(('bool', 'uint160', 'bytes16'), D), T[bo, i, by])  # fixed case 40780
ck(f(('bool', 'uint160', 'bytes17'), D), T[bo, i, by])  # fixed case 40781
ck(f(('bool', 'uint160', 'bytes18'), D), T[bo, i, by])  # fixed case 40782
ck(f(('bool', 'uint160', 'bytes19'), D), T[bo, i, by])  # fixed case 40783
ck(f(('bool', 'uint160', 'bytes2'), D), T[bo, i, by])  # fixed case 40784
ck(f(('bool', 'uint160', 'bytes20'), D), T[bo, i, by])  # fixed case 40785
ck(f(('bool', 'uint160', 'bytes21'), D), T[bo, i, by])  # fixed case 40786
ck(f(('bool', 'uint160', 'bytes22'), D), T[bo, i, by])  # fixed case 40787
ck(f(('bool', 'uint160', 'bytes23'), D), T[bo, i, by])  # fixed case 40788
ck(f(('bool', 'uint160', 'bytes24'), D), T[bo, i, by])  # fixed case 40789
ck(f(('bool', 'uint160', 'bytes25'), D), T[bo, i, by])  # fixed case 40790
ck(f(('bool', 'uint160', 'bytes26'), D), T[bo, i, by])  # fixed case 40791
ck(f(('bool', 'uint160', 'bytes27'), D), T[bo, i, by])  # fixed case 40792
ck(f(('bool', 'uint160', 'bytes28'), D), T[bo, i, by])  # fixed case 40793
ck(f(('bool', 'uint160', 'bytes29'), D), T[bo, i, by])  # fixed case 40794
ck(f(('bool', 'uint160', 'bytes3'), D), T[bo, i, by])  # fixed case 40795
ck(f(('bool', 'uint160', 'bytes30'), D), T[bo, i, by])  # fixed case 40796
ck(f(('bool', 'uint160', 'bytes31'), D), T[bo, i, by])  # fixed case 40797
ck(f(('bool', 'uint160', 'bytes32'), D), T[bo, i, by])  # fixed case 40798
ck(f(('bool', 'uint160', 'bytes4'), D), T[bo, i, by])  # fixed case 40799
ck(f(('bool', 'uint160', 'bytes5'), D), T[bo, i, by])  # fixed case 40800
ck(f(('bool', 'uint160', 'bytes6'), D), T[bo, i, by])  # fixed case 40801
ck(f(('bool', 'uint160', 'bytes7'), D), T[bo, i, by])  # fixed case 40802
ck(f(('bool', 'uint160', 'bytes8'), D), T[bo, i, by])  # fixed case 40803
ck(f(('bool', 'uint160', 'bytes9'), D), T[bo, i, by])  # fixed case 40804
ck(f(('bool', 'uint160', 'int104'), D), T[bo, i, i])  # fixed case 40805
ck(f(('bool', 'uint160', 'int112'), D), T[bo, i, i])  # fixed case 40806
ck(f(('bool', 'uint160', 'int120'), D), T[bo, i, i])  # fixed case 40807
ck(f(('bool', 'uint160', 'int128'), D), T[bo, i, i])  # fixed case 40808
ck(f(('bool', 'uint160', 'int136'), D), T[bo, i, i])  # fixed case 40809
ck(f(('bool', 'uint160', 'int144'), D), T[bo, i, i])  # fixed case 40810
ck(f(('bool', 'uint160', 'int152'), D), T[bo, i, i])  # fixed case 40811
ck(f(('bool', 'uint160', 'int16'), D), T[bo, i, i])  # fixed case 40812
ck(f(('bool', 'uint160', 'int160'), D), T[bo, i, i])  # fixed case 40813
ck(f(('bool', 'uint160', 'int168'), D), T[bo, i, i])  # fixed case 40814
ck(f(('bool', 'uint160', 'int176'), D), T[bo, i, i])  # fixed case 40815
ck(f(('bool', 'uint160', 'int184'), D), T[bo, i, i])  # fixed case 40816
ck(f(('bool', 'uint160', 'int192'), D), T[bo, i, i])  # fixed case 40817
ck(f(('bool', 'uint160', 'int200'), D), T[bo, i, i])  # fixed case 40818
ck(f(('bool', 'uint160', 'int208'), D), T[bo, i, i])  # fixed case 40819
ck(f(('bool', 'uint160', 'int216'), D), T[bo, i, i])  # fixed case 40820
ck(f(('bool', 'uint160', 'int224'), D), T[bo, i, i])  # fixed case 40821
ck(f(('bool', 'uint160', 'int232'), D), T[bo, i, i])  # fixed case 40822
ck(f(('bool', 'uint160', 'int24'), D), T[bo, i, i])  # fixed case 40823
ck(f(('bool', 'uint160', 'int240'), D), T[bo, i, i])  # fixed case 40824
ck(f(('bool', 'uint160', 'int248'), D), T[bo, i, i])  # fixed case 40825
ck(f(('bool', 'uint160', 'int256'), D), T[bo, i, i])  # fixed case 40826
ck(f(('bool', 'uint160', 'int32'), D), T[bo, i, i])  # fixed case 40827
ck(f(('bool', 'uint160', 'int40'), D), T[bo, i, i])  # fixed case 40828
ck(f(('bool', 'uint160', 'int48'), D), T[bo, i, i])  # fixed case 40829
ck(f(('bool', 'uint160', 'int56'), D), T[bo, i, i])  # fixed case 40830
ck(f(('bool', 'uint160', 'int64'), D), T[bo, i, i])  # fixed case 40831
ck(f(('bool', 'uint160', 'int72'), D), T[bo, i, i])  # fixed case 40832
ck(f(('bool', 'uint160', 'int8'), D), T[bo, i, i])  # fixed case 40833
ck(f(('bool', 'uint160', 'int80'), D), T[bo, i, i])  # fixed case 40834
ck(f(('bool', 'uint160', 'int88'), D), T[bo, i, i])  # fixed case 40835
ck(f(('bool', 'uint160', 'int96'), D), T[bo, i, i])  # fixed case 40836
ck(f(('bool', 'uint160', 'string'), D), T[bo, i, s])  # fixed case 40837
ck(f(('bool', 'uint160', 'string[]'), D), T[bo, i, A])  # fixed case 40838
ck(f(('bool', 'uint160', 'uint104'), D), T[bo, i, i])  # fixed case 40839
ck(f(('bool', 'uint160', 'uint112'), D), T[bo, i, i])  # fixed case 40840
ck(f(('bool', 'uint160', 'uint120'), D), T[bo, i, i])  # fixed case 40841
ck(f(('bool', 'uint160', 'uint128'), D), T[bo, i, i])  # fixed case 40842
ck(f(('bool', 'uint160', 'uint136'), D), T[bo, i, i])  # fixed case 40843
ck(f(('bool', 'uint160', 'uint144'), D), T[bo, i, i])  # fixed case 40844
ck(f(('bool', 'uint160', 'uint152'), D), T[bo, i, i])  # fixed case 40845
ck(f(('bool', 'uint160', 'uint16'), D), T[bo, i, i])  # fixed case 40846
ck(f(('bool', 'uint160', 'uint160'), D), T[bo, i, i])  # fixed case 40847
ck(f(('bool', 'uint160', 'uint168'), D), T[bo, i, i])  # fixed case 40848
ck(f(('bool', 'uint160', 'uint176'), D), T[bo, i, i])  # fixed case 40849
ck(f(('bool', 'uint160', 'uint184'), D), T[bo, i, i])  # fixed case 40850
ck(f(('bool', 'uint160', 'uint192'), D), T[bo, i, i])  # fixed case 40851
ck(f(('bool', 'uint160', 'uint200'), D), T[bo, i, i])  # fixed case 40852
ck(f(('bool', 'uint160', 'uint208'), D), T[bo, i, i])  # fixed case 40853
ck(f(('bool', 'uint160', 'uint216'), D), T[bo, i, i])  # fixed case 40854
ck(f(('bool', 'uint160', 'uint224'), D), T[bo, i, i])  # fixed case 40855
ck(f(('bool', 'uint160', 'uint232'), D), T[bo, i, i])  # fixed case 40856
ck(f(('bool', 'uint160', 'uint24'), D), T[bo, i, i])  # fixed case 40857
ck(f(('bool', 'uint160', 'uint240'), D), T[bo, i, i])  # fixed case 40858
ck(f(('bool', 'uint160', 'uint248'), D), T[bo, i, i])  # fixed case 40859
ck(f(('bool', 'uint160', 'uint256'), D), T[bo, i, i])  # fixed case 40860
ck(f(('bool', 'uint160', 'uint32'), D), T[bo, i, i])  # fixed case 40861
ck(f(('bool', 'uint160', 'uint40'), D), T[bo, i, i])  # fixed case 40862
ck(f(('bool', 'uint160', 'uint48'), D), T[bo, i, i])  # fixed case 40863
ck(f(('bool', 'uint160', 'uint56'), D), T[bo, i, i])  # fixed case 40864
ck(f(('bool', 'uint160', 'uint64'), D), T[bo, i, i])  # fixed case 40865
ck(f(('bool', 'uint160', 'uint72'), D), T[bo, i, i])  # fixed case 40866
ck(f(('bool', 'uint160', 'uint8'), D), T[bo, i, i])  # fixed case 40867
ck(f(('bool', 'uint160', 'uint80'), D), T[bo, i, i])  # fixed case 40868
ck(f(('bool', 'uint160', 'uint88'), D), T[bo, i, i])  # fixed case 40869
ck(f(('bool', 'uint160', 'uint96'), D), T[bo, i, i])  # fixed case 40870
ck(f(('bool', 'uint160', '?'), D), T[bo, i, A])  # fixed case 40871
ck(f(('bool', 'uint168', 'address'), D), T[bo, i, HA])  # fixed case 40872
ck(f(('bool', 'uint168', 'address[]'), D), T[bo, i, A])  # fixed case 40873
ck(f(('bool', 'uint168', 'bool'), D), T[bo, i, bo])  # fixed case 40874
ck(f(('bool', 'uint168', 'bool[]'), D), T[bo, i, A])  # fixed case 40875
ck(f(('bool', 'uint168', 'bytes'), D), T[bo, i, by])  # fixed case 40876
ck(f(('bool', 'uint168', 'bytes1'), D), T[bo, i, by])  # fixed case 40877
ck(f(('bool', 'uint168', 'bytes10'), D), T[bo, i, by])  # fixed case 40878
ck(f(('bool', 'uint168', 'bytes11'), D), T[bo, i, by])  # fixed case 40879
ck(f(('bool', 'uint168', 'bytes12'), D), T[bo, i, by])  # fixed case 40880
ck(f(('bool', 'uint168', 'bytes13'), D), T[bo, i, by])  # fixed case 40881
ck(f(('bool', 'uint168', 'bytes14'), D), T[bo, i, by])  # fixed case 40882
ck(f(('bool', 'uint168', 'bytes15'), D), T[bo, i, by])  # fixed case 40883
ck(f(('bool', 'uint168', 'bytes16'), D), T[bo, i, by])  # fixed case 40884
ck(f(('bool', 'uint168', 'bytes17'), D), T[bo, i, by])  # fixed case 40885
ck(f(('bool', 'uint168', 'bytes18'), D), T[bo, i, by])  # fixed case 40886
ck(f(('bool', 'uint168', 'bytes19'), D), T[bo, i, by])  # fixed case 40887
ck(f(('bool', 'uint168', 'bytes2'), D), T[bo, i, by])  # fixed case 40888
ck(f(('bool', 'uint168', 'bytes20'), D), T[bo, i, by])  # fixed case 40889
ck(f(('bool', 'uint168', 'bytes21'), D), T[bo, i, by])  # fixed case 40890
ck(f(('bool', 'uint168', 'bytes22'), D), T[bo, i, by])  # fixed case 40891
ck(f(('bool', 'uint168', 'bytes23'), D), T[bo, i, by])  # fixed case 40892
ck(f(('bool', 'uint168', 'bytes24'), D), T[bo, i, by])  # fixed case 40893
ck(f(('bool', 'uint168', 'bytes25'), D), T[bo, i, by])  # fixed case 40894
ck(f(('bool', 'uint168', 'bytes26'), D), T[bo, i, by])  # fixed case 40895
ck(f(('bool', 'uint168', 'bytes27'), D), T[bo, i, by])  # fixed case 40896
ck(f(('bool', 'uint168', 'bytes28'), D), T[bo, i, by])  # fixed case 40897
ck(f(('bool', 'uint168', 'bytes29'), D), T[bo, i, by])  # fixed case 40898
ck(f(('bool', 'uint168', 'bytes3'), D), T[bo, i, by])  # fixed case 40899
ck(f(('bool', 'uint168', 'bytes30'), D), T[bo, i, by])  # fixed case 40900
ck(f(('bool', 'uint168', 'bytes31'), D), T[bo, i, by])  # fixed case 40901
ck(f(('bool', 'uint168', 'bytes32'), D), T[bo, i, by])  # fixed case 40902
ck(f(('bool', 'uint168', 'bytes4'), D), T[bo, i, by])  # fixed case 40903
ck(f(('bool', 'uint168', 'bytes5'), D), T[bo, i, by])  # fixed case 40904
ck(f(('bool', 'uint168', 'bytes6'), D), T[bo, i, by])  # fixed case 40905
ck(f(('bool', 'uint168', 'bytes7'), D), T[bo, i, by])  # fixed case 40906
ck(f(('bool', 'uint168', 'bytes8'), D), T[bo, i, by])  # fixed case 40907
ck(f(('bool', 'uint168', 'bytes9'), D), T[bo, i, by])  # fixed case 40908
ck(f(('bool', 'uint168', 'int104'), D), T[bo, i, i])  # fixed case 40909
ck(f(('bool', 'uint168', 'int112'), D), T[bo, i, i])  # fixed case 40910
ck(f(('bool', 'uint168', 'int120'), D), T[bo, i, i])  # fixed case 40911
ck(f(('bool', 'uint168', 'int128'), D), T[bo, i, i])  # fixed case 40912
ck(f(('bool', 'uint168', 'int136'), D), T[bo, i, i])  # fixed case 40913
ck(f(('bool', 'uint168', 'int144'), D), T[bo, i, i])  # fixed case 40914
ck(f(('bool', 'uint168', 'int152'), D), T[bo, i, i])  # fixed case 40915
ck(f(('bool', 'uint168', 'int16'), D), T[bo, i, i])  # fixed case 40916
ck(f(('bool', 'uint168', 'int160'), D), T[bo, i, i])  # fixed case 40917
ck(f(('bool', 'uint168', 'int168'), D), T[bo, i, i])  # fixed case 40918
ck(f(('bool', 'uint168', 'int176'), D), T[bo, i, i])  # fixed case 40919
